/** 
 * \file ConfigFile.cpp
 * \brief This file defines the ConfigFile class which contains generic XML parsing functions
 * as well as functions specific to loading proxy details from the proxy config file.
 */

#include "ConfigFile.h"
#include "common.h"
#include "utils/utilFunctions.h"
#include "Proxy.h"
#include "StringTokenizer.h"

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <string>
#include <list>
#include <stdio.h>
#include <ctype.h>

using namespace std;

#include "ocd/OCD.h"


/**
  * Constructs a new ConfigFile object from the data given in the
  * specified file.
  * @param fileName the name of the configuration file.
  * @param rootTagName the name of the root tag of the config file.
  * TODO: Make it validating parser.  SO that looking for errors or
  * missed out fields in the config file is made easier.
  */
ConfigFile::ConfigFile (string fileName, string rootTagName) {

    this->configFileName= fileName;
    this->rootTagName = rootTagName;
   
	xmlInitParser ();
    docPtr = xmlReadFile(fileName.c_str(), NULL, 0);
    if (docPtr == NULL) {
        xmlErrorPtr tmpErr = xmlGetLastError();

        PRINT_DEBUG2 (
                DEBUG_LEVEL_FATAL, 
                "Error while parsing the XML configuration file %s.\n%s\nExiting.\n", 
                fileName.c_str(),
				tmpErr->message
            );
        
        PRINT_INFO2(
                INFO_LEVEL_FATAL_ERROR, 
                "Error while parsing the XML configuration file %s.\n%s\nExiting.\n", 
                fileName.c_str(),
				tmpErr->message
            );
        
        EXIT_ON_ERROR;
    }

    xmlNodePtr root;
    root = xmlDocGetRootElement(docPtr);
    
    if (root == NULL) {
        
        PRINT_INFO1(
                INFO_LEVEL_FATAL_ERROR, 
                "Empty XML configuration file %s. Exiting.\n", 
                fileName.c_str()
            );
        
        PRINT_DEBUG0(
                DEBUG_LEVEL_FATAL, 
                "Root element of the configuration file not found.\n"
            );
        
        EXIT_ON_ERROR;
    }
    
    if (xmlStrcmp(root->name, (const xmlChar *) rootTagName.c_str())) {
        
        PRINT_INFO1(
                INFO_LEVEL_FATAL_ERROR,
                "Invalid XML configuration file %s. Exiting.\n", 
                fileName.c_str()
            );
        
        PRINT_DEBUG2(
                DEBUG_LEVEL_FATAL, 
                "Root element of configuration file is \"%s\", expected \"%s\"\n", 
                root->name, rootTagName.c_str()
            );
        
        EXIT_ON_ERROR;
    }
}

/** Destructor */
ConfigFile::~ConfigFile() {
    xmlFreeDoc(docPtr);
    //xmlCleanupParser();
}   


xmlXPathObjectPtr ConfigFile::getNodeSet(xmlChar *xpath) {
    xmlXPathContextPtr context;
    xmlXPathObjectPtr result;
 
    context = xmlXPathNewContext(docPtr);
    result = xmlXPathEvalExpression(xpath, context);
    if(xmlXPathNodeSetIsEmpty(result->nodesetval)) {
        xmlXPathFreeObject(result);
        xmlXPathFreeContext(context);
        return NULL;
    }

    xmlXPathFreeContext(context);
    return result;
}

/**
  * Reads the value of an attribute of the element specified by the path
  * @param elementPath The XPath path to the element
  * @param attribName The name of the attribute we are interested in.
  * @param strBuf The buffer into which the value is to be placed.
  * @param requiredFlag If true and no value for the specified attribute
  * is found, an error message is displayed and the program is
  * terminated.
  */
void ConfigFile::readStringAttrib(
	char* elementPath, char *attribName,char* strBuf, bool requiredFlag) {

    xmlChar *xpath = xmlCharStrdup(elementPath);
    xmlXPathObjectPtr result = getNodeSet(xpath);
    xmlFree(xpath);

    if ( requiredFlag && result == NULL ) {
        
        PRINT_INFO2(
                INFO_LEVEL_FATAL_ERROR, 
                "Parameter %s required but not found in file: %s\n", 
                elementPath, configFileName.c_str()
            );
        
        EXIT_ON_ERROR;
    }

    if ( result == NULL )
        return;

    xmlNodeSetPtr nodeset = result->nodesetval;

    if ( nodeset->nodeNr >= 2 ) {
        xmlXPathFreeObject(result);
        PRINT_INFO2(
                INFO_LEVEL_FATAL_ERROR, 
        	    "%s should appear atmost once in configuration file %s\n",
                elementPath, configFileName.c_str()
            );
        EXIT_ON_ERROR;
    }

    xmlChar* resultstr = xmlGetProp(nodeset->nodeTab[0], (xmlChar*) attribName);
    if (requiredFlag && resultstr == NULL) {
		PRINT_INFO2(
			INFO_LEVEL_FATAL_ERROR,
			"%s attribute not found in element %s.\n",
			attribName, elementPath
			);
		EXIT_ON_ERROR;
    }

	if (resultstr == NULL) {

    	xmlXPathFreeObject(result);
		return;

	}
	
    strcpy(strBuf,(char*)resultstr);
    stripWS(strBuf);
    xmlFree(resultstr);
    xmlXPathFreeObject(result);
}

/**
  * Reads the value of an attribute of the element specified by the path
  * @param elementPath The XPath path to the element
  * @param attribName The name of the attribute we are interested in.
  * @param strBuf The buffer into which the value is to be placed.
  * @param defaultVal The default value to be returned
  * @return true if the parameter was successfully read. 
  * If the parameter was not found and that default value was used, returns false.
  */
bool ConfigFile::readStringAttrib(
	char* elementPath, char *attribName, char* strBuf,
	char *defaultVal) {

    xmlChar *xpath = xmlCharStrdup(elementPath);
    xmlXPathObjectPtr result = getNodeSet(xpath);
    xmlFree(xpath);

    if ( result == NULL ) {
        strcpy(strBuf, defaultVal);
    	return false;
   }
        
    xmlNodeSetPtr nodeset = result->nodesetval;

    if ( nodeset->nodeNr >= 2 ) {
        xmlXPathFreeObject(result);
        PRINT_INFO2(
                INFO_LEVEL_FATAL_ERROR, 
        	"%s should appear atmost once in configuration file %s\n",
                elementPath, configFileName.c_str()
            );
        EXIT_ON_ERROR;
    }

    xmlChar* resultstr = xmlGetProp(nodeset->nodeTab[0], (xmlChar *)attribName);
    if (resultstr == NULL) {
        strcpy(strBuf, defaultVal);
    	return false;
    }
    strcpy(strBuf,(char*)resultstr);
    stripWS(strBuf);
    xmlFree(resultstr);
    xmlXPathFreeObject(result);

    return true;
}


/**
  * Reads the string parameter specified by the given path.
  * Note that this function returns the text value enclosed between
  * the start and end tags of the element and NOT the value of any
  * attribute of the element.  For eg: <elementA>This text is returned</elementA>
  * @param path The XPath path to the parameter
  * @param strBuf The buffer into which the value is to be placed.
  * @param requiredFlag If true and no value for the specified parameter
  * is found, an error message is displayed and the program is
  * terminated.
  */
void ConfigFile::readStringParam(char* path, char* strBuf, bool requiredFlag) {

    xmlChar *xpath = xmlCharStrdup(path);
    xmlXPathObjectPtr result = getNodeSet(xpath);
    xmlFree(xpath);

    if ( requiredFlag && result == NULL ) {
        
        PRINT_INFO2(
                INFO_LEVEL_FATAL_ERROR, 
                "Parameter %s required but not found in file: %s\n", 
                path, configFileName.c_str()
            );
        
        PRINT_DEBUG2(
                DEBUG_LEVEL_FATAL, 
                "Parameter %s required but not found in file: %s\n", 
                path, configFileName.c_str()
                );
        
        EXIT_ON_ERROR;
    }

    if ( result == NULL ) {
        return;
	}

    xmlNodeSetPtr nodeset = result->nodesetval;

    if ( nodeset->nodeNr >= 2 ) {
        xmlXPathFreeObject(result);
        PRINT_INFO2(
                INFO_LEVEL_FATAL_ERROR, 
        	"%s should appear atmost once in configuration file %s\n",
                path, configFileName.c_str()
		);
        exit(-1);
    }

    xmlChar* resultstr = xmlNodeListGetString(docPtr, nodeset->nodeTab[0]->xmlChildrenNode, 1);
	if (requiredFlag && resultstr == NULL) {
		PRINT_INFO1 (INFO_LEVEL_FATAL_ERROR,
						"Error while reading param %s.\n",
						path
					);
		EXIT_ON_ERROR;
	}

    if ( resultstr == NULL ) {
    	xmlXPathFreeObject(result);
        return;
	}
    strcpy(strBuf,(char*)resultstr);
    stripWS(strBuf);
    xmlFree(resultstr);
    xmlXPathFreeObject(result);
}

/**
  * Reads the string parameter specified by the given path.  If the
  * specified value is not found, the default value is copied to the
  * buffer.
  * Note that this function returns the text value enclosed between
  * the start and end tags of the element and NOT the value of any
  * attribute of the element.  For eg: <elementA>This text is returned</elementA>
  * @param path The XPath path to the parameter
  * @param strBuf The buffer into which the value is to be placed.
  * @param defaultVal The default value to be used if parameter not
  * found.
  * @return true if the parameter was successfully read. 
  * If the parameter was not found and that default value was used, returns false.
  */
bool ConfigFile::readStringParam(char* path, char* strBuf, 
                                            char* defaultVal) {

    xmlChar *xpath = xmlCharStrdup(path);
    xmlXPathObjectPtr result = getNodeSet(xpath);
    xmlFree(xpath);

    if ( result == NULL ) {
        strcpy(strBuf, defaultVal);
        return false;
    }
    xmlNodeSetPtr nodeset = result->nodesetval;

    if ( nodeset->nodeNr >= 2 ) {
        xmlXPathFreeObject(result);
        PRINT_INFO2(
                INFO_LEVEL_FATAL_ERROR, 
        	    "%s should appear atmost once in configuration file %s\n",
                path, configFileName.c_str()
		);
        exit(-1);
    }

    xmlChar* resultstr = xmlNodeListGetString(docPtr, nodeset->nodeTab[0]->xmlChildrenNode, 1);
	if (resultstr == NULL) {
		strcpy(strBuf, defaultVal);
        return false;
	}
    strcpy(strBuf,(char*)resultstr);
    stripWS(strBuf);
    xmlFree(resultstr);
    xmlXPathFreeObject(result);
    
    return true;
}

/**
  * Reads the ushort parameter specified by the given path.
  * Note that this function returns the value enclosed between
  * the start and end tags of the element and NOT the value of any
  * attribute of the element.  
  * @param path The XPath path to the parameter
  * @param us The buffer into which the value is to be placed.
  * @param requiredFlag If true and no value for the specified parameter
  * is found, an error message is displayed and the program is
  * terminated.
  */
void ConfigFile::readUShortParam(char* path, unsigned short* us, bool requiredFlag) {
    char strBuf[200];
    readStringParam(path,strBuf,requiredFlag);
    *us = (unsigned short) atoi(strBuf);
}

/**
  * Reads the ushort attribute of the specificed element
  * @param elementPath The XPath to the element
  * @param attribName the attribute we are interested in.
  * @param us The buffer into which the value is to be placed.
  * @param requiredFlag If true and no value for the specified parameter
  * is found, an error message is displayed and the program is
  * terminated.
  */
void ConfigFile::readUShortAttrib(
	char* elementPath, char* attribName, unsigned short* us, bool requiredFlag) {
    char strBuf[200];
    readStringAttrib(elementPath, attribName, strBuf,requiredFlag);
    *us = (unsigned short) atoi(strBuf);
}

/**
  * Reads the ushort parameter specified by the given path.
  * Note that this function returns the value enclosed between
  * the start and end tags of the element and NOT the value of any
  * attribute of the element.  
  * @param path The XPath path to the parameter
  * @param us The buffer into which the value is to be placed.
  * @param defaultVal The default value to be used if parameter not * found.
  * @return true if the parameter was successfully read. 
  * If the parameter was not found and that default value was used, returns false.
  */
bool ConfigFile::readUShortParam(char* path, unsigned short* us, unsigned short defaultVal) {
    char strBuf[200];
    bool okFlag = readStringParam(path,strBuf,"");
    if (okFlag) {
        *us = (unsigned short) atoi(strBuf);
        return true;
    } else {
        *us = defaultVal;
        return false;
    }
}

/**
  * Reads the ushort attribute of the specified element.
  * @param elementPath The XPath to the element
  * @param attribName the attribute we are interested in.
  * @param us The buffer into which the value is to be placed.
  * @param defaultVal The default value to be used if parameter not * found.
  * @return true if the parameter was successfully read. 
  * If the parameter was not found and that default value was used, returns false.
  */
bool ConfigFile::readUShortAttrib
	(char* elementPath, char *attribName, unsigned short* us, unsigned short defaultVal) {
		
    char strBuf[200];
    bool okFlag = readStringAttrib(elementPath, attribName, strBuf,"");

    if (okFlag) {
        *us = (unsigned short) atoi(strBuf);
        return true;
    } else {
        *us = defaultVal;
        return false;
    }
}

/**
  * Reads the unsigned parameter specified by the given path.
  * @param path The XPath path to the parameter
  * @param val The buffer into which the value is to be placed.
  * @param defaultVal The default value to be used if parameter not * found.
  * @return true if the parameter was successfully read. 
  * If the parameter was not found and that default value was used, returns false.
  */
bool ConfigFile::readUIntParam(char* path, unsigned* val, unsigned defaultVal) {
    char strBuf[200];
    bool okFlag = readStringParam(path,strBuf,"");
    if (okFlag) {
        *val = atoi(strBuf);
        return true;
    } else {
        *val = defaultVal;
        return false;
    }
}

/**
  * Reads the string parameters specified by the given path.
  * Note that this function returns the value enclosed between
  * the start and end tags of the element and NOT the value of any
  * attribute of the element.  
  * @param path The XPath path to the parameter
  * @param num The number of string values found is put here
  * @return Array of string values satisfying the given path.
  */
char** ConfigFile::readStringParams(char* path, int* num) {
  
    xmlChar *xpath = xmlCharStrdup(path);
    xmlXPathObjectPtr result = getNodeSet(xpath);
    xmlFree(xpath);
  
    if ( result == NULL ) {
        *num=0;
        return NULL;
    }

    xmlNodeSetPtr nodeset = result->nodesetval;
    char** toret = (char**) malloc(nodeset->nodeNr * sizeof(char*));
    *num=nodeset->nodeNr;

    int i;
    for(i=0;i<nodeset->nodeNr;i++) {
          
        xmlChar* resultstr = 
            xmlNodeListGetString(docPtr, nodeset->nodeTab[i]->xmlChildrenNode, 1);
        toret[i] = strdup((char*)resultstr);
        stripWS(toret[i]);
        xmlFree(resultstr);
    
    }

    xmlXPathFreeObject (result);
    
    return toret;

}

/**
 * Loads the configuration parameters of the proxy into
 * the specified proxy object.
 */
void ConfigFile::loadProxyDetails(Proxy *proxy) {
       
    string path = XPATH_PROXY_CONFIG_FILE ;
    
    xmlChar *xpath = xmlCharStrdup(path.c_str());
    xmlXPathObjectPtr result = getNodeSet(xpath);
    xmlFree(xpath);
//TODO if tags are missing, make sure the proxy does not crash.

    if(result == NULL) {
        PRINT_INFO0(
                INFO_LEVEL_VERBOSE, 
                "No ProxyConfigFile tags found in the configuration file.\n"
            );
        return;
    }
    xmlNodeSetPtr nodeset = result->nodesetval;
 
    int count = nodeset->nodeNr;

   
    if ( count >= 2 ) {
        xmlXPathFreeObject(result);
        PRINT_INFO1 (
		        INFO_LEVEL_WARNING,
		        "%s should appear atmost once in configuration file\n",
		        path.c_str()
            );
        exit(-1);
    }

    xmlChar *attrib = NULL;
    xmlNodePtr currNode = nodeset->nodeTab[0];

    attrib = xmlGetProp(currNode, (xmlChar *)TAG_FAKE_ADDRESS_RANGE_START);
    if (attrib == NULL) {
        PRINT_INFO3(
                    INFO_LEVEL_FATAL_ERROR,
                    "Required attribute %s missing in %s element in file %s.\n",
                    TAG_FAKE_ADDRESS_RANGE_START,
                    XPATH_PROXY_CONFIG_FILE,
                    configFileName.c_str()
                );
        EXIT_ON_ERROR;              
    }

    string fakeAddrRangeStart ((char *)attrib);
    proxy->fakeAddrRangeStart = fakeAddrRangeStart;
    xmlFree(attrib);
    

    attrib = xmlGetProp(currNode, (xmlChar *)TAG_FAKE_ADDRESS_RANGE_MASK);
    if (attrib == NULL) {
        PRINT_INFO3(
                    INFO_LEVEL_FATAL_ERROR,
                    "Required attribute %s missing in %s element in file %s.\n",
                    TAG_FAKE_ADDRESS_RANGE_MASK,
                    XPATH_PROXY_CONFIG_FILE,
                    configFileName.c_str()
                );
        EXIT_ON_ERROR;              
    }
    string fakeAddrRangeMask((char *)attrib);
    proxy->fakeAddrRangeMask = fakeAddrRangeMask;
    xmlFree(attrib);
    
    attrib = xmlGetProp(currNode, (xmlChar *)TAG_STD_DNS_REQUEST_PORT);
    if (attrib == NULL) {
        PRINT_INFO3(
                    INFO_LEVEL_FATAL_ERROR,
                    "Required attribute %s missing in %s element in file %s.\n",
                    TAG_STD_DNS_REQUEST_PORT,
                    XPATH_PROXY_CONFIG_FILE,
                    configFileName.c_str()
                );
        EXIT_ON_ERROR;              
    }
    proxy->stdDNSRequestPort = atoi ((char *)attrib);
    xmlFree(attrib);
        
    attrib = xmlGetProp(currNode, (xmlChar *)TAG_LEGACY_DNS_REQUEST_PORT);
    if (attrib == NULL) {
        PRINT_INFO3(
                    INFO_LEVEL_FATAL_ERROR,
                    "Required attribute %s missing in %s element in file %s.\n",
                    TAG_LEGACY_DNS_REQUEST_PORT,
                    XPATH_PROXY_CONFIG_FILE,
                    configFileName.c_str()
                );
        EXIT_ON_ERROR;              
    }
    proxy->dnsRequestPort = atoi((char *)attrib);
    xmlFree(attrib);

    // START - client remote proxy attribs
    attrib = xmlGetProp(currNode, (xmlChar *)TAG_FUNCTION_AS_CLIENT_REMOTE_PROXY);
    if (attrib == NULL) {
        proxy->functionAsClientRemoteProxy = false;
    
    } else {
        if ( (strcasecmp ((char *) attrib, "yes") == 0) 
                || (strcasecmp ((char *)attrib, "true") == 0)) {
            
            proxy->functionAsClientRemoteProxy = true;
        
        } else {
        
            proxy->functionAsClientRemoteProxy = false;
        
        }
        xmlFree(attrib);
    }

    proxy->clientRemoteProxyDNSSuffixList.clear();
    attrib = xmlGetProp(currNode, (xmlChar *)TAG_CLIENT_REMOTE_PROXY_DNS_SUFFIX);
    if (attrib != NULL) {
        StringTokenizer st (string((char *) attrib), string(","));
        while (st.hasMoreTokens()) {
            proxy->clientRemoteProxyDNSSuffixList.push_back (st.nextToken());
        }
        xmlFree(attrib);
    }

    // STOP - client remote proxy attribs

    attrib = xmlGetProp(currNode, (xmlChar *)TAG_FUNCTION_AS_SERVER_PROXY);
    if (attrib == NULL) {
        proxy->functionAsServerProxy = false;
    
    } else {
        if ( (strcasecmp ((char *) attrib, "yes") == 0) 
                || (strcasecmp ((char *)attrib, "true") == 0)) {
            
            proxy->functionAsServerProxy = true;
        
        } else {
        
            proxy->functionAsServerProxy = false;
        
        }
        xmlFree(attrib);
    }

    attrib = xmlGetProp(currNode, (xmlChar *)TAG_SERVER_PROXY_CUSTOM_NAT);
    if (attrib == NULL) {
        proxy->mySoftwareNATFlag = false;
    
    } else {
        if ( (strcasecmp ((char *) attrib, "yes") == 0) 
                || (strcasecmp ((char *)attrib, "true") == 0)) {
            
            proxy->mySoftwareNATFlag = true;
        
        } else {
        
            proxy->mySoftwareNATFlag = false;
        
        }
        xmlFree(attrib);
    }


    proxy->middleBoxTypeName = "";
    attrib = xmlGetProp(currNode, (xmlChar *)TAG_FUNCTION_AS_MIDDLE_BOX);
    if (attrib == NULL) {
        proxy->functionAsMiddleBox = false;
        
    } else {
        if ( (strcasecmp ((char *) attrib, "yes") == 0) 
                || (strcasecmp ((char *)attrib, "true") == 0)) {
            
            proxy->functionAsMiddleBox = true;
            
            //Which middlebox to run ? - Bro? Gateway?
            
        } else {
        
            proxy->functionAsMiddleBox = false;
            
        }
        xmlFree(attrib);
    }

    proxy->middleBoxTypeName = "";
    attrib = xmlGetProp(currNode, (xmlChar *)TAG_MIDDLE_BOX_TYPE);
    if (attrib != NULL) {
        proxy->middleBoxTypeName = (char *) attrib;
        xmlFree(attrib);
    }
 
    proxy->middleBoxClassName = "";
    attrib = xmlGetProp(currNode, (xmlChar *)TAG_MIDDLE_BOX_CLASS);
    if (attrib != NULL) {
        proxy->middleBoxClassName = (char *) attrib;
        xmlFree(attrib);
    }
    
    attrib = xmlGetProp(currNode, (xmlChar *)TAG_MIDDLE_BOX_NAME);
    if (attrib != NULL) {
        strcpy (proxy->middleBoxName, (char *)attrib);
        xmlFree(attrib);
    }
   
//START - BAD HACK 
     attrib = xmlGetProp(currNode, (xmlChar *)"RONEnableFastStartup");
    if (attrib == NULL) {
        proxy->enableRONFastStartup = false;
    
    } else {
        if ( (strcasecmp ((char *) attrib, "yes") == 0) 
                || (strcasecmp ((char *)attrib, "true") == 0)) {
            
            proxy->enableRONFastStartup = true;
        
        } else {
        
            proxy->enableRONFastStartup = false;
        
        }
        xmlFree(attrib);
    }


// END - BAD HACK

    xmlXPathFreeObject (result);
}

/**
  * Loads the list of overlay networks specified in this configuration file.
  * An overlay network is identified by the tag OCD.
  * @param overlayList Pointer to the list to which the OCD
  * objects instantiated from the configuration file are to be added.
  */
void ConfigFile::initOCDList(list<OCD*> *overlayList) {
      
    string path = XPATH_OVERLAY_NETWORK;
    
    xmlChar *xpath = xmlCharStrdup(path.c_str());
    xmlXPathObjectPtr result = getNodeSet(xpath);
    xmlFree(xpath);
//TODO if tags are missing, make sure the proxy does not crash.

    if(result == NULL) {
        PRINT_INFO0(
                INFO_LEVEL_VERBOSE, 
                "No OCD tags found in the configuration file.\n"
            );
        return;
    }
    xmlNodeSetPtr nodeset = result->nodesetval;
 
    int count = nodeset->nodeNr;
    // go through all OCD elements and add them to the list
    for(int i=0; i < count; i++) {
        xmlChar *attrib;
        xmlNodePtr currNode = nodeset->nodeTab[i];
       
        attrib = xmlGetProp(currNode, (xmlChar *)TAG_OVERLAY_NETWORK_ID);
        if (attrib == NULL) {
            PRINT_INFO3(
                    INFO_LEVEL_FATAL_ERROR,
                    "Required attribute %s missing in %s element in file %s.\n",
                    TAG_OVERLAY_NETWORK_ID,
                    XPATH_OVERLAY_NETWORK,
                    configFileName.c_str()
                );
            EXIT_ON_ERROR;              
        }
        string id((char *)attrib);
        xmlFree(attrib);
     
        attrib = xmlGetProp(currNode, (xmlChar *)TAG_OVERLAY_NETWORK_TYPE);
        if (attrib == NULL) {
            PRINT_INFO3(
                    INFO_LEVEL_FATAL_ERROR,
                    "Required attribute %s missing in %s element in file %s.\n",
                    TAG_OVERLAY_NETWORK_TYPE,
                    XPATH_OVERLAY_NETWORK,
                    configFileName.c_str()
                );
            EXIT_ON_ERROR;              
        }
        string type((char *)attrib);
        xmlFree(attrib);

        string libFileName = "NOT_DEFINED";
#if defined(__APPLE__)
        // Get the name of the library which contains the OCD code.
        attrib = xmlGetProp(currNode, (xmlChar *)TAG_OVERLAY_NETWORK_LIB_FILE_APPLE);
        if (attrib == NULL) {
            if (type != OCD_TYPE_IPV4) {
                PRINT_INFO3(
                        INFO_LEVEL_FATAL_ERROR,
                        "Required attribute %s missing in %s element in file %s.\n",
                        TAG_OVERLAY_NETWORK_LIB_FILE_APPLE,
                        XPATH_OVERLAY_NETWORK,
                        configFileName.c_str()
                    );
                EXIT_ON_ERROR;              
            }
        
        } else {
        
            libFileName = (char *) attrib;
            xmlFree(attrib);
            
        }
        
#elif defined(_WIN32) 
        //Windows 
        // Get the name of the library which contains the OCD code.
        attrib = xmlGetProp(currNode, (xmlChar *)TAG_OVERLAY_NETWORK_LIB_FILE_WINDOWS);
        if (attrib == NULL) {
            if (type != OCD_TYPE_IPV4) {
                PRINT_INFO3(
                        INFO_LEVEL_FATAL_ERROR,
                        "Required attribute %s missing in %s element in file %s.\n",
                        TAG_OVERLAY_NETWORK_LIB_FILE_WINDOWS,
                        XPATH_OVERLAY_NETWORK,
                        configFileName.c_str()
                    );
                EXIT_ON_ERROR;              
            }
        
        } else {
        
            libFileName = (char *) attrib;
            xmlFree(attrib);
            
        }

#else
//Linux
        // Get the name of the library which contains the OCD code.
        attrib = xmlGetProp(currNode, (xmlChar *)TAG_OVERLAY_NETWORK_LIB_FILE_LINUX);
        if (attrib == NULL) {
            if (type != OCD_TYPE_IPV4) {
                PRINT_INFO3(
                        INFO_LEVEL_FATAL_ERROR,
                        "Required attribute %s missing in %s element in file %s.\n",
                        TAG_OVERLAY_NETWORK_LIB_FILE_LINUX,
                        XPATH_OVERLAY_NETWORK,
                        configFileName.c_str()
                    );
                EXIT_ON_ERROR;              
            }
        
        } else {
        
            libFileName = (char *) attrib;
            xmlFree(attrib);
            
        }
 
#endif

        attrib = xmlGetProp(currNode, (xmlChar *)TAG_OVERLAY_NETWORK_DESCRIPTION);
        if (attrib == NULL) {
            PRINT_INFO3(
                    INFO_LEVEL_FATAL_ERROR,
                    "Required attribute %s missing in %s element in file %s.\n",
                    TAG_OVERLAY_NETWORK_DESCRIPTION,
                    XPATH_OVERLAY_NETWORK,
                    configFileName.c_str()
                );
            EXIT_ON_ERROR;              
        }
        string desc((char *)attrib);
        xmlFree(attrib);
       
        
        attrib = xmlGetProp(currNode, (xmlChar *)TAG_OVERLAY_NETWORK_CONFIG_FILE);
        string fileName;
        if (attrib == NULL) {
            PRINT_INFO3(
                    INFO_LEVEL_WARNING,
                    "Attribute %s missing in %s element in file %s.\n",
                    TAG_OVERLAY_NETWORK_CONFIG_FILE,
                    XPATH_OVERLAY_NETWORK,
                    configFileName.c_str()
                );
         
            fileName = "NOT_SPECIFIED";
        } else {
            fileName = (char *)attrib;
            xmlFree(attrib);
        }
   
        
        attrib = xmlGetProp(currNode, (xmlChar *)TAG_OVERLAY_NETWORK_ACTIVE);
        if (attrib == NULL) {
            PRINT_INFO3(
                    INFO_LEVEL_FATAL_ERROR,
                    "Required attribute %s missing in %s element in file %s.\n",
                    TAG_OVERLAY_NETWORK_ACTIVE,
                    XPATH_OVERLAY_NETWORK,
                    configFileName.c_str()
                );
            EXIT_ON_ERROR;              
        }
        string active((char *)attrib);
        xmlFree(attrib);
   
        
        attrib = xmlGetProp(currNode, (xmlChar *)TAG_OVERLAY_NETWORK_HANDLE_LEGACY_URLS);
    	string handleLegacyURLs;
	    if (attrib != NULL) {        
		    handleLegacyURLs = (char *)attrib;
	    } else {
		    handleLegacyURLs = "no";
	    }
        xmlFree(attrib);

        bool handleLegacyURLsFlag;
        if (handleLegacyURLs == "yes") {
            handleLegacyURLsFlag = true;
        } else {
            handleLegacyURLsFlag = false;
        }
    
        
        attrib = xmlGetProp(currNode, (xmlChar *)TAG_OVERLAY_NETWORK_URL_SUFFIXES);
	    string urlSuffixes;
	    if (attrib != NULL) {
        	urlSuffixes =(char *)attrib;
	    } else {
		    urlSuffixes = "";
	    }

        xmlFree(attrib);


        // Based on the type field for active overlays, we should instantiate objects of the
        // appropriate OCD subclass.
        if (toLowerCase(active) == "yes" || toLowerCase(active) == "true") {
         
            OCD* overlayNet = 
                        proxy->getNewOCD(
                                        libFileName,
                                        type, id, desc, 
                                        fileName, urlSuffixes, 
                                        handleLegacyURLsFlag
                                    );

 	        if (overlayNet == NULL) {
                PRINT_INFO1 ( INFO_LEVEL_FATAL_ERROR,
			        "Unknown OCD type %s!\n",
                    type.c_str()        
             		);
		        exit(-1);
	        }
            overlayNet->setActive (true);
            overlayList->push_back(overlayNet);


            // go through all the OCD specific attributes of the current OCD element.
	        // We add all these name value pairs to a hash table.  Even the non-OCD specific
	        // attributes are added to the hash table.  The OCD module need not use all the values.
            for (xmlAttr *currAttr = currNode->properties; currAttr; currAttr = currAttr->next) {

                xmlChar* attrib = xmlGetProp(currNode, currAttr->name);

                PRINT_DEBUG3 (DEBUG_LEVEL_VERBOSE, "Adding property to overlay %s: %s --> %s.\n",
                        id.c_str(), (char *) currAttr->name, (char *) attrib
                    );
		        overlayNet->addConfigProperty ((char *)currAttr->name, (char *) attrib);
                xmlFree(attrib);
	        }
	    
            PRINT_DEBUG3(DEBUG_LEVEL_VERBOSE,
                "Reading config values for network with id %s type %s overlayaddr=%p\n",
                id.c_str(),
                type.c_str(),
                overlayNet
            );

        }
        
    }

  xmlXPathFreeObject (result);
}

/**
  * Reads the uint attribute of the specified element.
  * @param elementPath The XPath to the element
  * @param attribName the attribute we are interested in.
  * @param us The buffer into which the value is to be placed.
  * @param defaultVal The default value to be used if parameter not * found.
  * @return true if the parameter was successfully read. 
  * If the parameter was not found and that default value was used, returns false.
  */
bool ConfigFile::readUIntAttrib
	(char* elementPath, char *attribName, unsigned int* ui, unsigned int defaultVal) {
		
    char strBuf[200];
    bool okFlag = readStringAttrib(elementPath, attribName, strBuf,"");

    if (okFlag) {
        *ui = (unsigned int) atoi(strBuf);
        return true;
    } else {
        *ui = defaultVal;
        return false;
    }
}

/**
  * Update the "active" status of OCDs in the specified list.
  * This function parse the XML file all over again - does not use any 
  * previous instance of the parser
  * If the XML configuration contain any information about an OCD in ocdList,
  * that OCD is assumed to be inactive.
  * @return true if the updating was successful.
  */
bool ConfigFile::updateOCDActiveStatus (list<OCD*>* ocdList) {

    PRINT_DEBUG0 (  DEBUG_LEVEL_VERY_VERBOSE,
                    "Updating OCD Active Status ...\n"
            );
    
    // Set all OCDs status to inactive.
    list<OCD*>::iterator ocdItr = ocdList->begin();
    while (ocdItr != ocdList->end()) {
        OCD* currOCD = *ocdItr;
        currOCD->setActive (false);
        ocdItr++;
    }


    string path = XPATH_OVERLAY_NETWORK;
    
    xmlChar *xpath = xmlCharStrdup(path.c_str());
    xmlXPathObjectPtr result = getNodeSet(xpath);
    xmlFree(xpath);
//TODO if tags are missing, make sure the proxy does not crash.

    if(result == NULL) {
        PRINT_INFO0(
                INFO_LEVEL_VERBOSE, 
                "No OCD tags found in the configuration file.\n"
            );
        return false;
    }
    xmlNodeSetPtr nodeset = result->nodesetval;
 
    int count = nodeset->nodeNr;
    // go through all OCD elements and add them to the list
    for(int i=0; i < count; i++) {
        xmlChar *attrib;
        xmlNodePtr currNode = nodeset->nodeTab[i];
       
        attrib = xmlGetProp(currNode, (xmlChar *)TAG_OVERLAY_NETWORK_ID);
        if (attrib == NULL) {
            PRINT_INFO3(
                    INFO_LEVEL_FATAL_ERROR,
                    "Required attribute %s missing in %s element in file %s.\n",
                    TAG_OVERLAY_NETWORK_ID,
                    XPATH_OVERLAY_NETWORK,
                    configFileName.c_str()
                );
            return false;              
        }
        string id((char *)attrib);
        xmlFree(attrib);
     
        
        attrib = xmlGetProp(currNode, (xmlChar *)TAG_OVERLAY_NETWORK_ACTIVE);
        if (attrib == NULL) {
            PRINT_INFO3(
                    INFO_LEVEL_FATAL_ERROR,
                    "Required attribute %s missing in %s element in file %s.\n",
                    TAG_OVERLAY_NETWORK_ACTIVE,
                    XPATH_OVERLAY_NETWORK,
                    configFileName.c_str()
                );
            return false;
        }
        string active((char *)attrib);
        xmlFree(attrib);

        // Find the OCD with the id of the OCD element just read in and set its active status.
        ocdItr = ocdList->begin();
        while (ocdItr != ocdList->end()) {
            OCD* currOCD = *ocdItr;
            if (currOCD->id == id) {
                if (toLowerCase(active) == "yes" || toLowerCase(active) == "true") {
                    PRINT_DEBUG1 (DEBUG_LEVEL_VERY_VERBOSE, "Setting status of OCD %s to ACTIVE.\n", currOCD->id.c_str());
                    currOCD->setActive (true);
                } else {
                    PRINT_DEBUG1 (DEBUG_LEVEL_VERY_VERBOSE, "Setting status of OCD %s to INACTIVE.\n", currOCD->id.c_str());
                    currOCD->setActive (false);
                }
                break;
            }
            ocdItr++;
        }
    }

    xmlXPathFreeObject (result);

    return true;
}

