/**
 *  \file ConfigFile.h
 *  \brief Declaration of the various XML tags used in the config file.
 *  Also the ConfigFile parser class is defined here.
*/

#ifndef CONFIG_FILE_H
#define CONFIG_FILE_H 1

#ifndef _WIN32
#include <libxml2/libxml/xmlmemory.h>
#include <libxml2/libxml/parser.h>
#include <libxml2/libxml/xpath.h>
#else
extern "C" {
	#include <libxml/xmlmemory.h>
	#include <libxml/parser.h>
	#include <libxml/xpath.h>
};
#endif

#include <string>
#include <list>


#include "ocd/OCDTypes.h"

#include "OCALA_Export.h"

class OCD;
class Proxy;

/** This is the root tag for the proxy config file. */
#define TAG_PROXY_CONFIG_FILE           "ProxyConfigFile"
#define XPATH_PROXY_CONFIG_FILE           "//ProxyConfigFile"

// XML tags and XPath associated with the overlay networks
#define XPATH_OVERLAY_NETWORK                   "//ProxyConfigFile/OC-D-List/OC-D"
#define TAG_OVERLAY_NETWORK_ID                  "id"
#define TAG_OVERLAY_NETWORK_TYPE                "type"
#define TAG_OVERLAY_NETWORK_DESCRIPTION         "description"
#define TAG_OVERLAY_NETWORK_CONFIG_FILE         "configFile"
#define TAG_OVERLAY_NETWORK_ACTIVE              "active"
#define TAG_OVERLAY_NETWORK_LIB_FILE_LINUX      "libFileName_linux"
#define TAG_OVERLAY_NETWORK_LIB_FILE_APPLE      "libFileName_apple"
#define TAG_OVERLAY_NETWORK_LIB_FILE_WINDOWS    "libFileName_windows"
#define TAG_OVERLAY_NETWORK_URL_SUFFIXES        "urlSuffixes"
#define TAG_OVERLAY_NETWORK_HANDLE_LEGACY_URLS  "handleLegacyURLs"

#define TAG_FAKE_ADDRESS_RANGE_START  "fakeAddrRangeStart"
#define TAG_FAKE_ADDRESS_RANGE_MASK   "fakeAddrRangeMask"

#define TAG_STD_DNS_REQUEST_PORT    "stdDNSRequestPort"
#define TAG_LEGACY_DNS_REQUEST_PORT "legacyDNSRequestPort"

#define ATTRIB_MY_NAMES_CONFIG_FILE_NAME  "myNamesConfigFile"
#define ATTRIB_OUTBOUND_CONNECTION_CONFIG_FILE_NAME  "outboundConnectionConfigFile"

#define TAG_FUNCTION_AS_SERVER_PROXY "functionAsServerProxy"
#define TAG_SERVER_PROXY_CUSTOM_NAT  "serverProxyCustomNAT"
#define TAG_FUNCTION_AS_MIDDLE_BOX "functionAsMiddleBox"
#define TAG_MIDDLE_BOX_TYPE "middleBoxType"
#define TAG_MIDDLE_BOX_CLASS "middleBoxClassName"
#define TAG_MIDDLE_BOX_NAME     "middleBoxName"

#define TAG_FUNCTION_AS_CLIENT_REMOTE_PROXY 	"functionAsClientRemoteProxy"
#define TAG_CLIENT_REMOTE_PROXY_DNS_SUFFIX	"clientRemoteProxyDNSSuffix"

// XML tags associated with Legacy URL rules
#define XPATH_LEGACY_URL_RULE	"//LegacyURLRule"
#define TAG_LEGACY_URL_RULE_URL_PATTERN	"URLPattern"
#define TAG_LEGACY_URL_RULE_OVERLAY_ID	"OverlayId"
#define TAG_LEGACY_URL_RULE_SERVER_PROXY_ID "ServerProxyId"

// XPATHs related to the tunnelling device
#define XPATH_TUNNEL_DEVICE_PATH    "//Tunnel/DevicePath"
#define XPATH_TUNNEL_DEVICE_NAME    "//Tunnel/DeviceName"
#define XPATH_TUNNEL_WINTAP_NAME    "//Tunnel/WinTAPName"
#define XPATH_TUNNEL_IP_ADDRESS     "//Tunnel/IPAddress"
#define XPATH_TUNNEL_ROUTING_TABLE  "//Tunnel/RoutingTable"
#define XPATH_TUNNEL_MTU            "//Tunnel/MTU"
#define XPATH_TUNNEL_APPLE_CAPTURE_PORT  "//Tunnel/AppleCapturePort"
#define XPATH_TUNNEL_IFCONFIG_CMD   "//Tunnel/ifconfigCmd"
#define XPATH_TUNNEL_IPTABLES_CMD   "//Tunnel/iptablesCmd"
#define XPATH_TUNNEL_IP_CMD         "//Tunnel/ipCmd"
#define XPATH_TUNNEL_MODPROBE_CMD   "//Tunnel/modprobeCmd"

#define XPATH_SECURITY                          "//Security"
#define ATTRIB_SECURITY_CA_CERT_FILENAME        "CA_CertificateFile"
#define ATTRIB_SECURITY_LEVEL                   "SecurityLevel"

/**
 * This class encapsulates an XML configuration file.
 * It provides functions for easy retrieval of various parameters.
 */  
class ConfigFile {
protected:
    /** The name of the root tag of this configuration file. */
    std::string rootTagName;

    /** The root of the XML configuration file */
    xmlDocPtr docPtr;

    xmlXPathObjectPtr ConfigFile::getNodeSet(xmlChar *xpath);

public:
    /** The filename (including path info) of the config file. */
    std::string configFileName;

    /** Pointer to the proxy using this config file */
    Proxy* proxy;

    OCALA_LIB_EXPORT_IMPORT ConfigFile(std::string fileName, std::string rootTagName);    
    
    OCALA_LIB_EXPORT_IMPORT ~ConfigFile();

    OCALA_LIB_EXPORT_IMPORT void readStringParam(char* path, char* strBuf, bool requiredFlag);
    OCALA_LIB_EXPORT_IMPORT bool readStringParam(char* path, char* strBuf, char* defaultVal);
    OCALA_LIB_EXPORT_IMPORT void readStringAttrib(
													char* elementPath, 
													char *attribName, 
													char* strBuf, 
													bool requiredFlag
												);
    
    OCALA_LIB_EXPORT_IMPORT bool readStringAttrib(
													char* elementPath, 
													char *attribName, 
													char* strBuf,
													char *defaultVal);

    OCALA_LIB_EXPORT_IMPORT void readUShortParam(char* path, unsigned short* us, bool requiredFlag);
    OCALA_LIB_EXPORT_IMPORT bool readUShortParam(char* path, unsigned short* us, unsigned short defaultVal);

    OCALA_LIB_EXPORT_IMPORT void readUShortAttrib(
													char* elementPath, 
													char *attribName, 
													unsigned short *us, 
													bool requiredFlag
												);

    OCALA_LIB_EXPORT_IMPORT bool readUShortAttrib(
													char* elementPath, 
													char *attribName, 
													unsigned short *us, 
													unsigned short defaultVal
												);
    
    OCALA_LIB_EXPORT_IMPORT bool readUIntParam(char* path, unsigned* val, unsigned defaultVal);
    OCALA_LIB_EXPORT_IMPORT bool readUIntAttrib (char* elementPath, char *attribName, unsigned int* ui, unsigned int defaultVal) ;

    OCALA_LIB_EXPORT_IMPORT char** readStringParams(char* path, int* num);
    OCALA_LIB_EXPORT_IMPORT void initOCDList(std::list<OCD*> *overlayList);

    OCALA_LIB_EXPORT_IMPORT void loadProxyDetails(Proxy *proxy);
    OCALA_LIB_EXPORT_IMPORT bool updateOCDActiveStatus (std::list<OCD*>* ocdList);
};

#endif
