/** 
  * \file DNSPacket.cpp
  * \brief This class contains the defintions of a DNS packet.
  * DNSPacket is a subclass of IPPacket containing methods to manipulate
  * DNS requests and replies.
  */

#if defined( __APPLE__ )
#include <sys/types.h>
#endif

#ifndef _WIN32
	#include <netinet/udp.h>
	#include <sys/socket.h>
	#include <netinet/in.h>
	#include <arpa/inet.h>
#else
extern "C" {
	#include "netwrap.h"
}
#endif

#include <string.h>
#include "DNSPacket.h"
#include "common.h"

/** Constructs a DNS packet with just the requested URL 
 *  field initialized.  This constructor is mainly used during
 *  crash recovery.
 */
DNSPacket::DNSPacket(const char* reqURL) {
    this->reqURL = reqURL;
    this->ipHeader = NULL;
    this->pktContents = NULL;
    this->pktSize = -1;
    
    this->dnsHdr = NULL;
    this->dnsFlags = 0;
    this->dnsNumQuest = -1;

    this->vn = NULL;
}

/** Constructs a DNSPacket */
DNSPacket::DNSPacket(int pktSize, char *pktContents, IPHdr *ipHeader) {

    this->pktSize = pktSize;
    this->pktContents = pktContents;
    this->ipHeader = ipHeader;

    struct udphdr *tmpUDPHdr = (struct udphdr *) &pktContents[20];
    dnsHdr = (char *) tmpUDPHdr + 8;

    dnsFlags = ntohs(*((uint16_t *) dnsHdr + 1));
    dnsNumQuest = ntohs(*((uint16_t *) dnsHdr + 2));
    char *tmpDNSQuest = dnsHdr + 12;
  
    char tmpURLStr[MAX_URL_LEN];
    parseDNSQueryName(tmpDNSQuest, tmpURLStr);
    reqURL = tmpURLStr;

    vn = NULL;
}


/** Extracts the domain name in dns query and stores it in tmpURLStr
  * @return the domain name extracted, pointer to tmpURLStr
  */
char* DNSPacket::parseDNSQueryName(char *dnsQuest, char *tmpURLStr) {
  char  cnt = dnsQuest[0];
  uint  len = 0;
  
  while(cnt)
  {
    strncpy(tmpURLStr + len, &dnsQuest[len +1], cnt);
    len += cnt;
    tmpURLStr[len] = '.'; 
    len++;
    cnt = dnsQuest[len];
  }
  tmpURLStr[len -1] = 0x00;
  
  return tmpURLStr;
}


/**
  * This function constructs a DNS reply for this DNS packet.
  * @param ansIP IP address returned in the reply.  If this is null,  
  * a NAME ERROR DNS reply is returned.
  * @param buf The buffer into which the answer dns packet should be written
  */
void DNSPacket::getAnswer(char *buf, const IPAddress *ansIP) {
    uint16_t chksum;
    
    //TODO IPV4 address type check

    memcpy (buf, pktContents, pktSize);

    // change ip addresses
    memcpy (buf + 12, pktContents + 16, 4);
    memcpy (buf + 16, pktContents + 12, 4);

    // Exchange UDP port numbers
    memcpy (buf + 20, pktContents + 22, 2);
    memcpy (buf + 22, pktContents + 20, 2);
    memset (buf + 26, 0, 2);

    // Set the fields of the DNS answer
    buf[30] = buf[30] | 0x84;
    buf[34] = 0x00;
    memset (buf + 36, 0, 4);

    if (ansIP != NULL) {                
        // answer IP address
        // IP total length += 16
        buf[3] += 16;
        // update IP header checksum
        chksum = ntohs (((uint16_t *) buf)[5]);
        chksum -= 0x10;
        ((uint16_t *) buf)[5] = htons (chksum);

        buf[25] += 16;      // UDP length += 16

        buf[31] = 0x80;     // NO_ERROR
        buf[35] = 0x01;     // Answer RRs: 1

        buf[pktSize + 0] = 0xc0;
        buf[pktSize + 1] = 0x0c;
        buf[pktSize + 2] = 0x00;
        buf[pktSize + 3] = 0x01;
        buf[pktSize + 4] = 0x00;
        buf[pktSize + 5] = 0x01;
        buf[pktSize + 6] = 0x00;    // TTL
        buf[pktSize + 7] = 0x00;    // TTL
        buf[pktSize + 8] = 0x00;    // TTL
        buf[pktSize + 9] = 0x00;    // TTL
        buf[pktSize + 10] = 0x00;   // data length
        buf[pktSize + 11] = 0x04;   // data length
    
        struct in_addr temp_addr;
#ifndef _WIN32
        inet_aton(ansIP->c_str(),  &temp_addr);
#else
        inet_pton(AF_INET, ansIP->c_str(), &temp_addr);
#endif

        // At last, put in the fake address, that represents the DNS answer
#ifndef CCURED
        memcpy (buf + pktSize + 12, &temp_addr.s_addr,
              sizeof (unsigned long));
#else
        *(in_addr_t *) (buf + pktSize + 12) = temp_addr.s_addr;
#endif
    } else {
      // return "DNS_NAME_ERROR(No such name)"
      buf[31] = 0x83;       // NAME_ERROR
      buf[35] = 0x00;       // Answer RRs: 0
    }

}

/**
  * Constructs a DNS request packet for the specified requestedURL in buf.
  * @return The length of the DNS packet formed.
  */
int DNSPacket::makeDNSQueryPacket (char *buf, const char *requestedURL, uint16_t transactionId) {
  int i, char_cnt;
  
  //((uint16_t *)buf)[0] = transactionId; //set transaction ID
  ((uint16_t *)buf)[0] = htons (transactionId); //set transaction ID

  //printf ("\n\ninMakeDNSQueryPacket tid = %x \n\n", transactionId);

  buf[2] = 0x01; buf[3] = 0x00; //Flags: 0x0100 (Standard query)
  buf[4] = 0x00; buf[5] = 0x01; //Questions: 1
  memset(buf + 6, 0, 6); //Answer RRs, Authority RRs, Additional RRs: 0
  //format Name
  char_cnt = 0;
  for (i = 0; i < strlen(requestedURL); i++) {
    if (requestedURL[i] == '.') {
      buf[12 + i - char_cnt] = char_cnt;
      memcpy(buf + 13 + i - char_cnt, requestedURL + i - char_cnt, char_cnt);
      char_cnt = 0;
    } else if (i == strlen(requestedURL) - 1) {
      buf[12 + i - char_cnt] = char_cnt + 1;
      memcpy(buf + 13 + i - char_cnt, requestedURL + i - char_cnt, char_cnt + 1);
    } else {
      char_cnt++;
    }
  }
  buf[i + 13] = 0x00;
  buf[i + 14] = 0x00; buf[i + 15] = 0x01; // Type: Host address
  buf[i + 16] = 0x00; buf[i + 17] = 0x01; // Class: inet
  return i+18;

}

/**
  * Strip off the last section of the DNS preference string.
  * The stripped of section is returned.
  * @param dnsPrefStr The last section of this parameter is stripped off.
  * This parameter is actually affected.
  * @return the last section of the specified dns Pref Str. 
  * The return buffer should be deallocated after use.
  */
char *DNSPacket::stripDNSPrefLastSection (char *dnsPrefStr) {
    
    char leftDelim = '_';
   
    int len = strlen (dnsPrefStr);
    

    for (int i = len - 1; i >= 0; i--) {
        if (dnsPrefStr[i] == leftDelim) {
            char *lastStr = strdup (dnsPrefStr + i + 1);
            dnsPrefStr[i] = '\0'; //strip the last section
            return lastStr;
        }
    }
    char *retVal = strdup (dnsPrefStr);
    dnsPrefStr [0] = '\0';

    return retVal;
}

void DNSPacket::extractActualName (char *buf, const char *dnsPrefSection) {

    int len = strlen (dnsPrefSection);

    int dotCount = 0;
    int rightPos = len - 1;
    
    for (int i = len - 1; i >=0 ; i--) {
        if (dnsPrefSection[i] == '.') {
            dotCount++;
        }

        if (dotCount == 2) {
            rightPos = i;
            break;
        }
    }

    memcpy (buf, dnsPrefSection, rightPos);
    buf[rightPos] = '\0';
    
}

/** 
  * Extracts the first section of the given dnsPRefString into buf 
  */
void DNSPacket::extractFirstSection (char *buf, const char *dnsPrefSection) {

    strcpy (buf, dnsPrefSection);
    
    for (int i = 0; i < strlen (buf) ; i++) {
    
        if ( buf[i] == '_') {
            buf[i] = '\0';
            return;
        }
    }
}

/** Retrieves only the preference section of the specified dnsPrefStr.
  */
void DNSPacket::getPrefSection (const char *dnsPrefStr, char *prefSection) {

    //gets the part of the string between the secondlast and last dots.
    int len = strlen (dnsPrefStr);

    int numDots = 0;
    
    int rightPos = len - 1;
    int leftPos = 0;
    
    for (int i = len - 1; i >= 0; i--) {
    
        if (dnsPrefStr[i] == '.') {
            numDots ++;
            if (numDots == 1) {
                
                rightPos = i - 1;
            
            } else if (numDots == 2) {
                
                leftPos = i + 1;
                break;
            }
        }
        
    }
    int prefLen = rightPos - leftPos + 1;
    memcpy (prefSection, dnsPrefStr + leftPos , prefLen);
    prefSection [prefLen] = '\0';
}

/**
  * Returns true if the preference prefToCheck is present in the
  * preferences associated with dnsPrefStr.
  */
bool DNSPacket::checkPreference (const char *dnsPrefStr, char prefToCheck) {

    // Get only the preferences part of dnsPrefStr
    char onlyPrefStr [MAX_PREF_STRING_LEN];

    DNSPacket::getPrefSection (dnsPrefStr, onlyPrefStr);

    char *tmpPtr = strchr (onlyPrefStr, prefToCheck);
    
    if (tmpPtr == NULL) {
        
        return false;
    
    } else {
    
        return true;
    
    }
    
}
