/**
 * \file FDCallback.cpp
 * \brief This file defines a file descriptor callback.
 * An FDCallback encapsulates the callback to be invoked when there
 * is activity (for eg: packet arrival) on a file descriptor.
*/
#include <string.h>
#include <stdio.h>
#include "debug.h"
#include "FDCallback.h"
#include "TFunctor.h"

/**
  * Constructs a callback with the specified file descriptor and function pointer.
  * @param fd The file descriptor on which callback occurs.
  * @param ordinaryFnPtr Pointer to the function to be called.
  * This should be an ordinary function and not some member function of an object.
  */
FDCallback::FDCallback(int fd, int (*ordinaryFnPtr)(char *buf)) {
    this->fd = fd;
    this->functor = (TFunctor*) new OrdinaryFunctor(ordinaryFnPtr);
    this->disabledFlag = false;
    this->cbkFreeFlag = true; // we created the TFunctor, we free it ourselves.
}

/**
  * Constructs a callback with the specific file descriptor.
  * The function to be called in this callback is a member of some object. 
  * So it is passed via functors.
  * @param fd The file descriptor on which callback occurs.
  * @param functor The functor encapsulating the member function to be called.
  * @param cbkFreeFlag If true, the TFunctor functor will be freed when the FDCallback is deleted.
  */
FDCallback::FDCallback(int fd, TFunctor* functor, bool cbkFreeFlag) {
    this->fd = fd;
    this->functor = functor;
    this->disabledFlag = false;
    this->cbkFreeFlag = cbkFreeFlag;
}

/**
  * This is just a wrapper function to call the function pointed
  * to by fptr.
  */
void FDCallback::call() {
    if (!disabledFlag) {
        functor->callIntArg(this->fd);
    } else {
        PRINT_DEBUG0 (DEBUG_LEVEL_VERBOSE, "Executing disabled FD callback.\n");
    }
}

/**
  * Destructor 
  */
FDCallback::~FDCallback () {
    if (functor != NULL && cbkFreeFlag) {
        delete (functor);
    }
}
