/** 
  * \file FakeAddressManager.cpp
  * \brief Definitions for FakeAddressManager class.
  * This class manages the pool of fake IP addresses available to the
  * proxy.  It has methods for requesting and returning fake addresses
  * from and to the pool of available addresses.
  */

#include "FakeAddressManager.h"
#include "ConfigFile.h"
#include "common.h"

#ifdef _WIN32
extern "C" {
	#include "netwrap.h"
}
#else
	#include <netinet/in.h>
	#include <sys/socket.h>
	#include <arpa/inet.h>
#endif

#include <string>
#include <math.h>

#include <bitset>

//Free pool cannot be managed as a list as there are too many
//items.

/**
  * Allocates a fake IP address as a hash of the specified URL.
  * @return NULL if fake IP address to which it hashes has already been taken.
  */
IPAddress* FakeAddressManager::getNewFakeIP(string url) {

    int n = 0;
    char tmpName [MAX_URL_LEN];
    strcpy (tmpName, url.c_str());

    for (int i = 0; i < strlen (tmpName); i++) {
        n += tmpName[i];
    }

    n = n % numIPs;

    if (!freeIPSet.test (n)) {
        freeIPSet.set (n);
        struct in_addr tempAddr;
        tempAddr.s_addr = htonl(startFakeIP + n);
        IPAddress * newFakeIP = new IPAddress (inet_ntoa (tempAddr));
        return newFakeIP;
    }
    return NULL;
}

/**
  * Checks if the specified address is a fake IP address that had been
  * allocated by this FakeIPAddress manager.
  * @return true if yes.
  */
bool FakeAddressManager::isFakeIP(string testIPAddrStr) {

    struct in_addr testIPAddr;
    if (inet_aton (testIPAddrStr.c_str(), &testIPAddr) == 0) {
        
        PRINT_INFO1 (    
                    INFO_LEVEL_WARNING,
                    "IP address %s is invalid.\n",
                    testIPAddrStr.c_str()
                );
        return false;

    }
    
    // Now check if this address is in the fake range and was allocated
    unsigned long testIPAddrLong = ntohl(testIPAddr.s_addr); 
    if (testIPAddrLong >= startFakeIP && testIPAddrLong < (startFakeIP + numIPs)) {
        // falls in range of fake IP addresses.
        // Now check if it was allocated.
        if (freeIPSet.test (testIPAddrLong - startFakeIP)) {
            // Yes, it was allocated.
            return true;
        } else {
            PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
                            "IP address %s in fake IP address range, but not allocated.\n",
                            testIPAddrStr.c_str()
                        );
            return false;
        }
    } else {
        PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
                        "IP address %s not in fake IP address range.\n",
                        testIPAddrStr.c_str()
                    );
        return false;
    }
}


/** 
  * Returns a new fake IP address from its free pool.
  * TODO:  Right now, free pool management is very naive, assuming that
  * there are a very large number of fake IP addresses available.
  * @return free fake IPAddress if available, NULL if not available
  */
IPAddress* FakeAddressManager::getNewFakeIP() {

    for (unsigned long count = 0; count < numIPs; count++) {
    
        if (nextFreeIP == numIPs) {
            //cycle back to 0 once end of list is reached.
            nextFreeIP = 0;
        }
        
        if ( ! freeIPSet.test(nextFreeIP)) {
            //we have found a free IP
            freeIPSet.set(nextFreeIP);
            
            struct in_addr tempAddr;
            tempAddr.s_addr = htonl(startFakeIP + nextFreeIP);
            IPAddress* newFakeIP = new IPAddress(inet_ntoa(tempAddr));
            
            nextFreeIP ++;
   
            return newFakeIP;
            
        } 
        
        nextFreeIP++;
    }
    //no free IP found

    return NULL;
}

/** 
  * Allocates a random IP address from the pool of available addresses.
  * @return free fake IPAddress if available, NULL if a free random address
  * could not be obtained even after a large number of tries.
  */
IPAddress* FakeAddressManager::getNewRandomFakeIP() {

    /** How many times to try to obtain a random IP address */
    int maxAttempts = 100;

    for (int numAttempts = 0; numAttempts < maxAttempts; numAttempts ++) {

        //Select one of the IPs between 0 and numIPs -1 randomly
	float tmpRandVal = 0;
	
	tmpRandVal = rand();
        
	int randIndex =  (int) ((1.0 * tmpRandVal)/RAND_MAX * numIPs - 1);
      
        if ( ! freeIPSet.test(randIndex)) {
            //This is a free IP
            freeIPSet.set(randIndex);
            
            struct in_addr tempAddr;
            tempAddr.s_addr = htonl(startFakeIP + randIndex);
            IPAddress* newFakeIP = new IPAddress(inet_ntoa(tempAddr));
            return newFakeIP;
        } 
                
    }

    PRINT_DEBUG1 (
            DEBUG_LEVEL_MINIMAL,
            "Unable to find free random fake IP address even after %d tries.\n",
            maxAttempts
        );

    return NULL;
}


/**
 * This function marks the specified ipAddr as already allocated, i.e.
 * it won't be allocated again till it is released.
 */
bool FakeAddressManager::allocateFakeIP(const IPAddress* ipAddr) {

    struct in_addr tempAddr;
    inet_aton(ipAddr->c_str(), &tempAddr);
    
    unsigned long ipIndex = ntohl(tempAddr.s_addr) - startFakeIP;

    if (freeIPSet.test (ipIndex)) {
        return false;
    }
    
    freeIPSet.set(ipIndex);
    return true;        
}

/**
  * Releases the specified ip address to the free fake ip list.
  * @param ip Pointer to the IPAddress to be released.
  */
void FakeAddressManager::releaseFakeIP(const IPAddress *ip) {
    
    struct in_addr tempAddr;
    tempAddr.s_addr = htonl(inet_addr(ip->c_str()));
    unsigned long resetIndex = tempAddr.s_addr - startFakeIP;

    freeIPSet.reset(resetIndex);
    
}

/** 
  * Constructor
  */
FakeAddressManager::FakeAddressManager(string fakeAddrRangeStart_str, string fakeAddrRangeMask_str) {


    strncpy(fakeAddrRangeStart, fakeAddrRangeStart_str.c_str(), MAX_IP_ADDR_LEN);
    strncpy(fakeAddrRangeMask, fakeAddrRangeMask_str.c_str(), MAX_IP_ADDR_LEN);
    
    startFakeIP = ntohl(inet_addr(fakeAddrRangeStart));
    unsigned long fakeMask = ntohl(inet_addr(fakeAddrRangeMask));
#ifndef _WIN32
    numIPs = (unsigned long) exp2(NUM_BITS_IN_IP_ADDRESS) - fakeMask;
#else
	numIPs = (unsigned long) pow(2, NUM_BITS_IN_IP_ADDRESS) - fakeMask;
#endif
    PRINT_DEBUG3 (DEBUG_LEVEL_VERBOSE, "fakeAddrRange->start=%s fakeAddrRange->mask=%s numFakes=%d\n",
                    fakeAddrRangeStart, fakeAddrRangeMask,
                    numIPs);
    
    freeIPSet.reset();
    nextFreeIP = 0;
}

/**
  * Checks it the requeted fake IP address is available.
  * If yes, allocates it and return true.
  * If no, returns false.
  */
bool FakeAddressManager::checkAllocateFakeIP (struct in_addr requestedAddr) {
    unsigned long ipIndex = ntohl(requestedAddr.s_addr) - startFakeIP;
    
    if ( ! freeIPSet.test(ipIndex)) {
        //This is a free IP
        freeIPSet.set(ipIndex);
            
        return true;
        
    } else {

        return false;

    }
}

