/**
 * \file FakeAddressNegotiator.cpp
 * \brief This file contains the definition of the FakeAddressNegotiator.
 * This class is used to ensure that the same fake IP address is chosen at the 
 * sender and receiver.
 */

#include "FakeAddressNegotiator.h"

#ifndef _WIN32
#include <netinet/in.h>
#include <arpa/inet.h>

#endif

#include "Proxy.h"
#include "VirtualNode.h"

/**
  * @param vn The virtual node which is negotiating the address.
  * @param proxy Pointer to the generic proxy.
  */
FakeAddressNegotiator::FakeAddressNegotiator (
                    VirtualNode *vn,
                    Proxy* proxy) {
    
    this->proxy = proxy;
    this->vn = vn;
    proxy->getTunnelIP (&(this->requestedAddr));

    this->state = FAKE_ADDRESS_NEGOTIATION_IN_PROGRESS;
}

/**
  * Initiate the fake address negotiation 
  */
void FakeAddressNegotiator::initNewConnection () {
    

    FakeAddressNegotiationBlock* blk = new FakeAddressNegotiationBlock; 
    // TODO Remember to free this.
    
    blk->mainHeader.version = PROXY_VERSION;
    blk->mainHeader.blockType = PROXY_PACKET_BLOCK_FAKE_ADDRESS_NEGOTIATION;
    blk->mainHeader.nextBlock = sizeof (FakeAddressNegotiationBlock);
    blk->msgCode = FAKE_ADDRESS_NEGOTIATION_REQUEST;
    
    blk->requestedAddr = this->requestedAddr;
    
    // Send this block over the specific overlay associated with the VN.
    // Remember that the block may not be sent immediately.  
    //TODO FIX this->vn->sendBlock ( (ProxyPacketBlock *) blk);
}

/**
  * Received a fake address negotiation block 
  */
void FakeAddressNegotiator::receivedBlock (FakeAddressNegotiationBlock *blk) {
    
    switch (blk->msgCode) {
        
        case FAKE_ADDRESS_NEGOTIATION_REQUEST:
            PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL, "FAKE_ADDRESS_NEGOTIATION_REQUEST\n");
            receivedNegotiationRequest (blk);
        break;

        case FAKE_ADDRESS_NEGOTIATION_REPLY_YES:
            PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL, "FAKE_ADDRESS_NEGOTIATION_REPLY_YES\n");
            receivedNegotiationReplyYes(blk);
        break;
        
        case FAKE_ADDRESS_NEGOTIATION_REPLY_NO:
            PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL, "FAKE_ADDRESS_NEGOTIATION_REPLY_NO\n");
            receivedNegotiationReplyNo();
        break;
    }
}

void FakeAddressNegotiator::receivedNegotiationReplyNo () {
    this->state = FAKE_ADDRESS_NEGOTIATION_FAILED;

    // Attempt to send DNS replies.  DNS reply will be sent if all modules are ready.
    this->vn->answerDNSRequests();
}

/**
  * Receives a message indicating that the fake address negotiation request made earlier was 
  * successful.
  */
void FakeAddressNegotiator::receivedNegotiationReplyYes (FakeAddressNegotiationBlock *blk) {

    if ( proxy->checkAllocateFakeIP (blk->requestedAddr)) {
        
        PRINT_DEBUG1 (DEBUG_LEVEL_SUPER,
                "Requested fake IP address %s successfully allocated.\n",
                inet_ntoa (blk->requestedAddr)
            );

        // Now change the VN's fake ip address and the mappings in the proxy.
        proxy->changeVNFakeIP (vn, blk->requestedAddr);
        this->state = FAKE_ADDRESS_NEGOTIATION_SUCCESS;

    } else {
        this->state = FAKE_ADDRESS_NEGOTIATION_FAILED;
    }

    // Attempt to send DNS replies.  DNS reply will be sent if all modules are ready.
    this->vn->answerDNSRequests();

}

/**
  * This function is called when a fake address request is received.
  */
void FakeAddressNegotiator::receivedNegotiationRequest (FakeAddressNegotiationBlock * blk) {
    
    // Check if the requested fake IP address is available.  If yes, allocate it.
    if ( proxy->checkAllocateFakeIP (blk->requestedAddr)) {
        
        PRINT_DEBUG1 (DEBUG_LEVEL_SUPER,
                "Requested fake IP address %s successfully allocated.\n",
                inet_ntoa (blk->requestedAddr)
            );

        // Now change the VN's fake ip address and the mappings in the proxy.
        proxy->changeVNFakeIP (vn, blk->requestedAddr);

        // Send YES reply.
        FakeAddressNegotiationBlock* blk = new FakeAddressNegotiationBlock; 
        // TODO Remember to free this.
    
        blk->mainHeader.version = PROXY_VERSION;
        blk->mainHeader.blockType = PROXY_PACKET_BLOCK_FAKE_ADDRESS_NEGOTIATION;
        blk->mainHeader.nextBlock = sizeof (FakeAddressNegotiationBlock);
        blk->msgCode = FAKE_ADDRESS_NEGOTIATION_REPLY_YES;
    
        // Request that the originator host use the address of the tunnel of this machine
        // as the fake IP address in order to avoid address rewriting.
        proxy->getTunnelIP(&(blk->requestedAddr));
        
        // Send this block over the specific overlay associated with the VN.
        // Remember that the block may not be sent immediately.  
        //TODO FIX this->vn->sendBlock ( (ProxyPacketBlock *) blk);
   
        this->state = FAKE_ADDRESS_NEGOTIATION_SUCCESS;
        
    } else {
        PRINT_DEBUG1 (DEBUG_LEVEL_SUPER,
                "Requested fake IP address %s UNAVAILABLE.\n",
                inet_ntoa (blk->requestedAddr)
            );
        FakeAddressNegotiationBlock* blk = new FakeAddressNegotiationBlock; 
        // TODO Remember to free this.
    
        blk->mainHeader.version = PROXY_VERSION;
        blk->mainHeader.blockType = PROXY_PACKET_BLOCK_FAKE_ADDRESS_NEGOTIATION;
        blk->mainHeader.nextBlock = sizeof (FakeAddressNegotiationBlock);
        blk->msgCode = FAKE_ADDRESS_NEGOTIATION_REPLY_NO;
    
        // Send this block over the specific overlay associated with the VN.
        // Remember that the block may not be sent immediately.  
        //TODO FIX this->vn->sendBlock ( (ProxyPacketBlock *) blk);
        
        this->state = FAKE_ADDRESS_NEGOTIATION_FAILED;

    }

}

int FakeAddressNegotiator::getState () {

    return state;
}
