/**
 * \file IPPacket.cpp
 *  \brief This class represents an IP Packet.
 *  Subclasses like DNSPacket, overlay specific packets (I3Packet)
 *  may be derived from this class.
 */

#if defined( __APPLE__ )
#include <sys/types.h>
#endif

#ifndef _WIN32
	#include <netinet/in.h>
	#include <netinet/udp.h>
	#include <arpa/inet.h>
#else
extern "C" {
	#include "netwrap.h"
}
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "IPPacket.h"
#include "DNSPacket.h"

/** Constructs a packet.  
 * @param pktSize The total length of the IP packet (including
 * headers)
 * @param pktContents Pointer to the actual contents of the packet
 * (including headers). This memory will be freed when the packet is
 * destroyed.
 */
IPPacket::IPPacket(int pktSize, char * pktContents) {

    this->pktSize = pktSize;
    this->pktContents = pktContents;
    
    //interpret IP headers
    ipHeader = (struct IPHdr *) pktContents;
    
    //PRINT_DEBUG1(DEBUG_LEVEL_VERBOSE, "IPPacket constructor: source address = %s\n", inet_ntoa(ipHeader->src));
	
	this->freeMemFlag = true;
}

/**
  * Constructs a new IP packet.  
  * The contents buffer is not allocated.
  */
IPPacket::IPPacket() {

    pktSize = 0;
    pktContents = NULL;
	
	this->freeMemFlag = true;
}

/**
  * Destructor.
  * Deallocates the contents buffer if not NULL
  */
IPPacket::~IPPacket() {
	
    if (freeMemFlag && pktContents != NULL) {
        delete[] pktContents;
    }
}

/**
  * Returns the source IP address of this packet.  
  * @param tmpAddrBuf The source IP address is also copied into this
  * parameter.
  */
char * IPPacket::getSrcIPAddress(char* tmpAddrBuf) {

    strncpy(tmpAddrBuf, inet_ntoa(ipHeader->src), MAX_IP_ADDR_LEN);
    return tmpAddrBuf;
}

/**
  * Returns the source IP address of this packet in network byte order.
  */
ipaddr_n_t IPPacket::getSrcIPAddress() {
    return ipHeader->src.s_addr;
}

/**
  * Returns the destination IP address of this packet in network byte order.
  */
ipaddr_n_t IPPacket::getDestIPAddress() {
    return ipHeader->dst.s_addr;
}

/**
  * Returns the destination IP address of this packet.  
  * @param tmpAddrBuf The destination IP address is also copied into this
  * parameter.
  */
char * IPPacket::getDestIPAddress(char* tmpAddrBuf) {

    strncpy(tmpAddrBuf, inet_ntoa(ipHeader->dst), MAX_IP_ADDR_LEN);
    return tmpAddrBuf;
}

/**
  * Get the protocol number of this packet.
  */
protocol_n_t IPPacket::getProtocolNum() {
    return ipHeader->protocol;
}

/**
  * Checks if this IP packet is a DNS request
  * @return true if DNS request, else false
  */
bool IPPacket::isDNSRequest(){
  
    struct udphdr *tmpUDPHdr;
 
    if (pktSize < 21) {
	return false;
    }
    
    if (pktContents[0] == 0x45 && pktContents[9] == 0x11) { // Check if UDP?
        tmpUDPHdr = (struct udphdr *) &pktContents[20];
         
#if (defined(__APPLE__) || defined (_WIN32))
        if (ntohs(tmpUDPHdr->uh_dport) == 0x0035)  // DNS-Request
#else
        if (ntohs(tmpUDPHdr->dest) == 0x0035)  // DNS-Request
#endif
        return true;
    }
  
  return false; 
}

/**
  * This function converts this IP packet to a DNSPacket.
  * It does this by allocating a new DNSPacket and pointing its
  * pktContents to this packet's pktContents.
  */
DNSPacket* IPPacket::convertToDNSPacket() {

    DNSPacket *dnsPkt = new DNSPacket(
                            this->pktSize,
                            this->pktContents,
                            this->ipHeader
                            );
	dnsPkt->freeMemFlag = this->freeMemFlag;
    
    this->pktContents = NULL;
    this->pktSize = 0;
    this->ipHeader = NULL;  

    return dnsPkt;
}

void IPPacket::dumpPacket() {
    for (int index = 0; index < pktSize; index++) {
        if (index %64 ==0) {
            printf("\n");
        }
        printf("%x", pktContents[index]);
    }
}

/**
  * Calculate a unique identifer for the flow to which this packet belongs by 
  * using the src/dst ipaddr/port number fields (if applicable) and the protocol number.
  */
uint32_t IPPacket::getFlowHash() {

    uint16_t srcPort = getSrcPortNum();
    uint16_t dstPort = getDestPortNum();
    uint32_t srcIP = ipHeader->src.s_addr;
    uint32_t dstIP = ipHeader->dst.s_addr;
    uint8_t protocolNum = ipHeader->protocol;


    // Calculate a hash of the src/dst ip/port and protocol fields.
    // Right now we are justing adding them up to generate the hash. 
    // TODO better hash
    return (srcIP + dstIP + srcPort + dstPort + protocolNum);
}


/** Get the source port number(int network byte order) of this packet if TCP or UDP, else returns 0.
  */
portnum_n_t IPPacket::getSrcPortNum() {
    
    portnum_n_t tmpPort = 0;
    uint8_t headerLen = (0x0F & ipHeader->vihl) * 4;  // The header length (in 32bit words) is the last 4 bits of the first octet.
    
    switch (ipHeader->protocol) {
        
        case 1: //ICMP
            // don't do anything
        break;

        case 6: //TCP
        case 17: //UDP
            // get the port numbers.
            tmpPort = * ((uint16_t *) ( ((uint8_t *) ipHeader) + headerLen));

        break;
    }

    return tmpPort;
}

/** Get the destination port number (in network byte order) of this packet if TCP or UDP, else returns 0.
  */
portnum_n_t IPPacket::getDestPortNum() {
    
    portnum_n_t tmpPort = 0;
    uint8_t headerLen = (0x0F & ipHeader->vihl) * 4;  // The header length (in 32bit words) is the last 4 bits of the first octet.
   
    PRINT_DEBUG1 (DEBUG_LEVEL_VERBOSE, "header len = %d.\n", headerLen);
    switch (ipHeader->protocol) {
        
        case 1: //ICMP
            // don't do anything
        break;

        case 6: //TCP
        case 17: //UDP
            // get the port numbers.
            tmpPort = * ((uint16_t *) ( ((uint8_t *) ipHeader) + headerLen + 2));

        break;
    }

    return tmpPort;
}

/**
  * Prints the contents of the packet.
  */
void IPPacket::printPacket() {

    // network byte order
    uint16_t tmpSrcPort = 0;
    uint16_t tmpDstPort = 0;
    
    char tmpProtocol[2000];
    switch (getProtocolNum()) {
        case IPPROTO_TCP:
        {
            uint8_t* tmpTCPHdr = ((uint8_t *) pktContents) + sizeof (IPHdr);
            uint32_t tmpSeq = ntohl ( * ( ((uint32_t *) tmpTCPHdr) + 1));
            uint32_t tmpAck = ntohl (* ( ((uint32_t *) tmpTCPHdr) + 2));
            uint8_t tmpIPHdrLen = ipHeader->vihl & 0x0f;
            sprintf (tmpProtocol, "iphdrlen=%d tcp - seq=%u ack=%u\n", tmpIPHdrLen, tmpSeq, tmpAck);
            tmpSrcPort = getSrcPortNum();
            tmpDstPort = getDestPortNum();
           
            uint8_t tmpOptionKind = *(tmpTCPHdr);
            if (tmpOptionKind == 2) {
                uint8_t tmpOptionLength = *(tmpTCPHdr + 1);
                uint16_t tmpMSS = * ( ((uint16_t *) tmpTCPHdr) + 1);
                char tmpMSSStr[20];
                sprintf (tmpMSSStr, "<MSS %d>", ntohs (tmpMSS));
                strcat (tmpProtocol, tmpMSSStr);
            } else {
                char tmpIPOpt [100];
                sprintf (tmpIPOpt, "<IPOPT: %x%x%x%x>", *tmpTCPHdr, *(tmpTCPHdr + 1) , *(tmpTCPHdr + 2), * (tmpTCPHdr + 3));
                strcat (tmpProtocol, tmpIPOpt);
            }
            /*
			for (int i = 0 ; i < pktSize - 40; i++) {
                char tmpChar[20];
				sprintf ( tmpChar, "%c", *( (char *) (tmpTCPHdr + 20 + i)));
                strcat (tmpProtocol, tmpChar);
			}
            */

			strcat (tmpProtocol, "\n");
            uint8_t tmpFlags = *( (uint8_t *)pktContents + sizeof (IPHdr) + 13);
            if (tmpFlags & 0x02) {
                strcat (tmpProtocol, " syn");
            } 
            if (tmpFlags & 0x01) {
                strcat (tmpProtocol, " fin");
            } 
            if (tmpFlags & 0x08) {
                strcat (tmpProtocol, " psh");
            } 
            if (tmpFlags & 0x10) {
                strcat (tmpProtocol, " ack");
            } 
            if (tmpFlags & 0x04) {
                strcat (tmpProtocol, " rst");
            } 
            if (tmpFlags & 0x20) {
                strcat (tmpProtocol, " urg");
            }
        }
        break;

        case IPPROTO_UDP:
            strcpy (tmpProtocol, "udp");
            tmpSrcPort = getSrcPortNum();
            tmpDstPort = getDestPortNum();
        break;

		case 1: // ICMP
		{
			strcpy (tmpProtocol, "icmp - ");
			uint8_t icmpType = * ((uint8_t *) pktContents + sizeof (IPHdr));
			switch (icmpType) {
				case 0:
					strcat (tmpProtocol, "echo reply");
				break;

				case 3:
                {
					strcat (tmpProtocol, "destination unreachable - ");
					uint8_t tmpCode = * ((uint8_t *) pktContents + sizeof (IPHdr) + 1);
                    switch (tmpCode) {
                        case 0:
                            strcat (tmpProtocol, "Network unreachable error");
                        break;

                        case 1:
                            strcat (tmpProtocol, "Host unreachable error");
                        break;

                        case 2:
                            strcat (tmpProtocol, "Protocol unreachable error");
                        break;

                        case 3:
                            strcat (tmpProtocol, "Port unreachable error");
                        break;
                        
                        case 4:
                            strcat (tmpProtocol, "Datagram is too big");
                        break;
                        
                        case 5:
                            strcat (tmpProtocol, "Source route failed error");
                        break;

                        case 6:
                            strcat (tmpProtocol, "Destination Network Unknown error");
                        break;

                        case 7:
                            strcat (tmpProtocol, "Destination host unknown error");
                        break;

                        case 8:
                            strcat (tmpProtocol, "Source host isolated. Obsolete");
                        break;

                        case 9:
                            strcat (tmpProtocol, "The destination network is administratively prohibited.");
                        break;

                        case 10:
                            strcat (tmpProtocol, "The destination host is administratively prohibited.");
                        break;

                        case 11:
                            strcat (tmpProtocol, "The network is unreachable for Type of Service.");
                        break;

                        case 12:
                            strcat (tmpProtocol, "The host is unreachable for Type of Service.");
                        break;

                        case 13:
                            strcat (tmpProtocol, "Communication administratively prohibited.");
                        break;

                        case 14:
                            strcat (tmpProtocol, "Host precedence violation.");
                        break;

                        case 15:
                            strcat (tmpProtocol, "Precedence cutoff in effect");
                        break;

                        default:
                        {
					        char tmpNum [200];
					        sprintf (tmpNum, "unknown code %d", tmpCode);
					        strcat (tmpProtocol, tmpNum);
                        }
                    }
                }
                break;

				case 8:
					strcat (tmpProtocol, "echo request");
				break;

				case 11:
				{
					uint8_t tmpCode = * ((uint8_t *) pktContents + sizeof (IPHdr) + 1);
					if (tmpCode == 0) {
                        strcat (tmpProtocol, "time exceeded - TTL exceeded.");
					} else {
						strcat (tmpProtocol, "time exceeded - Fragment reassembly timeout.");
					}
				}
				break;

				default:
				{
					char tmpNum [200];
					sprintf (tmpNum, "unknown icmp type %d", icmpType);
					strcat (tmpProtocol, tmpNum);
				}

			}
		}
		break;
        default:
        
            sprintf (tmpProtocol, "unknown_proto %d", getProtocolNum());
            tmpSrcPort = 0;
            tmpDstPort = 0;
        break;
    }

    int tmpTotalLen = pktSize;

    char tmpSrc[100];
    char tmpDest[100];
    char tmpBuf[100];
    sprintf (tmpSrc, "%s:%d", getSrcIPAddress(tmpBuf), ntohs(tmpSrcPort));
    sprintf (tmpDest, "%s:%d", getDestIPAddress(tmpBuf), ntohs(tmpDstPort));
    
    PRINT_DEBUG4 (
                    DEBUG_LEVEL_SUPER,
                    "IPPacket [protocol=%s, totalLen=%d, %s-->%s]\n",
                    tmpProtocol,
                    tmpTotalLen, 
                    tmpSrc,
                    tmpDest
                );
}
