/** 
 * \file JournalManager.cpp
 * \brief The JournalManager class is defined here.  JournalManager is responsible for writing
 * the records of some JournalSource to disk periodically as well as for helping with crash recovery.
*/
#include "Proxy.h"
#include "JournalManager.h"
#include "JournalRecord.h"
#include "oci/OCI.h"
#include "Params.h"
#include "TFunctor.h"

/**
 * Instantiates the journal manager.
 * @param oci A pointer to the OCI.  This is needed for adding timer events.
 * @param journalSrc The source of journal records.
 * @param fileName The file to which the journal entries are to be periodically dumped.
 */
JournalManager::JournalManager( OCI* oci,  JournalSource *journalSrc, char * fileName, bool recoverFromCrash) {

    this->oci = oci;

    this->journalSrc = journalSrc;
    strcpy (this->fileName, fileName);

    if (recoverFromCrash) {
       doCrashRecovery(); 
       start();
    } else {
        // Start the periodic flushing
        start();
    }

}

/**
  * Goes through the journal records one by one and passes it to the journal source.
  */
void JournalManager::doCrashRecovery() {
    
    FILE *fd = fopen(fileName, "r");
    if (fd == NULL) {
        PRINT_INFO1 (INFO_LEVEL_WARNING,
			"Unable to read journal from file %s.\n",
			fileName
		  );

        //Schedule next journal dump
        start();
        return;
    }

    int count = 0;

    char tmpJournalBuf[MAX_JOURNAL_RECORD_LEN];
    while (fgets(tmpJournalBuf, MAX_JOURNAL_RECORD_LEN, fd) != NULL) {

        PRINT_DEBUG2 (
                        DEBUG_LEVEL_SUPER,
                        "Recovering record %d: %s\n",
                        count,
                        tmpJournalBuf
                    );
        journalSrc->recoverRecord (tmpJournalBuf);
        count ++;
    }

    fclose (fd);
}

/**
 * This function flushes all the journal entries to disk.
 */
void JournalManager::flushToDisk() {

    if (oci->proxy->get_stopProxyFlag()) {
        return;
    }
   
    list<JournalRecord *> recList;
    
    journalSrc->getJournalRecordList(&recList);
   
    PRINT_DEBUG2 (
                    DEBUG_LEVEL_VERY_VERBOSE,
                    "JournalManager: Flushing %d journal records to file %s\n",
                    recList.size(),
                    fileName
                );
     
    // Get the list of records.  Each record is a string
    // which can be directly dumped to the file.
    list<JournalRecord *>::iterator itr = recList.begin();
    list<JournalRecord *>::iterator endItr = recList.end();

    FILE *fd = fopen(fileName, "w");
    if (fd == NULL) {
        PRINT_INFO1 (
	            INFO_LEVEL_WARNING, 
		        "Unable to flush journal to to file: %s\n", 
		        fileName
	        );
	    return;
    }
    
    
    while (itr != endItr) {
        JournalRecord *currRec = (*itr);

        char tmpBuf[MAX_JOURNAL_RECORD_LEN];
        currRec->getJournalRecord (tmpBuf);
       
        //printf ("%p: %s\n", currRec, tmpBuf);
        //write the current record to disk
	    fputs(tmpBuf, fd);
	    fputc('\n',fd);
	    itr++;
    }
    
    fclose(fd);


    // Schedule the next flush to disk
    TObjectFunctor<JournalManager> *tmpFunctor = 
                        new TObjectFunctor<JournalManager> (this, &JournalManager::flushToDisk);
    timer = TimerEvent::create(JOURNAL_FLUSH_PERIOD, 0, (TFunctor *) tmpFunctor, true);
    oci->addTimerEvent(timer);
 
}

/**
  * This function stops the periodic flushing to disk.
  */
void JournalManager::stop() {

    if (timer != NULL) {
        timer->setDisabled();
        timer = NULL;
    }
}

/**
  * This function starts the periodic flush.
  */
void JournalManager::start() {
    TObjectFunctor<JournalManager> *tmpFunctor = 
                        new TObjectFunctor<JournalManager> (this, &JournalManager::flushToDisk);
    timer = TimerEvent::create (JOURNAL_FLUSH_PERIOD, 0, (TFunctor *) tmpFunctor, true);
    oci->addTimerEvent(timer);
 
}
