/**
 * \file LegacyAppLayer.cpp
 * \brief This file defines the LegacyAppLayer.  The LegacyAppLayer acts as an interface to legacy applications.
 * It has functions to send packets to legacy apps.  Also the packets intercepted from legacy apps first come here.
 */

#include "LegacyAppLayer.h"
#include "FakeAddressManager.h"
#include "VirtualNode.h"
#include "oci/OCI.h"
#include "oci/OCIState.h"
#include "common.h"
#include "debug.h"
#include "Tunnel.h"
#include "Proxy.h"
#include "IPPacket.h"

#include "utilFunctions.h"

#include <assert.h>

#ifndef _WIN32
	#include <sys/socket.h>
	#include <sys/stat.h>
	#include <arpa/inet.h>
	#include <netinet/in.h>
#else 
extern "C" {
	#include "netwrap.h"
}
#endif

LegacyAppLayer::LegacyAppLayer (Proxy *proxy) {

    this->proxy = proxy;
    this->tunnel = NULL;
}

void LegacyAppLayer::setTunnel (Tunnel *tunnel) {
    this->tunnel = tunnel;
}


/**
  * This function is called when the specified packet is to be sent
  * to the legacy application associated with the given VN.
  */
void LegacyAppLayer::sendToLegApp (VirtualNode *vn, char *pkt, int pktLen) {

    struct in_addr myFakeAddr; //Will be in network byte order
    if (inet_aton(vn->fakeIP->c_str(), &myFakeAddr) == 0) {
        PRINT_DEBUG1 (
                DEBUG_LEVEL_MINIMAL, 
                "Error while fixing checksum: Invalid fake IP :%s\n", 
                vn->fakeIP->c_str()
            );          
    }


    struct in_addr finalDestAddr; //this is also in network byte order
    
    if (vn->flowInitiatedFlag) {
        // If the vn was created due to a normal IP packet and not a DNS request,
        // the from address of packets sent out would have been saved in vn->fromIPAddr.
        // For reply packets, we need to put that same address as the destination address.
        finalDestAddr.s_addr = vn->fromIPAddr; 
    
    } else {
        // For connections set up on DNS requests, the from address of packets sent out will
        // be the tunnel IP address, as the application binds to the tunnel IP.
        // To send the reply packets to the application, we need to change the destination IP address
        // to the tunnel IP.
#ifdef __APPLE__
        finalDestAddr = proxy->my_addr_nf;
#else
        inet_aton (tunnel->getTunnelIP(), &finalDestAddr);
#endif
    }
        
    Tunnel::change_addr_fix_checksum (pkt, myFakeAddr, finalDestAddr);

    int dataLen = pktLen;
/*  TODO change address depending on address type

    if (vn->localI3Address.type == I3_ADDR_TYPE_IPv4) {
        change_addr_fix_checksum (ip_hdr, vn);
    }
    
#ifndef __CYGWIN__
    else if (vn->localI3Address.type == I3_ADDR_TYPE_IPv6) {
        PRINT_INFO0 (
                INFO_LEVEL_WARNING, 
                "\n IPv6 is not implemented yet in handle_data_packet(...)"
                );
        return;
    }
#endif
    else {
        PRINT_INFO0 (
                INFO_LEVEL_FATAL_ERROR, 
                "\n ERROR: Unknown address type in handle_data_packet(...)"
                );
        return;
    }
*/
    //send the packet to the legacy app over the tunnel.
    PRINT_DEBUG3(
            DEBUG_LEVEL_VERBOSE, 
            "Writing to tunnel len=%d from fakeIP=%s to %s.\n ",
            dataLen, vn->fakeIP->c_str(),
            inet_ntoa (finalDestAddr)
        );
    
    if (vn->dnsPacket != NULL) {
        PRINT_DEBUG1(
                DEBUG_LEVEL_VERBOSE, 
                " for dnsPacket: %s", 
                vn->dnsPacket->reqURL.c_str());     
    }
    
    tunnel->writeToTunnel(pkt, dataLen);


}


/**
  * This function is called to set up a new VirtualNode to interface to a legacy application.
  * @param ociState The OCI state associated with this VirtualNode.
  * @return Pointer to newly created virtual node.  NULL if unable to create new virtual node.
  */
VirtualNode* LegacyAppLayer::getNewVN (OCIState *ociState) {

    VirtualNode* vn = new VirtualNode (ociState, proxy);
    
    vn->fakeIP = proxy->fakeAddrManager->getNewFakeIP();

    if (vn->fakeIP == NULL) {
        PRINT_DEBUG0(
                DEBUG_LEVEL_MINIMAL,
                "LegacyAppLayer setup failed.\n"
            );
        return NULL;
    }

    addVN_FakeIP(vn);
	    
    // create the URL to VN mapping in the proxy.
	if (vn->dnsPacket != NULL) {
		addVN_URL(vn);
	}

    return vn;
}

/** 
 * This function is called when the tunnel captures a packet
 * sent by a legacy application.
 */
void LegacyAppLayer::tunnelCapturedPacket(IPPacket *captIPPkt) {
        
    if (captIPPkt->isDNSRequest()) {
        // if the captured packet is a DNS request  

        char tmpDestIP[MAX_IP_ADDR_LEN];
        string tmpDestIPStr(captIPPkt->getDestIPAddress(tmpDestIP));
        char tmpSrcIP[MAX_IP_ADDR_LEN];
        string tmpSrcIPStr(captIPPkt->getSrcIPAddress(tmpSrcIP));

        DNSPacket *dnsPkt = captIPPkt->convertToDNSPacket();
        //captIPPkt's buffer would have been set to NULL. So the packet contents won't
        // be freed by the next line
        delete(captIPPkt);


        PRINT_DEBUG3(DEBUG_LEVEL_VERBOSE, "Packet captured by Tunnel is a DNS Packet (%s-->%s) !!! Requested URL=%s\n\n",
                            tmpSrcIPStr.c_str(),
                            tmpDestIPStr.c_str(),
                            (dnsPkt->reqURL).c_str()
                        );
           

	if (	isSuffix(dnsPkt->reqURL.c_str(), "in-addr.arpa") || // reverse DNS look ups
		isPrefix(dnsPkt->reqURL.c_str(), "_")	// SRV lookups TODO: Do they always start with _?
			) { 
	    // These are special kinds of DNS requests.
	    // Handle them like legacy DNS requests.
            // Don't send to the server proxy if present.

            // processLegacyDNSRequest is responsible for freeing dnsPkt
            proxy->processLegacyDNSRequest(dnsPkt);
            return;
        }
        
  	    if (proxy->functionAsClientRemoteProxy) { 
            char tmpDNSReqURL[MAX_URL_LEN];
            strcpy (tmpDNSReqURL, dnsPkt->reqURL.c_str());

            // Try to strip all possible suffixes. Only the first matching suffix will be stripped.
            list<string>::iterator itr = proxy->clientRemoteProxyDNSSuffixList.begin();
            while (itr != proxy->clientRemoteProxyDNSSuffixList.end()) {
                
                bool tmpStripFlag = stripSuffix (tmpDNSReqURL, (*itr).c_str());
                PRINT_DEBUG2 (DEBUG_LEVEL_VERBOSE, "Stripped %s to %s\n", dnsPkt->reqURL.c_str(), tmpDNSReqURL);
                if (tmpStripFlag) {
                    //match, look no more
                    break;
                }
                itr++;
            }
            dnsPkt->reqURL = tmpDNSReqURL;
         
 	    }

        // Check if a Virtual Node already exists for this URL.
        VirtualNode* vn = lookupVN_URL(dnsPkt->reqURL);
           
        if (vn != NULL && vn->ociState != NULL && ! vn->ociState->killMeFlag) {
            char tmpBuf[MAX_PRINT_BUF];
            PRINT_DEBUG1 (
                                DEBUG_LEVEL_MINIMAL,
                                "FOUND VirtualNode !!!\n%s",
                                vn->toString(tmpBuf)
                            );
               
            if (vn->ociState->isReady()) {

                    tunnel->answerDNSRequest(dnsPkt, vn->fakeIP);

                    // We no longer need this dnsPkt
                    delete (dnsPkt);
                    return;
            } else {
                    vn->queueDNSRequest(dnsPkt);

                    // This dnsPkt has been queued.  We can't free it yet.
                    // It will be freed when the answer is written.
                    return;
            }

        } else {
                
            //Virtual Node for this DNS URL does not exist, so
            //create new one.
          
		    bool dropConnectionFlag;
            OCIState *ociState = proxy->oci->setUpNewConnection (
                                                    &(dnsPkt->reqURL),
                                                    NULL, // no captured IP packet
                                                    NULL, 0,
													&dropConnectionFlag);
                   

            if (!dropConnectionFlag && ociState != NULL) {
                    
                VirtualNode *newVN = new VirtualNode (ociState, proxy);
                ociState->initiatorVN = newVN;

                // Some OC-D wants to handle this connection.
                // So allocate fake IP address and do further setup.

                IPAddress* tmpFakeIP = NULL;
                /* TODO FIX
                if (prefs.negotiateFakeAddress) {
                    //Should we try to negotiate the same fake address at both ends?
                    tmpFakeIP = proxy->getNewRandomFakeIP();
                } else {
                */
               

                // try to allocate a fake IP address which is a hash of the requested url.
                tmpFakeIP = proxy->getNewFakeIP (dnsPkt->reqURL);

                if (tmpFakeIP == NULL) {
                    // could not allocate fake ip as hash of URL.  Get random.
                    tmpFakeIP = proxy->getNewFakeIP();
                }
                //}
                
                if (tmpFakeIP == NULL) {

                    //We have run out of free fake IP address
                    PRINT_INFO1 (
                                INFO_LEVEL_MINIMAL, 
                                "\nNo more fake IP addresses available. "//
                                "Cannot accept this new connection to %s.\n", 
                                (dnsPkt->reqURL).c_str()
                            ); 
                    
                    PRINT_DEBUG1(
                                DEBUG_LEVEL_FATAL, 
                                "\nNo more fake IP addresses available. "//
                                "Cannot accept this new connection to %s.\n", 
                                (dnsPkt->reqURL).c_str()
                            ); 
                    tunnel->answerDNSRequest(dnsPkt, NULL);
                    // We no longer need this dnsPkt
                    delete (dnsPkt);

                    return;
                    
                } else {

                    // we have successfully obtained a fake IP for this new connection
                    PRINT_DEBUG2(
                                DEBUG_LEVEL_VERBOSE, 
                                "Obtained new fake ip %s for %s.\n", 
                                tmpFakeIP->c_str(), dnsPkt->reqURL.c_str()
                            );
                    newVN->dnsPacket = dnsPkt;
                    newVN->fakeIP = tmpFakeIP;
                    
                    // create hashtable entry for the new VN keyed by
                    // the URL.
                    addVN_URL(newVN);
                    
                    // add the VN to the hash table with the vn's fake IP address as the key.
                    addVN_FakeIP(newVN);

                    // Don't free this DNS packet yet.  This packet will be freed only
                    // when new VN is deleted; not even when the DNS reply is sent
                    newVN->queueDNSRequest(dnsPkt);

                    return;
                }
                
            } else { //OCIState != NULL
            
				if (! dropConnectionFlag) {
	                // None of the active OC-Ds wish to handle
    	            // the URL associated with the captured DNS packet.
        	        // So we use normal IP for this connection.
            		// No OCD has signalled that this packet be dropped either.
            	    PRINT_DEBUG1(
                	            DEBUG_LEVEL_VERBOSE, 
                    	        "\nUsing normal IP for connection to %s.\n", 
                        	    dnsPkt->reqURL.c_str()
                        	);
	                // processLegacyDNSRequest is responsible for freeing dnsPkt
    	            proxy->processLegacyDNSRequest(dnsPkt);
        	        return;
					
				} else {
					// This connection is to be dropped.
					// Send a NXDOMAIN DNS reply
            	    PRINT_DEBUG1(
                	            DEBUG_LEVEL_MINIMAL, 
                    	        "\nDropping connection to %s.\n", 
                        	    dnsPkt->reqURL.c_str()
                        	);
                    tunnel->answerDNSRequest(dnsPkt, NULL);
                    // We no longer need this dnsPkt
                    delete (dnsPkt);
					return;
				}
        
            } //OCIState != NULL

            // Processing will resume when the callback LegacyAppLayer->setupFailed or setupSucceeded
            // is invoked by the overlay layer.
                
        }
        
    } else {    //captured packet is a data packet
       
        // Retrieve the VirtualNode associated with the fake IP address
        char tmpDestIP[MAX_IP_ADDR_LEN];
        string tmpDestIPStr(captIPPkt->getDestIPAddress(tmpDestIP));
        char tmpSrcIP[MAX_IP_ADDR_LEN];
        string tmpSrcIPStr(captIPPkt->getSrcIPAddress(tmpSrcIP));
        PRINT_DEBUG2 (DEBUG_LEVEL_VERY_VERBOSE, "Tunnel captured data packet to %s from %s.\n", 
                tmpDestIPStr.c_str(),
                tmpSrcIPStr.c_str());
        captIPPkt->printPacket();
        VirtualNode *vn = NULL;
        
        // In this advanced version of the proxy, not only packets with fake destination IP
        // addresses are captured.  Packets are also captured on the basis of the dst/src addr/port
        // number Match rules in outbound.xml.
        
        // First we check if the destionation IP address is a fake IP address. If yes,
        // we just lookup the virtual node associated with it using the fake ip address.
        // The mapping would have been established when the DNS request was received.
        
        if ( ! proxy->fakeAddrManager->isFakeIP(tmpDestIPStr)) {
            // The destination IP address of this packet is not a fake IP.
            // So there are two options.
            // Option 1: This packet is the first packet of a new flow that was set up to be 
            // handled by the proxy (through outbound.xml config file)
            // Option 2: This packet is part of some existing flow handled by the proxy
            // (flow setup previously through Option 1)
            // First check if it is option 1 or option 2.
            vn = lookupVN_FlowHash (captIPPkt);
            if (vn == NULL) {
                // This is a new flow, i.e. Option 1.
                // We need to set up the connection.  This is similar to what we do when we get a new 
                // DNS request.
                PRINT_DEBUG2 (
                                DEBUG_LEVEL_MINIMAL,
                                "New flow to %s:%d.\n",
                                tmpDestIPStr.c_str(), ntohs(captIPPkt->getDestPortNum())
                        );
        
                // Set up a new connection for this captured packet. 
				bool dropConnectionFlag;
                OCIState *ociState = proxy->oci->setUpNewConnection (
                                                NULL, // No DNS packet
                                                captIPPkt,
                                                NULL, 0,
												&dropConnectionFlag);
                
                if (ociState != NULL) {
                    // Some OC-D wants to handle this connection.
        
                    VirtualNode *newVN = new VirtualNode (ociState, proxy);
                    ociState->initiatorVN = newVN;
                
                    // Set the flow hash of this connection
                    newVN->flowHash = captIPPkt->getFlowHash();
                    newVN->flowInitiatedFlag = true;
                    newVN->fromIPAddr = captIPPkt->getSrcIPAddress();
        
                    // There is no DNS request associated with this connection
                    newVN->dnsPacket = NULL;
                    // The actual destination IP address plays the role of the fake ip address.
                    // We may be unnecessarily rewriting the destination IP address to itself later in the
                    // code.  This is inefficient, but we can reuse existing code for handling these
                    // type of connections without modifications.
                    newVN->fakeIP = new IPAddress (tmpDestIPStr.c_str());
            
                    // add the VN to the hash table with the vn's flow hash
                    addVN_FlowHash(newVN);
        
                } else {
            
                    char tmpSrcBuf[50];
                    PRINT_INFO4 (
                            INFO_LEVEL_WARNING,
                            "None of the active OC-Ds wish to handle the connection (%s:%d->%s:%d).\n"
                            "Then this packet should not have been captured. There is some error.",
                            captIPPkt->getSrcIPAddress(tmpSrcBuf),
                            ntohs(captIPPkt->getSrcPortNum()),
                            tmpDestIPStr.c_str(),
                            ntohs(captIPPkt->getDestPortNum())
                    );
                }
                // Processing will resume when the callback LegacyAppLayer->setupFailed or setupSucceeded
                // is invoked by the overlay layer.
        
            } else {
                // This is a known flow, i.e. Option 2
                // There is nothing more to do here.  Since the vn has been set,
                // the following code will take care of the rest.
                PRINT_DEBUG2 (
                                DEBUG_LEVEL_VERY_VERBOSE,
                                "Known flow to %s:%d.\n",
                                tmpDestIPStr.c_str(), ntohs(captIPPkt->getDestPortNum())
                        );
            }
            
        } else {
        
            // It is a fake IP
            vn = lookupVN_FakeIP(tmpDestIPStr);
            PRINT_DEBUG1 (DEBUG_LEVEL_VERBOSE, "%s is a fake IP address.\n", tmpDestIPStr.c_str());
        }
                        
        if (vn != NULL) {
            
            // Data has been received from the legacy application for this OCIState.
            // So it is active
            vn->ociState->setActive();
            
            if (vn->isReady()) {
            
                if (vn->ociState == NULL) {
                    char tmpBuf[MAX_PRINT_BUF];
                    PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,"\nOverlayto use null for VN %s.\n", vn->toString(tmpBuf));
                
                } else { 
        
                    proxy->oci->send( vn, captIPPkt->pktContents, captIPPkt->pktSize);
                }
                delete (captIPPkt);
            
            } else {
                // This VN is not ready - it has probably timed out and being in the state 
                // of refresh.  The legacy application should not have started sending data packets
                // on the fake IP address before getting a DNS reply, which doesn't happen till
                // the VN state is VN_STATUS_READY.
                
                PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL, "\n\nQueueing data packets\n");
                vn->queueDataPacket(captIPPkt);
                // the captIPPkt should be freed when the queue is flushed or destroyed.
            }
        } else {
	        //Beep(1000,30);	
            delete (captIPPkt);
            PRINT_DEBUG1(DEBUG_LEVEL_MINIMAL, "Received data packet on unknown fake IP address %s.\n", tmpDestIPStr.c_str());
        }
    }

}


/**
  * This function is used to add a virtual node to a hashtable keyed by its 
  * fake IP address. This is used to lookup the VN when a data packet is 
  * received on a fake IP address.
  *
  * @param vn The virtual node to be added into the hash table.  The fakeIP 
  * field should have already been set.
  */
void LegacyAppLayer::addVN_FakeIP (VirtualNode *vn) {
    map_VN_FakeIP.insert(pair<IPAddress, VirtualNode*>(*vn->fakeIP, vn));
}

/**
  * This function is used to add a virtual node to a hashtable keyed by the 
  * URL of the end point it is masquerading .
  *
  * @param vn The virtual node to be inserted into the hash table.  
  * The url field of the vn should have already been set.
  */
void LegacyAppLayer::addVN_URL (VirtualNode *vn) {
    if (vn->dnsPacket != NULL) {

        map_VN_URL.insert(pair<URL, VirtualNode*>(vn->dnsPacket->reqURL, vn));
    }
}

/** Adds the virtual node to the hash table keyed by its flow hash.
  */
void LegacyAppLayer::addVN_FlowHash (VirtualNode *vn) {
    map_VN_FlowHash.insert(pair<uint32_t, VirtualNode*>(vn->flowHash, vn));
}

/**
  * Remove the specified VN from the hashtable of VNs keyed by the VN's URL 
  */
void LegacyAppLayer::removeVN_URL(VirtualNode *vn) {

    // The if check is need as this function may be called as part of destroyVN
    // for VNs created on a connection request from a remote machine, i.e. the 
    // VN was not created on receiving a DNS request.
    if (vn->dnsPacket != NULL) {
        map_VN_URL.erase( vn->dnsPacket->reqURL);
    }
}

/**
  * Remove the specified VN from the hashtable of VNs keyed by the VN's FakeIP
  */
void LegacyAppLayer::removeVN_FakeIP(VirtualNode *vn) {
    if (vn->fakeIP != NULL) {
        map_VN_FakeIP.erase( *vn->fakeIP);
    }
}

/**
  * Remove the specified VN from the hash table of VNs keyed by VN's FlowHash.
  */
void LegacyAppLayer::removeVN_FlowHash (VirtualNode *vn) {
    if (vn->flowInitiatedFlag) {
        // This makes sense only for flow initiated VNs
        map_VN_FlowHash.erase (vn->flowHash);
    }
}

/**
  * This function is used to lookup a virtual node based on its fake IP address.
  * @param fakeIP  Retrieves the virtual node associated with this fakeIP.
  */
VirtualNode* LegacyAppLayer::lookupVN_FakeIP( IPAddress fakeIP) {
    map<IPAddress, VirtualNode*>::iterator itr = map_VN_FakeIP.find(fakeIP);
    
    if (itr != map_VN_FakeIP.end()) {
        return (VirtualNode*)itr->second;
        
    } else {
            
        return NULL;
    }
}

/**
  * This function is used to lookup a virtual node based on the flow details
  * of the captured packet.  Depending on the packet protocol, the flow details
  * include the src/dst ip/port numbers and protocol number.
  * @param ipPkt Pointer to the packet whose VN is being looked up.
  */
VirtualNode* LegacyAppLayer::lookupVN_FlowHash (IPPacket* ipPkt) {

    // Calculate a unique flow identifier for this packet
    uint32_t flowHash = ipPkt->getFlowHash();

    map<uint32_t, VirtualNode*>::iterator itr = map_VN_FlowHash.find(flowHash);
    
    if (itr != map_VN_FlowHash.end()) {
        return (VirtualNode*)itr->second;
        
    } else {
            
        return NULL;
    }
}
/**
  * This function is used to lookup a virtual node based on the URL of the 
  * end point it is masquerading.
  * @param url  Retrieves the virtual node associated with this url
  */
VirtualNode* LegacyAppLayer::lookupVN_URL( URL url) {
    map<URL, VirtualNode*>::iterator itr = map_VN_URL.find(url);
    
    if (itr != map_VN_URL.end()) {
        return (VirtualNode*)itr->second;
        
    } else {
            
        return NULL;
    }
}


/** Destroy the specified virtual node, i.e.
  * a. Remove it from all hash tables, 
  * b. Release its fake IP,
  * c. Call the VN's destructor to free up memory.
  */
void LegacyAppLayer::destroyVN (VirtualNode *vn) {

    if (vn != NULL) {
        
        removeVN_URL(vn);
        removeVN_FakeIP(vn);
        removeVN_FlowHash(vn);

        if (vn->fakeIP != NULL && !vn->flowInitiatedFlag) {
            // The fake ip address of only those flows initiated by DNS requests
            // should be released
            proxy->releaseFakeIP(vn->fakeIP);
        }

        //TODO XXX Major problem while freeing memory.  Needs detailed study
        delete (vn);
    }
}
