/**
 * \file MyNameConfigFile.cpp 
 * \brief This file defines the parser which extracts name information from the myNames.xml file.
 */

#include "MyNameConfigFile.h"
#include "MyName.h"

#include <list>
#include <stdlib.h>
#include "common.h"
#include "Proxy.h"
#include "ocd/OCD.h"
#include "security/lib/SecurityManager.h"

using namespace std;

MyNameConfigFile::MyNameConfigFile(string fileName, string rootTagName)
            :ConfigFile(fileName, rootTagName) {
    //TODO nothing
}

/**
  * Load the active myNames into the nameLists of the appropriate OCDs.
  * @param proxy The proxy which contains all the OCDs.
  */
void MyNameConfigFile::loadNameList(Proxy *proxy) {
    
    string path = XPATH_MY_NAME;
    
    xmlChar *xpath = xmlCharStrdup(path.c_str());
    xmlXPathObjectPtr result = getNodeSet(xpath);
    xmlFree(xpath);

    if(result == NULL) {
        PRINT_INFO2(
					INFO_LEVEL_VERBOSE, 
					"No %s tags found in the configuration file %s.\n", 
					XPATH_MY_NAME,
					configFileName.c_str());
        return;
    }
    
    xmlNodeSetPtr nodeset = result->nodesetval;
 
    int count = nodeset->nodeNr; 
    
    // go through all the MyName elements and add them to the list
    for(int i=0; i < count; i++) {
        xmlChar *attrib;
        xmlNodePtr currNode = nodeset->nodeTab[i];
        
        attrib = xmlGetProp(currNode, (xmlChar *) ATTRIB_MY_NAME_NAME);
        if (attrib == NULL) {
            PRINT_INFO3(
                    INFO_LEVEL_FATAL_ERROR,
                    "Required attribute %s missing in %s element in file %s.\n",
                    ATTRIB_MY_NAME_NAME,
                    XPATH_MY_NAME,
                    configFileName.c_str()
                );
            EXIT_ON_ERROR;              
        }
        string name((char *)attrib);
        xmlFree(attrib);
        
        attrib = xmlGetProp(currNode, (xmlChar *)ATTRIB_MY_NAME_OCD_ID);
        if (attrib == NULL) {
            PRINT_INFO3(
                    INFO_LEVEL_FATAL_ERROR,
                    "Required attribute %s missing in %s element in file %s.\n",
                    ATTRIB_MY_NAME_OCD_ID,
                    XPATH_MY_NAME,
                    configFileName.c_str()
                );
            EXIT_ON_ERROR;              
        }
        string ocdId((char *)attrib);
        xmlFree(attrib);
        
        attrib = xmlGetProp(currNode, (xmlChar *) ATTRIB_MY_NAME_ACTIVE);
        if (attrib == NULL) {
            PRINT_INFO3(
                    INFO_LEVEL_FATAL_ERROR,
                    "Required attribute %s missing in %s element in file %s.\n",
                    ATTRIB_MY_NAME_ACTIVE,
                    XPATH_MY_NAME,
                    configFileName.c_str()
                );
            EXIT_ON_ERROR;              
        }
        string active((char *)attrib);
        xmlFree(attrib);

        string certificateStr = "";
        attrib = xmlGetProp(currNode, (xmlChar *) ATTRIB_MY_NAME_CERTIFICATE);
        if (attrib != NULL) {
            certificateStr = (char *) attrib;
            xmlFree(attrib);
        }

        
        if (active == "yes" || active == "true") {

            //Create the new MyName object.
            MyName* myName = new MyName (name);
            
            //Now add all attribute value pairs to the map associated with this myName
            for (xmlAttr *currAttr = currNode->properties; currAttr; currAttr = currAttr->next) {

                xmlChar* attrib = xmlGetProp(currNode, currAttr->name);

                PRINT_DEBUG2 (DEBUG_LEVEL_VERBOSE, "Adding property to myName %s --> %s.\n",
                        (char *) currAttr->name, (char *) attrib
                    );
		        myName->addProperty ((char *)currAttr->name, (char *) attrib);
                xmlFree(attrib);
	        }

            OCD* currOCD = proxy->getOCDWithId (ocdId);
            if (currOCD != NULL) {
                currOCD->addMyName (myName);
            } else {
                PRINT_INFO2 (
                                INFO_LEVEL_WARNING,
                                "Unknown OCD Id %s specified for my name %s.\n",
                                ocdId.c_str(), name.c_str()
                            );
                
            }

            /** Load the certificate associated with this OCD if it is available. */
            if (certificateStr != "") {
               proxy->securityManager->loadCertificate (certificateStr.c_str()); 
            }
        }
    }

    xmlXPathFreeObject (result);
}
