/**
 * \file Proxy.cpp
 * \brief Implmentation of the main Proxy class.
 * All initialization and the main proxy select loop are defined in this file.
 */

#if !defined(_WIN32) && !defined(__APPLE__)
	#include "config.h"
#endif

/** 
  * Definitions of the Proxy class
  */
#include <errno.h>

#ifndef _WIN32
	#include <unistd.h>	
	#include <sys/socket.h>
	#include <sys/stat.h>
	#include <arpa/inet.h>
	#include <netinet/in.h>
#else
	extern "C" {
		#include "netwrap.h"
	}
	#include <time.h>
	#include <stdio.h>
    #include "fwint.h"
    #include <sys/types.h> // _ftime()
    #include <sys/timeb.h> // _ftime()
    #include <Windows.h> // struct timeval
#endif

extern "C" {
	#include "netwrap.h"
}
#include <assert.h>

#include "FakeAddressManager.h"
#include "FakeAddressNegotiator.h"
#include "Proxy.h"
#include "ConfigFile.h"
#include "DNSReplyCallback.h"
#include "MyNameConfigFile.h"
#include "oci/OCIState.h"
#include "oci/OCI.h"
#include "ocd/OCD.h"
#include "ocd/OCDTypes.h"
#include "ocd/OCDStateInfo.h"
#include "urlresolver/URLResolver.h"
#include "utils/utilFunctions.h"

#ifndef _WIN32
	// Remote server proxy functionality only in Linux.
	#include "lsip/LSIPv4_OCD.h"
#endif

#include "TimerEvent.h"
#include "FDCallback.h"
#include "TFunctor.h"
#include "Tunnel.h"

#include "LegacyAppLayer.h"
#include "Params.h"


#include "utils/utilFunctions.h"

#include "security/lib/SecurityManager.h"
#include "security/lib/CryptoState.h"

#include "compression/CompressionState.h"

//NEW_OVERLAY

#ifdef ENABLE_RON
    #include "RONOverlay/RONOverlayNetwork.h"
    #include "RONOverlay/RONOCDStateInfo.h"
#endif

#ifdef ENABLE_SAMPLE_OCD
    #include "sampleOCD/SampleOCD.h"
    #include "sampleOCD/SampleOCDStateInfo.h"
#endif



#ifndef _WIN32
	#include <dlfcn.h>
#endif

#include "ocd/OCD.h"

// This is the main proxy object in the OCALA_proxy application.
OCALA_LIB_EXPORT_IMPORT Proxy* Proxy::proxy;

/** 
  * Instantiates and initializes the Generic Proxy object.
  */
Proxy::Proxy(ConfigFile *configFile, bool recoverFromCrash) {

	this->tunnel = NULL;

    mutexInit();

	OCIPathSetupBlock t1;

    stopProxyFlag = false;

    legacyAppLayer = new LegacyAppLayer (this);

    // Initialize Random Number Generator
#if defined (_WIN32)
    srand ( time (NULL) );

#else
    srand(getpid() ^ time(0));	
#endif	
  
    bool addressOKFlag = false; 
    
    my_addr = getCurrentIP(&addressOKFlag);
    
    while ( ! addressOKFlag) {
	    PRINT_INFO0( INFO_LEVEL_WARNING,
			"Unable to obtain local IP address.  Retrying.\n"
		  );
#if defined (_WIN32)
		Sleep ( 5 * 1000); // time in millseconds
#else
	    sleep(5); //sleep for some time before making the next get address attempt. Time in seconds.
#endif
    	my_addr = getCurrentIP(&addressOKFlag);
    }
    //Beep(1000,200);

#if defined (_WIN32) 
	// Sleep for some time.
	// This is required when we are resuming the laptop in Windows.
	// There is a small lag between when the IP address is obtainable
	// the connection is fully established.  During that time,
	// windows blasts dns requests mercilessly.
	Sleep (10 * 1000);
#endif

    my_addr_nf.s_addr = htonl(my_addr.s_addr);
    PRINT_DEBUG1(DEBUG_LEVEL_MINIMAL, "Proxy is running at IP address %s\n", inet_ntoa(my_addr_nf));

    this->configFile = configFile;
    this->configFile->proxy = this;
    
    // read in the overlay independent configuration parameters
    //
    readConfigValues();
  
    initSecurityManager();

	bool tmp_initNameServerFlag = initNameServer();
	while (tmp_initNameServerFlag == false) {
		// Keep trying to initialize the name server
		PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL,
					"Retrying initNameServer() ...\n"
			);

		// Sleep for some time before retrying.
#		if defined (_WIN32) 
			Sleep (5 * 1000);	//time in milliseconds
#		else
			sleep(5);   //sleep 5 seconds
#		endif

		tmp_initNameServerFlag = initNameServer();
	}
       	
    // initialize the fake address manager.
    fakeAddrManager = new FakeAddressManager(fakeAddrRangeStart, fakeAddrRangeMask);

    
    // Intialize the socket used for forwarding legacy DNS requests (for eg: www.yahoo.com) to
    // the default (legacy) DNS servers.
    initDNSSocket();

    oci = new OCI(this);
}

/**
  * Returns the current local (real, not virtual) IP address of the proxy.
  * @param addressFound This flag is set to true on return to indicate that
  * a valid address has been found.  If this flag is set to false,
  * it means that no valid address could be found.  The address returned will be 
  * junk.  This situation was first noticed when resuming windows after a suspend.
  * If you call this function before the wireless interface (the only network interface
  * present on the system), a junk value is returned.
  */
struct in_addr Proxy::getCurrentIP(bool *addressFound) {

    return ::get_my_addr(addressFound);
}

/**
  * Read in the config values from the config file. 
  */
void Proxy::readConfigValues() {

    configFile->loadProxyDetails(this);

    // some sanity checks
    if ( functionAsServerProxy) { //server remote proxy
        if (functionAsMiddleBox) {
	    PRINT_INFO0 (
			INFO_LEVEL_MINIMAL,
			"functionAsMiddleBox flag turned off as functionAsServerProxy flag is true.\n"
			);
	    functionAsMiddleBox = false;
        }

        if (functionAsClientRemoteProxy) {
	    PRINT_INFO0 (
			INFO_LEVEL_MINIMAL,
			"functionAsClientRemoteProxy flag turned off as functionAsServerProxy flag is true.\n"
			);
	    functionAsClientRemoteProxy = false;
        }
    } else if (functionAsClientRemoteProxy) {
        if (functionAsMiddleBox) {
	    PRINT_INFO0 (
			INFO_LEVEL_MINIMAL,
			"functionAsMiddleBox flag turned off as functionAsClientRemoteProxy flag is true.\n"
			);
	    functionAsMiddleBox = false;
        }
    }

    PRINT_DEBUG2(DEBUG_LEVEL_VERBOSE,"Standard DNS request port = %u\nProxy DNS request port =  %u \n", stdDNSRequestPort, dnsRequestPort);

   if (functionAsClientRemoteProxy) {
       string tmpSuffixStr = "";
       list<string>::iterator itr = clientRemoteProxyDNSSuffixList.begin();
       while (itr != clientRemoteProxyDNSSuffixList.end()) {
            tmpSuffixStr += *itr;
           itr++;
       }
       PRINT_INFO1 (
			INFO_LEVEL_MINIMAL,
			"Functioning as Client Remote Proxy.  DNS Suffix = %s\n", tmpSuffixStr.c_str()
   		);
   }

   if (functionAsServerProxy) {
	PRINT_INFO0 (
			INFO_LEVEL_MINIMAL,
			"Functioning as Remote Server Proxy.\n"
		);
   }

   if (functionAsMiddleBox) {
   	PRINT_INFO1 (
			INFO_LEVEL_MINIMAL,
			"Functioning as a middlebox (type = %s).\n", middleBoxTypeName.c_str()
		);
   }
}

/**
  * Reads in the security manager related parameters from the config file
  * and initializes the security manager.
  */
void Proxy::initSecurityManager() {

    unsigned short tmpSecurityLevel;
    configFile->readUShortAttrib(
                            XPATH_SECURITY, 
                            ATTRIB_SECURITY_LEVEL, 
                            &tmpSecurityLevel, 
                            (unsigned short) SECURITY_LEVEL_2
                        );

    char caCertFileName [MAX_FILENAME_LEN];
    configFile->readStringAttrib(
                            XPATH_SECURITY, 
                            ATTRIB_SECURITY_CA_CERT_FILENAME, 
                            caCertFileName,
                            ""
                        );
   
    
    this->securityManager = new SecurityManager (tmpSecurityLevel, "", caCertFileName);
  

}

/**
 * This function marks the specified ipAddr as allocated.
 * This function is usually called when recovering from a crash to mark
 * the fakeIP read in from the journal as already been used.
 * @return true if allocation was successful
 */
bool Proxy::allocateFakeIP(const IPAddress* ipAddr) {

    return fakeAddrManager->allocateFakeIP(ipAddr);
}

/**
  * Releases the specified fake ip address back to the
  * pool of free fake IP addresses.
  */
void Proxy::releaseFakeIP(const IPAddress* fIP) {
    fakeAddrManager->releaseFakeIP(fIP);
}

/**
  * Looks up the path information associated with the specified URL.
  * @return NULL if the url does not match anything in the configuration file.
  * Else returns the path information associated with the url.
  */
PathInfo* Proxy::resolvePath(const URL* url) {

        return urlResolver->resolveURL(*url);
}

/**
  * Looks up the path information associated with the specified IP packet.
  * @return NULL if the packet does not match anything in the configuration file.
  * Else returns the path information associated with the packet
  */
PathInfo* Proxy::resolvePath(IPPacket* pkt) {

        return urlResolver->resolvePacket(pkt);
}

/**
  * Depending on the <code>type</code> field, instantiate a specific subclass of
  * <code>OCD</code> and return it.
  * <br>
  * This function needs to be modified to support a new overlay network.  The
  *  <code>if</code> statement should be extended to include the type of the
  * new overlay network being included.
  * @param type     The type of the overlay network, For eg: I3
  * @param id       The id of the overlay network, For eg: I3.PlanetLab
  * @param desc     A small description of the overlay network
  * @param fileName The name of the overlay specific configuration file.
  * @return a subclass of OCD, NULL if no type matches.
  * NEW_OVERLAY
  */

/**
  * @param libFileName The name of the library file containing the OCD's code.
  */
OCD* Proxy::getNewOCD(
                        string libFileName,
                        string type, string id, 
                        string desc, string fileName,
                        string urlSuffixes, bool handleLegacyURLs) {
   
#ifndef _WIN32
	// Remote server proxy only in Linux
    if (type == OCD_TYPE_IPV4) {
        // LSIP ocds are special OCDs which are compiled into the proxy.
        // So there is no need to dynamically load them.
        return (OCD*) new LSIPv4_OCD (type, id, desc);
    }
#endif

#ifndef _WIN32
    void* dlib = dlopen (libFileName.c_str(), RTLD_NOW |RTLD_LOCAL);
    if (dlib == NULL) {
        PRINT_INFO2 (    INFO_LEVEL_WARNING, 
                        "Unable to open %s: %s\n",
                        libFileName.c_str(),dlerror()
                    );
        return NULL;
    
    } else {
        PRINT_INFO1 (    INFO_LEVEL_MINIMAL,
                        "Successfully loaded %s.\n",
                        libFileName.c_str()
                    );
        OCD* retOCD = ocdMap[type] (    
                                type, id,
                                desc,fileName,
                                urlSuffixes, handleLegacyURLs
                            );
        return retOCD;
    }
#endif

#ifdef _WIN32
	
	
	typedef  OCD* (*makeOCDFull_fn_t)(	string type, string id, string description, string configFileName, string urlSuffixes, bool handleLegacyURLs, FILE* debugFD);

	
	HINSTANCE hDLL;               // Handle to DLL
	makeOCDFull_fn_t makeOCDFull_fn;    // Function pointer

	
	hDLL = LoadLibrary(libFileName.c_str());

	if (hDLL != NULL) {
		makeOCDFull_fn = (makeOCDFull_fn_t)GetProcAddress(hDLL,"makeOCDFull");
	
		if (!makeOCDFull_fn)	{
			// handle the error
			FreeLibrary(hDLL);     
			PRINT_INFO1 (INFO_LEVEL_WARNING,
					"Unable to find OCD maker function in library %s.\n",
					libFileName.c_str()
				);
			return NULL;
		} else {
			// call the function
			OCD* retOCD = makeOCDFull_fn (    
                                type, id,
                                desc,fileName,
                                urlSuffixes, handleLegacyURLs,
								debugFD
                            );
			return retOCD;        
		}
	}
	
	PRINT_INFO1 (INFO_LEVEL_WARNING,
				"Unable to load library %s.\n",
				libFileName.c_str()
			);
	return NULL;

#endif
}
 
/**
  * Loads the list of inbound names associated with this proxy.
  */
void Proxy::initMyNames () {

    char tmpName [MAX_PARAM_LEN];
    configFile->readStringAttrib (XPATH_PROXY_CONFIG_FILE, ATTRIB_MY_NAMES_CONFIG_FILE_NAME, tmpName, true /*required */);
    MyNameConfigFile tmpConfigFile (tmpName, MY_NAME_ROOT_TAG);
    tmpConfigFile.loadNameList (this);

    //Now go through all the load OCDs and initialize the names.
    
    list<OCD*>::iterator itr = OCD_List.begin();
    while (itr != OCD_List.end()) {
        OCD* currOCD = (*itr);
        currOCD->initMyNames();
        itr++;
    }
}

/**
  * Instantiates and initializes the various overlay networks specified in
  * the proxy configuration file.
  * <br>
  * This function needs to be extended if you wish to include a overlay network.
  * Modifications are necessary even for multiple instances of the same type of 
  * overlay.  This can be avoided by having run time type identification as
  * in the case of Java (FUTURE_WORK).
  * NEW_OVERLAY
  */
void Proxy::initOverlays(bool recoverFromCrash) {

    list<OCD*> tmpOverlayList;
    
    configFile->initOCDList(&tmpOverlayList);
   
    if (tmpOverlayList.empty()) {
        PRINT_INFO1(
            INFO_LEVEL_FATAL_ERROR, 
            "The configuration %s does not specific any active overlay networks.\n" //
            "It does not make sense for the proxy to continue. Exiting.\n", 
            (configFile->configFileName).c_str());
        EXIT_ON_ERROR;
    }
    

    // Go through all the overlay networks and initialize them.
    list<OCD*>::iterator itr = tmpOverlayList.begin();
    while (itr != tmpOverlayList.end()) {
        OCD *currNet = (*itr);

        currNet->init (this, oci, recoverFromCrash);
        OCD_List.push_back(currNet);

#ifndef _WIN32
		//LSIP only for Linux
	  	if (currNet->type == OCD_TYPE_IPV4) {
			lsipv4_OCD = (LSIPv4_OCD*) currNet;
		}
#endif
        itr++;
    }

#ifndef _WIN32
    /** Sanity check - If the proxy is functioning as a server proxy,
     *  an LSIP module must have been loaded.
     */
    if (functionAsServerProxy && lsipv4_OCD == NULL) {
	PRINT_INFO0 (INFO_LEVEL_FATAL_ERROR,
 			"The proxy is functioning as a legacy server IP gateway.\n"
			"At least one LSIP module should be specified as active in proxy-config.xml.\n"
			"Exiting.\n"
		);
 	EXIT_ON_ERROR;
    }
#else
	if (functionAsServerProxy) {
		PRINT_INFO0 (INFO_LEVEL_FATAL_ERROR,
					"The windows version of the proxy cannot function as a server proxy.\n"
					"Please set the functionAsServerProxy attribute to false.\n"
				);
		EXIT_ON_ERROR;
	}
#endif

    // You don't need to immediately check OCDs if they are active.  But it gets the
    // periodic process started.
    ocdActiveCheck();
   
}

/** Initialize the URL Resolver */
void Proxy::initURLResolver() {

    char tmpName[MAX_FILENAME_LEN];
    configFile->readStringAttrib (XPATH_PROXY_CONFIG_FILE, ATTRIB_OUTBOUND_CONNECTION_CONFIG_FILE_NAME, tmpName, true /*required */);
    
    urlResolver = new URLResolver(this, tmpName);
}

/**
  * Pretty prints the list of currently loaded (active) overlay networks.
  */
void Proxy::printOverlayList() {

    list<OCD*>::iterator itr = OCD_List.begin();

    PRINT_INFO0(INFO_LEVEL_VERBOSE, "List of currently loaded overlay networks:\n");
    while (itr != OCD_List.end()) {
        char tmpBuf[MAX_PRINT_BUF];
        PRINT_INFO1(INFO_LEVEL_VERBOSE, "%s\n", (*itr)->toString(tmpBuf)); /* Flawfinder: ignore */
        itr++;
    }
}

/**
  * This is the main loop in which the proxy services activity on the
  * timer events and the file descriptors associated with the different overlays.
  * This is an infinite loop.
  */
void Proxy::coreSelect() {
    
    while ( ! stopProxyFlag) {
      
        // Clear any FDCallbacks that have been marked for deletion in the previous iteration.
        clearToDeleteFDCallbackList();

		// Lock the main loop
		lockMainLoop();

        char tmpStr[MAX_PRINT_BUF];
        fd_set rfds;

        int retval;
        
        // Set the read set.
        FD_ZERO(&rfds);
        
        int highestFD = -1;
  
        if (timerEventQueue.empty() && fdCallbackTable.empty()) {
            // if both timer event queue and fd callback table are empty,
            // we sleep for some time and then check again.
            // If select blocks on empty read set and null timer, it will
            // never return
			
			unlockMainLoop();

#if defined (_WIN32) 
			Sleep (1 * 1000);	//time in milliseconds
#else
            sleep(1);   //sleep 1 second
#endif
            continue;
        }
        map<int, FDCallback*>::iterator itr = fdCallbackTable.begin();
        while (itr != fdCallbackTable.end()) {
                    
            FDCallback *cbk = (FDCallback *) itr->second;
            FD_SET(cbk->fd, &rfds);


            if (cbk->fd > highestFD) {
                highestFD = cbk->fd;
            }
            itr++;
        }
        highestFD++;
            
        if (!timerEventQueue.empty()) {
                
            TimerEvent *nextEvent = timerEventQueue.top();
            struct timeval* nextEventTimeVal = nextEvent->getTimeVal();
        
            //current time of day
            struct timeval currentTime;

#ifdef _WIN32
			struct timeb t;
			ftime(&t);
			currentTime.tv_sec = t.time;
			currentTime.tv_usec = t.millitm * 1000;			
#else

            gettimeofday(&currentTime, NULL);
#endif
        
            // subtract current time from the next event time to get the 
            // interval to wait for.
            struct timeval waitInterval;
            waitInterval.tv_sec = nextEventTimeVal->tv_sec - currentTime.tv_sec;
            suseconds_t usecDiff = nextEventTimeVal->tv_usec - currentTime.tv_usec;
            /*
	    printf ("currentTime=%d:%d nextEventTimeVal=%d:%d waitInterval=%d:%d\n",
			currentTime.tv_sec, currentTime.tv_usec,
			nextEventTimeVal->tv_sec, nextEventTimeVal->tv_usec,
			waitInterval.tv_sec, waitInterval.tv_usec	
		);
            */

            // TODO problem if you give 1.5million microseconds
            if (usecDiff < 0) {
                usecDiff += SECOND_USEC;
                waitInterval.tv_sec -= 1;
            }
            waitInterval.tv_usec = usecDiff;
        
            if (waitInterval.tv_sec < 0) {
                // This means that we have passed the deadline for this timer event.
                // This is bad!  But we don't crash - instead we continue and execute the action
                // associated with this event. Hopefully, it wasn't more than a few seconds late!

                PRINT_DEBUG1(DEBUG_LEVEL_VERBOSE, "Late timer event: %s.\n", nextEvent->toString(tmpStr));
                
                nextEvent->execute();

                delete(nextEvent);

                timerEventQueue.pop();

				unlockMainLoop();
                continue;
            }   
       
    	    //printf("-->highestFD =%d   rfds=%d\n", highestFD, rfds);	    
            retval = select(highestFD, &rfds, NULL, NULL, &waitInterval);
            /* Dont rely on the value of tv now! */
       
            //printf ("\naaaaaaaaa retVal = %d\n", retval);
        } else {
            //The timer queue is empty.
            //Just wait indefinitely for activity on the file descriptors.
            //This is because any new TimerEvents can be added only then. Remember
            //we are single threaded.
            
            retval = select(highestFD, &rfds, NULL, NULL, NULL);
        }

#ifdef BENCHMARK
	addCheckpoint (EVENT_DATA_NET_TO_OC_D, 1);
#endif
        if (stopProxyFlag) {
	        //Beep(500,200);
	        //printf("proxyStopFlag true in the middle\n");
			unlockMainLoop();
   	        break;	   
	    }	
        if (retval < 0) {
            
            PRINT_INFO3(
                    INFO_LEVEL_FATAL_ERROR, 
                    "Error in main SELECT statement!\n highestFD=%d,  rfds=%x, errorMsg=%s\n",
			highestFD,  rfds, 
                    strerror(errno)
                    );
            
            perror("select()");
		assert (false);


	    // iterate through all file descriptors and invoke the callbacks.
	    // Note that in this case, all callbacks are invoked because,
	    // an error occurred and the read set was unmodified.
            // The callback function of the fd which caused the problem,
            // should realize that and remove the fd from callback table.
            map<int, FDCallback*>::iterator fitr = fdCallbackTable.begin();
            while (fitr != fdCallbackTable.end()) {
                    
                FDCallback *cbk = (FDCallback *) fitr->second;
                if (FD_ISSET(cbk->fd, &rfds)) {
                    //Don't read the data in here.
                    //Allow the callback function to read from the fd.
                    //This allows the callback to read the data into whatever internal buffer 
                    //it wants to and avoids extra copying.  
                    //For eg: i3 reads it into a cl_buf struct.
                    
                    cbk->call();
                }
                
                fitr++;
            
                /* FD_ISSET(0, &rfds) will be true. */
            }


#ifdef HACK1111
//TODO HACK
//REMOVE THIS CODE AFTER TESTING
            //HACK: start
            // This hack will be fixed very soon
            list<OCD*>::iterator itr = OCD_List.begin();
            while (itr != OCD_List.end()) {
                OCD* tmpOCD = (*itr);

                //printf ("\n************************ %s\n\n", tmpOCD->type.c_str());
                if (tmpOCD->type == "i3" ||
                        tmpOCD->type == "I3" ) {

                    I3OverlayNetwork *i3OCD = (I3OverlayNetwork *) tmpOCD;
                    if (i3OCD->context->g_ctx->is_tcp && i3OCD->context->g_ctx->tcp_fd > 0) {

                        //printf ("\n\n\n\n\n\n\n\n\n\n\nDeleting fd %d\n\n\n\n\n\n", i3OCD->context->g_ctx->tcp_fd);
                        delFDCallback (i3OCD->context->g_ctx->tcp_fd);
                        i3OCD->context->g_ctx->tcp_fd  = -1;
                    }
                }
                itr++;
            }
            //HACK: end
#endif
           
        } else if (retval) {
            
            //Some activity on the file descriptors
        
            // iterate through all file descriptors and check for event.
            map<int, FDCallback*>::iterator itr = fdCallbackTable.begin();
            while (itr != fdCallbackTable.end()) {
                    
                FDCallback *cbk = (FDCallback *) itr->second;
                if (FD_ISSET(cbk->fd, &rfds)) {
                    //Don't read the data in here.
                    //Allow the callback function to read from the fd.
                    //This allows the callback to read the data into whatever internal buffer 
                    //it wants to and avoids extra copying.  
                    //For eg: i3 reads it into a cl_buf struct.
                    
                    cbk->call();
                }
                
                itr++;
            
                /* FD_ISSET(0, &rfds) will be true. */
            }
        } else {
            
            //Timeout occurred
            
            TimerEvent *currEvent = timerEventQueue.top();
            PRINT_DEBUG1(DEBUG_LEVEL_VERY_VERBOSE, "Timeout occurred: %s\n\n", currEvent->toString(tmpStr));
            timerEventQueue.pop();
            currEvent->execute();
            delete(currEvent);
        }

		unlockMainLoop();
    }

    clearToDeleteFDCallbackList();

    closeAllCallbackFDs();
    PRINT_DEBUG0(DEBUG_LEVEL_VERBOSE, "Proxy coreSelect stopped.\n");

}

/**
  * This function looks up the fd callback table for the callback associated
  * with the specified file descriptor.
  * @param fd the file descriptor whose callback we are looking for.
  * @return the associated FDCallback or NULL if fd was not found.
  */
FDCallback* Proxy::lookupFDCallback (int fd) {
    
    map<int,FDCallback*>::iterator itr = fdCallbackTable.find(fd);
    
    if (itr != fdCallbackTable.end()) {
        return (FDCallback*)itr->second;
        
    } else {
            
        return NULL;
    }
}

/**
  * Looks up the DNS packet by a key generated from
  * the dns packet. 
  * @return the associated DNSPacket or NULL if not found.
  */
DNSPacket* Proxy::lookupDNSPacket (unsigned short key) {
    
    map<unsigned short,DNSPacket*>::iterator itr = map_key_DNSPacket.find(key);
    
    if (itr != map_key_DNSPacket.end()) {
        return (DNSPacket*)itr->second;
        
    } else {
            
        return NULL;
    }
}



/**
  * Adds a FDCallback to the list of callbacks in a thread safe manner.
  * The callback function should do the actual read from the socket.
  * If it finds some error (like the tcp socket was closed), it should remove itself
  * from the callback table by calling delFDCallback.
  * @param fd The fd to be called back upon.
  * @param ordinaryFnPtr Pointer to the function to be called.
  */
void Proxy::addFDCallback(int fd, int (*ordinaryFunPtr)(char*)) {

    assert (fd >= 0);

    pthread_mutex_lock(&mutex_fdCallbackTable);

    FDCallback* cbk = new FDCallback(fd, ordinaryFunPtr);
    fdCallbackTable.insert(pair<int, FDCallback *>(cbk->fd, cbk));
    
    pthread_mutex_unlock(&mutex_fdCallbackTable);
}

/**
  * Adds a FDCallback to the list of callbacks in a thread safe manner.
  * @param fd The fd to be called back upon.
  * @param functor The functor encapsulating the member function of some object.
  * @param functorFreeFlag If this flag is set to true, the specified functor will
  * be deleted when the FDCallback is deleted. Default value is false
  */
void Proxy::addFDCallback(int fd, TFunctor* functor, bool functorFreeFlag) {

    assert (fd >= 0);

    pthread_mutex_lock(&mutex_fdCallbackTable);
   
    //delete the existing callback on the specified fd, if any
    delFDCallback (fd);
        
    FDCallback* cbk = new FDCallback(fd, functor, functorFreeFlag);
    fdCallbackTable.insert(pair<int, FDCallback *>(cbk->fd, cbk));
    
    pthread_mutex_unlock(&mutex_fdCallbackTable);
}

/**
  * This function deletes the callback associated with the specified fd.
  */
void Proxy::delFDCallback (int fd) {

    FDCallback *cbk = lookupFDCallback (fd);
    if (cbk == NULL) {
        return;
    }
    
    if (this->mainLoopLockedFlag) {
        // We are in the middle of an iteration of the main loop.
        // We will screw up the fd iterators if we remove the callback now.
        // So we add it to a toDeleteFDList and remove it once the current iteration is completed.
        toDeleteFDList.push_back (cbk->fd);

        // We also mark the cbk as deleted.  So it won't get invoked if encountered even in the current iteration
        // of the main loop.
        cbk->disabledFlag = true;
        
    } else {
        // Delete the cbk immediately
        this->delFDCallback_internal (cbk);
    }
}

/**
  * Actually removes the specified FDCallback.
  * This is a function that is internal to Proxy.cpp
  */
void Proxy::delFDCallback_internal (FDCallback* cbk) {

    assert (cbk != NULL);
    
    // Erase the fd from the actual callback table.
    fdCallbackTable.erase (cbk->fd);


    // Free the memory associated with the callback.
    delete(cbk);

}

/** Actually goes and deletes all the FDCallbacks that have been marked for deletion.
  */
void Proxy::clearToDeleteFDCallbackList () {
    list<int>::iterator itr = toDeleteFDList.begin();
    while (itr != toDeleteFDList.end()) {
        FDCallback* tmpCbk = lookupFDCallback (*itr);
        if (tmpCbk != NULL) {
            delFDCallback_internal (tmpCbk);
        }
        itr++;
    }

    toDeleteFDList.clear();
}

/**
  * Makes the mapping from DNS transaction id to the DNSReplyCallback that must
  * be notified when the reply of the DNS answer is received.
  */
void Proxy::add_DNSReplyCallback_DNSKey (uint16_t transactionId, DNSReplyCallback* replyCbk) {

    //printf ("\n\nAdding tid = %x %p\n", transactionId, ocdState);
    map_DNSReplyCallback_DNSKey.insert(pair<uint16_t, DNSReplyCallback*>(transactionId, replyCbk));

}

void Proxy::del_DNSReplyCallback_DNSKey (uint16_t transactionId) {

    map_DNSReplyCallback_DNSKey.erase(transactionId);

}

DNSReplyCallback* Proxy::lookup_DNSReplyCallback_DNSKey (uint16_t transactionId) {

    map<uint16_t, DNSReplyCallback*>::iterator itr = map_DNSReplyCallback_DNSKey.find(transactionId);
    
    if (itr != map_DNSReplyCallback_DNSKey.end()) {
        return itr->second;
        
    } else {
            
        return NULL;
    }   
}


/**
  * This function adds a timer event in a thread safe manner.
  * @param seconds The number of seconds after which the timer should go off.
  * @param microSeconds The number of micro seconds after which the timer should go off.  
  * This number is appropriately added to the number of seconds parameter.
  * @param description A brief description of the event for debuggin purposes.
  * @param cbk The functor to be invoked when this event happens
  * XXX: Do we need to mutex, since we have only one thread?
  * CHANGE: The cbk won't be freed when the timer object is destroyed, unless cbkFreeFlag is
  * set to true.  By default, it is false.
  * @param cbkFreeFlag Should the cbk which is passed to the TimerEvent be deallocated
  * when the TimerEvent is deallocated? true = yes, false = no
  */
void Proxy::addTimerEvent(
            time_t seconds, suseconds_t microSeconds, 
            TFunctor* cbk,  char *description, bool cbkFreeFlag) {

    // Don't add any timers if the proxy is supposed to be stopped.	
    if (stopProxyFlag) {
    	return;
    }

    pthread_mutex_lock(&mutex_timerEventQueue);
    TimerEvent* tmpTimerEvent = TimerEvent::create (seconds, microSeconds, cbk, description, cbkFreeFlag);
    tmpTimerEvent->addedToQueueFlag = true;
    timerEventQueue.push(tmpTimerEvent);
    pthread_mutex_unlock(&mutex_timerEventQueue);
    
}

/**
  * This function adds a preconstructed timer event to the timer queue.
  * @param tEvent The timer event to be added.
  */
void Proxy::addTimerEvent(TimerEvent* tEvent) {
    // Don't add any timers if the proxy is supposed to be stopped.	
    if (stopProxyFlag) {
    	return;
    }

    pthread_mutex_lock(&mutex_timerEventQueue);
    tEvent->addedToQueueFlag = true;
    timerEventQueue.push(tEvent);
    pthread_mutex_unlock(&mutex_timerEventQueue);
}

/**
  * This function just initializes all the mutexes used in the proxy.
  */
void Proxy::mutexInit() {

    pthread_mutex_init(&mutex_fdCallbackTable, NULL);
    pthread_mutex_init(&mutex_timerEventQueue, NULL);
	pthread_mutex_init(&mutex_mainLoop, NULL);
}

/**
  * This function initializes the socket to be used
  * for relaying DNS requests for legacy urls to legacy DNS servers
  * and for receiving the replies.
  */
void Proxy::initDNSSocket() {

    if ( (dnsFD = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
        PRINT_INFO0(
                INFO_LEVEL_WARNING, 
                "Error opening socket for relaying legacy DNS requests.\n"
                );
        return;
    }

    sockaddr_in saddr;
    memset(&saddr, 0, sizeof(struct sockaddr_in));
    saddr.sin_family =  AF_INET;
    saddr.sin_addr = my_addr_nf;
    
    bool dnsFDOKFlag = false;
    
        saddr.sin_port = htons(dnsRequestPort);

    	if (bind(dnsFD, (struct sockaddr *) &saddr, sizeof(struct sockaddr_in)) < 0) {
            PRINT_INFO2(
                 INFO_LEVEL_WARNING, 
                 "Unable to bind legacy DNS request socket to %s:%d.\n", 
                 inet_ntoa(saddr.sin_addr), ntohs(saddr.sin_port)
                );
    	} else {
	    dnsFDOKFlag = true;
        PRINT_DEBUG2(
                 DEBUG_LEVEL_VERBOSE, 
                 "Bound legacy DNS request socket to %s:%d.\n", 
                 inet_ntoa(saddr.sin_addr), ntohs(saddr.sin_port)
            );
	}
    
    PRINT_DEBUG1(DEBUG_LEVEL_VERBOSE, "Initialized DNS socket %d\n", dnsFD);
    
    // add the function to be called whenever a packet (dns reply) is received on the dns socket.
    TObjectFunctor<Proxy> *fdfunc = new TObjectFunctor<Proxy>(this, &Proxy::receivedLegacyDNSReply);

    if (dnsFDOKFlag) {
    	this->addFDCallback(dnsFD, fdfunc, true);
    }

}

/**
  * Processes a legacy DNS request to be sent via normal IP (as against
  * overlay specific handling of legacy DNS requests).
  * The DNS request is forwarded to its original DNS server and 
  * it is added to a hashtable, waiting for a reply from the server.
  * This function frees up the specified dns pkt if it is no longer needed.
  * In any case, callbacks setup by this function are responsible for freeing the pkt.
  * No one else need to further bother about pkt, nor should they try to use it.
  */
void Proxy::processLegacyDNSRequest(DNSPacket *pkt) {

    if (pkt == NULL || pkt->pktContents == NULL) {
        PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL,
			"Ignoring NULL legacy DNS request packet.\n"
		);

        if (pkt != NULL) {
            delete (pkt);
        }
	    return;
			
    }
    
    
	if (pkt->pktSize < 29) {
	    PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
			"Ignoring invalid legacy DNS packet of length %d.\n",
			pkt->pktSize
		);
        delete (pkt);
	    return;
    }

	PRINT_DEBUG0(DEBUG_LEVEL_VERBOSE, "Processing legacy DNS request.\n");

	sockaddr_in daddr;
    memset(&daddr, 0, sizeof(struct sockaddr_in));
    daddr.sin_family =  AF_INET;
    daddr.sin_port = htons(stdDNSRequestPort);
    /* Flawfinder: ignore */

    
#ifdef _WIN32
	daddr.sin_addr.s_addr = nameServerAddr.sin_addr.s_addr;
#else
	//Linux
    memcpy(&(daddr.sin_addr.s_addr), pkt->pktContents + 16, sizeof (struct in_addr));
#endif

   //TODO Get rid of the numbers 20, 28, etc if practical.
    sendto(dnsFD, pkt->pktContents + 28, 
            pkt->pktSize - 28, 0, (struct sockaddr *) &daddr, 
            sizeof(struct sockaddr)
            );

    // TODO: Make sure that the DNS packet has no options, etc.
    // size must be 28 bytes
    memcpy (pkt->packetData, pkt->pktContents, 20 + 8);
    
    //unsigned short key = ( (uint8_t) pkt->pktContents[29] << 8) + (uint8_t) pkt->pktContents[28];
    unsigned short key = ntohs ( *( (uint16_t *) (pkt->pktContents + 28)));
    
    //pkt->printfPacket();
    map_key_DNSPacket.insert(pair<unsigned short, DNSPacket*>(key, pkt));

	PRINT_DEBUG2(DEBUG_LEVEL_VERBOSE, "map_key_DNSPacket.size = %d (last URL added = %s)\n",
				map_key_DNSPacket.size(),
				pkt->reqURL.c_str()
				);

    // Don't delete this packet yet.  We will delete it when the DNS reply comes back.
    // TODO: Add a periodic purge of map_key_DNSPacket.  Otherwise, DNSPackets for which replies never come
    // back will never get deleted.
}

/**
  * This function is called when a packet (hopefully a DNS reply!) is
  * is received on the socket listening for DNS replies.
  * If there is a VN associated with the received DNS reply, control is handed
  * off to the VN.  Else the packet is is just written to the tunnel for the legacy app
  * to process.
  */
void Proxy::receivedLegacyDNSReply() {
    
    PRINT_DEBUG0(DEBUG_LEVEL_VERY_VERBOSE, "Received Legacy DNS Reply.\n");
  
    char buf[MAX_PACKET_SIZE];
    int len = recv (dnsFD, buf + 28, MAX_PACKET_SIZE - 28, 0);
    if (len <= 0) {
       
        PRINT_INFO0(INFO_LEVEL_WARNING, "Error while receiving DNS reply.\n");
        
    } else {


        //uint16_t key = (((uint8_t)buf[29])<< 8) + (uint8_t) buf[28];
	uint16_t key = ntohs ( *( (uint16_t *) (buf + 28)));

        //printf ("\n\n***** %x %x\n", (uint8_t) buf[29], (uint8_t) buf[28]);
        //printf ("\n\n\n++++++++ %x\n", key);

        DNSPacket *tmpPkt = lookupDNSPacket(key);
     
        if (tmpPkt != NULL) {
       
            // TODO Lots of numbers here.  Is it useful/practical
            // to replace the numbers by consts?
            
            map_key_DNSPacket.erase(key);
            
            memcpy (buf, tmpPkt->pktContents, 20 + 6);

            // change ip addresses
            memcpy (buf + 12, tmpPkt->pktContents + 16, 4);
            memcpy (buf + 16, tmpPkt->pktContents + 12, 4);

            ((uint16_t *) buf)[1] = htons (len + 28);
            ((uint16_t *) buf)[5] = 0;
            ((uint16_t *) buf)[5] = in_cksum ((uint16_t *) buf, 20);

            memcpy (buf + 20, tmpPkt->pktContents + 22, 2);
            memcpy (buf + 22, tmpPkt->pktContents + 20, 2);
            ((uint16_t *) buf)[12] = htons (len + 8);
            memset (buf + 26, 0, 2);
           
            if (tmpPkt->vn != NULL) {
                tmpPkt->vn->receivedDNSReply(tmpPkt);
            } else {
                // send the DNS reply to the legacy application
                tunnel->writeToTunnel(buf, len + 28);
            }

            // We no longer need this DNS packet
            delete (tmpPkt);

        } else {
          
            DNSReplyCallback *tmpDNSReplyCbk = lookup_DNSReplyCallback_DNSKey (key);
            del_DNSReplyCallback_DNSKey (key);
           
            char *replyBuf = buf + 28;
            
           if (tmpDNSReplyCbk != NULL) {
                
                if (!((replyBuf[2] & 0xf8) == 0x80 &&
	                (replyBuf[3] & 0x0f) == 0x00 &&
	                    ntohs(((uint16_t *)replyBuf)[3]) != 0)) {
	
                        PRINT_DEBUG0( DEBUG_LEVEL_MINIMAL, "DNS answer != (Standard query response, No error), or num of Answers != 0\n");
	        
                        tmpDNSReplyCbk->receivedLegacyDNSReply (0);
                
                } else {
	    
                    int i;
                    char  legacyurl[MAX_URL_LEN];
                    uint32_t hostaddr;

                    DNSPacket::parseDNSQueryName (replyBuf + 12, legacyurl);
	                i = 18 + strlen(legacyurl);
	        
                    while (i < len) {
                        if ((replyBuf[i] & 0xc0) == 0xc0) {
	                        
                            i = i + 2;
                        
                        } else {
	                        
                            while (replyBuf[i] != 0) {
	                            i++;
                            }
	                        
                            i++;
	                    }
	                        
                        if (replyBuf[i] == 0x00 && replyBuf[i+1] == 0x01) { //Type: Host address

	                        // retrieve IP address
	                        memcpy(&hostaddr, replyBuf+ i + 10, 4);
                        
                            struct in_addr tmpAddr;
                            tmpAddr.s_addr = hostaddr;
                            tmpDNSReplyCbk->receivedLegacyDNSReply (ntohl(hostaddr));
            
                            return;
                            
                        } else {
	                            
                            i = i + 10 + replyBuf[i+8] * 256 + replyBuf[i+9];
	                    
                        }
	                }
                }
            }
        }
    }
}
                            
            


void Proxy::tunnelCapturedPacket(IPPacket *captIPPkt) {

#ifndef _WIN32
	    
    // Remote server proxy functionality only in Linux

    if ( functionAsServerProxy) {
        // Functioning as remote server proxy
        if (mySoftwareNATFlag) {
            // use custom NAT
            lsipv4_OCD->tunnelCapturedPacket (captIPPkt);
        } else {
            // use Linux NAT
            lsipv4_OCD->tunnelCapturedPacket (captIPPkt);
        }
    } else {
        // Not functioning as remote server proxy
        legacyAppLayer->tunnelCapturedPacket (captIPPkt);
    }

#else
    // Windows, only simple proxy supported.
    legacyAppLayer->tunnelCapturedPacket (captIPPkt);
#endif
}

/** Simply sets the pointer to the tunnel.
  */
void Proxy::setTunnel (Tunnel* tunnel) {

    this->tunnel = tunnel;
}

/**
  * Sets the tunnel (packet capturing device) which interfaces with 
  * the legacy applications.
  */
void Proxy::initTunnel() {
    
    legacyAppLayer->setTunnel (tunnel);

#ifndef _WIN32
    // Do this only if not windows.  In the case of windows, we have a separate 
    // thread for capturing packets from the tunnel.
    // add the function to be called whenever a packet is captured by the tunnel.
    TObjectFunctor<Tunnel> *fdfunc = 
                new TObjectFunctor<Tunnel>(tunnel, &Tunnel::capturePacket);
   

    if ( ! functionAsMiddleBox) {
        // We need the tunnel only if we are interfacing with legacy apps.
        this->addFDCallback(tunnel->getFD(), fdfunc, true);

    }
    
#endif

    
}


/**
 * This function is called to stop the proxy.
 * Only the stop proxy flag is set to true.
 * The proxy may not stop immediately.
 * If this flag is true, we don't wait for any locks.  We immediately stop the proxy.
 */
void Proxy::stopProxy(bool immediateFlag) {

    if (!immediateFlag) {
	    lockMainLoop();
    }

    stopProxyFlag = true;
    
	// Delete the proxy's configuration file object.
	// This ensures that XML parsing is shutdown.
	if (configFile != NULL) {
		delete configFile;
		configFile = NULL;
	} else {
		PRINT_DEBUG0 (DEBUG_LEVEL_WARNING,
						"Proxy's configFile should not be NULL at this point.\n"
					);
	}

	// Close the dns socket
	nw_close(dnsFD);

    tunnel->cleanup();

    //TODO do ocd sleep proxy in Linux also
#ifdef _WIN32

	// Go through all the active OCDs and call stopProxy on them
    // Right now, we assume that only windows version of the proxy can do sleep/resume
	list<OCD*>::iterator itr = OCD_List.begin();
	while (itr != OCD_List.end()) {
		OCD* currOCD = *itr;
		currOCD->sleepProxy();
		itr++;
	}

#endif
    //wait for small time
#ifdef _WIN32
	//Sleep (1 * 1000);	// time in milliseconds
#else
    sleep (1);	// 1 second
#endif

    if (!immediateFlag) {
	    unlockMainLoop();
    }
}

/**
 * Close all the FDs registered for callbacks */
void Proxy::closeAllCallbackFDs() {
    map<int, FDCallback*>::iterator itr = fdCallbackTable.begin();
    while (itr != fdCallbackTable.end()) {
                    
        FDCallback *cbk = (FDCallback *) itr->second;
#if defined(_WIN32)
        nw_close(cbk->fd);
#else
		close (cbk->fd);
#endif

        itr++;
    }
    fdCallbackTable.clear();
}


/**
 * Returns the OCD object with specified id.
 * Returns NULL if not found.
 */
OCD* Proxy::getOCDWithId(string id) {
    list<OCD *>::iterator itr = OCD_List.begin();

    while (itr != OCD_List.end()) {
        
	    OCD* currOverlay = (OCD *) (*itr);
	    if ( currOverlay->id == id) {
	        return currOverlay;
	    }
        
        itr ++;
    }

    return NULL;
}


/**
  * This function is called when an overlay receives a data packet meant for a legacy
  * application.
  * @param vn The virtual node for which the packet is meant.
  * @param data The actual packet contents, may be encrypted.
  * @param dataLen The length of the packet.
  */
void Proxy::receivedDataPacketFromOverlay (VirtualNode *vn, char *data, int dataLen) {

}


/**
  * This function checks if the specified fake IP address is available.
  * If yes, allocates it and returns true.
  * If no, returns false.
  */
bool Proxy::checkAllocateFakeIP (struct in_addr requestedAddr) {

    bool retVal = fakeAddrManager->checkAllocateFakeIP (requestedAddr);
    return retVal;
}

/**
  * Change the fake IP address of this Virtual Node to newAddr.
  * Also update the Proxy's fakeIP to VN mappings.
  */
void Proxy::changeVNFakeIP (VirtualNode *vn, struct in_addr newAddr) {

    /*
    // Remove the old fake IP --> VN mapping
    removeVN_FakeIP(vn);

    //set the new fake IP address.
    if (vn->fakeIP != NULL) {
        delete vn->fakeIP;
    } 
    
    vn->fakeIP = new string (inet_ntoa (newAddr));

    // Add the new fake IP --> VN mapping
    addVN_FakeIP (vn);
    
*/
}

IPAddress* Proxy::getNewFakeIP (string url) {
    return fakeAddrManager->getNewFakeIP(url);
}

IPAddress* Proxy::getNewFakeIP() {
    return fakeAddrManager->getNewFakeIP();
}

IPAddress* Proxy::getNewRandomFakeIP() {

    return fakeAddrManager->getNewRandomFakeIP();
}

void Proxy::getTunnelIP(struct in_addr* a) {

    inet_aton (tunnel->getTunnelIP(), a);
}

/**
  * Returns a new DNS transaction Id.
  */
uint16_t Proxy::getNewDNSTransactionId() {
   
    uint16_t tmpId =  (uint16_t) ((rand() * 1.0)/ RAND_MAX * 65535);
    return tmpId; 
    //dnsTransactionId++;
    //return dnsTransactionId;
}


/**
  * Setup the name server associated with the proxy.
  * This name server is used when the proxy wants to make DNS query on its own.
  * For example, when functioning as a remote proxy, it may need to do DNS lookups.
  * @return true on successful initialization, else returns false.
  */
bool Proxy::initNameServer() {
    
#ifdef _WIN32
	//windows
	//Get the address of the name server.
	PIP_ADAPTER_INFO pAdapterInfo = NULL;
	PIP_PER_ADAPTER_INFO pPerAdapterInfo = NULL;
	ULONG OutBufLen = 0;
	    
	GetAdaptersInfo(NULL, &OutBufLen);
	pAdapterInfo = (PIP_ADAPTER_INFO)malloc(OutBufLen);
	if (pAdapterInfo == NULL) {

		PRINT_DEBUG0 (
				DEBUG_LEVEL_MINIMAL,
				"Unable to allocate PIP_ADAPTER_INFO for the received DNS request.\n"
			);
		return false;
	}
	
	DWORD retVal = GetAdaptersInfo(pAdapterInfo, &OutBufLen);
    if (retVal != ERROR_SUCCESS) {
        if (pAdapterInfo != NULL) {
	        free(pAdapterInfo);
        }     
        if (pPerAdapterInfo != NULL) {
	        free(pPerAdapterInfo);
        }

        PRINT_DEBUG0 (
                    DEBUG_LEVEL_MINIMAL,
                    "Unable to get list of adapters while initializing name server\n"
                );
        //TODO Use FormatMessage to get actual error message
        return false;
    }


	PIP_ADAPTER_INFO pai = pAdapterInfo;
	while (pai) {
	  if (pai->GatewayList.IpAddress.String[0] != 0) {
	    OutBufLen=0;
	    GetPerAdapterInfo(pai->Index, NULL, &OutBufLen);
	    pPerAdapterInfo = (PIP_PER_ADAPTER_INFO)malloc(OutBufLen);
	    if (pPerAdapterInfo == NULL) {
			PRINT_DEBUG0 (
				DEBUG_LEVEL_MINIMAL,
				"Unable to allocate PIP_PER_ADAPTER_INFO for the received DNS request.\n"
			);

			if (pAdapterInfo != NULL) {
				free (pAdapterInfo);
			}

			return false;
	    }
	    
	    GetPerAdapterInfo(pai->Index, pPerAdapterInfo, &OutBufLen);
	    if (pPerAdapterInfo->DnsServerList.IpAddress.String[0] != 0) {
	      
		  struct in_addr tmpAddr;
		  tmpAddr.s_addr = inet_addr(pPerAdapterInfo->DnsServerList.IpAddress.String);

		  if (tunnel == NULL || tmpAddr.s_addr != tunnel->tunDNSServerAddr) {
			  // we should never set the nameserver of the proxy to 
			  // the name server of the tun-device.
			  // we will get into an infinite loop if we do so.
			nameServerAddr.sin_addr.s_addr = tmpAddr.s_addr;
			nameServerAddr.sin_family = AF_INET;
			nameServerAddr.sin_port = htons(stdDNSRequestPort);
	      
			PRINT_INFO1 ( INFO_LEVEL_MINIMAL,
						"Using name server %s.\n",
						inet_ntoa (nameServerAddr.sin_addr)
						);
			char *tmpttt = inet_ntoa (nameServerAddr.sin_addr);
			break;
		  } 
	    }
	  }
	  pai = pai->Next;
	}
	if (pAdapterInfo != NULL) {
		free(pAdapterInfo);
	}
	if (pPerAdapterInfo != NULL) {
		free (pPerAdapterInfo);
	}

	return true;
# else 
	//Linux
    FILE *resolvconf = fopen("/etc/resolv.conf", "r");
    
    if (!resolvconf) {
        PRINT_INFO0 (INFO_LEVEL_WARNING,
                "Could not open /etc/resolv.conf while initializing the Proxy's name server\n");
        return false;

    } else {
        
        char addr_string[MAX_URL_LEN];
        char buf[MAX_PARAM_LEN], *ptr;
        nameServerAddr.sin_family = AF_UNSPEC;
        
        while (fgets(buf, MAX_PARAM_LEN, resolvconf) != NULL) {
            ptr = strtok(buf, " \t");
            
            if (strcmp(ptr, "nameserver") == 0) {
	            
                ptr = strtok(NULL, " ");
	            strcpy(addr_string, ptr);
	            nameServerAddr.sin_family = AF_INET;
	            nameServerAddr.sin_addr.s_addr = inet_addr(addr_string);
	            nameServerAddr.sin_port = htons(stdDNSRequestPort);
	            break;
            }
        }
        fclose(resolvconf);
        if (nameServerAddr.sin_family == AF_UNSPEC) {
            PRINT_INFO0 (INFO_LEVEL_WARNING,
                "Could not find any nameserver in /etc/resolv.conf while intializing the Proxy's name servers.\n"
                );
            return false;
        }
    }
	return true;
#endif // end Linux

}

/**
  * This function is used to make DNS request for the specified name.
  * When the DNS reply is received, the proxy will invoke the receivedLegacyDNSReply
  * callback function on the specified DNSReplyCallback.  The resolved IP address (host byte order)
  * is passed as a parameter to the callback.  0 is passed if the DNS request failed.
  * @param The DNSReplyCallback which should be invoked when the DNS reply is received.
  * @return true if the request was successfully sent (irrespective of reply in the future), else returns false.
  */
bool Proxy::makeDNSRequest (DNSReplyCallback* replyCbk, const char *requestedName) {

    char pktBuf [MAX_PACKET_SIZE];

    uint16_t transactionId = getNewDNSTransactionId();
    replyCbk->dnsTransactionId = transactionId;

    int pktSize = DNSPacket::makeDNSQueryPacket (pktBuf, requestedName, transactionId);
    
    add_DNSReplyCallback_DNSKey (transactionId, replyCbk);

    int numSent = sendto (dnsFD, pktBuf, pktSize, 0 /*flags */, (struct sockaddr *) &nameServerAddr, sizeof (struct sockaddr_in));

    PRINT_DEBUG2 (DEBUG_LEVEL_VERBOSE, "Sending DNS packet for %s to %s.\n", requestedName,
                inet_ntoa (nameServerAddr.sin_addr)
            );

    if (numSent < pktSize) {
        PRINT_DEBUG3 (DEBUG_LEVEL_MINIMAL,  
                "Error while making DNS request. Sent only %d of %d bytes: %s.\n",
                numSent, pktSize, strerror (errno)
            );
        return false;
    }

    return true;

 
}

void Proxy::clearDNSRequest (DNSReplyCallback* replyCbk) {

    if (replyCbk != NULL) {
        del_DNSReplyCallback_DNSKey (replyCbk->dnsTransactionId);
    }
}

/** If the proxy is behaving as a middlebox, it returns the middlebox's name.
  * Else returns NULL.
  */
char *Proxy::getMiddleBoxName() {

    if (functionAsMiddleBox) {
        return middleBoxName;
    } else {
        return NULL;
    }
}

void Proxy::addDstIPMatch (ipaddr_n_t addr) {
    tunnel->addDstIPMatch (addr);
}
void Proxy::addSrcIPMatch (ipaddr_n_t addr) {
    tunnel->addSrcIPMatch (addr);
}
void Proxy::addDstPortMatch (port_protocol_n_t p) {
    tunnel->addDstPortMatch (p);
}
void Proxy::addSrcPortMatch (port_protocol_n_t p) {
    tunnel->addSrcPortMatch (p);
}

/**
 *  This function returns the IP address assigned to the tunnel device.
 */
char* Proxy::getTunnelIP() {
	return tunnel->getTunnelIP(); 
}

/**
  * This function is used only for Remote Server Proxies when our custom software NAT is used.
  */
void Proxy::setNATPortRange (int natPortRangeStart, int natPortRangeEnd) {
	tunnel->setNATPortRange (natPortRangeStart, natPortRangeEnd);
}

void Proxy::lockMainLoop() {
	PRINT_DEBUG0 (DEBUG_LEVEL_VERY_VERBOSE, "Locking Main Loop\n");
	pthread_mutex_lock(&mutex_mainLoop);
    this->mainLoopLockedFlag = true;
    
}

void Proxy::unlockMainLoop() {
	PRINT_DEBUG0 (DEBUG_LEVEL_VERY_VERBOSE, "Unlocking Main Loop\n");
	pthread_mutex_unlock(&mutex_mainLoop);
    this->mainLoopLockedFlag = false;
}

//#ifdef DYN_LOAD_OCD
#include "ocd/OCD.h"
/**
  * This function should be called by an OCD module when it is loaded.
  * @param majorVersion The major version number of the OCALA against which the registering OCD was compiled.
  * @param minorVersion The minor version number of the OCALA against which the registering OCD was compiled.
  * @return true if the registration was successful.
  */
bool registerOCD (string ocdType, makeOCDFull_t* fnPtr, int majorVersion, int minorVersion) {
 
    if (majorVersion != OCALA_MAJOR_VERSION) {
        //version mismatch
        PRINT_INFO5 (    
                        INFO_LEVEL_FATAL_ERROR,
                        "Version mismatch! The OCD module for ocd type \"%s\" uses OCALA version %d.%d "
                        "while this is OCALA version %d.%d.\n",
                        ocdType.c_str(),
                        majorVersion, minorVersion,
                        OCALA_MAJOR_VERSION, OCALA_MINOR_VERSION
                    );
        EXIT_ON_ERROR;
    }

    PRINT_INFO3 (
                    INFO_LEVEL_MINIMAL,
                    "OCD module for OCD type \"%s\" compiled with version %d.%d of OCALA.\n",
                    ocdType.c_str(),
                    majorVersion, minorVersion
                );

    if (Proxy::proxy != NULL) {
        if (Proxy::proxy->ocdMap.find(ocdType) == Proxy::proxy->ocdMap.end()) {
            // An OCD with the same type has not already been loaded.
            Proxy::proxy->ocdMap[ocdType] = fnPtr;
            return true;
        } else {
            PRINT_INFO1 (
                        INFO_LEVEL_WARNING,
                        "Unable to load duplicate OCD with type %s.\n",
                        ocdType.c_str()
                    );
            exit (0);
            return false;
        }
    } else {
        PRINT_INFO0 (INFO_LEVEL_FATAL_ERROR,
                    "registerOCD(): proxy not defined.\n"
                );
        return false;
    }

}

/**
  * Schedule a timer event which will periodi check if the different OCDs are active or not.
  */
void Proxy::ocdActiveCheck  () {

    if (stopProxyFlag) {
        return;
    }

    PRINT_DEBUG0 (DEBUG_LEVEL_VERY_VERBOSE, "OCD Active Check\n");
    // Do the check
    ConfigFile tmpConfigFile (this->configFile->configFileName, TAG_PROXY_CONFIG_FILE);
    tmpConfigFile.updateOCDActiveStatus (&this->OCD_List);

    // Schedule the next check
    
    TObjectFunctor<Proxy> * ocdActiveCheckFunctor = 
                        new TObjectFunctor<Proxy>(
                                        this, 
                                        &Proxy::ocdActiveCheck
                                    );
    
    TimerEvent *ocdActiveCheckTimer = TimerEvent::create (
                                                        5,  //seconds
                                                        0, 
                                                        (TFunctor *) ocdActiveCheckFunctor, 
                                                        "OCD Active Check Functor",
                                                        true
                                                    );
    
    addTimerEvent(ocdActiveCheckTimer);


}
