/**
 *  \file Proxy.h
 *  \brief Declaration of the main proxy class.
*/

#ifndef PROXY_H
#define PROXY_H 1

#include "OCALA_version.h"
#include "common.h"

#include "TimerEvent.h"

#ifdef _WIN32
#	include "pthread.h"
#	include <windows.h>
#endif

#include "OCALA_Export.h"

class OCI;
class OCD;
class OCDStateInfo;
class PathInfo;
class LegacyAppLayer;
class ConfigFile;
class FakeAddressManager;
class VirtualNode;
class Tunnel;
class FDCallback;
class TFunctor;
class IPPacket;
class DNSPacket;
class URLResolver;
class SecurityManager;
class DNSReplyCallback;

#ifndef _WIN32
	class LSIPv4_OCD;
#endif

//DYN_LOAD_OCD
#include "ocd/OCD.h"

#define SETUP_DIRECTION_DOWN 1
#define SETUP_DIRECTION_UP 2


#include <map>
#include <list>
#include <queue>

using namespace std;
/**
  * This file contains the declarations of the generic Proxy.
  */
class Proxy {

private:
    // This is the main proxy object in the OCALA_proxy application.
    OCALA_LIB_EXPORT_IMPORT static Proxy* proxy;

public:
    OCALA_LIB_EXPORT_IMPORT void addTimerEvent(time_t seconds, suseconds_t microSeconds,TFunctor *, char *description, bool cbkFlag);
    OCALA_LIB_EXPORT_IMPORT void addTimerEvent(TimerEvent* tEvent);

    OCALA_LIB_EXPORT_IMPORT uint16_t getNewDNSTransactionId();
    OCALA_LIB_EXPORT_IMPORT bool makeDNSRequest (DNSReplyCallback* , const char *requestedName);
    OCALA_LIB_EXPORT_IMPORT void clearDNSRequest (DNSReplyCallback*);
    
    OCALA_LIB_EXPORT_IMPORT OCD* getOCDWithId(string id);
    OCALA_LIB_EXPORT_IMPORT bool get_stopProxyFlag() {return stopProxyFlag;} 

    OCALA_LIB_EXPORT_IMPORT struct in_addr get_my_addr_nf() {return my_addr_nf;}
    OCALA_LIB_EXPORT_IMPORT struct in_addr get_my_addr() {return my_addr;}
    OCALA_LIB_EXPORT_IMPORT struct in_addr getCurrentIP(bool *);

    OCALA_LIB_EXPORT_IMPORT void getTunnelIP(struct in_addr *);
    OCALA_LIB_EXPORT_IMPORT char* getTunnelIP();

    // START BAD HACK
  bool enableRONFastStartup;
 // END BAD HACK
 //


private:

    /** The list of OCDs currently loaded in the proxy */
    list<OCD*> OCD_List;

    map<uint16_t, DNSReplyCallback*> map_DNSReplyCallback_DNSKey;

    /** The OCI layer. */
    OCI* oci;
    
    //DYN_LOAD_OCD
    /** This is a map of OCD type -> OCD makeOCDFull() function pointers.
     * This map is populated by each OCD shared lib when it is loaded.
     */
    map<string, makeOCDFull_t*, less<string> > ocdMap;

  
    // NOTE: Only one of the functionAs flags should be true

    /** This flag is set to true if this proxy is to function as a remote server proxy.
      * Please note that if this flag is true, the proxy won't function
      * as a local client proxy.
      */
    bool functionAsServerProxy;

    /** This flag is set to true if we are using our own custom software NAT instead of the 
      * Linux software NAT.
      * This flag is relevant only when functionAsServerProxy is also set.
      */
    bool mySoftwareNATFlag;

    /** This flag is set to true if this proxy is to function as a middlebox.
    */
    bool functionAsMiddleBox;

    /** Is this proxy functioning as a remote client proxy. */
    bool functionAsClientRemoteProxy;

    /** If functioning as a client remote proxy, what is the DNS suffix to be stripped off. 
     *  There can be multiple suffixes (separated by commas) specified here.
     *  We start stripping with the first one in the list.
     *  Once a matching suffix is found, we stop.
     *  So order of suffixes in the list is important
     */
    list<string> clientRemoteProxyDNSSuffixList;

  
    /** This object is used to decide which overlay to use for a particular URL */
    URLResolver *urlResolver;
    
    /** The timer event queue sorted by earliest event first */
    priority_queue<TimerEvent*, vector<TimerEvent*>, TimerEventComparator> timerEventQueue;

    // ORIGINAL PLACE OF mutex_timerEventQueue;
    
    /** The list of callbacks on file descriptors currently being monitored. */
    map<int, FDCallback*> fdCallbackTable;
    void delFDCallback_internal (FDCallback *);
    /** This list contains the fds of the FD callbacks that have been marked for deletion. */
    list<int> toDeleteFDList;
    void clearToDeleteFDCallbackList ();

    /** The name server to be used by the proxy to make its own DNS queries */
    struct sockaddr_in nameServerAddr;
    
    /** This hash table is used to store legacy DNS requests
      * keyed on a key derived from the request packet, till the DNS reply comes back from
      * the original DNS server to which the request was sent.
      */
    map<unsigned short, DNSPacket*> map_key_DNSPacket;
    
   /** Mutex for the fdCallbackTable. */
    pthread_mutex_t mutex_fdCallbackTable;
   
    /** This is used to get unique transaction ids while making DNS requests. */
    uint16_t dnsTransactionId;

    /** The packet capturing device used by the proxy to
     * communicate with legacy applications.
     * This tunnel is different from OC-D to OC-D tunnels.
     */
    Tunnel* tunnel;

 
    void addFDCallback(int fd, int (*ordinaryFunPtr)(char *));
    void addFDCallback(int fd, TFunctor *, bool functorFreeFlag);
    void delFDCallback(int fd);
    FDCallback* lookupFDCallback (int fd);

    DNSPacket* lookupDNSPacket (unsigned short key);

    // Used only custom NAT is used for Remote Server Proxy
	void setNATPortRange (int natPortRangeStart, int natPortRangeEnd);

	/** This mutex is used to serialize the mainloop if they are multiple threads
	 * which affect the state inside the mainloop.
	 * For eg: In Windows, suspend messages are asynchronous, but they affect the 
	 * internal state of the Proxy.
	 * This needs to be serialized.
	 */
	pthread_mutex_t mutex_mainLoop;
    /** This variable follows the state of mutex_mainLoop - true if mutex mainLoop is locked, else false. */
    bool mainLoopLockedFlag;
	void lockMainLoop();  
	void unlockMainLoop();

    void mutexInit();

    void initDNSSocket();

    void readConfigValues();

    void closeAllCallbackFDs();

    bool initNameServer();

    
    void ocdActiveCheck ();
    
    /** The proxy's security manager handles security related aspects for all OCDs. */
    SecurityManager* securityManager;
    
    //TODO make private
    /** This is flag is set to true to indicate the proxy to shutdown */
    volatile bool stopProxyFlag;
    
    /** This is just temporary storage.  Actual value is in the FakeAddressManager */
    string fakeAddrRangeStart;
    /** This is just temporary storage.  Actual value is in the FakeAddressManager */
    string fakeAddrRangeMask;
    
    
    /** Mutex for the timer event priority queue. */
    pthread_mutex_t mutex_timerEventQueue;
    
    /** The IP address of local host (this machine).
      * This is in host byte order.
      */
    struct in_addr    my_addr;

    /** The same IP address as my_addr.  But this is in network byte order. */
    struct in_addr    my_addr_nf;

   /** This layer directly interfaces with the legacy applications */
    LegacyAppLayer *legacyAppLayer;

#ifndef _WIN32
    /** This is needed if the proxy is functioning as a remote server proxy. */
    LSIPv4_OCD* lsipv4_OCD;
#endif
    
    /** The standard port from which DNS requests are to be captured */
    unsigned stdDNSRequestPort;

    /** The port from which the proxy makes DNS requests */
    unsigned dnsRequestPort;

    /** The socket used to relay DNS requests for legacy URLs to
      * to legacy DNS servers.
      */
    int dnsFD;
    
    /**
     * The file containing configuration info about the proxy.
     */
    ConfigFile* configFile;
    
    /** This class manages the pool of fake IP addresses available to
     * the proxy.
     */
    FakeAddressManager* fakeAddrManager;
    

    /**
      * Constructor 
      */
    OCALA_LIB_EXPORT_IMPORT Proxy(ConfigFile *configFile, bool recoverFromCrash);


    void releaseFakeIP(const IPAddress *);
    bool allocateFakeIP(const IPAddress *);

    PathInfo* resolvePath(const URL* url);
    PathInfo* resolvePath(IPPacket* pkt);
    
    OCD* getNewOCD(
                            string libFileName,
                            string type, string id, 
                            string desc, string fileName,
                            string urlSuffixes, bool handleLegacyURLs);

    OCALA_LIB_EXPORT_IMPORT void initOverlays(bool);
    OCALA_LIB_EXPORT_IMPORT void initMyNames ();
    OCALA_LIB_EXPORT_IMPORT void initURLResolver();
    OCALA_LIB_EXPORT_IMPORT void initSecurityManager();
    OCALA_LIB_EXPORT_IMPORT void printOverlayList();

    OCALA_LIB_EXPORT_IMPORT void setTunnel(Tunnel *);
    OCALA_LIB_EXPORT_IMPORT void initTunnel();

    OCALA_LIB_EXPORT_IMPORT void coreSelect();

  
    void processLegacyDNSRequest(DNSPacket *);
    void receivedLegacyDNSReply();
    

    OCALA_LIB_EXPORT_IMPORT void stopProxy(bool immediateFlag = false);

    bool checkAllocateFakeIP (struct in_addr requestedAddr);
    void changeVNFakeIP (VirtualNode *vn, struct in_addr newAddr);
    IPAddress* getNewRandomFakeIP();
    IPAddress* getNewFakeIP();
    IPAddress* getNewFakeIP(string url);


    void receivedDataPacketFromOverlay (VirtualNode* vn, char *pkt, int pktSize);

    void tunnelCapturedPacket(IPPacket *captIPPkt);

    char *getMiddleBoxName();
    char middleBoxName[MAX_URL_LEN];

    /** What type of middlebox is this, if it all a middlebox?
      * bro, gateway, etc..
      */
    string middleBoxTypeName;
    /** The name of the class which implements the middlebox functionality. */
    string middleBoxClassName;
    
    void add_DNSReplyCallback_DNSKey (uint16_t transactionId, DNSReplyCallback* replyCbk);
    void del_DNSReplyCallback_DNSKey (uint16_t transactionId);
    DNSReplyCallback* lookup_DNSReplyCallback_DNSKey (uint16_t transactionId);

    void addDstIPMatch (ipaddr_n_t);
    void addSrcIPMatch (ipaddr_n_t);
    void addDstPortMatch (port_protocol_n_t);
    void addSrcPortMatch (port_protocol_n_t);

    

    friend class Tunnel;
    friend class ConfigFile;
    friend class LegacyAppLayer;
    friend class VirtualNode;
    friend class OCI;

#ifndef _WIN32
    friend class LSIPv4_OCD;
#endif

    friend class OCIState;
    friend class URLResolver;
    friend class FakeAddressNegotiator;
    friend class MyNameConfigFile;
        
    friend int main (int argc, char* argv[]);
    friend void startGenericProxy_Linux(); 

#if defined (_WIN32)
    friend void* proxySelect(void *args);
	friend int init_generic_proxy(int argc, char* argv[]);
	friend int resumeGenericProxy();
	friend void updateOverlayStatus();
	friend void* statusUpdater(void *args);
	friend DWORD WINAPI HandlerEx ( DWORD dwControl, DWORD dwEventType, LPVOID lpEventData, LPVOID lpContext);
	friend LRESULT CALLBACK MainWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);
#endif


	friend void * resume_generic_proxy(void *dummyArgs);

    friend void proxy_cleanup (int signum);

    friend class MyNAT;


    friend void OCD::addFDCallback(int fd, TFunctor* functor, bool functorFreeFlag);
    friend void OCD::delFDCallback(int fd);

    friend bool registerOCD (string ocdType, makeOCDFull_t* fnPtr, int majorVersion, int minorVersion);
};
#endif

