/**
 *  \file TFunctor.h
 *  \brief  This class defines a functor, i.e. a function pointer object.
 *  This way, we can have pointer to member functions of classes.
*/

#ifndef FUNCTOR_H
#define FUNCTOR_H 1

#include <stdlib.h>

class VirtualNode;

/**
  * This class is used to encapsulate function pointers
  * (both member function pointers as well as normal pointer.).
  * Different inherited versions of this class are used for that
  * purpose.
  */
class TFunctor {
public:
    virtual void operator()()=0;  // call using operator
    virtual void callIntArg (int intArg) {
    }
};


// derived template class
template <class TClass> class TObjectFunctor : public TFunctor {
private:
    
    /** Pointer to a function taking no arguments */
    void (TClass::*funcPtr_noArgs)();
    
    /** Pointer to a function taking a void* argument */
    void (TClass::*funcPtr_voidArgs)(void *);
    
    /** Pointer to a function taking an int argument */
    void (TClass::*funcPtr_intArg)(int);
  
    /** The data associated with the funcPtr_voidArgs */
    void *voidArgs_data;

    /** Pointer to a function take a VirtualNode *argument */
    void (TClass::*funcPtr_vnArgs)(VirtualNode *);
    
    /** The data associated with funcPtr_vnArgs */
    VirtualNode *vnArgs_data;
    
    TClass* pt2Object;                  // pointer to object

public:

    /**
      * Constructs a ObjectFunctor which takes in a pointer to a function with
      * no arguments.
      */
    TObjectFunctor(TClass* _pt2Object, void(TClass::*funcPtr_noArgs)()) {
        pt2Object = _pt2Object;
        this->funcPtr_noArgs = funcPtr_noArgs;
        this->funcPtr_voidArgs = NULL;
        this->funcPtr_intArg = NULL;
        this->voidArgs_data = NULL;
        this->funcPtr_vnArgs = NULL;
        this->vnArgs_data = NULL;
    }

    /**
      * Constructs a ObjectFunctor which takes in a pointer to a function with
      * a VirtualNode * argument.
      */
    TObjectFunctor(TClass* _pt2Object, void(TClass::*funcPtr_vnArgs)(VirtualNode*), VirtualNode *vn) {
        pt2Object = _pt2Object;
        this->funcPtr_noArgs = NULL;
        this->funcPtr_intArg = NULL;
        this->funcPtr_voidArgs = NULL;
        this->voidArgs_data = NULL;
        this->funcPtr_vnArgs = funcPtr_vnArgs;
        this->vnArgs_data = vn;
        
    }
 
    /**
      * Constructs a ObjectFunctor which takes in a pointer to a function with
      * an int argument.
      */
    TObjectFunctor(TClass* _pt2Object, void(TClass::*funcPtr_intArg)(int)) {
        pt2Object = _pt2Object;
        this->funcPtr_noArgs = NULL;
        this->funcPtr_intArg = funcPtr_intArg;
        this->funcPtr_voidArgs = NULL;
        this->voidArgs_data = NULL;
        this->funcPtr_vnArgs = NULL;
        this->vnArgs_data = NULL;
        
    }
    
    /**
      * Constructs a ObjectFunctor which takes in a pointer to a function with
      * a void * argument
      */
    TObjectFunctor(TClass* _pt2Object, void(TClass::*funcPtr_voidArgs)(void *), void *voidArgs_data) {
        pt2Object = _pt2Object;
        this->funcPtr_voidArgs = funcPtr_voidArgs;
        this->funcPtr_noArgs = NULL;
        this->funcPtr_intArg = NULL;
        this->voidArgs_data = voidArgs_data;
        this->funcPtr_vnArgs = NULL;
        this->vnArgs_data = NULL;
    }
    
    /**
      * This calls the function pointed to by funcPtr_noArgs or
      * funcPtr_voidArgs depending on how the functor was constructed.
      * This call is valid only for ObjectFunctors created with the 
      * constructor accepting a pointer to a function taking no arguments
      * (but the callback may involved a stored parameter).
      */
    virtual void operator()() {
        if (funcPtr_voidArgs != NULL) {
            (*pt2Object.*funcPtr_voidArgs)(voidArgs_data);
        } else if (funcPtr_noArgs != NULL) {
            (*pt2Object.*funcPtr_noArgs)();

        } else {
            (*pt2Object.*funcPtr_vnArgs)(vnArgs_data);
            
        }
    }

   virtual void callIntArg (int intArg) {

       if (funcPtr_intArg != NULL) {

            (*pt2Object.*funcPtr_intArg)(intArg);

       } else if (funcPtr_voidArgs != NULL) {
            (*pt2Object.*funcPtr_voidArgs)(voidArgs_data);
       } else if (funcPtr_noArgs != NULL) {
            (*pt2Object.*funcPtr_noArgs)();

       } else if (funcPtr_vnArgs != NULL) {
            (*pt2Object.*funcPtr_vnArgs)(vnArgs_data);
            
       }

   }
};


class OrdinaryFunctor : public TFunctor {
private:
    /** The pointer to the function to be called. */
    int (*fptr)(char *);
public:
    OrdinaryFunctor(int (*fptr)(char *)) {
        this->fptr = fptr;
    }
    virtual void operator()(char* buf) {
        fptr(buf);
    }
    virtual void operator()() {
        //don't do anything right now
    }
};


#endif
