/**
 * \file TimerEvent.cpp
 * \brief This file defines the TimerEvent object.  TimerEvent is a callback to be invoked at
 * a specific time.
*/

#include "TimerEvent.h"
#include <string.h>
#include <assert.h>

#ifndef _WIN32
	#include <sys/time.h>
#else
    #include <sys/types.h> // _ftime()
    #include <sys/timeb.h> // _ftime()
    #include <Windows.h> // struct timeval
	#include <time.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include "debug.h"

/**
  * Create a new timer event to go off after the specified amount of time - the number of seconds + micro seconds.
  * @param seconds The number of seconds after which the timer should go off.
  * @param microSeconds The number of micro seconds after which the timer should go off.
  * @param cbk The functor to be invoked when this event happens
  * CHANGE: The cbk won't be freed when the timer object is destroyed, unless cbkFreeFlag is
  * set to true.  By default, it is false.
  * @param cbkFreeFlag Should the cbk which is passed to the TimerEvent be deallocated
  * when the TimerEvent is deallocated? true = yes, false = no
  */
TimerEvent* TimerEvent::create (time_t seconds, suseconds_t microSeconds, TFunctor *cbk, bool cbkFreeFlag ) {
    return new TimerEvent (seconds, microSeconds, cbk, cbkFreeFlag);
}

TimerEvent::TimerEvent(time_t seconds, suseconds_t microSeconds, TFunctor *cbk, bool cbkFreeFlag ) 
{

#ifdef _WIN32
	struct timeb t;
	ftime(&t);
	tv.tv_sec = t.time + seconds;
	tv.tv_usec = t.millitm * 1000 + microSeconds;	
#else
    gettimeofday(&tv, NULL);

    // TODO overflow
    tv.tv_sec += seconds;
    tv.tv_usec += microSeconds;
#endif

    this->addedToQueueFlag = false;
    this->cbk = cbk;
    this->cbkFreeFlag = cbkFreeFlag;
    
    disabledFlag = false;
    description[0] = '\0';

    // Make sure that microseconds is between 0 and 10^6 by adjusting seconds
    adjustTimer();
}

/**
  * Create a new timer event to go off after the specified amount of time - the number of seconds + micro seconds.
  * This constructs a string descriptor of the event.
  * @param seconds The number of seconds after which the timer should go off.
  * @param microSeconds The number of micro seconds after which the timer should go off.
  * @param desc A small description of this event.
  * @param cbk The functor to be invoked when this event happens
  * CHANGE: The cbk won't be freed when the timer object is destroyed, unless cbkFreeFlag is
  * set to true.  By default, it is false.
  * @param cbkFreeFlag Should the cbk which is passed to the TimerEvent be deallocated
  * when the TimerEvent is deallocated? true = yes, false = no
  */
TimerEvent* TimerEvent::create(time_t seconds, suseconds_t microSeconds, TFunctor* cbk, char* desc, bool cbkFreeFlag)  {

    return new TimerEvent (seconds, microSeconds, cbk, cbkFreeFlag, desc);

}

TimerEvent::TimerEvent(time_t seconds, suseconds_t microSeconds, TFunctor* cbk, bool cbkFreeFlag, char* desc) 
{

#ifdef _WIN32
	struct timeb t;
	ftime(&t);
	tv.tv_sec = t.time + seconds;
	tv.tv_usec = t.millitm * 1000 + microSeconds;	
#else
	gettimeofday(&tv, NULL);
    tv.tv_sec += seconds;
    tv.tv_usec += microSeconds;
#endif

	if (desc != NULL) {
		strncpy (this->description, desc, TIMER_EVENT_DESCRIPTION_MAX_LEN);
	} else {
		this->description[0] = '\0';
	}
    this->cbk = cbk;
    this->cbkFreeFlag = cbkFreeFlag;
    this->addedToQueueFlag = false;
    disabledFlag = false;
    
    // Make sure that microseconds is between 0 and 10^6 by adjusting seconds
    adjustTimer();
}


/**
 * Make sure that the microseconds of the time at which this timer event is to be fired
 * is between 0 and 10^6.  If not, bump up the seconds field in order to make it so.
 */ 
void TimerEvent::adjustTimer() {
	tv.tv_sec += (tv.tv_usec / SECOND_USEC);
	tv.tv_usec = (tv.tv_usec % SECOND_USEC);	
}

/**
  * Returns a reference to the time value of this event. 
  */
struct timeval *TimerEvent::getTimeVal() {
    return &tv;
}

/** 
  * This function is called when the TimerEvent is to be invoked.
  */
void TimerEvent::execute() {
    if (!disabledFlag) {
        (*cbk)();
    }
}

/**
  * This disables the timer event.  This means that calling the excute method
  * will not do anything useful.
  */
void TimerEvent::setDisabled() {
    disabledFlag = true;
}

/**
  * Returns a string describing the TimerEvent.
  * @param tmpStr This should be a preallocated character of sufficient size to contain the description string.
  * @return a pointer to tmpStr.
  */
char *TimerEvent::toString(char *tmpStr) {
    struct timeval tempTime;


#ifdef _WIN32
	struct timeb t;
	ftime(&t);
	tempTime.tv_sec = t.time;
	tempTime.tv_usec = t.millitm * 1000;	
#else 
	gettimeofday(&tempTime, NULL);
#endif

    sprintf(tmpStr, "TimerEvent (%s) currentTime=(%d,%d) eventTime=(%d,%d)", 
                    description,
                    tempTime.tv_sec, tempTime.tv_usec,
                    tv.tv_sec, tv.tv_usec);
    return tmpStr;
}

/**
  * Safely destroys the specified TimerEvent.
  * Note that you never need to manually destroy a TimerEvent object if you have used
  * it as part of an addTimerEvent call.  This function is only for TimerEvents which you 
  * never call in an addTimerEvent call.  I can't imagine any normal situation where
  * you will create a TimerEvent object and not call addTimerEvent on it. 
  * So you will probably never have to call this function.
  */
void TimerEvent::destroy (TimerEvent* t) {

    if (t == NULL) {
        PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL,
                    "NULL timer even passed to TimerEvent::destroy.\n"
                );
        assert (false);
    }

    if ( !t->addedToQueueFlag) {
        delete (t);
    } else {
        // Just disable the timer
        t->disabledFlag = true;
    }
}

/** Destructor */
TimerEvent::~TimerEvent() {
   

    if (cbk != NULL && cbkFreeFlag == true) {
		//PRINT_INFO1 (INFO_LEVEL_MINIMAL, "TimerEvent::Freeing %p.\n", cbk);
        delete cbk;
    }
    
}

/** This function compares the specified two TimerEvent objects
  * on the basis of their time values.  Earlier time value is lesser.
  */
int TimerEventComparator::operator() (TimerEvent* &x, TimerEvent* &y) {
    if (x->tv.tv_sec < y->tv.tv_sec) {
        return 0;
    } else if (x->tv.tv_sec > y->tv.tv_sec) {
        return 1;
    } else if (x->tv.tv_usec < y->tv.tv_usec) {
        return 0;
    } else if (x->tv.tv_usec > y->tv.tv_usec) {
        return 1;
    } else {
        return 0;
    }
}





