/**
 *  \file TimerEvent.h
 *  \brief This class declares a TimerEvent - a callback function which should be invoked at 
 *   a particular time instant.
*/

#ifndef TIMER_EVENT_H
#define TIMER_EVENT_H 1

#include "OCALA_Export.h"

#ifndef _WIN32
#	include <sys/time.h>
#	include <sys/types.h>
#else
extern "C" {
	#include "netwrap.h"
}
#endif

#include "common.h"
#include "TFunctor.h"

/** The maximum length of the descriptive string associated with a timer event. */
#define TIMER_EVENT_DESCRIPTION_MAX_LEN 100

/**
  * This is the base Timer Event class. 
  * This class is overloaded to perform different actions when the timerevent happens.
  */
class TimerEvent {
protected:
    
    /** The clock time at which this event should be activated. */
    struct timeval tv;
    
    /** A small description of this event. */
    char description [TIMER_EVENT_DESCRIPTION_MAX_LEN];
    
    /** This flag is set to indicate that a timer event is disabled.
      * This implies that the execute function should not do anything.
      * Keep this in mind while extending the TimeEvent class.
      */
    bool disabledFlag;

    /** This flag is set to true if this TimerEvent has been added to the timer queue. */
    bool addedToQueueFlag;
    
    /** The function to be executed when this event happens. */
    TFunctor *cbk;
  
    /** Should the cbk TFunctor be freed when the TimerEvent is destroyed. */
    bool cbkFreeFlag;
   
    void adjustTimer();
 
    /** The constructor and destructors of TimerEvent have been made private on purpose.
      * So now the only way you can create or destroy timers is by calling the 
      * createTimer() and destroyTimer factory methods.
      * If you ever add a TimerEvent to the main loop using the addTimer function,
      * you will never need to call destroyTimer().  Destroying the timer is automatically taken care 
      * for you.
      */
    OCALA_LIB_EXPORT_IMPORT TimerEvent(time_t seconds, suseconds_t microSeconds, TFunctor *t, bool cbkFreeFlag);
    OCALA_LIB_EXPORT_IMPORT TimerEvent(time_t seconds, suseconds_t microSeconds,TFunctor *, bool cbkFreeFlag, char* desc);
    OCALA_LIB_EXPORT_IMPORT virtual ~TimerEvent();
public:
    
    OCALA_LIB_EXPORT_IMPORT static TimerEvent* create ( time_t seconds, suseconds_t microSeconds, TFunctor *t, bool cbkFreeFlag = false);
    OCALA_LIB_EXPORT_IMPORT static TimerEvent* create (time_t seconds, suseconds_t microSeconds,TFunctor *, char* desc, bool cbkFreeFlag = false);
    OCALA_LIB_EXPORT_IMPORT static void destroy (TimerEvent* t);
    
    OCALA_LIB_EXPORT_IMPORT struct timeval *getTimeVal();
    OCALA_LIB_EXPORT_IMPORT char *toString(char *);

    OCALA_LIB_EXPORT_IMPORT void setDisabled();

    OCALA_LIB_EXPORT_IMPORT virtual void execute();

	friend class TimerEventComparator;
    friend class Proxy;
};


/**
  * This class is a function object used to compare two TimerEvent objects on the basis of their time values.
  */
class TimerEventComparator {
public:
    int operator() (TimerEvent* &x, TimerEvent* &y);
};


#endif
