/**
  * \file Tunnel.cpp
  * \brief Implementation of functions of the Packet Capturing 
  * tunnel device (both windows and Linux).
  *
  */

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <signal.h>

#if !defined(_WIN32) 
	#include <unistd.h>
	#include <sys/ioctl.h>
	#include <sys/socket.h>
	#include <netinet/in.h>
	#include <arpa/inet.h>
	#include <net/if.h>
#endif

#ifdef __APPLE__
#	include "appleSocket.h"
#	include <netinet/ip.h>
#endif


#if !defined(_WIN32) && !defined(__APPLE__)
    #include <linux/if_packet.h>
    #include <linux/if_ether.h>

extern "C" {
	#include "netwrap.h"
}
#endif

#include "common.h"

#if defined (_WIN32) 
	#include <windows.h>
	#include <winioctl.h>
	#include <iphlpapi.h>
	#include <iprtrmib.h>
	#include <tchar.h>
#endif

#if defined (_WIN32)
#define TAP_CONTROL_CODE(request,method) \
	CTL_CODE(FILE_DEVICE_UNKNOWN, request, method, FILE_ANY_ACCESS)

//  CTL_CODE (FILE_DEVICE_PHYSICAL_NETCARD | 8000, \
//            request, method, FILE_ANY_ACCESS)

#define TAP_IOCTL_GET_VERSION           TAP_CONTROL_CODE (2, METHOD_BUFFERED)
#define TAP_IOCTL_GET_MTU               TAP_CONTROL_CODE (3, METHOD_BUFFERED)
#define TAP_IOCTL_GET_INFO              TAP_CONTROL_CODE (4, METHOD_BUFFERED)
#define TAP_IOCTL_CONFIG_POINT_TO_POINT TAP_CONTROL_CODE (5, METHOD_BUFFERED)
#define TAP_IOCTL_SET_MEDIA_STATUS      TAP_CONTROL_CODE (6, METHOD_BUFFERED)
#define TAP_IOCTL_CONFIG_DHCP_MASQ      TAP_CONTROL_CODE (7, METHOD_BUFFERED)
#define TAP_IOCTL_GET_LOG_LINE          TAP_CONTROL_CODE (8, METHOD_BUFFERED)
#define TAP_IOCTL_CONFIG_DHCP_SET_OPT   TAP_CONTROL_CODE (9, METHOD_BUFFERED)

 #define REG_CONTROL_NET "SYSTEM\\CurrentControlSet\\Control\\Network\\{4D36E972-E325-11CE-BFC1-08002BE10318}"
 #define REG_INTERFACES_KEY "SYSTEM\\CurrentControlSet\\Services\\Tcpip\\Parameters\\Interfaces"


#define USERMODEDEVICEDIR "\\\\.\\"
#define TAPSUFFIX ".tap"


#elif !defined( __APPLE__ )
  
    #include <linux/if_tun.h>

#endif


// The following includes are required only when the proxy is to be started on a planetlab
// node with Alpine4Linux user level network stack.
#ifndef _WIN32
    #include <fcntl.h>
    #include <sys/types.h>
    #include <sys/stat.h>
#endif
// End planetlab


#include "debug.h"
#include "Tunnel.h"
#include "Proxy.h"
#include "IPPacket.h"
#include "DNSPacket.h"
#include "VirtualNode.h"
#include "FakeAddressManager.h"
#include "utils/utilFunctions.h"
#include "utils/InetFns.h"
#include "utils/StringTokenizer.h"

#ifdef __APPLE__
	unsigned int get_prefix_len(char *prefix); /* XXXX */
#endif

#ifdef _WIN32

	string tunDNSServerStr;

	// These functions are outside any class because their pointers are to be
	// passed to set_terminate and on_exit.
	/** Map to hold the original (not tunDNSServerAddr) DHCPNameServers of the various interfaces.
	 */
	map <string, string> interface_NameServer_map;
#define NAME_SERVER_FIELD_NAME "NameServer"
#endif

/**
  * Constructs a packet capturing tunnel.
  * @param tunIPAddress The IP address to be set for the tunnel.
  * If this is NULL, the value in the config file is used.
  */
Tunnel::Tunnel(Proxy *proxy, IPAddress* tunIPAddress) {

    this->proxy = proxy;

    readConfigValues();

    
    if (tunIPAddress != NULL) {
        // This will override whatever was read from config file.
        strcpy (this->tunIPAddr, tunIPAddress->c_str());
        PRINT_DEBUG1 (DEBUG_LEVEL_VERBOSE,
                "Tunnel IP Address = %s\n", tunIPAddress->c_str());
    }
   
    delete (tunIPAddress);

}

void Tunnel::setProxy(Proxy* proxy) {
    this->proxy = proxy;
}

void Tunnel::readConfigValues() {

    proxy->configFile->readStringParam(XPATH_TUNNEL_DEVICE_PATH,
                            tunDevPath, DEFAULT_TUNNEL_DEVICE_PATH);
    
    proxy->configFile->readStringParam(XPATH_TUNNEL_DEVICE_NAME,
                            tunDevName, DEFAULT_TUNNEL_DEVICE_NAME);
   
    proxy->configFile->readStringParam(XPATH_TUNNEL_IP_ADDRESS,
                            tunIPAddr, DEFAULT_TUNNEL_IP_ADDRESS);
    
    proxy->configFile->readStringParam(XPATH_TUNNEL_ROUTING_TABLE,
                            tunTabName, DEFAULT_TUNNEL_ROUTING_TABLE);
    
    proxy->configFile->readStringParam(XPATH_TUNNEL_MTU,
                            tunMTU, DEFAULT_TUNNEL_MTU);

#ifdef __APPLE__
    appleCapturePort = DEFAULT_APPLE_CAPTURE_PORT; 
    proxy->configFile->readUShortParam(XPATH_TUNNEL_APPLE_CAPTURE_PORT,
                            &appleCapturePort, true);
    PRINT_DEBUG (DEBUG_LEVEL_VERBOSE, "\n\n\t\tApple Capture Port = %u\n\n\n", appleCapturePort);
#endif 

#if !defined (_WIN32) 
	//Linux only
    proxy->configFile->readStringParam(XPATH_TUNNEL_IFCONFIG_CMD,
                            ifconfigCmd, DEFAULT_TUNNEL_IFCONFIG_CMD);
    
    proxy->configFile->readStringParam(XPATH_TUNNEL_IPTABLES_CMD,
                            iptablesCmd, DEFAULT_TUNNEL_IPTABLES_CMD);
    
    proxy->configFile->readStringParam(XPATH_TUNNEL_IP_CMD,
                            ipCmd, DEFAULT_TUNNEL_IP_CMD);
    
    proxy->configFile->readStringParam(XPATH_TUNNEL_MODPROBE_CMD,
                            modprobeCmd, DEFAULT_TUNNEL_MODPROBE_CMD);
#endif

#if defined (_WIN32)
   tunWinTAPName_defined = proxy->configFile->readStringParam(XPATH_TUNNEL_WINTAP_NAME,
                                                 tunWinTAPName, DEFAULT_TUNNEL_WINTAP_NAME);
#endif

    
}

#if ! defined (_WIN32)	// initTunnel() (Linux version)


/**
  * Initializes the packet capture device.
  * This function is specific to the Linux platform vtun device.
  */
void Tunnel::initTunnel() {

    if (planetlabFlag) {
        // This proxy is supposed to work on planetlab.
        // So we can't use the tun device.  
        // Instead, we use a FIFO to communicate between the legacy app and the proxy.
        // The proxy then processes the packets as if they had been captured by the 
        // tun device.
  
        // Create the FIFO for the app --> proxy communication
        mknod(FIFO_APP2PROXY, S_IFIFO | 0666, 0);

        // Create the FIFO for the proxy --> app communication
        mknod(FIFO_PROXY2APP, S_IFIFO | 0666, 0);
   
        // Open the app --> proxy FIFO for reading.
        // OBSOLETE: Note that the legacy app (alpine server) should first open the app --> proxy FIFO for writing
        // OBSOLETE: Otherwise there will be deadlock.
        PRINT_DEBUG (
                    DEBUG_LEVEL_MINIMAL,
                    "Waiting for Alpine Server to open app --> proxy FIFO for writing.\n"
                );
        fifo_fd_app2proxy = open(FIFO_APP2PROXY, O_RDONLY); // |O_NDELAY);
        PRINT_DEBUG (
                    DEBUG_LEVEL_MINIMAL,
                    "app --> proxy FIFO got a writer.\n"
                );

        // Now open the proxy --> app FIFO for writing
        PRINT_DEBUG (
                    DEBUG_LEVEL_MINIMAL,
                    "Waiting for Alpine Server to open proxy --> app FIFO for reading.\n"
                );
        fifo_fd_proxy2app = open(FIFO_PROXY2APP, O_WRONLY);// |O_NDELAY);
        PRINT_DEBUG (
                    DEBUG_LEVEL_MINIMAL,
                    "proxy --> FIFO got a writer.\n"
                );

        // We don't need to do anything further for initializing the tunnel.
        // Especially, we don't need the tun device.
        return;
    }  // end of if (planetlabFlag)
    
    // If functioning as server proxy and we have our own NAT,
    // don't do anything
    if (proxy->mySoftwareNATFlag) {
        if (proxy->functionAsServerProxy) {
            return;
        }
    }
    
    // If the proxy is functioning as a middle box, 
    // we don't need to initialize the tunnel because
    // we are just sending and receiving packets over the 
    // network - no legacy apps here.
    // Hence we don't also need root permissions.
    if ( proxy->functionAsMiddleBox) {
        return;
    }


    char  commandBuf[CMD_BUF_LEN];  //temp buf space to construct command
#ifndef __APPLE__    
    sprintf(commandBuf, "%s tun", modprobeCmd);
    //printf("Running %s\n",commandBuf);
    system(commandBuf); /* Flawfinder: ignore */
#endif

#ifdef __APPLE__
#   define APPLE_TUN_KEXT "/Library/Extensions/tun.kext"
    // Check if the kernel module has been installed
    {
        FILE *tmpFD = fopen (APPLE_TUN_KEXT, "r");
        if (tmpFD == NULL) {
            PRINT_INFO0 (	INFO_LEVEL_FATAL_ERROR, 
				"tun kernel extension not installed in " APPLE_TUN_KEXT ". Aborting.\n"
			)
	    EXIT_ON_ERROR;
        } else {
	   fclose (tmpFD);
        }    
    }
    // Load the tun kernel module
    sprintf (commandBuf, "kextload " APPLE_TUN_KEXT);
    system(commandBuf);
#endif

    if ((tunFD = open(tunDevPath, O_RDWR)) < 0) {
        
        
        PRINT_INFO2 (INFO_LEVEL_FATAL_ERROR, "Opening %s failed (%s)!\n", tunDevPath, strerror (errno));
        close(tunFD);
        EXIT_ON_ERROR;
    }
  
#ifndef __APPLE__
    struct ifreq   ifr;
    memset(&ifr, 0, sizeof(ifr));
    ifr.ifr_flags = IFF_TUN | IFF_NO_PI;
    strncpy(ifr.ifr_name, tunDevName, 4);

    if (ioctl(tunFD, TUNSETIFF, (void *) &ifr) < 0) {
        close(tunFD);
        
        PRINT_INFO1 (INFO_LEVEL_FATAL_ERROR , "Error while configuring tunnelling device %s.\n", tunDevName);
        PRINT_DEBUG1 (DEBUG_LEVEL_FATAL, "Error in ioctl on device %s.\n", tunDevName);
        EXIT_ON_ERROR;
    }
#endif
 
#ifdef __APPLE__   
    sprintf(commandBuf, 
            "%s %s %s %s mtu %s up", 
            ifconfigCmd, tunDevName, tunIPAddr, tunIPAddr, tunMTU);
#else
    //Linux
    sprintf(commandBuf, 
            "%s %s %s pointopoint %s mtu %s up", 
            ifconfigCmd, tunDevName, tunIPAddr, tunIPAddr, tunMTU);

#endif
    
    PRINT_DEBUG1(DEBUG_LEVEL_VERBOSE, "Executing command: %s\n", commandBuf);
    //int retVal = 
    system(commandBuf); /* Flawfinder: ignore */


    /** TODO:  How to know if the ifconfig succeeded or not?  The
     * following does not work.
    if ( ! retVal ) {
        PRINT_INFO1 (INFO_LEVEL_FATAL_ERROR, "Error while configuring tunnelling device %s.\n", tunDevName);
        PRINT_DEBUG1 (DEBUG_LEVEL_FATAL, "Error in ifconfig on device %s.\n", tunDevName);
        EXIT_ON_ERROR;
    }
    */
    
    // rp_filtering START
    // Now we turn of rp_filtering.
    // This fix is needed for the proxy to correctly send packets to the legacy applications in Debian Linux.
    // Problem pointed out by Lev Popov (MIT) for Ubuntu Linux.
    // Solution by Jayanth
    char rp_filter_fileName[1000];
    sprintf (rp_filter_fileName, "/proc/sys/net/ipv4/conf/%s/rp_filter", tunDevName);
    printf ("---> %s\n", rp_filter_fileName);
    // First check if this file exists.
    FILE* tmp_rp_filter_file = fopen (rp_filter_fileName, "r");
    if (tmp_rp_filter_file != NULL) {
	    // The file does exist.  We put 0 into it.
	    fclose (tmp_rp_filter_file);
	    tmp_rp_filter_file = fopen (rp_filter_fileName, "w");
	    if (tmp_rp_filter_file != NULL) {
	        fprintf (tmp_rp_filter_file, "0\n");
	        fclose (tmp_rp_filter_file);	
	    } else {
	        PRINT_INFO1 ( INFO_LEVEL_FATAL_ERROR,
			            "Unable to write to file %s.\n",
			            rp_filter_fileName
			        );
	        EXIT_ON_ERROR;
	    }
    } else {
	    // the file does not exist. We do not do anything.
    }
    // rp_filtering STOP

#ifdef __APPLE__
    initCaptureSocket();
#endif
    initIPTables();
}

#endif  //initTunnel() (Linux version)

#if defined (_WIN32)		// initTunnel() (Windows version)
void Tunnel::initTunnel() {


    // If the proxy is functioning as a middle box, 
    // we don't need to initialize the tunnel because
    // we are just sending and receiving packets over the 
    // network - no legacy apps here.
    // Hence we don't also need root permissions.
    if ( proxy->functionAsMiddleBox) {
        return;
    }


  tunFD = INVALID_HANDLE_VALUE;

  HKEY key;
  int enum_index;
  char devid[1024], devname[1024];
  DWORD len;

  LONG ret;
  char connection_string[1024];
  HKEY connection_key;
  DWORD reg_type;
  char reg_data[1024];

  if(RegOpenKeyEx(HKEY_LOCAL_MACHINE, REG_CONTROL_NET, 0, KEY_READ, &key)) {
    PRINT_INFO0(INFO_LEVEL_FATAL_ERROR, "Unable to initialize packet capturing tunnel: Cannot to read registry.\n");
	EXIT_ON_ERROR;
  }

  // search an adapter with TAPSUFFIX

  enum_index = 0;
  while (1) {
    len = sizeof(devid);
    if(RegEnumKeyEx(key, enum_index, devid, &len, 0, 0, 0, NULL) != ERROR_SUCCESS) {
      RegCloseKey(key);
	PRINT_INFO0(INFO_LEVEL_FATAL_ERROR, "Unable to initialize packet capturing tunnel.\n");
	EXIT_ON_ERROR;
    }

    if (tunWinTAPName_defined) {
      snprintf(connection_string, sizeof(connection_string),
              "%s\\%s\\Connection",
              REG_CONTROL_NET, devid);
      ret = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                        connection_string,
                        0,
                        KEY_READ,
                        &connection_key);
      if (ret != ERROR_SUCCESS) {
       enum_index++;
       continue;
      }

      ret = RegQueryValueEx(connection_key, "Name", NULL, &reg_type, (BYTE *)reg_data, &len);
      if (ret != ERROR_SUCCESS || reg_type != REG_SZ) {
       enum_index++;
       continue;
      }

      if (strcmp(tunWinTAPName, reg_data) != 0) {
       enum_index++;
       continue;
      }
    }

    snprintf(devname, sizeof(devname), USERMODEDEVICEDIR "%s" TAPSUFFIX, devid);
    tunFD = CreateFile(devname, GENERIC_WRITE | GENERIC_READ, 0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_SYSTEM, 0);
    
    if(tunFD != INVALID_HANDLE_VALUE) {
      RegCloseKey(key);
      CloseHandle(tunFD);
      break;
    }
    enum_index++;
  }
  // Open TAP-Win32 device
  tunFD = CreateFile(devname, GENERIC_WRITE|GENERIC_READ,  0,  0,
		      OPEN_EXISTING, FILE_ATTRIBUTE_SYSTEM | FILE_FLAG_OVERLAPPED, 0);
  if (tunFD == INVALID_HANDLE_VALUE) {
    PRINT_INFO0(INFO_LEVEL_FATAL_ERROR, "Unable to initialize packet capturing tunnel:Could not open Windows tap device.\n");
	EXIT_ON_ERROR;
  }

  // set TAP-Win32 device in TUN mode
  {
    in_addr_t ep[2];
    ep[0] = inet_addr(tunIPAddr);
    ep[1] = ep[0] ^ inet_addr("0.0.0.3");
    if (!DeviceIoControl (tunFD, TAP_IOCTL_CONFIG_POINT_TO_POINT,
			  ep, sizeof (ep),
			  ep, sizeof (ep), &len, NULL)) {
      PRINT_INFO0(INFO_LEVEL_FATAL_ERROR, "Unable to initialize packet capturing tunnel The TAP-Win32 driver rejected a DeviceIoControl call to set Point-to-Point mode\n");
	EXIT_ON_ERROR;
    }
  }

  // address configuration
  {
    in_addr_t ep[4];
	//TODO remove the constants
    ep[0] = inet_addr(tunIPAddr); // IP address
    ep[1] = inet_addr("255.255.255.252"); // netmask
    ep[2] = ep[0] ^ inet_addr("0.0.0.3"); // remote IP address
    ep[3] = 31536000; // DHCP lease time (1year)
    
    if (!DeviceIoControl (tunFD, TAP_IOCTL_CONFIG_DHCP_MASQ,
			  ep, sizeof (ep),
			  ep, sizeof (ep), &len, NULL)) {
      PRINT_INFO0(INFO_LEVEL_FATAL_ERROR, "Unable to initialize packet capturing tunnel: Failed to configure addresses.\n");
	EXIT_ON_ERROR;
    }
  }

  {
    uint8_t dhcp_options[6];
//    uint8_t u8;
    tunDNSServerAddr = inet_addr(tunIPAddr) ^ inet_addr("0.0.0.3");
	struct in_addr tmpAddr;
	tmpAddr.s_addr = tunDNSServerAddr;
	tunDNSServerStr = string (inet_ntoa (tmpAddr));

    dhcp_options[0] = (uint8_t) 6;
    dhcp_options[1] = (uint8_t) 4;
    memcpy(dhcp_options + 2, &tunDNSServerAddr, 4);

    if (!DeviceIoControl (tunFD, TAP_IOCTL_CONFIG_DHCP_SET_OPT,
                         &dhcp_options, 6,
                         &dhcp_options, 6, &len, NULL)) {
      PRINT_INFO0(INFO_LEVEL_FATAL_ERROR, "Unable to initialize packet capturing tunnel: Failed to set dhcp options.\n");
      EXIT_ON_ERROR;
    }
  }

  
  ULONG status = TRUE;
  if (!DeviceIoControl (tunFD, TAP_IOCTL_SET_MEDIA_STATUS,
			&status, sizeof (status),
			&status, sizeof (status), &len, NULL)) {
    PRINT_INFO0(INFO_LEVEL_FATAL_ERROR, "Unable to initialize packet capturing tunnel:Failed to set TAP-Win32 status as 'connected'.\n");
	EXIT_ON_ERROR;
  }
  
  Sleep(100);

    //ASSERT: tunFD should be a valid file handle at this point.
    initIPTables();

	initNameServerRedirector ();

}

void Tunnel::initNameServerRedirector () {

	nameServerUpdate ();

	atexit (restoreNameServerSettings);
	set_terminate (restoreNameServerSettings);
	signal (SIGTERM, restoreNameServerSettings_intarg);
	signal (SIGINT, restoreNameServerSettings_intarg);

	if (SetConsoleCtrlHandler((PHANDLER_ROUTINE)restoreNameServerSettings_CtrlHandler, TRUE) == 0) {
		PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL,
				"Unable to set console handler to restore name server settings on exit.\n"
			);
	} else {
		PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL,
				"Set console handler to restore name server settings on exit.\n"
				);
	}
}

void do_netsh_setdns (char* interfaceName, char* dnsCmd) {
	STARTUPINFO si;
    PROCESS_INFORMATION pi;
	char netshSetTunDNSCmd [1000];
	sprintf (netshSetTunDNSCmd, "netsh interface ip set dns \"%s\" %s",
				interfaceName, dnsCmd);
    LPTSTR szCmdline=_tcsdup(TEXT(netshSetTunDNSCmd));

	PRINT_DEBUG1 (DEBUG_LEVEL_VERBOSE, 
					"Executing netsh command: %s.\n",
					netshSetTunDNSCmd
					);
    ZeroMemory( &si, sizeof(si) );
    si.cb = sizeof(si);
	si.dwFlags = STARTF_USESHOWWINDOW;
	// set the window display to HIDE
	si.wShowWindow = SW_HIDE;
	ZeroMemory( &pi, sizeof(pi) );

    // Start the child process. 
    if( !CreateProcess( NULL,   // No module name (use command line)
        szCmdline,      // Command line
        NULL,           // Process handle not inheritable
        NULL,           // Thread handle not inheritable
        FALSE,          // Set handle inheritance to FALSE
        0,              // No creation flags
        NULL,           // Use parent's environment block
        NULL,           // Use parent's starting directory 
        &si,            // Pointer to STARTUPINFO structure
        &pi )           // Pointer to PROCESS_INFORMATION structure
    ) {
		PRINT_DEBUG3 ( DEBUG_LEVEL_MINIMAL, "CreateProcess failed while doing netsh %s %s :(%d).\n", 
						interfaceName,
						dnsCmd,
						GetLastError() 
					);
        return;
    }

    // Wait until child process exits.
    WaitForSingleObject( pi.hProcess, INFINITE );

    // Close process and thread handles. 
    CloseHandle( pi.hProcess );
    CloseHandle( pi.hThread );
}

#endif

#ifndef _WIN32	// initIPTables() (Linux version)
/**
  * Initialize iptables to capture packets from legacy applications and
  * send them across to the tunnel.
  *
  * TODO : This part should be configurable - specify which apps to be
  * tunnelled, etc.
  */
void Tunnel::initIPTables() {

    saveIPTables();

    char commandBuf[CMD_BUF_LEN];

//#define SERVER_PROXY 1
#ifdef __APPLE__    

	/* remove zombie entry (left here by a previous proxy instance) */
    system("ipfw delete 333");   

    /* forward all DNS requests sent by proxy to a legacy DNS server */ 
	sprintf(commandBuf, "%s add 333 accept udp from any %u to any %u out",
		"ipfw", proxy->dnsRequestPort, proxy->stdDNSRequestPort);
    system(commandBuf);   
	
	/* capture all DNS requests sent by processes (other than proxy) and send them to proxy */
	//ION sprintf(commandBuf, "%s add 333 forward %s,%u udp from any to any %u out",
	//ION	"ipfw", "127.0.0.1", appleCapturePort, proxy->stdDNSRequestPort); /* XXX */
	sprintf(commandBuf, "%s add 333 divert %d udp from any to any %u out",
		"ipfw", appleCapturePort, proxy->stdDNSRequestPort); /* XXX */
    system(commandBuf);   

	/* capture all packets sent to the tun device and forward them to proxy */
	sprintf(commandBuf, "%s add 333 divert %d udp from any to %s/%u out",
		"ipfw", appleCapturePort, proxy->fakeAddrManager->fakeAddrRangeStart, 
		get_prefix_len(proxy->fakeAddrManager->fakeAddrRangeMask));
    system(commandBuf);   
        /* capture all packets sent to the tun device and forward them to proxy */
        sprintf(commandBuf, "%s add 333 divert %d icmp from any to %s/%u out",
                "ipfw", appleCapturePort, proxy->fakeAddrManager->fakeAddrRangeStart,
                get_prefix_len(proxy->fakeAddrManager->fakeAddrRangeMask));
    system(commandBuf);

        sprintf(commandBuf, "%s add 333 divert %d tcp from any to %s/%u out",
                "ipfw", appleCapturePort, proxy->fakeAddrManager->fakeAddrRangeStart,
                get_prefix_len(proxy->fakeAddrManager->fakeAddrRangeMask));
    system(commandBuf);

	/* accept all IP and ICMP packets sent by the tun device---forward them to intended application */
	sprintf(commandBuf, "%s add 333 allow ip from %s/%u to any",
		"ipfw", proxy->fakeAddrManager->fakeAddrRangeStart, get_prefix_len(proxy->fakeAddrManager->fakeAddrRangeMask));
    system(commandBuf);   
	sprintf(commandBuf, "%s add 333 allow icmp from %s/%u to any",
		"ipfw", proxy->fakeAddrManager->fakeAddrRangeStart, get_prefix_len(proxy->fakeAddrManager->fakeAddrRangeMask));
    system(commandBuf);   

	sprintf(commandBuf, "%s add 333 allow tcp from %s/%u to any",
		"ipfw", proxy->fakeAddrManager->fakeAddrRangeStart, get_prefix_len(proxy->fakeAddrManager->fakeAddrRangeMask));
    system(commandBuf);   

#else /* __APPLE__ */



    sprintf(commandBuf, "%s -F -t mangle", iptablesCmd); /* Flawfinder: ignore */
    //printf("Running %s\n",commandBuf);
    system(commandBuf);     /* Flawfinder: ignore */
   
    sprintf(commandBuf, "%s -F -t nat", iptablesCmd);
    //printf("Running %s\n",commandBuf);    
    system(commandBuf);

    if ( ! proxy->functionAsServerProxy) {
        /* Flawfinder: ignore */
        /** Add a rule to blindly accept any DNS requests the proxy makes from its 
          * specified DNS request src port.
          */
        sprintf(commandBuf, "%s -A OUTPUT -t mangle  -p udp --sport %u --dport %u -j ACCEPT", 
                        iptablesCmd, 
                        proxy->dnsRequestPort,
                        proxy->stdDNSRequestPort);
	    //printf("Running %s\n",commandBuf);    
        system(commandBuf);     /* Flawfinder: ignore */
 
        /* Flawfinder: ignore */
        /** Set a mark 1 on DNS requests */
        sprintf(commandBuf, "%s -A OUTPUT -t mangle  -p udp --dport 53 -j MARK --set-mark 1", 
                        iptablesCmd);
	    //printf("Running %s\n",commandBuf);    
        system(commandBuf); /* Flawfinder: ignore */

        /* Flawfinder: ignore */
        // Set a mark 1 on packets with destination address falling in the fake address region.
        sprintf(commandBuf, "%s -A OUTPUT -t mangle  -d %s/%s -j MARK --set-mark 1",
                        iptablesCmd,
                        proxy->fakeAddrManager->fakeAddrRangeStart,
                        proxy->fakeAddrManager->fakeAddrRangeMask);
	    //printf("Running %s\n",commandBuf);    
        system(commandBuf); /* Flawfinder: ignore */

        initMatchRules();

        // TODO Add the rule to set the mark on IP addresses and port numbers specified
        // in the config file.
        // TODO In the config file, specify the protocol in addition to port number for the
        // Match element.
        
        /* Flawfinder: ignore */
        sprintf(commandBuf, "%s -A OUTPUT -t mangle -j ACCEPT", iptablesCmd);
	    //printf("Running %s\n",commandBuf);    
        system(commandBuf); /* Flawfinder: ignore */
 
        // ip rule add fwmark 1 table i3
	// sprintf(commandBuf, "%s rule add fwmark 1 table %s", ipCmd, tunTabName); /* Flawfinder: ignore */	
        // Add a forwarding rule so that packets with mark 1 (set for packets with dstIP = fake address
        // and DNS requests to port 53) to use the forwarding table 200.
        sprintf(commandBuf, "%s rule add fwmark 1 table 200", ipCmd); /* Flawfinder: ignore */	
	//printf("Running %s\n",commandBuf);    
        system(commandBuf); /* Flawfinder: ignore */
    }    
     
    // ip route add default table i3_mobility dev tun0
    // This rule specifies the default forwarding rule for the tun table (200)
    // as "send to the tun device"
    // TODO: In this rule the table is specified by tunTabName, in the previous name it is 200.
    // QQQ: Should we change it?
    sprintf(commandBuf, "%s route add default table %s dev %s", ipCmd,
                          tunTabName, tunDevName); /* Flawfinder: ignore */
    //printf("Running %s\n",commandBuf);
    system(commandBuf);   /* Flawfinder: ignore */
   
   
    if ( proxy->functionAsServerProxy) {
    
        sprintf(commandBuf,"echo 1 > /proc/sys/net/ipv4/ip_forward");
        system(commandBuf);
    
        sprintf(commandBuf,"%s -t nat -A POSTROUTING --source 127.0.0.1 -j ACCEPT",iptablesCmd); 
        system(commandBuf);

        char *myIP = inet_ntoa (proxy->my_addr_nf);
        sprintf (   commandBuf, 
                  "%s -t nat -A POSTROUTING --source ! %s -j SNAT --to-source %s",
                    iptablesCmd,myIP, myIP
                );
        system(commandBuf);
    }

    {
        // Add the route which says that all fake addresses should go through the tunnel.
        struct in_addr route;
        route.s_addr = inet_addr(proxy->fakeAddrManager->fakeAddrRangeStart) & inet_addr(proxy->fakeAddrManager->fakeAddrRangeMask);
        char *fake_addr_route = inet_ntoa(route);
    
        sprintf(commandBuf, "/sbin/route add -net %s netmask %s dev %s", fake_addr_route,
	        proxy->fakeAddrManager->fakeAddrRangeMask, tunDevName);
        //printf("Command: %s\n",commandBuf);
        system(commandBuf);
    }
#endif //__APPLE__
 
}
#endif //initIPTables() (Linux version)

#ifndef _WIN32 
/**
  * This function saves the current iptables configuration.
  * It is restored by calling restoreIPTables.
  */
void Tunnel::saveIPTables() {

    PRINT_DEBUG (DEBUG_LEVEL_VERBOSE,
                "Saving iptables configuration to %s.\n",
                IPTABLES_SAVE_FILE
            );
    system ("/sbin/iptables-save -t mangle > " IPTABLES_SAVE_FILE "_mangle");
    system ("/sbin/iptables-save -t filter > " IPTABLES_SAVE_FILE "_filter");

}
/**
  * This function restores iptables settings to those saved in
  * saveIPTables.
  */
void Tunnel::restoreIPTables () {
    
    PRINT_DEBUG (DEBUG_LEVEL_VERBOSE,
                "Restoring iptables configuration from %s_*.\n",
                IPTABLES_SAVE_FILE
            );
    system ("/sbin/iptables-restore < " IPTABLES_SAVE_FILE "_mangle");
    system ("/sbin/iptables-restore < " IPTABLES_SAVE_FILE "_filter");

}
#endif

#ifdef _WIN32  //initIPTables() (Windows version)
/**
 *  This function sets up the tunnel to capture DNS packets
 *  packets sent/recd to fake IP addresses.  In Linux, this
 *  process is achieved through iptables.  Hence the name for 
 *  this function.
 */ 
void Tunnel::initIPTables()
{
  in_addr_t fake_start_addr = inet_addr(proxy->fakeAddrManager->fakeAddrRangeStart);
  in_addr_t mask = inet_addr(proxy->fakeAddrManager->fakeAddrRangeMask);
  in_addr_t tun_addr = inet_addr(tunIPAddr);
  in_addr_t tun_mask = inet_addr("255.255.255.252");  //TODO
  if ((tun_addr & tun_mask) == (fake_start_addr & tun_mask)) {
    PRINT_INFO0(INFO_LEVEL_FATAL_ERROR, "FAKE_ADDR_RANGE_START should not be within the subnet of the tun interface\n");
    EXIT_ON_ERROR;

    /* TODO Verify if this is necessary
  } else if ((tun_addr & mask) == (fake_start_addr & mask) && tun_addr > fake_start_addr) {
    PRINT_INFO0(INFO_LEVEL_FATAL_ERROR, "FAKE_ADDR_RANGE_START should be greater than TUN_DEV_ADDRESS\n");
	EXIT_ON_ERROR;
*/
  }

  // wait for tun_dev to be up before calling CreateIpForwardEntry
  MIB_IPFORWARDROW bestroute;
  for (;;) {
    if (GetBestRoute(tun_addr ^ (~tun_mask), 0, &bestroute) != NO_ERROR){
      PRINT_INFO0(INFO_LEVEL_FATAL_ERROR, "Unable to initialise packet capturing tunnel:add_fake_route(): GetBestRoute failed\n");
	EXIT_ON_ERROR;
    }
    if (bestroute.dwForwardNextHop == tun_addr)
      break;
    Sleep(100);
  }

  MIB_IPFORWARDROW route;
  memset(&route, 0, sizeof(MIB_IPFORWARDROW));
  route.dwForwardDest = fake_start_addr & mask;
  route.dwForwardMask = mask;
  route.dwForwardNextHop = tun_addr ^ (~tun_mask);
  if (GetBestInterface(route.dwForwardNextHop,
		       &route.dwForwardIfIndex) != NO_ERROR) {
    PRINT_INFO0(INFO_LEVEL_FATAL_ERROR, "Unable to initialize packet capturing tunnel : add_fake_route(): GetBestInterface failed\n");
	EXIT_ON_ERROR;
  }
  route.dwForwardType = 4;
  route.dwForwardProto = 3;
  
  route.dwForwardMetric1 = 0;
  route.dwForwardMetric2 = 0;
  route.dwForwardMetric3 = 0;
  route.dwForwardMetric4 = 0;
  route.dwForwardMetric5 = 0;
  route.dwForwardPolicy = 0;

  int retVal;
  if ( (retVal = CreateIpForwardEntry(&route)) != NO_ERROR) {
    PRINT_INFO0(INFO_LEVEL_FATAL_ERROR,"Unable to initialize packet capturing tunnel: add_fake_route(): CreateIpForwardEntry failed.\n");

    	char *error_buf;
	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |
			  FORMAT_MESSAGE_IGNORE_INSERTS,NULL, retVal,
			  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),(LPTSTR) &error_buf,0,NULL);
	PRINT_INFO1 (INFO_LEVEL_FATAL_ERROR, "%s\n", error_buf);
	EXIT_ON_ERROR;
  }
  return;
}

#endif	//initIPTables() (Windows version)


#if !defined (_WIN32)

/**
  * Adds ip tables rules corresponding to the port/ip addresses
  * matches in dstPortList, dstIPAddrList, srcPortList, srcIPAddrList.
  */
void Tunnel::initMatchRules() {

    list<ipaddr_n_t>::iterator dstIPItr = dstIPAddrList.begin();
    while (dstIPItr != dstIPAddrList.end()) {
        addPacketCaptureByDstIP (*dstIPItr);
        dstIPItr++;
    }
    
    list<ipaddr_n_t>::iterator srcIPItr = srcIPAddrList.begin();
    while (srcIPItr != srcIPAddrList.end()) {
        addPacketCaptureBySrcIP (*srcIPItr);
        srcIPItr++;
    }

    list<port_protocol_n_t>::iterator dstPortItr = dstPortList.begin();
    while (dstPortItr != dstPortList.end()) {
        addPacketCaptureByDstPort (*dstPortItr);
        dstPortItr++;
    }

    list<port_protocol_n_t>::iterator srcPortItr = srcPortList.begin();
    while (srcPortItr != srcPortList.end()) {
        addPacketCaptureBySrcPort (*srcPortItr);
        srcPortItr++;
    }

    
}

/**
  * This function instructs the tunnel to capture all packets
  * sent to the specified destination IP address (in network byte order).
  */
void Tunnel::addPacketCaptureByDstIP (ipaddr_n_t dstIPAddr) {
    
    char commandBuf[CMD_BUF_LEN];
    
    char tmpDstIPAddrStr[20];
    
    // convert the dstIPAddr into A.B.C.D notation
    if (convert_ipaddr_int2str (dstIPAddr, tmpDstIPAddrStr)) {
    
        /* Flawfinder: ignore */
        // Set a mark 1 on packets with the specified destination address.
        // We would have already put rules for capturing marked packets in
        // the initTunnel function.
        sprintf(commandBuf, "%s -A OUTPUT -t mangle  -d %s -j MARK --set-mark 1",
                        iptablesCmd,
                        tmpDstIPAddrStr);
	    printf("\n\n\n\nRunning %s\n",commandBuf);    
        system(commandBuf); /* Flawfinder: ignore */
    
    } else {
        PRINT_INFO1 (
                    INFO_LEVEL_WARNING,
                    "Tunnel:: Unable to add packet capture for destination ip adddress %d.\n",
                    dstIPAddr
                );
    }
 
}

/**
  * This function instructs the tunnel to capture all packets
  * sent to the specified source IP address (in network byte order).
  */
void Tunnel::addPacketCaptureBySrcIP (ipaddr_n_t srcIPAddr) {
    
    char commandBuf[CMD_BUF_LEN];
    
    char tmpSrcIPAddrStr[20];
    
    // convert the srcIPAddr into A.B.C.D notation
    if (convert_ipaddr_int2str (srcIPAddr, tmpSrcIPAddrStr)) {
    
        /* Flawfinder: ignore */
        // Set a mark 1 on packets with the specified destination address.
        // We would have already put rules for capturing marked packets in
        // the initTunnel function.
        sprintf(commandBuf, "%s -A OUTPUT -t mangle  -s %s -j MARK --set-mark 1",
                        iptablesCmd,
                        tmpSrcIPAddrStr);
	    //printf("\n\n\n\nRunning %s\n",commandBuf);    
        system(commandBuf); /* Flawfinder: ignore */
    
    } else {
        PRINT_INFO1 (
                    INFO_LEVEL_WARNING,
                    "Tunnel:: Unable to add packet capture for source ip adddress %d.\n",
                    srcIPAddr
                );
    }
 
}


void Tunnel::addPacketCaptureBySrcPort (port_protocol_n_t port) {
    char commandBuf[CMD_BUF_LEN];
    
    char tmpProtStr[20];
    if (port.protocol == PROTOCOL_UDP) {
        strcpy (tmpProtStr, "udp");
    } else {
        strcpy (tmpProtStr, "tcp");
    }

    /* Flawfinder: ignore */
    // Set a mark 1 on packets with the specified destination address.
    // We would have already put rules for capturing marked packets in
    // the initTunnel function.
    sprintf(commandBuf, "%s -A OUTPUT -t mangle  -p %s --source-port %d -j MARK --set-mark 1",
                        iptablesCmd,
                        tmpProtStr,
                        ntohs (port.port));
	//printf("\n\n\n\nRunning %s\n",commandBuf);    
    system(commandBuf); /* Flawfinder: ignore */
    
}
void Tunnel::addPacketCaptureByDstPort (port_protocol_n_t port) {
    char commandBuf[CMD_BUF_LEN];
    
    char tmpProtStr[20];
    if (port.protocol == PROTOCOL_UDP) {
        strcpy (tmpProtStr, "udp");
    } else {
        strcpy (tmpProtStr, "tcp");
    }

    /* Flawfinder: ignore */
    // Set a mark 1 on packets with the specified destination address.
    // We would have already put rules for capturing marked packets in
    // the initTunnel function.
    sprintf(commandBuf, "%s -A OUTPUT -t mangle  -p %s --destination-port %d -j MARK --set-mark 1",
                        iptablesCmd,
                        tmpProtStr,
                        ntohs (port.port));
	printf("\n\n\n\nRunning %s\n",commandBuf);    
    system(commandBuf); /* Flawfinder: ignore */
 
}
#else 


#endif

/**
  * Writes the contents of buf into the tunnel.
  * @param buf the contents of the packet.
  * @param len The length of the packet.
  */
void Tunnel::writeToTunnel(char *buf, int len) {
        IPPacket captIPPkt (len, buf);
        captIPPkt.freeMemFlag = false; // This is important, otherwise there will be a double free
        // Moreover, we did not allocate the buf.  So we do not free it.
        char tmpIPAddrBuf1[MAX_IP_ADDR_LEN];
        char tmpIPAddrBuf2[MAX_IP_ADDR_LEN];
        
        PRINT_DEBUG3( DEBUG_LEVEL_VERBOSE, "\tWrote to legacy application bytes%d, srcIP=%s destIP=%s\n\n",
                        len,
                        captIPPkt.getSrcIPAddress(tmpIPAddrBuf1),
                        captIPPkt.getDestIPAddress(tmpIPAddrBuf2));
 
#if defined (_WIN32)

    DWORD lenin;
    OVERLAPPED overlapped = {0};
    if (!WriteFile(tunFD, buf, len, &lenin, &overlapped))  {
         PRINT_DEBUG0(DEBUG_LEVEL_MINIMAL, "Failed to write packet to tun device\n");
    }

#elif defined(__APPLE__)

    char *enpacket = buf;
    struct sockaddr_in dst;
    struct ip *ip;
    int rc;


    ip = (struct ip *)enpacket;

    bzero(&dst, sizeof(dst));
#ifdef RON_HAS_STRUCT_SINLEN
    dst.sin_len = sizeof(dst);
#endif
    dst.sin_family = AF_INET;
    dst.sin_addr = ip->ip_dst;

    /*
     * This next bit is to counteract some BSD stupidity.
     * The kernel insists on fixing up the IP len and IP
     * offset fields, so it wants them in host byte order...
     */
    ip->ip_sum = 0;
    ip->ip_len = ntohs(ip->ip_len);
    ip->ip_off = ntohs(ip->ip_off);  /* Gag me with a spoon */
    ip->ip_ttl--;   /* Make sure we don't permit routing loops */
    
    if ((rc = sendto(appleOutSocket, enpacket, len, 0,
		     (struct sockaddr *)&dst, sizeof(dst))) <= 0) {
	perror("ip_conduit::recv could not sendto");
	assert (false);
    }
    PRINT_DEBUG (DEBUG_LEVEL_VERBOSE, "ip_conduit::recv sendto rc %d, msgLen=%d\n", rc, len);
    
#else
    //Linux version
#   ifdef BENCHMARK
	    addCheckpoint (EVENT_DATA_OC_I_TO_TUN, 1);
#   endif

    if (planetlabFlag) {
    
        struct myEthernetHeader {
            uint8_t src[6];
            uint8_t dst[6];
            uint16_t frameType;
        };

        // Generate an ethernet header
        char frameBuf[10000];

        struct myEthernetHeader *tt = (struct myEthernetHeader *) frameBuf;
        tt->src[0] = 0x00;
        tt->src[1] = 0x22;
        tt->src[2] = 0x33;
        tt->src[3] = 0x44;
        tt->src[4] = 0x55;
        tt->src[5] = 0x66;

        // Destination is ethernet broadcast address
        // Actually, we should set the destination hwaddress to the
        // mac address of the ethernet card.
        tt->dst[0] = 0xff;
        tt->dst[1] = 0xff;
        tt->dst[2] = 0xff;
        tt->dst[3] = 0xff;
        tt->dst[4] = 0xff;
        tt->dst[5] = 0xff;

        tt->frameType = htons(0x800);
    
        int offset = sizeof (struct myEthernetHeader);

        // TODO Yet another memcpy to avoid!
        memcpy (frameBuf + offset, buf, len); 
        write(fifo_fd_proxy2app, frameBuf, len + offset);
        
        // This sleep is necessary to slow down the proxy so that Alpine4Linux network stack
        // can keep up with the processing.
        // Ugly hack by DILIP.
        struct timespec tmp_sleep_time;
        tmp_sleep_time.tv_sec = 0;
        tmp_sleep_time.tv_nsec = 100;
        nanosleep (&tmp_sleep_time, NULL);
 
    } else {
    
        // not Planetlab, use normal tun device
        write (tunFD, buf, len);
    }
#endif // End Linux version
}


/**
  * This function writes a DNS reply for the specified DNS Packet.
  * @param dnsPkt The original DNS request we are replying to.
  * @param fakeIP The fake IP address to be returned in the DNS reply.
  * If fakeIP is NULL, a DNS Not found reply is sent.
  * This function does NOT delete dnsPkt
  */
void Tunnel::answerDNSRequest(DNSPacket *dnsPkt, const IPAddress* fakeIP) {

    static char* answerBuf = NULL;
	if (answerBuf == NULL) {
		// This initialization happens only the first time the function is called.
		answerBuf = new char[MAX_PACKET_SIZE];
		memset (answerBuf, 0, MAX_PACKET_SIZE);
	}

	if (fakeIP != NULL) {
		PRINT_DEBUG2 (DEBUG_LEVEL_VERBOSE,
							"Writing DNS reply for %s (%s) to tunnel.\n",
							dnsPkt->reqURL.c_str(), fakeIP->c_str()
						);
	} else {
		PRINT_DEBUG1 (DEBUG_LEVEL_VERBOSE,
							"Writing DNS NOT found reply for %s to tunnel.\n",
							dnsPkt->reqURL.c_str()
						);
	}

    dnsPkt->getAnswer(answerBuf, fakeIP);


#ifdef BENCHMARK
	addCheckpoint (EVENT_DNS_REPLY_BEFORE_WRITE_TO_TUN, 1);
#endif

    writeToTunnel(answerBuf, dnsPkt->pktSize + 16);
    
#ifdef BENCHMARK
	addCheckpoint (EVENT_DNS_REPLY_AFTER_WRITE_TO_TUN, 1);
#endif

}

/**
  * LINUX: This function reads from tunFD and processes the captured packet appropriately.
  * This function can be called in an infinite while loop (in which case it will block when
  * no packets are available to be read) or when a select() detects read activity on tunFD
  * (in which case it will read the packet and return immediately).
  *
  * Windows: The packet was earlier captured by the separate thread listening on the tunnel.
  * The captured packet is passed as a an argument.  So there is no explicit reading from
  * the tunnel device here, nor does it block.
  */
#if defined (_WIN32)
void Tunnel::capturePacket(char* pktBuf, int pktLen) {

#else

// Linux and APPLE version
void Tunnel::capturePacket() {

        //PRINT_DEBUG0(DEBUG_LEVEL_MINIMAL, "Packet Capture loop start\n");

		// read the packet from the tun device
        int pktLen;
        char *pktBuf = new char[MAX_PKT_CAPT_LEN];
#	ifdef __APPLE__
        pktLen = read (appleCaptureSocket, pktBuf, MAX_PKT_CAPT_LEN);
#	else         
	pktLen = read (getFD(), pktBuf, MAX_PKT_CAPT_LEN);
#	endif

#endif // End Linux Version

#ifdef BENCHMARK
	//addCheckpoint (EVENT_READ_FROM_TUN, 1);
#endif

	if (proxy->stopProxyFlag) {
	    return;
	}

        if (pktLen < 0) {
            PRINT_DEBUG1(DEBUG_LEVEL_FATAL, "Read from tunnel (fd=%d) failed.\n", tunFD);
            PRINT_INFO0 (INFO_LEVEL_FATAL_ERROR, "Error while reading from packet capturing device.  Exiting.\n\n");
            EXIT_ON_ERROR;
        }
    
        IPPacket *captIPPkt = new IPPacket(pktLen, pktBuf);

#if defined (_WIN32)
		// The pktBuf is a static buffer in this case, unlike in Linux version where we dynamically 
		// allocate a buffer each time.  Since the buffer is static here, we must not free it.
		captIPPkt->freeMemFlag = false;
#endif

#ifdef BENCHMARK
	if ( ! captIPPkt->isDNSRequest()) {
    		addCheckpoint (EVENT_DATA_TUN_TO_OC_I, 1);
	} else {
		addCheckpoint (EVENT_DNS_REQUEST_FROM_TUN, 1);
	}
#endif

        char tmpIPAddrBuf1[MAX_IP_ADDR_LEN];
        char tmpIPAddrBuf2[MAX_IP_ADDR_LEN];
        
        PRINT_DEBUG3(DEBUG_LEVEL_VERBOSE, "Tunnel received packet of length %d, srcIP=%s destIP=%s\n\n",
                        pktLen,
                        captIPPkt->getSrcIPAddress(tmpIPAddrBuf1),
                        captIPPkt->getDestIPAddress(tmpIPAddrBuf2));

	// pass the captured packet to the Proxy for processing.	
	proxy->tunnelCapturedPacket(captIPPkt);

}

/**
 * This function returns the IP address assigned to the tunnel device.
 */
char* Tunnel::getTunnelIP() {
    return tunIPAddr;
}

#if defined (__APPLE__) // Apple version

/**
  * Returns the file descriptor of the socket which captures the packets.
  */
int Tunnel::getFD() {

    return appleCaptureSocket;
}


#elif !defined (_WIN32)  // Linux version
/**
  * Returns the file descriptor of the socket which captures the packets.
  */
int Tunnel::getFD() {

    if (planetlabFlag) {
        return fifo_fd_app2proxy;
    } else {
        return tunFD;
    }
}
#endif //getFD() Linux version

#if defined (_WIN32) // Windows version
/**
  * Returns the file descriptor of the socket which captures the packets.
  */
HANDLE Tunnel::getFD() {
    return tunFD;
}
#endif // getFD() Windows version

#if defined (_WIN32)
/**
 *  This function reads from the TUN device and writes to 
 *  a normal file handle, from where it can be read in as in
 *  a normal socket.  This function should be running in a
 *  thread of its own.
 */
void Tunnel::tunReader()
{
  DWORD len;
  char buf[MAX_PACKET_BUF_LEN];
  OVERLAPPED overlapped;

  overlapped.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

  while ( ! proxy->stopProxyFlag ) {
    overlapped.Offset = 0;
    overlapped.OffsetHigh = 0;
    ResetEvent(overlapped.hEvent);

	len = 0;
    int status = ReadFile(tunFD, buf, MAX_PACKET_BUF_LEN, &len, &overlapped);

    if (proxy->stopProxyFlag) {
        return;
    }

    if (!status) {
		if (GetLastError() == ERROR_IO_PENDING) {
			WaitForSingleObject(overlapped.hEvent, INFINITE);
			if (proxy->stopProxyFlag) {
				return;
			}
	
			if (!GetOverlappedResult(tunFD, &overlapped, &len, FALSE))
				continue;
		} else {
			PRINT_DEBUG0(DEBUG_LEVEL_MINIMAL,"tunreader: error while reading from tun device\n");
			continue;
		}
    }

    if (proxy->stopProxyFlag) {
        return;
    }

	
	capturePacket(buf, len);
    
	//write(readsp[0], buf, len);
  }

  PRINT_DEBUG0 (DEBUG_LEVEL_VERBOSE, "Proxy (tunReader) stopped.");
  //CloseHandle(tunFD);
  return;
}
#endif


/** Adds the specified destination port number to the list of destination ports
  * whose packets are to be captured by the tunnel.
  */
void Tunnel::addDstPortMatch (port_protocol_n_t port) {
    dstPortList.push_back (port);
}

/** Adds the specified source port number to the list of source ports
  * whose packets are to be captured by the tunnel.
  */
void Tunnel::addSrcPortMatch (port_protocol_n_t port) {
    srcPortList.push_back (port);
}

/** Adds the specified destination ipaddress to the list of destination ip addresses
  * whose packets are to be captured by the tunnel.
  */
void Tunnel::addDstIPMatch (ipaddr_n_t addr) {
    dstIPAddrList.push_back (addr);
}

/** Adds the specified source ipaddress to the list of source ip addresses
  * whose packets are to be captured by the tunnel.
  */
void Tunnel::addSrcIPMatch (ipaddr_n_t addr) {
    srcIPAddrList.push_back (addr);
}


/**
  * This function fixes replaces the source port of the specified packet with
  * the new source port.
  * @param ip_hdr Pointer to the ip header
  * @param newSrcPort The new source port to be put into the packet (in network byte order)
  */
void Tunnel::change_src_port_fix_checksum (char *ip_hdr, portnum_n_t newSrcPort) {
  
	
  	struct IPHdr* ipHeader = (struct IPHdr*) ip_hdr;
    uint8_t headerLen = (0x0F & ipHeader->vihl) * 4;  // The header length (in 32bit words) is the last 4 bits of the first octet.
    
    switch (ipHeader->protocol) {
        
        case 6: //TCP
		{
            // get the old source port
			uint16_t* portLoc = ((uint16_t *) ( ((uint8_t *) ipHeader) + headerLen));
            portnum_n_t oldSrcPort = * portLoc;
			// Now put the new port number in.
			*portLoc = newSrcPort;
			// Now incrementally update the checksum based on RFC 1141

			uint32_t tmpNewPort = ntohs (newSrcPort);
			uint32_t tmpOldPort = ntohs (oldSrcPort);
			


            uint32_t checksum = ntohs (((uint16_t *) ip_hdr)[18]);
			checksum += (~tmpNewPort & 0xffff) + (tmpOldPort & 0xffff);
            checksum += ((~tmpNewPort) >> 16) + (tmpOldPort >> 16);
            checksum = (checksum & 0xffff) + (checksum >> 16);
			((uint16_t *) ip_hdr)[18] = htons ((checksum & 0xffff) + (checksum >> 16));
     	
		}
		break;

        case 17: //UDP
		{
            // get the old source port
			uint16_t* portLoc = ((uint16_t *) ( ((uint8_t *) ipHeader) + headerLen));
            portnum_n_t oldSrcPort = * portLoc;
			// Now put the new port number in.
			*portLoc = newSrcPort;
			// Now incrementally update the checksum based on RFC 1141
			uint16_t* checkSumPtr = ((uint16_t*) ip_hdr) + 16;
		
			uint32_t tmpNewPort = ntohs (newSrcPort);
			uint32_t tmpOldPort = ntohs (oldSrcPort);
			
            uint32_t checksum = ntohs (*checkSumPtr);
			checksum += (~tmpNewPort & 0xffff) + (tmpOldPort & 0xffff);
            checksum += ((~tmpNewPort) >> 16) + (tmpOldPort >> 16);
            checksum = (checksum & 0xffff) + (checksum >> 16);
			*checkSumPtr = htons ((checksum & 0xffff) + (checksum >> 16));
     	
		}
        break;
    }
}

/**
  * This function fixes replaces the destination port of the specified packet with
  * the new destination port.
  * @param ip_hdr Pointer to the ip header
  * @param newDstPort The new destination port to be put into the packet (in network byte order)
  */
void Tunnel::change_dst_port_fix_checksum (char *ip_hdr, portnum_n_t newDstPort) {
  
	
  	struct IPHdr* ipHeader = (struct IPHdr*) ip_hdr;
    uint8_t headerLen = (0x0F & ipHeader->vihl) * 4;  // The header length (in 32bit words) is the last 4 bits of the first octet.
    
    switch (ipHeader->protocol) {
        
        case 6: //TCP
		{


			// get the old destination port
			uint16_t* portLoc = ((uint16_t *) ( ((uint8_t *) ipHeader) + headerLen + 2));
            portnum_n_t oldDstPort = * portLoc;
			// Now put the new port number in.
			*portLoc = newDstPort;
			// Now incrementally update the checksum based on RFC 1141

			uint32_t tmpNewPort = ntohs (newDstPort);
			uint32_t tmpOldPort = ntohs (oldDstPort);
			


            uint32_t checksum = ntohs (((uint16_t *) ip_hdr)[18]);
			checksum += (~tmpNewPort & 0xffff) + (tmpOldPort & 0xffff);
            checksum += ((~tmpNewPort) >> 16) + (tmpOldPort >> 16);
            checksum = (checksum & 0xffff) + (checksum >> 16);
			((uint16_t *) ip_hdr)[18] = htons ((checksum & 0xffff) + (checksum >> 16));
     	
		}
		break;

     	case 17: //UDP
		{
            // get the old source port
			uint16_t* portLoc = ((uint16_t *) ( ((uint8_t *) ipHeader) + headerLen));
            portnum_n_t oldDstPort = * portLoc;
			// Now put the new port number in.
			*portLoc = newDstPort;
			// Now incrementally update the checksum based on RFC 1141
			uint16_t* checkSumPtr = ((uint16_t*) ip_hdr) + 16;
		
			uint32_t tmpNewPort = ntohs (newDstPort);
			uint32_t tmpOldPort = ntohs (oldDstPort);
			
            uint32_t checksum = ntohs (*checkSumPtr);
			checksum += (~tmpNewPort & 0xffff) + (tmpOldPort & 0xffff);
            checksum += ((~tmpNewPort) >> 16) + (tmpOldPort >> 16);
            checksum = (checksum & 0xffff) + (checksum >> 16);
			*checkSumPtr = htons ((checksum & 0xffff) + (checksum >> 16));
     	
		}
        break;

    }
}

/**
  * This function fixes replaces the destination address of the specified
  * IP header with the specified address.
  * @param ip_hdr The IP header whose address is to be replaced.
  * @param newDstAddr The new destination address (in network byte order)
  */
void Tunnel::change_dst_ip_fix_checksum (char *ip_hdr, struct in_addr newDstAddr) {
    
    uint16_t old_ip_chk, new_ip_chk;
    unsigned long tmp_addr, old_rcv_addr, new_rcv_addr;

    memcpy (&old_rcv_addr, ip_hdr + 16, 4);
    tmp_addr = newDstAddr.s_addr;
    memcpy (ip_hdr + 16, &tmp_addr, 4);
    new_rcv_addr = ntohl(newDstAddr.s_addr);
	
    old_rcv_addr = ntohl (old_rcv_addr);

    old_ip_chk = ((uint16_t *) ip_hdr)[5];
    ((uint16_t *) ip_hdr)[5] = 0;
    new_ip_chk = in_cksum ((uint16_t *) ip_hdr, 20);
    ((uint16_t *) ip_hdr)[5] = new_ip_chk;

    if (ip_hdr[0] == 0x45) {
        uint32_t checksum;

        switch (ip_hdr[9]) {
            
            case 1:     // ICMP
                // Do nothing!
            break;

            case 2:     // IGMP
                // Do nothing!
            break;

            case 6:     // TCP
                // incrementally update TCP checksum (see RFC 1141) 
                checksum = ntohs (((uint16_t *) ip_hdr)[18]);

                checksum += (~new_rcv_addr & 0xffff) + (old_rcv_addr & 0xffff);
                checksum += ((~new_rcv_addr) >> 16) + (old_rcv_addr >> 16);
                checksum = (checksum & 0xffff) + (checksum >> 16);

                ((uint16_t *) ip_hdr)[18] = htons ((checksum & 0xffff) + (checksum >> 16));
            break;

            case 17:        // UDP
                // incrementally update UDP checksum (see RFC 1141)
                checksum = ntohs (((uint16_t *) ip_hdr)[16]);

                checksum += (~new_rcv_addr & 0xffff) + (old_rcv_addr & 0xffff);
                checksum += ((~new_rcv_addr) >> 16) + (old_rcv_addr >> 16);
                checksum = (checksum & 0xffff) + (checksum >> 16);

                ((uint16_t *) ip_hdr)[16] = htons ((checksum & 0xffff) + (checksum >> 16));
            break;

  	   case 50:            //ESP
           break;

           case 99:            //HIP
           break;


            default:
                PRINT_DEBUG1 (
                    DEBUG_LEVEL_MINIMAL, 
                    "\n Unsupported layer4-protocol in handle_data_packet(), " //
                    "packet-type: %u\n", ip_hdr[9]
                );
            break;
        }
    }
}





/**
  * This function fixes replaces the source and destination of the specified
  * IP header with the real and fake addresses of the given VN.
  * @param myFakeAddr The fake address to be put as the new source address (in network byte order)
  * @param finalDestAddr The address to be put in the destination field of the packet (in network byte order)
  */
void Tunnel::change_addr_fix_checksum (char *ip_hdr, struct in_addr myFakeAddr, struct in_addr finalDestAddr) {
    
    /*
    struct in_addr myFakeAddr; //Will be in network byte order
     if (inet_aton(vn->fakeIP->c_str(), &myFakeAddr) == 0) {
        PRINT_DEBUG (
                DEBUG_LEVEL_MINIMAL, 
                "Error while fixing checksum: Invalid fake IP :%s\n", 
                vn->fakeIP->c_str()
            );          
    }
   
    */
     {
        struct in_addr ttdestAddr;
        struct in_addr ttsrcAddr;
        memcpy (&ttdestAddr, ip_hdr + 16, sizeof (ttdestAddr));
        memcpy (&ttsrcAddr, ip_hdr + 12, sizeof (ttsrcAddr));
        
        PRINT_DEBUG4 (
                DEBUG_LEVEL_VERBOSE,
                "In change_addr_fix_chksum: src = %s\n"
                "dest = %s\n"
                "finalDestAddr = %s\n" 
                "myFakeIP = %s\n",
                inet_ntoa (ttsrcAddr),
                inet_ntoa (ttdestAddr),
                inet_ntoa (finalDestAddr),
                inet_ntoa (myFakeAddr)
            );
        
    }
  
    //struct in_addr finalDestAddrIP; //this is also in network byte order
    //getTunnelIP (&finalDestAddr);
    if ( (memcmp (&finalDestAddr, ip_hdr + 16, sizeof (finalDestAddr)) == 0) &&        //this packet is addressed to the local TUN device
         (memcmp (&myFakeAddr, ip_hdr + 12, sizeof (myFakeAddr)) == 0)        //this source of this packet matches the local VN
       ) {

        //No need to rewrite address
        PRINT_DEBUG0 ( 
                DEBUG_LEVEL_MINIMAL,
                "No need to rewrite addresses\n"
            );
        return;
    }

   
    uint16_t old_ip_chk, new_ip_chk;
    unsigned long tmp_addr, old_snd_addr, old_rcv_addr, new_snd_addr, new_rcv_addr;

    memcpy (&old_rcv_addr, ip_hdr + 16, 4);
    tmp_addr = finalDestAddr.s_addr;
    //tmp_addr = htonl (vn->localI3Address.t.v4.addr.s_addr);  TODO Problem when address of VN changes.  Must make the proxy do address change check too.
    
    memcpy (ip_hdr + 16, &tmp_addr, 4);
    
    new_rcv_addr = ntohl(finalDestAddr.s_addr);
    //new_rcv_addr = vn->localI3Address.t.v4.addr.s_addr; TODO when address of VN changes, Must make the proxy do address change check too.

    tmp_addr = myFakeAddr.s_addr;
    
    memcpy (&old_snd_addr, ip_hdr + 12, 4);
    //tmp_addr = vn->fake_addr.t.v4.addr.s_addr;
    memcpy (ip_hdr + 12, &tmp_addr, 4);
    new_snd_addr = ntohl(myFakeAddr.s_addr);

    old_rcv_addr = ntohl (old_rcv_addr);
    old_snd_addr = ntohl (old_snd_addr);

    old_ip_chk = ((uint16_t *) ip_hdr)[5];
    ((uint16_t *) ip_hdr)[5] = 0;
    new_ip_chk = in_cksum ((uint16_t *) ip_hdr, 20);
    ((uint16_t *) ip_hdr)[5] = new_ip_chk;

    if (ip_hdr[0] == 0x45) {
        uint32_t checksum;

        switch (ip_hdr[9]) {
            
            case 1:     // ICMP
                // Do nothing!
            break;

            case 2:     // IGMP
                // Do nothing!
            break;

            case 6:     // TCP
                // incrementally update TCP checksum (see RFC 1141) 
                checksum = ntohs (((uint16_t *) ip_hdr)[18]);

                checksum += (~new_rcv_addr & 0xffff) + (old_rcv_addr & 0xffff);
                checksum += (~new_snd_addr & 0xffff) + (old_snd_addr & 0xffff);
                checksum += ((~new_rcv_addr) >> 16) + (old_rcv_addr >> 16);
                checksum += ((~new_snd_addr) >> 16) + (old_snd_addr >> 16);
                checksum = (checksum & 0xffff) + (checksum >> 16);

                ((uint16_t *) ip_hdr)[18] = htons ((checksum & 0xffff) + (checksum >> 16));
            break;

            case 17:        // UDP
                // incrementally update UDP checksum (see RFC 1141)
                checksum = ntohs (((uint16_t *) ip_hdr)[16]);

                checksum += (~new_rcv_addr & 0xffff) + (old_rcv_addr & 0xffff);
                checksum += (~new_snd_addr & 0xffff) + (old_snd_addr & 0xffff);
                checksum += ((~new_rcv_addr) >> 16) + (old_rcv_addr >> 16);
                checksum += ((~new_snd_addr) >> 16) + (old_snd_addr >> 16);
                checksum = (checksum & 0xffff) + (checksum >> 16);

                ((uint16_t *) ip_hdr)[16] = htons ((checksum & 0xffff) + (checksum >> 16));
            break;

  	   case 50:            //ESP
           break;

           case 99:            //HIP
           break;


            default:
                PRINT_DEBUG1 (
                    DEBUG_LEVEL_MINIMAL, 
                    "\n Unsupported layer4-protocol in handle_data_packet(), " //
                    "packet-type: %u\n", ip_hdr[9]
                );
            break;
        }
    }
}

/** 
  * This function is used to clean up the state associated with the tunnel.
  */
void Tunnel::cleanup () {

    PRINT_DEBUG0 (
                    DEBUG_LEVEL_MINIMAL,
                    "Cleaning up tunnel.\n"
                );

#ifdef __APPLE__
    system("ipfw delete 333");   
#endif

#ifndef _WIN32
    if (planetlabFlag) {
   
        close (fifo_fd_app2proxy);
        close (fifo_fd_proxy2app);
    }
#endif

}


#ifdef __APPLE__
void Tunnel::initCaptureSocket() {
   
    PRINT_DEBUG (DEBUG_LEVEL_MINIMAL, "Initializing Apple Capture Socket at port %d.\n", appleCapturePort); 

    appleCaptureSocket = create_divert_socket(appleCapturePort);
    if (appleCaptureSocket == -1) {
	PRINT_DEBUG ( DEBUG_LEVEL_FATAL, "ip_conduit constructor:  divert socket");
        EXIT_ON_ERROR;
    }
    PRINT_DEBUG (DEBUG_LEVEL_MINIMAL, "appleCaptureFD = %d\n", appleCaptureSocket);
    appleOutSocket = raw_socket();
    if (appleOutSocket == -1) {
	PRINT_DEBUG(DEBUG_LEVEL_MINIMAL, "ip_conduit_output constructor:  raw_socket");
    }

}

/* get the prefix lenght, i.e., the number of leading ones */ 
unsigned int get_prefix_len(char *prefix)
{
#define MAX_PREFIX_STR_LEN 32	
	char  local_prefix[MAX_PREFIX_STR_LEN];
	char *first;
	char *last;
	char  ch;
	unsigned int prefix_len = 0;
	int   i;

	if (strlen(prefix) >= MAX_PREFIX_STR_LEN)
		abort();

	strcpy(local_prefix, prefix);
	local_prefix[strlen(prefix)] = '.';
	first = local_prefix;
	
	while (1) {
		if ((last = strstr(first, ".")) == NULL)
			break;
		*last = 0;
		ch = atoi(first);
		for (i = 0; i < 8; i++) {
			if (ch < 0)
				prefix_len++;
			else
				break;
			ch = ch << 1;
		}
		first = last + 1;
	}
	return prefix_len;
}
#endif

#ifdef _WIN32 

/** Self-rescheduling wrapper for updateRegistryDNSServers()
 */
void Tunnel::nameServerUpdate () {
	updateRegistryDNSServers  ();

	TObjectFunctor<Tunnel> * nameServerUpdateFunctor = 
                        new TObjectFunctor<Tunnel>(
                                        this, 
                                        &Tunnel::nameServerUpdate
                                    );
    
    TimerEvent *nameServerUpdateTimer = TimerEvent::create (
                                                        60,  //seconds
                                                        0, 
                                                        (TFunctor *) nameServerUpdateFunctor, 
                                                        "nameServerUpdate",
                                                        true
                                                    );
    
    proxy->addTimerEvent(nameServerUpdateTimer);
}

/**
 * This function is periodically called.
 */
void Tunnel::updateRegistryDNSServers () {

	if (proxy->stopProxyFlag) {
		return;
	}

	// Get current DNS server IP - start
	// We need to periodically get the DNS server's address
	// because it might change when we take our laptop from school
	// to home, for example.
	bool tmp_initNameServerFlag = proxy->initNameServer();
	while (tmp_initNameServerFlag == false) {
		// Keep trying to initialize the name server
		PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL,
					"Retrying initNameServer() ...\n"
			);

		// Sleep for some time before retrying.
#		if defined (_WIN32) 
			Sleep (5 * 1000);	//time in milliseconds
#		else
			sleep(5);   //sleep 5 seconds
#		endif

		tmp_initNameServerFlag = proxy->initNameServer();
	}
	// Get current DNS server IP - stop


	rememberRegDNSServerSettings();

	// Iterate through the interfaces
	PIP_ADAPTER_INFO pAdapterInfo;
    ULONG OutBufLen = 0;
    
    GetAdaptersInfo(NULL, &OutBufLen);
    pAdapterInfo = (PIP_ADAPTER_INFO)malloc(OutBufLen);
    GetAdaptersInfo(pAdapterInfo, &OutBufLen);
    
    PIP_ADAPTER_INFO pai = pAdapterInfo;
	char dnsCmd [500];
	sprintf (dnsCmd, "static %s primary", tunDNSServerStr.c_str());
    while (pai) {
		char humanInterfaceName[1000];
		DWORD humanInterfaceNameLen = 1000;
		getHumanReadableInterfaceName (pai->AdapterName, humanInterfaceName, &humanInterfaceNameLen);
		do_netsh_setdns (humanInterfaceName, dnsCmd);  
		pai = pai->Next;
    }
    free(pAdapterInfo);


	
}


/** 
 * This function is called when the proxy terminates normally or abnormally.
 * It restores the DHCP NameServer mappings in the registery
 * from the saved values.
 */
void restoreNameServerSettings_intarg(int signum) {
	restoreNameServerSettings();
	return;
}
 static BOOL restoreNameServerSettings_CtrlHandler( DWORD fdwCtrlType ) {

	 restoreNameServerSettings();
	 return FALSE;
}
void restoreNameServerSettings () {

	PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL, "Restoring name server settings\n\n\n\n");

	// Iterate through the interfaces
	PIP_ADAPTER_INFO pAdapterInfo;
    	ULONG OutBufLen = 0;
    	
    	GetAdaptersInfo(NULL, &OutBufLen);
    	pAdapterInfo = (PIP_ADAPTER_INFO)malloc(OutBufLen);
    	GetAdaptersInfo(pAdapterInfo, &OutBufLen);
    
    	PIP_ADAPTER_INFO pai = pAdapterInfo;
    	while (pai) {

		// First check if there is a NameServer entry for this interface.
		// This means that it is not DHCP
		string tmpDNSServerStr;
		if ( lookupInterfaceNameServerMap (string (pai->AdapterName), &tmpDNSServerStr) ) {

			char humanInterfaceName[1000];
			DWORD humanInterfaceNameLen = 1000;
			getHumanReadableInterfaceName (pai->AdapterName, humanInterfaceName, &humanInterfaceNameLen);

			StringTokenizer dnsST (tmpDNSServerStr.c_str(), ",");
			if (dnsST.hasMoreTokens()) {
				// We set only the first DNS server in the list
				// TODO: What to do about the rest - Dilip?
				char dnsCmd [500];
				sprintf (dnsCmd, "static %s primary", dnsST.nextToken().c_str());
				do_netsh_setdns (humanInterfaceName, dnsCmd);   
			} else {
				PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
					"No DNSServer fields found in \"%s\". Possibly malformed.\n",
					tmpDNSServerStr.c_str()
					);
			}

		} else {
			//DHCP
		
			char humanInterfaceName[1000];
			DWORD humanInterfaceNameLen = 1000;
			getHumanReadableInterfaceName (pai->AdapterName, humanInterfaceName, &humanInterfaceNameLen);
			do_netsh_setdns (humanInterfaceName, "dhcp");  
		}
		pai = pai->Next;
   	 }
    	free(pAdapterInfo);

}


/**
 * Goes through all interfaces in the registry
 * and remembers the value of "NameServer" if found.
 */
void rememberRegDNSServerSettings() {

	// Forget the currently remembered values	
	// What should be returned
	BOOL retCode = TRUE;
	LONG res;

	// Open the SYSTEM Key
	HKEY keySYSTEM;
	res = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
							 "SYSTEM",
							 0, /* Reserved must be 0 */
							 KEY_READ,
							 &keySYSTEM
						);

	if (res != ERROR_SUCCESS) {
		PRINT_INFO0 (INFO_LEVEL_WARNING,
			"Tunnel::rememberDNSServerSettings: Unable to open HKEY_LOCAL_MACHINE\\SYSTEM\n"
						);
		retCode = FALSE;
		return;
	}

	//SYSTEM\CurrentControlSet
	HKEY keyCurrentControlSet;	
	res = RegOpenKeyEx(		keySYSTEM,
							 "CurrentControlSet",
							 0, /* Reserved must be 0 */
							 KEY_READ,
							 &keyCurrentControlSet
						);
	if (res != ERROR_SUCCESS) {
		// CurrentControlSet key not found.
		PRINT_INFO0 (INFO_LEVEL_WARNING,
						"Tunnel::rememberDNSServerSettings: Unable to open HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\n"
					);
	
		retCode = FALSE;
		RegCloseKey (keySYSTEM);
		return;
	}

	//SYSTEM\CurrentControlSet\Services
	HKEY keyServices;	
	res = RegOpenKeyEx(		keyCurrentControlSet,
							 "Services",
							 0, /* Reserved must be 0 */
							 KEY_READ,
							 &keyServices
						);
	if (res != ERROR_SUCCESS) {

		PRINT_INFO0 (INFO_LEVEL_WARNING,
						"Tunnel::rememberDNSServerSettings: Unable to open HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\n"
					);
	
		retCode = FALSE;
		RegCloseKey (keyCurrentControlSet);
		RegCloseKey (keySYSTEM);
		return;
	}

	//SYSTEM\CurrentControlSet\Services\Tcpip
	HKEY keyTcpip;	
	res = RegOpenKeyEx(		keyServices,
							 "Tcpip",
							 0, /* Reserved must be 0 */
							 KEY_READ,
							 &keyTcpip
						);
	if (res != ERROR_SUCCESS) {

		PRINT_INFO0 (INFO_LEVEL_WARNING,
						"Tunnel::rememberNameServerSettings: Unable to open HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\Tcpip\n"
					);
	
		retCode = FALSE;
		RegCloseKey (keyServices);
		RegCloseKey (keyCurrentControlSet);
		RegCloseKey (keySYSTEM);
		return;
	}


	//SYSTEM\CurrentControlSet\Services\Tcpip\Parameters
	HKEY keyParameters;	
	res = RegOpenKeyEx(		keyTcpip,
							 "Parameters",
							 0, /* Reserved must be 0 */
							 KEY_READ | KEY_SET_VALUE,
							 &keyParameters
						);
	if (res != ERROR_SUCCESS) {

		PRINT_INFO0 (INFO_LEVEL_WARNING,
						"Tunnel::rememberNameServerSettings: Unable to open HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\Tcpip\\Parameters\n"
					);
	
		retCode = FALSE;
		RegCloseKey (keyTcpip);
		RegCloseKey (keyServices);
		RegCloseKey (keyCurrentControlSet);
		RegCloseKey (keySYSTEM);
		return;
	}

	// Get the value of the DHCPNameServer value
	char strDHCPNameServer [MAX_REGISTRY_STR_VALUE_LEN];
	DWORD strDHCPNameServerSize = MAX_REGISTRY_STR_VALUE_LEN;
	DWORD valType;

	strDHCPNameServerSize = MAX_REGISTRY_STR_VALUE_LEN;
	res = RegQueryValueEx(
						keyParameters,
						NAME_SERVER_FIELD_NAME,
						NULL,
						&valType,
						(BYTE*) strDHCPNameServer,
						&strDHCPNameServerSize
					);
	if (res != ERROR_SUCCESS) {
		// No DHCPNameServer found for SYSTEM\CurrentControlSet\Services\Tcpip\Parameters
		// This is okay.
		PRINT_DEBUG0 (DEBUG_LEVEL_VERBOSE,
					"No " NAME_SERVER_FIELD_NAME " found for SYSTEM\\CurrentControlSet\\Services\\Tcpip\\Parameters\n"
				);
		updateNameServer_map (KEY_PARAM_STR, NULL);
	} else {
		PRINT_DEBUG1 (DEBUG_LEVEL_VERBOSE,
					"SYSTEM\\CurrentControlSet\\Services\\Tcpip\\Parameters DHCPNameServer=%s\n",
					strDHCPNameServer
				);

		updateNameServer_map (KEY_PARAM_STR, & string(strDHCPNameServer));

	}

	// Open the Interfaces Key and Enumerate all the interfaces
	//SYSTEM\CurrentControlSet\Services\Tcpip\Parameters\Interfaces
	HKEY keyInterfaces;	
	res = RegOpenKeyEx(		keyParameters,
							 "Interfaces",
							 0, /* Reserved must be 0 */
							 KEY_READ,
							 &keyInterfaces
						);
	if (res != ERROR_SUCCESS) {

		PRINT_INFO0 (INFO_LEVEL_WARNING,
						"Tunnel::rememberNameServerSettings: Unable to open HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\Tcpip\\Parameters\\Interfaces\n"
					);
	
		retCode = FALSE;
		RegCloseKey (keyParameters);
		RegCloseKey (keyTcpip);
		RegCloseKey (keyServices);
		RegCloseKey (keyCurrentControlSet);
		RegCloseKey (keySYSTEM);
		return;
	}


	int currInterfaceIndex = 0;
	while (TRUE) {

		char strInterfaceId [MAX_REGISTRY_STR_VALUE_LEN];
		DWORD strInterfaceIdSize = MAX_REGISTRY_STR_VALUE_LEN;
		
		res = RegEnumKeyEx(
							keyInterfaces,
							currInterfaceIndex,
							strInterfaceId,
							&strInterfaceIdSize,
							NULL,
							NULL,
							NULL,
							NULL
					);
		if (res != ERROR_SUCCESS && res != ERROR_NO_MORE_ITEMS) {
			// Some error occurred, so we get out of the loop.		
			PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL,
						"Tunnel::rememberNameServerSettings: Error while iterating through HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\Tcpip\\Parameters\\Interfaces\n"
					);

			retCode = FALSE;
			break;
		} else if (res == ERROR_NO_MORE_ITEMS) {
			// We are done enumerating all interfaces. 
			// So we exit out of the loop.
			break;
		}
		currInterfaceIndex++;

		// Open the key associated with the current interface
		HKEY keyCurrInterface;
		res = RegOpenKeyEx( keyInterfaces,
							strInterfaceId,
							0, /* Reserved must be 0 */
							KEY_READ | KEY_SET_VALUE,
							&keyCurrInterface
					);

		if (res != ERROR_SUCCESS) {
			PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
						"Error while opening key associated with interface  HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\Tcpip\\Parameters\\Interfaces%s.\n",
						strInterfaceId
					);
			continue;
		}

		// Now read the avlue of DHCPNameServer of the current interface
		strDHCPNameServerSize = MAX_REGISTRY_STR_VALUE_LEN;
		res = RegQueryValueEx(
						keyCurrInterface,
						NAME_SERVER_FIELD_NAME,
						NULL,
						&valType,
						(BYTE*) strDHCPNameServer,
						&strDHCPNameServerSize
					);
		if (res != ERROR_SUCCESS) {
			// No DHCPNameServer found for SYSTEM\CurrentControlSet\Services\Tcpip\Parameters
			// This is okay.
			PRINT_DEBUG1 (DEBUG_LEVEL_VERBOSE,
					"No " NAME_SERVER_FIELD_NAME " found for SYSTEM\\CurrentControlSet\\Services\\Tcpip\\Parameters\\Interfaces\\%s\n",
					strInterfaceId
				);
			updateNameServer_map (string (strInterfaceId), NULL);
		} else {
			PRINT_DEBUG2 (DEBUG_LEVEL_VERBOSE,
					"SYSTEM\\CurrentControlSet\\Services\\Tcpip\\Parameters\\Interfaces\\%s DHCPNameServer=%s\n",
					strInterfaceId, strDHCPNameServer
				);
			updateNameServer_map (string (strInterfaceId), &string(strDHCPNameServer));
		}	

		RegCloseKey (keyCurrInterface);

  	} // while (TRUE)



	RegCloseKey (keyParameters);
	RegCloseKey (keyTcpip);
	RegCloseKey (keyServices);
	RegCloseKey (keyCurrentControlSet);
	RegCloseKey (keySYSTEM);

}

void updateNameServer_map ( string interfaceNameStr, string* nameServerStr) {

	if (nameServerStr == NULL || isBlank ( (*nameServerStr).c_str())) {
		interface_NameServer_map.erase (interfaceNameStr);
	} else if (*nameServerStr != tunDNSServerStr) {
		// DHCP must have filled in this value after we set it to tunDNSServerStr				// So let us save it.  This is the value that must be restored when the 
		// proxy exits.
		interface_NameServer_map.insert (pair<string, string> (interfaceNameStr, *nameServerStr));
		PRINT_DEBUG2 (DEBUG_LEVEL_MINIMAL,
				"Setting the DHCPNameServer mapping for interface %s to %s.\n",
				interfaceNameStr.c_str(),
				(*nameServerStr).c_str()
			);
	}
}


BOOL lookupInterfaceNameServerMap (string interfaceName, string* nameServerStr) {

	map <string, string>::iterator itr = interface_NameServer_map.find (interfaceName);
	if (itr != interface_NameServer_map.end()) {
		*nameServerStr = itr->second;
		return TRUE;
	} else {
		return FALSE;
	}
}
#endif
