/**
 *  \file Tunnel.h
 *  \brief Definition of some default Tunnel parameters and the declaration of the Tunnel class.
*/

#ifndef TUNNEL_H
#define TUNNEL_H 1

#include <list>
#include <map>
#include <string>
using namespace std;
#include <string.h>

#include "OCALA_Export.h"
#include "common.h"

#if defined (_WIN32)
	#include <windows.h>
	#include <winioctl.h>
	#include <iphlpapi.h>
	#include <iprtrmib.h>

	// These undefs are required to prevent conflict with stl library in windows
	#undef min
	#undef max

#endif


/** The maximum length of a packet captured by the tunnel. */
#define MAX_PKT_CAPT_LEN    2000

#define DEFAULT_TUNNEL_DEVICE_PATH      "/dev/net/tun"
#define DEFAULT_TUNNEL_DEVICE_NAME      "tun0"
#define DEFAULT_TUNNEL_WINTAP_NAME      "undef"
#define DEFAULT_TUNNEL_MTU              "1200"
#define DEFAULT_TUNNEL_IP_ADDRESS       "1.0.0.1" 
#define DEFAULT_TUNNEL_ROUTING_TABLE    "tuntab"

#ifdef __APPLE__
	#define DEFAULT_APPLE_CAPTURE_PORT	63322
#endif

#ifndef _WIN32
#   define DEFAULT_TUNNEL_IFCONFIG_CMD     "/sbin/ifconfig"
#   define DEFAULT_TUNNEL_IPTABLES_CMD     "/sbin/iptables"
#   define DEFAULT_TUNNEL_IP_CMD           "/sbin/ip"
#   define DEFAULT_TUNNEL_MODPROBE_CMD     "/sbin/modprobe"

#   define IPTABLES_SAVE_FILE               "OCALA_iptables-save-file"
#endif
  
#ifdef _WIN32
# define KEY_PARAM_STR string("Parameters")
#endif

class Proxy;
class VirtualNode;
class DNSPacket;
class IPPacket;

using namespace std;

/**
 *  This class represents the packet capture device.
 *  It can be thought of as a tunnel between the legacy application and the 
 *  proxy.  In this case, the tunnel is implemented via vtun.
 *  
 *  Independent of any overlay network.
 */

class Tunnel {

//#ifdef MY_SOFTWARE_NAT
	public:
		void receivedPacket (IPPacket *);
		void setNATPortRange (int natPortRangeStart, int natPortRangeEnd) {
			this->natPortRangeStart= natPortRangeStart;
			this->natPortRangeEnd = natPortRangeEnd;
		}
	private:
	
		/** The range of ports reserved for our software NAT. */
		int natPortRangeStart, natPortRangeEnd;
//#endif


private:

    /** This flag is by default false.
     *  If it is true, the proxy uses the Alpine4Linux user level networking
     *  stack for sending/receiving packets.  This is mainly used on Planetlab nodes.
     */
    bool planetlabFlag;
    void setPlanetlabFlag (bool f) { this->planetlabFlag = f; }
    
    
// START planetlab
// The following defs are necessary only when the proxy is used on a planetlab node
// with the Alpine4Linux user level network stack.
#   define FIFO_APP2PROXY   "/tmp/fifo_app2proxy"
#   define FIFO_PROXY2APP   "/tmp/fifo_proxy2app"
        
    /** File descriptor of the FIFO over which legacy apps send pkts to the proxy. */
    int fifo_fd_app2proxy;

    /** File descriptor of the FIFO over which proxy send pkts to legacy apps. */
    int fifo_fd_proxy2app;
// END planetlab

        
#if defined (_WIN32)

    HANDLE tunFD;  ///< The file descriptor of the packet caputure device.

        /** The name of the TAP-Win32 device */
    char tunWinTAPName[50];

    bool tunWinTAPName_defined;

#else
    int tunFD;  ///< The file descriptor of the packet caputure device.
#endif

#ifdef __APPLE__
	unsigned short appleCapturePort;
        int appleCaptureSocket;
        int appleOutSocket;
#endif


    /** The path to the tunnel device */
    char tunDevPath[50];
    
    /** The name of the tunnel device */
    char tunDevName[50];

    /** The MTU of the tunnel */
    char tunMTU[20];

    /** The name of the tunnel routing table */
    char tunTabName[25];

    /** The interface address assigned to the tunnel */
    char tunIPAddr[MAX_IP_ADDR_LEN];

    /** The command used to invoke iptables */
    char iptablesCmd[25];

    /** The command used to invoke ip */
    char ipCmd[25];

    /** The command used to invoke modprobe */
    char modprobeCmd[25];

    /** The command used to invoke ifconfig */
    char ifconfigCmd[25];

    /** A reference to the proxy served by this tunnel */
    Proxy* proxy;

    /** Dest/Src IP and port numbers whose packets are to be captured by
      * the tunnel 
      */
    list<port_protocol_n_t> dstPortList;
    list<ipaddr_n_t> dstIPAddrList;
    list<port_protocol_n_t> srcPortList;
    list<ipaddr_n_t> srcIPAddrList;
    void initMatchRules();
    void addPacketCaptureByDstIP (ipaddr_n_t dstIPAddr);
    void addPacketCaptureBySrcIP (ipaddr_n_t srcIPAddr);
    void addPacketCaptureByDstPort (port_protocol_n_t dstPort);
    void addPacketCaptureBySrcPort (port_protocol_n_t srcPort);
    

    void initIPTables();
    
    void readConfigValues();

    VirtualNode* lookupURL(URL reqURL);

public:

    void addDstPortMatch (port_protocol_n_t);
    void addSrcPortMatch (port_protocol_n_t);
    void addDstIPMatch (ipaddr_n_t);
    void addSrcIPMatch (ipaddr_n_t);

    
    OCALA_LIB_EXPORT_IMPORT void initTunnel();
	OCALA_LIB_EXPORT_IMPORT void initNameServerRedirector ();

    OCALA_LIB_EXPORT_IMPORT void cleanup();

    OCALA_LIB_EXPORT_IMPORT Tunnel(Proxy* proxy, IPAddress *tunIPAddress);
   
    OCALA_LIB_EXPORT_IMPORT void setProxy(Proxy *proxy);

    void answerDNSRequest(DNSPacket *dnsPkt, const IPAddress* fakeIP);

#if defined(_WIN32)
    OCALA_LIB_EXPORT_IMPORT void tunReader();
#endif

    void writeToTunnel(char *buf, int len);

#if defined (_WIN32)
	void capturePacket(char* pktBuf, int pktLen);
#else
    void capturePacket();
#endif
    static void change_addr_fix_checksum (char *ip_hdr, struct in_addr myFakeAddr, struct in_addr finalDestAddr);
	static void change_dst_ip_fix_checksum (char *ip_hdr, struct in_addr newDstAddr);
	static void change_dst_port_fix_checksum (char *ip_hdr, portnum_n_t newDstPort);
	static void change_src_port_fix_checksum (char *ip_hdr, portnum_n_t newSrcPort);

 
    char* getTunnelIP();

#ifdef __APPLE__
    void initCaptureSocket();
#endif

#ifndef _WIN32
    void saveIPTables();
    void restoreIPTables();
#endif

#if defined (_WIN32)
    HANDLE getFD();
#else
    int getFD();
#endif


#if defined (_WIN32)
	/** The IP address of the DNS server associated with the tun device. 
	    Usually 10.0.0.2
	 */
	in_addr_t tunDNSServerAddr;
	void nameServerUpdate();
	void updateRegistryDNSServers ();

#endif

   friend int main (int argc, char* argv[]);
   friend void proxy_cleanup (int signum); 
};

#ifdef _WIN32

	OCALA_LIB_EXPORT_IMPORT void restoreNameServerSettings ();
	static BOOL restoreNameServerSettings_CtrlHandler( DWORD fdwCtrlType );
	void restoreNameServerSettings_intarg (int);
	void setDNSServers2TunDNSServer();
	void do_netsh_setdns (char* interfaceName, char* dnsCmd);


	void rememberRegDNSServerSettings();
	void updateNameServer_map ( string interfaceNameStr, string* nameServerStr);
	BOOL lookupInterfaceNameServerMap (string interfaceName, string* nameServerStr);
#endif

#endif

