/** 
  * \file VirtualNode.cpp
  * \brief This file contains the class definitions of a Virtual Node.
  * A virtual node represents an end point (remote computer) to the
  * legacy application.
  */

#include "Proxy.h"
#include "VirtualNode.h"
#include "common.h"
#include "Params.h"
#include "TFunctor.h"
#include "TimerEvent.h"
#include "Tunnel.h"
#include "oci/OCIState.h"

#include <assert.h>

/**
  * Constructs a new Virtual Node for the specified DNS URL.
  * This constructor is used when we are constructing a new VN on
  * receiving a DNS request for a previously unseen URL
  */
VirtualNode::VirtualNode( OCIState *ociState, Proxy* proxy) {

    this->ociState = ociState;
    this->proxy = proxy;
    this->dnsPacket = NULL;
    this->fakeIP = NULL;
    this->flowHash = 0;
    this->flowInitiatedFlag = false;

}

/**
  * Gives a string representation (pretty printed) of the Virtual Node.
  * @param tmpStr a character buffer in which the string is stored.
  * Sufficient space should be allocated for tmpStr.
  * @return pointer to tmpStr
  */
char* VirtualNode::toString(char *tmpStr) {
    string str = "Virtual Node\n";

    if (dnsPacket != NULL) {
        str += "\tURL : " + dnsPacket->reqURL + "\n";
    } else {
        str += "\tURL : NULL\n";
    }
   
    if (fakeIP != NULL) {
        str += "\tFakeIP Addr: " + *fakeIP + "\n";
    } else {
        str += "\tFakeIP Addr: NULL\n";
    }

    strncpy(tmpStr, str.c_str(), MAX_PRINT_BUF);
    return tmpStr;
}


/**
 * This adds the specified DNS request to the queue of DNS requests for the
 * the URL of this VN.
 * @param dnsPkt the DNSPacket corresponding to the request to be queued.
 */
void VirtualNode::queueDNSRequest(DNSPacket *dnsPkt) {
    dnsRequestQueue.push_back(dnsPkt);
}

/**
  * This function answers all the DNS queries queued on this virtual node.
  * @param fakeIP The fakeIP address to be returned.
  * @param tunnel A pointer to the tunnel to be used to send back the reply.
  */
void VirtualNode::answerDNSRequests(const IPAddress* fakeIP) {
   
    while ( ! dnsRequestQueue.empty()) {
        
        DNSPacket* currPacket = (DNSPacket*) dnsRequestQueue.front();
        proxy->tunnel->answerDNSRequest(currPacket, fakeIP);
        dnsRequestQueue.pop_front();

        // Do not free the DNS packet if this was the first DNS packet which set up this VN.
        if (currPacket != dnsPacket) {
            delete(currPacket);
        }
    }
}

/**
 * This adds the specified data packet to the queue of data packets
 * waiting for this VN status to be READY.
 * @param dataPkt The data packet to be queued.
 */
void VirtualNode::queueDataPacket(IPPacket *dataPkt) {
    dataPacketQueue.push_back(dataPkt);
}

/**
  * This is function is usually called from the VirtualNode's destructor.
  * It is used to free up the packets on the dataPacketQueue.
  */
void VirtualNode::destroyDataQueue() {
    while ( ! dataPacketQueue.empty()) {
        IPPacket* currPacket = (IPPacket*) dataPacketQueue.front();
        dataPacketQueue.pop_front();
        delete(currPacket);
    }
}

/**
  * This is function is usually called from the VirtualNode's destructor.
  * It is used to free up the packets on the DNS request Queue.
  */
void VirtualNode::destroyDNSQueue() {
    while ( ! dnsRequestQueue.empty()) {
        DNSPacket* currPacket = (DNSPacket*) dnsRequestQueue.front();
        dnsRequestQueue.pop_front();
        
        /** Free up all packets except for the first one, which caused the creation of this VN.
          * The first DNS request will be deleted separately.
          */
        if (currPacket != dnsPacket) {
            delete(currPacket);
        }
    }
}
/** Destructor */
VirtualNode::~VirtualNode() {
    //TODO release fake IP address back to free list

    destroyDNSQueue();
    destroyDataQueue();

    if (fakeIP != NULL) {
        delete (fakeIP);
    }
   
    if (dnsPacket != NULL && dnsPacket->pktContents != NULL) {

    	unsigned short key = ( (uint8_t) dnsPacket->pktContents[29] << 8) + (uint8_t) dnsPacket->pktContents[28];
	    proxy->map_key_DNSPacket.erase (key);

        delete (dnsPacket);
    }
}


/** Is this virtual node ready to relay data? 
 *  Returns true if ALL setup has been completed.  
 */
bool VirtualNode::isReady() {
    
    bool retVal = true;
    if (ociState->status != OCI_STATE_STATUS_READY) {  
        retVal = false;
        printf ("isReady -- 1\n");
    }
    
    return retVal;
}

/**
  * If all modules are ready, the DNS replies is sent to any
  * waiting legacy application.
  */
void VirtualNode::answerDNSRequests() {

    if (isReady()) {
        //answer DNS requests only after everyone is ready 
        answerDNSRequests(fakeIP);
    }
}
