/**
 *  \file VirtualNode.h
 *  \brief Declaration of the VirtualNode class.
*/

#ifndef VIRTUAL_NODE_H
#define VIRTUAL_NODE_H 1

#include "common.h"
#include "DNSPacket.h"

#include <list>
using namespace std;

class TimerEvent;
class Proxy;
class OCIState;
class Tunnel;
class FakeAddressNegotiator;
class CompressionState;

class VirtualNode {

private:

    Proxy* proxy;

    OCIState *ociState;
    
    /** Queue of DNSRequests for the URL of this VN*/
    list<DNSPacket *> dnsRequestQueue;

    /** Queue of data packets waiting to be sent once this VN status is READY. */
    list<IPPacket *> dataPacketQueue;

  
    //TODO List of data packets captured while VN is not ready.

    void addVNCheck();

public:

    /** This is used only when the VN was constructed for a flow not initiated by a DNS request.
      */
    uint32_t flowHash;
    
    /** This flag is set to true for those VNs which were constructed NOT as a result of getting a DNS request.
      */
    bool flowInitiatedFlag;

    /** The fake IP address of this Virtual Node */
    IPAddress* fakeIP;

    /** This field is used only if flowInitiatedFlag is true.
      * It is the source IP address of the captured packets which initiated this connection.
      * When reply packets are written back to the application, the destination address should be this.
      * Note: For VNs where flowInitiatedFlag is false, i.e. DNS initiated connections,
      * the destination address of reply packets is the tunnel address.
      */
    ipaddr_n_t fromIPAddr;
    
    /** Pointer to the DNS packet which lead to the creation of this Virtual node.
      * Maybe NULL if the VN was created not on receipt of a DNS request, but at some
      * other overlay specific moment.
      * The reqURL of the dns packet is the URL masqueraded by this VN.
      */
    DNSPacket *dnsPacket;
    
    /** 
      *
      * VN_STATUS_READY - implies that overlay specific set up is complete.
      * However security and other plugin setup may not be complete.  
      * The node may not be able to masquerade as the destination till
      * all set up is complete.
      *
      * VN_STATUS_NOT_READY - implies that the virtual node is not yet ready
      * to receive packets.Some more initialization steps need to be 
      * completed by the overlay network.  This is in addition to security setup
      * etc, which may still be pending
      * TODO
      * VN_STATUS_TIMED_OUT - implies that virtual node has timed out due 
      * to long period of inactivity (overlay specific meaning)
      *
      * VN_STATUS_REFERSHING - implies that the virutal node had timed out, 
      * but the overlay network is now refreshing its status.
      */

    VirtualNode(DNSPacket *dnsPacket, IPAddress* fakeIP, Proxy *proxy);
    VirtualNode(OCIState *, Proxy *proxy);

    virtual ~VirtualNode();

    char *toString(char *tmpStr);
    void queueDNSRequest(DNSPacket *dnsPkt);
    void answerDNSRequests(const IPAddress* fakeIP);

    void queueDataPacket(IPPacket *dataPkt);
    void destroyDataQueue();
    void destroyDNSQueue();
    
    /**
      * This function is called when the reply to a DNS 
      * request made by this virtual node is received.
      */
    virtual void receivedDNSReply(DNSPacket *pkt) {
        //DUMMY : Overload to do something useful.
    }

    bool isReady();
    
    void answerDNSRequests();

    friend class LegacyAppLayer;
    friend class OCI;
};



#endif
