/**
 * \file dns_benchmark.cpp
 * \brief This program generates DNS packets.  It is used while benchmarking the proxy.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>
#include <string.h>
#include <fcntl.h>

extern "C" {
	#include "checkpoint.h"
}

//
// #defines for cleanliness
// 
#define UMILLION 1000000ULL
#define THRESHOLD_SLEEP_TIME_US 10000

#define UNDEFINED	0
#define IsDefined(v) (UNDEFINED != v)

#define PROTO_TCP	1
#define PROTO_UDP	2
#define PROTO_ECHO	3
#define SENDER		1
#define RECEIVER	2

#define DEFAULT_TOTAL_KB	1024
#define DEFAULT_NUM_PKTS	100
#define DEFAULT_SIZE_BYTES	1200
#define DEFAULT_SPACING_US	1000000
#define MAX_PKT_SIZE_BYTES	1500


//
// Global variables
//
int send_recv = UNDEFINED;
int proto = UNDEFINED;
char *name = UNDEFINED;
uint32_t addr = UNDEFINED;
uint16_t port = UNDEFINED;

int total_KB = DEFAULT_TOTAL_KB;
int num = DEFAULT_NUM_PKTS;
int size = DEFAULT_SIZE_BYTES;
int spacing = DEFAULT_SPACING_US;


//
// Usage
// 
void
Usage()
{
    fprintf(stderr, "Usage: ./benchmark -<s|r> -<U|T|E> -p <port>\n");
    fprintf(stderr, "      only for sender: -h <hostname>\n");
    fprintf(stderr, "      only for TCP sender: -B <total KB>\n");
    fprintf(stderr, "      only for UDP sender: -k <num pkts>");
    fprintf(stderr, " -b <packet size(bytes)> -t <spacing (us)>\n");

    exit(-1);
}

//
// Read command line options
// 
void
ProcessArgs(int argc, char *argv[])
{
    int ch;

    while (-1 != (ch = getopt(argc, argv, "srEUTp:h:k:b:t:B:"))) {
	switch (ch) {
	    case 's':
		if (IsDefined(send_recv)) Usage();
		else send_recv = SENDER;
		break;

	    case 'r':
		if (IsDefined(send_recv)) Usage();
		else send_recv = RECEIVER;
		break;

	    case 'U':
		if (IsDefined(proto)) Usage();
		else proto = PROTO_UDP;
		break;

	    case 'T':
		if (IsDefined(proto)) Usage();
		else proto = PROTO_TCP;
		break;

	    case 'E':
		if (IsDefined(proto)) Usage();
		else proto = PROTO_ECHO;
		break;

	    case 'p':
		port = atoi(optarg);
		break;
		
	    case 'h':
		name = optarg;
		break;

	    case 'k':
		num = atoi(optarg);
		break;

	    case 'b':
		size = atoi(optarg);
		break;

	    case 't':
		spacing = atoi(optarg);
		if (spacing < THRESHOLD_SLEEP_TIME_US) {
		    printf("Warning: usleep() granularity not small enough to allow this spacing\n");
		}
		break;
		
	    case 'B':
		total_KB = atoi(optarg);
		break;

	    default:
		Usage();
		break;
	}
    }

    if (!(IsDefined(send_recv) && IsDefined(proto) && 
	(RECEIVER == send_recv || IsDefined(name)) && IsDefined(port)))
	Usage();
}

//
// Sleep for appropriate amount of time
// 
// void SleepSpacing()
// {
//     usleep(spacing);
// }

//
// UDP/Echo Receiver
//
void UDPEchoReceiver()
{
    if (PROTO_ECHO == proto)
	printf("Starting Echo Receiver\n");
    else if (PROTO_UDP == proto)
	printf("Starting UDP Receiver\n");

    // UDP socket initialization
    int sock;
    if (-1 == (sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP))) {
	perror("socket()");
	abort();
    }

    struct sockaddr_in raddr;
    raddr.sin_family = AF_INET;
    raddr.sin_addr.s_addr = htonl(INADDR_ANY);
    raddr.sin_port = htons(port);

    if (-1 == (bind(sock, (struct sockaddr *) &raddr, sizeof(raddr)))) {
	perror("bind()");
	abort();
    }

    // Receive structures initialization
    struct sockaddr_in fromaddr;
    socklen_t addrlen = sizeof (struct sockaddr_in);
    char data[MAX_PKT_SIZE_BYTES];
    int pkt_count = 0;
    
    // Eternal recv
    uint32_t j, len;
    while (1) {
	if (-1 == (len = recvfrom(sock, data, MAX_PKT_SIZE_BYTES, 0, 
		    (struct sockaddr *)&fromaddr, &addrlen))) {
            perror("recvfrom()");
	    abort();
        }
	memmove(&j, data, sizeof(uint32_t));
	j = ntohl(j);
	addCheckpoint(UDP_RECV, j);

	if (PROTO_ECHO == proto) {
	    addCheckpoint(UDP_ECHO_SEND_START, j);
	    if (-1 == sendto(sock, data, len, 0, 
			(const struct sockaddr *) &fromaddr, sizeof(raddr))) {
		perror("sendto()");
		abort();
	    }
	    addCheckpoint(UDP_ECHO_SEND_END, j);
	}
    }
}

//
// UDP/Echo Sender
//
void UDPEchoSender()
{
    if (PROTO_ECHO == proto)
	printf("Starting Echo Sender\n");
    else if (PROTO_UDP == proto)
	printf("Starting UDP Sender\n");

    // UDP socket initialization
    int sock;
    if (-1 == (sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP))) {
	perror("socket()");
	abort();
    }

    // non blocking
    int flags = fcntl(sock, F_GETFL);
    flags |= O_NONBLOCK;
    fcntl(sock, F_SETFL, flags);

    struct sockaddr_in raddr;
    raddr.sin_family = AF_INET;
    raddr.sin_addr.s_addr = addr;
    raddr.sin_port = htons(port);

    if (connect(sock, (const struct sockaddr *)&raddr, sizeof(raddr)) == -1) {
	perror("connect()");
	abort();
    }

    // Generate random data
    char data[MAX_PKT_SIZE_BYTES];
    socklen_t addrlen = sizeof (struct sockaddr_in);
    for (int i = 0; i < MAX_PKT_SIZE_BYTES; i++)
	data[i] = rand() % 256;

    // Send data
    uint32_t j;
    for (uint32_t i = 0; i < num; i++) {
	j = htonl(i);
	memmove(data, &j, sizeof(uint32_t));

	addCheckpoint(UDP_SEND_START, i);
	if (0 > send(sock, data, size, 0)) {
	    perror("send");
	    abort();
	}
	addCheckpoint(UDP_SEND_END, i);

	struct timeval tv;
	tv.tv_sec  = spacing / UMILLION;
	tv.tv_usec = spacing % UMILLION;

	if (PROTO_ECHO == proto) {
	    int ret;
	    fd_set rset;
	    FD_ZERO(&rset);
	    FD_SET(sock, &rset);
	    
	    if (-1 == (ret = select(sock + 1, &rset, 0, 0, &tv))) {
		perror("select()");
		abort();
	    } else if (ret > 0) {
		uint32_t j, len;
		addCheckpoint(UDP_ECHO_RECV_START, i);
		if (-1 == (len = recvfrom(sock, data, MAX_PKT_SIZE_BYTES, 0, 
				(struct sockaddr *)&raddr, &addrlen))) {
		    perror("recvfrom()");
		    abort();
		}
		memmove(&j, data, sizeof(uint32_t));
		j = ntohl(j);
		addCheckpoint(UDP_ECHO_RECV_END, j);
	    }
	}
	
	usleep(tv.tv_sec * UMILLION + tv.tv_usec);
    }

    DumpEventInfo(0);
}

//
// TCP Receiver
//
void TCPReceiver()
{
    printf("Starting TCP Receiver\n");

    // init socket
    int listen_sock, recv_sock;
    if (-1 == (listen_sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP))) {
        perror("socket()");
        abort();
    }

    int sin_len = sizeof(struct sockaddr_in);
    struct sockaddr_in raddr, faddr;
    raddr.sin_family = AF_INET;
    raddr.sin_addr.s_addr = htonl(INADDR_ANY);
    raddr.sin_port = htons(port);

    if (-1 == bind(listen_sock, (struct sockaddr *) &raddr, sizeof(raddr))) {
	perror("bind()");
	abort();
    }

    if (-1 == listen(listen_sock, 1)) {
        perror("listen()");
        abort();
    }

    // accept new connection
    if (-1 == (recv_sock = accept(listen_sock, (struct sockaddr *) &faddr, 
		    (socklen_t *) &sin_len))) {
	perror("accept");
	abort();
    }

    // recv packets
#define Q_KB 100
    char data[Q_KB*MAX_PKT_SIZE_BYTES];
    int recv_len, total_recv_len = 0;

    recv(recv_sock, data, 0, 0);

    addCheckpoint(TCP_RECV_START, 0);
    while (0 < (recv_len = recv(recv_sock, data, Q_KB*MAX_PKT_SIZE_BYTES, 0))) {
	total_recv_len += recv_len;
    }
    addCheckpoint(TCP_RECV_END, 0);

    if (recv_len < 0)
	perror("recv");

    DumpEventInfo(0);
}

//
// TCP Sender
//
void TCPSender()
{
    printf("Starting TCP Sender\n");

    // init socket
    int sock;
    if (-1 == (sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP))) {
        perror("socket()");
        abort();
    }

    int sin_len = sizeof(struct sockaddr_in);
    struct sockaddr_in raddr;
    raddr.sin_family = AF_INET;
    raddr.sin_addr.s_addr = addr;
    raddr.sin_port = htons(port);

    if (connect(sock, (const struct sockaddr *)&raddr, sizeof(raddr)) == -1) {
	perror("connect()");
	abort();
    }

    // Generate random data
#define Q_KB 100
    char data[Q_KB*1024];
    for (int i = 0; i < Q_KB*1024; i++)
	data[i] = rand() % 256;

    // Send data
    int sent_bytes = 0;
    addCheckpoint(TCP_SEND_START, 0);
    for (uint32_t i = 0; i < total_KB; i++) {
	int len;

	if (total_KB - i < Q_KB)
	    len = (total_KB - i) * 1024;
	else
	    len = Q_KB * 1024;
	
	if (0 > send(sock, data, 8*MAX_PKT_SIZE_BYTES, 0)) {
	    perror("send");
	    abort();
	}
    }
    addCheckpoint(TCP_SEND_END, 0);
    
    close(sock);
    DumpEventInfo(0);
}

//
// Resolve host name
//
void ResolveHost()
{
    struct hostent *host;
    
    addCheckpoint(DNS_START, 0);
    host = gethostbyname(name);
    addCheckpoint(DNS_END, 0);

    DumpEventInfo(0);

    exit(0);
    
    if (NULL == host) {
	fprintf(stderr, "Host %s not found\n", name);
	exit(-1);
    }

    addr = *((uint32_t *) host->h_addr);

    struct in_addr ia;
    ia.s_addr = addr;
    printf("Contacting host: %s\n", inet_ntoa(ia));
}


// function pointer to all 4 funs.
void (*func_arr[3][2])(void) = {{TCPSender, TCPReceiver}, {UDPEchoSender, UDPEchoReceiver}, {UDPEchoSender, UDPEchoReceiver}};

//
// Initialization goes here
// 
void
initialize()
{
    // signal term
    signal(SIGINT, DumpEventInfo);
    signal(SIGABRT, DumpEventInfo);

    // rand num init
    srand(getpid() ^ time(NULL));
}

int
main(int argc, char *argv[])
{
    // signal term
    signal(SIGINT, DumpEventInfo);
    
    // cmd line args
    ProcessArgs(argc, argv);

    // resolve name if sender
    if (SENDER == send_recv)
	ResolveHost();
    
    // call function
    func_arr[proto-1][send_recv-1]();
}
