/**
 * \file CompressionState.cpp
 * \brief The functions to manipulate the compression state associated 
 * with the connection.  Also functions to implement the compression parameter
 * negotiation protocol.
 * Not used currently.
 */
#include <stdio.h>

#include "../VirtualNode.h"
#include "CompressionState.h"
#include "CompressionPacketBlocks.h"
#include "CompressionBlackBox.h"

extern "C" {
    #include <zlib.h>
};

/** Compression decompression buffer size */
#define CHUNK 128 * 1024

CompressionState::CompressionState (VirtualNode *vn, int level) {

    this->vn = vn;
    this->compressionLevel = level;
    this->compressionScheme  = COMPRESSION_SCHEME_ZLIB;
    this->blackBox = new CompressionBlackBox();
    
}

CompressionState::CompressionState (VirtualNode *vn) {

    this->vn = vn;
    this->blackBox = new CompressionBlackBox();
}

CompressionState::~CompressionState () {

    if (this->blackBox != NULL) {
        delete (this->blackBox);
        this->blackBox = NULL;
        
    }
}
/**
  * Compress plain data. Put the output in compressed data.
  * @param compressedDataLen will contain the length of the compressed data on exit.  However it must be initially set to the size of the compressedData buffer. The compressedDataBuffer should be atleast 0.1% + 12 bytes larger than plainData buffer!
  */
void CompressionState::compress (
                const char *plainData, unsigned long plainDataLen, 
                char *compressedData, unsigned long *compressedDataLen) {


    blackBox->compress (plainData, plainDataLen, compressedData, compressedDataLen, this->compressionLevel);

}


/** 
  * Decompress compressed data and put the output in plain data.
  * @param plainDataLen will be used to return the length of the decompressed data. However it must initialized to the size of the plain data buffer.
  */
void CompressionState::decompress (
        const char *compressedData, unsigned long compressedDataLen,
        char *plainData, unsigned long *plainDataLen) {

    blackBox->decompress (compressedData, compressedDataLen, plainData, plainDataLen);

}

/** Setup negotiation with the remote end point to use compression
  */
void CompressionState::initNewConnection () {
    
    CompressionPacketBlock *blk = new CompressionPacketBlock;
    blk->mainHeader.version = PROXY_VERSION;
    blk->mainHeader.blockType = PROXY_PACKET_BLOCK_COMPRESSION;
    blk->mainHeader.nextBlock = sizeof (CompressionPacketBlock);
    blk->compressionScheme = this->compressionScheme;
    blk->compressionLevel = this->compressionLevel;
    blk->msgCode = COMPRESSION_BLOCK_REQUEST;

    this->state = COMPRESSION_STATE_NEGOTIATION_IN_PROGRESS;

    //TODO FIX this->vn->sendBlock ( (OCIPacketBlock *) blk);
}

/**
  * This function is called when a compression block is received for
  * this compression state manager.
  */
void CompressionState::receivedBlock (CompressionPacketBlock *blk) {

    switch (blk->msgCode) {
        
        case COMPRESSION_BLOCK_REQUEST:
            // A request for enabling compression
            handleCompressionRequest (blk);

        break;

        case COMPRESSION_BLOCK_REPLY_OK:
            // Compression was enabled as requested.
            handleCompressionReplyOK (blk);
            
        break;

        case COMPRESSION_BLOCK_REPLY_FAIL:
            // Compression could not be enabled as requested.
            handleCompressionReplyFail (blk);
        break;

        default:
            PRINT_DEBUG0 (
                    DEBUG_LEVEL_SUPER,
                    "Unknown Compression block received.\n"
                );
        break;
    }

}

void CompressionState::handleCompressionRequest (CompressionPacketBlock *reqBlock) {


    //TODO Check if requested compression scheme is supported
    //For now, assume it is.
   
    //set local params
    this->compressionLevel = reqBlock->compressionLevel;
    this->compressionScheme = reqBlock->compressionScheme;
    
    CompressionPacketBlock *blk = new CompressionPacketBlock;
    blk->mainHeader.version = PROXY_VERSION;
    blk->mainHeader.blockType = PROXY_PACKET_BLOCK_COMPRESSION;
    blk->mainHeader.nextBlock = sizeof (CompressionPacketBlock);
    blk->compressionScheme = this->compressionScheme;
    blk->compressionLevel = this->compressionLevel;
    blk->msgCode = COMPRESSION_BLOCK_REPLY_OK;

    this->state = COMPRESSION_STATE_OK;
    
    // TODO FIX this->vn->sendBlock ( (OCIPacketBlock *) blk);   

}

void CompressionState::handleCompressionReplyOK (CompressionPacketBlock *blk) {

    //TODO check if the compression parameters in the reply match what we had requested.
    this->state = COMPRESSION_STATE_OK;
}

void CompressionState::handleCompressionReplyFail (CompressionPacketBlock *blk) {

    this->state = COMPRESSION_STATE_FAIL;
}
