/**
 *  \file debug.h
 *  \brief  This file contains the functions to print out debug/info messages, the debug/info level definitions
 *  and the current debug/info level.
 */
#ifndef _OCALA_DEBUG_H
#define _OCALA_DEBUG_H 1

#include <time.h>
#include "OCALA_Export.h"

#ifdef _WIN32
#include <stdio.h>
#	ifdef __cplusplus
	extern "C" 	OCALA_LIB_EXPORT_IMPORT FILE* debugFD;		
#	else
		extern OCALA_LIB_EXPORT_IMPORT FILE* debugFD;
#	endif
#endif

/** Irrespective of current levels, debug information will be printed only DEBUG_ENABLED is defined.
  * This is like a master switch.
  */
#define DEBUG_ENABLED

/**
  * The current debug level.
  * Only debug messages with level less than or equal to this value will
  * be printed.
  */
#define CURRENT_DEBUG_LEVEL 89

/**
  * The current info level. This is used to restrict the amount of
  * informational messages displayed to the user.
  * Only those messages whose level is less than or equal to the
  * CURRENT_INFO_LEVEL will be displayed.
  */
#define CURRENT_INFO_LEVEL 100

/**
  * This info level is used for displaying fatal error messages.
  * Always displayed.
  */
#define INFO_LEVEL_FATAL_ERROR 0

/**
  * This info level is used for messages which are very essential.
  * Messages at this level will always be displayed.
  */
#define INFO_LEVEL_MINIMAL  50

/**
  * This info level is used for warning messages.
  */
#define INFO_LEVEL_WARNING  60

/**
  * The messages at this level are not critical.  So they may not be
  * displayed (depends on CURRENT_INFO_LEVEL).
  */
#define INFO_LEVEL_VERBOSE  70

/**
  * This is used for error messages which are probably caused due to bugs in the 
  * code.  i.e. they are not INFO_LEVEL_WARNINGs
  */
#define DEBUG_LEVEL_WARNING   20


/**
  * This is used for messages describing fatal errors.
  */
#define DEBUG_LEVEL_FATAL   0

/**
  * This is used for messages that are not fatal, but quite important
  * or high level.
  */
#define DEBUG_LEVEL_MINIMAL     30

/** 
 * This debug level is used when we want to turn off all other debug messages
 * and display just a few selected messages.  This is used temporarily while
 * debugging some particular piece of code.
 */
#define DEBUG_LEVEL_SUPER -1

/**
  * This is used for messages that are not very important.
  */
#define DEBUG_LEVEL_VERBOSE     90

#define DEBUG_LEVEL_VERY_VERBOSE 95

/**
 *  This macro is used to terminate the program when some fatal error
 *  occurs.
 */
#define EXIT_ON_ERROR   exit(-1)


/** Chops off any \n, \r at the end of the specified string */
#define CHOP(str) \
    if (str[strlen(str) - 1] == '\n' || str[strlen(str) - 1] == '\r') str[strlen(str) - 1] = '\0';



#	ifndef PRINT_DEBUG
#		if !defined(_WIN32) || defined(__CYGWIN__)
#			ifdef DEBUG_ENABLED
			/**
   			 * This macro is used to print debugging information. 
			 * The message is printed only if the current debugging level is
			 * greater than that of the level specified in the macro call.
			 */

#               ifdef __cplusplus
#				    define PRINT_DEBUG(debugLevel, msg, ... )  \
                        if(debugLevel <= CURRENT_DEBUG_LEVEL) { \
                            time_t tmpTime = ::time(NULL); \
                            char* tmpTimeStr = ctime(&tmpTime); \
                            CHOP(tmpTimeStr);                  \
                            printf("[%s\b: Line:%d in file:%s] ", tmpTimeStr, __LINE__, __FILE__);\
                            printf(msg, ##__VA_ARGS__);\
                        }
#               else
#				    define PRINT_DEBUG(debugLevel, msg, ... )  \
                        if(debugLevel <= CURRENT_DEBUG_LEVEL) { \
                            time_t tmpTime = time(NULL); \
                            char* tmpTimeStr = ctime(&tmpTime); \
                            CHOP(tmpTimeStr);                  \
                            printf("[%s\b: Line:%d in file:%s] ", tmpTimeStr, __LINE__, __FILE__);\
                            printf(msg, ##__VA_ARGS__);\
                        }
#               endif   //__cplusplus
        
#			else //DEBUG_ENABLED

#				define PRINT_DEBUG(debugLevel, msg, ... ) 

#			endif //DEBUG_ENABLED

#			define PRINT_DEBUG0	PRINT_DEBUG
#			define PRINT_DEBUG1	PRINT_DEBUG
#			define PRINT_DEBUG2	PRINT_DEBUG
#			define PRINT_DEBUG3	PRINT_DEBUG
#			define PRINT_DEBUG4	PRINT_DEBUG
#			define PRINT_DEBUG5	PRINT_DEBUG
#			define PRINT_DEBUG6	PRINT_DEBUG
#			define PRINT_DEBUG7	PRINT_DEBUG

#		else
		// stg: the MS compiler will support variadic macros only from version 7 on. Until then, we will have to live with warnings about too few arguments to this macro
		#define PRINT_DEBUG0(debugLevel, msg) if (debugLevel <= CURRENT_DEBUG_LEVEL) {fprintf (debugFD, msg);fflush(debugFD);}
		#define PRINT_DEBUG1(debugLevel, msg, arg1) if (debugLevel <= CURRENT_DEBUG_LEVEL) {fprintf (debugFD, msg, arg1);fflush(debugFD);}
		#define PRINT_DEBUG2(debugLevel, msg, arg1, arg2) if (debugLevel <= CURRENT_DEBUG_LEVEL) {fprintf (debugFD, msg, arg1, arg2);fflush(debugFD);}
		#define PRINT_DEBUG3(debugLevel, msg, arg1, arg2, arg3) if (debugLevel <= CURRENT_DEBUG_LEVEL) {fprintf (debugFD, msg, arg1, arg2, arg3);fflush(debugFD);}
		#define PRINT_DEBUG4(debugLevel, msg, arg1, arg2, arg3, arg4) if (debugLevel <= CURRENT_DEBUG_LEVEL) {fprintf (debugFD, msg, arg1, arg2, arg3, arg4);fflush(debugFD);}
		#define PRINT_DEBUG5(debugLevel, msg, arg1, arg2, arg3, arg4, arg5) if (debugLevel <= CURRENT_DEBUG_LEVEL) {fprintf (debugFD, msg, arg1, arg2, arg3, arg4, arg5);fflush(debugFD);}
		#define PRINT_DEBUG6(debugLevel, msg, arg1, arg2, arg3, arg4, arg5, arg6) if (debugLevel <= CURRENT_DEBUG_LEVEL) {fprintf (debugFD, msg, arg1, arg2, arg3, arg4, arg5, arg6);fflush(debugFD);}
		#define PRINT_DEBUG7(debugLevel, msg, arg1, arg2, arg3, arg4, arg5, arg6, arg7) if (debugLevel <= CURRENT_DEBUG_LEVEL) {fprintf (debugFD, msg, arg1, arg2, arg3, arg4, arg5, arg6, arg7);fflush(debugFD);}
		//#define PRINT_DEBUG(debugLevel, msg)
#	endif //WIN32
#endif  //PRINT_DEBUG

/**
  * This macro is used to print messages irrespective of debug level.
  * These are messages are not for debugging, but for giving information
  * to the user.
  */
#	ifndef PRINT_INFO
#		if !defined(_WIN32) || defined(__CYGWIN__)
#               ifdef __cplusplus
#				    define PRINT_INFO(infoLevel, msg, ... )  \
                        if(infoLevel <= CURRENT_INFO_LEVEL) { \
                            time_t tmpTime = ::time(NULL); \
                            char* tmpTimeStr = ctime(&tmpTime); \
                            CHOP(tmpTimeStr);                  \
                            printf("[%s\b: Line:%d in file:%s] ", tmpTimeStr, __LINE__, __FILE__);\
                            printf(msg, ##__VA_ARGS__);\
                        }
#               else
#				    define PRINT_INFO(infoLevel, msg, ... )  \
                        if(infoLevel <= CURRENT_INFO_LEVEL) { \
                            time_t tmpTime = time(NULL); \
                            char* tmpTimeStr = ctime(&tmpTime); \
                            CHOP(tmpTimeStr);                  \
                            printf("[%s\b: Line:%d in file:%s] ", tmpTimeStr, __LINE__, __FILE__);\
                            printf(msg, ##__VA_ARGS__);\
                        }
#               endif   //__cplusplus
 


        
#			define PRINT_INFO0	PRINT_INFO
#			define PRINT_INFO1	PRINT_INFO
#			define PRINT_INFO2	PRINT_INFO
#			define PRINT_INFO3	PRINT_INFO
#			define PRINT_INFO4	PRINT_INFO
#			define PRINT_INFO5	PRINT_INFO
#			define PRINT_INFO6	PRINT_INFO
#			define PRINT_INFO7	PRINT_INFO

#		else
			// stg: the MS compiler will support variadic macros only from version 7 on. Until then, we will have to live with warnings about too few arguments to this macro
//#			define PRINT_INFO(infoLevel, msg)
#			define PRINT_INFO0(infoLevel, msg) if (infoLevel <= CURRENT_INFO_LEVEL) {fprintf (debugFD, msg);fflush(debugFD);}
#			define PRINT_INFO1(infoLevel, msg, arg1) if (infoLevel <= CURRENT_INFO_LEVEL) {fprintf (debugFD, msg, arg1);fflush(debugFD);}
#			define PRINT_INFO2(infoLevel, msg, arg1, arg2) if (infoLevel <= CURRENT_INFO_LEVEL) {fprintf (debugFD, msg, arg1, arg2);fflush(debugFD);}
#			define PRINT_INFO3(infoLevel, msg, arg1, arg2, arg3) if (infoLevel <= CURRENT_INFO_LEVEL) {fprintf (debugFD, msg, arg1, arg2, arg3);fflush(debugFD);}
#			define PRINT_INFO4(infoLevel, msg, arg1, arg2, arg3, arg4) if (infoLevel <= CURRENT_INFO_LEVEL) {fprintf (debugFD, msg, arg1, arg2, arg3, arg4);fflush(debugFD);}
#			define PRINT_INFO5(infoLevel, msg, arg1, arg2, arg3, arg4, arg5) if (infoLevel <= CURRENT_INFO_LEVEL) {fprintf (debugFD, msg, arg1, arg2, arg3, arg4, arg5);fflush(debugFD);}
#			define PRINT_INFO6(infoLevel, msg, arg1, arg2, arg3, arg4, arg5, arg6) if (infoLevel <= CURRENT_INFO_LEVEL) {fprintf (debugFD, msg, arg1, arg2, arg3, arg4, arg5, arg6);fflush(debugFD);}
#			define PRINT_INFO7(infoLevel, msg, arg1, arg2, arg3, arg4, arg5, arg6, arg7) if (infoLevel <= CURRENT_INFO_LEVEL) {fprintf (debugFD, msg, arg1, arg2, arg3, arg4, arg5, arg6, arg7);fflush(debugFD);}
#		endif
#endif //PRINT_INFO


#endif //_OCALA_DEBUG_H

