/** 
  * \file generic_proxy.cpp
  * \brief This file contains the main function used to launch the proxy.  In the case UNIX,
  * the program starts from this file.  In the case of Windows, win-main.cpp has the main function.
  * But still all the proxy related action starts in this file.
  */
#include "OCALA_version.h"
#include "Proxy.h"
#include "oci/OCI.h"
#include "Tunnel.h"
#include "common.h"
#include "ConfigFile.h"
#include "security/lib/SecurityManager.h"
#include "utils/utilFunctions.h"

#include <assert.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <pthread.h>

/** The following hash define controls if the OCALA proxy uses
 * Mike Freedman's OASIS service when functioning as a remote server proxy.
 * Comment it out, if you do not want to use OASIS.
 */
#define OASIS 1
#ifdef OASIS
	// was OASIS initialized successfully?
	bool oasisInitFlag;
    pthread_t oasisThread;
    int oasisFD;
    uint32_t oasisSrvId;
    uint16_t oasisPort;
#endif //OASIS


extern "C" {
#	include "netwrap.h"
}

#ifndef _WIN32
	#include <unistd.h>
    #include <sys/types.h>
    #include <sys/stat.h>
    #include <netinet/in.h>
    #include <arpa/inet.h>
#else
extern "C" {
	#include "getopt.h"
}	
	void startGenericProxy_Windows();
	
#endif

#include <ctype.h>

#include "Params.h"
#include "generic_proxy.h"

#ifndef _WIN32
    #include <signal.h>
#endif

#ifdef _WIN32
    void* proxySelect(void *args);
    void* tunnelReader(void *args);

#else
    /** Should the proxy be started as a daemon */
    bool isDaemonFlag;
    
    /** Should the proxy be started in PlanetLab (Alpine4Linux) mode? */
    bool isPlanetLabFlag;
   
    void startGenericProxy_Linux();    
    void proxy_cleanup (int signum);
#endif

void printUsage();
void processCommandLineArgs(int argc, char* argv[]);

/** This will store the name of the config file to be used */
char configFileName[MAX_FILENAME_LEN];

/** Should crash recovery be attempted? */
bool recoverFromCrash;

ConfigFile* configFile = NULL;
Tunnel *tunnel = NULL;

#ifdef _WIN32
/** This function is called from win32/win-main.cpp
 * when the proxy is used in Windows.  This function 
 * can be called as a service or as a console based 
 * application.
 */
int init_generic_proxy(int argc, char* argv[]) {
#else
int main (int argc, char* argv[]) {

    signal (SIGPIPE, proxy_cleanup);
    signal (SIGINT, proxy_cleanup);
    signal (SIGHUP, proxy_cleanup);
    signal (SIGSEGV, proxy_cleanup);
    signal (SIGILL, proxy_cleanup);
    signal (SIGABRT, proxy_cleanup);
    signal (SIGTERM, proxy_cleanup);

#endif

#ifdef _WIN32
	// Initialize the winsock networking
	if (nw_init() != 0) {
		PRINT_INFO0 ( INFO_LEVEL_FATAL_ERROR,
						"Unable to initialize winsock networking. Exiting.\n"
						);
		EXIT_ON_ERROR; 
	}
#endif

#ifndef _WIN32
#	ifdef BENCHMARK
		signal (SIGINT, DumpEventInfo);
#	endif
#endif
   
    PRINT_INFO2(
                    INFO_LEVEL_MINIMAL , 
                    "\nStarting generic proxy (OCALA version %d.%d) ...\n",
                    OCALA_MAJOR_VERSION, OCALA_MINOR_VERSION
            ); 

    processCommandLineArgs(argc, argv);

#if !defined(_WIN32) && !defined(__APPLE__)
    // Reference: http://www.linuxprofilm.com/articles/linux-daemon-howto.html
    if (isDaemonFlag) {
        // Start the proxy as a daemon.

        //Fork off the parent process
        pid_t pid = fork();
        if (pid < 0) {
            PRINT_INFO0 ( INFO_LEVEL_FATAL_ERROR,
                            "Forking failed: Unable to start OCALA proxy as a daemon.  Please run without -d option.\n"
                        );
            EXIT_ON_ERROR;
        }
        // If we got a good PID, then we can exit the parent process.
        if (pid > 0) {

            printf ( "Starting OCALA_proxy as daemon.");
            exit(EXIT_SUCCESS);
        }

        // The child process execution continues from here.
        
        // Change the file mode mask
        // By setting the umask to 0, we will have full access to the files generated by the daemon. 
        // Even if you aren't planning on using any files, it is a good idea to set the umask here 
        // anyway, just in case you will be accessing files on the filesystem.
        umask(0);

        // Create a new SID for the child process
        // From here, the child process must get a unique SID from the kernel in order to operate. 
        // Otherwise, the child process becomes an orphan in the system. 
        pid_t sid = setsid();
        if (sid < 0) {
            PRINT_INFO0 (INFO_LEVEL_FATAL_ERROR,
                            "Unable to get a new SID for OCALA_Proxy child process. Exiting.\n"
                        );
            EXIT_ON_ERROR;
        }

/*
        // Change the current working directory to /tmp.
        // The logs will be created in /tmp so that we don't have to worry about cleaning up.
        if ((chdir("/tmp")) < 0) {
            PRINT_INFO0 (INFO_LEVEL_FATAL_ERROR,
                            "Unable to change working directory to \\tmp. Exiting.\n"
                        );
            EXIT_ON_ERROR;
        } 
*/        
        // Close the standard input as we won't need it.
        close(STDIN_FILENO);
        
        // Redirect stdout and stderr to our debug file.
        close(STDOUT_FILENO);
        stdout = fopen ("debug_OCALA.txt", "w");
        close(STDERR_FILENO);
        stderr = fopen ("debug_OCALA.txt", "a");
    } //end daemon
#endif //_WIN32
    
    configFile = new ConfigFile (configFileName, TAG_PROXY_CONFIG_FILE);
    
    Proxy::proxy = new Proxy(configFile, recoverFromCrash);
 
    // Instantiate the tunnel (packet capturing device)
    tunnel = new Tunnel(Proxy::proxy, NULL /*proxy->getNewRandomFakeIP()*/);
#ifndef _WIN32
	tunnel->setPlanetlabFlag (isPlanetLabFlag);
#endif

    // Just sets the pointer to the tunnel in the proxy.
   Proxy:: proxy->setTunnel (tunnel);
    
    
    Proxy::proxy->initOverlays(recoverFromCrash);
    Proxy::proxy->initURLResolver();
   
    
    if (Proxy::proxy->functionAsServerProxy) {
        // Functioning as a remote server proxy
        // We don't need to initialize the tunnel if we are using our own
        // custom software NAT.
        if ( ! Proxy::proxy->mySoftwareNATFlag) {
            // Using Linux NAT for remote server proxy.
            // initTunnel should be called only after initURLResolver because
            // the URLResolver might add some dest/src IP/ports whose packets need to be 
            // captured by the tunnel.  It is much more convenient to do the packet
            // capturing initializing in one go.
            tunnel->initTunnel();

            // proxy->initTunnel should be called only after tunnel->initTunnel
            // as it uses the tunnel fds (adds them to select loop)
            Proxy::proxy->initTunnel();
        } 
    } else {

        // functioning as a local client proxy
    
        // initTunnel should be called only after initURLResolver because
        // the URLResolver might add some dest/src IP/ports whose packets need to be 
        // captured by the tunnel.  It is much more convenient to do the packet
        // capturing initializing in one go.
        tunnel->initTunnel();

        // proxy->initTunnel should be called only after tunnel->initTunnel
        // as it uses the tunnel fds (adds them to select loop)
        Proxy::proxy->initTunnel();
    }
    
    Proxy::proxy->printOverlayList();
    Proxy::proxy->securityManager->printCertificateList();

    // initMyNames should be called only after the Tunnel is completely
    // initialized (i.e. initTunnel() has been called).  This is because
    // initMyNames might involve sending out packets (for eg: insert trigger
    // pkts in i3).  The Tunnel constructor called earlier had set up some
    // pkt capturing state.  It is not a good idea to send pkts from our program
    // before completely initializing the tunnel packet capturing rules.
    Proxy::proxy->initMyNames ();

    
    Proxy::proxy->oci->initJournalManager(recoverFromCrash);

#ifdef _WIN32
    //startGenericProxy_Windows() must be called after init_generic_proxy().
    //This call is made from win32/win-main.cpp file.
#else 
    startGenericProxy_Linux();
#endif

	return 1;
}


#ifdef _WIN32
int resumeGenericProxy() {

	// Proxy is not active - starting up
	setProxyStatusCode (5);

	// restore the normal DNS settings
	// Some init activities might depend on the DNS server value being correctly set.
	restoreNameServerSettings();

    //sleep(30);	
    //Beep(1000,10);
    PRINT_INFO0(INFO_LEVEL_MINIMAL , "\nResuming generic proxy ...\n");

    ConfigFile* configFile = new ConfigFile (configFileName, TAG_PROXY_CONFIG_FILE);
    
	// When we are resuming the proxy after a laptop wakes up from sleep,
	// we need to reload all the saved state - just like in the case of a crash.

	recoverFromCrash = true;
    Proxy::proxy = new Proxy(configFile, recoverFromCrash);
  
    // We are resuming the proxy here.
    // We don't have re instantiate the tunnel.
    tunnel->setProxy(Proxy::proxy);

    Proxy::proxy->setTunnel(tunnel);
    
    Proxy::proxy->initOverlays(recoverFromCrash);
    Proxy::proxy->initURLResolver();
    Proxy::proxy->printOverlayList();

    Proxy::proxy->initTunnel();

	tunnel->initNameServerRedirector();

    Proxy::proxy->printOverlayList();
    Proxy::proxy->securityManager->printCertificateList();

    // initMyNames should be called only after the Tunnel is completely
    // initialized (i.e. initTunnel() has been called).  This is because
    // initMyNames might involve sending out packets (for eg: insert trigger
    // pkts in i3).  The Tunnel constructor called earlier had set up some
    // pkt capturing state.  It is not a good idea to send pkts from our program
    // before completely initializing the tunnel packet capturing rules.
	Proxy::proxy->initMyNames ();

    
    Proxy::proxy->oci->initJournalManager(recoverFromCrash);


	// Proxy is now active.
	setProxyStatusCode (1);

    startGenericProxy_Windows();

	return 1;
}

#endif


#ifdef _WIN32

/**
 * This function is used only in Windows.
 * Starts separate threads for listening to tunnel and overlays.
 */
void startGenericProxy_Windows() {
    //A separate thread for the tunnel might be needed in the case of windows. 
    
    // The position of the next 2 statements affects deadlock avoidance.
    // So be careful.
    pthread_t selectThread;
    int selectRetVal = pthread_create(
                                        &selectThread, NULL, 
                                        proxySelect, (void *) NULL
                                    );
    
    if (selectRetVal != 0) {
        //some error occurred during the creation of the select thread
        perror("Error while creating select thread");
    }

    pthread_t tunnelReaderThread;
    int tunnelReaderRetVal = pthread_create(
                                    &tunnelReaderThread, NULL, 
                                    tunnelReader, (void *) NULL
                                );
    if (tunnelReaderRetVal != 0) {
        //some error occurred during the creation of the tunnel reader thread.
        perror("Error while creating tunnel reader thread");
    }

    pthread_join(tunnelReaderThread, NULL);
    pthread_join(selectThread, NULL);

    //pthread_join(journalThread, NULL);
    
    pthread_exit(NULL); //so that other threads are not killed.

}

#else
// Linux specific code

#   ifdef OASIS
/**
 *  This function continuously listens for TCP connections on a particular port.
 *  It must be run in a separate thread.
 */
void* oasisListen (void* args) {
   

    // Create the TCP socket for OASIS to listen on
    if ((oasisFD = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
        
        PRINT_INFO1 (INFO_LEVEL_FATAL_ERROR,
                    "Unable to open OASIS socket: %s.\n",
                    strerror (errno)
                );
        EXIT_ON_ERROR;
    }

    
    struct sockaddr_in my_addr;    // my address information
    my_addr.sin_port = htons (oasisPort);     // short, network byte order
    my_addr.sin_addr.s_addr = INADDR_ANY; // automatically fill with my IP
    memset(&(my_addr.sin_zero), '\0', 8); // zero the rest of the struct

    // Bind the OASIS socket
    if (bind(oasisFD, (struct sockaddr *)&my_addr, sizeof(struct sockaddr)) == -1) {
        PRINT_INFO1 (INFO_LEVEL_FATAL_ERROR,
                    "Unable to bind OASIS socket: %s. Exiting.\n",
                    strerror (errno)
                );
        EXIT_ON_ERROR;

    }

	int yes = 1;
    if (setsockopt(oasisFD, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1) {
        PRINT_DEBUG1(
                        DEBUG_LEVEL_MINIMAL,
                        "Unable to set SO_REUSEADDR socket option for the oasis socket: %s\n",
                        strerror (errno)
                    );
    }


    // Listen on the OASIS socket.
    if (listen(oasisFD , 10 /* backlog*/) == -1) {
        PRINT_INFO1 (INFO_LEVEL_FATAL_ERROR,
                    "Unable to listen on OASIS socket: %s.\n",
                    strerror (errno)
                );
        EXIT_ON_ERROR;
    }
    
    PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
                    "OASIS listening on port %d.\n",
                    ntohs (my_addr.sin_port)
                );
    
    // Start the indefinite accept-wait loop.
   	while(1) {
        socklen_t sin_size = sizeof(struct sockaddr_in);
        
        // FD associated with the current connection
        int currFD;
    	struct sockaddr_in their_addr; // connector's address information

        if ((currFD = accept(oasisFD, (struct sockaddr *) &their_addr, &sin_size)) == -1) {
            PRINT_INFO2 (INFO_LEVEL_WARNING,
                    "Error while accepting connection on OASIS socket from %s : %s.\n",
                    inet_ntoa (their_addr.sin_addr),
                    strerror (errno)
                );
           	continue;
        }

        PRINT_DEBUG2 (DEBUG_LEVEL_VERY_VERBOSE,
                    "OASIS got connection from %s:%d.\n",
                    inet_ntoa (their_addr.sin_addr),
                    ntohs (their_addr.sin_port)
                );
        
        // Write the status info to the OASIS daemon which initiated the connection.
        struct {
            u_int32_t srvid;
            u_int32_t load;
            u_int32_t cap;
        } oasserv_localapp_res; 

        // TODO: Use some global load variables.  Calculate load properly.
        uint32_t proxyCurrentLoad = 1;
        uint32_t proxyMaxLoad = 10;
        
        oasserv_localapp_res.srvid = htonl (oasisSrvId);
        oasserv_localapp_res.load = htonl (proxyCurrentLoad);
        oasserv_localapp_res.cap = htonl (proxyMaxLoad);

        write ( currFD, &oasserv_localapp_res, sizeof (oasserv_localapp_res));

        // We are done with the current connection.
        close (currFD);
        
	}
}

/**
  * This function is called in order to stop listening on the OASIS socket.
  * The thread is stopped; the socket is closed.
  */
void stopOASIS () {
   
    if (oasisInitFlag) {
	
    	int retVal = pthread_kill (oasisThread, SIGKILL);
    	if (retVal != 0) {
        	PRINT_INFO1 (INFO_LEVEL_MINIMAL,
                    "Unable to stop OASIS thread : %s.\n",
                    strerror (errno)
                );
    	}
		PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL,
						"\n\n\nClosing OASIS fd\n\n\n"
					);
    	close (oasisFD);
	}
}
#   endif  //OASIS


/** When this function is called, the proxy starts listening for packets
 * captured by the tunnel as well as those received by the overlay.
 * Everything runs in a single thread.  
 * This function is to be used only in *nix.
 */
void startGenericProxy_Linux() {
    //NOT WINDOWS

#   ifdef OASIS
    // Read the oasis configuration from the config file
    assert (configFile != NULL);
    char tmpParam  [MAX_PARAM_LEN];
	oasisInitFlag = false;
    configFile->readStringAttrib("//OASIS", "enable", tmpParam, "no");

    if (strcasecmp (tmpParam, "yes") == 0) {
        // OASIS enabled

        if (Proxy::proxy->functionAsServerProxy || Proxy::proxy->functionAsClientRemoteProxy) {
        
            if ( ! configFile->readUShortAttrib("//OASIS", "listenPort", &oasisPort, (short unsigned int) 0)) {
                PRINT_INFO (INFO_LEVEL_FATAL_ERROR,
                            "OASIS is enabled.  listenPort must be specified.\n"
                        );
                EXIT_ON_ERROR;

            }
            if (! configFile->readUIntAttrib("//OASIS", "serviceId", &oasisSrvId, 0)) {
                PRINT_INFO (INFO_LEVEL_FATAL_ERROR,
                            "OASIS is enabled.  serviceId must be specified.\n"
                        );
                EXIT_ON_ERROR;
            }
 
            PRINT_DEBUG (DEBUG_LEVEL_MINIMAL,
                        "OASIS: listenPort=%d serviceId = %d.\n",
                        oasisPort, oasisSrvId
                    );

            int oasisThreadRetVal = pthread_create(
                                        &oasisThread, NULL, 
                                        oasisListen, (void *) NULL
                                    );
    
            if (oasisThreadRetVal != 0) {
                PRINT_INFO1 (INFO_LEVEL_FATAL_ERROR,
		            "Error while creating oasis listening threadselect thread: %s",
		            strerror (errno)
	                );
  	            EXIT_ON_ERROR;
            }
			oasisInitFlag = true;
        } else {
            PRINT_INFO0 (INFO_LEVEL_WARNING,
                        "Ignoring OASIS initialization because the proxy is not function as a remote server or remote client proxy.\n"
                    );
        }
    }
#    endif //OASIS


    // The single thread of execution - all action takes place here.
    Proxy::proxy->coreSelect();    
    
}

#endif // Linux specific code
    

#if defined (_WIN32)
    /**
     * This function is just a wrapper to call pthread_create.
    */
    void* proxySelect(void *args) {
		
		// At least in Windows, every thread needs to call its own
		// srand.  Otherwise the rand () called in this thread
		// always return the same value.
		// I learned this by trial and error.
		// - Dilip, March 18, 2006
		srand (time(NULL));

        Proxy::proxy->coreSelect();

		return NULL;
    }


    /** This function is just a wrapper to call pthread_create */
void* tunnelReader(void *args) {
	srand (time(NULL));

	tunnel->tunReader();
	return NULL;
}
#endif

/**
  * This function processes command line arguments.  
  * If incorrect number of arguments are used, a message showing the correct usage
  * is printed and the proxy exits.
  */
void processCommandLineArgs(int argc, char* argv[]) {
    //Set the default values   
    strncpy(configFileName, DEFAULT_CONFIG_FILENAME, MAX_FILENAME_LEN);    
    recoverFromCrash = true;

#ifndef _WIN32
	isDaemonFlag = false;
    isPlanetLabFlag = false;
#endif

    opterr = 1;
    char choice;

    while ((choice = getopt (argc, argv, ":Hf:cdp")) != -1) {
        switch (choice) {
            case 'c': //cold start
                recoverFromCrash = false;
            break;
           
            case 'f': //explicitly specify the configuration file to use
	            strcpy(configFileName, optarg);
            break;
#ifndef _WIN32
            case 'd':   // start as a daemon
                isDaemonFlag = true;
            break;

            case 'p': // start proxy in planetlab mode, uses Alpine4Linux 
                isPlanetLabFlag = true;
            break;
#endif            
	        case ':':   //DILIP- TODO This case seems unnecessary. Verify
	    	    if (optopt == 'f') { 
		            PRINT_INFO1(INFO_LEVEL_WARNING, 
			                    "Proxy config file not specified. "
				                "Using default config file %s.\n",
				                DEFAULT_CONFIG_FILENAME
			                );
		        }
	        break;
	    
	        case '?':
                if (isprint (optopt)) {
                    PRINT_INFO1(INFO_LEVEL_WARNING, "Unknown option `-%c'.\n", optopt);
	            } else {
                    PRINT_INFO1(INFO_LEVEL_WARNING, "Unknown option `\\x%x'.\n", optopt);
	            }
    	        printUsage();	       
	            EXIT_ON_ERROR;
	        break;

	        case 'H':
    		    printUsage();	       
		        EXIT_ON_ERROR;
	         break;
	   
	        default:
	   	        printUsage();
		        EXIT_ON_ERROR;
        }//end switch
    }
}

/** 
 * This function just prints the command line options associated
 * with the proxy.
 */
void printUsage() {
    //incorrect number of command line arguments are specified.
    PRINT_INFO0(INFO_LEVEL_MINIMAL, 
            "\nUsage: generic_proxy [-c] [-f ConfigFileName]"
            "\n\n-f ConfigFileName\t ConfigFileName contains the overlay independent configuration information. "
	                                    "`proxy-config.xml' is used by default if this option is omitted."
            "\n\n-c\t\t The proxy by default loads previous state from its journal while starting up.  This option tells the proxy to start with a clean slate."
	    "\n"
            "\n\n-d\t\t Start the proxy as a daemon."
            "\n\n-p\t\t Start the proxy in PlanetLab (Alpine4Linux) mode."
	    );
}



#if !defined (_WIN32) 
/**
  * This function is called whenever the proxy receives signal
  * which forces it to quit.
  */
void proxy_cleanup (int signum) {

    char tmpStr[30];
    
    switch (signum) {
        
        case SIGSEGV:
            strcpy (tmpStr, "SIGSEGV");
        break;

        case SIGABRT:
            strcpy (tmpStr, "SIGABRT");
        break;

        case SIGPIPE:
            strcpy (tmpStr, "SIGPIPE");
        break;

        case SIGHUP:
            strcpy (tmpStr, "SIGHUP");
        break;

        case SIGINT:
            strcpy (tmpStr, "SIGINT");
        break;

        default:
            strcpy (tmpStr, "Default");
        break;
        
    }
    PRINT_DEBUG (
                    DEBUG_LEVEL_MINIMAL,
                    "Stopping proxy on receiving signal %d  (%s) \n ",
                    signum, tmpStr
                );
    
  
    Proxy::proxy->tunnel->restoreIPTables();

    Proxy::proxy->stopProxy(true /* Immediate, don't wait for lock */);

#ifdef OASIS
    stopOASIS ();
#endif
    
    exit (0);
}

#endif



#ifdef _WIN32
	
/**
 * Set the specified status code in LOCAL_MACHINE/SOFTWARE/OCALA
 * registry to indicate the current status of the proxy.
 * @return FALSE if unable to set the status code.
 */
BOOL setProxyStatusCode (DWORD statusCodeToSet) {

	LONG res;

	DWORD valType;
	char stringVal [MAX_REGISTRY_STR_VALUE_LEN];
	DWORD stringValSize = MAX_REGISTRY_STR_VALUE_LEN;
	DWORD dwordVal;
	DWORD dwordValSize = sizeof (DWORD);

	// Open the Software Key
	HKEY keySoftware;
	res = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
							 "Software",
							 0, /* Reserved must be 0 */
							 KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS,
							 &keySoftware
	
					);
	if (res != ERROR_SUCCESS) {
		PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL,
						"Unable to open the HKEY_LOCAL_MACHINE/Software key.\n"
					);
		return FALSE;
	}

	// Open the OCALA key.
	HKEY keyOCALA;	
	res = RegOpenKeyEx(		keySoftware,
							 "OCALA",
							 0, /* Reserved must be 0 */
							 KEY_SET_VALUE,
							 &keyOCALA
						);

	if (res != ERROR_SUCCESS) {

		PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
						"Unable to open the HKEY_LOCAL_MACHINE/Software key/%s key.  OCALA probably never started.\n",
						"OCALA"
					);

		// Create it
		
		res = RegCreateKeyEx(
								keySoftware,
								"OCALA",
								0,
								NULL,
								REG_OPTION_NON_VOLATILE,
								KEY_ALL_ACCESS,
								NULL,
								&keyOCALA,
								NULL
							);
		if (res != ERROR_SUCCESS) {
			PRINT_INFO1 ( INFO_LEVEL_WARNING,
							"Unable to create HKEY_LOCAL_MACHINE\\Software\\%s.\n", 
							"OCALA");
			RegCloseKey (keySoftware);
			return FALSE;
		}
		if (!MakeRegKeyAccessible (keyOCALA, "OCALA")) {
			
			PRINT_DEBUG1 ( DEBUG_LEVEL_WARNING,
							"Unable to make the key %s accessible to all.\n",
							"OCALA"
						);
			RegCloseKey (keyOCALA);
			RegCloseKey (keySoftware);

			return FALSE;
		}
	}

	res = RegSetValueEx(
							keyOCALA,
							"statusCode",
							0,
							REG_DWORD,
							(BYTE*) &statusCodeToSet,
							sizeof (DWORD)
						);


	RegCloseKey (keyOCALA);
	RegCloseKey (keySoftware);

	if (res != ERROR_SUCCESS) {
		return FALSE;
	} else {
		return TRUE;
	}

}

#endif

//#endif //DYN_LOAD_OCD
