#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <strings.h>

#include "BlackList.h"
#include "../common.h"
#include "../debug.h"
#include "../utils/utilFunctions.h"

#ifndef TEST_MODE 
#   include "../TimerEvent.h"
#   include "../Proxy.h"
#endif //TEST_MODE

/**
  * Construct the BlackList object.
  * @param forbiddenListFilename The file containing the list of forbidden URLS and 
  * URL suffixes, one per line.  
  * @param reloadCheckPeriod How often (in seconds) should we check if forbiddenListFilename is to be reloaded.
  * @param reloadCheckFile This file should contain only one line.  If the line is "yes",
  * the forbiddenListFilename is reloaded during the periodic check.
  */
#ifndef TEST_MODE
BlackList::BlackList (  const char* forbiddenListFilename, 
                        const char* reloadCheckFilename, 
                        Proxy* proxy, 
                        uint16_t reloadCheckPeriod = BLACKLIST_DEFAULT_RELOAD_CHECK_PERIOD
                    ) {

    this->reloadCheckFilename = string (reloadCheckFilename);
    this->reloadCheckPeriod = reloadCheckPeriod;
    this->mainLoop = proxy;
#else 
BlackList::BlackList (  const char* forbiddenListFilename) {
#endif //TEST_MODE

    this->forbiddenListFilename = string (forbiddenListFilename);

    this->forbiddenMap = NULL;

    if (! loadMap()) {

        PRINT_INFO2 ( INFO_LEVEL_FATAL_ERROR,
                        "BlackList First Time: Unable to open %s for reading: %s.\n",
                        this->forbiddenListFilename.c_str(),
                        strerror (errno)
                    );
        EXIT_ON_ERROR;
    } else {
        PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
                        "Blacklist loaded from %s.\n",
                        this->forbiddenListFilename.c_str()
                    );
        dumpForbiddenList();
    }

#ifndef TEST_MODE
    TObjectFunctor<BlackList> *reloadFunctor = 
                        new TObjectFunctor<BlackList> (this, &BlackList::doPeriodicReload);
    TimerEvent* reloadTimer = TimerEvent::create(
                            reloadCheckPeriod ,    //seconds
                            0,    //microseconds
                            (TFunctor *) reloadFunctor,
                            true
                        );
    mainLoop->addTimerEvent(reloadTimer);
#endif //TEST_MODE
}

#ifndef TEST_MODE
/**
  * This function is periodically called in order to reload the black list.
  */
void BlackList::doPeriodicReload () {

    // See if a reload is necessary.
    bool reloadNecessaryFlag = false;
    FILE* fd = fopen (reloadCheckFilename.c_str(), "r");
    if (fd != NULL) {
        char tmpLine [100];
        if (fgets (tmpLine, 100, fd) != NULL) {
            chomp (tmpLine);
            if ( strcasecmp (tmpLine, "yes") == 0) {
                reloadNecessaryFlag = true;
            }
        }
        fclose (fd);
    }

    if (reloadNecessaryFlag) {
        if (! loadMap()) {

            PRINT_INFO2 ( INFO_LEVEL_MINIMAL,
                        "Unable to reload BlackList: Unable to open %s for reading: %s.\n",
                        this->forbiddenListFilename.c_str(),
                        strerror (errno)
                    );
        } else {
            PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
                        "Blacklist reloaded from %s.\n",
                        this->forbiddenListFilename.c_str()
                    );
            dumpForbiddenList();
        }
    }
    
    // schedule the next reload
    TObjectFunctor<BlackList> *reloadFunctor = 
                        new TObjectFunctor<BlackList> (this, &BlackList::doPeriodicReload);
    TimerEvent* reloadTimer = TimerEvent::create(
                            reloadCheckPeriod ,    //seconds
                            0,    //microseconds
                            (TFunctor *) reloadFunctor,
                            true
                        );
    mainLoop->addTimerEvent(reloadTimer);

}
#endif //TEST_MODE

/** This function loads the URLs/patterns from the forbiddenListFile.
  * @return true if the load was successful, else returns false.
  */
bool BlackList::loadMap () {
    // Read in the forbidden URLs and URL suffixes
    FILE* fd = fopen (forbiddenListFilename.c_str(), "r");
    if (fd == NULL) {
        PRINT_INFO2 ( INFO_LEVEL_FATAL_ERROR,
                        "BlackList: Unable to open %s for reading: %s.\n",
                        forbiddenListFilename.c_str(),
                        strerror (errno)
                    );
        return false;
    }

    char tmpURL [MAX_URL_LEN];

    list<string> tmpList;
    minPatternLen = 1000000;
    while ( fgets (tmpURL, MAX_URL_LEN, fd) != NULL) {
        chomp (tmpURL);
        if (!isBlank (tmpURL)) {

            // Non-blank URL/pattern
            // first store it in a temporary list
            tmpList.push_back (string(tmpURL));
            
            // Part of finding the shortest URL/pattern
            int currURLLen = strlen (tmpURL);
            if (currURLLen < minPatternLen) {
                minPatternLen = currURLLen;
            }
        }
    }
    fclose (fd);

    // Now populate forbiddenMap  - first use a copy.
    map <string, list<string>*>* tmpForbiddenMap = new map<string, list<string>*>;
    
    list<string>::iterator itr = tmpList.begin();
    while (itr != tmpList.end()) {
        string currURL = *itr;
        int currURLLen = currURL.length();
        string currSuffix = currURL.substr (currURLLen - minPatternLen);
        
        list<string>* currList = NULL;
        map<string, list<string>*>::iterator mapItr = tmpForbiddenMap->find (currSuffix);
        if (mapItr == tmpForbiddenMap->end()) {
            // No entry for currSuffix.  Create a new one.
            currList = new list<string>;
            tmpForbiddenMap->insert ( pair<string, list<string>*> (currSuffix, currList) );

        } else {
            currList = mapItr->second;
        }
       
        // Push the URL into the associated list.
        currList->push_back (currURL);

        itr++;
    }

    // Delete the existing forbiddenMap
    if (forbiddenMap != NULL) {
        delete (forbiddenMap);
    } 
    // Set the new map
    forbiddenMap = tmpForbiddenMap;

    return true;
}

/**
  * Dumps the list of forbidden URLs/URL suffixes to screen.
  */
void BlackList::dumpForbiddenList () {
    PRINT_INFO0 (INFO_LEVEL_MINIMAL, "BlackList: The forbidden list:\n");
    map<string, list<string>*>::iterator mapItr = forbiddenMap->begin();
    while (mapItr != forbiddenMap->end()) {
       
        string tmpStr = mapItr->first + string(" : ");
        list<string>* currList = mapItr->second;
        list<string>::iterator listItr = currList->begin();
        while (listItr != currList->end()) {
            tmpStr += *listItr + string (" ");
            listItr++;
        }
        PRINT_INFO0 (INFO_LEVEL_MINIMAL, "%s\n", tmpStr.c_str());
        mapItr++;
    }
}

/**
  * Returns true if the specified url is in the forbidden list exactly,
  * or one of the entries of the forbidden list is a suffix of the 
  * specified url.
  */
bool BlackList::isBlackListed (const char* url) {

    string urlStr = url;
    string urlSuffix = urlStr.substr (urlStr.length() - minPatternLen);
    
    map<string, list<string>*>::iterator mapItr = forbiddenMap->find (urlSuffix);
    if (mapItr == forbiddenMap->end()) {
        // Definitely not blacklisted.
        return false;
    }

    // Full comparison against all patterns with the specified suffix necessary.
    list<string>* currList = mapItr->second;
    list<string>::iterator listItr = currList->begin();
    while (listItr != currList->end()) {

        if ( isSuffix (url, (*listItr).c_str()) ) {
            return true;
        }
        listItr++;
    }

    // Nothing matched.  So this url is not black listed.
    return false;
}

/**
  * Destructor.
  */
BlackList::~BlackList() {

    map<string, list<string>*>::iterator itr = forbiddenMap->begin();
    while (itr != forbiddenMap->end()) {
        
        list<string>* currList = itr->second;
        if (currList != NULL) {
            delete (currList);
        } else {
            PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL,
                        "A list inside the BlackList forbidden Map should not be NULL.  Something fishy here.  Aborting.\n"
                    );
            assert (false);
        }
        itr++;
    }
}   
