/**
 *  \file LSIPv4_OCD.cpp
 *  \brief This file defines the Legacy Server IP module.  This is a special
 *  OCD module which forms the core of the remote server proxy.
 */
#include <stddef.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <assert.h>

#include "../common.h"
#include "../utils/utilFunctions.h"

#include "../ocd/OCD.h"
#include "../oci/OCI.h"
#include "../oci/OCIPacketBlock.h"
#include "../Proxy.h"
#include "../Tunnel.h"
#include "../DNSPacket.h"
#include "../debug.h"

#include "LSIPv4_OCD.h"
#include "LSIPv4_OCDState.h"
#include "LSIPXMLTags.h"
#include "MyNAT.h"
#include "BlackList.h"
#include "../oci/OCIState.h"

LSIPv4_OCD::LSIPv4_OCD (string type, string id, string desc) {

    this->type = type;
    this->id = id;
    this->description = desc;
}

LSIPv4_OCD::LSIPv4_OCD (OCD *o) {

    this->type = o->type;
    this->id = o->id;
    this->description = o->description;
    this->configFileName = o->configFileName;
    this->urlSuffixes = o->urlSuffixes;
    this->handleLegacyURLs = o->handleLegacyURLs;
}

void LSIPv4_OCD::init (Proxy *proxy, OCI *oci, bool recoverFromCrash) {
    this->proxy = proxy;
    this->oci = oci;
    this->tunnel = proxy->tunnel;

    if ( ! proxy->functionAsServerProxy) {
        PRINT_INFO (
                        INFO_LEVEL_WARNING,
                        "LSIPv4_OCD module loaded when Proxy not set to function as a remote server proxy.\n"
                    );
    } else {

		// functioning as server proxy

        // Load black list
        char blackListFilename [MAX_FILENAME_LEN];
       if ( readStringAttrib (LSIP_ATTRIB_BLACKLIST_FILE, blackListFilename, false)) {
            char reloadCheckFilename [MAX_FILENAME_LEN];
            reloadCheckFilename [0] = '\0';
            readStringAttrib (LSIP_ATTRIB_BLACKLIST_RELOAD_CHECK_FILE, reloadCheckFilename, false);
            int reloadCheckPeriod;
            readIntAttrib (LSIP_ATTRIB_BLACKLIST_RELOAD_CHECK_PERIOD, &reloadCheckPeriod, BLACKLIST_DEFAULT_RELOAD_CHECK_PERIOD);
            
            blackList = new BlackList (blackListFilename, reloadCheckFilename, proxy, reloadCheckPeriod);
            blackList->dumpForbiddenList();
       } else {
            blackList = NULL;
       }

        if (proxy->mySoftwareNATFlag) {
            // we are using our own custom software NAT instead of the Linux NAT
	
	        /**
	        * natPortCount ports between natPortRangeStart and natPortRangeEnd are reserved
	        * for the use of the LSIP's software NAT.
	        */
	        int natPortRangeStart;
	        int natPortRangeEnd;
	        int natPortCount;       
	        readIntAttrib (LSIP_ATTRIB_NAT_PORT_RANGE_START, &natPortRangeStart, 10000);
		    readIntAttrib (LSIP_ATTRIB_NAT_PORT_RANGE_END, &natPortRangeEnd, 20000);
		    readIntAttrib (LSIP_ATTRIB_NAT_PORT_COUNT, &natPortCount, 200);
		    PRINT_INFO (	
						INFO_LEVEL_MINIMAL,
						"LSIP using software NAT with %d ports between %d and %d.\n",
						natPortCount, natPortRangeStart, natPortRangeEnd
					);

	
		    // Tell the tunnel to capture packets with destination ports in the specified range
		    // We need a continuous range of port numbers.  Suppose some of the ports in the range
		    // do not belong to our applications, capturing those packets will break the applications which use them.
		    // TODO TODO TODO
		    proxy->setNATPortRange (natPortRangeStart, natPortRangeEnd);
	
		    myNAT = new MyNAT (proxy, this, natPortRangeStart, natPortRangeEnd, natPortCount);
        }
	}
}

OCDHandle LSIPv4_OCD::setup (
                const string endPointName, 
                int endPointNameType,
                char *optionBlocks, 
                int optionLen,
                OCIHandle ociHandle,
                OCI *oci ) {


    if (isSuffix(endPointName.c_str(), "arpa")) {
        PRINT_DEBUG (
			DEBUG_LEVEL_MINIMAL,
                        "Dropping connection associated with reverse DNS lookup to : %s\n",
                        endPointName.c_str()
                    );
        return NULL_HANDLE;
    }

    if (blackList != NULL && blackList->isBlackListed (endPointName.c_str()) ) {
        PRINT_DEBUG (
                DEBUG_LEVEL_MINIMAL,
                "Dropping connection for blacklisted URL %s.\n",
                endPointName.c_str()
            );
        return NULL_HANDLE;
    }

    // Check if the OCI statte corresponding to ociHandle already has a 
    // a valid rightOCDHandle.  If yes, we need to re-use the the OCDState
    // corresponding to that rightOCDHandle.  If not, we might forget to
    // free the old OCDState.  Worse, we have to do extra work to change
    // the MyNAT mappings
    // HACK!!!! - Dilip: change OCD:setup() to incorporate any existing OCDHandle info.

    OCIState* tmpOCIState = oci->getStateInfo (ociHandle);
    assert (tmpOCIState != NULL);

    LSIPv4_OCDState* ocdState = NULL;
    OCDHandle currLayerHandle = NULL_HANDLE;
    
    if (tmpOCIState->rightOCD == this && tmpOCIState->rightOCDHandle != NULL_HANDLE) {
        // The OCIState invoking the setup call already has an LSIP OCD State.
        // We re-use it.  Note that we are not sharing LSIP OCDStates - we are 
        // just re-using an existing OCDState for its own parent OCIState
        // during path re-establishment.
        ocdState = (LSIPv4_OCDState *) getStateInfo (tmpOCIState->rightOCDHandle);
        assert (ocdState != NULL);
        currLayerHandle = ocdState->myHandle;

    } else {
        // This is the first time the OCIState is invoking a setup.
        // We need to create a new LSIP OCDState for this OCIState. 

    
        // LSIP OCDs are not shared.
        // This is because there is no OCI at the other end.
        // Having a unique OCD State for each OCI state makes 
        // multiplexing/demultiplexing easy.

        currLayerHandle = getNewOCDHandle();
        ocdState = new LSIPv4_OCDState (currLayerHandle, oci, this);
        putStateInfo (currLayerHandle, ocdState);

        strcpy(ocdState->endPointName, endPointName.c_str());

        ocdState->ociHandle = ociHandle;
    
        PRINT_DEBUG  (
                DEBUG_LEVEL_VERBOSE,
                "\nLSIPv4 setting up connection to %s\n\n", 
                endPointName.c_str()
            );

	    // We are relying on Linux's NAT. We demultiplex packets on the basis of
	    // the fake destionation IPs (or source IPs when the reply packets come in)
        ocdState->fakeIP = proxy->getNewFakeIP();
        addVN_FakeIP (ocdState);

    }

    // Mark that this OCIState should be informed when setup completes o frails.

    ocdState->addParentHandle ( ociHandle);
    //now make the DNS request for the actual requested URL.
    // If the DNS resolution is successful, it calls setupSuccess()
    makeDNSRequest (ocdState);



    return currLayerHandle;
}

void LSIPv4_OCD::send ( OCDHandle handle, char *pkt, int pktLen) {

    LSIPv4_OCDState* remoteIPState = (LSIPv4_OCDState *) getStateInfo (handle);

    assert (remoteIPState != NULL);

    OCIPacketBlock *ociPacketBlock = (OCIPacketBlock *) pkt;
	// OCIPacketBlock passed to the LSIP will be in network byte format
	// as it is intended for an OCI:receive() at some remote node.
	// Since LSIP is a special OC-D module, there is no remote node.
	// It fakes an OCI:receive.
	// So we convert the packet into host byte order for processing within the LSIP. 
	ntoh (ociPacketBlock);

    // We have to send only the actual IP pkt to the remote end point.  
    // This is because there is no proxy at the other end.
    // The LSIP OCD module needs to understand OCI packets.
    switch (ociPacketBlock->blockType) {

        case OCI_PATH_SETUP_BLOCK:

        {
            //Since there is no OCI at the other end,
            //the LSIP answers OCI_PATH_SETUP_BLOCK.
            OCIPacketBlock tmpBlock;
            tmpBlock.version = PROXY_VERSION;
            tmpBlock.blockType = OCI_PATH_SETUP_COMPLETED_BLOCK;
            tmpBlock.nextBlock = 0;
            memcpy ( tmpBlock.dstHandleArray, ociPacketBlock->dstHandleArray, sizeof (OCIHandle));
   
            int tmpBlockLen = sizeof (OCIPacketBlock);
			
			// OCI::receive expects the OCI packet to be in network byte order
			hton (&tmpBlock);

            // Fake the receipt of the OCI_PATH_SETUP_COMPLETED_BLOCK.
			//printf ("LSIP called oci receive 2.\n");
            oci->receive ( (char *) &tmpBlock, this, remoteIPState->get_myHandle());

        }
        break;


        case OCI_KEEPALIVE_BLOCK:

		{
			OCIHandle tmpDstHandle = *((OCIHandle *) ociPacketBlock->dstHandleArray);
            PRINT_DEBUG (
                  DEBUG_LEVEL_VERBOSE,
                  "Handling KeepAlive Block for handle " UINT64_FORMAT_CODE "\n",
                  tmpDstHandle
                );

            // Send the KeepAlive Reply
            OCIPacketBlock tmpBlock;
            tmpBlock.version = PROXY_VERSION;
            tmpBlock.blockType = OCI_KEEPALIVE_REPLY_BLOCK;
            tmpBlock.nextBlock = 0;
            memcpy (tmpBlock.dstHandleArray, ociPacketBlock->dstHandleArray, sizeof (OCIHandle));

			// OCI::receive expects the OCI packet to be in network byte order
			hton (&tmpBlock);

            // Fake the receipt of the OCI_KEEPALIVE_REPLY_BLOCK
			//printf ("LSIP called oci receive 3.\n");
            oci->receive ( (char *) &tmpBlock, this, remoteIPState->get_myHandle());
		}
        break;
        
        case OCI_DATA_BLOCK:
        {
            // Extract data out of the OCI data packet and send to the 
            // legacy server.
            OCIDataBlock *proxyDataBlock = (OCIDataBlock *) pkt;
            char *data = proxyDataBlock->data;
    
            struct in_addr myFakeAddr; //Will be in network byte order
            if (inet_aton(remoteIPState->fakeIP->c_str(), &myFakeAddr) == 0) {
                PRINT_DEBUG (
                        DEBUG_LEVEL_MINIMAL, 
                        "Error while converting fakeIP str to inet_addr: Invalid fake IP :%s\n", 
                        remoteIPState->fakeIP->c_str()
                    );          
            }

			
			// Check if it is an ICMP packet destination unreachable 
           
            if (proxy->mySoftwareNATFlag) {
                
			    // For the 4-tuple associated with this packet, check if we already have
			    // state in our software NAT.
			    uint16_t newSrcPort; //host byte order
			    IPPacket tmpIPPacket (proxyDataBlock->dataLen, data);

                tmpIPPacket.printPacket();

			    // This indicates that the data part of the tmpIPPacket should not be freed 
			    // while the IPPacket object is destroyed.  By default, freeMemFlag is set to
			    // true.  Here we are using the IPPacket class just as a temporary wrapper - 
			    // we don't want to free the data associated with the packet just yet.
			    tmpIPPacket.freeMemFlag = false;
		

                if (!myNAT->sendPacket (&tmpIPPacket, remoteIPState)) {
                    PRINT_DEBUG0 (
                                DEBUG_LEVEL_WARNING,
                                "MyNAT unable to send IP packet.\n"
                            );
                    return;
                }
                
            } else {
                
			    struct in_addr finalDestAddr; //this is also in network byte order
                finalDestAddr.s_addr = htonl(remoteIPState->targetIP_h);

                PRINT_DEBUG (DEBUG_LEVEL_SUPER, ">>>>> LSIP sending the following packet from %s to %s.\n",
                        remoteIPState->fakeIP->c_str(),
                        inet_ntoa (finalDestAddr)
                    );
           
                IPPacket tmpIPPacket (proxyDataBlock->dataLen, data);
			    tmpIPPacket.freeMemFlag = false;
                tmpIPPacket.printPacket();
                printf ("\n");
		

                Tunnel::change_addr_fix_checksum (data, myFakeAddr, finalDestAddr);

			    tunnel->writeToTunnel (data, proxyDataBlock->dataLen);
            }
        }
        break;

        default:
            PRINT_DEBUG (
                    DEBUG_LEVEL_MINIMAL,
                    "LSIPv4 received unknown OCI packet of type %d.\n",
                    ociPacketBlock->blockType
                );
        break;
    }
}

/**
  * Make a DNS request for the url requested when the specified VN behaves
  * as a remote server proxy.
  */
void LSIPv4_OCD::makeDNSRequest (LSIPv4_OCDState *ocdState) {

    // if the endPointName is already in the form of an IP address, we 
    // do not need to make a DNS request.
    struct in_addr tmpIP;
    if (inet_aton (ocdState->get_endPointName().c_str(), &tmpIP) != 0) {
        //valid IP address, no need to resolve
	    ocdState->targetIP_h = htonl(tmpIP.s_addr);
        ocdState->resolvedFlag = true;
        ocdState->setupCompletedFlag = true;
        
        //We should not call notifySetupSuccess() at this point because the functions that get called by notifySetupSuccess
        //require some values which are set by statements which occur only after the function call that led to this point.
        //ocdState->notifySetupSuccess();
        
    } else {
        // DNS name, must resolve
        
        if (proxy->makeDNSRequest (ocdState, ocdState->endPointName)) {
    
            //DNS successfully sent to the name server of the machine on which the proxy is running

        } else {
            PRINT_INFO (
                INFO_LEVEL_WARNING,
                "Unable to send DNS requested to the name server of this machine.\n"
            );
        }
    }
}

/** 
 * This function is called when the tunnel captures a packet
 * sent from the remote endpoint.
 * This function is used only if we are using the Linux software NAT.
 */
void LSIPv4_OCD::tunnelCapturedPacket(IPPacket *captIPPkt) {
    this->tunnelCapturedPacket (captIPPkt, NULL);
}


/**
  * This function is called when a packet arrives over a raw socket
  * for the specificied ocdState.
  */
void LSIPv4_OCD::tunnelCapturedPacket(IPPacket *captIPPkt, LSIPv4_OCDState* ocdState) {

        
    if (captIPPkt->isDNSRequest()) {

        PRINT_INFO (
                INFO_LEVEL_WARNING,
                "Captured DNS packet while behaving as server proxy!\n"
                "This should not happen!\n"
            );
        
        } else {    //captured packet is a data packet

            
			char tmpDestIP[MAX_IP_ADDR_LEN];
            string tmpDestIPStr(captIPPkt->getDestIPAddress(tmpDestIP));
			
            char tmpSrcIP[MAX_IP_ADDR_LEN];
            string tmpSrcIPStr(captIPPkt->getSrcIPAddress(tmpSrcIP));


            // Retrieve the VirtualNode associated with the fake IP address
            PRINT_DEBUG(DEBUG_LEVEL_SUPER, "<<<< LSIPv4 received the following packet from %s to %s.\n", 
                        tmpSrcIPStr.c_str(),
                        tmpDestIPStr.c_str()
                    );
            captIPPkt->printPacket();
            printf ("\n");
            
            if ( !proxy->mySoftwareNATFlag) {
                ocdState = lookupVN_FakeIP(tmpDestIPStr);
            }
            
            if (ocdState != NULL) {
                
                //Pack the data packet from the remote end point so that the 
                // generic proxy can understand it and pass it to the legacy application.
                // We have to do this because the remote end point is a plain IP end point;
                // no proxy running there.
                char finalBuf [MAX_PACKET_BUF_LEN];

                OCIDataBlock *proxyDataBlk = (OCIDataBlock *) finalBuf;
                proxyDataBlk->header.version = PROXY_VERSION;
                proxyDataBlk->header.blockType = OCI_DATA_BLOCK;
                proxyDataBlk->header.nextBlock = 0;
                memcpy (proxyDataBlk->header.dstHandleArray, &(ocdState->ociHandle), sizeof (OCIHandle));
 
                memcpy (proxyDataBlk->data, captIPPkt->pktContents, captIPPkt->pktSize);
                proxyDataBlk->dataLen = captIPPkt->pktSize;
    
				// OCI::receive expects the OCI packet to be in network byte order
				hton ((OCIPacketBlock *) finalBuf);

                oci->receive(finalBuf, this, ocdState->myHandle);
                
                delete (captIPPkt);
                
            } else {

                delete (captIPPkt);
                
                if ( ! proxy->mySoftwareNATFlag) {
                    PRINT_DEBUG(DEBUG_LEVEL_MINIMAL, "Received data packet on unknown destination IP address %s:%d.\n", tmpDestIPStr.c_str(), ntohs(captIPPkt->getDestPortNum()));
                } else {
                    PRINT_DEBUG(DEBUG_LEVEL_MINIMAL, "Received data packet for NULL OCD dest IP address %s:%d.\n", tmpDestIPStr.c_str(), ntohs(captIPPkt->getDestPortNum()));
                }
            }
        }

}

/**
  * This function is used to add a virtual node to a hashtable keyed by its 
  * fake IP address. This is used to lookup the VN when a data packet is 
  * received on a fake IP address.
  *
  * @param ocdState The virtual node to be added into the hash table.  The fakeIP 
  * field should have already been set.
  */
void LSIPv4_OCD::addVN_FakeIP (LSIPv4_OCDState *ocdState) {
    map_VN_FakeIP.insert(pair<IPAddress, LSIPv4_OCDState *>(*ocdState->fakeIP, ocdState));
}

/**
  * This function is used to lookup a virtual node based on its fake IP address.
  * @param fakeIP  Retrieves the virtual node associated with this fakeIP.
  */
LSIPv4_OCDState* LSIPv4_OCD::lookupVN_FakeIP( IPAddress fakeIP) {
    map<IPAddress, LSIPv4_OCDState*>::iterator itr = map_VN_FakeIP.find(fakeIP);
    
    if (itr != map_VN_FakeIP.end()) {
        return (LSIPv4_OCDState*)itr->second;
        
    } else {
            
        return NULL;
    }
}

/**
  * Remove the specified VN from the hashtable of VNs keyed by the VN's FakeIP
  */
void LSIPv4_OCD::removeVN_FakeIP(LSIPv4_OCDState *ocdState) {
    map_VN_FakeIP.erase( *ocdState->fakeIP);
}

void LSIPv4_OCD::recoverFromCrash() {

}

char* LSIPv4_OCD::toString (char *tmpStr) {


    sprintf( tmpStr,
            "Remote IPv4 OCD\n"
            "\ttype\t\t:\t%s\n"
            "\tid\t\t:\t%s\n"
            "\tdescription\t:\t%s\n",
            type.c_str(),
            id.c_str(),
            description.c_str()
           );
    return tmpStr;
}

