#include "MyNAT.h"


#include <unistd.h>
#include <errno.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "Proxy.h"
#include "MyNATPort.h"
#include "LSIPv4_OCD.h"
#include "LSIPv4_OCDState.h"
#include "Tunnel.h"


/**
 * Constructs and initialize the NAT.
 * As part of the initialization, we try to grab (bind) natPortCount ports
 * from the given range.
 * @param natPortRangeStart in host byte order
 * @param natPortRangeEnd in host byte order
 * @param lsip Pointer to the LSIP being served by this NAT
 */
MyNAT::MyNAT (Proxy* proxy, LSIPv4_OCD* lsip, uint16_t natPortRangeStart, uint16_t natPortRangeEnd, int natPortCount) {
    this->lsip = lsip;
    this->proxy = proxy;
	this->natPortRangeStart = natPortRangeStart;
	this->natPortRangeEnd = natPortRangeEnd;
	this->natPortCount = natPortCount;

/*
	int numPortsGrabbed = grabPorts();
	PRINT_INFO (
				INFO_LEVEL_MINIMAL,
				"MyNAT: Grabbed %d ports between %d and %d.\n",
				numPortsGrabbed, natPortRangeStart, natPortRangeEnd
			);
*/
}

/**
 * Reserve up to natPortCount ports in the range allocated to us.
 * We grab/reserve a port by binding to it.
 * @return The number of ports successfully grabbed.
 */
int MyNAT::grabPorts() {
	int grabCount = 0;
	int currPort = natPortRangeStart;
	while (grabCount < natPortCount && currPort <= natPortRangeEnd) {

		// try to grab the current port
		bool udpSuccessFlag = false;
		bool tcpSuccessFlag = false;

		// TCP
		int sockfd;
		
		if ((sockfd = socket(PF_INET, SOCK_STREAM, 0)) != -1) {
			struct sockaddr_in my_addr;
			my_addr.sin_family = AF_INET;         // host byte order
			my_addr.sin_port = htons(currPort);     // short, network byte order
			my_addr.sin_addr.s_addr = INADDR_ANY;
			memset(&(my_addr.sin_zero), '\0', 8); // zero the rest of the struct
								    
			// don't forget your error checking for bind():
			if (bind(sockfd, (struct sockaddr *)&my_addr, sizeof(struct sockaddr)) != -1) {
				tcpSuccessFlag = true;
			}
		}

		// UDP
		if ((sockfd = socket(PF_INET, SOCK_DGRAM, 0)) != -1) {
			struct sockaddr_in my_addr;
			my_addr.sin_family = AF_INET;         // host byte order
			my_addr.sin_port = htons(currPort);     // short, network byte order
			my_addr.sin_addr.s_addr = INADDR_ANY;
			memset(&(my_addr.sin_zero), '\0', 8); // zero the rest of the struct
								    
			// don't forget your error checking for bind():
			if (bind(sockfd, (struct sockaddr *)&my_addr, sizeof(struct sockaddr)) != -1) {
				udpSuccessFlag = true;
			}
		}

		
		if (udpSuccessFlag && tcpSuccessFlag) {
			// port was successfully captured in both TCP and UDP
			grabCount++;
		}

		currPort++;
	}

	return grabCount;
}

/**
  * @param relPort in host byte order.
  */
void MyNAT::releasePort (uint16_t relPort) {

    uint16_t tmpOffset = relPort - natPortRangeStart;
    freePortSet.reset(tmpOffset);
}

/** Get the next free port from our allocated range into freePort. The port is in host byte order.
 *  @return true if a free port was successfully found, else returns false.
 */
bool MyNAT::getFreePort(uint16_t *freePort) {

	for (unsigned long count = 0; count < natPortCount; count++) {
    
        if (nextFreePort == natPortCount) {
            //cycle back to 0 once end of list is reached.
            nextFreePort = 0;
        }
        
        if ( ! freePortSet.test(nextFreePort)) {
            //we have found a free port
            freePortSet.set(nextFreePort);
            
            nextFreePort++;

            *freePort =(natPortRangeStart + nextFreePort);
            return true;
        } 
        
        nextFreePort ++;
    }
    //no free Port found

    return false;

}


/** Checks if this packet is associated with some previously established state.
 * If yes, just returns the previously established NATPort state. Else creates new state.
 * This function is called when a packet is to be sent out.
 * @param ocdState the OCD state associated with this flow, needs to be looked up
 * when a reply packet comes in.
 * @return pointer to the NATState (preexisting or newly created), NULL if an error occurred.
 */
MyNATPort* MyNAT::lookupNATState (IPPacket *pkt, LSIPv4_OCDState* ocdState) {

	uint32_t pktFlowHash = pkt->getFlowHash();
	map<uint32_t, MyNATPort*>::iterator itr = map_NATPort_flowHash.find (pktFlowHash);
	if (itr != map_NATPort_flowHash.end()) {
		// Found existing entry.
        MyNATPort* tmpNATPort = itr->second;
		return tmpNATPort;
	} else {
        
        // host byte order
        uint16_t tmpSrcPort;
        
		// no existing state found.  
		// So establish new one.
		if (!getFreePort(&tmpSrcPort)) {
			PRINT_INFO (INFO_LEVEL_WARNING,
					"No more free NAT ports available.\n"
					);
			return NULL;
		} else {
			
            // A free port number has been put into srcPort.

            // Initialize the raw socket to listen on the newly allocated srcPort.
            int tmpSockFD = initPort (tmpSrcPort, pkt->getProtocolNum());
            if (tmpSockFD == -1) {

                PRINT_DEBUG2 (
                                DEBUG_LEVEL_WARNING,
                                "Unable to init NAT port %d, protocol %d.\n",
                                tmpSrcPort, pkt->getProtocolNum()
                            );
                return NULL;
            }

            //Create a NATPort state object
            MyNATPort* tmpNATPort = new MyNATPort (
                                                    tmpSrcPort,   // new src port
                                                    ntohs(pkt->getSrcPortNum()),  // original src port
                                                    pkt->getProtocolNum(), 
													pkt->getDestIPAddress(),
                                                    tmpSockFD, 
                                                    pktFlowHash,
                                                    ocdState,
                                                    this
                                                );
            
			// Associate this NATPort state object with the flowHash so that it can be retrieved
            // for future packets with the same flow hash.
			map_NATPort_flowHash.insert(pair<uint32_t, MyNATPort*>(pktFlowHash, tmpNATPort));
           
            // Associate this NATPort state object with the sockFD so that it can be retrieved when a
            // reply packet is received on the raw socket associated with this NAT state.
            map_NATPort_sockFD.insert (pair<int, MyNATPort*>(tmpSockFD, tmpNATPort));

			return tmpNATPort;
		}
	}
}

/**
  * Send the following specified packet to the final destination
  * given in the ocdState after doing the necessary natting.
  */
bool MyNAT::sendPacket (IPPacket* ipPkt, LSIPv4_OCDState* ocdState) {

    // Get the NATstate for this packet.  If this is a new flow,
	// create new NAT state.

	struct in_addr finalDestAddr; //this is also in network byte order
    finalDestAddr.s_addr = htonl(ocdState->targetIP_h);
    
    PRINT_DEBUG2 (
        				DEBUG_LEVEL_VERBOSE,
					    "Sending packet of size %d to final destination %s.\n",
                        ipPkt->pktSize,
                        inet_ntoa (finalDestAddr)
				);
    ipPkt->printPacket();

	// Change the source address to the local IP address.
	// Change the destination address to the final destination address.
	Tunnel::change_addr_fix_checksum (ipPkt->pktContents, proxy->my_addr_nf, finalDestAddr);

    MyNATPort* tmpNATPort = NULL;
    
	if ((tmpNATPort = lookupNATState (ipPkt, ocdState)) == NULL) {
        // Unable to lookup NAT state or allocate new state.
		PRINT_INFO (
        				INFO_LEVEL_WARNING,
					    "Unable to allocate new NAT state for final destination %s.\n",
                        inet_ntoa (finalDestAddr)
				);
		return false;
	}

    // Sending a packet on this NAT port.  So it is active.
    tmpNATPort->activeFlag = true;

    PRINT_DEBUG2 (DEBUG_LEVEL_VERBOSE,
				"@@@@@@@@@@@@2Source port for this packet = %d, old source port = %d; in pkt src=%d dst=%d.\n",
				tmpNATPort->portNum, tmpNATPort->originalPortNum,
                htons(ipPkt->getSrcPortNum()),
                htons(ipPkt->getDestPortNum())
	        );


	// We now rewrite the source Port number to the new source port number
	// allotted to this connection earlier
	Tunnel::change_src_port_fix_checksum (ipPkt->pktContents, htons(tmpNATPort->portNum));



    PRINT_DEBUG2 (DEBUG_LEVEL_VERBOSE,
				"$$$$$$$$$$$$Source port for this packet = %d, old source port = %d; in pkt src=%d dst=%d.\n",
				tmpNATPort->portNum, tmpNATPort->originalPortNum,
                htons(ipPkt->getSrcPortNum()),
                htons(ipPkt->getDestPortNum())
	        );

	struct sockaddr_in mysocket;
	memset(&mysocket,'\0',sizeof(mysocket));
   
	mysocket.sin_family = AF_INET;
	mysocket.sin_port = ipPkt->getDestPortNum();
	mysocket.sin_addr = finalDestAddr;
		   
	if( sendto(
                tmpNATPort->sockFD,
                ipPkt->pktContents,
                ipPkt->pktSize, 
                0x0/*flags*/,
                (struct sockaddr *)&mysocket,
				sizeof(mysocket)
            ) != ipPkt->pktSize)  {
	
        perror("raw socket sendto");
		PRINT_INFO (
					INFO_LEVEL_WARNING,
					"LSIPv4_OCD: Unable to send out packet to %s:%d (pktsize = %d).\n",
					inet_ntoa(finalDestAddr), ntohs (mysocket.sin_port), ipPkt->pktSize
				);
        return false;
	}

    return true;

}

/**
  * Closes the socket and frees the port for future use.
  */
void MyNAT::destroyPort (MyNATPort *natState) {
   
    PRINT_DEBUG1 (DEBUG_LEVEL_SUPER, 
                "Destroying NAT state for port %d.\n", natState->portNum
            );
    proxy->delFDCallback (natState->sockFD);
    close (natState->sockFD);

	map_NATPort_flowHash.erase (natState->pktFlowHash);
    map_NATPort_sockFD.erase(natState->sockFD);
}

/** 
  * Open a raw socket and bind it to the specified protocol and port.
  * @param port in host byte order
  * @param protocol IPPROTO_UDP or IPPROTO_TCP
  * @return the socketFD of the opened raw socket if successful. Else returns -1.
  */
int MyNAT::initPort (uint16_t port, int protocol) {
     	
    int sockFD;

	if ((sockFD = socket(PF_INET, SOCK_RAW, protocol)) < 0) {
        PRINT_DEBUG3 (
                        DEBUG_LEVEL_MINIMAL,
                        "Unable to open raw socket for port %d, protocol %d: %s.\n",
                        port, protocol,
                        strerror (errno)
                    );
        return -1;
	}

    struct sockaddr_in sin;

    memset(& sin, 0, sizeof(sin));
    sin.sin_port = htons (port);
              

    if (bind(sockFD, (struct sockaddr *) &sin, sizeof(sin)) == -1) {
       
        PRINT_DEBUG3 (
                        DEBUG_LEVEL_MINIMAL,
                        "Unable to bind raw socket for port %d, protocol %d: %s.\n",
                        port, protocol,
                        strerror (errno)
                    );
        close (sockFD);
        return -1;
    }
    
    int yes = 1;
    if (setsockopt(sockFD, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1) {
        PRINT_DEBUG3(
                        DEBUG_LEVEL_MINIMAL,
                        "Unable to set SO_REUSEADDR socket option for socket on port %d, protocol %d: %s.\n",
                        port, protocol,
                        strerror (errno)
                    );
    }

    yes = 1;
    if (setsockopt(sockFD, 0, IP_HDRINCL, &yes, sizeof(yes)) == -1) {
        PRINT_DEBUG3(
                        DEBUG_LEVEL_MINIMAL,
                        "Unable to set IP_HDRINCL socket option for socket on port %d, protocol %d: %s.\n",
                        port, protocol,
                        strerror (errno)
                    );
        return -1;
    }

    TObjectFunctor<MyNAT> *fdfunc = new TObjectFunctor<MyNAT>(this, &MyNAT::receiveRawPacket);
    proxy->addFDCallback (sockFD, fdfunc, true);

    return sockFD;
}

/**
  * This callback is invoked whenever there is a raw packet to be read on the specified sockFD.
  */
void MyNAT::receiveRawPacket (int sockFD) {
   
    // Retrieve the NATPort state associated with this socket FD.
    map<int, MyNATPort*>::iterator itr = map_NATPort_sockFD.find (sockFD);
    if (itr == map_NATPort_sockFD.end()) {
        // No state associated with this socket.
        // This should not happen.
        PRINT_DEBUG1 (
                        DEBUG_LEVEL_WARNING,
                        "No NAT state associated with socket FD %d.\n"
                        "This should NOT happen!!! Closing the socket.\n",
                        sockFD
                    );
        // Close the socket.
        close (sockFD);
        return;
        
    } else {
        // The state associated with this socket was found.
        MyNATPort* tmpNATPort = itr->second;
      
        // Received a packet on this NAT port.  So it is active
        tmpNATPort->activeFlag = true;

        // Read the packet from the socket.
        char pktBuf[MAX_PACKET_SIZE];
        int pktLen = read (sockFD, pktBuf, MAX_PACKET_SIZE);
        
        PRINT_DEBUG1 (
                        DEBUG_LEVEL_VERY_VERBOSE,
                        "Received packet of length %d.\n",
                        pktLen
                    );

        // Rewrite the destination port of the packet to the original source port.
        // We do this because this reply packet would have been addressed to the 
        // original source port if we had not rewritten it.
        Tunnel::change_dst_port_fix_checksum (pktBuf, htons (tmpNATPort->originalPortNum));

        // Create an IP packet and pass it to the LSIP
        IPPacket* tmpIPPkt = new IPPacket (pktLen, pktBuf);
        tmpIPPkt->freeMemFlag = false;

		// Check if the packet is from the remoteAddr of this MyNATPort.
		if (  tmpIPPkt->getSrcIPAddress() == tmpNATPort->remoteAddr) {
			// Great! Valid packet from the remote host known by this MyNATPort state.
			// Actually, this MyNATPort state was created when the first packet was 
			// sent to tmpNATPort->remoteAddr.
        	lsip->tunnelCapturedPacket (tmpIPPkt, tmpNATPort->ocdState);
		} else {
			struct in_addr t1;
			struct in_addr t2;
			t1.s_addr = tmpIPPkt->getSrcIPAddress();
			t2.s_addr = tmpNATPort->remoteAddr;
			char ts1 [100];
			char ts2 [100];
			strcpy (ts1, inet_ntoa (t1));
			strcpy (ts2, inet_ntoa (t2));
			PRINT_DEBUG2 (DEBUG_LEVEL_MINIMAL,
					"Dropping packet because source address (%s) of packet does not match the remote address (%s) known to NAT state.\n",
					ts1, ts2
				);
			tmpIPPkt->printPacket();
			PRINT_DEBUG (DEBUG_LEVEL_MINIMAL, "\n\n");

			// We don't need this IP packet anymore. Free it.
			delete (tmpIPPkt);
        }
    }
    
}

