/**
 *  \file BroMiddleBox.cpp
 *  \brief This file defines the shim layer which interacts with the proxy and interfaces
 *  to the actual bro intrusion detection process running on the same machine.
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>

#include "BroMiddleBox.h"
#include "BroState.h"
#include "../common.h"
#include "../debug.h"
#include "../oci/OCIState.h"
#include "../Tunnel.h"
#include "../FakeAddressManager.h"


/**
  * Constructs the bro middle box.
  * Set up the interface to the actual bro process.
  */
BroMiddleBox::BroMiddleBox (char *configFileName) {

    //TODO Read this in from the config file.
    string fakeAddrRangeStart = "10.0.0.2";
    string fakeAddrRangeMask = "255.255.255.0";
    
    fakeAddrManager = new FakeAddressManager(fakeAddrRangeStart, fakeAddrRangeMask);

    openFIFO();
}

/** 
 *  Dummy destructor, just to make the destructor virtual and make the warning in OS X go away.
 */
BroMiddleBox::~BroMiddleBox () {

}

/**
  * Try to open the FIFO to the Bro process.
  * If it fails, it just prints an error message,
  * outFD remains at -1 and returns false.
  */
bool BroMiddleBox::openFIFO() {
    
    PRINT_DEBUG (DEBUG_LEVEL_MINIMAL, "Trying to open pipe to Bro process\n");
    outFD = open(BRO_FIFO_NAME,O_WRONLY|O_NONBLOCK);
   
    if (outFD == -1) {

        PRINT_INFO (
                        INFO_LEVEL_FATAL_ERROR,
                        "Unable to open FIFO %s to the Bro process.\n",
                        BRO_FIFO_NAME
                    );
        // We will again try to open the fifo when the bro receives a packet.
        return false;
    } else {

        PRINT_DEBUG (
                    DEBUG_LEVEL_MINIMAL,
                    "Fifo %s to Bro successfully opened.\n",
                    BRO_FIFO_NAME
                );
    
        //TODO Ask KJK what this does
        write(outFD, hexheader,24);
        //TODO
        //flush (outFD);
        return true;
    }

}

/**
  * This function is called when a packet reaches the middlebox.
  */
bool BroMiddleBox::receive (OCIState *ociState, const char* pkt, uint16_t *pktLen) {

    PRINT_DEBUG ( DEBUG_LEVEL_VERBOSE,
            "Bro received packet of len %u from OCIState " UINT64_FORMAT_CODE ".\n",
            *pktLen, ociState->myHandle
        );
    
    if (outFD == -1) {
        //FIFO to the Bro process not open.  Try to open it again.
        if ( ! openFIFO()) {
            // We failed again.  This means that we couldn't send the packet to Bro.
            // This is against security. So this packet should be dropped.

            return false;
        }
    }   
    
    // Find the source and destination IP addresses to be used for this OCIState.
    BroState *broState = lookupBroState (ociState->myHandle);
    
    if (broState == NULL) {
        // If it is the first time we are seeing this OCI handle.
        IPAddress *tmpSrcIP = fakeAddrManager->getNewFakeIP();
        IPAddress *tmpDstIP = fakeAddrManager->getNewFakeIP();
       
        struct in_addr tmp_src_ip;
        struct in_addr tmp_dst_ip;
        inet_aton (tmpSrcIP->c_str(), &tmp_src_ip);
        inet_aton (tmpDstIP->c_str(), &tmp_dst_ip);
        
        broState = new BroState(tmp_src_ip, tmp_dst_ip);
        addBroState (ociState->myHandle, broState);

        delete tmpSrcIP;
        delete tmpDstIP;
    }

    // Copy the packet into a temporary buffer.
    // This is because bro does not modify packets.
    char dupPkt[MAX_PACKET_BUF_LEN];
    memcpy (dupPkt, pkt, MAX_PACKET_BUF_LEN);
    
    //Rewrite the source and destination address of the duplicated packet.
    Tunnel::change_addr_fix_checksum (dupPkt, broState->srcAddr, broState->dstAddr);

    int dataLen = *pktLen;
    // Now send the packet to bro over the pipe.
    struct timeval t;
    gettimeofday(&t,NULL);
    write(outFD, (char*)(&(t.tv_sec)), sizeof (t.tv_sec));
    write(outFD, (char*)(&(t.tv_usec)),sizeof (t.tv_usec));
    dataLen += 16; // incl link layer header
    write(outFD, (char*)(&(dataLen)),sizeof (dataLen));
    write(outFD, (char*)(&(dataLen)),sizeof (dataLen));
    write(outFD, (char*)linkheader,16);
    write(outFD, dupPkt,dataLen-16);

    //TODO
    //flush(outFD);
                        
    return true;
}

/**
  * Looks up the bro state associated with the specified oci handle.
  * Returns NULL if it does not exist.
  */
BroState* BroMiddleBox::lookupBroState (OCIHandle ociHandle) {
    map<OCIHandle, BroState*>::iterator itr = broStateMap.find (ociHandle);
    
    if (itr != broStateMap.end()) {
        return (BroState*)itr->second;
        
    } else {
            
        return NULL;
    }
}

/**
  * Adds the specified BroState keyed by the specified oci handle. 
  */
void BroMiddleBox::addBroState (OCIHandle ociHandle, BroState *state) {
    
    broStateMap.insert(pair<OCIHandle, BroState*> (ociHandle, state));
}
