/**
 *  \file OCD.cpp
 *  \brief Thie file contains the definition of the Overlay Convergence Dependent Layer class.
 *  The OCD class serves as the base class for specific overlays . Utility functions and features 
 *  required by all overlays are implemented in this class.
*/
#include <assert.h>

#include "OCDStateInfo.h"
#include "OCD.h"
#include "../common.h"
#include "../oci/OCI.h"
#include "../Proxy.h"
#include "../TimerEvent.h"
#include "../JournalManager.h"
#include "../MyName.h"
#include "../utils/utilFunctions.h"
#include "../DNSReplyCallback.h"


/**
  * Constructor.
  * @param type         The type of the overlay network. For eg: I3
  * @param id           The id of the overlay network. For eg: I3.PlanetLab
  * @param desc         A brief description of the overlay.
  * @param fileName     The name of the overlay specific configuration
  * file
  * @param urlSuffixes  The string of space separated suffixes of URLs which are to be
  * handled by this overlay.
  * @param tunnel The tunnelling (packet capturing device) used for
  * sending packets to the legacy application.
  */
OCD::OCD(
                string type, string id, string
                desc, string fileName, 
                string urlSuffixes, bool handleLegacyURLs
            ) {
   
    this->proxy = NULL;
    this->oci = NULL;
    this->id = id;
    this->type = type;
    this->description = desc;
    this->configFileName = fileName;
    this->urlSuffixes = urlSuffixes;
    this->handleLegacyURLs = handleLegacyURLs;
   
    this->journalManager = NULL;

#ifdef _WIN32
	setMonitor_Type (type);
#endif
}


/**
  * Do nothing constructor.
  */
OCD::OCD() {

    this->proxy = NULL;
    this->oci = NULL;
}

/** 
 * Dummy desctructor.
 * Just to make the base class destructor virtual.
 */
OCD::~OCD() {

}


/**
  * Adds the specified timer event to the proxy's event queue.
  */
void OCD::addTimerEvent(TimerEvent *t) {
    
    if (proxy != NULL) {
       
        proxy->addTimerEvent(t);
        
    } else {
        
        PRINT_DEBUG0(DEBUG_LEVEL_WARNING, "Proxy not set before trying to add timer event. Event not added!\n");
    }
}

/**
  * Just a wrapper function to the proxy's makeDNSRequest function.
  */
OCALA_LIB_EXPORT_IMPORT bool OCD::makeDNSRequest (DNSReplyCallback* replyCbk, const char *requestedName) {
    
    if (proxy != NULL) {
        return proxy->makeDNSRequest (replyCbk, requestedName);
    } else {
        PRINT_DEBUG0 (DEBUG_LEVEL_WARNING,
                "Unable to make DNS request as proxy equals NULL.\n"
            );
        return false;
    }
}

/**
 * This function returns the list of virtual nodes associated
 * with this overlay network.
 */
list<OCDStateInfo *> OCD::getVirtualNodeList() {
    return ocdList;
}

/**
 * Adds the specified OCDState to the list of OCDStates in this OCD 
 */
void OCD::addOCDState (OCDStateInfo *ocd) {
    ocdList.push_back(ocd);
}


/**
 * Returns a dummy status.  This function should be overloaded
 * to return the status of a specific overlay.
 */
int OCD::getStatus() {
    return 0;
}

#ifdef _WIN32
/**
 * Returns the windows registry key under which status information is stored.
 * This function is windows specific.
 * This function simply returns the id.  May be overloaded for specific overlays.
 */
const char* OCD::getWinRegistryKey() {
    return id.c_str();
}	

#endif

/**
  * Adds a FDCallback to the list of callbacks in a thread safe manner.
  * @param fd The fd to be called back upon.
  * @param functor The functor encapsulating the member function of some object to be invoked
  * when there is some activity on the fd.
  * @param functorFreeFlag If this flag is set to true, the specified functor will
  * be deleted when the FDCallback is deleted. Default value is false
  */
void OCD::addFDCallback(int fd, TFunctor* functor, bool functorFreeFlag) {
    
    proxy->addFDCallback (fd, functor, functorFreeFlag);
}

/**
  * This function is used to remove the callback associated with the 
  * specified fd.
  */
void OCD::delFDCallback(int fd) {
    proxy->delFDCallback (fd);
}

/**
  * Adds the specified state information indexed by handle.
  */
void OCD::putStateInfo (OCDHandle handle, OCDStateInfo *state) {
    
    handleMap.insert ( pair<OCDHandle, OCDStateInfo *> (handle, state));
}

/**
  * Retrieves the state information associated with the specified handle.
  */
OCDStateInfo* OCD::getStateInfo (OCDHandle handle) {

    map<OCDHandle, OCDStateInfo *>::iterator itr = handleMap.find (handle);
    
    if (itr != handleMap.end()) {
        return (OCDStateInfo*)itr->second;
        
    } else {
            
        return NULL;
    }

}

int OCD::getStatus (OCDHandle handle) {

    OCDStateInfo * tmpState = getStateInfo (handle);
    assert (tmpState != NULL);    
    return tmpState->status;
}

void OCD::setStatus (OCDHandle handle, int status) {

    OCDStateInfo * tmpState = getStateInfo (handle);
    assert (tmpState != NULL);
    
    tmpState->status = status;
}

OCDHandle OCD::createDuplicateHandle (OCDHandle originalHandle) {

    OCDStateInfo* stateInfo = getStateInfo (originalHandle);
    OCDHandle newOCDHandle = getNewOCDHandle();
    putStateInfo (newOCDHandle, stateInfo);
    return newOCDHandle;
}

/**
  * Returns a new unique layer handle
  */
OCDHandle OCD::getNewOCDHandle () {

    OCDHandle retHandle;

    bool okFlag = false;
    do {
        // repeat till we find a random new OCD handle that is not currently
        // in local use.
        #if defined (_WIN32)
	        retHandle =  (OCDHandle) (1.0 * rand() / RAND_MAX * OCD_MAX_HANDLE);
        #else
            retHandle =  (OCDHandle) (1.0 * random() / RAND_MAX * OCD_MAX_HANDLE);
        #endif

        map<OCDHandle, OCDStateInfo*>::iterator chkItr = handleMap.find (retHandle);
        if (chkItr == handleMap.end()) {
            // retHandle is not currently in use.
            // So we are done.
            okFlag = true;
        } else {
            PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
                    "getNewOCDHandle(): Generated OCD handle already in use %lu."
                    "Retrying generation of random OCD handle.\n",
                    retHandle
                );
            okFlag = false;
        }

        if (retHandle == NULL_HANDLE || retHandle == DROP_OCD_HANDLE) {
            // We are not interested in the NULL handle.
            okFlag = false;
        }
    } while (!okFlag);

    return retHandle;
}


/** 
  * Looks up the state associated with the specified end point name.
  * NULL is returned if no state is found.
  */
OCDStateInfo* OCD::lookupEndPoint (const string endPointName) {

    map<string, OCDStateInfo *>::iterator itr = endPointMap.find (endPointName);
    
    if (itr != endPointMap.end()) {
        return (OCDStateInfo*)itr->second;
        
    } else {
            
        return NULL;
    }
}

/**
  * Adds the specified state information indexed by end point name.
  */
void OCD::putEndPoint (const string endPointName, OCDStateInfo *state) {
    
    endPointMap.insert ( pair<string, OCDStateInfo *> (endPointName, state));
}

/**
  * Delete the end point mapping associated with the specified OCD state.
  */
void OCD::delEndPoint (OCDStateInfo *state) {
    endPointMap.erase (state->endPointName);
}

/**
  * Adds the specified state information indexed by end point id.
  * NOTE: that the endPointName field of the OCDState is overloaded to represent
  * an id or name.
  */
void OCD::putEndPointId (const string endPointId, OCDStateInfo *state) {
    
    endPointIdMap.insert ( pair<string, OCDStateInfo *> (endPointId, state));
}

/** 
  * Looks up the state associated with the specified end point id.
  * NULL is returned if no state is found.
  */
OCDStateInfo* OCD::lookupEndPointId (const string endPointId) {

    map<string, OCDStateInfo *>::iterator itr = endPointIdMap.find (endPointId);
    
    if (itr != endPointIdMap.end()) {
        return (OCDStateInfo*)itr->second;
        
    } else {
            
        return NULL;
    }
}


/** 
  * Delete the end point id mapping associated with the specified OCD state.
  * NOTE: that the endPointName field of the OCDState is overloaded to represent
  * an id or name.
  */
void OCD::delEndPointId (OCDStateInfo *state) {
    endPointIdMap.erase (state->endPointName);
}

/**
  * Deletes the state info mapping associated with the specified state.
  */
void OCD::delStateInfo (OCDStateInfo *state) {
    handleMap.erase (state->myHandle);
    ocdList.remove (state);
}

/**
  * This function is called by an OCIState (with handle ociHandle) when it stops using this OCDState.
  * ociHandle is removed from the ocdHandle's OCI reference list.
  * If the reference list becomes empty and no OCI is waiting for the setup 
  * completion of this OCD state, it is deleted.
  */
void OCD::freeOCDState (OCDHandle ocdHandle, OCIHandle ociHandle) {
   
    
    OCDStateInfo *state = getStateInfo (ocdHandle);
    if (state != NULL) {

        // The ocistate with the specified handle is not waiting for 
        // setup completion of this ocd state, if at all.
        state->delWaitingParentHandle (ociHandle);
        
        //Decrease the number of OCIStates using this OCDState.
        state->delOCIRefList (ociHandle);

        if (state->getOCIRefCount() <= 0 && state->getWaitingParentCount() <= 0) {

            PRINT_DEBUG2 (
                            DEBUG_LEVEL_MINIMAL,
                            "OCDState to %s with handle %lu being deleted as no OCI is using it.\n",
                            state->endPointName,
                            state->myHandle
                        );
            delete state;
        
        } 
    }
}

/**
  * Adds the specified ociHandle to the OCDHandle's OCI reference list. 
  * This means that ociHandle uses ocdHandle.
  */
void OCD::addOCIRefList (OCDHandle ocdHandle, OCIHandle ociHandle) {

    OCDStateInfo *state = getStateInfo (ocdHandle);
    assert (state != NULL);

    state->addOCIRefList (ociHandle);
}

/**
  * Get the OCD-specific next hop if any.
  */
string OCD::getNextHopId (OCDHandle handle) {

    //Dummy, must be overridden if necessary
    return "";
}

string OCD::getMiddleBoxName (OCDHandle handle) {

    // Dummy, must be overridden to do anything useful.
    return "";
}

/** 
  * Puts the the list of journal records associated with this OCD into
  * the supplied list.
  */
void OCD::getJournalRecordList ( list<JournalRecord*> * recList) {

    assert (recList != NULL);
        
    list<OCDStateInfo *>::iterator itr = ocdList.begin();
    while (itr != ocdList.end()) {
        recList->push_back ( (JournalRecord *) (*itr));
        itr++;
    }
}

void OCD::setActive (bool stateFlag) {

    this->activeFlag = stateFlag;

#ifdef _WIN32
	updateProxyMonitorStatus ();
#endif
	
}

bool OCD::isActive () {
    return this->activeFlag;
}

void OCD::recoverRecord (char *tmpLine) {
    //dummy must be overridden 
}

/**
  * Adds the specified name to the list of names associated with this OCD.
  */
void OCD::addMyName (MyName *myName) {

    myNameList.push_back (myName);
}

/**
  * Initializes the inbound names associated with this OCD.
  */
void OCD::initMyNames () {
    //Dummy function.  Must be overloaded by specific OCD if necessary.
}



#ifdef _WIN32

#include "RegistryParams.h"
/**
 * Display either the specified error message or
 * if no error message is specified (i.e. szError is NULL),
 * display an error message corresponding to the last error 
 * that occurred.
 */
void ShowErrorMessage(LPCSTR szError) {
    
	LPVOID          lpMsgBuf = NULL;

	if (szError) {
		// Error Message was specified.
		MessageBox(NULL, szError, "Error", MB_OK | MB_ICONEXCLAMATION);
	
	}else {
		// Display the message corresponding to the error
		// that occurred last.
		FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
		      FORMAT_MESSAGE_FROM_SYSTEM |
		      FORMAT_MESSAGE_IGNORE_INSERTS,
		      NULL, GetLastError(), 0, (LPSTR) & lpMsgBuf, 0,
		      NULL);
		MessageBox(NULL, (LPCSTR) lpMsgBuf, "Error",
					   MB_OK | MB_ICONEXCLAMATION);
		LocalFree(lpMsgBuf);
    }
}

/**
 * Sets the value of the specified key
 * in the registry entry corresponding to this OCD.
 */
BOOL OCD::SetRegistryValue (string keyName, string keyVal) {
	
	// What should be returned
	BOOL retCode = TRUE;

	LONG res;
	DWORD stringValSize = MAX_REGISTRY_STR_VALUE_LEN;

	// Open the Software Key
	HKEY keySoftware;
	res = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
							 "Software",
							 0, /* Reserved must be 0 */
							 KEY_ALL_ACCESS,
							 &keySoftware
						);

	if (res != ERROR_SUCCESS) {
		PRINT_INFO0 (INFO_LEVEL_WARNING,
						"Unable to open HKEY_LOCAL_MACHINE\\Software\n"
						);
		return FALSE;
	}

	// Open the OCALA key.
	HKEY keyOCALA;	
	res = RegOpenKeyEx(		keySoftware,
							 REGISTRY_APP_NAME,
							 0, /* Reserved must be 0 */
							 KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS |
							 KEY_CREATE_SUB_KEY | KEY_SET_VALUE,
							 &keyOCALA
						);

	
	if (res != ERROR_SUCCESS) {
		// OCALA Key not found. So create it.
		res = RegCreateKeyEx(
								keySoftware,
								REGISTRY_APP_NAME,
								0,
								NULL,
								REG_OPTION_NON_VOLATILE,
								KEY_ALL_ACCESS,
								NULL,
								&keyOCALA,
								NULL
							);
		if (res != ERROR_SUCCESS) {
			PRINT_INFO1 ( INFO_LEVEL_WARNING,
							"Unable to create HKEY_LOCAL_MACHINE\\Software\\%s.\n", 
							REGISTRY_APP_NAME);
			RegCloseKey (keySoftware);
			return FALSE;
		}

		
        if (!MakeRegKeyAccessible (keyOCALA, REGISTRY_APP_NAME)) {
			
			PRINT_DEBUG1 ( DEBUG_LEVEL_WARNING,
							"Unable to make the key %s accessible to all.\n",
							REGISTRY_APP_NAME
						);
			RegCloseKey (keyOCALA);
			RegCloseKey (keySoftware);

			return FALSE;
		}


	}


	// All the individual OCDs are under the key OCD.
	// Now open that key.
	HKEY keyOCDs;	
	res = RegOpenKeyEx( keyOCALA,
							"OCDs",
							 0, /* Reserved must be 0 */
							 KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS,
							 &keyOCDs
						);
	if (res != ERROR_SUCCESS) {
		// OCDs subkey not found under OCALA, so create it
		res = RegCreateKeyEx(
								keyOCALA,
								"OCDs",
								0,
								NULL,
								REG_OPTION_NON_VOLATILE,
								KEY_ALL_ACCESS,
								NULL,
								&keyOCDs,
								NULL
							);
		if (res != ERROR_SUCCESS) {
			PRINT_INFO1 (INFO_LEVEL_WARNING,
						"Unable to create HKEY_LOCAL_MACHINE\\Software\\%s\\OCDs\n",
						REGISTRY_APP_NAME
					);
			RegCloseKey (keyOCALA);
			RegCloseKey (keySoftware);
			return FALSE;
		}

		if (!MakeRegKeyAccessible (keyOCDs, "OCDs")) {
			
			PRINT_DEBUG1 ( DEBUG_LEVEL_WARNING,
							"Unable to make the key %s accessible to all.\n",
							"OCDs"
						);
			RegCloseKey (keyOCDs);
			RegCloseKey (keyOCALA);
			RegCloseKey (keySoftware);

			return FALSE;
		}

	}

	// Now open the subkey for the current OCD (ocdId)
	// under the OCDs subkey
	HKEY keyCurrOCD;
	res = RegOpenKeyEx( keyOCDs,
							id.c_str(),
							 0, /* Reserved must be 0 */
							 KEY_ALL_ACCESS,
							 &keyCurrOCD
						);

	if (res != ERROR_SUCCESS) {
		// key with the name id of this OCD not found under OCDs
		res = RegCreateKeyEx(
								keyOCDs,
								id.c_str(),
								0,
								NULL,
								REG_OPTION_NON_VOLATILE,
								KEY_ALL_ACCESS,
								NULL,
								&keyCurrOCD,
								NULL
							);
		if (res != ERROR_SUCCESS) {
			PRINT_INFO2 (INFO_LEVEL_WARNING,
						"Unable to create HKEY_LOCAL_MACHINE\\Software\\%s\\OCDs\\%s\n",
						REGISTRY_APP_NAME, id.c_str()
					);
			RegCloseKey (keyOCDs);
			RegCloseKey (keyOCALA);
			RegCloseKey (keySoftware);
			return FALSE;
		}

		if (!MakeRegKeyAccessible (keyCurrOCD, id.c_str())) {
			
			PRINT_DEBUG1 ( DEBUG_LEVEL_WARNING,
							"Unable to make the key %s accessible to all.\n",
							id.c_str()
						);
			RegCloseKey (keyCurrOCD);
			RegCloseKey (keyOCDs);
			RegCloseKey (keyOCALA);
			RegCloseKey (keySoftware);

			return FALSE;
		}


	}

	// Now set the value of the specified key-value pair.
	res = RegSetValueEx(
							keyCurrOCD,
							keyName.c_str(),
							0,
							REG_SZ,
							(BYTE*) keyVal.c_str(),
							keyVal.length() + 1
						);
	if (res != ERROR_SUCCESS) {
		PRINT_INFO4 (INFO_LEVEL_WARNING,
						"Unable to open HKEY_LOCAL_MACHINE\\Software\\%s\\OCDs\\%s[%s=%s]\n",
						REGISTRY_APP_NAME, id.c_str(),
						keyName.c_str(), keyVal.c_str()
					);
		ShowErrorMessage (NULL);
		retCode = FALSE;
	}

	RegCloseKey (keyCurrOCD);
	RegCloseKey (keyOCDs);
	RegCloseKey (keyOCALA);
	RegCloseKey (keySoftware);
	
	return retCode;
}





/**
 * Sets the value of the specified key
 * in the registry entry corresponding to this OCD.
 */
BOOL OCD::SetRegistryValue (string keyName, DWORD keyVal) {
	
	// What should be returned
	BOOL retCode = TRUE;

	LONG res;
	DWORD valType;
	char stringVal [MAX_REGISTRY_STR_VALUE_LEN];
	DWORD stringValSize = MAX_REGISTRY_STR_VALUE_LEN;

	// Open the Software Key
	HKEY keySoftware;
	res = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
							 "Software",
							 0, /* Reserved must be 0 */
							 KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS |
							 KEY_CREATE_SUB_KEY | KEY_SET_VALUE,
							 &keySoftware
						);

	if (res != ERROR_SUCCESS) {
		PRINT_INFO0 (INFO_LEVEL_WARNING,
						"Unable to open HKEY_LOCAL_MACHINE\\Software\n"
						);
		return FALSE;
	}

	// Open the OCALA key.
	HKEY keyOCALA;	
	res = RegOpenKeyEx(		keySoftware,
							 REGISTRY_APP_NAME,
							 0, /* Reserved must be 0 */
							 KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS |
							 KEY_CREATE_SUB_KEY | KEY_SET_VALUE,
							 &keyOCALA
						);

	if (res != ERROR_SUCCESS) {
		// OCALA Key not found. So create it.
		res = RegCreateKeyEx(
								keySoftware,
								REGISTRY_APP_NAME,
								0,
								NULL,
								REG_OPTION_NON_VOLATILE,
								KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS |
								 KEY_CREATE_SUB_KEY | KEY_SET_VALUE,
								NULL,
								&keyOCALA,
								NULL
							);
		if (res != ERROR_SUCCESS) {
			PRINT_INFO1 ( INFO_LEVEL_WARNING,
							"Unable to open HKEY_LOCAL_MACHINE\\Software\\%s.\n", 
							REGISTRY_APP_NAME);
			RegCloseKey (keySoftware);
			return FALSE;
		}
	}

	// All the individual OCDs are under the key OCD.
	// Now open that key.
	HKEY keyOCDs;	
	res = RegOpenKeyEx( keyOCALA,
							"OCDs",
							 0, /* Reserved must be 0 */
							 KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS,
							 &keyOCDs
						);
	if (res != ERROR_SUCCESS) {
		// OCDs subkey not found under OCALA, so create it
		res = RegCreateKeyEx(
								keyOCALA,
								"OCDs",
								0,
								NULL,
								REG_OPTION_NON_VOLATILE,
								KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS |
								 KEY_CREATE_SUB_KEY | KEY_SET_VALUE,
								NULL,
								&keyOCDs,
								NULL
							);
		if (res != ERROR_SUCCESS) {
			PRINT_INFO1 (INFO_LEVEL_WARNING,
						"Unable to open HKEY_LOCAL_MACHINE\\Software\\%s\\OCDs\n",
						REGISTRY_APP_NAME
					);
			RegCloseKey (keyOCALA);
			RegCloseKey (keySoftware);
			return FALSE;
		}
	}

	// Now open the subkey for the current OCD (ocdId)
	// under the OCDs subkey
	HKEY keyCurrOCD;
	res = RegOpenKeyEx( keyOCDs,
							id.c_str(),
							 0, /* Reserved must be 0 */
							 KEY_ALL_ACCESS,
							 &keyCurrOCD
						);

	if (res != ERROR_SUCCESS) {
		// key with the name id of this OCD not found under OCDs
		res = RegCreateKeyEx(
								keyOCDs,
								id.c_str(),
								0,
								NULL,
								REG_OPTION_NON_VOLATILE,
								KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS |
								 KEY_CREATE_SUB_KEY | KEY_SET_VALUE,
								NULL,
								&keyCurrOCD,
								NULL
							);
		if (res != ERROR_SUCCESS) {
			PRINT_INFO2 (INFO_LEVEL_WARNING,
						"Unable to open HKEY_LOCAL_MACHINE\\Software\\%s\\OCDs\\%s\n",
						REGISTRY_APP_NAME, id.c_str()
					);
			RegCloseKey (keyOCDs);
			RegCloseKey (keyOCALA);
			RegCloseKey (keySoftware);
			return FALSE;
		}
	}

	// Now set the value of the specified key-value pair.
	res = RegSetValueEx(
							keyCurrOCD,
							keyName.c_str(),
							0,
							REG_DWORD,
							(BYTE*) &keyVal,
							sizeof (DWORD)
						);
	if (res != ERROR_SUCCESS) {

		char tmpErrBuf[MAX_PRINT_BUF];
		PRINT_INFO5 (INFO_LEVEL_WARNING,
			"Unable to open HKEY_LOCAL_MACHINE\\Software\\%s\\OCDs\\%s[%s=%d] : %s\n",
						REGISTRY_APP_NAME, id.c_str(),
						keyName.c_str(), keyVal,
						GetErrorMsgStr (tmpErrBuf)
					);
		retCode = FALSE;
	
	}

	RegCloseKey (keyCurrOCD);
	RegCloseKey (keyOCDs);
	RegCloseKey (keyOCALA);
	RegCloseKey (keySoftware);
	
	return retCode;
}



/** This function is used to the set the current icon to be
 * displayed for this OCD in the sytem tray.
 * @param iconFileName The name of the icon file.  Only the name
 * please, the path if formed as follows icons\OCD_type\iconFileName
 */
void OCD::setMonitor_CurrentIcon (string iconFileName) {
	SetRegistryValue (REGISTRY_OCD_CURR_ICON_NAME, iconFileName);
	currentIcon = iconFileName;
}

void OCD::setMonitor_CurrentStatusMsg (string msg) {
	SetRegistryValue (REGISTRY_STATUS_MSG, msg);
	currentStatusMsg = msg;
}

void OCD::setMonitor_CurrentStatusCode (DWORD statusCode) {
	SetRegistryValue (REGISTRY_STATUS_CODE, statusCode);
}

void OCD::setMonitor_CurrentDescription (string description) {
	SetRegistryValue (REGISTRY_OCD_DESCRIPTION, description);
}

void OCD::setMonitor_Type (string type) {
	
	SetRegistryValue (REGISTRY_OCD_TYPE, toLowerCase(type));
}
	
/** Sets the list of icons to be used by this OCD.
 * @param iconList Icon filenames separated by commas.
 */
void OCD::setMonitor_IconList (string iconList) {
	SetRegistryValue (REGISTRY_OCD_ICON_LIST, iconList);
}

void OCD::setMonitor_BalloonMsg (string msg) {
	SetRegistryValue (REGISTRY_OCD_BALLOON_MSG, msg);
}

/**
 * Override this method if this OCD needs to do some special
 * processing when the proxy goes to sleep.
 * One common thing to do is to close all active sockets, for
 * they will be reopened later.  Bind might fail if they are not
 * properly closed now.
 */
void OCD::sleepProxy() {

}


/**
 * This function is called periodically as well as when 
 * some event that could have changed the status of this OCD
 * occurs.
 * In this function, the OCD will update the Windows Registry
 * with its status function.
 * This function must be overriden to do something more useful.
 */
void OCD::updateProxyMonitorStatus () {
	setMonitor_CurrentStatusCode (0); //inactive
	setMonitor_CurrentIcon ( toLowerCase (type) + string ("_nosrv.ico"));
}

#endif



