/**
 *  \file OCD.h
 *  \brief This file delcares the OCD class.  Some parts of the OCD class definition are 
 * located in the file OCD_API.h file which gets included in this file.
 */
#ifndef OCD_H
#define OCD_H 1

#include "../ConfigFile.h"
#include "../VirtualNode.h"
#include "../IPPacket.h"
#include "../TFunctor.h"
#include "../JournalSource.h"

#include "../OCALA_Export.h"

#include "OCDHandle.h"
#include "../oci/OCIHandle.h"

#include <string>
#include <list>
#include <map>
using namespace std;

class ConfigFile;
class Proxy;
class OCI;
class TimerEvent;
class OCDStateInfo;
class JournalManager;
class MyName;
class DNSReplyCallback;


/**
 *  Class definition for OCD objects.
 *  The class is used as the base class for specific overlay networks 
 *  - for eg: the i3 overlay
 */
class OCALA_LIB_EXPORT_IMPORT OCD : public JournalSource {

public:
    /**
      * Adds the specified timer event to the proxy's timer event queue.
      */
    void addTimerEvent(TimerEvent *t);
    bool makeDNSRequest (DNSReplyCallback* , const char *requestedName);

    virtual ~OCD();

    void setActive (bool stateFlag);
    bool isActive ();

private:

// ----------------------- Data members - START ---------------------------
    /** This hash table stores handle->stateinfo mappings */
    map<OCDHandle, OCDStateInfo *> handleMap;
  
    /** This hash table stores endpointname --> stateinfo mappings */
    map<string, OCDStateInfo *> endPointMap;

    /** This hash table stores endpointId --> stateInfo mappings */
    map<string, OCDStateInfo *> endPointIdMap;
      
    /**
      * List of OCDStates associated with this overlay network 
      */
    list<OCDStateInfo *> ocdList;
 

    /** Is this OCD active or not?
      * If activeFlag is false, this OCD will not handle any new connections.
      * Note that connections that were started before this OCD was set to inactive
      * state will continue to go through the OCD.
      */
    bool activeFlag;
  
protected:
    /** Journal to keep state on disk */
    JournalManager *journalManager;

     /**
       * A reference to the OCI inside the proxy.
       */
     OCI* oci;

     /**
       * This is a reference to the Proxy of the the overlay is a part of. 
       * This pointer is to be used by the OCD for general utility functions - 
       * for example, finding the current IP address.
      */
     Proxy* proxy;
 
public:
     /** The journal used to keep track of this overlay's state */
     //VirtualNodeJournal* ocdJournal;
     
    /** The type of this overlay network.  For eg: I3 */
    string type;
    
    /** The id of this overlay network.  For eg: I3.PlanetLab */
    string id;

    /** A brief description of this overlay network. */
    string description;

    /** Space delimited string of suffixes of URLs to be handled by this overlay */
    string urlSuffixes;

    /** This flag is true if this overlay is capable of/wants to handle 
      * legacy URLS like www.abc.com.
      */
    bool handleLegacyURLs;
    
    /** Name of the overlay specific configuration file */
    string configFileName;
 
    /** Map of name-value pairs read in from under the OCD element in
      * in proxy config file.
      */
    map<string, char *> configPropertyMap;

    /** 
      * List of inbound names associated with this OCD.
      */
    list<MyName *> myNameList;
   

// ----------------------- Data members - END ---------------------------

#include "OCD_API.h"

protected:
    OCDHandle getNewOCDHandle();
    

    void putEndPoint (string endPointName, OCDStateInfo* state);
    OCDStateInfo *lookupEndPoint(const string endPointName);
    void putEndPointId (string endPointId, OCDStateInfo *state);
    OCDStateInfo *lookupEndPointId(const string endPointNameId);
    void delEndPointId (OCDStateInfo *state);
    
    void addMyName (MyName *);

    OCDHandle createDuplicateHandle (OCDHandle handle);

    void putStateInfo (OCDHandle handle, OCDStateInfo* state);
    OCDStateInfo *getStateInfo (OCDHandle handle);
    void delStateInfo (OCDStateInfo *state);

    int getStatus (OCDHandle handle);
    void setStatus (OCDHandle handle, int status);

    void delEndPoint (OCDStateInfo *state);

    void freeOCDState ( OCDHandle ocdHandle, OCIHandle ociHandle);
    void addOCIRefList (OCDHandle ocdHandle, OCIHandle ociHandle);

    void getJournalRecordList ( list<JournalRecord*> *recList);

  
    /** Each OCD has its specific properties in the configuration file.
     *  These are read in and stored in a hash table keyed by the property name.
     *  The following functions are used to add/lookup entries from this
     *  hash table.
     */ 
    void addConfigProperty (string name, char *value);
    char* getConfigProperty (string name);
    void readUShortAttrib( char* attribName, unsigned short* us, bool requiredFlag);
    bool readUShortAttrib (char *attribName, unsigned short* us, unsigned short defaultVal);
    void readIntAttrib( char* attribName, int* us, bool requiredFlag);
    bool readIntAttrib (char *attribName, int* us, int defaultVal);
    bool readStringAttrib( char *attribName,char* strBuf, bool requiredFlag) ;

    /**
     * Returns a list of VirtualNodes associated with this overlay.
     */
    list<OCDStateInfo *> getVirtualNodeList();

    void addOCDState(OCDStateInfo *);

    /** 
     * This function is used to add a callback on a file descriptor.
     */
    void addFDCallback(int fd, TFunctor* functor, bool functorFreeFlag);
    
    /**
      * This function is used to remove the callback associated with the specified file descriptor.
      */
    void delFDCallback(int fd);

#ifdef _WIN32
	BOOL SetRegistryValue (string keyName, string keyVal);
	BOOL SetRegistryValue (string keyName, DWORD keyVal);

	string currentIcon;
	string currentStatusMsg;

	// These functions are used to control the ProxyMonitor display */
	void setMonitor_CurrentIcon (string iconFileName);
	void setMonitor_CurrentStatusMsg (string msg);
	void setMonitor_CurrentStatusCode (DWORD statusCode);
	void setMonitor_CurrentDescription (string description);
	void setMonitor_Type (string type);
	void setMonitor_IconList (string iconList);
	void setMonitor_BalloonMsg (string msg);
#endif

    friend class ConfigFile;
    friend class MyNameConfigFile;
    friend class OCDStateInfo;
    friend class OCI;
    friend class OCIState;
    friend class CryptoState;
    friend class Proxy;
};

//DYN_LOAD_OCD
// The following is a typedef is used to refer to the function used for 
// registering dynamically loaded OCD modules with the proxy.
typedef OCD* makeOCDFull_t (string, string, string, string, string, bool);
// Each OCD module inserts its type->makeFullOCD() fn ptr entry into ocdMap when
// it gets loaded.
extern map<string, makeOCDFull_t*, less<string> > ocdMap;


#endif
