/**
 * \file OCDConfigProperties.cpp
 * \brief This file defines the functions used by overlays to read
 * overlay specific configuration parameters specified as 
 * name-value pairs in the proxy-config.xml file.
*/

#include "OCD.h"
#include "../debug.h"
#include "../common.h"
#include "../utils/utilFunctions.h"

/**
  * Adds the specified name-value pair to the config value map
  * associated with this OCD.
  */
void OCD::addConfigProperty (string name, char *value) {

    configPropertyMap.insert (pair<string, char *> (name, strdup(value)));
}

/**
  * Get the property with the specified name from the config property map.
  * @return NULL if not found.
  */
char* OCD::getConfigProperty (string name) {

    map<string, char *>::iterator itr = configPropertyMap.find (name);
    
    if (itr != configPropertyMap.end()) {
        return itr->second;
        
    } else {
            
        return NULL;
    }
}

/**
  * Reads the ushort attribute from the config property map
  * @param attribName the attribute we are interested in.
  * @param us The buffer into which the value is to be placed.
  * @param requiredFlag If true and no value for the specified parameter
  * is found, an error message is displayed and the program is
  * terminated.
  */
void OCD::readUShortAttrib(
	char* attribName, unsigned short* us, bool requiredFlag) {
   
    string tmpStr = attribName;
    
    char *attribVal = getConfigProperty (tmpStr);
    if (attribVal == NULL) {
        if (requiredFlag) {
            PRINT_INFO2 (INFO_LEVEL_FATAL_ERROR,
                    "Could not find required attribute %s in OCD %s.  Exiting.\n",
                    attribName, id.c_str()
                );
            EXIT_ON_ERROR;
        }
        return;
    }
    
    *us = (unsigned short) atoi(attribVal);
}

/**
  * Reads the ushort attribute from the config properties map.
  * @param attribName the attribute we are interested in.
  * @param us The buffer into which the value is to be placed.
  * @param defaultVal The default value to be used if parameter not * found.
  * @return true if the parameter was successfully read. 
  * If the parameter was not found and that default value was used, returns false.
  */
bool OCD::readUShortAttrib
	(char *attribName, unsigned short* us, unsigned short defaultVal) {
		
    string tmpStr = attribName;

    char *attribVal = getConfigProperty (tmpStr);
    
    if (attribVal != NULL) {
        *us = (unsigned short) atoi(attribVal);
        return true;
    } else {
        *us = defaultVal;
        return false;
    }
}

/**
  * Reads the value of an attribute of the element with the specified name from the config
  * property map.
  * @param attribName The name of the attribute we are interested in.
  * @param strBuf The buffer into which the value is to be placed.
  * @param requiredFlag If true and no value for the specified attribute
  * is found, an error message is displayed and the program is
  * terminated.
  * @return true if the parameter was successfully read. 
  * If the parameter was not found, returns false.
   */
bool OCD::readStringAttrib(
	    char *attribName,char* strBuf, bool requiredFlag) {

	strBuf[0] = '\0';

    string tmpStr = attribName;
    char *attribVal = getConfigProperty (tmpStr);
    
    if ( requiredFlag && attribVal == NULL ) {
        
        PRINT_INFO2(
                INFO_LEVEL_FATAL_ERROR, 
                "Parameter %s required but not found in OCD with id %s\n", 
                attribName, id.c_str()
            );
        
        EXIT_ON_ERROR;
    }

    if ( attribVal == NULL )
        return false;
    
    strcpy(strBuf,(char*) attribVal);
    stripWS(strBuf);
	return true;
}


/**
  * Reads the integer attribute from the config property map
  * @param attribName the attribute we are interested in.
  * @param us The buffer into which the value is to be placed.
  * @param requiredFlag If true and no value for the specified parameter
  * is found, an error message is displayed and the program is
  * terminated.
  */
void OCD::readIntAttrib(
    char* attribName, int* us, bool requiredFlag) {
   
    string tmpStr = attribName;
    
    char *attribVal = getConfigProperty (tmpStr);
    if (attribVal == NULL) {
        if (requiredFlag) {
            PRINT_INFO2  (INFO_LEVEL_FATAL_ERROR,
                    "Could not find required attribute %s in OCD %s.  Exiting.\n",
                    attribName, id.c_str()
                );
            EXIT_ON_ERROR;
        }
        return;
    }
    
    *us = atoi(attribVal);
}

/**
  * Reads the integer attribute from the config properties map.
  * @param attribName the attribute we are interested in.
  * @param us The buffer into which the value is to be placed.
  * @param defaultVal The default value to be used if parameter not * found.
  * @return true if the parameter was successfully read. 
  * If the parameter was not found and that default value was used, returns false.
  */
bool OCD::readIntAttrib
	(char *attribName, int* us, int defaultVal) {
		
    string tmpStr = attribName;

    char *attribVal = getConfigProperty (tmpStr);
    
    if (attribVal != NULL) {
        *us = atoi(attribVal);
        return true;
    } else {
        *us = defaultVal;
        return false;
    }
}


