/**
 * \file OCDStateInfo.cpp
 * \brief This file defines the OCDStateInfo class.  OCDStateInfo encapsulates
 * the properties and data structures that are common across all overlays.  
 * Overlay specific state info classes derive from the OCDStateInfo class.
 */

#include "OCDStateInfo.h"
#include "../oci/OCI.h"

#include <string.h>
#include <assert.h>

OCDStateInfo::OCDStateInfo (OCDHandle myHandle, OCI *oci, OCD* ocd) {

    this->setupOptionBlocks = NULL;
    this->setupOptionLen = 0;
    this->myHandle = myHandle;
    //dummy

    this->oci = oci;
    this->ocd = ocd;

    this->endPointName[0] = '\0';

    this->setupCompletedFlag = false;

}

OCDStateInfo::OCDStateInfo (OCDHandle myHandle, OCD *ocd) {

    this->setupOptionBlocks = NULL;
    this->setupOptionLen = 0;
    this->myHandle = myHandle;
    //dummy
    
    this->oci = NULL;
    this->ocd = ocd;
    
    this->endPointName[0] = '\0';
    this->setupCompletedFlag = false;
}

/**
  * Adds the specified handle to the list of parent handles waiting
  * for setup success or failure callbacks from this OCDState.
  */
void OCDStateInfo::addParentHandle (OCIHandle handle) {
  
    
    currWaitingParentHandleList.push_back (handle);
}

/**
  * Deletes the specified handle to the list of parent handles waiting
  * for setup success or failure callbacks from this OCDState.
  */
void OCDStateInfo::delWaitingParentHandle (OCIHandle ociHandle) {

    currWaitingParentHandleList.remove (ociHandle);

}

/** Notify all OCI states which were waiting for the setup completion
  * of this OCD state about the failure.
  * Remove the notified OCI states from the waiting list.
  */
void OCDStateInfo::notifySetupFailure () {

    assert (oci != NULL);

    printf ("notifySetupFailure size of currWaitingParentHandleList = %d\n", currWaitingParentHandleList.size());

    //Remove all duplicate handles.  They might have been added 
    //as a result of multiple unsuccessful setup attempts.
    //currWaitingParentHandleList.unique();

    OCIHandle *tmpHandleList = new OCIHandle [currWaitingParentHandleList.size()];
    
    list<OCIHandle>::iterator itr = currWaitingParentHandleList.begin();
    int index = 0;
    while (itr != currWaitingParentHandleList.end()) {

        PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL, "notifySetupFailure on waiting oicd parent handle %d\n", (*itr));

        tmpHandleList[index] = *itr;

        index++;
        itr++;
    }

    currWaitingParentHandleList.clear();

    for (index = 0; index < currWaitingParentHandleList.size(); index++) {
    
	    oci->setupFailed( tmpHandleList[index]);
    }


    delete[] (tmpHandleList);
}

/** Notify all OCI states which were waiting for the setup completion
  * of this OCD state about the setup success.
  * Remove the notified OCI states from the waiting list.
  */
void OCDStateInfo::notifySetupSuccess () {

    setupCompletedFlag = true;

    assert (oci != NULL);

    //Remove all duplicate handles.  They might have been added 
    //as a result of multiple unsuccessful setup attempts.
    currWaitingParentHandleList.unique();

    list<OCIHandle>::iterator itr = currWaitingParentHandleList.begin();
    while (itr != currWaitingParentHandleList.end()) {

        PRINT_DEBUG2 (
                    DEBUG_LEVEL_MINIMAL,
                    "notifySetupSuccess at ocdstate %p --> ocihandle = %d\n", this, (*itr)
                );
	    oci->setupSucceeded ( (*itr));
        itr++;
    }

    currWaitingParentHandleList.clear();
}

void OCDStateInfo::receivedLegacyDNSReply (uint32_t resolvedIP) {

    //Dummy.  Overload if necessary
}

/**
  * Destructor.
  */
OCDStateInfo::~OCDStateInfo () {

    PRINT_DEBUG2 (
                    DEBUG_LEVEL_MINIMAL,
                    "Destructing OCDState Info %lu %p\n",
                   myHandle, this
                 );

    assert (ocd != NULL);
    ocd->delStateInfo (this);
    ocd->delEndPoint (this);
    ocd->delEndPointId (this);

}

/**
  * Adds the specified OCI handle to the oci reference list.
  */
void OCDStateInfo::addOCIRefList (OCIHandle ociHandle) {

    ociRefList.push_back (ociHandle);
    ociRefList.unique();
}

/**
  * Deletes the specified OCI handle from the oci reference list
  * of this OCD state.
  */
void OCDStateInfo::delOCIRefList (OCIHandle ociHandle) {

    ociRefList.remove (ociHandle);
}

/**
  * Returns the number of OCI states currently referring to this
  * OCD state.
  */
int OCDStateInfo::getOCIRefCount() {
    
    return ociRefList.size();
}

/**
  * Returns the number of parent OCI States waiting for the setup completion
  * of this OCDstate.
  */
int OCDStateInfo::getWaitingParentCount() {

    return currWaitingParentHandleList.size();
}

void OCDStateInfo::getJournalRecord (char *buf) {
        // Dummy, must overload
        strcpy (buf, "ERROR: Function getJournalRecord not correctly overloaded.\n");
        
}

    /** This flag is set to indicate that this trigger needs to be destroyed.
      * This is usually used only for private triggers.  When the VN which inserted
      * a private trigger is destroyed, we cannot immediately free up the private trigger as
      * it may still be on the trigger refresh queue.  So the trigger is freed the next
      * time a trigger refresh event occurs if this flag is set to true. 
      */
 



