/** 
 * \file OCD_API.h
 * \brief This file contains the OCD API declaration.
 * The functions that have to defined by a new overlay are given here.
 * This file must be included within the definition of the OCD class.
 * A separate file exists only to clearly list out the functions
 * that must be implemented by a new overlay.
 */

// NOTE: This file is to be included inside the OCD class definition
// It is desirable that a new OCD module implement each of the following functions.
// The functions that MUST be overloaded are marked with MUST_OVERLOAD.

protected:


    /** 
    * This constructor is usually used when the OCD module is constructored
    * from a base OCD module created after reading the config files.
    * The data fields in the base OCD module are used to initialize the fields
    * of the specific OCD module being constructed (see example).
    */
    OCD();

    /**
     * This constructor is used when all the basic details of the OCD module
     * are already available.
     */
    OCD(string type, string id, string description, 
            string configFileName, 
            string urlSuffixes, bool handleLegacyURLs
         );

 /** 
      * This function is called by the overlay to initialize itself
      * before it is used.
      * For eg: An I3 overlay will insert public triggers here.
      * This function must be overloaded (MUST_OVERLOAD).
      * @param prxy A pointer to the proxy of this overlay is a part of.
      * @param oci The OCI associated with this OCD.
      * @param recoverFromCrash This parameter is set to true by the proxy if it is 
      * is recovering from a crash. The OCD is expected to reconstruct its
      * state from the journal.  This flag is usually simply passed on to
      * the OCD's JournalManager which is instantiated in the init() function .
      */
    virtual void init(Proxy* prxy, OCI* oci, bool crashRecoveryFlag)=0;


   /**
      * This function sets up a tunnel to the specified endPointName and returns a
      * handle to the corresponding OCDState.  Note that existing tunnels may be reused
      * if the OCD wishes.
      * This function must be overloaded (MUST_OVERLOAD).
      * @param ociHandle Handle to the OCI state to be notified through a callback when the 
      * tunnel setup succeeds or fails.
      * @param endPointNameType 
      *     OCI_NAME - indicates that the parameter endPointName is an OCI-level name which needs to be
      *                 resolved to an OCD-specific id.
      *     OCD_ID  - indicates that the parameter endPointName is already an OCD-specified id.
      */
    virtual OCDHandle setup (
                const string endPointName, 
                int endPointNameType,
                char *optionBlocks, 
                int optionLen,
                OCIHandle ociHandle,
                OCI* oci ) = 0;
    /**
      * Sends the specified data via the connection represented by handle.
      * This function must be overloaded (MUST_OVERLOAD).
      */
    virtual void send ( OCDHandle handle, char* data, int dataLen) = 0;


public: 
  /**
      * This function returns a human readable short
      * description of this overlay.
      */
    virtual char* toString(char *tempStr) = 0;

protected:
    /**
     * Initializes the inbound names associated with this OCD.
     * For example, in the case of i3 all the public triggers must be inserted at this point.
     */
    virtual void initMyNames();

    /** This function is called to recover the state of the OCD state encoded in the string
      * read from the journal.  This is called during crash recovery of the proxy.
      * The specified string is encoded in an OCD specific manner.
      */
    virtual void recoverRecord (char *);

    /**
      * If the specified handle correponds to an OCD-specified middlebox, the name of the middlebox is
      * returned.  Else the empty string "" is returned.
      */
    virtual string getMiddleBoxName(OCDHandle handle);
   
    /**
      * This function is used to ask the left side OCD (the OCD on which the tunnel setup request was received)
      * if it wants to specify the next hop id.  If yes, this function returns the OCD-id which constitutes the next id.
      * Else, it returns the empty string "".
      */
    virtual string getNextHopId (OCDHandle handle);

 
public:
#ifdef _WIN32
    /**
     * Returns the windows registry key which contains status information about
     * this overlay.
     */
    virtual const char* getWinRegistryKey();

	/** 
	 * This function is called is when the proxy is about to be temporarily
	 * stopped.  For example: when a laptop goes to sleep.
	 */
	virtual void sleepProxy();

#endif



    /**
     * Returns an integer which represents the status of this overlay.
     * The return value has an overlay specific meaning.
     * For example, the windows proxy monitor service uses the value returned by this function
     * to change the displayed status icon.
     */
    virtual int getStatus();

#ifdef _WIN32
	/**
	* This function is called periodically as well as when 
	* some event that could have changed the status of this OCD
	* occurs.
	* In this function, the OCD will update the Windows Registry
	* with its status function.
	* This function must be overriden to do something more useful.
	*/
	virtual void updateProxyMonitorStatus ();
#endif 

