/**
 *  \file OCI.cpp
 *  \brief This file defines the Overlay Convergence Independent Layer class.
 */

#include "common.h"
#include "OCI.h"
#include "OCIState.h"
#include "OCIKeepAliveState.h"
#include "../ocd/OCDStateInfo.h"
#include "../LegacyAppLayer.h"
#include "../DNSPacket.h"
#include "../Proxy.h"
#include "../Params.h"
#include "../TFunctor.h"

#ifndef _WIN32
#include "../middlebox/MiddleBox.h"
#include "../middlebox/BroMiddleBox.h"
#endif

#include "../urlresolver/HopInfo.h"
#include "../urlresolver/PathInfo.h"

#include "../JournalManager.h"

#include "../security/lib/CryptoState.h"
#include "../security/lib/SecurityManager.h"
#include "../compression/CompressionState.h"

#include <assert.h>

OCI::OCI (Proxy *proxy) {

    this->proxy = proxy;

#ifndef _WIN32
	// We allow the proxy to function as a middlebox only in Linux.
    if (proxy->functionAsMiddleBox) {
        //TODO allow other middleboxes

        if (proxy->middleBoxTypeName == "bro") {
            this->middleBox = new BroMiddleBox(NULL);
        } else {
            this->middleBox = NULL;
        }
    } else {
    
        this->middleBox = NULL;

    }
#endif

    addStateCleanUpTimer();

}

/**
 * Dummy destructor.
 * Just to make it virtual.
 */
OCI::~OCI() {

}

void OCI::initJournalManager (bool crashRecoveryFlag) {
    
    this->journalManager = new JournalManager (this, this, OCI_STATE_JOURNAL_FILENAME, crashRecoveryFlag);
}

/**
  * This function is called by the OCD when a tunnel setup requested by the OCI failed.
  * @param ociHandle The handle of the OCI state which had requested the tunnel setup.
  */
void OCI::setupFailed (OCIHandle ociHandle) {

    OCIState* ociState = (OCIState *) getStateInfo (ociHandle);

    if (ociState == NULL) {
        return;
    }

    if (ociState->myRole == OCI_ROLE_MIDDLEBOX) {
        //Send a path setup failed message back to the initiator.
        ociState->sendPathSetupFailedMessage();
    }
    
    // Don't destroy the state immediately.  Wait for it time out.
    // If we kill it immediately, its associated OCIstate might be left hanging.
    // Note that we can immediately delete the OCDstates associated with this OCIState
    // even if no one else is using this, becuase this function call returns to
    // the OCD state just deleted --> segfault!
    ociState->activeFlag = false;
    destroyOCIState(ociState);
    
}

void OCI::destroyOCIState (OCIState *ociState) {

    PRINT_DEBUG1 ( DEBUG_LEVEL_MINIMAL, 
		"Marking ociState %p for deletion\n",
		ociState
               );

    if (ociState->killMeFlag) {
    // already marked for killing
    	PRINT_DEBUG1 ( DEBUG_LEVEL_MINIMAL, 
		"**** RE **** Marking ociState %p for deletion\n",
		ociState
               );
        return;
    }

    // Remove this state from all maps and lists.
    delStateInfo (ociState);

    if (ociState->initiatorVN != NULL) {
        ociState->initiatorVN->answerDNSRequests (NULL);
    }

    proxy->legacyAppLayer->destroyVN (ociState->initiatorVN);
    ociState->initiatorVN = NULL;
    
    proxy->legacyAppLayer->destroyVN (ociState->receiverVN);
    ociState->receiverVN = NULL;

    
    if (ociState->cryptoState != NULL) delete ociState->cryptoState;
    ociState->cryptoState = NULL;

    if (ociState->compressionState != NULL) delete ociState->compressionState;
    ociState->compressionState = NULL;

    if (ociState->pathInfo != NULL) delete ociState->pathInfo;
    ociState->pathInfo = NULL;

    if (ociState->timer != NULL) {
        ociState->timer->setDisabled();
    }
    ociState->timer = NULL;

    if (ociState->activeCheckTimer != NULL) {
        ociState->activeCheckTimer->setDisabled();
    }
    ociState->activeCheckTimer = NULL;

    if (ociState->keepAliveState != NULL) {
        PRINT_DEBUG2 (DEBUG_LEVEL_VERBOSE, "11111: Deleting ociState (%p) keepAliveState (%p).\n",
                    ociState, ociState->keepAliveState
                );
        delete (ociState->keepAliveState);
    } else {
        PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL, "11111; keep alive state = NULL\n");
    }
    ociState->keepAliveState = NULL;

    // Free up the OCD states associated with this OCI State.
    // NOTE: The OCDStates will be completely deleted only when all OCIStates associated
    // with them have freed them.
    if (ociState->leftOCD != NULL) {

        printf ("left ocd %d of oci " UINT64_FORMAT_CODE " freed\n", ociState->leftOCDHandle, ociState->myHandle);
        ociState->leftOCD->freeOCDState (ociState->leftOCDHandle, ociState->myHandle);

    }
    ociState->leftOCD = NULL;

    if (ociState->rightOCD != NULL) {
        printf ("right ocd %d of oci " UINT64_FORMAT_CODE " freed\n", ociState->rightOCDHandle, ociState->myHandle);
        ociState->rightOCD->freeOCDState (ociState->rightOCDHandle, ociState->myHandle);
    }
    ociState->rightOCD = NULL;

    ociState->killMeFlag = true;
}

/**
  * This function is invoked by an OCD when the requested tunnel has been set up.
  * @param ociHandle The handle of the OCI state which had requested the tunnel to be setup.
  */
void OCI::setupSucceeded (OCIHandle ociHandle) {

    OCIState* ociState = (OCIState *) getStateInfo (ociHandle);
    if (ociState == NULL) {
        PRINT_DEBUG1 (
                DEBUG_LEVEL_MINIMAL,
                "setupSucceeded callback to unknown OCI handle %d.\n",
                ociHandle
            );
        return;

    }
    
    // Add this OCI to the OCI refList of its associated OCDStates.
    if (ociState->leftOCD != NULL) {
        ociState->leftOCD->addOCIRefList (ociState->leftOCDHandle, ociState->myHandle);
    }
    if (ociState->rightOCD != NULL) {
        ociState->rightOCD->addOCIRefList (ociState->rightOCDHandle, ociState->myHandle);
    }
    
    if (ociState != NULL) {
      
        if (ociState->timer != NULL) { 
            //Disable the OCD setup timer.
            ociState->timer->setDisabled();
            ociState->timer = NULL;
        }

        // Now send the OCI-path-setup message using the tunnel which has just been set up.
        ociState->sendPathSetupMessage();

    }

 
#ifdef BENCHMARK
	addCheckpoint (EVENT_DNS_OC_D_TO_OC_I, 0);
#endif
    
}

/**
  * This function is called when a packet captured from a legacy application
  * is to be sent across.
  * @param vn The VirtualNode on which the packet was caught.
  */
void OCI::send (VirtualNode *vn, const char* data, int dataLen) {

    assert (vn->ociState != NULL);

    // The OCI State is definitely active when it is getting data from the legacy application
    vn->ociState->setActive();

    if (vn->ociState->initiatorVN == vn) {
        // Packet was caught at the initiator.  Send to right.
        send (vn->ociState, data, dataLen, OCI_RIGHT);
    } else {
        // Packet was caught at the end point.  Send to left.
        send (vn->ociState, data, dataLen, OCI_LEFT);
    }
}

/**
  * Send the specified data through the OCD associated with ociState.
  * If the direction is OCI_LEFT, the data is is sent via the left OCD.
  * If the direction is OCI_RIGHT, the data is sent via the right OCD.
  */
void OCI::send ( OCIState* ociState, const char *data, int dataLen, int direction) {


    //Assume that all packets sent to the proxy from above are data packets.
    //Pack it in a generic proxy data block;
    char finalBuf [MAX_PACKET_BUF_LEN];
    OCIDataBlock *proxyDataBlk = (OCIDataBlock *) finalBuf;
    proxyDataBlk->header.version = PROXY_VERSION;
    proxyDataBlk->header.blockType = OCI_DATA_BLOCK;
    
	memcpy ( proxyDataBlk->header.dstHandleArray, &(ociState->myHandle), sizeof (OCIHandle));
    proxyDataBlk->header.nextBlock = 0;

    char compressBuf [MAX_PACKET_BUF_LEN];
    unsigned long compressedBufLen = sizeof (compressBuf);
    uint16_t finalBufLen;

    //TODO avoid extra copying somehow.
    
    if (ociState->compressionState != NULL
                    && ociState->compressionState->state == COMPRESSION_STATE_OK) {

        
        //Compress the data
        ociState->compressionState->compress (
                                        data,    
                                        dataLen,
                                        (char *)compressBuf,
                                        &compressedBufLen
                            );
                       
        PRINT_DEBUG2 (DEBUG_LEVEL_MINIMAL, "Compressed data of size %d to size %d\n", dataLen, compressedBufLen);
        
        if (ociState->cryptoState != NULL) {
            // Encrypt the packet
            ociState->cryptoState->encrypt (
                                    (unsigned char*) compressBuf,
                                    compressedBufLen,
                                    (unsigned char *)proxyDataBlk->data,
                                    &finalBufLen,
                                    direction
                            );
  
           
            char tmpBuf[MAX_PRINT_BUF];
            PRINT_DEBUG1 (
                            DEBUG_LEVEL_VERBOSE, 
                            "Encrypting with symkey %s.\n",
                            ociState->cryptoState->getSymKeyString (direction, tmpBuf)
                        );
        } else {
                                
            //TODO remove this completely unnecessary copying
            memcpy (proxyDataBlk->data, compressBuf, compressedBufLen);
            finalBufLen = compressedBufLen ;
        }
    } else {
                            
        //No compression
                            
        if (ociState->cryptoState != NULL) {
            // Encrypt the packet
            ociState->cryptoState->encrypt (
                                        (unsigned char*) data,
                                        dataLen,
                                        (unsigned char*) proxyDataBlk->data,
                                        &finalBufLen,
                                        direction
                                    );

            char tmpBuf[MAX_PRINT_BUF];
            PRINT_DEBUG1 (
                            DEBUG_LEVEL_VERBOSE, 
                            "Encrypting with symkey %s.\n",
                            ociState->cryptoState->getSymKeyString (direction, tmpBuf)
                        );
        } else {
            //TODO remove this completely unnecessary copying
            memcpy (proxyDataBlk->data, data, dataLen);
            finalBufLen = dataLen;
        }
    }
    proxyDataBlk->dataLen = finalBufLen;
    PRINT_DEBUG1 (
                DEBUG_LEVEL_VERY_VERBOSE,
                "\nOCI::send dataLen = %d\n", 
                proxyDataBlk->dataLen
            );


#ifdef BENCHMARK
	addCheckpoint (EVENT_DATA_OC_I_TO_OC_D, 0);
#endif

    // Convert the fields in the packet to network byte order
    hton ((OCIPacketBlock *) proxyDataBlk);

    if (direction == OCI_LEFT) {
        ociState->leftOCD->send (ociState->leftOCDHandle, (char *) proxyDataBlk, finalBufLen + sizeof (OCIDataBlock));
    } else {
        ociState->rightOCD->send (ociState->rightOCDHandle, (char *) proxyDataBlk, finalBufLen + sizeof (OCIDataBlock));
    }

}

/**
  * This callback is invoked when the OCI Layer receives a packet from
  * some the OCD layer.
  * @param pkt The received packet.  The OCI may modify the contents of the pkt buffer.
  * However, it will not store a pointer to it for further use or free the memory.
  * This means that whoever allocated the memory corresponding to pkt, can do whatever
  * they want with it.  
  * @param pkt An OCI packet in network byte order
  * @param ocd The OCD on which this packet was received.
  * @param ocdHandle The OCD handle on which this packet was received.
  */
void OCI::receive (char *pkt, OCD* ocd, OCDHandle ocdHandle) {

    PRINT_DEBUG1 (DEBUG_LEVEL_VERY_VERBOSE, 
                "OCI received packet on handle ocd handle %d\n",
                ocdHandle
            );
    
    if (pkt != NULL) {
        OCIPacketBlock *ociPkt = (OCIPacketBlock *) pkt;
        
        if (ociPkt->version != PROXY_VERSION) {
            // Version mismatch!
            PRINT_DEBUG3 ( 
                            DEBUG_LEVEL_MINIMAL,
                            "Version mismatch for OCI Packet (version %u, expected version %u) "
                            "received from OCDHandle %lu. "
                            "Ignoring this packet\n"
                            ,
                            ociPkt->version,
                            PROXY_VERSION,
                            ocdHandle
                    );
            //Ignore this packet.
            return;
        }

		// Convert all the fields of the packet to host byte order
      	ntoh (ociPkt);
 
		OCIHandle tmpDstHandle = *((OCIHandle *) ociPkt->dstHandleArray);
        OCIState* ociState = (OCIState *) getStateInfo (tmpDstHandle);
        if (ociState == NULL) {
            // There is no OCI state associated with the received packet.
            // Check if it is a OCI Path Setup message.
            
            if (ociPkt->blockType == OCI_PATH_SETUP_BLOCK) {
                // This is a new connection.
               
                ociState = new OCIState (this, tmpDstHandle);
                putStateInfo (ociState);
                
            } else {
                // No OCI state exists for this packet.
                // Moreover it is not a path setup message.
                // So just ignore it.
                PRINT_DEBUG1 (
                                DEBUG_LEVEL_MINIMAL,
                                "OCI received packet on unknown handle %llu.\n",
                                tmpDstHandle
                            );
		dumpOCIPacket (ociPkt, DEBUG_LEVEL_MINIMAL);
                return;
            }
        }
        
        ociState->processBlocks (ociPkt, ocd, ocdHandle);
    }
}


/**
  * This function is called by the legacy app layer when a new DNS request is captured
  * or when the first IP packet of a flow (which does not rely on DNS requests) is captured.
  * Only one of the params url or captIPPkt should be non NULL.
  * @param url When this function is called in response to a new DNS request, this param is
  * set to the url in the DNS request.
  * @param captIPPkt When this function is called in response to the first packet of a new flow,
  * this param is set to point to the captured packet.
  * @param dropConnectionFlag is set to true to indicate that this connection should be dropped.
  * If this flag is set, the return value should be ignored.
  * @return The OCIState of the new connection.  NULL if some error occurred and the connection
  * was not set up.
  */
OCIState* OCI::setUpNewConnection (
                    const URL *url,
                    IPPacket* captIPPkt,
                    char *setupOptionBlocks,
                    int setupOptionLen,
					bool* dropConnectionFlag
                ) {

    PathInfo* pathInfo = NULL;
    
    string tmpURL = "";
    
    if (url != NULL) {

        // New DNS packet request
        PRINT_DEBUG1 ( DEBUG_LEVEL_VERBOSE, 
                "OCI setting up new connection for URL %s\n",
                url->c_str()
            );

        pathInfo = proxy->resolvePath (url);
        tmpURL = *url;
        
    } else if (captIPPkt != NULL) {
        // Due to first packet of new flow
        char tmpSrcBuf[50];
        char tmpDestBuf[50];

        PRINT_DEBUG4 ( DEBUG_LEVEL_VERBOSE, 
                "OCI setting up new connection for flow %s:%d->%s:%d\n",
                captIPPkt->getSrcIPAddress (tmpSrcBuf),
                ntohs(captIPPkt->getSrcPortNum()),
                captIPPkt->getDestIPAddress (tmpDestBuf),
                ntohs(captIPPkt->getDestPortNum())
            );

        // What is the path to be used for this flow?
        // From the info loaded from the config file, figure this out.
        pathInfo = proxy->resolvePath (captIPPkt); 
   
        // The final destination IP address will be the destination URL.
        // When the remote proxy does a gethostby name for an IP address, the same IP address will be returned.
        tmpURL = tmpDestBuf;
        
    } else {
        PRINT_DEBUG0 (
                    DEBUG_LEVEL_WARNING,
                    "Both url and captIPPkt are NULL!\n"
                );
    }

  
    
    OCIState *ociState = NULL;
    if (pathInfo != NULL) {
            
        //set up the OCI state for this connection
        OCIHandle newOCIHandle = getNewOCIHandle();
        ociState = new OCIState (this, newOCIHandle);
        ociState->myRole = OCI_ROLE_INITIATOR;
        ociState->pathInfo = new PathInfo(pathInfo);
        ociState->nextHopNameType = OCI_NAME;

        // Set the name of the last hop in the path to the requested url.
        HopInfo *tmpLastHop = ociState->pathInfo->getHop ( ociState->pathInfo->getHopCount() - 1);
        tmpLastHop->hopEndPointName = tmpURL;
        
        // Set the name of the final oci end point to the same requested URL.
        ociState->endPointName = tmpURL;
        putStateInfo (ociState);

        
        //TODO setup compression state.this->setup ( proxyTmpHandle, pathInfo, setupOptionBlocks, setupOptionLen);

     
        HopInfo* firstHop = ociState->pathInfo->getHop(0);
        assert (firstHop != NULL);
    
        ociState->rightOCD  = firstHop->ocd;
        assert (ociState->rightOCD != NULL);
    
        if (ociState->pathInfo->getHopCount() == 1) {
            // If there is only one hop, the name of the end point is the URL for which
            // the new connection is being setup.
            ociState->nextHopName = ociState->endPointName;    
        
        } else {
            // The end point name is the name of the first hop end point.
            ociState->nextHopName = firstHop->hopEndPointName;
        }

        ociState->nextHopNameType = OCI_NAME;

        //Setup the remainder of the path. In this case, remainder = full path.
		OCDHandle tmpSetupPathHandle = setupPath (ociState);
        if ( tmpSetupPathHandle == NULL_HANDLE || tmpSetupPathHandle == DROP_OCD_HANDLE) {
            // Some error occurred while trying to set up the path.
            destroyOCIState (ociState);
			if (tmpSetupPathHandle == DROP_OCD_HANDLE) {
				*dropConnectionFlag = true;
			} else {
				*dropConnectionFlag = false;
			}
            return NULL;
        }
       
   }

	*dropConnectionFlag = false;
    return ociState;
}

/**
  * This function is used to setup the path associated
  * with the ociState.
  * @return true if the setup path initialization is successful (path not set up yet), else returns false.
  */
OCDHandle OCI::setupPath (OCIState * ociState) {
 
    assert (ociState != NULL);
  
    ociState->numPathSetupAttempts++;
    // Retrieve the first hop information
  
    char ociStateOptionBuf [MAX_PACKET_SIZE];
    int ociStateOptionLen = ociState->packToSendBlocks (ociStateOptionBuf);
              
#ifdef BENCHMARK
	addCheckpoint (EVENT_DNS_OC_I_TO_OC_D, 0);
#endif

    assert (ociState->rightOCD != NULL);

    // Set up the OC-D level tunnel
    OCDHandle ocdHandle = ociState->rightOCD->setup (
                                ociState->nextHopName, 
                                ociState->nextHopNameType, // OCI_NAME or OCD_ID
                                ociStateOptionBuf, 
                                ociStateOptionLen, 
                                ociState->myHandle,
                                this
                            );

    if (ocdHandle == NULL_HANDLE) {
        PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
                         "Right OCD setup for oci State " UINT64_FORMAT_CODE " returned NULL_HANDLE\n",
                         ociState->myHandle
                    );
        //TODO free this ociState memory later.
        return NULL_HANDLE;
    }

	if (ocdHandle == DROP_OCD_HANDLE) {
		return DROP_OCD_HANDLE;
	}

    ociState->rightOCDHandle = ocdHandle;
       
    // Add ocdSetupTimer
    // Disable any existing timer
    if (ociState->timer != NULL) {
        ociState->timer->setDisabled();
	ociState->timer = NULL;
    }
    
    TObjectFunctor<OCIState> *ocdSetupTimerFunctor = 
                        new TObjectFunctor<OCIState> (ociState, &OCIState::ocdSetupTimerExpired);
    ociState->timer = TimerEvent::create (OCD_SETUP_TIMER, 0, (TFunctor *) ocdSetupTimerFunctor, true);
    addTimerEvent(ociState->timer);
 
    // Processing continues on setupSucceeded() callback from OCD or 
    // when ocdSetupTimer expires.

    return ocdHandle;
}

/**
  * Adds the specified state information indexed by its handle to the handleMap.
  * Also adds the specified state to the list of active OCI states.
  */
void OCI::putStateInfo (OCIState *state) {
    
    handleMap.insert ( pair<OCIHandle, OCIState *> (state->myHandle, state));
    ociStateList.push_back (state);
}

/**
  * Removes the specified OCI state from the handle map and from the list of active
  * OCI states.
  */
void OCI::delStateInfo (OCIState *state) {

    handleMap.erase (state->myHandle);
    
}

/**
  * Retrieves the state information associated with the specified handle.
  */
OCIState* OCI::getStateInfo (OCIHandle handle) {

    map<OCIHandle, OCIState *>::iterator itr = handleMap.find (handle);
    
    if (itr != handleMap.end()) {
        return (OCIState*)itr->second;
        
    } else {
            
        return NULL;
    }

}

int OCI::getStatus (OCIHandle handle) {

    OCIState * tmpState = getStateInfo (handle);
    
    return tmpState->status;
}

/**
  * Returns a new unique layer handle
  */
OCIHandle OCI::getNewOCIHandle () {

    OCIHandle retHandle;

    bool okFlag = false;
    do {
        // repeat till we find a random new OCI handle that is not currently
        // in local use.
		uint32_t tmpLower;
		uint32_t tmpHigher;
		uint32_t* pRetHandleLower = (uint32_t *) &retHandle;
		uint32_t* pRetHandleHigher = ((uint32_t *) &retHandle + 1);

        #if defined (_WIN32) 
			tmpLower = rand(); 
			tmpHigher = rand();
	        //retHandle =  (OCIHandle) ( 1.0 * tt / RAND_MAX * OCI_MAX_HANDLE);
        #else
			tmpLower = rand();
			tmpHigher = rand();
            //retHandle = (OCIHandle) ( 1.0 * random() / RAND_MAX * OCI_MAX_HANDLE);
        #endif
		memcpy (pRetHandleLower, &tmpLower, sizeof (uint32_t));
		memcpy (pRetHandleHigher, &tmpHigher, sizeof (uint32_t));

        map<OCIHandle, OCIState*>::iterator chkItr = handleMap.find (retHandle);
        if (chkItr == handleMap.end()) {
            // retHandle is not currently in use.
            // So we are done.
            okFlag = true;
        } else {
            PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
                    "getNewOCIHandle(): Generated OCI handle already in use " UINT64_FORMAT_CODE "."
 		    "Retrying generation of random OCI handle.\n",
                    retHandle
                );
            okFlag = false;
        }

        if (retHandle == NULL_HANDLE) {
            // We are not interested in the NULL handle.
            okFlag = false;
        }
    } while (!okFlag);

    return retHandle;
}

/**
  * Adds a timer event.  Just a wrapper for the proxy's addTimerEvent.
  */
void OCI::addTimerEvent (TimerEvent *t) {
    
    proxy->addTimerEvent (t);
}


/**
  * Put all active OCIStates into the specified list of JournalRecords.
  */
void OCI::getJournalRecordList (list<JournalRecord*>* recList) {

    list<OCIState*>::iterator itr = ociStateList.begin();

    while (itr != ociStateList.end()) {

        OCIState *currRec = *itr;
        recList->push_back ( currRec);
        itr++;
    }
}

/**
  * Restore the state associated with the specified journal string.
  */
void OCI::recoverRecord (char *tmpJournalStr) {

    OCIState* newState = OCIState::constructState (tmpJournalStr, this);

    if (newState != NULL) {
        
        putStateInfo (newState);
       
        if (newState->initiatorVN != NULL) {
            // create hashtable entry for the new VN keyed by
            // the URL.
            proxy->legacyAppLayer->addVN_URL(newState->initiatorVN);
            // add the VN to the hash table with the vn's fake IP address as the key.
            proxy->legacyAppLayer->addVN_FakeIP(newState->initiatorVN);
        }
 
        if (newState->receiverVN != NULL) {
            // add the VN to the hash table with the vn's fake IP address as the key.
            proxy->legacyAppLayer->addVN_FakeIP(newState->receiverVN);
        }

        if (newState->myRole == OCI_ROLE_INITIATOR ||
                newState->myRole == OCI_ROLE_INITIATOR_ENDPOINT) {

            //TODO Change this first hop setting when OCD-imposed middleboxes.
            HopInfo* firstHop = newState->pathInfo->getHop(0);
            assert (firstHop != NULL);
    
            newState->rightOCD = firstHop->ocd;
            assert (newState->rightOCD != NULL);
    
            if (newState->pathInfo->getHopCount() == 1) {
                // If there is only one hop, the name of the end point is the URL for which
                // the new connection is being setup.
                newState->nextHopName = newState->endPointName;    
        
            } else {
                // The end point name is the name of the first hop end point.
                newState->nextHopName = firstHop->hopEndPointName;
            }

            newState->nextHopNameType = OCI_NAME;
         
            //if (newState->pathInfo->securityInfo.protocol == SECURITY_PROTOCOL_CUSTOM_SSL) {
            //    newState->cryptoState = new CryptoState (proxy, newState, proxy->securityManager);
            //}

            setupPath (newState);    

	    OCDHandle tmpSetupPathHandle = setupPath (newState);
            if ( tmpSetupPathHandle == NULL_HANDLE || tmpSetupPathHandle == DROP_OCD_HANDLE) {
                // Some error occurred while trying to set up the path.
                destroyOCIState (newState);
            }

        }

    } else {
        PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
                        "Error while recovering OCIState form %s.\n",
                        tmpJournalStr
                    );
    }
}

/**
  * Go through all OCI States and clean up all those marked for purging.
  */
void OCI::stateCleanUp() {

    PRINT_DEBUG0 (DEBUG_LEVEL_VERY_VERBOSE,
                "OCI State Cleanup Start!\n\n\n");
   
    OCIState** tmpOCIStateList = new OCIState* [ociStateList.size()];
    list<OCIState *>::iterator itr = ociStateList.begin();
    
    int i = 0;
    while (itr != ociStateList.end()) {
        if ((*itr)->killMeFlag) {
            tmpOCIStateList[i] = (OCIState *) (*itr);
        } else {
            tmpOCIStateList[i] = NULL;
        }
        itr++;
        i++;
    }

    for (i = 0; i < ociStateList.size(); i++) {
    	ociStateList.remove (tmpOCIStateList[i]);
        delete (tmpOCIStateList[i]);
    }

    delete[] (tmpOCIStateList);
    
    PRINT_DEBUG0 (DEBUG_LEVEL_VERY_VERBOSE,
                "OCI State Cleanup Stop!\n\n\n");

    // Schedule the next clean up
    addStateCleanUpTimer();
}

/**
  * Schedule a timer event which will periodically go through the list of all available OCIStates and
  * clean the ones marked killMeFlag.
  */
void OCI::addStateCleanUpTimer () {

    PRINT_DEBUG0 (DEBUG_LEVEL_VERY_VERBOSE, "Adding OCI State Cleanup timer\n");
    
    TObjectFunctor<OCI> * stateCleanupFunctor = 
                        new TObjectFunctor<OCI>(
                                        this, 
                                        &OCI::stateCleanUp
                                    );
    
    TimerEvent *stateCleanUpTimer = TimerEvent::create (
                                                        OCI_STATE_CLEANUP_PERIOD, 
                                                        0, 
                                                        (TFunctor *) stateCleanupFunctor, 
                                                        "OCI State Cleanup",
                                                        true
                                                    );
    
    addTimerEvent(stateCleanUpTimer);


}
