/**
 *  \file OCIKeepAliveState.cpp
 *  \brief This file defines the OCDKeepAliveState.  The OCIKeepAliveState is 
 *  to periodically send OCI keep alive messages and also to respond to
 *  Keep aliver replies not being received on time.
 */

#include "OCIKeepAliveState.h"
#include "OCIState.h"
#include "OCI.h"
#include "../TimerEvent.h"
#include "../TFunctor.h"
#include "../Params.h"

OCIKeepAliveState::OCIKeepAliveState(OCIState *ociState) {

    this->ociState = ociState;

    currSendPeriod = OCI_KEEPALIVE_NORMAL_SEND_INTERVAL;
    currReplyPeriod = OCI_KEEPALIVE_NORMAL_REPLY_INTERVAL;
    
    this->sendTimer = NULL;
    this->replyTimer = NULL;

}

/**
  * Destructor
  */
OCIKeepAliveState::~OCIKeepAliveState () {

    if (sendTimer != NULL) {
        sendTimer->setDisabled();
	    sendTimer = NULL;
        printf ("keepalive state send timer disabled\n");
    } else {

        printf ("keepalive state send timer already null\n");
    }
    
    if (replyTimer != NULL) {
        replyTimer->setDisabled();
	    replyTimer = NULL;
        printf ("keepalive state reply timer disabled\n");
    } else {

        printf ("keepalive state reply timer already null\n");
    }
}

/**
  * Start sending KeepAlives.
  */
void OCIKeepAliveState::start() {

    if (sendTimer != NULL) {
        sendTimer->setDisabled();
        sendTimer = NULL;
    }

    if (replyTimer != NULL) {
        replyTimer->setDisabled();
        replyTimer = NULL;
    }

    // Fake the expiry of the send Timer in order to start the Keep Alive sender.
    sendTimerExpired();

    // schedule the first reply timeout timer
    receivedKeepAliveReply();
}

/**
  * This function is invoked when the KeepAlive reply is not received on time.
  */
void OCIKeepAliveState::replyTimerExpired() {

    PRINT_DEBUG1 (
                DEBUG_LEVEL_SUPER,
                "Keep Alive reply not received on time for OCIState " UINT64_FORMAT_CODE ".\n",
                ociState->myHandle
            );
    if (sendTimer != NULL) {
        sendTimer->setDisabled();
        sendTimer = NULL;
    }

    if (replyTimer != NULL) {
        replyTimer->setDisabled();
        replyTimer = NULL;
    }


    // Notify the OCI that the path is broken
    ociState->notifyPathBroken();
}

/**
  * This function is called when it is time to send the next KeepAlive message.
  */
void OCIKeepAliveState::sendTimerExpired() {

    if (sendTimer != NULL) {
        sendTimer->setDisabled();
        sendTimer = NULL;
    }

    // Send the message.
    ociState->sendKeepAlive();
    
    // Now wait for the KeepAlive reply or the reply timer to expire.
    TObjectFunctor<OCIKeepAliveState> *keepAliveSendFunctor = 
                        new TObjectFunctor<OCIKeepAliveState> (this, &OCIKeepAliveState::sendTimerExpired);
    sendTimer = TimerEvent::create(
                            currSendPeriod / 1000 ,    //seconds
                            (currSendPeriod % 1000) * 1000,    //microseconds
                            (TFunctor *) keepAliveSendFunctor,
                            true
                        );
    
    ociState->oci->addTimerEvent(sendTimer);
   
}

/**
  * This function is called when an OCI Keep Alive Reply is received.
  * This function is called only at the initiator of the KeepAlives.
  * Intermediate OCIState's have no KeepAliveStates.  They just relay
  * or reflect the keep alive messages.
  */
void OCIKeepAliveState::receivedKeepAliveReply() {


    // disable the currenlty active reply timeout timer.
    if (replyTimer != NULL) {
         replyTimer->setDisabled();
         replyTimer = NULL;
    }

    //Schedule the next send.
    TObjectFunctor<OCIKeepAliveState> *keepAliveReplyFunctor = 
                        new TObjectFunctor<OCIKeepAliveState> (this, &OCIKeepAliveState::replyTimerExpired);
    replyTimer = TimerEvent::create(
                            currReplyPeriod / 1000 ,    //seconds
                            (currReplyPeriod % 1000) * 1000,    //microseconds
                            (TFunctor *) keepAliveReplyFunctor,
                            true
                        );
    
    ociState->oci->addTimerEvent(replyTimer);
 
}

