/** 
 * \file OCIPacketBlock.cpp
 * \brief This file contains some functions to allocate OCI-level protocol packets.
 */

#include "OCIPacketBlock.h"
#include "../debug.h"
#include "../utils/utilFunctions.h"

/**
  * Allocates an OCI Path setup message with the specified number of hops.
  */
OCIPathSetupBlock * allocateOCIPathSetupBlock (int numHops) {

	int tmpBlkSize = sizeof (OCIPathSetupBlock) + numHops * sizeof (HopInfo_pkt);
	void* retBlk =  malloc (tmpBlkSize);
	if (retBlk == NULL) {
		PRINT_INFO0 (INFO_LEVEL_FATAL_ERROR, "OCIPathSetupBlock allocateOCIPathSetupBlock() Unable to allocate memory.\n");
		EXIT_ON_ERROR;
	}
	
	memset (retBlk, 0, tmpBlkSize);

	return (OCIPathSetupBlock *) retBlk;

    //Remember to free this memory. FREE
}


// Functions to convert blocks to network byte order and vice versa
void ntoh (OCIPacketBlock* pkt) {
	pkt->nextBlock = ntohs (pkt->nextBlock);

	OCIHandle tmpDstHandleH = ntoh64 ( * ((OCIHandle *) pkt->dstHandleArray));
	memcpy ( pkt->dstHandleArray, &tmpDstHandleH, sizeof (tmpDstHandleH)); 
	switch (pkt->blockType) {
		case OCI_DATA_BLOCK:
			((OCIDataBlock *)pkt)->dataLen = ntohs (((OCIDataBlock *) pkt)->dataLen);
		break;
	}
}
void hton (OCIPacketBlock* pkt) {
	pkt->nextBlock = htons (pkt->nextBlock);
	
	OCIHandle tmpDstHandleN = hton64 (* ((OCIHandle *) pkt->dstHandleArray));
	memcpy (pkt->dstHandleArray, &tmpDstHandleN, sizeof (tmpDstHandleN));
	switch (pkt->blockType) {
		case OCI_DATA_BLOCK:
			((OCIDataBlock *)pkt)->dataLen = htons (((OCIDataBlock *) pkt)->dataLen);
		break;
	}
}


/**
 * This function dumps the contents of the OCI packet to screen/debug file
 */
void dumpOCIPacket (OCIPacketBlock *blk, int debugLevel) {


	// We avoid doing all the packet pretty printing
	// if not at appropriate debug levels.
#ifndef DEBUG_ENABLED
	return;
#endif
	if( !(debugLevel <= CURRENT_DEBUG_LEVEL) ) {
		return;
	}

	int currBlockPos = 0;
	string tmpPrintBufStr;

	// Process each block
    while (1) {
          
        const OCIPacketBlock *currBlock = (OCIPacketBlock *) ( (char *) blk + currBlockPos);
	
		OCIHandle tmpDstHandle = ntoh64 ( * ((OCIHandle *) currBlock->dstHandleArray));        

        //Process the current block
        switch (currBlock->blockType) {
             
            case OCI_PATH_SETUP_BLOCK:
			{
				char tmpPrintBuf [ MAX_PRINT_BUF];
				OCIPathSetupBlock *tmpPathSetupBlk = (OCIPathSetupBlock *) currBlock;
				sprintf (	tmpPrintBuf,
						"OCI_PATH_SETUP_BLOCK: dstHandle=%llu numHops=%d currHop=%d ",
						tmpDstHandle,
						tmpPathSetupBlk->path.numHops,
						tmpPathSetupBlk->path.currHop
					);
				tmpPrintBufStr = string(tmpPrintBuf);
				for (int hopIndex = 0; hopIndex < tmpPathSetupBlk->path.numHops; hopIndex++) {
					sprintf (tmpPrintBuf, " [Hop: %s,%s] ", 
							tmpPathSetupBlk->path.hops[hopIndex].endPointName,
							tmpPathSetupBlk->path.hops[hopIndex].ocdId
						);
					tmpPrintBufStr += tmpPrintBuf;
				}
			}
            break;
           
            case OCI_PATH_SETUP_COMPLETED_BLOCK:
			{
				char tmpPrintBuf [ MAX_PRINT_BUF];
				sprintf (	tmpPrintBuf,
						"OCI_PATH_SETUP_COMPLETED_BLOCK: dstHandle=%llu ",
						tmpDstHandle
					);
				tmpPrintBufStr = string(tmpPrintBuf);

			}
            break;
        
            case OCI_PATH_SETUP_FAILED_BLOCK:
			{
				char tmpPrintBuf [ MAX_PRINT_BUF];
				sprintf (	tmpPrintBuf,
						"OCI_PATH_SETUP_FAILED_BLOCK: dstHandle=%llu ",
						tmpDstHandle
					);
				tmpPrintBufStr = string(tmpPrintBuf);
			}
			break;
            
            case OCI_KEEPALIVE_BLOCK:
			{
				char tmpPrintBuf [ MAX_PRINT_BUF];
				sprintf (	tmpPrintBuf,
						"OCI_KEEPALIVE_BLOCK: dstHandle=%llu ",
						tmpDstHandle
					);
				tmpPrintBufStr = string(tmpPrintBuf);
			}
			break;

            case OCI_KEEPALIVE_REPLY_BLOCK:
			{
				char tmpPrintBuf [ MAX_PRINT_BUF];
				sprintf (	tmpPrintBuf,
						"OCI_KEEPALIVE_REPLY_BLOCK: dstHandle=%llu ",
						tmpDstHandle
					);
				tmpPrintBufStr = string(tmpPrintBuf);
			}
			break;
            
            case OCI_DATA_BLOCK:
			{
				OCIDataBlock* tmpDataBlock = (OCIDataBlock *) currBlock;
				char tmpPrintBuf [ MAX_PRINT_BUF];
				sprintf (	tmpPrintBuf,
						"OCI_DATA_BLOCK: dstHandle=%llu, dataLen=%d ",
						tmpDstHandle, 
						tmpDataBlock->dataLen
					);
				tmpPrintBufStr = string(tmpPrintBuf);
			}
            break;

            
            case PROXY_PACKET_BLOCK_SECURITY:
            {
				char tmpPrintBuf [ MAX_PRINT_BUF];
				sprintf (	tmpPrintBuf,
						"PROXY_PACKET_BLOCK_SECURITY: dstHandle=%llu ",
						tmpDstHandle
					);
				tmpPrintBufStr = string(tmpPrintBuf);
			}
            break;
           
            default:
			{
				char tmpPrintBuf [ MAX_PRINT_BUF];
				sprintf (	tmpPrintBuf,
						"UNKNOWN OCI PACKET TYPE: dstHandle=%llu ",
						tmpDstHandle
					);
				tmpPrintBufStr = string(tmpPrintBuf);
			}
        }

        if (currBlock->nextBlock == 0) { 
            //No more blocks to process.
            break;
        } else {
            currBlockPos += currBlock->nextBlock;
        }
    }

	PRINT_DEBUG0 (
		debugLevel,
		tmpPrintBufStr.c_str()
	);
}
