/**
 * \file OCIPacketBlock.h
 * \brief The format of packets sent and interpreted by the OCI layer.
 * All data fields of the structures in this file should be in host byte order.
 * They should be converted to network byte order just before transmission.
 */
#ifndef OCI_PACKET_BLOCK_H
#define OCI_PACKET_BLOCK_H

#include "../common.h"
#include "OCI.h"
#include "OCIHandle.h"

#ifndef _WIN32
#include <netdb.h>
#endif

/** The proxy version.  This is put as the first byte of every packet */
#define PROXY_VERSION   0x01 

/** General Setup block */
#define PROXY_PACKET_BLOCK_SETUP    1

/** Security related setup block */
#define PROXY_PACKET_BLOCK_SECURITY 2

/** Fake Address Negotiation block */
#define PROXY_PACKET_BLOCK_FAKE_ADDRESS_NEGOTIATION 3

/** Compression negotiation */
#define PROXY_PACKET_BLOCK_COMPRESSION 4

/** Data block meant for application */
#define OCI_DATA_BLOCK 5

/** Used to denote an OCI Path set up block */
#define OCI_PATH_SETUP_BLOCK    10

/** Used to denote an OCI Path set up Completed block */
#define OCI_PATH_SETUP_COMPLETED_BLOCK    11

/** Used to denote an OCI Path set up Completed block */
#define OCI_PATH_SETUP_FAILED_BLOCK    12

/** Used to denote an OCI Keep Alive block */
#define OCI_KEEPALIVE_BLOCK    13

/** Used to denote an OCI Keep Alive reply block */
#define OCI_KEEPALIVE_REPLY_BLOCK    14


/**
  * The maximum length of the identifier (eg: Planetlab.i3) of the OC-D (overlay)
  */
#define MAX_OCD_ID_LEN  20

/**
  * The maximum length of the end point of a particular hop.
  */
#define MAX_OCD_ENDPOINT_NAME   50

/**
  * This structure specifies the format in which hop information
  * will be embedded in the path setup message.
  */
typedef struct {
    char ocdId [MAX_OCD_ID_LEN];
    char endPointName [MAX_OCD_ENDPOINT_NAME];
    
} HopInfo_pkt;

/**
  * This structure specifies the format in which path information
  * will be embedded in the path setup messages.
  */
typedef struct {

    /** The total number of hops in this path */
    uint8_t numHops;

    /** The index to the current hop */
    uint8_t currHop;

    /** Placeholder pointer to the hops */
    HopInfo_pkt hops[0];
    

} PathInfo_pkt;


/**
  * This file defines a basic OCI control or data block.
  * A OCI back consists of a sequence of these kinds of blocks.
  */
typedef struct {

	// The ordering of fields matters in Windows.
	// Otherwise, Visual C++ adds padding which makes portability difficult.
	// Also, for similar reasons, we must make sure that the sizeof OCIPacketHeader
	// is a multible of 4 bytes.

	/** Offset to the start of the next block from the start of this block.
      * A value of 0 implies that there is no nextBlock.
      */
    uint16_t nextBlock;

    /** The proxy version. */
    uint8_t version;
    
    /** The type of the block - SETUP, SECURITY, DATA */
    uint8_t blockType;  

    /** The following array of 8 bytes represent the 64 bit OCIHandle.
	 *  We cannot use the OCIHandle type here because of the alignment and padding
	 *  problems caused by Visual C++.
	 *  The OCIHandle is used to retreive the OCI state for which this block is meant. 
	 */
	uint8_t dstHandleArray[8];
  
} OCIPacketBlock;

typedef struct {

    OCIPacketBlock header;

    /** Length of the actual data */
    uint16_t dataLen;

    /** Just a handle to refer to the actual data.
      * No space is allocated in the struct for it.
      */
    char data[0];
} OCIDataBlock;

/**
  * This structure represents a Path Setup message.
  */
typedef struct {

    OCIPacketBlock header;
    
    /** The path to be setup */
    PathInfo_pkt path;

} OCIPathSetupBlock;

/**
  * Allocates an OCIPathSetupBlock with the specified number of hops.
  */
OCIPathSetupBlock* allocateOCIPathSetupBlock (int numHops);

void ntoh (OCIPacketBlock* pkt);
void hton (OCIPacketBlock* pkt);

void dumpOCIPacket (OCIPacketBlock *blk, int debugLevel);

#endif
