/**
 * \file OCIState.cpp
 * \brief This file defins the Overlay Convergence Independ state information
 * associated with a connection. 
 */

#include "OCIState.h"
#include "OCI.h"
#include "OCIPacketBlock.h"
#include "OCIKeepAliveState.h"
#include "../Proxy.h"
#include "../Params.h"
#include "../ocd/OCD.h"
#include "../ocd/OCDStateInfo.h"
#include "../middlebox/MiddleBox.h"
#include "../LegacyAppLayer.h"
#include "../urlresolver/PathInfo.h"
#include "../urlresolver/HopInfo.h"

#include "../compression/CompressionState.h"
#include "../security/lib/CryptoState.h"
#include "../security/lib/certificate.h"
#include "../security/lib/SecurityPacketBlocks.h"
#include "../security/lib/SecurityManager.h"

#include <assert.h>

OCIState::OCIState (OCI *oci, OCIHandle myHandle) {

    this->oci = oci;
    this->pathInfo = NULL;

    this->myHandle = myHandle;

    this->leftOCD = NULL;
    this->leftOCDHandle = NULL_HANDLE;
    this->rightOCD = NULL;
    this->rightOCDHandle = NULL_HANDLE;

    this->initiatorVN = NULL;
    this->receiverVN = NULL;

    this->cryptoState = NULL;
    this->compressionState = NULL;

    this->myRole = OCI_ROLE_NULL;
    this->status = OCI_STATE_STATUS_NOT_READY;

    this->timer = NULL;
    this->activeCheckTimer = NULL;

    /** The OCIState is active when it is first created. */
    this->activeFlag = true;

    keepAliveState = NULL;
    numPathSetupAttempts = 0;

    setupOptionBlocks = NULL;

    ocdMiddleBoxFlag = false;

    killMeFlag = false;
}

/**
  * This function sends the specified block to be send out 
  * by the overlay.
  * Please note that this block may not be sent out immediately.
  * It will be sent only when the specific overlay decides to.
  * @param direction Which OCD to use - the left one or the right one?
  * @param blk The OCI block to be sent.  The fields of the block are in host byte order.
  * They will be converted to network byte order by this function.
  */
void OCIState::sendBlock (OCIPacketBlock * blk, int blkSize, int direction) {

    // Convert the fields of the block to network byte order.
    hton (blk);

    if (direction == OCI_LEFT) {
       if (leftOCD->getStatus(leftOCDHandle) == OCI_STATE_STATUS_READY) {

	        PRINT_DEBUG0(DEBUG_LEVEL_VERBOSE, "SEnding immediately\n");
            blk->nextBlock = 0;
            leftOCD-> send (leftOCDHandle, (char *)blk, blkSize);

        } else {
	        PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL, "\nQueueing to send later\n");
            toSendBlockList.push_back (blk);
        }

    } else if (direction == OCI_RIGHT) {
       if (rightOCD->getStatus(rightOCDHandle) == OCI_STATE_STATUS_READY) {

	        PRINT_DEBUG0 (DEBUG_LEVEL_VERBOSE, "SEnding immediately\n");
            blk->nextBlock = 0;
            rightOCD-> send (rightOCDHandle, (char *)blk, blkSize);

        } else {
	        PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL, "\nQueueing to send later\n");
            toSendBlockList.push_back (blk);
        }

    }
}

/**
  * This function is called when a data block meant for the legacy application is 
  * received.
  * @param ocd The OCD on which the block was received.
  * @param ocdHandle The OCDHandle on which the block was received.
  */
void OCIState::processDataBlock (OCIDataBlock* blk, OCD* ocd, OCDHandle ocdHandle) {

    // The OCI State is active if it is gets some data block or a PathSetupCompleted block
    // Other blocks like KeepAliveReply should not set the activeFlag to true.
    activeFlag = true;
    
    char *finalBuf;
    uint16_t finalLen;
   
    char decryptedDataBuf [MAX_PACKET_BUF_LEN];
    uint16_t decryptedDataBufLen = 0;

    char decompressedDataBuf [MAX_PACKET_BUF_LEN];
    unsigned long decompressedDataLen = sizeof (decompressedDataBuf);
   
    if (cryptoState != NULL) { //TODO make sure that crypto state is not NULL only when authentication is required.
    
        int direction;
        if (ocd == leftOCD && ocdHandle == leftOCDHandle) {
            direction = OCI_LEFT;
        } else {
            direction = OCI_RIGHT;
        }
        
        // This connection is using some security features
        // The data packet received will be encrypted.
        // First decrypt it before handing it to the legacy application.
        cryptoState->decrypt ( 
                    (unsigned char*) blk->data,
                    blk->dataLen,
                    (unsigned char*) decryptedDataBuf,
                    &decryptedDataBufLen,
                    direction
                );

        finalBuf  = decryptedDataBuf;
        finalLen = decryptedDataBufLen;

        char tmpBuf[MAX_PRINT_BUF];
        PRINT_DEBUG1 (
                DEBUG_LEVEL_VERBOSE,
                "Decrypting with key symmetric key %s.\n",
                cryptoState->getSymKeyString (direction, tmpBuf)
            );

        if (compressionState != NULL && compressionState->state == COMPRESSION_STATE_OK) {
            
            //both encryption and compression

            compressionState->decompress (
                        decryptedDataBuf, decryptedDataBufLen, 
                        decompressedDataBuf, &decompressedDataLen
                    );
    
            PRINT_DEBUG2 (DEBUG_LEVEL_MINIMAL, "Pkt of original length %d decompressed to %d\n", decryptedDataBufLen, decompressedDataLen);
            finalLen = decompressedDataLen;
            finalBuf = decompressedDataBuf;
        }
    
    }   else { // no encryption
       
        
        if (compressionState != NULL && compressionState->state == COMPRESSION_STATE_OK) {
            // Only compression
            
            compressionState->decompress (
                            blk->data, blk->dataLen, 
                            decompressedDataBuf, &decompressedDataLen
                        );
        
            finalLen = decompressedDataLen;
            finalBuf = decompressedDataBuf;
            PRINT_DEBUG2 (DEBUG_LEVEL_MINIMAL, "Pkt of original length %d decompressed to %d\n", blk->dataLen, decompressedDataLen);
        
        } else { 
            
            // no encryption or compression.  Pass the received packet to the legacy application unmodified.
            finalBuf = blk->data;
            finalLen = blk->dataLen;
        }
    }

    //send to legacy app or middlebox

    if (myRole == OCI_ROLE_MIDDLEBOX) {
        assert (this != NULL);

        if (oci->middleBox != NULL) {
            //There exists a middlebox
            
            // We pass this packet up to the middle box.
            // The middle box may change the contents as well
            // as the length of the packet (in finalBuf and finalLen).
            // The middlebox should return only after the required
            // transformations have been done.
            
            bool okFlag = oci->middleBox->receive (this, finalBuf, &finalLen);
        
            if ( !okFlag) {
                
                PRINT_DEBUG2 (
                                INFO_LEVEL_WARNING,
                                "MiddleBox dropped packet for OCIState " UINT64_FORMAT_CODE ", %s\n",
                                myHandle, endPointName.c_str()
                            );
                return;
            }
        }
        
        
        //middlebox: forward it to the other side.
        if ( ocd == leftOCD && ocdHandle == leftOCDHandle) {

            oci->send(this, finalBuf, finalLen, OCI_RIGHT);
        } else if (ocd == rightOCD && ocdHandle == rightOCDHandle) {
            oci->send (this, finalBuf, finalLen, OCI_LEFT);
        } else {
             PRINT_DEBUG6 (
                            DEBUG_LEVEL_MINIMAL,
                            "In processDataBlock middlebox: received packet on unknown OCD %p %d.  left=%p %d, right=%p %d\n",
                            ocd, ocdHandle,
                            leftOCD, leftOCDHandle,
                            rightOCD, rightOCDHandle
                        );
        }
        
    } else {
        // One of the end points.  Send to legacy application
        
		// Dump pkt - start
		IPPacket tmpIPPkt (finalLen, finalBuf);
		tmpIPPkt.freeMemFlag = false;
		tmpIPPkt.printPacket ();

		// Dump pkt - end

        if ( ocd == leftOCD && ocdHandle == leftOCDHandle) {
            
            assert (receiverVN != NULL);
            // This block was received on the left side.
            oci->proxy->legacyAppLayer->sendToLegApp (receiverVN, finalBuf, finalLen);
        } else if (ocd == rightOCD && ocdHandle == rightOCDHandle) {
            // This block was received on the left side.
            assert (initiatorVN != NULL);
            oci->proxy->legacyAppLayer->sendToLegApp (initiatorVN, finalBuf, finalLen);
        } else {
            PRINT_DEBUG6 (
                            DEBUG_LEVEL_MINIMAL,
                            "In processDataBlock: received packet on unknown OCD %p %d.  left=%p %d, right=%p %d\n",
                            ocd, ocdHandle,
                            leftOCD, leftOCDHandle,
                            rightOCD, rightOCDHandle
                        );
        }
    }
}

/**
 * Process the list of generic blocks meant for this OCI State.
 * @param blk All fields of the block are in host byte order
 */
void OCIState::processBlocks (OCIPacketBlock *blk, OCD* ocd, OCDHandle ocdHandle) {

    int currBlockPos = 0;
    
    // Process each block
    while (1) {
          
        const OCIPacketBlock *currBlock = (OCIPacketBlock *) ( (char *) blk + currBlockPos);
        
        //Process the current block

        switch (currBlock->blockType) {
             
            case OCI_PATH_SETUP_BLOCK:
                PRINT_DEBUG0 (
                        DEBUG_LEVEL_MINIMAL,
                        "OCI_PATH_SETUP_BLOCK\n"
                    );
                handlePathSetupBlock ( (OCIPathSetupBlock *) blk, ocd, ocdHandle);
            break;
           
            case OCI_PATH_SETUP_COMPLETED_BLOCK:
                PRINT_DEBUG0 (
                        DEBUG_LEVEL_VERY_VERBOSE,
                        "OCI_PATH_SETUP_COMPLETED_BLOCK\n"
                    );
                handlePathSetupCompletedBlock (blk, ocd, ocdHandle);

            break;
        
            case OCI_PATH_SETUP_FAILED_BLOCK:
                handlePathSetupFailedBlock( blk, ocd, ocdHandle);
            break;
            
            case OCI_KEEPALIVE_BLOCK:
                PRINT_DEBUG0 (
                        DEBUG_LEVEL_VERBOSE,
                        "OCI_KEEPALIVE_BLOCK\n"
                    );
                handleKeepAliveBlock ( (OCIPacketBlock *) blk, ocd, ocdHandle);
            break;

            case OCI_KEEPALIVE_REPLY_BLOCK:
                PRINT_DEBUG0 (
                        DEBUG_LEVEL_VERBOSE,
                        "OCI_KEEPALIVE_REPLY_BLOCK\n"
                    );
                handleKeepAliveReplyBlock ( (OCIPacketBlock *) blk, ocd, ocdHandle);
            break;
            
            case OCI_DATA_BLOCK:

                PRINT_DEBUG0 (
                        DEBUG_LEVEL_VERBOSE,
                        "OCI_DATA_BLOCK\n"
                    );
                processDataBlock ( (OCIDataBlock *)blk, ocd, ocdHandle);

            break;

            
            case PROXY_PACKET_BLOCK_SECURITY:
            {

                PRINT_DEBUG0 (
                        DEBUG_LEVEL_VERBOSE,
                        "PROXY_PACKET_BLOCK_SECURITY\n"
                    );

                if (cryptoState == NULL) {
                
                    PRINT_DEBUG0 (
                            DEBUG_LEVEL_VERBOSE,
                            "Received security block for Virtual Node with NULL crypto state."
                            "Creating new crypto state for this VN.\n"
                        );

                   cryptoState = new CryptoState (this->oci->proxy, this, this->oci->proxy->securityManager);
				}
                    
                cryptoState->receivedSecurityBlock ((SecurityBlock *) currBlock);
    
            }
            break;
           
            /*
            case PROXY_PACKET_BLOCK_FAKE_ADDRESS_NEGOTIATION:
                PRINT_DEBUG0 (
                        DEBUG_LEVEL_MINIMAL,
                        "PROXY_PACKET_BLOCK_FAKE_ADDRESS_NEGOTIATION\n"
                    );
                
                if (fakeAddressNegotiator == NULL) {
                
                    PRINT_DEBUG0 (
                            DEBUG_LEVEL_MINIMAL,
                            "Received Fake Address Negotiation block for Virtual Node with NULL fakeAddressNegotiator."
                            "Creating new fakeAddressNegotiator for this VN.\n"
                        );

                    fakeAddressNegotiator = new FakeAddressNegotiator (this, this->proxy);
                } 
                    
                fakeAddressNegotiator->receivedBlock ((FakeAddressNegotiationBlock *) currBlock);

            break;
           
            case PROXY_PACKET_BLOCK_COMPRESSION:
                PRINT_DEBUG0 (
                        DEBUG_LEVEL_MINIMAL,
                        "PROXY_PACKET_BLOCK_COMPRESSION\n"
                    );
                if ( compressionState == NULL) {
                    PRINT_DEBUG0 (
                            DEBUG_LEVEL_MINIMAL,
                            "Received compression message at NULL compression state.\n"
                            "Creating new Compression State\n"
                        );
                    
                    compressionState = new CompressionState (this);
                }
                compressionState->receivedBlock ((CompressionPacketBlock *) currBlock);
            break;

            */

            default:
                PRINT_DEBUG1 (
                            DEBUG_LEVEL_MINIMAL,
                            "Unknown generic block with number %d received.\n",
                            currBlock->blockType
                        );
            break;
            
        }

        if (currBlock->nextBlock == 0) { 
            //No more blocks to process.
            break;
        } else {
            currBlockPos += currBlock->nextBlock;
        }
    }

}


/**
  * Pack all the outstanding blocks into the specified buffer.
  * The blocks which are packed are also removed from the toSendBlock list as
  * a side effect.
  * Return the total size of the blocks packed.
  */
int OCIState::packToSendBlocks (char *buf) {

    list<OCIPacketBlock *>::iterator itr = toSendBlockList.begin();

    /* List to keep track of blocks to delete.  A block is deleted
     * after it is packed.
     */
    list<list<OCIPacketBlock*>::iterator> deleteList;
    
    // Counter to keep track of next index at which to write into the buffer
    int tmpOffset = 0;
    
    // The total size of all the packed blocks 
    int totalPktSize = 0;
        
    OCIPacketBlock *currBlk;
  
    int numBlocks = toSendBlockList.size();
    int blockIndex = 0;
    
    while (itr != toSendBlockList.end()) {
        currBlk = *itr;
        
        //When this block is created, this field is set to the length of the block.
        //Only when the block is packed along with some other blocks this field changes.
        int currBlkSize = currBlk->nextBlock;   
            
        totalPktSize += currBlkSize;
 
        if (blockIndex == numBlocks - 1) {

            currBlk->nextBlock = 0;  // this is the last block in the packet.
        }
           
    
        //TODO Make sure that we don't have too many packets that we exceed MTU
        memcpy ((buf + tmpOffset), currBlk, currBlkSize);
        tmpOffset += currBlkSize;
           
        deleteList.push_back (itr);
        itr++;

        blockIndex++;
    
    }

     //now delete all the done blocks which we recorded in the first iteration.
    list<list<OCIPacketBlock *>::iterator>::iterator delItr = deleteList.begin();
    while (delItr != deleteList.end()) {
        toSendBlockList.erase (*delItr);
        delItr++;
    }
   
    return totalPktSize;   
}

/** Is this OCIState ready for data transfer?
 */
bool OCIState::isReady() {
    
    bool retVal = true;
    if (status != OCI_STATE_STATUS_READY) {
        retVal = false;
        printf ("isReady -- 1\n");
    }
    
    if (cryptoState != NULL &&
            ! cryptoState->isReady()) {

        retVal = false;
        printf ("isReady -- 2\n");
    }
    
    /*
    //TODO Include fake address negotiation result in the decision process
    if (fakeAddressNegotiator != NULL 
            && fakeAddressNegotiator->getState() == FAKE_ADDRESS_NEGOTIATION_IN_PROGRESS) {
        retVal = false;
        printf ("isReady -- 3\n");
    }

    if ( compressionState != NULL && compressionState->state != COMPRESSION_STATE_OK) {
        retVal = false;
        printf ("compression state not ready in VN->isReady()\n");
    }
    */
    return retVal;
            
}

/**
  * Constructs a path setup block for the path corresponding to this
  * oci state.
  * @param blockSize The size of the path setup block is returned via this parameter.
  */
OCIPathSetupBlock* OCIState::getPathSetupBlock (int *blockSize, const string finalEndPointName) {

    OCIPathSetupBlock* tmpBlk = pathInfo->getPathSetupBlock (blockSize, finalEndPointName);
    //PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL, "tmpBlok->pathInfoPkt->currHop = %d. \n", tmpBlk->path.currHop);
    memcpy (&(tmpBlk->header.dstHandleArray), &myHandle, sizeof (myHandle));

    return tmpBlk;
    
}

/**
  * Send KeepAlive message */
void OCIState::sendKeepAlive () {
        
    PRINT_DEBUG1 (
                DEBUG_LEVEL_VERY_VERBOSE,
                "Sending Keep Alive at OCIState " UINT64_FORMAT_CODE ".\n",
                myHandle
            );

    OCIPacketBlock tmpBlock;
    tmpBlock.version = PROXY_VERSION;
    tmpBlock.blockType = OCI_KEEPALIVE_BLOCK;
    tmpBlock.nextBlock = 0;
    memcpy (&(tmpBlock.dstHandleArray), &myHandle, sizeof (myHandle));

    int tmpBlockLen = sizeof (OCIPacketBlock);

   
    hton (&tmpBlock); 

	PRINT_DEBUG1 (DEBUG_LEVEL_VERBOSE, "Sending Keep Alive of size %d.\n",
					tmpBlockLen);
    rightOCD->send (rightOCDHandle, (char *) &tmpBlock, tmpBlockLen);

}

/**
  * Send KeepAliveReply message */
void OCIState::sendKeepAliveReply () {
 
    PRINT_DEBUG1 (
                DEBUG_LEVEL_VERY_VERBOSE,
                "Sending Keep Alive Reply at OCIState " UINT64_FORMAT_CODE ".\n",
                myHandle
            );

       
    OCIPacketBlock tmpBlock;
    tmpBlock.version = PROXY_VERSION;
    tmpBlock.blockType = OCI_KEEPALIVE_REPLY_BLOCK;
    tmpBlock.nextBlock = 0;
    memcpy ( tmpBlock.dstHandleArray, &myHandle, sizeof (myHandle));

    int tmpBlockLen = sizeof (OCIPacketBlock);

    hton (&tmpBlock);
    leftOCD->send (leftOCDHandle, (char *) &tmpBlock, tmpBlockLen);

}
/**
* This function is called when this OCI state receives
* a KeepAlive message
* @param blk The KeepAlive message.
* @param ocd The OCD on which the keep alive block was received.
* @param ocdHandle The OCD handle on which the keep alive block was received.
*/
void OCIState::handleKeepAliveBlock (OCIPacketBlock *blk, OCD* ocd, OCDHandle ocdHandle) {

    /** If this OCI is receiving a KeepAlive, that means that the path upstream is working.
      * Changing the status to READY on a keep alive is useful, especially when an end point
      * recovers after a crash, but the initiator hasn't timed out and doesn't send a new PathSetup
      * message.  Originally, for an endpoint of the connection, status was set to READY
      * when the path setup message is received and replied to.  In this case, there is no path setup
      * message sent.  So this is needed.
      */
    status = OCI_STATE_STATUS_READY;

    PRINT_DEBUG1 (
                  DEBUG_LEVEL_VERBOSE,
                  "Handling KeepAlive Block for handle " UINT64_FORMAT_CODE "\n",
                  myHandle
              );

    if ( myRole == OCI_ROLE_MIDDLEBOX ) {
        // Just relay the KeepAlive message
        hton (blk);
        rightOCD->send (rightOCDHandle, (char *) blk, sizeof (OCIPacketBlock));
        
    } else if (myRole == OCI_ROLE_ENDPOINT || myRole == OCI_ROLE_INITIATOR_ENDPOINT) {
        
        // Send the KeepAlive Reply
        sendKeepAliveReply();
        
    } else {
        
        PRINT_DEBUG0 (
                    DEBUG_LEVEL_MINIMAL,
                    "ERROR: KeepAliveMessage received at INITIATOR. \n"
                );
    }
}




/**
* This function is called when this OCI state receives
* a path setup message.
* @param blk The Path Setup block.
* @param ocd The OCD on which the path setup block was received.
* @param ocdHandle The OCD handle on which the path setup block was received.
*/
void OCIState::handlePathSetupBlock (OCIPathSetupBlock *blk, OCD* ocd, OCDHandle ocdHandle) {

   
    PRINT_DEBUG1 (
                  DEBUG_LEVEL_VERBOSE,
                  "Handling Path Setup Block for handle " UINT64_FORMAT_CODE "\n",
                  myHandle
              );

    PathInfo *currPathInfo = new PathInfo (& (blk->path));
  
    if (pathInfo == NULL) {
        //This is the first time this OCI is seeing a path setup message.
        pathInfo = currPathInfo;
   
    }  else {
       
        // pathInfo has been already set previously.

        if ( leftOCD == NULL && leftOCDHandle == NULL_HANDLE) {
            // We have to consider the case when the initiator == the receiver.
            myRole = OCI_ROLE_INITIATOR_ENDPOINT;
        }

        if ( myRole != OCI_ROLE_INITIATOR_ENDPOINT && ! pathInfo->equals (currPathInfo)) {
            
            // Previous Path Info does not match with current Path Info
            PRINT_DEBUG1 (
                            DEBUG_LEVEL_MINIMAL,
                            "Path Setup Message path inconsistency on OCI handle " UINT64_FORMAT_CODE ".\n",
                            myHandle
                        );
       
            delete currPathInfo;
            currPathInfo = NULL;

            oci->setupFailed(myHandle);

            return;
        }
    
        delete currPathInfo;
	    currPathInfo = NULL;	// currPathInfo variable should no longer be used.

    }

    leftOCD = ocd;
    leftOCDHandle = ocdHandle;

    //Set the end point of this OCIState to the (final) end point of
    // the path setup message.
    endPointName = pathInfo->getEndPointName();
    
    // First ask the left OCD if it wants to determine the next hop path.
    // The OCD level next hop can override the OCI level path info.
    // If the OCD level does specify a next hop, the OCI level pathinfo is carried
    // unmodified to the OCD specified next hop where it is processed.

    string ocdNextHopId = leftOCD->getNextHopId (leftOCDHandle);

    if (ocdNextHopId == "") {
        // The OCD does not want to specify the next hop path.
        // Process the OCI path info.
        
        int tmpHopCount = pathInfo->getHopCount();
       
        // Note: When the same node is the initiator as well as end point,
        // the ociState of that connection will have currHop always set to -1.
        // This is because there is only one OCIState corresponding to the connection
        // at the node even though the node has two roles.  You cannot have two OCIState
        // objects because they both should have the unique oci id.
       
        // BUG_FIX: Sept 5, 2005: Dilip:
        // The myRole=OCI_ROLE_INITIATOR_ENDPOINT  check was moved a level up
        // from the following if.
        if ( myRole != OCI_ROLE_INITIATOR_ENDPOINT && pathInfo->currHop < (tmpHopCount - 1)) {
            // If the connection loops back to the initiator, we do not allow it continue further.
            // i.e. we don't allow connections of the form A-->...-->A-->...
            //More hops remaining.  Do the remaining set up.
            
            myRole = OCI_ROLE_MIDDLEBOX;
    
            //Advance the current hop.
            //NOTE: This pathInfo object exclusively belongs to this OCI state.  But still
            //it needs to remain identical so that duplicate path setup messages don't get
            //confused
    
            //Setup the remainder of the path.
   
            HopInfo* nextHop = pathInfo->getHop(pathInfo->currHop + 1);
            
            nextHop->ocd = oci->proxy->getOCDWithId (nextHop->ocdId);
                
            OCD* nextHopOCD = nextHop->ocd;
    
            if (nextHopOCD == NULL) {
                
                PRINT_DEBUG1 (
                            DEBUG_LEVEL_MINIMAL,
                            "Unknown OCD with id %s.\n",
                            nextHop->ocdId.c_str()
                        );
                return;
    
            } 
                
            PRINT_DEBUG1 (
                                DEBUG_LEVEL_VERBOSE,
                                "Setting up next hop of path to %s.\n",
                                nextHop->hopEndPointName.c_str()
                            );
    
            char ociStateOptionBuf [MAX_PACKET_SIZE];
            int ociStateOptionLen = packToSendBlocks (ociStateOptionBuf);
                
            this->nextHopName = nextHop->hopEndPointName;
            this->nextHopNameType = OCI_NAME;
            /* OCI_NAME tells the OCD that the previous argument is a name and not an OCD-specific id */
                
            // Set up the OC-D level tunnel
            rightOCD = nextHopOCD;
            OCDHandle nextHopOCDHandle = nextHopOCD->setup (
                            this->nextHopName,
                            this->nextHopNameType,
                            ociStateOptionBuf, 
                            ociStateOptionLen, 
                            myHandle,
                            oci
                        );
    
            rightOCDHandle = nextHopOCDHandle;
           
            // Start checking for activity on this OCI State.
            scheduleActiveCheck();
               
            // Sometimes, for some OCDs, the entire setup is completed during the setup() function call;
            // there is won't be any future event that will trigger the setupSucceeded() callback.
            // For eg: LSIPv4 ocd when the name to be resolved is an actual IP address.
            // In those cases, we make a check if the ocd setup has already been completed.
            // If yes, we directly invoke setupSucceeded callback from here and don't add the setup timers.
          
            OCDStateInfo* nextHopOCDState = nextHopOCD->getStateInfo (nextHopOCDHandle);
            if (nextHopOCDState != NULL && nextHopOCDState->hasSetupCompleted ()) {
                // ocd setup already completed.
                PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL, "ocd setup completed quickly.\n");
                nextHopOCDState->notifySetupSuccess();

            } else {
                // Add ocdSetupTimer
		// Disable any existing timer
		if (timer != NULL) {
			timer->setDisabled();
			timer = NULL;
		}
                TObjectFunctor<OCIState> *ocdSetupTimerFunctor = 
                            new TObjectFunctor<OCIState> (this, &OCIState::ocdSetupTimerExpired);
                timer = TimerEvent::create(OCD_SETUP_TIMER, 0, (TFunctor *) ocdSetupTimerFunctor, true);
                oci->addTimerEvent(timer);
   
                // Processing continues on setupSucceeded() callback from OCD or 
                // when ocdSetupTimer expires.
            }
        } else {
            //We have reached the last hop.
    
            if ( myRole != OCI_ROLE_INITIATOR_ENDPOINT) {
                myRole = OCI_ROLE_ENDPOINT;
            }
    
            if (receiverVN == NULL) {
                // Setup Virtual Node interface to legacy application if this is the first time.
                
                receiverVN = oci->proxy->legacyAppLayer->getNewVN(this);
            }
    
            if (receiverVN == NULL) {
    
                PRINT_INFO1(
                                    INFO_LEVEL_MINIMAL, 
                                    "\nNo more fake IP addresses available. "//
                                    "Cannot accept this new connection at OCIHandle " UINT64_FORMAT_CODE "\n", 
                                    myHandle
                                ); 
                        
                //TODO Return PathSetupFailed message 
            } else {
    
                status = OCI_STATE_STATUS_READY;
                // Send OCI-path-setup-completed message.
                sendPathSetupCompletedMessage ();
              
                // Start the periodic check for activity on this oci state.
                scheduleActiveCheck();
                
            }
        }
    
    } else {
        // The OCD has specified the next hop path.
        myRole = OCI_ROLE_MIDDLEBOX;
  
        // This is an OCD-level middlebox.
        ocdMiddleBoxFlag = true;

        // OCD specific paths lie within the same OCD.
        rightOCD = leftOCD;
        
        PRINT_DEBUG1 (
                        DEBUG_LEVEL_MINIMAL,
                        "Setting up next OCD level hop of path to %s.\n",
                        ocdNextHopId.c_str()
                    );
        
        this->nextHopNameType = OCD_ID;
        /* tells the OCD that the previous arg is an OCD-specific name and not an OCI name */       
        this->nextHopName = ocdNextHopId;

        // Set up the OC-D level tunnel
        rightOCDHandle = rightOCD->setup (
                        this->nextHopName,
                        this->nextHopNameType,
                        NULL,
                        0,
                        myHandle,
                        oci
                    );
    
        // Add ocdSetupTimer
	// Disable any existing timer
	if (timer != NULL) {
	   timer->setDisabled();
	   timer = NULL;
	}
        TObjectFunctor<OCIState> *ocdSetupTimerFunctor = 
                    new TObjectFunctor<OCIState> (this, &OCIState::ocdSetupTimerExpired);
        timer = TimerEvent::create(OCD_SETUP_TIMER, 0, (TFunctor *) ocdSetupTimerFunctor, true);
        oci->addTimerEvent(timer);

        // Start the periodic activity checking on this OCI State
        scheduleActiveCheck ();
        
        // Processing continues on setupSucceeded() callback from OCD or 
        // when ocdSetupTimer expires.
    } 
}
    
/**
  * Send an OCI path setup message to the remote
  * end point.
  */
void OCIState::sendPathSetupMessage () {

    PRINT_DEBUG1 (
                DEBUG_LEVEL_VERY_VERBOSE,
                "Sending path setup message on oci handle " UINT64_FORMAT_CODE ".\n",
                myHandle
            );

    int tmpBlockLen = 0;
    
    // Start timer for receipt of SetupSuccess or SetupFailed block.
    assert (timer == NULL);
    
    TObjectFunctor<OCIState> *tmpFunctor = 
                new TObjectFunctor<OCIState> (this, &OCIState::ociSetupTimerExpired);
    timer = TimerEvent::create(OCI_SETUP_TIMER, 0, (TFunctor *) tmpFunctor, true);
    oci->addTimerEvent(timer);

    assert (rightOCD != NULL);

    OCIPathSetupBlock* tmpBlock = getPathSetupBlock(&tmpBlockLen, endPointName);
    
    hton ((OCIPacketBlock *) tmpBlock);
    rightOCD->send (rightOCDHandle, (char *)tmpBlock, tmpBlockLen);

    free (tmpBlock);
}

/**
  * Send a path setup completed message back to the initiator.
  */
void OCIState::sendPathSetupCompletedMessage () {

    OCIPacketBlock tmpBlock;
    tmpBlock.version = PROXY_VERSION;
    tmpBlock.blockType = OCI_PATH_SETUP_COMPLETED_BLOCK;
    tmpBlock.nextBlock = 0;
    memcpy ( tmpBlock.dstHandleArray, &myHandle, sizeof (myHandle));
   
    int tmpBlockLen = sizeof (OCIPacketBlock);

    hton (&tmpBlock);
    leftOCD->send (leftOCDHandle, (char *) &tmpBlock, tmpBlockLen);

}

/**
  * Send a path setup failed message back to the initiator.
  */
void OCIState::sendPathSetupFailedMessage () {

    OCIPacketBlock tmpBlock;
    tmpBlock.version = PROXY_VERSION;
    tmpBlock.blockType = OCI_PATH_SETUP_FAILED_BLOCK;
    tmpBlock.nextBlock = 0;
    memcpy ( tmpBlock.dstHandleArray, &myHandle, sizeof (myHandle));
   
    int tmpBlockLen = sizeof (OCIPacketBlock);

    hton (&tmpBlock);
    leftOCD->send (leftOCDHandle, (char *) &tmpBlock, tmpBlockLen);

}


/**
  * This function is called when a PathSetupCompleted Block is received 
  * @param blk The Path Setup Completed block
  * @param ocd The OCD on which the block was received.
  * @param ocdHandle The OCD handle on which the block was received.
  */
void OCIState::handlePathSetupCompletedBlock (OCIPacketBlock *pkt, OCD* ocd, OCDHandle ocdHandle) {

#ifdef BENCHMARK
	addCheckpoint (EVENT_PATH_SETUP_COMPLETED, 0);
#endif
 
    // The OCI State is active if when it gets a PathSetupCompleted block or a data block.
    // Other blocks like KeepAliveReply should not set the activeFlag to true.
    activeFlag = true;
    
    status = OCI_STATE_STATUS_READY;

    // We have successfully established end to end connection.
    // So we can reset number of path setup attempts made to zero,
    // so that we can start afresh if this path breaks at some later stage.
    numPathSetupAttempts = 0;

    // Disable the ociSetupTimer
    if (timer != NULL) {
        timer->setDisabled();
        timer = NULL;
    }
    
    if (myRole == OCI_ROLE_MIDDLEBOX) {
        // Relay the path setup message towards the initiator.
        hton (pkt);
        leftOCD->send (leftOCDHandle, (char *) pkt, sizeof (OCIPacketBlock));
    
    } else if (myRole == OCI_ROLE_INITIATOR || 
                    myRole == OCI_ROLE_INITIATOR_ENDPOINT) {

        // We are the initiator of the PathSetupMessage.
        // The end point should never receive a PathSetupCompleted message
        // as it is the one which originates it.

		if (pathInfo->securityInfo.protocol == SECURITY_PROTOCOL_CUSTOM_SSL) {
            // Initialize the custom security state.  
            // The message exchange required for security state will be done a little later.

			if (cryptoState != NULL) {
    			PRINT_DEBUG1 (
                    DEBUG_LEVEL_MINIMAL,
                    "Rare condition trigger: Pre-existing cryptoState deleted when PathSetupReply message received at OCIState " UINT64_FORMAT_CODE ".\n",
                    myHandle
                );
				delete (cryptoState);
				cryptoState = NULL;
			}
			// Create the new CryptoState
			// TODO: Will this cause a glitch in the established connection if this triggered by a duplicate stray PathSetupCompleted message?
            cryptoState = new CryptoState (oci->proxy, this, oci->proxy->securityManager);
            // This path requires some security features.
            // Start the security protocol
            cryptoState->initNewConnection ();
        
		} else {
            // No security is needed for this connection.
            // So we are done!
            connectionEstablished();
		}
 
        
        /* TODO FIX 
        if (vn->crashRecoveryInProgress && vn->dnsPacket != NULL) {
		    PRINT_DEBUG1( 
		            DEBUG_LEVEL_SUPER, 
				    "CrashRecovery: Changing status of VN (%s) to OK\n", 
				    vn->dnsPacket->reqURL.c_str()
			    );
        }
 
        */

    } else {
        PRINT_DEBUG0 (
                DEBUG_LEVEL_MINIMAL,
                "Error: ENDPOINT OCI state has received a path setup completed message.\n"
            );

        printf ("myRole = %d\n", myRole);
    }
}


/**
  * This function is called when a KeepAliveReply  Block is received 
  * @param blk The KeepAliveReply block
  * @param ocd The OCD on which the block was received.
  * @param ocdHandle The OCD handle on which the block was received.
  */
void OCIState::handleKeepAliveReplyBlock (OCIPacketBlock *pkt, OCD* ocd, OCDHandle ocdHandle) {
    
    PRINT_DEBUG1 (
                    DEBUG_LEVEL_VERBOSE,
                    "Handling Keep Alive Reply block at OCIState " UINT64_FORMAT_CODE ".\n",
                    myHandle
                );

    if (myRole == OCI_ROLE_MIDDLEBOX) {
        // Relay the KeepAliveReply block message towards the initiator.
	hton (pkt);
        leftOCD->send (leftOCDHandle, (char *) pkt, sizeof (OCIPacketBlock));
    
    } else if (myRole == OCI_ROLE_INITIATOR || 
                    myRole == OCI_ROLE_INITIATOR_ENDPOINT) {
    
		if (keepAliveState != NULL) {
			keepAliveState->receivedKeepAliveReply();
		} else {
			// This can occur when the proxy resumes from a suspend resume quickly.
			// The other end is still active, but the OCI state at this end did not get
			// a chance to send a Keep Alive message and construct the KeepAliveState 
			// before this message arrived.
			PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
					"OCIState " UINT64_FORMAT_CODE ": Ignoring Keep Alive Reply because KeepAliveState == NULL.\n",
					this->myHandle
				);
		}

    } else {
        PRINT_DEBUG0 (
                DEBUG_LEVEL_MINIMAL,
                "Error: ENDPOINT OCI state has received a KeepAlive reply completed message.\n"
            );
    }
}

/**
  * This function is called when a PathSetupFailed block is received.
  * @param blk The received block
  * @param ocd The OCD on which the block was received.
  * @param ocdHandle The OCD handle on which the block was received.
  */
void OCIState::handlePathSetupFailedBlock (OCIPacketBlock *pkt, OCD* ocd, OCDHandle ocdHandle) {

    PRINT_DEBUG1 (
                    DEBUG_LEVEL_MINIMAL,
                    "Handling Path Setup Failed block at OCIState " UINT64_FORMAT_CODE ".\n",
                    myHandle
                );

    // Disable the ociSetupTimer
    if (timer != NULL) {
        timer->setDisabled();
        timer = NULL;
    } 

    if (myRole == OCI_ROLE_MIDDLEBOX) {
        // Relay the PathSetupFaield block message towards the initiator.
        hton (pkt);
        leftOCD->send (leftOCDHandle, (char *) pkt, sizeof (OCIPacketBlock));
    
    } else if (myRole == OCI_ROLE_INITIATOR || 
                    myRole == OCI_ROLE_INITIATOR_ENDPOINT) {
  
        oci->destroyOCIState (this);
       
    } else {
        PRINT_DEBUG0 (
                DEBUG_LEVEL_MINIMAL,
                "Error: ENDPOINT OCI state has received a PathSetupFailed message.\n"
            );
    }
}

/**
  * This function is called periodically to check if this
  * OCIState is still active.
  * If the activeFlag of this OCIState is false, this OCI state is killed.
  * Else it just sets the flag to 0 and schedule itself to be called again
  * after some time.
  */
void OCIState::activeCheck () {

    activeCheckTimer = NULL;

    if ( ! activeFlag ) {
        PRINT_DEBUG1 (
                        DEBUG_LEVEL_MINIMAL,
                        "OCIState with handle " UINT64_FORMAT_CODE " found to be INACTIVE. Destroying it.\n\n",
                        myHandle
                    );
        
        // There has been no activity since the last time we checked.
        oci->destroyOCIState (this);
    } else {
       
        PRINT_DEBUG1 (
                        DEBUG_LEVEL_VERY_VERBOSE,
                        "OCIState with handle " UINT64_FORMAT_CODE " found to be ACTIVE.  Scheduling next active check.\n\n",
                        myHandle
                    );
        // Reset the flag to false.
        activeFlag = false;

        scheduleActiveCheck ();
    }
}

/**
  * This function schedules the activeCheck() function to be called
  * after the designated amount of time.
  */
void OCIState::scheduleActiveCheck() {
   
    // First disable any existing active Check timer before we add a new one.
    if (activeCheckTimer != NULL) {
        activeCheckTimer->setDisabled();
        activeCheckTimer = NULL;
    }
    
    TObjectFunctor<OCIState> *tmpFunctor = 
                new TObjectFunctor<OCIState> (this, &OCIState::activeCheck);
    activeCheckTimer = TimerEvent::create(OCI_ACTIVE_CHECK_PERIOD, 0, (TFunctor *) tmpFunctor, true);
    oci->addTimerEvent(activeCheckTimer);
}

/**
  * This function is called when a PathSetupCompleted or PathSetupFailed message
  * is not received within a designated time after sending the PathSetupRequest.
  */
void OCIState::ociSetupTimerExpired() {

    timer = NULL;

    PRINT_DEBUG1 (
                    DEBUG_LEVEL_MINIMAL,
                    "OCIState (" UINT64_FORMAT_CODE ") ociSetupTimer expired as no PathSetup Completed "
                    "or Failed message was received.\n",
                    myHandle
                );

    oci->destroyOCIState(this);
    
}

/**
  * This function is called when the setup request made to the OCD by this OCI state
  * as part of a setUpNewConnection() does not respond within a specified time.
  */
void OCIState::ocdSetupTimerExpired() {

    timer = NULL;

    PRINT_DEBUG1 (
                DEBUG_LEVEL_MINIMAL,
                "\n\n\t\t\t\tOCD Setup Timer expired for OCI with handle " UINT64_FORMAT_CODE ".\n\n\n",
                myHandle
            );

    //printf ("\n\nnumber path setup attempts = %d\n\n\n", numPathSetupAttempts);
    if ( numPathSetupAttempts < MAX_PATH_SETUP_ATTEMPTS) {
       
        oci->setupPath (this);
    
    } else {
        // We have tried too many times.  Give up.
        oci->destroyOCIState (this);
    }
}

void OCIState::startKeepAlives () {

    // Free up any existing keep alive states
    if (keepAliveState != NULL) {
        delete (keepAliveState);
        keepAliveState = NULL;
    }
    
    keepAliveState = new OCIKeepAliveState(this);
    keepAliveState->start();
}

/**
  * Destructor
  */
OCIState::~OCIState () {

    PRINT_DEBUG2 (
                    DEBUG_LEVEL_MINIMAL,
                    "Destroying OCI state " UINT64_FORMAT_CODE " (%p).\n",
                    myHandle, this
                );

    if (keepAliveState != NULL) {
        delete (keepAliveState);
        keepAliveState = NULL;
    }

    if (compressionState != NULL) {
        delete (compressionState);
        compressionState = NULL;
    }

    if (cryptoState != NULL) {
        delete (cryptoState);
        cryptoState = NULL;
    }

    if (timer != NULL) {
       timer->setDisabled();
       timer = NULL;
    }

    if (activeCheckTimer != NULL) {
        activeCheckTimer->setDisabled();
        activeCheckTimer = NULL;
    }
}

/**
  * This function is to notify the OCI State that its path is broken 
  */
void OCIState::notifyPathBroken () {

#ifdef BENCHMARK
	addCheckpoint (EVENT_PATH_BROKEN, 0);
#endif

    PRINT_DEBUG1 (
                    DEBUG_LEVEL_MINIMAL,
                    "\n\n\n\t\t\tPath broken for OCI State " UINT64_FORMAT_CODE ".\n\n\n\n",
                    myHandle
                );

    this->status = OCI_STATE_STATUS_NOT_READY;
    this->rightOCD->setStatus (this->rightOCDHandle, OCI_STATE_STATUS_NOT_READY);

    if (numPathSetupAttempts < MAX_PATH_SETUP_ATTEMPTS) {
        
        // Retry path setup.
     
        oci->setupPath (this);
        
    }
}

/**
  * This function is called when the security state associated with this OCIState
  * completes the security setup protocol - i.e. the exchange of keys, nonces,etc.
  */
void OCIState::securitySetupSuccess() {

    PRINT_DEBUG2 (
                    DEBUG_LEVEL_MINIMAL,
                    "Security protocol for OCIState[" UINT64_FORMAT_CODE ", %s] has completed successfully.\n",
                    myHandle, endPointName.c_str()
                );

    connectionEstablished();

}

/**
  * This function is called when the security state associated with this OCIState
  * detects that the security protocol has failed.
  */
void OCIState::securitySetupFailure() {
    PRINT_DEBUG2 (
                    DEBUG_LEVEL_MINIMAL,
                    "Security protocol for OCIState[" UINT64_FORMAT_CODE ", %s] has FAILED.\n",
                    myHandle, endPointName.c_str()
                );

    oci->setupFailed (myHandle);
    
}

/**
  * Put this state information into the specified buf */
void OCIState::getJournalRecord( char *buf) {

    char tmp_initiatorFakeIP[MAX_PRINT_BUF];
    if (initiatorVN != NULL &&  initiatorVN->fakeIP != NULL) {
        strcpy(tmp_initiatorFakeIP, initiatorVN->fakeIP->c_str());
    } else {

        strcpy(tmp_initiatorFakeIP, "<<NULL>>");
    }

    char tmp_receiverFakeIP[MAX_PRINT_BUF];
    if (receiverVN != NULL &&  receiverVN->fakeIP != NULL) {
        strcpy(tmp_receiverFakeIP, receiverVN->fakeIP->c_str());
    } else {

        strcpy(tmp_receiverFakeIP, "<<NULL>>");
    }
   
    char tmp_pathInfo[MAX_PRINT_BUF];
    if (pathInfo != NULL) {
        pathInfo->getJournalString (tmp_pathInfo);
    } else {
        strcpy (tmp_pathInfo, "<<NULL>>");
    }
    
    char tmp_leftOCDId[MAX_PRINT_BUF];
    if (leftOCD != NULL) {
        strcpy (tmp_leftOCDId, leftOCD->id.c_str());
    } else {
        strcpy (tmp_leftOCDId, "<<NULL>>");
    }
     
    char tmp_rightOCDId[MAX_PRINT_BUF];
    if (rightOCD != NULL) {
        strcpy (tmp_rightOCDId, rightOCD->id.c_str());
    } else {
        strcpy (tmp_rightOCDId, "<<NULL>>");
    }

    char tmp_endPointName[MAX_PRINT_BUF];
    if (endPointName == "") {
        sprintf (tmp_endPointName, "<<NULL>>");
    } else {
        sprintf (tmp_endPointName, "%s", endPointName.c_str());
    }

    char tmp_cryptoState [MAX_PRINT_BUF];
    if (cryptoState != NULL) {
        sprintf (tmp_cryptoState, ".cryptoState." UINT64_FORMAT_CODE , myHandle);
        //write the crypto state to disk right here.
        //We are assuming that this function call is made only when the Journal is about to be flushed 
        //to disk.  So this is okay.
        cryptoState->writeKeysToDisk (tmp_cryptoState);
    } else {
        sprintf (tmp_cryptoState, "<<NULL>>");
    }

    sprintf ( buf,
                
                UINT64_FORMAT_CODE "\t"    //myhandle
                "%s\t"    //initiator Fake IP
                "%s\t"    //receiver  Fake IP
                "%s\t"    //pathInfo
                "%s\t"    //end point name
                "%s\t"    //leftOCD id
                "%ld\t"    //leftOCD handle
                "%s\t"    //rightOCD id
                "%ld\t"    //right OCD handle
                "%d\t"    //ocdMiddleBoxFlag
                "%d\t"    //myRole
		"%s\t"	 // cryptoState

            ,
                myHandle,
                tmp_initiatorFakeIP,
                tmp_receiverFakeIP,
                tmp_pathInfo,
                tmp_endPointName,
                tmp_leftOCDId,
                leftOCDHandle,
                tmp_rightOCDId,
                rightOCDHandle,
                ocdMiddleBoxFlag,
                myRole,
		tmp_cryptoState
            );
                
	
}


/**
  * This function is called when the complete path, including compression, security
  * (if required) have been successfully setup.
  * This function answers the DNS queries from legacy applications,
  * starts sending keepAlives and schedule inactivity checks on this oci state.
  */
void OCIState::connectionEstablished() {

    char tmpBuf[MAX_PRINT_BUF];

    PRINT_DEBUG1 (
                    DEBUG_LEVEL_MINIMAL,
                    "Connection established for OCIState\n %s\n",
                    toString(tmpBuf)
                );
    // Notify the legacy application that connection
    // has been established.
    assert (initiatorVN != NULL);
    initiatorVN->answerDNSRequests(initiatorVN->fakeIP);

    //Start the keep alives
    startKeepAlives();

    // Start the periodic checks for activity on this OCIState.
    scheduleActiveCheck();

}

/**
  * Dumps a string representation of this OCI State into 
  * buf.  Returns a pointer to the same buf.
  */
char* OCIState::toString (char *buf) {

    sprintf (buf,   "########################## OCI STATE #####################################\n"
                    "handle=" UINT64_FORMAT_CODE " endpoint=%s \n"
                    "leftOCDHandle = %d rightOCDHandle=%d\n"
                    "status=%d role=%d ocdMiddleBoxFlag=%d\n",
                    myHandle, endPointName.c_str(), 
                    leftOCDHandle, rightOCDHandle, 
                    status, myRole, ocdMiddleBoxFlag
                );
    char tmpBuf[MAX_PRINT_BUF];
    if (initiatorVN != NULL && initiatorVN->fakeIP != NULL) {
        sprintf ( tmpBuf, "\t\tInitiator Fake IP : %s\n", initiatorVN->fakeIP->c_str());
        strcat(buf, tmpBuf);
    }

    if (receiverVN != NULL && receiverVN->fakeIP != NULL) {
        sprintf ( tmpBuf, "\t\tReceiver Fake IP : %s\n", receiverVN->fakeIP->c_str());
        strcat(buf, tmpBuf);
    }

    if (pathInfo != NULL) {
        sprintf (tmpBuf, "\t\tPathInfo: ");
        list<HopInfo *>::iterator itr = pathInfo->hopList.begin();
        while (itr != pathInfo->hopList.end()) {
            HopInfo *currHop = (*itr);
            strcat (tmpBuf, currHop->hopEndPointName.c_str());
            strcat (tmpBuf, ",");
            itr ++;
        }
        strcat (buf, tmpBuf);
        strcat (buf, "\n");
    }

    if (cryptoState != NULL) {

        unsigned char tmpLeftKey [EVP_MAX_KEY_LENGTH];
        unsigned char tmpRightKey [EVP_MAX_KEY_LENGTH];
        memcpy (tmpLeftKey, cryptoState->leftSymKey, EVP_MAX_KEY_LENGTH);
        memcpy (tmpRightKey, cryptoState->rightSymKey, EVP_MAX_KEY_LENGTH);

        tmpLeftKey[10] = '\0';
        tmpRightKey[10] = '\0';
        
        sprintf (tmpBuf, "\t\tCryptoState: left:%s right:%s\n", tmpLeftKey, tmpRightKey);
        strcat (buf, tmpBuf);
    }
   
    strcat (buf, "##########################################################################\n\n");
    return buf;
}

/**
  * Marks this OCIState as active.
  */
void OCIState::setActive () {

    activeFlag = true;
}

/**
  * Constructs a new OCI state from the specified journal string.
  * Returns NULL if something goes wrong.
  */
OCIState* OCIState::constructState (char *journalRecStr, OCI *oci) {


    OCIHandle tmp_myHandle;
    char tmp_initiatorFakeIP [MAX_PRINT_BUF];
    char tmp_receiverFakeIP [MAX_PRINT_BUF];
    char tmp_pathInfo [MAX_PRINT_BUF];
    char tmp_endPointName [MAX_PRINT_BUF];
    char tmp_leftOCDId [MAX_PRINT_BUF];
    OCDHandle tmp_leftOCDHandle;
    char tmp_rightOCDId [MAX_PRINT_BUF];
    OCDHandle tmp_rightOCDHandle;
    int tmp_ocdMiddleBoxFlag;
    int tmp_myRole;

    char tmp_cryptoState [MAX_PRINT_BUF];

    int numAssigned = sscanf ( journalRecStr,
                
                UINT64_FORMAT_CODE "\t"    //myhandle
                "%s\t"    //initiator Fake IP
                "%s\t"    //receiver  Fake IP
                "%s\t"    //pathInfo
                "%s\t"    //end point name
                "%s\t"    //leftOCD id
                "%ld\t"    //leftOCD handle
                "%s\t"    //rightOCD id
                "%ld\t"    //right OCD handle
                "%d\t"    //ocdMiddleBoxFlag
                "%d\t"    //myRole
		"%s\t"	  //cryptoState

            ,
                &tmp_myHandle,
                tmp_initiatorFakeIP,
                tmp_receiverFakeIP,
                tmp_pathInfo,
                tmp_endPointName,
                tmp_leftOCDId,
                &tmp_leftOCDHandle,
                tmp_rightOCDId,
                &tmp_rightOCDHandle,
                &tmp_ocdMiddleBoxFlag,
                &tmp_myRole,
                tmp_cryptoState
            );
                
    if (numAssigned < 12) {
        PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL,
                "Error while parsing journal entry.\n"
            );
        return NULL;
    }
    OCIState* newState = new OCIState (oci, tmp_myHandle);

    newState->myRole = tmp_myRole;
    newState->ocdMiddleBoxFlag = tmp_ocdMiddleBoxFlag;
    newState->endPointName = tmp_endPointName;
    
    if (strcmp (tmp_initiatorFakeIP, "<<NULL>>") != 0) {
        newState->initiatorVN = new VirtualNode (newState, oci->proxy);
        string tmpFakeIP = tmp_initiatorFakeIP;
        
        if (oci->proxy->allocateFakeIP (&tmpFakeIP) ) {
        
            newState->initiatorVN->fakeIP = new IPAddress (tmpFakeIP);
        
        } else {
            PRINT_DEBUG1 ( DEBUG_LEVEL_MINIMAL,
                            "Unable to allocate fakeIP %s during crash recovery\n",
                            tmpFakeIP.c_str()
                        );
            return NULL;
        }
        
        newState->initiatorVN->dnsPacket = new DNSPacket (newState->endPointName.c_str());
    } else {
        newState->initiatorVN = NULL;
    }
     
    if (strcmp (tmp_receiverFakeIP, "<<NULL>>") != 0) {
        newState->receiverVN = new VirtualNode (newState, oci->proxy);
        string tmpFakeIP = tmp_receiverFakeIP;
        if (oci->proxy->allocateFakeIP (&tmpFakeIP)) {
            
            newState->receiverVN->fakeIP = new IPAddress (tmpFakeIP);
         
        } else {
            PRINT_DEBUG1 ( DEBUG_LEVEL_MINIMAL,
                            "Unable to allocate fakeIP %s during crash recovery\n",
                            tmpFakeIP.c_str()
                        );
            return NULL;
        }
    } else {
        newState->receiverVN = NULL;
    }
    
    if (strcmp (tmp_leftOCDId, "<<NULL>>") != 0) {
        newState->leftOCD = oci->proxy->getOCDWithId (tmp_leftOCDId);
        newState->leftOCDHandle = tmp_leftOCDHandle;
    } else {
        newState->leftOCD = NULL;
        newState->leftOCDHandle = NULL_HANDLE;
    }

    if (strcmp (tmp_rightOCDId, "<<NULL>>") != 0) {
        newState->rightOCD = oci->proxy->getOCDWithId (tmp_rightOCDId);
        newState->rightOCDHandle = tmp_rightOCDHandle;
    } else {
        newState->rightOCD = NULL;
        newState->rightOCDHandle = NULL_HANDLE;
    }
   
    if (strcmp (tmp_cryptoState, "<<NULL>>") != 0) {
	newState->cryptoState = new CryptoState (oci->proxy, newState, oci->proxy->securityManager);
        newState->cryptoState->readKeysFromDisk (tmp_cryptoState);
    } else {
        newState->cryptoState = NULL;
    }

    newState->pathInfo = PathInfo::constructPathInfo (tmp_pathInfo, oci->proxy);

    return newState;

}
