/**
 *  \file OCIState.h
 *  \brief This file declares the object to keep track of 
 *  OCI level state information of a connection.
 */

#ifndef OCI_STATE_H
#define OCI_STATE_H

#include <list>
using namespace std;

#include "OCI.h"
#include "OCIHandle.h"
#include "../ocd/OCDHandle.h"
#include "OCIPacketBlock.h"
#include "../TimerEvent.h"

#include "../JournalRecord.h"


class CryptoState;
class CompressionState;
class OCIKeepAliveState;
class PathInfo;
class VirtualNode;
class OCD;

/** Temporary placeholder role */
#define OCI_ROLE_NULL       0

/** Indicates OCI state at the initiator of a path */
#define OCI_ROLE_INITIATOR  100

/** Indicates OCI state at the end point of a path */
#define OCI_ROLE_ENDPOINT   200

/** Indicates OCI state in the middle of a path */
#define OCI_ROLE_MIDDLEBOX  300

/** Indicate OCI state which is both initiator and the endpoint, at the same time */
#define OCI_ROLE_INITIATOR_ENDPOINT 400

#define OCI_STATE_STATUS_NOT_READY 0
#define OCI_STATE_STATUS_READY 1
#define OCI_STATE_STATUS_FAILED 3

class OCIState : public JournalRecord {

    /** List of blocks waiting to be sent out from this virtual node. */
    list<OCIPacketBlock *> toSendBlockList;
  
    OCI* oci;

    /** The virtual node for this OCI state to talk to legacy applications. 
     *  This VN is used at the initiator end.
     *  We need two separate VNs only because the same OCIState can be both 
     *  initiator and receiver.  For eg: A->A.
     */
    VirtualNode* initiatorVN;
    
    /** The virtual node for this OCI state to talk to legacy applications. 
     *  This VN is used at the receiver(end point) end.
     *  We need two separate VNs only because the same OCIState can be both 
     *  initiator and receiver.  For eg: A->A.
     */
    VirtualNode* receiverVN;
    
    /** The end to end path associated with this OCI state. */
    PathInfo *pathInfo;
   
    /** The name of the end point of this path.
     *  This may be valid only at initiator OCIstates.*/
    string endPointName;
   
    /** This denotes the name/OCD id of the next hop of this OCIState. */
    string nextHopName;

    /** Indicates whether nextHopName is an OCI_NAME or an OCD_ID */
    int nextHopNameType;
    
    /** The handle referring to this state */
    OCIHandle myHandle;

    /*
            Sender                   Receiver
               OCI                      OCI
       leftOCD    rightOCD      leftOCD    rightOCD
    */

    /** OCD facing the sender (initiator) of a connection. */
    OCD* leftOCD;
    /** The handle of the OCD state facing the send (initiator) of a connection. */
    OCDHandle leftOCDHandle;
     
    /** OCD facing the receiver of a connection. */
    OCD* rightOCD;
    /** The handle of the OCD state facing the receiver of a connection. */
    OCDHandle rightOCDHandle;
    
    char *setupOptionBlocks;
    int setupOptionLen;
 
    int status;

    /** This flag is true if this OCI state is assocated with an OCD-level middle box, i.e.,
      * this middle box was not visible at the OCI path level.
      */
    bool ocdMiddleBoxFlag;
    
    /**
      * What is the role of this OCI state?
      * OCI_ROLE_INITIATOR, OCI_ROLE_ENDPOINT, OCI_ROLE_MIDDLEBOX 
     */
    int myRole;
    
    /** The security state associated with this virtual node */
    CryptoState *cryptoState;

    CompressionState *compressionState;

    /** This object manages the periodic send and receipt of keep alive messages.
      * It notifies the OCIState whenver it detects that the path has broken.
      */
    OCIKeepAliveState *keepAliveState;
    
    /** General purpose timer event, if any, associated with this OCI state */
    TimerEvent *timer;

    /** This timer is used to periodically check this OCI State for activitiy */
    TimerEvent* activeCheckTimer;

    /** This flag is set to true whenever there is activity on this OCIState.
      * This flag is set to false each time activeCheckTimer expires.
      * If the activeCheck finds this flag to be false, then this OCIState is destroyed.
      */
    bool activeFlag;
    
    /** The number of times we have attempted to set up a path to the end point.  */
    int numPathSetupAttempts;

    /** This flag is set to true when this OCIState needs to be destroyed */
    bool killMeFlag;
    

    OCIState (OCI* oci, OCIHandle myHandle);
    virtual ~OCIState ();
    static OCIState* constructState (char *, OCI*);

    void processBlocks (OCIPacketBlock *blk, OCD*, OCDHandle);
    void processDataBlock (OCIDataBlock *blk, OCD*, OCDHandle layerHandle);
    

    void handlePathSetupBlock (OCIPathSetupBlock *, OCD*, OCDHandle);
    void handlePathSetupCompletedBlock (OCIPacketBlock *, OCD*, OCDHandle);
    void handlePathSetupFailedBlock (OCIPacketBlock *, OCD*, OCDHandle);
    void handleKeepAliveReplyBlock (OCIPacketBlock *pkt, OCD* ocd, OCDHandle ocdHandle);
    void handleKeepAliveBlock (OCIPacketBlock *pkt, OCD* ocd, OCDHandle ocdHandle);
    
    void sendBlock (OCIPacketBlock * blk, int blkSize, int direction);
   
    void sendKeepAlive();
    void sendKeepAliveReply();

    void sendPathSetupMessage ();
    void sendPathSetupCompletedMessage ();
    void sendPathSetupFailedMessage ();
 
    int packToSendBlocks(char *buf);

    OCIPathSetupBlock *getPathSetupBlock (int *, const string finalEndPointName);


    void ocdSetupTimerExpired();
    void ociSetupTimerExpired();

    void scheduleActiveCheck ();
    void activeCheck();
    
    void startKeepAlives();
    void notifyPathBroken();

    bool isReady();

    void getJournalRecord (char *);

    void securitySetupSuccess();
    void securitySetupFailure();

    void connectionEstablished();
    void connectionFailed();

    void setActive();

    char *toString (char *);

    friend class OCI;
    friend class LegacyAppLayer;
    friend class VirtualNode;
    friend class OCIKeepAliveState;
    friend class CryptoState;
    friend class BroMiddleBox;
    friend class Proxy;

    friend class LSIPv4_OCD; //HACK - Dilip
    
};

#endif
