/**
 *  \file CryptoState.cpp
 *  \brief CryptoState encapsulates all the state information associated with our custom SSL protocol.
 *  This file contains functions which handle various security protocol messages as well as functions which
 *  send out different security protocol messages.  In short, this file implements our custom SSL protocol.
*/

#include "SecurityManager.h"

#include "EncryptDecryptBlackBox.h"
#include "SecurityPacketBlocks.h"
#include "CryptoState.h"

#include "../../TimerEvent.h"
#include "../../TFunctor.h"
#include "../../Proxy.h"
#include "../../oci/OCIState.h"
#include "../../oci/OCI.h"
#include "../../ocd/OCD.h"
#include "../../common.h"

#include "../../debug.h"
#include "../../urlresolver/PathInfo.h"
#include "../../urlresolver/HopInfo.h"

#include <assert.h>

#include "cryptoFunctions.h"
#include "SecurityPacketBlocks.h"

CryptoState::CryptoState (Proxy *proxy, OCIState *vn, SecurityManager *securityManager) {

    this->proxy = proxy;
    this->vn = vn;
    this->securityManager = securityManager;

    this->state = CRYPTO_STATE_INIT;

    leftEncryptDecryptBlackBox = NULL;
    rightEncryptDecryptBlackBox = NULL;

    remotePublicKey = NULL;
    myPublicKey = NULL;
    myPrivateKey = NULL;

    timer = NULL;

    passThroughFlag = false;
}

/**
  * Sets the symmetric keys to be used for encrypting and
  * decrypting data.
  * @param direction OCI_LEFT or OCI_RIGHT: Which side symmetric key is to be set?
  */
void CryptoState::setSymKey (
        unsigned char symKey [EVP_MAX_KEY_LENGTH], 
        unsigned char symIV[EVP_MAX_IV_LENGTH],
        int direction
    ) {

    if (direction == OCI_LEFT) {
        
        memcpy (this->leftSymKey, symKey, EVP_MAX_KEY_LENGTH);
        memcpy (this->leftSymIV, symIV, EVP_MAX_IV_LENGTH);

        if (leftEncryptDecryptBlackBox != NULL) {
            delete leftEncryptDecryptBlackBox;
        }
        leftEncryptDecryptBlackBox = new EncryptDecryptBlackBox (symKey, symIV);
    } else {

        memcpy (this->rightSymKey, symKey, EVP_MAX_KEY_LENGTH);
        memcpy (this->rightSymIV, symIV, EVP_MAX_IV_LENGTH);

        if (rightEncryptDecryptBlackBox != NULL) {
            delete rightEncryptDecryptBlackBox;
        }
        rightEncryptDecryptBlackBox = new EncryptDecryptBlackBox (symKey, symIV);
    }
}

/**
  * Encrypts the specified data with the specified symmetric key.
  */
void CryptoState::encrypt (
            unsigned char* dataToEncrypt, uint16_t dataToEncryptLen,
            unsigned char* encryptedData, uint16_t *encryptedDataLen,
            SymKey * k
        ) {

    EncryptDecryptBlackBox::encrypt (
                            dataToEncrypt,
                            dataToEncryptLen,
                            encryptedData,
                            encryptedDataLen,
                            k
                        );
}
        
void CryptoState::encrypt ( 
            unsigned char* dataToEncrypt, uint16_t dataToEncryptLen,
            unsigned char* encryptedData, uint16_t *encryptedDataLen,
            int direction
        ) {

    if (passThroughFlag) {
        // Don't do anything.  Just copy the dataToEncrypt into the encryptedData field as it is.
        // TODO: Optimize, avoid copying.
        *encryptedDataLen = dataToEncryptLen;
        memcpy (encryptedData, dataToEncrypt, dataToEncryptLen);
    } else {
        
        if (direction == OCI_LEFT && leftEncryptDecryptBlackBox != NULL) {
        
            leftEncryptDecryptBlackBox->encrypt (dataToEncrypt, dataToEncryptLen, encryptedData, encryptedDataLen);
        
        } else if (direction == OCI_RIGHT && rightEncryptDecryptBlackBox != NULL) {
            
            rightEncryptDecryptBlackBox->encrypt (dataToEncrypt, dataToEncryptLen, encryptedData, encryptedDataLen);
        
        } else {
                // no encryption box was specified
                // So just plainly relay the data
                *encryptedDataLen = dataToEncryptLen;
                memcpy (encryptedData, dataToEncrypt, dataToEncryptLen);
        }
    }
    
}

/** Decrypts the specified data with the specified symmetric key.
  */
void CryptoState::decrypt (
            unsigned char* dataToDecrypt, uint16_t dataToDecryptLen,
            unsigned char* decryptedData, uint16_t *decryptedDataLen,
            SymKey *k
        ) {

    EncryptDecryptBlackBox::decrypt (dataToDecrypt, dataToDecryptLen, decryptedData, decryptedDataLen, k);

}            

void CryptoState::decrypt (
            unsigned char* dataToDecrypt, uint16_t  dataToDecryptLen,
            unsigned char* decryptedData, uint16_t *decryptedDataLen,
            int direction
        ) {

    if (passThroughFlag) {
        // Don't do anything.  Just copy the dataToDecrypt into the decryptedData field as it is.
        // TODO: Optimize, avoid copying.
        *decryptedDataLen = dataToDecryptLen;
        memcpy (decryptedData, dataToDecrypt, dataToDecryptLen);
    
    } else {
        
        if (direction == OCI_LEFT && leftEncryptDecryptBlackBox != NULL) {
    
            leftEncryptDecryptBlackBox->decrypt (dataToDecrypt, dataToDecryptLen, decryptedData, decryptedDataLen);
        } else if (direction == OCI_RIGHT && rightEncryptDecryptBlackBox != NULL) {
        
            rightEncryptDecryptBlackBox->decrypt (dataToDecrypt, dataToDecryptLen, decryptedData, decryptedDataLen);
        } else {
            
            // no decryption box present.  So just copy the data over.
            *decryptedDataLen = dataToDecryptLen;
            memcpy (decryptedData, dataToDecrypt, dataToDecryptLen);
        }
    }
}

/**
  * Encrypt with the specified public key.
  */
void CryptoState::encryptWithPublicKey (unsigned char *buf, uint16_t bufLen, unsigned char *encryptedBuf, uint16_t* encryptedLen, RSA* publicKey) {

    if (bufLen > RSA_size (publicKey) - 11) {
        PRINT_INFO0 (INFO_LEVEL_WARNING,
                "The buffer to  is too long to encrypt with the public key.\n"
            );
        return;
    }

    *encryptedLen = RSA_public_encrypt (bufLen,  buf, encryptedBuf, publicKey, CRYPTO_STATE_RSA_PADDING);

}

/**
  * Decrypts the specified symmetric key and IV using the private RSA key.
  * The decrypted keys are stored for later user.
  */
void CryptoState::setEncryptedSymKey ( 
        unsigned char *encryptedSymKeyBuf, int encryptedSymKeyLen,
        unsigned char *encryptedSymKeyIVBuf, int encryptedSymKeyIVLen
        ) {
 

}

/**
  * Sets the public key of the remote end point.
  * The public key is extracted from the certificate.
  */
void CryptoState::setRemotePublicKey(Certificate *cert) {
    this->remotePublicKey = loadPublicKeyFromBuffer (cert->publicKeyBuf, cert->publicKeyLen);
}

/**
  * Sets the name, public key and private key of this 
  * crypto state.  The name and public key are obtained
  * from the certificate.
  */
void CryptoState::setMyName_and_Keys (Certificate *cert, RSA* privKey) {
    this->myName = cert->name;
    this->myPublicKey = loadPublicKeyFromBuffer (cert->publicKeyBuf, cert->publicKeyLen);

    this->myPrivateKey = privKey;
}

/**
  * Generate a random symmetric key for this crypto state
  * @param direction OCI_LEFT or OCI_RIGHT : On which side should the symmetric key be set.
  */
void CryptoState::setRandomSymKey (int direction) {
    
    if (direction == OCI_LEFT) {
        
        generateRandomSymKey (this->leftSymKey, this->leftSymIV);
    
        if (leftEncryptDecryptBlackBox != NULL) {
            delete leftEncryptDecryptBlackBox;
        }
        leftEncryptDecryptBlackBox = new EncryptDecryptBlackBox (this->leftSymKey, this->leftSymIV);
    
    } else {
        
        generateRandomSymKey (this->rightSymKey, this->rightSymIV);
    
        if (rightEncryptDecryptBlackBox != NULL) {
            delete rightEncryptDecryptBlackBox;
        }
        rightEncryptDecryptBlackBox = new EncryptDecryptBlackBox (this->rightSymKey, this->rightSymIV);
    }

}

/**
  * Generate a random nonce for this crypto state object.
  */
void CryptoState::setRandomNonce () {
    generateNonce (this->nonce);
}


/** Returns the nonce associated with this cryptostate object */
unsigned char* CryptoState::getNonce() {
    return nonce;
}

/**
  * This function extracts the symmetric key from the sym key block.
  * The nonce contained in the block is decrypted and checked against
  * the nonce which was sent earlier.
  * The function returs true if the nonce matches.  Else returns false.
  */
bool CryptoState::processSymKeyBlock (SymKeyBlock *blk) {
	return true;
	// TODO Is this function actually used somewhere?
}

/**
  * Decrypts the proof string in the sym key reply block with the symmetric key
  * and checks if matches the mutually agreed value.A
  * @param k The symmetric key used for checking.
  */
bool CryptoState::verifyProofString (SymKeyReplyBlock *blk, SymKey *k) {
   
    // Actually the proof string is of maximum size PROOF_STRING_LENGTH.
    // We are allocating some extra space to protect against possible 
    // buffer overflow attacks by encrypted text which decrypts to more than
    // size PROOF_STRING_LENGTH
    char tmpProofString [CERT_RSA_KEY_SIZE/8]; 
    
    uint16_t tmpProofStringLen = 0;
    
    decrypt (
                (unsigned char *) blk->encryptedProofString,
                blk->encryptedProofStringLen,
                (unsigned char *) tmpProofString,
                &tmpProofStringLen, 
                k
            );
    
    tmpProofString[PROOF_STRING_LENGTH] = '\0'; //sanitizing
    
    if (strcmp (tmpProofString, PROOF_STRING) == 0) {
    
        // proof string matches
        return true;
    
    } else {

        // proof string does not match
        return false;

    }

}

/**
  * Cancel any existing timer and set a new timeout
  * for the specified event.
  * @functor the function to be called after the specified delay.
  */
void CryptoState::setTimer (
        long delaySeconds, long delayMilliseconds, 
        TFunctor* functor, char *description) {

    if (timer != NULL) {
        timer->setDisabled();
    }

    timer = TimerEvent::create (delaySeconds, delayMilliseconds, functor, description, true);
    proxy->addTimerEvent (timer);
}

void CryptoState::setNonce (unsigned char *n) {
    
    memcpy (this->nonce, n, NONCE_BIT_LEN/8);

}

/**
  * Disable the timer associated with this crypto state, if any *
  */
void CryptoState::disableTimer() {
    if (timer != NULL) {
        timer->setDisabled();
    }
}

/**
  * This function returns true if the receiver, i.e. the other end of the connection
  * initiated by the sender needs to be authenticated.
  */
bool CryptoState::requireReceiverAuth () {
    return true;
}

/**
  * This function returns true if the sender, i.e. the initiator of a connection needs to 
  * be authenticated.  This is of concern to the receiver.
  */
bool CryptoState::requireSenderAuth() {
    return false;
}

/**
  * This function returns true if the communication between the sender and receiver 
  * is to be encrypted.
  */
bool CryptoState::requireEncryption() {
    return true;
}

/**
  * Returns the name/id of this node.
  * This was the name for which the node received the certificate request.
  */
const char* CryptoState::getMyName() {
    
    return myName.c_str();
}

/**
  * This function is called when a sym key reply block is received for myId.
  * @param blk The received sym key reply block
  */
void CryptoState::receivedSymKeyReply (SymKeyReplyBlock *blk) {

    if (this->passThroughFlag) {
        PRINT_DEBUG0 (DEBUG_LEVEL_WARNING,
                    "SymKeyReply received for CryptoState which was set to pass through. Something is fishy!\n");
        return;
    }
    
    if (vn->myRole == OCI_ROLE_MIDDLEBOX) {
        // If this node is a middlebox, just relay the SymKeyReply block to
        // to the initiator.
        PRINT_DEBUG1 (
                        DEBUG_LEVEL_MINIMAL,
                        "SymKeyReply block received at middlebox [" UINT64_FORMAT_CODE "].  Relaying to initiator.\n",
                        vn->myHandle
                    );
        
        sendBlock ( (SecurityBlock *) blk, sizeof (SymKeyReplyBlock), OCI_LEFT);
        return;
    }

    PRINT_DEBUG2 (
                    DEBUG_LEVEL_MINIMAL,
                    "SymKeyReply block received at initiator[" UINT64_FORMAT_CODE ", %s].\n",
                    vn->myHandle, vn->endPointName.c_str()
                );
    

    //Get the symmetric key associated with the hop which sent the SymKeyReply
    string currHopName = blk->name;
    SymKey *currHopSymKey = getSymKeyFromMap (currHopName);
    
    //Verify the proof string.
    if (verifyProofString(blk, currHopSymKey)) {
        // Proof string matches.

        PRINT_DEBUG1 (
                    DEBUG_LEVEL_MINIMAL,
                    "Proof string correctly verified from hop %s.\n",
                    blk->name
                );
 
        //Remove the currHop symKey from the map.
        delSymKeyFromMap (currHopName);
    }

    if (hop_symKey_map.size() == 0) {
        
        //We have received sym key replies from all hops.
        // The connection is established.

        PRINT_DEBUG2 (
                        DEBUG_LEVEL_MINIMAL,
                        "All SymKeyReplies received by OCIState[" UINT64_FORMAT_CODE ", %s].\n"
                        "Security setup complete.\n",
                        vn->myHandle, vn->endPointName.c_str()
                    );

        if (state == CRYPTO_STATE_WAITING_FOR_SYMKEY_REPLY) {
           
            state = CRYPTO_STATE_CONNECTION_ESTABLISHED;

            vn->securitySetupSuccess ();

            // Disable the security setup timer.
            assert (timer != NULL);
            timer->setDisabled();
            timer = NULL;

            // clear all maps and lists.
            certificateWaitList.clear();
            recdCertBlockMap.clear();
            hop_symKey_map.clear();
            ocdCertNonceList.clear();
        

        } else {
                // The VN was not waiting for a sym key reply block when this sym key reply was
                // received.  
                PRINT_INFO2 (
                            INFO_LEVEL_WARNING,
                            "OCIState(" UINT64_FORMAT_CODE ", %s)->cryptoState received unexpected SymKeyReply block.",
                            vn->myHandle, vn->endPointName.c_str()
                        );
        }
 
    } 
}
    

/** Starts a new connection for the associated OCI State
  * The first security related message is sent.
  */
void CryptoState::initNewConnection() {

    state = CRYPTO_STATE_INIT;

    // Construct certificate request block
    int tmpNumCerts = vn->pathInfo->getHopCount();
   
    int tmpCertReqBlkSize = sizeof (CertificateRequestBlock) +
                                            tmpNumCerts * sizeof (CertificateName) ;

 
    char tmpReqBuf [MAX_PACKET_BUF_LEN];
    CertificateRequestBlock *certReqBlk = (CertificateRequestBlock *) tmpReqBuf;

    certReqBlk->mainHeader.mainHeader.version = PROXY_VERSION;
    certReqBlk->mainHeader.mainHeader.blockType = PROXY_PACKET_BLOCK_SECURITY;
    certReqBlk->mainHeader.mainHeader.nextBlock = sizeof (CertificateRequestBlock);
    memcpy ( certReqBlk->mainHeader.mainHeader.dstHandleArray, &(vn->myHandle), sizeof (OCIHandle));
    certReqBlk->mainHeader.msgCode = SECURITY_BLOCK_CERTIFICATE_REQUEST;

    certReqBlk->numCertificates = tmpNumCerts;
    certReqBlk->numOCDCertificates = 0;
    certReqBlk->currHop = 0;

    // Go through each Hop in the PathInfo
    // and add the name of the hop to the list of certificates requested.
    int certIndex = 0;
    list<HopInfo *>::iterator itr = vn->pathInfo->hopList.begin();
    while (itr != vn->pathInfo->hopList.end()) {

        HopInfo* currHop = (*itr);

        if (currHop->avoidAuthFlag) {
            // We don't want to authenticate this hop.
            // So we put the blank "" name in the slot for this hop
            // in the certificate request packet.
            strcpy(certReqBlk->nameList[certIndex].name, "");

            // Note that we don't wait for any certificate reply for this hop.
            
        } else {
            
            strcpy(certReqBlk->nameList[certIndex].name, currHop->hopEndPointName.c_str());

            // We need to wait for a certificate reply from each of the sender controlled
            // middleboxes before proceeding.
            certificateWaitList.push_back (currHop->hopEndPointName);
        }
        
        itr++;
        certIndex++;
    }


    // Send the certificate request block to the right, i.e. towards the end point.
    sendBlock ( (SecurityBlock *) certReqBlk, tmpCertReqBlkSize, OCI_RIGHT);

    // Set the timer for crypto setup completion.
    assert (timer == NULL); //no prior active timer exists.

            
    // Add ocdSetupTimer
    TObjectFunctor<CryptoState> *tmpFunctor = 
                        new TObjectFunctor<CryptoState> (this, &CryptoState::cryptoSetupTimerExpired);
    timer = TimerEvent::create(CRYPTO_SETUP_TIMER, 0, (TFunctor *) tmpFunctor, true);
    vn->oci->addTimerEvent(timer);

}

/**
  * This function is called when the setup timer associated with this crypto initialization expires.
  * i.e. we haven't finished all crypto message exchanges within the specified amount of time.
  * Hence, we abort.
  */
void CryptoState::cryptoSetupTimerExpired () {

    PRINT_DEBUG2 (
                    DEBUG_LEVEL_MINIMAL,
                    "CryptoSetup timer expired for OCIState[" UINT64_FORMAT_CODE ", %s]\n",
                    vn->myHandle, vn->endPointName.c_str()
                );
    this->vn->securitySetupFailure();
}


/**
  * This function is called when a security block intended for this 
  * crypto state is received.
  * @param securityBlk The SecurityBlock received. All the OCI header fields of this block 
  * are already in host byte order.  However, the security specific fields are 
  * still in network byte order.  This is because the OCI does not know anything
  * about the security protocol we implement.  Hence, we will convert all the 
  * security related fields of this block to host byte order first.
  */
void CryptoState::receivedSecurityBlock ( SecurityBlock *securityBlk) {
          
    ntoh (securityBlk);
     
    switch (securityBlk->msgCode) {
                  
        case SECURITY_BLOCK_CERTIFICATE_REQUEST:
            PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL, "SECURITY_BLOCK_CERTIFICATE_REQUEST\n");
            receivedCertificateRequest ((CertificateRequestBlock *) securityBlk);                
        break;

        case SECURITY_BLOCK_CERTIFICATE_REPLY:
            PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL, "SECURITY_BLOCK_CERTIFICATE_REPLY\n");
            receivedCertificateReplyBlock ((CertificateReplyBlock *) securityBlk);
        break;

        case SECURITY_BLOCK_CERTIFICATE_UNAVAILABLE:
            PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL, "SECURITY_BLOCK_CERTIFICATE_UNAVAILABLE\n");
            receivedCertificateNotFoundBlock( (SecurityBlock *) securityBlk);
            return; //TODO Will be interested in anything more once a CERTIFICATE NOT FOUND is received?
        break;

        case SECURITY_BLOCK_SYMKEY:
            PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL, "SECURITY_BLOCK_SYMKEY\n");
            receivedSymKey ((SymKeyBlock *) securityBlk);

            return; 
            // We are assuming that no other block is sent with this block.
            // This assumption won't hold when we are piggybacking data blocks
            // with this packet.

        break;

        case SECURITY_BLOCK_SYMKEY_OCD:
            PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL, "SECURITY_BLOCK_SYMKEY_OCD\n");
            receivedSymKeyOCD ( (SymKeyBlock *) securityBlk);
        break;
                       
        case SECURITY_BLOCK_SYMKEY_REPLY:
        {
                        
            PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL, "SECURITY_BLOCK_SYMKEY_REPLY\n");
                         
                           
            receivedSymKeyReply ((SymKeyReplyBlock *) securityBlk);

            return; 
                            
            // We are assuming that no other block is sent with this block.
            // This assumption won't hold when we are piggybacking data blocks
            // with this packet.
        }
        break;

        default:
            PRINT_INFO1 (
                        INFO_LEVEL_WARNING,
                        "Unknown SECURITY block code %d\n",
                        securityBlk->msgCode
                    );
            break;
    }
}


/** 
  * This function is called when a certificate not found message is received for
  * this crypto state.
  */
void CryptoState::receivedCertificateNotFoundBlock(SecurityBlock *blk) {

    if (this->passThroughFlag) {
        PRINT_DEBUG0 ( DEBUG_LEVEL_WARNING,
                "CertificateNotFoundBlock received for CryptoState which was set to pass through. Something is fishy!\n");
        return;
    }
    
    if (vn->myRole == OCI_ROLE_INITIATOR || vn->myRole == OCI_ROLE_INITIATOR_ENDPOINT) {
        // The certificate not found message has reached the initiator.
        
        vn->status = OCI_STATE_STATUS_FAILED;
        vn->securitySetupFailure();
    
    } else {
        
        // The certificate not found message has reached a middlebox.  Just relay it to the left.
        sendBlock ((SecurityBlock *) blk, sizeof (SecurityBlock), OCI_LEFT);
    }
}

/**
  * This function is called when a certificate reply block is received for this crypto state.
  */
void CryptoState::receivedCertificateReplyBlock (CertificateReplyBlock * certReplyBlock) {
 
    if (this->passThroughFlag) {
        PRINT_DEBUG0 ( DEBUG_LEVEL_WARNING,
                    "CertificateReplyBlock received for CryptoState which was set to pass through. Something is fishy!\n"
                );
        return;
    }
    
    if ( vn->myRole == OCI_ROLE_INITIATOR || vn->myRole == OCI_ROLE_INITIATOR_ENDPOINT ) {
        // The certificate reply has reached the initiator node.
        
        // Verify the received certificate.  If correct, remove it from the wait list.
    
        // Verify the received certificate.
        Certificate* tmpCert = &certReplyBlock->certificate;
  
        // First check if the certificate belongs to whom it says.
        if ( ! securityManager->checkCertificate (tmpCert->name, tmpCert) ) {
                
            PRINT_INFO1 (
                        INFO_LEVEL_WARNING,
                        "The certificate given by the host named %s is INVALID!\n",
                        tmpCert->name
                );
                
            return;
    
        } 
        
        PRINT_DEBUG1 (
                            DEBUG_LEVEL_MINIMAL,
                            "The certificate given by the host named %s verified to be CORRECT!\n",
                            tmpCert->name
                    );
    
        
        // Remove this certificate from the wait list.
        removeCertificateFromWaitList (tmpCert->name);

        CertificateReplyBlock *newTmpCertReplyBlk = new CertificateReplyBlock;
        //TODO FREE
        memcpy (newTmpCertReplyBlk, certReplyBlock, sizeof (CertificateReplyBlock));

        // Remember this certificate reply block.  
        // We need the public key and nonce contained in this when we are
        // going to send the symmetric keys.
        addCertBlockToRecdMap (newTmpCertReplyBlk);
        
    
        if ( certificateWaitList.size() == 0) {
            // We have received all certificates we are waiting for.
            // We can now proceed to send SymKeyReplies.
            PRINT_DEBUG2 (
                            DEBUG_LEVEL_MINIMAL,
                            "OCIState [" UINT64_FORMAT_CODE ", %s] has received ALL necessary certificate replies.\n",
                            vn->myHandle, vn->endPointName.c_str()
                        );

            // Now send the symmetric keys to all the sender controlled middleboxes.
            sendSymKeyBlock();


        }
    
    } else {

        //The certificate reply has reached an intermediate node.
        //Just relay it to the left, towards the initiator.

        sendBlock ((SecurityBlock *) certReplyBlock, sizeof (CertificateReplyBlock), OCI_LEFT);

    }
    
}

/**
  * This function sends the symmetric keys to be used by each of the sender controlled middleboxes.
  *
  */
void CryptoState::sendSymKeyBlock () {

    PRINT_DEBUG2 (
                    DEBUG_LEVEL_MINIMAL, 
                    "Sending SymKeyPacket at OCIState [" UINT64_FORMAT_CODE ", %s]\n",
                    vn->myHandle, vn->endPointName.c_str()
                );
    
        
    state = CRYPTO_STATE_WAITING_FOR_SYMKEY_REPLY;


    // First generate a symmetric key for the initiator, i.e. myself
    // We set the key on the right side of the connection.
    // The left side of the initiator is NULL, unless the initiator is also 
    // the receiver.
    //  (left =  NULL) A (right=K1)  ---->  (left=K1) M1 (right=K2) ---> (left=K2) M2 (right=K3) -->  (left=K3) B (right = NULL)
    setRandomSymKey(OCI_RIGHT);
    
    unsigned char prevSymKey[EVP_MAX_KEY_LENGTH];
    unsigned char prevSymIV[EVP_MAX_IV_LENGTH];

    memcpy (prevSymKey, this->rightSymKey, EVP_MAX_KEY_LENGTH);
    memcpy (prevSymIV, this->rightSymIV, EVP_MAX_IV_LENGTH);
    
    // Note:
    // In the SymKey block we only include symkeys for the hops to be authenticated.
    // Unlike in the certificate request, we don't include placeholders for hops that
    // don't need to be authenticated.
    int numAuthHops = vn->pathInfo->getAuthHopCount();

    // Construct the sym key reply block header
    char tmpPktBuf [MAX_PACKET_BUF_LEN];
    SymKeyBlock *symKeyBlk = (SymKeyBlock *) tmpPktBuf;
    int symKeyBlkSize = sizeof (SymKeyBlock) + numAuthHops * sizeof (SymKeyPair);
        
    symKeyBlk->mainHeader.mainHeader.version = PROXY_VERSION;
    symKeyBlk->mainHeader.mainHeader.blockType = PROXY_PACKET_BLOCK_SECURITY;
    symKeyBlk->mainHeader.mainHeader.nextBlock = symKeyBlkSize;
    memcpy ( symKeyBlk->mainHeader.mainHeader.dstHandleArray, &(vn->myHandle), sizeof (OCIHandle));
    
    symKeyBlk->mainHeader.msgCode = SECURITY_BLOCK_SYMKEY;

    symKeyBlk->currHop = 0;
    symKeyBlk->numHops = numAuthHops;

    int hopIndex = 0;

    // For each hop to be authenticated
    list<HopInfo *>::iterator itr = vn->pathInfo->hopList.begin();
    while (itr != vn->pathInfo->hopList.end()) {

        HopInfo* currHop = (*itr);
        
        if ( !currHop->avoidAuthFlag) {
            // This hop needs to be authenticated.
            
            CertificateReplyBlock *certReplyBlock = getCertBlockFromRecdMap (currHop->hopEndPointName);
            assert (certReplyBlock != NULL);
        
            printf ("\n\n\t\t\tSending symkey to %s|%s.\n", certReplyBlock->certificate.name, currHop->hopEndPointName.c_str());
    
            RSA* currHopPubKey  = loadPublicKeyFromBuffer (
                                            certReplyBlock->certificate.publicKeyBuf, 
                                            certReplyBlock->certificate.publicKeyLen
                                        );

            assert (currHopPubKey != NULL);
        
            //TODO FREE

            // Remember that this key is associated with this hop name.  
            // We need this info to select the appropriate sym key for different
            // SymKeyReply blocks.
            SymKey* tmpSymKey = new SymKey;    //TODO FREE
            memcpy (tmpSymKey->key, prevSymKey, EVP_MAX_KEY_LENGTH);
            memcpy (tmpSymKey->iv, prevSymIV, EVP_MAX_IV_LENGTH);
            addSymKeyToMap (currHop->hopEndPointName, tmpSymKey);


            //Encrypt the nonce in the certReply block with the symmetric key of the previous hop.
            encrypt (
                    certReplyBlock->nonce, NONCE_BIT_LEN/8,
                    symKeyBlk->symKeyPairs[hopIndex].encryptedNonce, 
                    &(symKeyBlk->symKeyPairs[hopIndex].encryptedNonceLen),
                    tmpSymKey
                    
            );
        
            //Encrypt the previous hop private key and put it in the left of the symkey pair.
            encryptWithPublicKey ( 
                    prevSymKey,
                    EVP_MAX_KEY_LENGTH,
                    symKeyBlk->symKeyPairs[hopIndex].encryptedLeftSymKey,
                    &symKeyBlk->symKeyPairs[hopIndex].encryptedLeftSymKeyLen,
                    currHopPubKey
                );
                    
            encryptWithPublicKey ( 
                    prevSymIV,
                    EVP_MAX_IV_LENGTH,
                    symKeyBlk->symKeyPairs[hopIndex].encryptedLeftSymIV,
                    &symKeyBlk->symKeyPairs[hopIndex].encryptedLeftSymIVLen,
                    currHopPubKey
                );
       
            // If we are not at the last hop, generate and fill in the symmetric key for the next hop
            if (hopIndex < numAuthHops - 1) {

                generateRandomSymKey (prevSymKey, prevSymIV);

                //This is the symmetric key to be used on the right side of this hop.
                //Encrypt this with the symmetric key previously set for the left side of
                //this hop.  The hop decrypts the left symmetric key using its public key.
                //The right symmetric key is decrypted using the left symmetric key as
                //symmetric key decryption is faster.
            
                encrypt (
                        prevSymKey, EVP_MAX_KEY_LENGTH,
                        symKeyBlk->symKeyPairs[hopIndex].encryptedRightSymKey, 
                        &(symKeyBlk->symKeyPairs[hopIndex].encryptedRightSymKeyLen),
                        tmpSymKey
                    );
            
                encrypt (
                        prevSymIV, EVP_MAX_IV_LENGTH,
                        symKeyBlk->symKeyPairs[hopIndex].encryptedRightSymIV, 
                        &(symKeyBlk->symKeyPairs[hopIndex].encryptedRightSymIVLen),
                        tmpSymKey
                    );


            } else {
                 // We are the last hop.
                //The last hop does not need a right side symmetric key.
                symKeyBlk->symKeyPairs[hopIndex].encryptedRightSymKeyLen = 0;
                symKeyBlk->symKeyPairs[hopIndex].encryptedRightSymIVLen = 0; 
             
            }
            hopIndex++;
        } 
            
        itr++;

    }    
    
    // end the SymKeyBlock towards the end point, i.e. towards right.
    sendBlock ((SecurityBlock *) symKeyBlk, symKeyBlkSize, OCI_RIGHT);
}

/**
  * Received a request for a certificate.
  * @param certReqBlock The certificate request block that was received.
  */
void CryptoState::receivedCertificateRequest(CertificateRequestBlock *certReqBlock) {
    
    // The node receiving the certificate request checks if the currHop of the CertificateRequest
    // corresponds to its name.  
    // If yes, it sends CertificateReply to the initiator.  If it is not the last hop, it relays the certificate further.
    // If no, it sends CertificateUnavalable message to the initiator.  It does not relay the certificate further.

    // Extract the name of the certificate requested from this hop.
    char *tmpReqName = certReqBlock->nameList[certReqBlock->currHop].name;
    tmpReqName [CERT_MAX_NAME_LEN - 1] = '\0'; //sanitizing

    PRINT_DEBUG2 (
                    DEBUG_LEVEL_MINIMAL,
                    "Received certificate request.  CurrHop = %d, %s.\n",
                    certReqBlock->currHop, tmpReqName
                );
   
         
    // Get the requested certificate from the security manager
    Certificate *tmpMyCert = securityManager->getCertificate (tmpReqName);
    RSA* tmpMyPrivKey = securityManager->getPrivateKey(tmpReqName);
  
    // If the private key associated with the requested certificate does not
    // exist and the OCIState is associated with an OCD-specified middlebox,
    // Relay this CertificateRequest unmodified, but appended with the OCD-specified
    // certificate and nonce
   if ( (tmpMyPrivKey == NULL || tmpMyCert == NULL) && vn->ocdMiddleBoxFlag) {

        char tmpBuf [MAX_PACKET_BUF_LEN];
        CertificateRequestBlock *relayBlk = (CertificateRequestBlock *) tmpBuf;
        
        //Copy the original part of the block into the new block to be relayed forward.
        int reqBlkSize = sizeof (CertificateRequestBlock) + certReqBlock->numOCDCertificates * sizeof (CertNonce);
        memcpy (relayBlk, certReqBlock, reqBlkSize);
       
        string tmpMiddleBoxName = vn->leftOCD->getMiddleBoxName(vn->leftOCDHandle);
       
        if (tmpMiddleBoxName == "") {
            
            PRINT_DEBUG1 (
                            DEBUG_LEVEL_MINIMAL,
                            "MiddleBox Name not specified for OCD-level middlebox with handle %d.\n",
                            vn->leftOCDHandle
                        );
            
        }
        
        // Get the name of this OCD-specified middlebox.
        tmpMyCert = securityManager->getCertificate (tmpMiddleBoxName.c_str());
        tmpMyPrivKey = securityManager->getPrivateKey(tmpMiddleBoxName.c_str());
       
        if (tmpMyCert != NULL && tmpMyPrivKey != NULL) {
            
            // Set the certificate and private key of this crypto state
            setMyName_and_Keys (tmpMyCert, tmpMyPrivKey);
 
            // Generate a new random nonce.
            setRandomNonce();

            relayBlk->numOCDCertificates ++;
       
            CertNonce tmpCertNonce;
            memcpy ( &tmpCertNonce.cert, tmpMyCert, sizeof (Certificate));
            memcpy ( &tmpCertNonce.nonce, getNonce(), NONCE_BIT_LEN/8);
            
            // There is only one certificate in the chain, in this case.
            memcpy ( &relayBlk->ocdCertificates[relayBlk->numOCDCertificates - 1], &tmpCertNonce, sizeof(CertNonce));
 
            // We are adding one more certificate and nonce to the end of this packet.
            int relayBlkSize = reqBlkSize + sizeof (CertNonce);
        
            // Relay the augmented message to the right.
            sendBlock ((SecurityBlock *) relayBlk, relayBlkSize, OCI_RIGHT);

           
        } else {
            PRINT_DEBUG1 (
                            DEBUG_LEVEL_MINIMAL,
                            "Certificate or private key for %s not found!\n",
                            tmpMiddleBoxName.c_str()
                        );

            //TODO Can we send an error message back to the initiator?
            // But the initiator may not be aware of this middlebox in the first place!!!
        }

        return;
    }


    // This is not an OCD middle box and the requested certificate name is "".
    // This means that this hop need not be authenticated.
    if (strcmp(tmpReqName,"") == 0) {
        // Set the pass through flag to indicate that no encryption/decryption is to be done
        // by this CryptoState object.  It just passes the messages through as it is.
        this->passThroughFlag = true;

        // If the requested name is the blank string, it means that no certificate was
        // requested for this hop.  So we need to just relay the message across.
        // Also if this is the last hop, we need to authenticate any certificate inserted
        // by ocd specified middleboxes.
        goto relay_cert_request;
    }


    // This hop needs to be authenticated 
    //Check if certificate and the corresponding private key exists.  
    if (tmpMyCert == NULL || tmpMyPrivKey == NULL) {
        //If not, send CERTIFICATE_NOT_FOUND message straight away.
    
        
        /* Create a Certificate Not Found message */
        SecurityBlock *securityBlk = new SecurityBlock;

        securityBlk->mainHeader.version = PROXY_VERSION;
        securityBlk->mainHeader.blockType = PROXY_PACKET_BLOCK_SECURITY;
        securityBlk->mainHeader.nextBlock = sizeof (SecurityBlock);
        memcpy ( securityBlk->mainHeader.dstHandleArray, &(vn->myHandle), sizeof (OCIHandle));
      
        securityBlk->msgCode = SECURITY_BLOCK_CERTIFICATE_UNAVAILABLE;

        PRINT_INFO2 ( INFO_LEVEL_WARNING,
                    "OCIState(" UINT64_FORMAT_CODE "): Certificate for %s UNAVAILABLE. Sending CERT_NOT_FOUND message to initiator.\n",
                    this->vn->myHandle, tmpReqName);

    
        assert (this->vn != NULL);

        // Send this Certificate Unavailable message towards the initiator, i.e. towards the left.
        sendBlock ((SecurityBlock *) securityBlk, sizeof (SecurityBlock), OCI_LEFT);
   
        // Free up the memory allocated for the certificate not found message.
        delete (securityBlk);

        return;
    } 

    
        
    // The requested certificate and its associated private key are available.
    {   
        // Set the certificate and private key of this crypto state
        setMyName_and_Keys (tmpMyCert, tmpMyPrivKey);
        // Generate a new random nonce.
        setRandomNonce();

        int numCert = 1;
        
        int tmpBlkBufLen = sizeof (CertificateReplyBlock) + numCert * sizeof (Certificate);
        
        char tmpBlkBuf [MAX_PACKET_BUF_LEN];
    
        // Construct the certificate reply block
        CertificateReplyBlock *certReplyBlk = (CertificateReplyBlock *) tmpBlkBuf;
       
        certReplyBlk->mainHeader.mainHeader.version = PROXY_VERSION;
        certReplyBlk->mainHeader.mainHeader.blockType = PROXY_PACKET_BLOCK_SECURITY;
        certReplyBlk->mainHeader.mainHeader.nextBlock = tmpBlkBufLen;
        memcpy ( certReplyBlk->mainHeader.mainHeader.dstHandleArray, &(vn->myHandle), sizeof (OCIHandle));
     
        certReplyBlk->mainHeader.msgCode = SECURITY_BLOCK_CERTIFICATE_REPLY;
            
        // There is only one certificate in the chain, in this case.
        memcpy ( &certReplyBlk->certificate, tmpMyCert, sizeof(Certificate));
 
        // Copy the nonce to the certificate reply block
        memcpy ( certReplyBlk->nonce, getNonce(), NONCE_BIT_LEN/8);

        // Send the certificate reply block towards the initiator, i.e. to the left. 
        sendBlock ((SecurityBlock *) certReplyBlk, tmpBlkBufLen, OCI_LEFT);
    }

    // This code is ugly. TODO FIX, Dilip
relay_cert_request:
    
    // If this node is not the last node on the path, i.e. more certificate request exist,
    // Relay this certificate request further down the path.
    if (certReqBlock->currHop < certReqBlock->numCertificates - 1) {
         // Not the last hop
         PRINT_DEBUG2 (
                            DEBUG_LEVEL_MINIMAL,
                            "Certificate request on OCIState (" UINT64_FORMAT_CODE ") being relayed from %s to next hop.\n",
                            vn->myHandle, tmpReqName
                        );
         certReqBlock->currHop++;
         sendBlock ( 
                        (SecurityBlock *) certReqBlock, 
                        sizeof (CertificateRequestBlock) + certReqBlock->numCertificates * sizeof (CertificateName),
                        OCI_RIGHT
                    );
        
    } else {
        // We have reached the last hop.
        PRINT_DEBUG2 (
                         DEBUG_LEVEL_MINIMAL,
                         "Certificate request on OCIState (" UINT64_FORMAT_CODE ") has reached last hop (%s).\n",
                         vn->myHandle, tmpReqName
                     );
           
        // Now we need to authenticate any of the receiver-OCD specified middleboxes which
        // tagged their Certificates along with this CertificateRequest.
         
        for (int certIndex = 0; certIndex < certReqBlock->numOCDCertificates; certIndex++) {
          
            // Verify the certificate
            Certificate* tmpCert = &certReqBlock->ocdCertificates[certIndex].cert;
  
            // First check if the certificate belongs to whom it says.
            if ( ! securityManager->checkCertificate (tmpCert->name, tmpCert) ) {
                
                 PRINT_INFO1 (
                        INFO_LEVEL_WARNING,
                        "The certificate given by the host named %s is INVALID!\n",
                        tmpCert->name
                    );
                    
                return;
            }

            CertNonce *tmpCertNonce = new CertNonce;
            //TODO FREE
            memcpy ( tmpCertNonce, &certReqBlock->ocdCertificates[certIndex], sizeof (CertNonce));
              
            // This certificate is valid.  Add it to the list of OCD certificates.
            addOCDCertNonce (tmpCertNonce);
        } 
    } 
  
}

/**
  * This function is called when a SymKey message is received from the 
  * receiver.  This should happen only at receiver-initiated OCD-level
  * middleboxes.
  */
void CryptoState::receivedSymKeyOCD (SymKeyBlock *blk) {

    // This message should be received only at OCD-level receiver initiated
    // middleboxes.
    assert (vn->ocdMiddleBoxFlag);

    PRINT_DEBUG2 (
                    DEBUG_LEVEL_MINIMAL,
                    "Received OCD SymKey Block at with currHop = %d at OCIState[" UINT64_FORMAT_CODE "]\n",
                    blk->currHop,
                    vn->myHandle
                );

    int symKeyBlkSize = sizeof (SymKeyBlock) + blk->numHops * sizeof (SymKeyPair);

    SymKeyPair *currKeyPair = &blk->symKeyPairs[blk->currHop];
    RSA_private_decrypt (
                            currKeyPair->encryptedLeftSymKeyLen, 
                            currKeyPair->encryptedLeftSymKey, 
                            this->leftSymKey, 
                            myPrivateKey, 
                            CRYPTO_STATE_RSA_PADDING
                        );
    
    RSA_private_decrypt (
                            currKeyPair->encryptedLeftSymIVLen, 
                            currKeyPair->encryptedLeftSymIV, 
                            this->leftSymIV, 
                            myPrivateKey, 
                            CRYPTO_STATE_RSA_PADDING
                        );
   
    // Now initialize the left side encryption blackbox.
    if (leftEncryptDecryptBlackBox != NULL) {
        delete leftEncryptDecryptBlackBox;
    }
    leftEncryptDecryptBlackBox = new EncryptDecryptBlackBox (leftSymKey, leftSymIV);
    

    //Decrypt the nonce with the newly obtained left symmetric key
    unsigned char tmpNonce [ NONCE_BIT_LEN / 8];
    uint16_t tmpNonceLen = 0;
        
    decrypt (
                currKeyPair->encryptedNonce,
                currKeyPair->encryptedNonceLen,
                tmpNonce,
                &tmpNonceLen,
                OCI_LEFT
            );
        
    
    // Check if the nonce matches the nonce which was stored in the
    // crypto state earlier.
    if (isNonceEqual (tmpNonce, nonce) != 1) {

        PRINT_INFO2 (
                    INFO_LEVEL_WARNING,
                    "Nonce does not match for VN %s(" UINT64_FORMAT_CODE ").\n"
                    "Aborting connection.\n",
                    vn->endPointName.c_str(),
                    vn->myHandle
                );

        return;
       
    }
 
    uint16_t dummyLen = 0;
    // Decrypt the right hand symmetric key, again using the left side symmetric key.
    decrypt (
                currKeyPair->encryptedRightSymKey,
                currKeyPair->encryptedRightSymKeyLen,
                this->rightSymKey,
                &dummyLen,
                OCI_LEFT
            );
       
    unsigned char tmpRightSymIV [EVP_MAX_IV_LENGTH + 100];
    //TODO decrypt seems to exceed its buffer
        
    dummyLen = 0;
    decrypt (
                currKeyPair->encryptedRightSymIV,
                currKeyPair->encryptedRightSymIVLen,
                tmpRightSymIV,
                &dummyLen,
                OCI_LEFT
            );
    memcpy (this->rightSymIV, tmpRightSymIV, EVP_MAX_IV_LENGTH);
        
        
    //set the right side encrypt decrypt box
    if (rightEncryptDecryptBlackBox != NULL) {
        delete rightEncryptDecryptBlackBox;
    }

    rightEncryptDecryptBlackBox = new EncryptDecryptBlackBox (rightSymKey, rightSymIV);
   

    // clear all maps and lists.
    certificateWaitList.clear();
    recdCertBlockMap.clear();
    hop_symKey_map.clear();
    ocdCertNonceList.clear();
    
    if ( blk->currHop > 0) {
        // We are not the last hop for this message. 
        // Again remmeber, we are going in the reverse direction of normal SymKey messages.
       
        // Relay this message to the left after decrementing currHop 
       
        blk->currHop --;
        
        // Forward the SymKey message towards the initiator, i.e. to the left.
        sendBlock ((SecurityBlock *) blk, symKeyBlkSize, OCI_LEFT);
    }
    
}

/**
  * This function is called when a sym key message (block) is received
  * at a node.  The node should extract the SymKeyPair meant for it and
  * and forward the rest of the message to downstream hops, if is not the
  * end point.
  * @param blk The received sym key block
  */
void CryptoState::receivedSymKey (SymKeyBlock *blk) {

    int symKeyBlkSize = sizeof (SymKeyBlock) + blk->numHops * sizeof (SymKeyPair);
   
    // TODO: If the last hop is configured to avoid authentication, then
    // authentication of OCD level middleboxes won't work.  This is because
    // the symkey for the ocd level middleboxes used to be distributed when the
    // symkey packet reached the last hop.  But now, if the last hop has authentication
    // turned off, it never receves the symkey packet!
    // TODO: This can be fixed.  I will do it later. Dilip

    if (this->passThroughFlag) {
        // Forward the SymKey message towards the receiver, i.e. to the right.
        // Note that we don't increment the hop count.
        if ( blk->currHop < blk->numHops - 1) {
            sendBlock ((SecurityBlock *) blk, symKeyBlkSize, OCI_RIGHT);
        }
        
        return;
    }
    
    PRINT_DEBUG2 (
                    DEBUG_LEVEL_MINIMAL,
                    "Received SymKey Block at with currHop = %d at OCIState[" UINT64_FORMAT_CODE "]\n",
                    blk->currHop,
                    vn->myHandle
                );

    assert (blk->currHop < blk->numHops);
   
    // These variables are used only when there were ocd imposed middleboxes in the path.
    // i.e. ocdCertNonceList was not empty.
    unsigned char prevSymKey[EVP_MAX_KEY_LENGTH];
    unsigned char prevSymIV[EVP_MAX_IV_LENGTH];


    if (vn->ocdMiddleBoxFlag) {
        // If this cryptostate is associated with an OCD level middlebox,
        // we ignore this SymKey reply.  
        // We just relay it to the right, towards the receiver.
        // We are only interested in SymKeyOCDBlocks.
        sendBlock ((SecurityBlock *) blk, symKeyBlkSize, OCI_RIGHT);

        return;
    }
    
    SymKeyPair *currKeyPair = &blk->symKeyPairs[blk->currHop];
    RSA_private_decrypt (
                            currKeyPair->encryptedLeftSymKeyLen, 
                            currKeyPair->encryptedLeftSymKey, 
                            this->leftSymKey, 
                            myPrivateKey, 
                            CRYPTO_STATE_RSA_PADDING
                        );
    
    RSA_private_decrypt (
                            currKeyPair->encryptedLeftSymIVLen, 
                            currKeyPair->encryptedLeftSymIV, 
                            this->leftSymIV, 
                            myPrivateKey, 
                            CRYPTO_STATE_RSA_PADDING
                        );
   
    // Now initialize the left side encryption blackbox.
    if (leftEncryptDecryptBlackBox != NULL) {
        delete leftEncryptDecryptBlackBox;
    }
    leftEncryptDecryptBlackBox = new EncryptDecryptBlackBox (leftSymKey, leftSymIV);
    

    //Decrypt the nonce with the newly obtained left symmetric key
    unsigned char tmpNonce [ NONCE_BIT_LEN / 8];
    uint16_t tmpNonceLen = 0;
        
    decrypt (
                currKeyPair->encryptedNonce,
                currKeyPair->encryptedNonceLen,
                tmpNonce,
                &tmpNonceLen,
                OCI_LEFT
            );
        
    
    // Check if the nonce matches the nonce which was stored in the
    // crypto state earlier.
    if (isNonceEqual (tmpNonce, nonce) != 1) {

        PRINT_INFO1 (
                    INFO_LEVEL_WARNING,
                    "Nonce does not match for VN %s.\n"
                    "Aborting connection.\n",
                    vn->endPointName.c_str()
                );

        vn->securitySetupFailure();
        return;
       
    }

    
    if ( blk->currHop < blk->numHops - 1) {
        // We are not at the last hop.  So we need to set the right sym key also.
       
        uint16_t dummyLen = 0;
    
        // Decrypt the right hand symmetric key, again using the left side symmetric key.
        decrypt (
                currKeyPair->encryptedRightSymKey,
                currKeyPair->encryptedRightSymKeyLen,
                this->rightSymKey,
                &dummyLen,
                OCI_LEFT
            );
       
        unsigned char tmpRightSymIV [EVP_MAX_IV_LENGTH + 100];
        //TODO decrypt seems to exceed its buffer
        
        dummyLen = 0;
        decrypt (
                currKeyPair->encryptedRightSymIV,
                currKeyPair->encryptedRightSymIVLen,
                tmpRightSymIV,
                &dummyLen,
                OCI_LEFT
            );
        memcpy (this->rightSymIV, tmpRightSymIV, EVP_MAX_IV_LENGTH);
        
        
        //set the right side encrypt decrypt box
        if (rightEncryptDecryptBlackBox != NULL) {
            delete rightEncryptDecryptBlackBox;
        }

        if (currKeyPair->encryptedRightSymKeyLen != 0 && currKeyPair->encryptedRightSymIVLen != 0) {
            rightEncryptDecryptBlackBox = new EncryptDecryptBlackBox (rightSymKey, rightSymIV);
        } else {

            // no right symkey specified (probably last hop)
            // so there is no need for a right EncryptBlackBox
            rightEncryptDecryptBlackBox = NULL;
        }
        
        // Increase the current hop.
        blk->currHop ++;
        
        // Forward the SymKey message towards the receiver, i.e. to the right.
        sendBlock ((SecurityBlock *) blk, symKeyBlkSize, OCI_RIGHT);
   
    } else if (ocdCertNonceList.size() > 0) {
        // We are the last hop and there were some OCD imposed middleboxes.
        // Now we need to establish the symmetric keys at any OCD-receiver imposed middleboxes.
        // This is done by going through the list of OCD certificates which were received along with
        // the original certificate request.

        PRINT_DEBUG2 (
                    DEBUG_LEVEL_MINIMAL, 
                    "Sending SymKeyOCDPacket at OCIState [" UINT64_FORMAT_CODE ", %s]\n",
                    vn->myHandle, vn->endPointName.c_str()
                );

        // This is the number of OCD level middleboxes.
        int numHops = ocdCertNonceList.size();

        // Construct the sym key reply block header
        char tmpPktBuf [MAX_PACKET_BUF_LEN];
        SymKeyBlock *symKeyBlk = (SymKeyBlock *) tmpPktBuf;
        int symKeyBlkSize = sizeof (SymKeyBlock) + numHops * sizeof (SymKeyPair);
        
        symKeyBlk->mainHeader.mainHeader.version = PROXY_VERSION;
        symKeyBlk->mainHeader.mainHeader.blockType = PROXY_PACKET_BLOCK_SECURITY;
        symKeyBlk->mainHeader.mainHeader.nextBlock = symKeyBlkSize;
        memcpy (symKeyBlk->mainHeader.mainHeader.dstHandleArray, &(vn->myHandle), sizeof (OCIHandle));
        
        symKeyBlk->mainHeader.msgCode = SECURITY_BLOCK_SYMKEY_OCD;
   
        // In OCD SymKey reply we go in reverse direction, each OCD middlebox decrements current hop by 1.
        symKeyBlk->currHop = numHops - 1;
        symKeyBlk->numHops = numHops;
 
        //The leftmost OCD middlebox should its left symmetric key to the
        //left symmetric key of the receiver.
        // NOTE:  We assume that OCD middleboxes and OCI-middleboxes are not
        // interspersed. i.e. we don't have topologies like A - OCI - OCD- OCI -OCD - M.
        memcpy (prevSymKey, this->leftSymKey, EVP_MAX_KEY_LENGTH);
        memcpy (prevSymIV, this->leftSymIV, EVP_MAX_IV_LENGTH);
        
        //We are not waiting for the SymKeyReply from OCD level middleboxes
        // as it involve an extra half RTT. So the following commented statement
        // is not necessary.  It is left commented just to note that we didn't 
        // forget about changing the state.
        //state = CRYPTO_STATE_WAITING_FOR_SYMKEY_REPLY;

        int hopIndex = 0;
        
        list<CertNonce *>::iterator itr = ocdCertNonceList.begin();
        while (itr != ocdCertNonceList.end()) {
            CertNonce *currCertNonce = (*itr);

            printf ("\n\n\t\t\tpacking symkey for %s.\n", currCertNonce->cert.name);
    
            RSA* currHopPubKey  = loadPublicKeyFromBuffer (
                                            currCertNonce->cert.publicKeyBuf, 
                                            currCertNonce->cert.publicKeyLen
                                        );

            assert (currHopPubKey != NULL);
            //TODO FREE

            SymKey* tmpSymKey = new SymKey;    //TODO FREE
            memcpy (tmpSymKey->key, prevSymKey, EVP_MAX_KEY_LENGTH);
            memcpy (tmpSymKey->iv, prevSymIV, EVP_MAX_IV_LENGTH);

            // We are not waiting for SymKey  replies from OCD level middleboxes.
            // So the following block of lines are commented out.
            // Remember that this key is associated with this hop name.  
            // We need this info to select the appropriate sym key for different
            // SymKeyReply blocks.
            //addSymKeyToMap (currHop->hopEndPointName, tmpSymKey);


            //Encrypt the nonce in the certReply block with the symmetric key of the previous hop.
            encrypt (
                    currCertNonce->nonce, NONCE_BIT_LEN/8,
                    symKeyBlk->symKeyPairs[hopIndex].encryptedNonce, 
                    &(symKeyBlk->symKeyPairs[hopIndex].encryptedNonceLen),
                    tmpSymKey
                    
            );
        
            //Encrypt the previous hop private key and put it in the left of the symkey pair.
            encryptWithPublicKey ( 
                    prevSymKey,
                    EVP_MAX_KEY_LENGTH,
                    symKeyBlk->symKeyPairs[hopIndex].encryptedLeftSymKey,
                    &symKeyBlk->symKeyPairs[hopIndex].encryptedLeftSymKeyLen,
                    currHopPubKey
                );
                    
            encryptWithPublicKey ( 
                    prevSymIV,
                    EVP_MAX_IV_LENGTH,
                    symKeyBlk->symKeyPairs[hopIndex].encryptedLeftSymIV,
                    &symKeyBlk->symKeyPairs[hopIndex].encryptedLeftSymIVLen,
                    currHopPubKey
                );
       

            generateRandomSymKey (prevSymKey, prevSymIV);

            //This is the symmetric key to be used on the right side of this hop.
            //Encrypt this with the symmetric key previously set for the left side of
            //this hop.  The hop decrypts the left symmetric key using its public key.
            //The right symmetric key is decrypted using the left symmetric key as
            //symmetric key decryption is faster.
            
            encrypt (
                    prevSymKey, EVP_MAX_KEY_LENGTH,
                    symKeyBlk->symKeyPairs[hopIndex].encryptedRightSymKey, 
                    &(symKeyBlk->symKeyPairs[hopIndex].encryptedRightSymKeyLen),
                    tmpSymKey
            );
            
            encrypt (
                    prevSymIV, EVP_MAX_IV_LENGTH,
                    symKeyBlk->symKeyPairs[hopIndex].encryptedRightSymIV, 
                    &(symKeyBlk->symKeyPairs[hopIndex].encryptedRightSymIVLen),
                    tmpSymKey
            );


            hopIndex++;
            itr++;
    
        }

        // There actually were some OCD middleboxes in the path.
        // send the OCDSymKeyBlock towards the initiator, i.e to the left.
        sendBlock ((SecurityBlock *) symKeyBlk, symKeyBlkSize, OCI_LEFT);

        // The end point, i.e. me, should use the right symkey sent to the last node
        // as its left symmetric key.
        // We still need to send the proof string to the initiator using the original 
        // left symmetric key (which was in fact proposed by the initiator itself).
        // So we do this step a little later in this same function.
    }
    
    // Connection at this hop is now complete.
    state = CRYPTO_STATE_CONNECTION_ESTABLISHED;
    //TODO should we do vn->connectionEstablished()
  
    char tmpPrintBuf[MAX_PRINT_BUF];
    PRINT_DEBUG1 (
                DEBUG_LEVEL_MINIMAL,
                "Security setup completed for %s",
                vn->toString(tmpPrintBuf)
            );
                
    // Now send the symKey reply back to the initiator.
    // NOTE: intermediate hops as well as the end point invidually send the symmetric key reply.
        
    //Construct the SymKeyReplyBlock
        
    char tmpBuf[MAX_PACKET_BUF_LEN];
    SymKeyReplyBlock *symKeyReplyBlk = (SymKeyReplyBlock *) tmpBuf;
    
    symKeyReplyBlk->mainHeader.mainHeader.version = PROXY_VERSION;
    symKeyReplyBlk->mainHeader.mainHeader.blockType = PROXY_PACKET_BLOCK_SECURITY;
    symKeyReplyBlk->mainHeader.mainHeader.nextBlock = sizeof (SymKeyReplyBlock);
    memcpy (symKeyReplyBlk->mainHeader.mainHeader.dstHandleArray, &(vn->myHandle), sizeof (OCIHandle));
    
    symKeyReplyBlk->mainHeader.msgCode = SECURITY_BLOCK_SYMKEY_REPLY;
    encrypt (
                    (unsigned char*) PROOF_STRING,
                    sizeof (PROOF_STRING),
                    (unsigned char*) symKeyReplyBlk->encryptedProofString,
                    &symKeyReplyBlk->encryptedProofStringLen,
                    OCI_LEFT
                );

    // Put the name for which we are replying. 
    // This is needed because the initiator will get symkey replies from the various hops
    // invidually.  It needs the name to figure out which symmetric key to use to verify the 
    // proof string.
    strcpy (symKeyReplyBlk->name, myName.c_str());
    // Send the SymKeyReply packet to the initiator, i.e. to the left.
    sendBlock ((SecurityBlock *) symKeyReplyBlk, sizeof (SymKeyReplyBlock), OCI_LEFT);
    

    if (ocdCertNonceList.size () > 0) {
        // There were some OCD receiver inititated middleboxes in the path.
        // So we change the left symmetric key of this node to the right symmetric key of
        // the last ocd -middlebox in the path.

        memcpy (leftSymKey, prevSymKey, EVP_MAX_KEY_LENGTH);
        memcpy (leftSymIV, prevSymIV, EVP_MAX_IV_LENGTH);

        // TODO We shouldn't have instantiated this earlier in the function.
        if (leftEncryptDecryptBlackBox != NULL) {
            delete leftEncryptDecryptBlackBox;
        }
        
        leftEncryptDecryptBlackBox = new EncryptDecryptBlackBox (leftSymKey, leftSymIV);
 
    }
}


bool CryptoState::isReady () {

    if ( state == CRYPTO_STATE_CONNECTION_ESTABLISHED ) {

        return true;
    }

    return false;
}

/**
  * Removes the certificate with the specified name from the wait list.
  */
void CryptoState::removeCertificateFromWaitList ( char *certName ) {

    string tmpCertName = certName;
    certificateWaitList.remove (tmpCertName);
}

/**
  * Adds the specified certificate to the received certificate map.
  */
void CryptoState::addCertBlockToRecdMap (CertificateReplyBlock *certBlk) {

    string tmpCertName = certBlk->certificate.name;
    recdCertBlockMap.insert (pair<string, CertificateReplyBlock *> (tmpCertName, certBlk));
}

/**
  * Retries the certificate reply block with the specified name from the 
  * map of received replies.  Returns NULL if no block is found is found.
  */
CertificateReplyBlock* CryptoState::getCertBlockFromRecdMap (string name) {
    map<string, CertificateReplyBlock*>::iterator itr = recdCertBlockMap.find(name);
    
    if (itr != recdCertBlockMap.end()) {
        return (CertificateReplyBlock*)itr->second;
        
    } else {
            
        return NULL;
    }
}

/**
  * Adds the specified certificate nonce to the list of received ocd certificates,
  * keyed by the name of the certificate.
  */
void CryptoState::addOCDCertNonce (CertNonce *c) {

    ocdCertNonceList.push_back (c);
}

/**
  * Insert a hop name to symmetric key mapping.
  */
void CryptoState::addSymKeyToMap (string name, SymKey *k) {

    hop_symKey_map.insert ( pair <string, SymKey*> (name, k));
}

/**
  * Removes the symmetric key from the hopname->symkey map.
  */
void CryptoState::delSymKeyFromMap (string name) {
    
    hop_symKey_map.erase (name);
}

/**
  * Retrieves the symmetric key associated with a particular hop.
  */
SymKey * CryptoState::getSymKeyFromMap (string hopName) {
    map<string, SymKey*>::iterator itr = hop_symKey_map.find(hopName);
    
    if (itr != hop_symKey_map.end()) {
        return (SymKey*)itr->second;
        
    } else {
            
        return NULL;
    }
}

/**
  * Returns a string representation symmetric key being used
  * in the particular direction.
  */
char * CryptoState::getSymKeyString (int direction, char *buf) {

    int i = 0;

    if (direction == OCI_LEFT) {

        for (i = 0; i < EVP_MAX_KEY_LENGTH; i++) {
            buf[i] = leftSymKey [i];
        }
    } else {
        for (i = 0; i < EVP_MAX_KEY_LENGTH; i++) {
            buf[i] = rightSymKey [i];
        }
    }

    buf[i] = '\0';
    return buf;
}

/**
  * Destructor.
  */
CryptoState::~CryptoState () {

    if (remotePublicKey != NULL) {
        RSA_free (remotePublicKey);
    }

    if (timer != NULL) {
        timer->setDisabled();
        timer = NULL;
    }   
    
}

/** Writes the symmetric keys associated with this crypto state to the 
 * specified file.
 */
void CryptoState::writeKeysToDisk (char *fileName) {
    assert (fileName != NULL);

    FILE* fp = fopen (fileName, "w");
    if (fp == NULL) {
        PRINT_DEBUG1 (   DEBUG_LEVEL_MINIMAL,
           		"Unable to open %s to dump the keys associated with cryptostate\n",
                        fileName
                    );
        return;
    }

    fwrite (leftSymKey, sizeof (unsigned char), EVP_MAX_KEY_LENGTH, fp);
    fwrite (leftSymIV, sizeof (unsigned char), EVP_MAX_IV_LENGTH, fp);
    
    fwrite (rightSymKey, sizeof (unsigned char), EVP_MAX_KEY_LENGTH, fp);
    fwrite (rightSymIV, sizeof (unsigned char), EVP_MAX_IV_LENGTH, fp);
   
    fclose (fp);
 
}

/**
 *  Reads the symmetric keys from the specified file name 
 */
void CryptoState::readKeysFromDisk (char *fileName) {
    assert (fileName != NULL);
    
    FILE *fp = fopen (fileName, "r");
    if (fp == NULL) {
        PRINT_DEBUG1 (
  			DEBUG_LEVEL_MINIMAL,
 			"Unable to read keys from %s\n",
			fileName
		  );
	return;
    }

    fread ( leftSymKey, sizeof (unsigned char), EVP_MAX_KEY_LENGTH, fp);
    fread ( leftSymIV, sizeof (unsigned char), EVP_MAX_IV_LENGTH, fp);
    if (leftEncryptDecryptBlackBox != NULL) {
         delete leftEncryptDecryptBlackBox;
    }
    leftEncryptDecryptBlackBox = new EncryptDecryptBlackBox (leftSymKey, leftSymIV);

    fread ( rightSymKey, sizeof (unsigned char), EVP_MAX_KEY_LENGTH, fp);
    fread ( rightSymIV, sizeof (unsigned char), EVP_MAX_IV_LENGTH, fp);

    if (rightEncryptDecryptBlackBox != NULL) {
         delete rightEncryptDecryptBlackBox;
    }
    rightEncryptDecryptBlackBox = new EncryptDecryptBlackBox (rightSymKey, rightSymIV);
    
    fclose (fp);
    
}

/**
  * Converts the security block to network byte order
  * and sends it out.
  */
void CryptoState::sendBlock (SecurityBlock * blk, int blkSize, int direction) {
    hton (blk);
    this->vn->sendBlock ((OCIPacketBlock *) blk, blkSize, direction);
}
