/**
 *  \file CryptoState.h
 *  \brief The definition of the state information need to be kept by
 *  our custom SSL protocol for each connection.
*/

#ifndef CRYPTO_STATE_H
#define CRYPTO_STATE_H

#include <openssl/evp.h>
#include <string>
#include <list>

#ifndef _WIN32
#    include <netdb.h>
#endif

#include "certificate.h"
#include "SecurityPacketBlocks.h"
#include "CryptoParams.h"


// CryptoState->state values
#define CRYPTO_STATE_INIT 0
#define CRYPTO_STATE_WAITING_FOR_SYMKEY 100
#define CRYPTO_STATE_WAITING_FOR_SYMKEY_REPLY 200
#define CRYPTO_STATE_CONNECTION_ESTABLISHED 300

/** How time (in MILLISECONDS) should a piggyback packet be held waiting
  * for a normal packet to piggyback on?  If no packet arrives in this time,
  * the piggyback packet is sent directly.
  */
#define CRYPTO_STATE_SEND_TIMEOUT   100

/**
 * Don't wait for piggybacking.  Send this packet immediately. 
 */
#define CRYPTO_STATE_SEND_TIMEOUT_IMMEDIATE 1

#include "nonce.h"

class TimerEvent;
class Proxy;
class TFunctor;
class OCIState;

using namespace std;

#define CRYPTO_STATE_RSA_PADDING RSA_PKCS1_PADDING

/** The number of seconds to wait for the security setup. */
#define CRYPTO_SETUP_TIMER  20

class EncryptDecryptBlackBox;
class SecurityManager;

/**
  * This class is used as a container for all cryptographic state
  * of a virtual node.
  */
class CryptoState {

    /** A reference to the proxy this CryptoState is part of */
    Proxy* proxy;
    
    /** The symmetric key used for encryption/decryption on the left side of the connection.*/
    unsigned char leftSymKey [EVP_MAX_KEY_LENGTH];

    /** The IV used for encryption/decryption on the left side of the connection*/
    unsigned char leftSymIV [EVP_MAX_IV_LENGTH];

    /** The symmetric key used for encryption/decryption on the right side of the connection.*/
    unsigned char rightSymKey [EVP_MAX_KEY_LENGTH];

    /** The IV used for encryption/decryption on the right side of the connection*/
    unsigned char rightSymIV [EVP_MAX_IV_LENGTH];
 
    /** This box is used for al the encryption and decryption on the left side of the connection*/
    EncryptDecryptBlackBox *leftEncryptDecryptBlackBox;

    /** This box is used for al the encryption and decryption on the right side of the connection*/
    EncryptDecryptBlackBox *rightEncryptDecryptBlackBox;
  
    /** The public key of the other end point of the connection.
     */
    RSA* remotePublicKey;

    /** This flag is set to true if this crypto state does no encryption or decryption, i.e.
      * it lets the messages simply pass through.
      * This kind of functionality is needed when we don't want to authenticate some of the hops
      * in an otherwise authenticated list of hops.
      */
    bool passThroughFlag;
    
    /**
      * This is the name of this node.
      * The other end point thinks that it is communicating
      * with a machine with this name.  In other words,
      * this is the name which the remote end point had
      * requested a certificate for.
      */
    string myName;
    
    /** This is the public key associated with myName */
    RSA* myPublicKey;
    
    /** This is the private key associated with myName */
    RSA* myPrivateKey;
  
    /** The nonce sent by the receiver to the sender */
    unsigned char nonce [NONCE_BIT_LEN/8];

    /** The generic state to which this crypto state is associated */
    OCIState *vn;
   
    /** Pointer to the security manager to be used by this crypto state */
    SecurityManager* securityManager;
    
    /**
      * This is the list of names from which we are expecting certificates.
      */
    list<string> certificateWaitList;

    /**
      * The certificates contained in the certificate replies, keyed by the certificate name.
      */
    map<string, CertificateReplyBlock*> recdCertBlockMap;

    /**
      * The list of OCD Certificate Nonces received along with a CertificateRequest block.
      * The order of this list is important.
      */
    list<CertNonce *>    ocdCertNonceList;
    
    /**
      * Hop name to symmetric key mappings.
      */
    map<string, SymKey *> hop_symKey_map;
    
    /** A generic timer used for all timeouts associated with the crypto state */
    TimerEvent* timer;

    void removeCertificateFromWaitList (char *certName);
    
    void cryptoSetupTimerExpired ();

    void addOCDCertNonce (CertNonce *);

    void addCertBlockToRecdMap (CertificateReplyBlock *);
    CertificateReplyBlock *getCertBlockFromRecdMap (string name);

    void addSymKeyToMap (string name, SymKey *);
    SymKey* getSymKeyFromMap (string name);
    void delSymKeyFromMap (string name);

    void sendSymKeyBlock ();

    void encryptWithPublicKey (unsigned char *buf, uint16_t bufLen, unsigned char *encryptedBuf, uint16_t* encryptedLen, RSA* publicKey) ;

    
    void writeKeysToDisk (char *fileName);
    void readKeysFromDisk (char *fileName);


    void sendBlock (SecurityBlock * blk, int blkSize, int direction);
    
public:

    /** What is the current state of the crypto protocol 
      * For eg: waiting for symmetric key, waiting for symmetric key reply.
      */
    int state;
    
   
    CryptoState (Proxy *, OCIState *, SecurityManager *);
    ~CryptoState();
    
    void encrypt ( 
            unsigned char* dataToEncrypt, uint16_t dataToEncryptLen,
            unsigned char* encryptedData, uint16_t *encryptedDataLen,
            int direction
        );
             
    void encrypt ( 
            unsigned char* dataToEncrypt, uint16_t dataToEncryptLen,
            unsigned char* encryptedData, uint16_t *encryptedDataLen,
            SymKey *k
        );
             

    void decrypt (
            unsigned char* dataToDecrypt, uint16_t dataToDecryptLen,
            unsigned char* decryptedData, uint16_t *decryptedDataLen,
            int direction
        );
 
    void decrypt (
            unsigned char* dataToDecrypt, uint16_t dataToDecryptLen,
            unsigned char* decryptedData, uint16_t *decryptedDataLen,
            SymKey *k
        );
    
    void setRemotePublicKey (Certificate *cert);
    void setMyName_and_Keys (Certificate *cert, RSA* privKey);
    
    void setSymKey (
            unsigned char key [EVP_MAX_KEY_LENGTH], 
            unsigned char iv[EVP_MAX_IV_LENGTH],
            int direction
            
        );

    void setEncryptedSymKey (
        unsigned char *encryptedSymKeyBuf, int encryptedSymKeyLen,
        unsigned char *encryptedSymKeyIVBuf, int encryptedSymKeyIVLen
        );

    void setRandomSymKey (int direction);
    void setRandomNonce ();

    bool processSymKeyBlock (SymKeyBlock *blk);

    bool verifyProofString (SymKeyReplyBlock *blk, SymKey* key);

    void setTimer (
            long delaySeconds, long delayMilliseconds, 
            TFunctor* functor, char *description);
   
    void disableTimer();

    void setNonce (unsigned char *);
    
    unsigned char * getNonce();

    bool requireReceiverAuth();
    bool requireSenderAuth();
    bool requireEncryption();

    void setRemoteName(char* name);
    const char * getMyName();
    const char * getRemoteName();

    void initNewConnection();



    void receivedCertificateNotFoundBlock(SecurityBlock *);
    void receivedSecurityBlock ( SecurityBlock *securityBlk);
    void receivedCertificateRequest(CertificateRequestBlock *certReqBlock);
    void receivedCertificateReplyBlock (CertificateReplyBlock * certReplyBlk);
    void receivedSymKey (SymKeyBlock *blk);
    void receivedSymKeyOCD (SymKeyBlock *blk);
    void receivedSymKeyReply (SymKeyReplyBlock *blk);

    bool isReady();

    char *getSymKeyString (int direction, char *buf);

    friend class OCIState;
};


#endif
