/**
 *  \file EncryptDecryptBlackBox.cpp
 *  \brief This file defines the EncryptDecryptBlackBox.
 *  The EncryptDecryptBlackBox can be used to encrypt and decrypt data after setting up a symmetric key.
 */

#include <openssl/err.h>
#include <string.h>

#include "EncryptDecryptBlackBox.h"
#include "../../debug.h"

/**
  * Constructs a new encryption decryption black box.
  * A random symmetric key is generated here is used
  * for all encryption and decryption purposes.
  * @param key The symmetric key to be used for encryption 
  * and decryption.
  */
EncryptDecryptBlackBox::EncryptDecryptBlackBox (unsigned char key [EVP_MAX_KEY_LENGTH], unsigned char iv[EVP_MAX_IV_LENGTH]) {

    memcpy (this->key, key, EVP_MAX_KEY_LENGTH);
    memcpy (this->iv, iv, EVP_MAX_IV_LENGTH);
    
}

/**
  * Destructor:
  * Clean up all encryption/decryption state.
  */
EncryptDecryptBlackBox::~EncryptDecryptBlackBox () {

}

/**
  * Make sure encryptedData buffer has atleast dataToEncryptLen + EVP_CIPHER_block_size() bytes allocated.
  * Encrypt the data with the stored symmetric key.
  */
void EncryptDecryptBlackBox::encrypt (
            unsigned char* dataToEncrypt, uint16_t dataToEncryptLen,
            unsigned char* encryptedData, uint16_t *encryptedDataLen
        ) {

    EVP_CIPHER_CTX_init (&encryptCtx);
    
    int tmpLen;
    *encryptedDataLen = 0;
    
    if ( ! EVP_CipherInit_ex (&encryptCtx, ENC_ALGO,  NULL, key, iv, 1 /* encrypt */) ) {

        PRINT_INFO0 (INFO_LEVEL_WARNING,
                "Unable to initialize encryption context!\n"
                );
        ERR_print_errors_fp(stderr);
    }
 
    if ( ! EVP_CipherUpdate (&encryptCtx, encryptedData, &tmpLen, dataToEncrypt, dataToEncryptLen)) {

        PRINT_INFO0 (INFO_LEVEL_WARNING,
                "Unable to encrypt!\n"
                );
        ERR_print_errors_fp(stderr);
    }
    *encryptedDataLen += tmpLen;

    // Write any remaining parts of the encrypted data to the out buffer
    EVP_CipherFinal_ex(&encryptCtx, (encryptedData + *encryptedDataLen), &tmpLen); 

    *encryptedDataLen += tmpLen;
    
    EVP_CIPHER_CTX_cleanup (&encryptCtx);
}

/**
  * Decrypt the data with the stored symmetric key.
  */
void EncryptDecryptBlackBox::decrypt (
            unsigned char* dataToDecrypt, uint16_t dataToDecryptLen,
            unsigned char* decryptedData, uint16_t *decryptedDataLen
        ) {
    
    EVP_CIPHER_CTX_init (&decryptCtx);
    
    if ( ! EVP_CipherInit_ex (&decryptCtx, ENC_ALGO,  NULL, key, iv, 0 /* decrypt */) ) {
    
        PRINT_INFO0 (INFO_LEVEL_WARNING,
                "Unable to initialize decryption context!\n"
                );
        ERR_print_errors_fp(stderr);
    }

    int tmpLen;
    *decryptedDataLen = 0;

    if ( ! EVP_CipherUpdate (&decryptCtx, decryptedData, &tmpLen, dataToDecrypt, dataToDecryptLen)) {

        PRINT_INFO0 (INFO_LEVEL_WARNING,
                "Unable to decrypt!\n"
                );
        ERR_print_errors_fp(stderr);
    }
    
    *decryptedDataLen += tmpLen;
    
    EVP_CipherFinal_ex(&decryptCtx, (decryptedData + *decryptedDataLen), &tmpLen); 

    *decryptedDataLen += tmpLen;
    
    EVP_CIPHER_CTX_cleanup (&decryptCtx);
}

/**
  * Decrypt the data with the specified symmetric key.
  */
void EncryptDecryptBlackBox::decrypt (
            unsigned char* dataToDecrypt, uint16_t dataToDecryptLen,
            unsigned char* decryptedData, uint16_t *decryptedDataLen,
            SymKey *k
        ) {
     
    EVP_CIPHER_CTX decryptCtx;

    EVP_CIPHER_CTX_init (&decryptCtx);
    
    if ( ! EVP_CipherInit_ex (&decryptCtx, ENC_ALGO,  NULL, k->key, k->iv, 0 /* decrypt */) ) {
    
        PRINT_INFO0 (INFO_LEVEL_WARNING,
                "Unable to initialize decryption context!\n"
                );
        ERR_print_errors_fp(stderr);
    }

    int tmpLen;
    *decryptedDataLen = 0;

    if ( ! EVP_CipherUpdate (&decryptCtx, decryptedData, &tmpLen, dataToDecrypt, dataToDecryptLen)) {

        PRINT_INFO0 (INFO_LEVEL_WARNING,
                "Unable to decrypt!\n"
                );
        ERR_print_errors_fp(stderr);
    }
    
    *decryptedDataLen += tmpLen;
    
    EVP_CipherFinal_ex(&decryptCtx, (decryptedData + *decryptedDataLen), &tmpLen); 

    *decryptedDataLen += tmpLen;
    
    EVP_CIPHER_CTX_cleanup (&decryptCtx);
        
} 

/**
  * Encrypt the data with the specified symmetric key.
  */
void EncryptDecryptBlackBox::encrypt (
            unsigned char* dataToEncrypt, uint16_t dataToEncryptLen,
            unsigned char* encryptedData, uint16_t *encryptedDataLen,
            SymKey *k
        ) {


    EVP_CIPHER_CTX encryptCtx;

    EVP_CIPHER_CTX_init (&encryptCtx);
    
    int tmpLen;
    *encryptedDataLen = 0;
    
    if ( ! EVP_CipherInit_ex (&encryptCtx, ENC_ALGO,  NULL, k->key, k->iv, 1 /* encrypt */) ) {

        PRINT_INFO0 (INFO_LEVEL_WARNING,
                "Unable to initialize encryption context!\n"
                );
        ERR_print_errors_fp(stderr);
    }
 
    if ( ! EVP_CipherUpdate (&encryptCtx, encryptedData, &tmpLen, dataToEncrypt, dataToEncryptLen)) {

        PRINT_INFO0 (INFO_LEVEL_WARNING,
                "Unable to encrypt!\n"
                );
        ERR_print_errors_fp(stderr);
    }
    *encryptedDataLen += tmpLen;

    // Write any remaining parts of the encrypted data to the out buffer
    EVP_CipherFinal_ex(&encryptCtx, (encryptedData + *encryptedDataLen), &tmpLen); 

    *encryptedDataLen += tmpLen;
    
    EVP_CIPHER_CTX_cleanup (&encryptCtx);
}

