/**
 *  \file SecurityManager.cpp
 *  \brief Functions for loading and retrieving certificates, keys associated with the proxy
 *  instantiation.
*/

#include "SecurityManager.h"

#include "cryptoFunctions.h"
#include "../../utils/StringTokenizer.h"
#include "../../common.h"
#include "../../Params.h"
#include "../../debug.h"

#include <string.h>


/**
  * Initialize the security manager at the specified security level.
  * @param securityLevel    The current security levelA
  * @param certFileStr List of space separated certFileName:privKeyFileName to be loaded.
  * @param caCertFile   The file containing the certificate of the CA to be used.
  */
SecurityManager::SecurityManager (int securityLevel, char *certFileStr,  char *caCertFile) {

    PRINT_INFO0 ( INFO_LEVEL_MINIMAL,
                    "Initializing Security Manager ...\n"
                );

    cryptoInit();

    this->securityLevel = securityLevel;

    switch (this->securityLevel) {
       
        case SECURITY_LEVEL_0:
            authenticationRequiredFlag = false;
            encryptionRequiredFlag = false;
        break;

        case SECURITY_LEVEL_1:
            authenticationRequiredFlag = true;
            encryptionRequiredFlag = false;
        break;

        case SECURITY_LEVEL_2:
            authenticationRequiredFlag = true;
            encryptionRequiredFlag = true;
        break;

    }

    // Go through each file name and try to load the certificate

    string tmpCertFileStr = certFileStr;
	StringTokenizer st1 (tmpCertFileStr, string (CERT_FILENAME_DELIM));

	while (st1.hasMoreTokens()) {
       
		string currToken = st1.nextToken();

		StringTokenizer st2 (currToken, string (CERT_PRIVKEY_DELIM));
        char* tmpCertFileName  = strdup (st2.nextToken().c_str());
		char* tmpPrivKeyFileName = strdup(st2.nextToken().c_str());
        //TODO Change the format of XML file or add error check here
        
        loadCertificate (tmpCertFileName, tmpPrivKeyFileName);

        free(tmpCertFileName);
        free(tmpPrivKeyFileName);
       
    }

    //printCertificateList();


	caPubKey = NULL;
   // Read the CA's public key from disk
	if (caCertFile != NULL && strcmp (caCertFile, "") != 0) {
		caPubKey = initPublicKey (caCertFile);
		if (caPubKey == NULL) {
			 PRINT_INFO0 (
				        INFO_LEVEL_FATAL_ERROR,
					    "Error while reading CA public key from CA certificate file.\n"
					);
			caPubKey = NULL;
		    return;
		}
	} else {
		caPubKey = NULL;
	}

 
}

/**
  * Loads the certificate from the specified file.
  * This file also loads the private key associated with
  * the certificate.
  */
void SecurityManager::loadCertificate(const char *certFileName, const char *privKeyFileName) {

    PRINT_INFO1 (
                INFO_LEVEL_VERBOSE,
                "Loading certificate from file %s.\n",
                certFileName
            );
    
    Certificate* myCert = (Certificate*) malloc (sizeof(Certificate));
    blankCertificate(myCert);

    int retVal = loadCertificateFromFile(certFileName, myCert);
    
    // Add the certificate to the certificate hash table if it loaded correctly.
    if (retVal == 1) {
        if (getCertificate(myCert->name) == NULL) {
            //If the certificate does not already exist.

            // Load the corresponding private key and put it into the private key hash table keyed by the name.
            EVP_PKEY* tmpKey = loadPrivateKeyFromDisk(privKeyFileName);
            if (tmpKey != NULL) {
                addCertificate (myCert);
                privKeyTable.insert(pair<string, RSA*>(myCert->name, EVP_PKEY_get1_RSA(tmpKey)));
                privEVPKeyTable.insert(pair<string, EVP_PKEY*>(myCert->name, tmpKey));
            } else {
                PRINT_INFO1 (INFO_LEVEL_WARNING,
                            "Unable to load private key for %s. Certificate not loaded.\n",
                            myCert->name
                        );
                return;
            }

        } else {
            // Certificate with the same name already exists.
            // Don't add
            PRINT_INFO1 (
                        INFO_LEVEL_WARNING,
                        "Duplicate certificate for name %s specified.\n"
                        "I will not load it.\n"
                        ,
                        myCert->name
                    );
            
        }
    }
}

/** Adds the specified certificate to the certificate hash table.
  */
void SecurityManager::addCertificate (Certificate *cert) {
    
    string name (cert->name);
    certificateTable.insert(pair<string, Certificate*>(name, cert));
}

/**
  * Retrieves the certificate with the specified name.
  * Returns NULL if not found.
  */
Certificate* SecurityManager::getCertificate (const char *name) {
    string tmpName (name);

    PRINT_DEBUG1 (
            DEBUG_LEVEL_VERBOSE,
            "Requesting certificate for name %s.\n",
            name
        );

    map<string, Certificate*>::iterator itr = certificateTable.find(tmpName);
    
    if (itr != certificateTable.end()) {
        return (Certificate*)itr->second;
        
    } else {
            
        return NULL;
    }
}

/**
  * Retrieves the private key in (EVP format) associated with the specified name.
  * Returns NULL if not found.
  */
EVP_PKEY* SecurityManager::getPrivateEVPKey (char *name) {
    string tmpName (name);

    PRINT_DEBUG1 (
            DEBUG_LEVEL_VERBOSE,
            "Requesting private EVP key for name %s.\n",
            name
        );

    map<string, EVP_PKEY*>::iterator itr = privEVPKeyTable.find(tmpName);
    
    if (itr != privEVPKeyTable.end()) {
        return (EVP_PKEY*)itr->second;
        
    } else {
            
        return NULL;
    }
}
/**
  * Retrieves the private key associated with the specified name.
  * Returns NULL if not found.
  */
RSA* SecurityManager::getPrivateKey (const char *name) {
    string tmpName (name);

    PRINT_DEBUG1 (
            DEBUG_LEVEL_VERBOSE,
            "Requesting private key for name %s.\n",
            name
        );

    map<string, RSA*>::iterator itr = privKeyTable.find(tmpName);
    
    if (itr != privKeyTable.end()) {
        return (RSA*)itr->second;
        
    } else {
            
        return NULL;
    }
}

/** Prints the list of currently loaded certificates */
void SecurityManager::printCertificateList() {
    
    map<string, Certificate*>::iterator itr = certificateTable.begin();

    PRINT_INFO0(INFO_LEVEL_VERBOSE, "\nSecurityManager: List of currently loaded certificates:\n");
    while (itr != certificateTable.end()) {
        printf_certificate((Certificate *)itr->second);
        printf ("\n");
        itr ++;
    }

    PRINT_INFO0(INFO_LEVEL_VERBOSE, "\n\n");
}

/**
  * Verify if the specificed certificate actually belongs to the name.
  */
bool SecurityManager::checkCertificate(const char *name, Certificate *cert) {
    
    if (caPubKey == NULL) {
        PRINT_DEBUG1 (
                DEBUG_LEVEL_MINIMAL,
                "CA public key not available. Cannot verify certificate for name %s.\n",
                name
            );
        return false;
    }
   
    if (verifyCertificate(cert, caPubKey)) {
        PRINT_DEBUG0 (
                DEBUG_LEVEL_MINIMAL,
                "CA's signature on the certificate Verified Ok.\n"
            );
    } else {
        PRINT_DEBUG0 (
                DEBUG_LEVEL_MINIMAL,
                "CA's signature on the certificate Verification FAILED.\n"
            );
        return false;
    }

   if (strcmp(name, cert->name) == 0) {
       PRINT_DEBUG1 (
               DEBUG_LEVEL_MINIMAL,
               "The requested name (%s) matches the name in the certificate.\n",
               name
            );
 
        return true;
   } else {
        // The name in the certificate does not match with the requested name.
       PRINT_DEBUG2 (
               DEBUG_LEVEL_MINIMAL,
               "The requested name (%s) does not match the name in the certificate (%s).\n",
               name, cert->name
            );
               
       return false;
   }

}

/**
  * Add the certificate and private key specified by cerStr.
  * @param certStr This string is of the form certFileName::privKeyFileName
  */
void SecurityManager::loadCertificate (const char *certStr) {

    StringTokenizer st = StringTokenizer (certStr, CERT_PRIVKEY_DELIM);
    
    char tmpCertFileName[MAX_FILENAME_LEN];
    strcpy ( tmpCertFileName, st.nextToken().c_str());
    
    char tmpPrivKeyFileName[MAX_FILENAME_LEN];
    strcpy ( tmpPrivKeyFileName, st.nextToken().c_str());
   
    //TODO Change the format of XML file or add error check here
    loadCertificate (tmpCertFileName, tmpPrivKeyFileName);

}


/**
  * Should the middle box with the specified name be allowed.
  */
bool SecurityManager::allowMiddleBox (char *middleBoxName) {

    //TODO  read in a list from configuration file.
    return true;
}
