/** 
 * \file SecurityPacketBlocks.cpp
 * \brief Functions to manipulate security packets.  Currently none.
*/
#include "SecurityPacketBlocks.h"
#include "certificate.h"


/** Converts all the fields in the specified SymKeyPair to host byte order.
 */
void ntoh (SymKeyPair *keyPair) {
    
    keyPair->encryptedNonceLen = ntohs (keyPair->encryptedNonceLen);
    keyPair->encryptedLeftSymKeyLen = ntohs (keyPair->encryptedLeftSymKeyLen);
    keyPair->encryptedLeftSymIVLen = ntohs (keyPair->encryptedLeftSymIVLen);
    keyPair->encryptedRightSymKeyLen = ntohs (keyPair->encryptedRightSymKeyLen);
    keyPair->encryptedRightSymIVLen = ntohs (keyPair->encryptedRightSymIVLen);
}

/** Converts all the fields in the specified SymKeyPair to network byte order.
 */
void hton (SymKeyPair *keyPair) {
    
    keyPair->encryptedNonceLen = htons (keyPair->encryptedNonceLen);
    keyPair->encryptedLeftSymKeyLen = htons (keyPair->encryptedLeftSymKeyLen);
    keyPair->encryptedLeftSymIVLen = htons (keyPair->encryptedLeftSymIVLen);
    keyPair->encryptedRightSymKeyLen = htons (keyPair->encryptedRightSymKeyLen);
    keyPair->encryptedRightSymIVLen = htons (keyPair->encryptedRightSymIVLen);
}



/**
 * Converts the security related fields into 
 * network byte order.  Does not touch the oci headers.
 */
void hton (SecurityBlock* blk) {

    switch (blk->msgCode) {
                  
        case SECURITY_BLOCK_SYMKEY:
        case SECURITY_BLOCK_SYMKEY_OCD:
	    {
		    SymKeyBlock* tmpBlk = (SymKeyBlock *) blk;
		    for (int hopIndex = 0; hopIndex < tmpBlk->numHops; hopIndex++) {
			    hton (&(tmpBlk->symKeyPairs[hopIndex]));
		    }

	    }	
        break;
                       
        case SECURITY_BLOCK_SYMKEY_REPLY:
        {
		    SymKeyReplyBlock* tmpBlk = (SymKeyReplyBlock*) blk;
		    tmpBlk->encryptedProofStringLen = htons (tmpBlk->encryptedProofStringLen);
        }
        break;

        case SECURITY_BLOCK_CERTIFICATE_REQUEST:
        {
            CertificateRequestBlock* tmpBlk = (CertificateRequestBlock *) blk;
            for (int i = 0; i < tmpBlk->numOCDCertificates; i++) {
                hton (&(tmpBlk->ocdCertificates[i].cert));
            }
        }
        break;
       
        case SECURITY_BLOCK_CERTIFICATE_REPLY:
        {
            CertificateReplyBlock* tmpBlk = (CertificateReplyBlock *) blk;
            hton (&(tmpBlk->certificate));
        }
        break;

        case SECURITY_BLOCK_CERTIFICATE_UNAVAILABLE:
        break;

        default:
            PRINT_INFO1 (
                        INFO_LEVEL_WARNING,
                        "hton: Unknown SECURITY block code %d\n",
                        blk->msgCode
                    );
        break;
    }
}

/**
 * Converts the security related fields into 
 * host byte order.  Does not touch the oci headers.
 */
void ntoh (SecurityBlock* blk) {

    switch (blk->msgCode) {
                  
        case SECURITY_BLOCK_SYMKEY:
        case SECURITY_BLOCK_SYMKEY_OCD:
	    {
		    SymKeyBlock* tmpBlk = (SymKeyBlock *) blk;
		    for (int hopIndex = 0; hopIndex < tmpBlk->numHops; hopIndex++) {
			    ntoh (&(tmpBlk->symKeyPairs[hopIndex]));
		    }

	    }	
        break;
                       
        case SECURITY_BLOCK_SYMKEY_REPLY:
        {
		    SymKeyReplyBlock* tmpBlk = (SymKeyReplyBlock*) blk;
		    tmpBlk->encryptedProofStringLen = ntohs (tmpBlk->encryptedProofStringLen);
        }
        break;

        case SECURITY_BLOCK_CERTIFICATE_REQUEST:
        {
            CertificateRequestBlock* tmpBlk = (CertificateRequestBlock *) blk;
            for (int i = 0; i < tmpBlk->numOCDCertificates; i++) {
                ntoh (&(tmpBlk->ocdCertificates[i].cert));
            }
        }
        break;
       
        case SECURITY_BLOCK_CERTIFICATE_REPLY:
        {
            CertificateReplyBlock* tmpBlk = (CertificateReplyBlock *) blk;
            ntoh (&(tmpBlk->certificate));
        }
        break;

        case SECURITY_BLOCK_CERTIFICATE_UNAVAILABLE:
        break;

        default:
            PRINT_INFO1 (
                        INFO_LEVEL_WARNING,
                        "ntoh: Unknown SECURITY block code %d\n",
                        blk->msgCode
                    );
        break;



    }
}
