/**
 *  \file SecurityPacketBlocks.h
 *  \brief Packet formats of the various messages used in our custom SSL protocol.
*/

#ifndef SECURITY_PACKET_BLOCKS_H
#define SECURITY_PACKET_BLOCKS_H

#include "../../oci/OCIPacketBlock.h"
#include "certificate.h"

#include "nonce.h"

/** Request a certificate from the remote end point */
#define SECURITY_BLOCK_CERTIFICATE_REQUEST  1

/** Reply with the requested certificate */
#define SECURITY_BLOCK_CERTIFICATE_REPLY    2

/** Indicate that a requested certificate was not available */
#define SECURITY_BLOCK_CERTIFICATE_UNAVAILABLE  3

/** This block contains the encrypted symkey, nonce etc. 
  * sent from the sender to the receiver.
  */
#define SECURITY_BLOCK_SYMKEY   4

/** This block contains the reply message sent for the SYMKEY message.
  */
#define SECURITY_BLOCK_SYMKEY_REPLY 5

/**
  * This message code is used for SymKeys sent by the Receiver
  * to the OCD-level-receiver initiated middleboxes.
  */
#define SECURITY_BLOCK_SYMKEY_OCD   6


#define SECURITY_TYPE_AUTH_RECV_ONLY    1
#define SECURITY_TYPE_SSL               2

/**
  * The top level security block
  */
typedef struct {

    /** The main proxy packet block header */
    OCIPacketBlock mainHeader;    

    /** AUTH_RECV_ONLY, SSL ... */
    //int securityType;
    
    /** What type of security message is contained in this security block */
    int8_t msgCode;
    
} SecurityBlock;

/**
  * This is a dummy structure to hold a certificate name.
  * Its sole purpose is to easily allow unsized arrays in the 
  * CertificateRequest block.
  */
typedef struct {

    char name[CERT_MAX_NAME_LEN];
    
} CertificateName;

/**
  * A structure to group together a Certificate and its associated nonce.
  */
typedef struct {

    Certificate cert;
    unsigned char nonce [NONCE_BIT_LEN/8];

} CertNonce;


// Packet types for SECURITY_TYPE_AUTH_RECV_ONLY
/**
  * Request a certificate from the other end point.
  * msgCode = SECURITY_BLOCK_CERTIFICATE_REQUEST
 */
typedef struct {

    SecurityBlock mainHeader; //Note this is SecurityBlock and NOT OCIPacketBlock.

    /** The number of certificates in the list required by initiator. */
    uint8_t numCertificates;
   
    /** The current hop in the sender initiated certificate request list.
      * This is set to 0 when the initiator sends out the certificate request.
      * Each node on the path knows what certificate from the nameList is being
      * requested at each hop based on the value of this field.
      * It is incremented as the message is relayed forward.
      */
    uint8_t currHop;
    
    /** The names of the certificates requested by the initiator. 
      * If no certificate is requested for a particular hop, the
      * entry in nameList corresponding to that hop is the blank string.
      */
    CertificateName nameList [MAX_NAMES_IN_CERT_REQ];

    /** The number of OCD Certificates (and associated nonces) tagged along with this
      * certificate request.
      */
    uint8_t numOCDCertificates;

    CertNonce ocdCertificates[0];

} CertificateRequestBlock;

/**
  * Reply with the requested certificate
  * msgCode = SECURITY_BLOCK_CERTIFICATE_REPLY
 */
typedef struct {

    SecurityBlock mainHeader; //Note this is SecurityBlock and NOT OCIPacketBlock.
   
    /** Nonce sent by receiver to sender */
    unsigned char nonce [NONCE_BIT_LEN/8];
    
    Certificate certificate;

} CertificateReplyBlock;

/**
  * This represents the symmetric key pair meant for a particular hop.
  */
typedef struct {
    
    /** The actual length of the encrypted nonce. */
    uint16_t  encryptedNonceLen;

    /** The encrypted nonce */
    unsigned char encryptedNonce [ NONCE_BIT_LEN / 8 + MAX_CIPHER_BLOCK_LENGTH];

    /* NOTE: For explanation of left and right sides, please look at the documentation in CryptoState.cpp
     */

    /** The length of the encrypted symmmetric key to be used on the left side of the connection */
    uint16_t encryptedLeftSymKeyLen;
    
    /** The symmetric key encrypted with public key of the other end to be used on the left side of the connection 
     */
    unsigned char encryptedLeftSymKey [CERT_RSA_KEY_SIZE/8];

    /** The length of the encrypted sym key IV to be used at the left side of the connection.*/
    uint16_t encryptedLeftSymIVLen;

    /** The symmetric key IV encrypted with the public key of the other end to be used at the left side of the connection.*/
    unsigned char encryptedLeftSymIV [CERT_RSA_KEY_SIZE/8];

    /** The length of the encrypted symmmetric key to be used on the right side of the connection */
    uint16_t encryptedRightSymKeyLen;
    
    /** The symmetric key encrypted with public key of the other end to be used on the right side of the connection 
     */
    unsigned char encryptedRightSymKey [CERT_RSA_KEY_SIZE/8];

    /** The length of the encrypted sym key IV to be used at the right side of the connection.*/
    uint16_t encryptedRightSymIVLen;

    /** The symmetric key IV encrypted with the public key of the other end to be used at the right side of the connection.*/
    unsigned char encryptedRightSymIV [CERT_RSA_KEY_SIZE/8];

} SymKeyPair;

/**
  * The sender sends the encrypted symmetric key and nonce
  * to the receiver via this block.
  * msgCode = SECURITY_BLOCK_SYMKEY
  */
typedef struct {

    SecurityBlock mainHeader; //Note this is SecurityBlock and NOT OCIPacketBlock.

    /** The current hop for which this packet is meant. */
    uint8_t currHop;

    /** The total number of hops */
    uint8_t numHops;
    
    /** The pairs of symmetric keys and encrypted nonces for each hop */
    SymKeyPair symKeyPairs[0];
    
} SymKeyBlock;



/** The maximum length of the mutually agreed upon proof string */
#define PROOF_STRING_LENGTH 50

#define PROOF_STRING "Hello Sender"

/**
  * The receiver replies with this block after receiving
  * the SymKeyBlock from the sender and verifying that
  * the nonce echoed back by the sender is correct.
  * msgCode = SECURITY_BLOCK_SYMKEY_REPLY
  */
typedef struct {

    SecurityBlock mainHeader; //Note this is SecurityBlock and NOT OCIPacketBlock.
   
    /** The sender and receiver apriori on a mutual string to serve as
      * proof that the receiver was able to decrypt the symmetric key sent by
      * the sender, i.e. the receiver possessed the private key of the
      * certificate sent to the sender.
      * This field of the block contains that mutually agreed string encrypted with
      * the symmetric key sent by the sender.  
      * This packet is sent by the receiver.  The sender will verify that 
      * it correctly decrypts to the mutually agreed upon string.
      * QQQ: Will it be more secure to have a nonce proposed by the sender
      * rather than have a fixed string here?
      */
    char encryptedProofString[CERT_RSA_KEY_SIZE/8];
    
    uint16_t  encryptedProofStringLen;

    /** The name of the certificate for which this sym key reply is being sent.
      * This name is needed for retrieving the correct sym key for verficiation of
      * the proof string at the initiator.
      */
    char name [CERT_MAX_NAME_LEN];
    
} SymKeyReplyBlock;


void ntoh (SecurityBlock* blk);
void hton (SecurityBlock* blk);
void ntoh (SymKeyPair *keyPair);
void hton (SymKeyPair* keyPair);


#endif
