/**
 *  \file certificate.c
 *  \brief This file contains definitions of functions used to manipulate 
 *  certificates issued by the CA.
*/

#include "certificate.h"
#include "cryptoFunctions.h"

#include "../../debug.h"
#include <string.h>
#include <stdio.h>

/**
  * Prints a text representation of the specificed certificate
  * into the std output.
  */
void printf_certificate ( Certificate* cert) {
    
    printf(     "Name\t\t: %s\n"
                "Real Name\t: %s\n"
                "Email\t\t: %s\n"
                , 
                cert->name,
                cert->realName,
                cert->emailAddr
           );
    
    //RSA_print_fp (stdout, cert->publicKey, 0); 
}

/**
  * Packs the data of the certificate which is to be signed
  * into the specified buffer buf.
  * The length of the buffer is placed in bufLen.
  */
void getDataToSign(Certificate *cert, char *buf, int* bufLen) {
    
    int offset = 0;
     
    memcpy(buf + offset, cert->name, CERT_MAX_NAME_LEN);
    
     offset += CERT_MAX_NAME_LEN;
    memcpy(buf + offset, cert->realName, CERT_MAX_REAL_NAME_LEN);
    
    offset += CERT_MAX_REAL_NAME_LEN;
    memcpy(buf + offset, cert->emailAddr, CERT_MAX_EMAIL_LEN);
    
    offset += CERT_MAX_EMAIL_LEN;
    memcpy(buf + offset, cert->publicKeyBuf, cert->publicKeyLen);
  
    //printf ("\nIn getDataToSign publicKeyLen = %d\n", cert->publicKeyLen);

    *bufLen = offset + cert->publicKeyLen;

}

/**
  * Write the certificate to the specified file in binary form.
  */
void writeCertificateToFile (char *fileName, Certificate *cert) {
    
    FILE *fp = fopen (fileName, "wb");
	int numWritten;

    // Store the certificate in network byte order for portability of the certificate file across different platforms
    hton (cert);
    
    if (fp == NULL) {
        PRINT_INFO1 (
                    INFO_LEVEL_FATAL_ERROR,
                    "Unable to open file %s for writing.\n",
                    fileName
                );
        perror ("writeCertificateToFile");
        return;
    }

    numWritten = fwrite (cert, sizeof(Certificate), 1, fp);
    if (numWritten == sizeof(Certificate)) {
        PRINT_INFO1 (
                INFO_LEVEL_WARNING,
                "Did not completely write the certificate to file %s.\n",
                fileName
            );
    }
    fclose (fp);
}

/**
  * Returns 1 on success.
  * Returns 0 on error.
  */
int  loadCertificateFromFile (const char *fileName, Certificate *cert) {
   
    FILE *fp = fopen (fileName, "rb");
	int numRead;

    if (fp == NULL) {
        PRINT_INFO1 (
                    INFO_LEVEL_FATAL_ERROR,
                    "Unable to open file %s for reading.\n",
                    fileName
                );
        perror ("readCertificateFromFile");
        return 0;
    }

    numRead = fread (cert, sizeof(Certificate), 1, fp);
    if (numRead < 1) {
        PRINT_INFO1 (
                INFO_LEVEL_WARNING,
                "Did not completely read the certificate from file %s.\n",
                fileName
            );
        fclose (fp);
        return 0;
    }

    ntoh (cert);
    fclose (fp);
    return 1;
 

}

/**
  * Construct a new certificate with the specified details.
  * Returns NULL on error.
  */
Certificate *constructNewCertificate (char *name, char* realName, char* emailAddr, RSA* publicKey) {

    Certificate *newCert = (Certificate*) malloc(sizeof(Certificate));
    blankCertificate (newCert);

    if (newCert == NULL) {
        PRINT_INFO1 (INFO_LEVEL_FATAL_ERROR,
                    "Unable to allocate new certificate for name %s\n", name);
        return NULL;
    }

    strncpy(newCert->name, name, CERT_MAX_NAME_LEN);
    strncpy(newCert->realName, realName, CERT_MAX_REAL_NAME_LEN);
    strncpy(newCert->emailAddr, emailAddr, CERT_MAX_EMAIL_LEN);

    newCert->sign[0] = '\0';
    newCert->publicKeyLen = savePublicKeyToBuffer(publicKey, newCert->publicKeyBuf, CERT_MAX_PUBLIC_KEY_LEN); 

    return newCert;
}

/**
  * Returns 1 it can be verified that the CA signed this
  * certificate with its private key.
  * Else returns 0
  */
int verifyCertificate (Certificate *cert, EVP_PKEY *caPubKey) {

    char signedData[MAX_DATA_TO_SIGN_SIZE];
    int signedDataLen;

    if (    cert->publicKeyLen > CERT_MAX_PUBLIC_KEY_LEN ||
            cert->signLen > CERT_MAX_SIGN_LEN
        ) {
        return false;
    }

    getDataToSign (cert, signedData, &signedDataLen); 
    return verifySignature (signedData, signedDataLen, cert->sign, cert->signLen, caPubKey);
}

/**
  * Retrieves the public key stored in the certificate.
  */
EVP_PKEY* getPublicKeyFromCertificate(Certificate *cert) {

    RSA* tmpPubKey = loadPublicKeyFromBuffer(cert->publicKeyBuf, cert->publicKeyLen);
    
    EVP_PKEY* pkey = EVP_PKEY_new();
    if (!EVP_PKEY_assign_RSA(pkey, tmpPubKey)) {
        PRINT_INFO0 (
                INFO_LEVEL_WARNING,
                "Unable to assign RSA public key to EVP key.\n"
                );
        return NULL;
    }
    
    return pkey;

}

/**
  * Fill the entire memory space of the specified 
  * certificate with 0s
  */
void blankCertificate (Certificate *cert) {
	memset(cert, 0, sizeof (Certificate));
}


/**
  * Converts all the fields in the specified Certificate to host byte order.
  */
void ntoh (Certificate* cert) {
    cert->publicKeyLen = ntohs (cert->publicKeyLen);
    cert->signLen = ntohs (cert->signLen);
}

/**
  * Converts all the fields in the specified Certificate to network byte order.
  */
void hton (Certificate* cert) {
    cert->publicKeyLen = htons (cert->publicKeyLen);
    cert->signLen = htons (cert->signLen);
}

void ntoh (CertPasswordMsg *passwordMsg) {
    // do nothing
}
void hton (CertPasswordMsg *passwordMsg) {
    // do nothing
}

