/**
 *  \file certificate.h
 *  \brief Defines the structure of a certificate and the various certificate request/response codes
 *  used by the nameAllocator.
*/

#ifndef I3_CERTIFICATE_H
#define I3_CERTIFICATE_H

#include "OCALA_Export.h"

#include "common.h"
#include "securityCommon.h"
#include "cryptoHeaders.h"

typedef struct {
  
    char name[CERT_MAX_NAME_LEN];
    
    uint8_t publicKeyBuf[CERT_MAX_PUBLIC_KEY_LEN];
    uint16_t publicKeyLen;

    uint8_t sign[CERT_MAX_SIGN_LEN];
    uint16_t signLen;

    char realName[CERT_MAX_REAL_NAME_LEN];
    char emailAddr[CERT_MAX_EMAIL_LEN];

} Certificate;

/**
  * This is used to construct the password reply packet.
  */
typedef struct {
    char name[CERT_MAX_NAME_LEN];
    char password[CERT_EMAIL_NONCE_LEN];

} CertPasswordMsg;

//Certificate Request Codes 

/** Code used to make a new name request */
#define REQCODE_NEW_REQUEST 1

/** Code used while supplying the emailed password for 
  * a prior request.
  */
#define REQCODE_NEW_REQUEST_CONFIRM 2


/** Code to indicate the requested name is not available.
  */
#define REQCODE_NAME_UNAVAILABLE 3

/** Code to indicate the requested name is available.
  */
#define REQCODE_NAME_AVAILABLE 4

/** The requested name had been previously allocated to the same email. */
#define REQCODE_NAME_REREQUEST 5 


/** Code to indicate that the name was previously verified */
#define REQCODE_NAME_VERIFY_DUPLICATE 6

/** Code to indicate that the certificate has successfully been signed. */
#define REQCODE_SIGNED  7

/** Code to indicate that some error occurred */
#define REQCODE_ERROR 8

/** Code to indicate that the password supplied while confirming the name is incorrect. */
#define REQCODE_INVALID_PASSWORD 9

/** The name you are trying to confirm was not found */
#define REQCODE_UNKNOWN_NAME 10

/** The current name registration policy forbids the registration of this name */
#define REQCODE_POLICY_DISALLOW 11

/**
  * Request to retrieve a certificate.
  */
#define REQCODE_RETRIEVE_CERTIFICATE 12
#define REQCODE_RETRIEVE_CERTIFICATE_SUCCESS 13
#define REQCODE_RETRIEVE_CERTIFICATE_FAILED 14

OCALA_LIB_EXPORT_IMPORT void printf_certificate (Certificate *cert);
OCALA_LIB_EXPORT_IMPORT Certificate* constructNewCertificate (char *name, char* realName, char* emailAddr, RSA* publicKey); 

OCALA_LIB_EXPORT_IMPORT void writeCertificateToFile (char *fileName, Certificate *cert);
OCALA_LIB_EXPORT_IMPORT int loadCertificateFromFile (const char *fileName, Certificate *cert);

OCALA_LIB_EXPORT_IMPORT int verifyCertificate (Certificate *cert, EVP_PKEY *caPubKey);

OCALA_LIB_EXPORT_IMPORT void getDataToSign(Certificate *cert, char *data, int *dataLen);

OCALA_LIB_EXPORT_IMPORT EVP_PKEY* getPublicKeyFromCertificate(Certificate *cert);

OCALA_LIB_EXPORT_IMPORT void blankCertificate (Certificate *cert);


OCALA_LIB_EXPORT_IMPORT void ntoh (Certificate* cert);
OCALA_LIB_EXPORT_IMPORT void hton (Certificate* cert);
OCALA_LIB_EXPORT_IMPORT void ntoh (CertPasswordMsg* passwordMsg);
OCALA_LIB_EXPORT_IMPORT void hton (CertPasswordMsg* passwordMsg);
#endif
