/**
 *  \file cryptoFunctions.c
 *  \brief  Generic cryptographic functions to sign a buffer, initialize and load keys.
 */

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "OCALA_Export.h"
#include "../../debug.h"

#include "common.h"
#include "securityCommon.h"
#include "cryptoHeaders.h"
#include "nonce.h"

/**
  * Sign the specificed data with the given private key
  * @return
  *     sigBuf - will contain the signature
  *     sigLen - will contain the length of the signature
  */
OCALA_LIB_EXPORT_IMPORT void sign (char *data, int dataLen, unsigned char *sigBuf, uint16_t *sigLen16, EVP_PKEY *privateKey) {

	int err;
	//printf ("Max signature size = %d\n", EVP_PKEY_size(privateKey));


    EVP_MD_CTX     md_ctx;
    EVP_SignInit   (&md_ctx, EVP_sha1());
    EVP_SignUpdate (&md_ctx, data, dataLen);

    unsigned sigLen;
    sigLen = sizeof(sigBuf);
    err = EVP_SignFinal (&md_ctx, sigBuf, &sigLen, privateKey);

    //printf ("in sign(), signlen = %d\n", sigLen);
    *sigLen16 = (uint16_t) sigLen;
    if (err != 1) {
	    ERR_print_errors_fp(stderr);
	    return;
    }

}


/**
  * Returns 1 if the signature is correct.
  * i.e. it was signed using the private key of the specified public key.
  * @param data The data whose signature is to be verified
  * @param sigBuf The signature to be verified.
  * @param sigLen THe length of the signature
  * @param publicKey The public Key used to verify the signature.
  */
OCALA_LIB_EXPORT_IMPORT int verifySignature(char* data, int dataLen, unsigned char* sigBuf, unsigned int sigLen, EVP_PKEY* publicKey) {
  
    int err;
    EVP_MD_CTX     md_ctx;
    
    EVP_VerifyInit   (&md_ctx, EVP_sha1());
    EVP_VerifyUpdate (&md_ctx, data, dataLen);
    err = EVP_VerifyFinal (&md_ctx, sigBuf, sigLen, publicKey);

    if (err != 1) {
        return 0;
    } else {
        return 1;
    }
}

OCALA_LIB_EXPORT_IMPORT void cryptoInit() {

    static bool cryptoInitFlag = false;

    if (cryptoInitFlag) {
        // We have already initialized the crypto
        return;
    }
#ifdef _WIN32
	srand (time(NULL));
#else
    srandom(time(NULL));
    srand (time(NULL));
#endif
    SSL_load_error_strings();
    
    cryptoInitFlag = true;

}

/**
  * Load the public key of the CA from the file containing the CA's certificate.
  */
OCALA_LIB_EXPORT_IMPORT EVP_PKEY* initPublicKey(char *certfile) {

	X509* x509 = NULL;
	EVP_PKEY* pkey = NULL;
    FILE* fp = fopen (certfile, "r");

    if (fp == NULL) {
        printf("Unable to read certificate from file %s.\n", certfile);
        return NULL;
    }
    
    x509 = PEM_read_X509(fp, NULL, NULL, NULL);
    fclose (fp);

    if (x509 == NULL) {
	    ERR_print_errors_fp (stderr);
	    return NULL;
    }
  
    /* Get public key - eay */
    pkey = X509_get_pubkey(x509);
    if (pkey == NULL) {
	    ERR_print_errors_fp (stderr);
        X509_free (x509);
	    return NULL;
    }

    X509_free (x509);
    return pkey;

}


/** Loads the private key from the specified file.
  */
OCALA_LIB_EXPORT_IMPORT EVP_PKEY* initPrivateKey(char *pkeyFileName) {
  
    FILE* fp = fopen (pkeyFileName, "r");
	EVP_PKEY* pkey = NULL;

    if (fp == NULL) {
        printf("Error while reading private key file %s.\n", pkeyFileName);
        return NULL;
    }
  
    pkey = PEM_read_PrivateKey(fp, NULL, NULL, NULL);
    fclose (fp);

    if (pkey == NULL) { 
	    ERR_print_errors_fp (stderr);
	    return NULL;
    }

    return pkey;
}

OCALA_LIB_EXPORT_IMPORT void freeKey(EVP_PKEY* pkey) {

    EVP_PKEY_free (pkey);
}

OCALA_LIB_EXPORT_IMPORT RSA* generateKeyPair() {

    return RSA_generate_key(CERT_RSA_KEY_SIZE, RSA_F4, NULL, NULL);
}

OCALA_LIB_EXPORT_IMPORT RSA* rsa_dup (RSA* key) {

	RSA* newKey=RSA_new();
	memcpy(newKey,key,sizeof(RSA));
	return newKey;
}

/**
  * Gets only the public key from the specified pair.
  */
OCALA_LIB_EXPORT_IMPORT RSA* getPublicKeyFromPair(RSA *keyPair) {
	RSA* newKey = rsa_dup(keyPair);
    newKey->d=NULL;
	newKey->p=NULL;
	newKey->q=NULL;
	newKey->dmp1=NULL;
	newKey->dmq1=NULL;
	newKey->iqmp=NULL;
    
    return newKey;
}

OCALA_LIB_EXPORT_IMPORT int savePublicKeyToBuffer(RSA* keypair, unsigned char *buf, long bufLen) {
    
    unsigned char *p = buf;    //temporary variable is mandatory
    int requiredLen = i2d_RSAPublicKey(keypair, NULL);
	int len;

    //printf("\nRequiredLen for public key = %d\n", requiredLen);
    if (requiredLen > bufLen) {
        PRINT_INFO2( INFO_LEVEL_FATAL_ERROR,
                "savePublicKeyToBuffer: Insufficient buffer space.  Required %d, found %d\n",
                requiredLen,
                bufLen
            );
        EXIT_ON_ERROR;
    }

    len = i2d_RSAPublicKey (keypair, &p);

    printf ("save Public Key to buffer: %d\n", len);
    return len;
}

OCALA_LIB_EXPORT_IMPORT int savePrivateKeyToBuffer(RSA* keypair, unsigned char *buf, long bufLen) {
    
    unsigned char *p = buf;  //temporary variable is mandatory
    int len;
    int requiredLen = i2d_RSAPrivateKey(keypair, NULL);

    if (requiredLen > bufLen) {
        PRINT_INFO2( INFO_LEVEL_FATAL_ERROR,
                "savePrivateKeyToBuffer: Insufficient buffer space. Required = %d, found = %d\n",
                requiredLen,
                bufLen
            );
        EXIT_ON_ERROR;
    }

    len = i2d_RSAPrivateKey (keypair, &p);
    return len;
}

OCALA_LIB_EXPORT_IMPORT RSA* loadPublicKeyFromBuffer(unsigned char *buf, long bufLen) {
    unsigned char *p = buf; //temporary variable is mandatory
    return d2i_RSAPublicKey (NULL, (const unsigned char**) &p, bufLen);
    
}

OCALA_LIB_EXPORT_IMPORT RSA* loadPrivateKeyFromBuffer(unsigned char *buf, long bufLen) {
    unsigned char *p = buf; //temporary variable is mandatory
    return d2i_RSAPrivateKey (NULL, (const unsigned char**) &p, bufLen);
    
}

/** This differs from initPublicKey.
  * The file this function reads is in a custom format.
  */
OCALA_LIB_EXPORT_IMPORT EVP_PKEY* loadPublicKeyFromDisk (const char *fileName) {
    FILE *fp = fopen (fileName, "rb");
	char keyBuf[CERT_MAX_PUBLIC_KEY_LEN];
    int keyLen;
    
	RSA* retRSA = NULL;
	EVP_PKEY* pkey = NULL;

    if (fp == NULL) {
        PRINT_INFO1(
                INFO_LEVEL_FATAL_ERROR,
                "Unable to read public key from %s.\n",
                fileName
            );
        perror("Error");
        return NULL;
    }

    if (fread (&keyLen, sizeof(keyLen), 1 , fp) < 1) {
        PRINT_INFO1 (INFO_LEVEL_WARNING,
                "Unable to read key len from %s.\n",
                fileName
            );
        return NULL;
    }
    
    if (keyLen >= CERT_MAX_PUBLIC_KEY_LEN) {
        PRINT_INFO1 (
                    INFO_LEVEL_WARNING,
                    "Key length of public key too long (%d).\n",
                    keyLen
                );
        return NULL;
    }
    
    if (fread (keyBuf, keyLen, 1, fp) < 1) {
        PRINT_INFO1 (INFO_LEVEL_WARNING,
                    "Unable to read public key from %s.\n",
                    fileName
                );
        return NULL;
    }
    
    retRSA = loadPublicKeyFromBuffer((unsigned char*) keyBuf, keyLen);
    if (retRSA == NULL) {
        PRINT_INFO1 (INFO_LEVEL_WARNING,
                    "Unable to load public key from %s.\n",
                    fileName
                );
        return NULL;
    }

    fclose(fp);

    pkey = EVP_PKEY_new();
    if (!EVP_PKEY_assign_RSA(pkey, retRSA)) {
        PRINT_INFO0 (
                INFO_LEVEL_WARNING,
                "Unable to assign RSA public key to EVP key.\n"
                );
        return NULL;
    }
    return pkey;
    
}

/** This differs from initPrivateKey.
  * The file this function reads is in a custom format.
  */
OCALA_LIB_EXPORT_IMPORT EVP_PKEY* loadPrivateKeyFromDisk (const char *fileName) {
    FILE *fp = fopen (fileName, "rb");
	char keyBuf[CERT_MAX_PRIVATE_KEY_LEN];
    int keyLen;
	RSA* retRSA = NULL;
	EVP_PKEY* pkey = NULL;

    if (fp == NULL) {
        PRINT_INFO1 (
                INFO_LEVEL_FATAL_ERROR,
                "Unable to read private key from %s.\n",
                fileName
            );
        perror("Error");
        return NULL;
    }

    if (fread (&keyLen, sizeof(keyLen), 1 , fp) < 1) {
        PRINT_INFO1 (INFO_LEVEL_WARNING,
                    "Unable to read key len from %s.\n",
                    fileName
                );
        return NULL;
    }
    
    if (keyLen >= CERT_MAX_PRIVATE_KEY_LEN) {
        PRINT_INFO1 (
                    INFO_LEVEL_WARNING,
                    "Key length of private key too long (%d).\n",
                    keyLen
                );
        return NULL;
    }
    
    if (fread (keyBuf, keyLen, 1, fp) < 1) {
        PRINT_INFO1 (INFO_LEVEL_WARNING,
                    "Unable to read key from %s.\n",
                    fileName
                );
        return NULL;
    }
    
    retRSA = loadPrivateKeyFromBuffer((unsigned char*) keyBuf, keyLen);
    if (retRSA == NULL) {
        PRINT_INFO1 (INFO_LEVEL_WARNING,
                    "Unable to load private key from %s.\n",
                    fileName
                );
        return NULL;
    }
    
    pkey = EVP_PKEY_new();
    if (!EVP_PKEY_assign_RSA(pkey, retRSA)) {
        PRINT_INFO0 (
                INFO_LEVEL_WARNING,
                "Unable to assign RSA private key to EVP key.\n"
                );
	    ERR_print_errors_fp(stderr);
        return NULL;
    }

    fclose(fp);

    return pkey;
    
}

/**
  * Generates a random symmetric key and IV and places it in outKey and outIV.
  */
void generateRandomSymKey (unsigned char *outKey, unsigned char* outIV) {

    unsigned char randomStr[NONCE_BIT_LEN/8];
    int randomStrLen = NONCE_BIT_LEN/8;
    // The random nonce function is just reused to generate a good random string.  We have nothing to do with nonce here.
    generateNonce (randomStr);
	EVP_BytesToKey(EVP_des_ede3_cbc(), EVP_md5(), (const unsigned char*) "some random salt", randomStr, randomStrLen, 1, outKey, outIV);
}
