/**
 *  \file createSecureName.cpp
 *  \brief This program creates the private/public key pair and the signed certificate for the specified name.
 *  This program needs access to the master key.  It does not check the database of assigned names.
 *  Use with care.
*/

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "../lib/cryptoHeaders.h"
#include "../lib/cryptoFunctions.h"
#include "../lib/certificate.h"
#include "../lib/securityCommon.h"
#include "../lib/utils.h"

#include "../../debug.h"
#include "../../common.h"


/** The private key used to sign the certificates */
EVP_PKEY* masterPrivateKey;


/**
  * Sign the specified certificate with the master
  * private key of the server.
  * The signature is returned in sigBuf.
  * The length of the signature is put in sigLen
  */
void signCertificate(Certificate *cert, unsigned char sigBuf[MAX_SIGN_BUF_SIZE], uint16_t *sigLen) {
    

   // Bring together the data to be signed.
   char dataToBeSigned[MAX_DATA_TO_SIGN_SIZE];
  
   int dataLen;
   getDataToSign(cert, dataToBeSigned, &dataLen);
   
   //Sign it
   sign(dataToBeSigned, dataLen, sigBuf, sigLen, masterPrivateKey);
   
}

   
      


void saveKeysToDisk (RSA *keypair, char* pubKeyFileName, char* privKeyFileName) {

    unsigned char pubKeyBuf[CERT_MAX_PUBLIC_KEY_LEN];
    int pubKeyLen = savePublicKeyToBuffer(keypair, pubKeyBuf, CERT_MAX_PUBLIC_KEY_LEN);
    
    FILE *fp = fopen (pubKeyFileName, "wb");
    if (fp == NULL) {
        PRINT_INFO1 (
                    INFO_LEVEL_FATAL_ERROR,
                    "Unable to open file %s for writing the public key.\n",
                    pubKeyFileName
                );
        EXIT_ON_ERROR;
    }
    fwrite( &pubKeyLen, sizeof(pubKeyLen), 1, fp);
    fwrite( pubKeyBuf, pubKeyLen, 1, fp);
    fclose(fp);
    
    unsigned char privKeyBuf[CERT_MAX_PRIVATE_KEY_LEN];
    int privKeyLen = savePrivateKeyToBuffer(keypair, privKeyBuf, CERT_MAX_PRIVATE_KEY_LEN);

    fp = fopen (privKeyFileName, "wb");
    if (fp == NULL) {
        PRINT_INFO1 (
                    INFO_LEVEL_FATAL_ERROR,
                    "Unable to open file %s for writing the private key.\n",
                    privKeyFileName
                );
        EXIT_ON_ERROR;
    }
    fwrite( &privKeyLen, sizeof(privKeyLen), 1, fp);
    fwrite( privKeyBuf, privKeyLen, 1, fp);
    fclose(fp);
    
}

/**
  * Do a consistency check for the keys and the certificate.
  * @return true if all tests succeed. Else returns false.
  */
bool testKeys(char* pubKeyFileName, char* privKeyFileName, char* certFileName, char* caCertFileName) {

    bool retVal = true;
    
    PRINT_INFO4 (
                INFO_LEVEL_MINIMAL,
                "Using files:\n"
                "My Public Key File\t: %s\n"
                "My Private Key File\t: %s\n"
                "My Certificate File\t: %s\n"
                "CA Certificate File\t: %s\n"
                "\n"
                ,
                pubKeyFileName,
                privKeyFileName,
                certFileName,
                caCertFileName
            );
    
    // Load the public key from the disk.
    EVP_PKEY* pubKey = loadPublicKeyFromDisk(pubKeyFileName);
    if (pubKey == NULL) {
        PRINT_INFO1 (INFO_LEVEL_FATAL_ERROR,
                    "Unable to load public key from %s.\n",
                    pubKeyFileName
                );
        return false;
    }
    
    // Load the private key from the disk.
    EVP_PKEY* privKey = loadPrivateKeyFromDisk(privKeyFileName);

    if (privKey == NULL) {
        PRINT_INFO1 (INFO_LEVEL_FATAL_ERROR,
                    "Unable to load private key from %s.\n",
                    privKeyFileName
                );
        return false;
    }
    // Load the certificate from the disk
    Certificate myCert;
    blankCertificate(&myCert);

    loadCertificateFromFile(certFileName, &myCert);
    PRINT_INFO0 ( INFO_LEVEL_MINIMAL,
                "Your signed certificate is:\n"
            );
    printf_certificate(&myCert);

    char testMsg[] = "This is a test message to check the signatures\n";

    // Check if my private and public keys are consistent.
    unsigned char sigBuf [MAX_SIGN_BUF_SIZE];
    uint16_t sigLen;
    sign (testMsg, strlen(testMsg), sigBuf, &sigLen, privKey);
    
    if (verifySignature(testMsg, strlen(testMsg), sigBuf, sigLen, pubKey)) {
        PRINT_INFO0(
                INFO_LEVEL_MINIMAL,
                "With Stored Keys: My Signature Verified Ok.\n"
            );
    } else {
        PRINT_INFO0(
                INFO_LEVEL_MINIMAL,
                "With Stored Keys: My Signature Verification FAILED.\n"
            );
        retVal = false;
    }


    // Check if my public key obtained from my certificate is consistent with my private key.
    sign (testMsg, strlen(testMsg), sigBuf, &sigLen, privKey);
   
    EVP_PKEY *pubKeyFromCert = getPublicKeyFromCertificate(&myCert);
    if (pubKeyFromCert == NULL) {
        PRINT_INFO0 (
                    INFO_LEVEL_FATAL_ERROR,
                    "Error while reading public key from certificate file.\n"
                );
        return false;
    }
    
    if (verifySignature(testMsg, strlen(testMsg), sigBuf, sigLen, pubKeyFromCert)) {
        PRINT_INFO0(
                INFO_LEVEL_MINIMAL,
                "With Public Key from Certificate: My Signature Verified Ok.\n"
            );
    } else {
        PRINT_INFO0(
                INFO_LEVEL_MINIMAL,
                "With Public Key from Certificate: My Signature Verification FAILED.\n"
            );
        retVal = false;
    }

    // Verify the CA's signature on your certificate.
    EVP_PKEY* caPubKey =  initPublicKey (caCertFileName);
    if (caPubKey == NULL) {
         PRINT_INFO0 (
                    INFO_LEVEL_FATAL_ERROR,
                    "Error while reading CA public key from CA certificate file.\n"
                );
        return false;
    
    }
    if (verifyCertificate(&myCert, caPubKey)) {
        PRINT_INFO0 (
                INFO_LEVEL_MINIMAL,
                "CA's signature on the certificate Verified Ok.\n"
            );
    } else {
        PRINT_INFO0 (
                INFO_LEVEL_MINIMAL,
                "CA's signature on the certificate Verification FAILED.\n"
            );
        retVal = false;
    }

    return retVal;
}

void usage() {

    PRINT_INFO0 ( INFO_LEVEL_MINIMAL,
                "\n\n"
                "./createSecureName\n"
                "\t-n <NAME TO CREATE>\n"
                "\t-r <REAL NAME>\n"
                "\t-e <EMAIL ADDR>\n"
                "\t-K <MASTER PRIVATE KEY>\n"
                "\t-C <CA's CERTIFICATE>\n"
                "\n\n"
            );
}


int main(int argc, char **argv) {

    opterr = 1; //getopt parameter


    char name[CERT_MAX_NAME_LEN];
    name[0] = '\0';

    char realName[CERT_MAX_REAL_NAME_LEN];
    strcpy (realName, "generated by createSecureName");
    
    char email[CERT_MAX_EMAIL_LEN];
    strcpy (email, "createSecureName@danjo.cs.berkeley.edu");
    
    char masterPrivKeyFileName [MAX_FILENAME_LEN];
    strcpy (masterPrivKeyFileName, CA_PRIVATE_KEY_FILE);
    
    char caCertFileName [MAX_FILENAME_LEN];
    strcpy (caCertFileName, CA_CERT_FILE);
    
    char choice;
    while ((choice = getopt (argc, argv, "hn:r:e:K:C:")) != -1) {
    
        switch (choice) {
            
            case 'h':
                usage();
                exit(0);
            break;
            
            case 'n':
                strcpy (name, optarg);
            break;

            case 'e':
                strcpy (email, optarg);
            break;

            case 'r':
                strcpy (realName, optarg);
            break;

            case 'K':
                strcpy (masterPrivKeyFileName, optarg);
            break;

            case 'C':
                strcpy (caCertFileName, optarg);
            break;

        }
    }

    cryptoInit();

    bool paramsOkFlag = true;
    
    if (strlen (name) == 0) {
        PRINT_INFO0 (INFO_LEVEL_FATAL_ERROR, "Please specify the name to be created using the -n option.\n");
        paramsOkFlag = false;
    }

    FILE* tmpFD;
    // Check if the masterPrivateKey exists
    tmpFD = fopen (masterPrivKeyFileName, "r" );
    if (tmpFD == NULL) {
        PRINT_INFO0 (INFO_LEVEL_FATAL_ERROR, "Unable to find master private key file %s.\n", masterPrivKeyFileName);
        paramsOkFlag = false;
    } else {
        fclose (tmpFD);
    }

    // Check if the CA's certificate file exists
    tmpFD= fopen (caCertFileName, "r");
    if (tmpFD == NULL) {
        PRINT_INFO0 (INFO_LEVEL_FATAL_ERROR, "Unable to find CA certificate file %s.\n", caCertFileName);
        paramsOkFlag = false;
    } else {
        fclose (tmpFD);
    }


    if ( !paramsOkFlag) {
        EXIT_ON_ERROR;
    }
    
    /** Load the private key with which the certificates are to be signed.*/
    masterPrivateKey = initPrivateKey(masterPrivKeyFileName);
   
    if (masterPrivateKey == NULL) {
        PRINT_INFO1 (INFO_LEVEL_FATAL_ERROR,
                    "Unable to load master private key from %s. Aborting.\n",
                    masterPrivKeyFileName
                );
        EXIT_ON_ERROR;
    }




    //get rid of the newline characters at the end
    stripNL(name);
    stripNL(realName);
    stripNL(email);

    PRINT_INFO3 (INFO_LEVEL_MINIMAL,
                "\nSecure Name Requested\n"
                "Name\t\t: %s\n"
                "Real Name\t: %s\n"
                "Email\t\t: %s\n",
                name, realName, email
            );

    //generate the public private key pair to be used for this name
    PRINT_INFO0 (INFO_LEVEL_MINIMAL,
                    "\nGenerating public/private key pair for this name...\n"
            );
    RSA* keypair = generateKeyPair();
    if (keypair == NULL) {
        PRINT_INFO1 ( INFO_LEVEL_FATAL_ERROR,
                    "Unable to generate public/private key pair for name %s. Aborting.\n",
                    name
                );
        EXIT_ON_ERROR;
    }
    
    RSA* pubKey = getPublicKeyFromPair(keypair);
    if (pubKey == NULL) {
        PRINT_INFO1 ( INFO_LEVEL_FATAL_ERROR,
                    "Unable to obtain public key from the generated key pair for name %s. Aborting.\n",
                    name
                );
        EXIT_ON_ERROR;
    }

    char pubKeyFileName[MAX_FILENAME_LEN];
    char privKeyFileName[MAX_FILENAME_LEN];
    char certFileName[MAX_FILENAME_LEN];
    sprintf (pubKeyFileName, "%s.pubkey", name);
    sprintf (privKeyFileName, "%s.privkey", name);
    sprintf (certFileName, "%s.cert", name);

    //save the generated public and private keys to disk
    saveKeysToDisk(keypair, pubKeyFileName, privKeyFileName);

    // create the new certificate, it is not signed yet.
    Certificate *newCert = constructNewCertificate(
                                    name,
                                    realName,
                                    email,
                                    pubKey
                            );


   
    // Sign the certificate     
    signCertificate(newCert, newCert->sign, &(newCert->signLen));

    // Write the signed certificate to disk.
    writeCertificateToFile (certFileName, newCert);

    if (testKeys (pubKeyFileName, privKeyFileName, certFileName, caCertFileName)) {
        PRINT_INFO1 (INFO_LEVEL_MINIMAL,
                    "Secure name %s successfully created.\n",
                    name
                );
        exit (0);
    } else {
        PRINT_INFO1 (INFO_LEVEL_MINIMAL,
                    "Verification of secure name %s failed.\n",
                    name
                );
        EXIT_ON_ERROR;
    }
}
