/**
 *  \file  db.c
 *  \brief This file contains functions to interact with the mysql database.
 */

#include <stdio.h>
#include <mysql.h>
#include <pthread.h>
#include <string.h>

#include "../../debug.h"
#include "db.h"
#include "../lib/certificate.h"

#define SOCKET_NAME "/var/lib/mysql/mysql.sock"

MYSQL  *dbConn;        /* pointer to connection handler */
pthread_mutex_t dbMutex;    /* The mutex for database operations */

void lockDB() {
    pthread_mutex_lock (&dbMutex);
}

void unlockDB() {
    pthread_mutex_unlock (&dbMutex);
}

void initDBConnection(char *dbName, char *hostName, char *userName, char *password) {

    pthread_mutex_init(&dbMutex, NULL);
   
    lockDB();
    
    dbConn = mysql_init (NULL);
    if (dbConn == NULL) {
        PRINT_INFO (INFO_LEVEL_FATAL_ERROR,
                    "Unable to initialize mysql (probably out of memory\n");
        unlockDB();
        EXIT_ON_ERROR;
    }
    
    if(mysql_real_connect (
                        dbConn,          /* pointer to connection handler */
                        hostName, /* host to connect to */
                        userName, /* user name */
                        password,  /* password */
                        dbName,   /* database to use */
                        0,             /* port (use default) */
                        SOCKET_NAME,          /* socket (use default) */
                        0              /* flags (none) */
                    ) == NULL ) {
        
        PRINT_INFO (INFO_LEVEL_FATAL_ERROR,
                    "Unable to connect to database %s @ %s with user name %s:\n%s",
                    dbName,
                    hostName,
                    userName,
                    mysql_error(dbConn)
                );
        unlockDB();
        EXIT_ON_ERROR;

    }

    unlockDB();

}

/**
  * Returns 1 if the name exists and has already been verified.
  * Else return 0.
  * Returns -1 on error.
  */
int checkIfNameVerified(char *name) {

    lockDB();
   
    char queryStr[1000];
    queryStr[0] = '\0';
    
    char encValueStr[500];
    mysql_real_escape_string (dbConn, encValueStr, name, strlen(name));
    
    sprintf(
                queryStr, 
                "SELECT * FROM " NAME_ALLOCATION_TABLE 
                " WHERE " COL_NAME "='%s'"
                ,
                encValueStr
           );
   
    PRINT_DEBUG (DEBUG_LEVEL_VERBOSE,
                "checkIfNameVerified:\n%s\n", 
                queryStr
            );

    if(mysql_query(dbConn, queryStr)){ /* Make query */
        PRINT_INFO (
                    INFO_LEVEL_WARNING,
                    "Unable to check if the name %s has been verified.\n"
                    "The following query failed:\n"
                    "%s\n",
                    name,
                    mysql_error(dbConn)
                );
        unlockDB();
        return -1;
    }
           
    
    MYSQL_RES* res_set = mysql_use_result(dbConn); /* Download result from server */
    
    if (res_set == NULL) {
        unlockDB();
        return 0;
    } 

    MYSQL_ROW row = mysql_fetch_row(res_set); /* Get a row from the results */
    if (row == NULL) {

        unlockDB();
        return 0;

    }

    MYSQL_FIELD   *field;
    
    /* determine column display widths */
    mysql_field_seek (res_set, 0);
    int numFields = mysql_num_fields(res_set);
    
    int retVal = 0;
    int i;
    for (i = 0; i < numFields; i++) {
          
        field = mysql_fetch_field (res_set);
        if (strcmp(field->name, COL_VERIFIED_FLAG) == 0) {
            if (atoi(row[i]) == 1) {
                retVal = 1;
                break;
            }
        }
    }
    mysql_free_result(res_set);

    unlockDB();
    return retVal;
}

/**
  * Loads the certificate details of the certificate with the specified name
  * from the database.
  * Returns 1 if successful.
  * Else return 0.
  */
int loadCertificateFromDB (char *name, Certificate *cert) {

    lockDB();
   
    char queryStr[1000];
    queryStr[0] = '\0';
    
    char encValueStr[500];
    mysql_real_escape_string (dbConn, encValueStr, name, strlen(name));
    
    sprintf(
                queryStr, 
                "SELECT * FROM " NAME_ALLOCATION_TABLE 
                " WHERE " COL_NAME "='%s'"
                ,
                encValueStr
           );
   
    PRINT_DEBUG (DEBUG_LEVEL_VERBOSE,
                "loadCertificateFromDB:\n%s\n", 
                queryStr
            );

    if(mysql_query(dbConn, queryStr)){ /* Make query */
        PRINT_INFO (
                    INFO_LEVEL_WARNING,
                    "Unable to load details of certificate with name %s.\n"
                    "The following query failed:\n"
                    "%s\n",
                    name,
                    mysql_error(dbConn)
                );
        unlockDB();
        return 0;
    }
           
    
    MYSQL_RES* res_set = mysql_use_result(dbConn); /* Download result from server */
    
    if (res_set == NULL) {
        unlockDB();
        return 0;
    } 

    MYSQL_ROW row = mysql_fetch_row(res_set); /* Get a row from the results */
    if (row == NULL) {

        unlockDB();
        return 0;

    }

    MYSQL_FIELD   *field;
    
    /* determine column display widths */
    mysql_field_seek (res_set, 0);
    int numFields = mysql_num_fields(res_set);
    
    int i;
    for (i = 0; i < numFields; i++) {
          
        field = mysql_fetch_field (res_set);
        if (strcmp(field->name, COL_EMAIL_ADDR) == 0) {
            
            strcpy(cert->emailAddr, row[i]);
            
        } else if (strcmp(field->name, COL_REAL_NAME) == 0) {

            strcpy(cert->realName, row[i]);

        } else if (strcmp(field->name, COL_SIGN_LEN) == 0) {
            cert->signLen =  (unsigned int) atoi (row[i]);
        
        } else if (strcmp(field->name, COL_PUB_KEY_LEN) == 0) {
            cert->publicKeyLen = (unsigned long) atol (row[i]);
        
        } else if (strcmp(field->name, COL_NAME) == 0) {
            strcpy(cert->name, row[i]);
        }
    }
    
    mysql_field_seek (res_set, 0);
    for (i = 0; i < numFields; i++) {
        field = mysql_fetch_field (res_set);
        if (strcmp(field->name, COL_PUB_KEY) == 0) {
            
            memcpy(cert->publicKeyBuf, row[i], cert->publicKeyLen);
       
        } else if (strcmp (field->name, COL_SIGN) == 0) {

            printf ("loading sign of lengt %d from database.\n", cert->signLen);

            memcpy(cert->sign, row[i], cert->signLen);
        }
    }
        
    mysql_free_result(res_set);

    unlockDB();
    return 1;
}


/**
  * Returns 1 if the specified email password/nonce of the given name
  * matches with value in the database.
  * Else return 0.
  * Returns -1 on error.
  */
int checkNamePassword(char *name, char *password) {

    lockDB();
   
    char queryStr[1000];
    queryStr[0] = '\0';
    
    char encValueStr[500];
    mysql_real_escape_string (dbConn, encValueStr, name, strlen(name));
    
    sprintf(
                queryStr, 
                "SELECT * FROM " NAME_ALLOCATION_TABLE 
                " WHERE " COL_NAME "='%s'"
                ,
                encValueStr
           );
   
    PRINT_DEBUG (DEBUG_LEVEL_VERBOSE,
                "checkNamePassword:\n%s\n", 
                queryStr
            );

    if(mysql_query(dbConn, queryStr)){ /* Make query */
        PRINT_INFO (
                    INFO_LEVEL_WARNING,
                    "Unable to check the password of the name %s.\n"
                    "The following query failed:\n"
                    "%s\n",
                    name,
                    mysql_error(dbConn)
                );
        unlockDB();
        return -1;
    }
           
    
    MYSQL_RES* res_set = mysql_use_result(dbConn); /* Download result from server */
    
    if (res_set == NULL) {
        unlockDB();
        return -2;
    } 

    MYSQL_ROW row = mysql_fetch_row(res_set); /* Get a row from the results */
    if (row == NULL) {

        unlockDB();
        return -2;

    }

    MYSQL_FIELD   *field;
    
    /* determine column display widths */
    mysql_field_seek (res_set, 0);
    int numFields = mysql_num_fields(res_set);
    
    int retVal = 0;
    int i;
    for (i = 0; i < numFields; i++) {
          
        field = mysql_fetch_field (res_set);
        if (strcmp(field->name, COL_EMAILED_PASSWORD) == 0) {
            
            if (strcmp(row[i], password) == 0) {
                retVal = 1;
                break;
            }
        }
    }
    mysql_free_result(res_set);

    unlockDB();
    return retVal;
}



/**
  * Returns NAME_EXISTS_EMAIL_SAME if the name exists under the same email address
  * Returns NAME_EXISTS_EMAIL_DIFF if the name exists with a different email address.
  * Returns NAME_NO_EXIST if the name does not exist.
  * Returns -1 on error.
  */
int checkIfNameExists(char *name, char *email) {

    lockDB();
   
    char queryStr[1000];
    queryStr[0] = '\0';
    
    char encValueStr[500];
    mysql_real_escape_string (dbConn, encValueStr, name, strlen(name));
    
    sprintf(
                queryStr, 
                "SELECT * FROM " NAME_ALLOCATION_TABLE 
                " WHERE " COL_NAME "='%s'"
                ,
                encValueStr
           );
   
    PRINT_DEBUG (DEBUG_LEVEL_VERBOSE,
                "checkIfNameExists:\n%s\n", 
                queryStr
            );

    if(mysql_query(dbConn, queryStr)){ /* Make query */
        PRINT_INFO (
                    INFO_LEVEL_WARNING,
                    "Unable to check the existence of name %s.\n"
                    "The following query failed:\n"
                    "%s\n",
                    name,
                    mysql_error(dbConn)
                );
        unlockDB();
        return -1;
    }
           
    
    MYSQL_RES* res_set = mysql_use_result(dbConn); /* Download result from server */
    
    if (res_set == NULL) {
        unlockDB();
        return NAME_NO_EXIST;
    } 



    MYSQL_ROW row = mysql_fetch_row(res_set); /* Get a row from the results */
    if (row == NULL) {

    	mysql_free_result(res_set);
        unlockDB();
        return NAME_NO_EXIST;

    }

    MYSQL_FIELD   *field;
    
    /* determine column display widths */
    mysql_field_seek (res_set, 0);
    int numFields = mysql_num_fields(res_set);
    
    int i;
    for (i = 0; i < numFields; i++) {
          
        field = mysql_fetch_field (res_set);
        if (strcmp(field->name, COL_EMAIL_ADDR) == 0) {
          
            if (strcmp(row[i], email) == 0) {

    		mysql_free_result(res_set);
		unlockDB();
		return NAME_EXISTS_EMAIL_SAME;
            } else {
    		mysql_free_result(res_set);
    		unlockDB();
		return NAME_EXISTS_EMAIL_DIFF;
	    }
        }
    }

    mysql_free_result(res_set);
    unlockDB();

    return NAME_EXISTS_EMAIL_DIFF;
}

/**
  * Change the verified status of the specified name to true.
  * Put the signature into the database.
  * Returns 1 on success, else returns 0.
  */
int  setNameSigned ( char *name, unsigned char *sign, unsigned int signLen) {

    lockDB();
    
    char encNameStr[1000];
    mysql_real_escape_string (dbConn, encNameStr, name, strlen(name));
    char encSignStr[1000];
    mysql_real_escape_string (dbConn, encSignStr, (char *)sign, signLen);
    
    char queryStr[5000];
    queryStr[0] = '\0';
    
    sprintf (queryStr, 
            "UPDATE " NAME_ALLOCATION_TABLE 
            " SET " COL_VERIFIED_FLAG "= 1, " 
            COL_SIGN "=\'%s\'," 
            COL_SIGN_LEN "= %d " " WHERE " COL_NAME "=\'%s\'",
            
            encSignStr,
            signLen,
            encNameStr
        );
   
    if ( mysql_query(dbConn, queryStr) != 0) {
        PRINT_INFO (INFO_LEVEL_FATAL_ERROR,
                "Unable to change the verified status of name %s.\n%s\n", 
                name, mysql_error(dbConn)
            );
        unlockDB();
        return 0;//error
    }

    PRINT_DEBUG (
                DEBUG_LEVEL_VERBOSE, 
                "setNameVerified:%s\n", 
                queryStr
            );
    unlockDB();
    return 1;   //ok

}


/**
  * Adds a new name record to the database.
  * This first deletes any record with the same name.
  */
int addNewNameRecord (
            char *name,
            char *realName,
            char *emailAddr,
            unsigned char *publicKey,
            unsigned int publicKeyLen,
            char *emailedPassword,
            char *revokePassword,
            int verifiedFlag
       ) {


    lockDB();
    
    char valueStr[4000];
    sprintf(valueStr,
            "'%s','%s','%s','%s','%s','%s','%d'"
            ,
            name,
            realName,
            emailAddr,
            publicKey,
            emailedPassword,
            revokePassword,
            verifiedFlag
        );

            
    char queryStr[5000];
    queryStr[0] = '\0';
   
    //first delete any existing record with the same name.
    sprintf (queryStr,
		"DELETE from " NAME_ALLOCATION_TABLE " where " COL_NAME " = '%s'",
	        name
            );
    if ( mysql_query(dbConn, queryStr) != 0) {
        PRINT_INFO (INFO_LEVEL_FATAL_ERROR,
                "Unable to delete existing name (%s) record before inserting new one .\n%s\n", name, mysql_error(dbConn)
            );
        unlockDB();
        return 0;//error
    }


    queryStr[0] = '\0';
 
    sprintf (queryStr, 
            "INSERT into " NAME_ALLOCATION_TABLE "(" COL_NAME "," COL_REAL_NAME "," COL_EMAIL_ADDR "," COL_PUB_KEY "," COL_PUB_KEY_LEN "," COL_EMAILED_PASSWORD "," COL_REVOKE_PASSWORD "," COL_VERIFIED_FLAG "," COL_SIGN_LEN ") "
            "VALUES ("
        );
   
    char encValueStr[1000];
    
    mysql_real_escape_string (dbConn, encValueStr, name, strlen(name));
    strcat(queryStr, "\'");
    strcat(queryStr, encValueStr);
    strcat(queryStr, "\',");
    
    mysql_real_escape_string (dbConn, encValueStr, realName, strlen(realName));
    strcat(queryStr, "\'");
    strcat(queryStr, encValueStr);
    strcat(queryStr, "\',");
    
    mysql_real_escape_string (dbConn, encValueStr, emailAddr, strlen(emailAddr));
    strcat(queryStr, "\'");
    strcat(queryStr, encValueStr);
    strcat(queryStr, "\',");
    
    mysql_real_escape_string (dbConn, encValueStr, (char *)publicKey, publicKeyLen);
    strcat(queryStr, "\'");
    strcat(queryStr, encValueStr);
    strcat(queryStr, "\',");
  
    sprintf(encValueStr, "\'%d\',", publicKeyLen);
    strcat (queryStr, encValueStr);

    mysql_real_escape_string (dbConn, encValueStr, emailedPassword, strlen(emailedPassword));
    strcat(queryStr, "\'");
    strcat(queryStr, encValueStr);
    strcat(queryStr, "\',");
    
    mysql_real_escape_string (dbConn, encValueStr, revokePassword, strlen(revokePassword));
    strcat(queryStr, "\'");
    strcat(queryStr, encValueStr);
    strcat(queryStr, "\',");
    
    sprintf(encValueStr, "\'%d\',", verifiedFlag);
    strcat (queryStr, encValueStr);
   
    int signLen = 0;
    sprintf(encValueStr, "\'%d\'", signLen);
    strcat (queryStr, encValueStr);
   
    strcat (queryStr, ")");
   
    if ( mysql_query(dbConn, queryStr) != 0) {
        PRINT_INFO (INFO_LEVEL_FATAL_ERROR,
                "Unable to insert new Name record into database.\n%s\n", mysql_error(dbConn)
            );
        unlockDB();
        return 0;//error
    }

    PRINT_DEBUG (
                DEBUG_LEVEL_VERBOSE, 
                "addNewNameRecord:%s\n", 
                queryStr
            );
    unlockDB();
    return 1;   //ok

}


void closeDBConnection() {
    mysql_close (dbConn);
}



