/**
 *  \file nameAllocatorClient.cpp
 *  \brief This is the client program used to request a certificate from
 *  nameAllocatorServer program.
*/
 
#include <stdio.h>
#include <string.h>


#include "OCALA_Export.h"

#ifndef _WIN32
#	include <unistd.h>
#	include <sys/socket.h>
#	include <sys/types.h>
#else
#   define MSG_WAITALL 0
#	include <Winsock2.h>
#	include <Windows.h>
extern "C" {
#	include "getopt.h"
}
#endif

extern "C" {
#	include "../../netwrap.h"
}

    #include "../lib/cryptoFunctions.h"
	#include "../lib/cryptoHeaders.h"
    #include "../lib/certificate.h"
    #include "../lib/securityCommon.h"
    #include "../lib/utils.h"

#include "../../debug.h"
#include "../../common.h"

#include "nameAllocatorClientExitCodes.h"

#define DEFAULT_SERVER_CONFIG_FILE "nameAllocator.config"

/** The name of the file in which the certificate is to be stored.
  */
char certFileName[MAX_FILENAME_LEN];
char pubKeyFileName[MAX_FILENAME_LEN];
char privKeyFileName[MAX_FILENAME_LEN];


char serverConfigFile[MAX_FILENAME_LEN];

void usage() {
    PRINT_INFO0 (INFO_LEVEL_MINIMAL,
            "nameAllocator_client\n"
            "\t\t-C <server config file>\n"
            "\t\t-r <name request file>\n"
            "\t\t-c <confirm request file>\n"
            "\t\t-t \t\tTest keys and the certificate\n"
            "\t\t-g <name> -f <filename> \tRetrieve the certificate for the specified name and store it in filename.\n"
            "\t\t-H\n"
            );
    exit(EXCODE_NORMAL_TERMINATION);
}

int connectToServer (char* serverName, int serverPort) {
    
    int sockfd, numbytes;  
    struct hostent *he;
    struct sockaddr_in their_addr; // connector's address information 

    if ((he=gethostbyname(serverName)) == NULL) {  // get the host info 
        PRINT_INFO1 (INFO_LEVEL_FATAL_ERROR,
                    "Unable to resolve server name %s.\n",
                    serverName
                );
        perror("gethostbyname");
        exit (EXCODE_CONNECT_RESOLVE_FAILED);
    }

    if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
        PRINT_INFO0 (INFO_LEVEL_FATAL_ERROR,
                    "Unable to open socket.\n"
                );
        perror("socket");
        exit (EXCODE_CONNECT_OPENSKT_FAILED);
    }

    their_addr.sin_family = AF_INET;    // host byte order 
    their_addr.sin_port = htons(serverPort);  // short, network byte order 
    their_addr.sin_addr = *((struct in_addr *)he->h_addr);
    memset(&(their_addr.sin_zero), '\0', 8);  // zero the rest of the struct 

    if (connect(sockfd, (struct sockaddr *)&their_addr, sizeof(struct sockaddr)) == -1) {
        PRINT_INFO2 (INFO_LEVEL_FATAL_ERROR,
                    "Unable to connect to %s:%d.\n",
                    serverName,
                    serverPort
                );
        perror("connect");
        exit (EXCODE_CONNECT_FAILED);
    }

    return sockfd;
}

void saveKeysToDisk (RSA *keypair) {

    unsigned char pubKeyBuf[CERT_MAX_PUBLIC_KEY_LEN];
    int pubKeyLen = savePublicKeyToBuffer(keypair, pubKeyBuf, CERT_MAX_PUBLIC_KEY_LEN);
    
    FILE *fp = fopen (pubKeyFileName, "wb");
    if (fp == NULL) {
        PRINT_INFO1 (
                    INFO_LEVEL_FATAL_ERROR,
                    "Unable to open file %s for writing the public key.\n",
                    pubKeyFileName
                );
        exit (EXCODE_KEY_WRITE_FAILED);
    }
    fwrite( &pubKeyLen, sizeof(pubKeyLen), 1, fp);
    fwrite( pubKeyBuf, pubKeyLen, 1, fp);
    fclose(fp);
    
    unsigned char privKeyBuf[CERT_MAX_PRIVATE_KEY_LEN];
    int privKeyLen = savePrivateKeyToBuffer(keypair, privKeyBuf, CERT_MAX_PRIVATE_KEY_LEN);

    fp = fopen (privKeyFileName, "wb");
    if (fp == NULL) {
        PRINT_INFO1 (
                    INFO_LEVEL_FATAL_ERROR,
                    "Unable to open file %s for writing the private key.\n",
                    privKeyFileName
                );
        exit (EXCODE_KEY_WRITE_FAILED);
    }
    fwrite( &privKeyLen, sizeof(privKeyLen), 1, fp);
    fwrite( privKeyBuf, privKeyLen, 1, fp);
    fclose(fp);
    
}


/** 
  * Reads a message from the specified socket and prints it to screen.
  */
void printMsgFromSocket (int sockFD) {
    
    char msgBuf[MAX_INFO_MSG_LEN];
    
    recv(sockFD, msgBuf, sizeof(msgBuf), MSG_WAITALL);
    msgBuf[MAX_INFO_MSG_LEN - 1] = '\0';    //Sanitizing the buffer

    PRINT_INFO1 (
            INFO_LEVEL_MINIMAL,
            "%s\n",
            msgBuf
        );
}


/**
  * Get the server to contact details from the 
  * server config file.
  */
void getServerDetails (char* serverConfigFile, char *serverName, int *serverPort) {
    FILE* srvFP = fopen (serverConfigFile, "r");
    if (srvFP == NULL) {
        PRINT_INFO1 (INFO_LEVEL_FATAL_ERROR,
                "Unable to open server configuration file %s.\n",
                serverConfigFile
            );
        exit (EXCODE_CONFIGFILE_ERROR);
    }
    
    fscanf(srvFP, "%s", serverName);
    fscanf(srvFP, "%d", serverPort);
    fclose(srvFP);

}

void makeNewNameRequest (char *nameReqFileName) {
    
    PRINT_INFO0 (
            INFO_LEVEL_VERBOSE,
            "Making a new Name request.\n"
        );
   
    //read the request from the file.
    char name[CERT_MAX_NAME_LEN];
    char realName[CERT_MAX_REAL_NAME_LEN];
    char email[CERT_MAX_EMAIL_LEN];

    FILE *fp = fopen(nameReqFileName, "r");
    if (fp == NULL) {
        PRINT_INFO1 (INFO_LEVEL_FATAL_ERROR,
                "Unable to open name request file %s.\n",
                nameReqFileName
            );
        exit (EXCODE_REQFILE_ERROR);
    }

    fgets(name, CERT_MAX_NAME_LEN, fp);
    fgets(realName, CERT_MAX_REAL_NAME_LEN, fp);
    fgets(email, CERT_MAX_EMAIL_LEN, fp);
    fclose(fp);

    //get rid of the newline characters at the end
    stripNL(name);
    stripNL(realName);
    stripNL(email);

    PRINT_INFO3 (INFO_LEVEL_MINIMAL,
                "\nName Request:\n"
                "Name\t\t: %s\n"
                "Real Name\t: %s\n"
                "Email\t\t: %s\n",
                name, realName, email
            );

    cryptoInit();
    
    //generate the public private key pair to be used for this name
    PRINT_INFO0 (INFO_LEVEL_MINIMAL,
                    "\nGenerating public/private key pair for this name...\n"
            );
    RSA* keypair = generateKeyPair();
    RSA* pubKey = getPublicKeyFromPair(keypair);

    // open a connection to the name allocation server
    // and send the certificate for signing.
    
    // Read the server host and port number from config file.
    char serverName[MAX_PARAM_LEN];
    int serverPort;
    getServerDetails(serverConfigFile, serverName, &serverPort);

    PRINT_INFO2 (INFO_LEVEL_MINIMAL,
                "Sending signing request to %s:%d\n",
                serverName, serverPort
            );
    // create the new certificate, it is not signed yet.
    Certificate *newCert = constructNewCertificate(
                                    name,
                                    realName,
                                    email,
                                    pubKey
                            );


   
    int sockFD = connectToServer (serverName, serverPort); 

    uint8_t requestCode = REQCODE_NEW_REQUEST;
    if (send(sockFD,  (char *)&requestCode, sizeof(requestCode), 0) == -1) {
        perror ("Error Sending Request Code");
        exit (EXCODE_SEND_REQ_ERROR);
    }

    hton (newCert);
    if (send(sockFD, (char *)newCert, sizeof(Certificate), 0) == -1) {
        perror ("Error Sending Request");
        exit (EXCODE_SEND_REQ_ERROR);
    }

    uint8_t replyCode;
    int numBytes = recv(sockFD, (char *)&replyCode, sizeof(replyCode), MSG_WAITALL);
    if (numBytes == -1) {
    
        perror("Error getting reply code");
        exit (EXCODE_SEND_REQ_ERROR);
        
    } else {
        char msgBuf[MAX_INFO_MSG_LEN];

        switch (replyCode) {
            case REQCODE_NAME_UNAVAILABLE:
                PRINT_INFO0 (
                        INFO_LEVEL_MINIMAL,
                        "Requested name is not available.\n"
                    );
                exit (EXCODE_NAME_UNAVAILABLE);
                
            break;

            case REQCODE_NAME_AVAILABLE:
	    case REQCODE_NAME_REREQUEST:
                PRINT_INFO0 (
                        INFO_LEVEL_MINIMAL,
                        "The requested name is AVAILABLE.\n"
                        );
                        
                       
                printMsgFromSocket (sockFD);
    
                //save the keys to disk
                saveKeysToDisk(keypair);
    
                exit (EXCODE_NAME_AVAILABLE);
            break;
   

            case REQCODE_POLICY_DISALLOW:

                printMsgFromSocket(sockFD);
                nw_close (sockFD);
                exit (EXCODE_POLICY_DISALLOW);
            break;
            
            case REQCODE_ERROR:
                printMsgFromSocket(sockFD);
                nw_close(sockFD);
                exit (EXCODE_SERVER_INTERNAL_ERROR);
            break;
            
            default:
                PRINT_INFO1 (
                        INFO_LEVEL_WARNING,
                        "Unexpected/Unknown reply code %d received.\n",
                        replyCode
                    );
                nw_close(sockFD);
                exit (EXCODE_UNEXPECTED_REPLY);
            break;
        }
    }
}


void makeNameRequestConfirmation (char *cnfFileName) {
    PRINT_INFO0 (
                INFO_LEVEL_VERBOSE,
                "Making Name Request Confirmation.\n"
            );
    
    //read the confirmation details from the file
    char name[CERT_MAX_NAME_LEN];
    char password[CERT_EMAIL_NONCE_LEN];

    FILE *fp = fopen(cnfFileName, "r");
    if (fp == NULL) {
        PRINT_INFO1 (INFO_LEVEL_FATAL_ERROR,
                "Unable to open confirmation file %s.\n",
                cnfFileName
            );
        exit (EXCODE_CONFIRMFILE_ERROR);
    }

    fgets(name, CERT_MAX_NAME_LEN, fp);
    fgets(password, CERT_EMAIL_NONCE_LEN, fp);
    fclose(fp);

    //get rid of the newline characters at the end
    stripNL(name);
    stripNL(password);

    PRINT_INFO2 (INFO_LEVEL_MINIMAL,
                "\nName Confirmation:\n"
                "Name\t\t: %s\n"
                "Password\t: %s\n",
                name,
                password
            );

    cryptoInit();
    
    // open a connection to the name allocation server
    // and send the name request confirmation.
    
    // Read the server host and port number from config file.
    char serverName[MAX_PARAM_LEN];
    int serverPort;
    getServerDetails (serverConfigFile, serverName, &serverPort);

    PRINT_INFO2 (INFO_LEVEL_MINIMAL,
                "Sending name request confirmation to %s:%d\n",
                serverName, serverPort
            );

    
    CertPasswordMsg passwordMsg;
    strcpy(passwordMsg.name, name);
    strcpy(passwordMsg.password, password);
    
    int sockFD = connectToServer (serverName, serverPort); 

    uint8_t requestCode = REQCODE_NEW_REQUEST_CONFIRM;
    if (send(sockFD, (char *) &requestCode, sizeof(requestCode), 0) == -1) {
        perror ("Error Sending Request Code");
    }

    hton (&passwordMsg);
    if (send(sockFD, (char *) &passwordMsg, sizeof(CertPasswordMsg), 0) == -1) {
        perror ("Error Sending Confirmation");
    }

    uint8_t replyCode;
    int numBytes = recv(sockFD, (char *)&replyCode, sizeof(replyCode), MSG_WAITALL);
    if (numBytes == -1) {
    
        perror("Error getting reply code");
    
    } else {
        char msgBuf[MAX_INFO_MSG_LEN];

        switch (replyCode) {

            case REQCODE_NAME_VERIFY_DUPLICATE:
                PRINT_INFO0 (
                        INFO_LEVEL_MINIMAL,
                        "The name has already been verified previously.\n"
                        );
                        
                
                printMsgFromSocket (sockFD);
                nw_close(sockFD);
                exit (EXCODE_NAME_ALREADY_VERIFIED);
            break;

            case REQCODE_SIGNED:
                // The certificate was successfully signed.
                Certificate signedCert;
                blankCertificate( &signedCert);

                printf ("waiting for complete certiciate\n");
                numBytes = recv (sockFD, (char *) &signedCert, sizeof(Certificate), MSG_WAITALL);
                ntoh (&signedCert);
                if (numBytes < sizeof(Certificate)) {
                    PRINT_INFO0 (INFO_LEVEL_WARNING,
                                "Received incomplete certificate.\n"
                            );
                }
                
                writeCertificateToFile (certFileName, &signedCert);

                PRINT_INFO2 (
                            INFO_LEVEL_MINIMAL,
                            "Success!  You have been allocated the name %s.\n"
                            "Your certificate file is stored in the file %s.\n"
                            ,
                            signedCert.name,
                            certFileName
                        );
    
                nw_close(sockFD);
                exit(EXCODE_NAME_CONFIRMED);
            break;
           
            case REQCODE_ERROR:
                printMsgFromSocket (sockFD);
                nw_close(sockFD);
                exit (EXCODE_SERVER_INTERNAL_ERROR);
            break;
            
            case REQCODE_UNKNOWN_NAME:
                
                printMsgFromSocket (sockFD);
                nw_close(sockFD);
                exit (EXCODE_UNKNOWN_NAME);
            break;
                
            case REQCODE_INVALID_PASSWORD:
                printMsgFromSocket (sockFD);
                nw_close(sockFD);
                exit(EXCODE_INVALID_PASSWORD);
            break;
          
            default:
                PRINT_INFO1 (
                        INFO_LEVEL_WARNING,
                        "Unexpected/Unknown reply code %d received.\n",
                        replyCode
                    );

                nw_close(sockFD);
                exit (EXCODE_UNEXPECTED_REPLY);
            break;
        }
    }
}

/**
  * Do a consistency check for the keys and the certificate.
  */
void testKeys(char* keyPrefix) {

    if (keyPrefix != NULL) {
        sprintf (pubKeyFileName, "%s.pubkey", keyPrefix);
        sprintf (privKeyFileName, "%s.privkey", keyPrefix);
        sprintf (certFileName, "%s.cert", keyPrefix);
    } else {
        PRINT_INFO0 (INFO_LEVEL_FATAL_ERROR,
                    "Please specify the prefix of the certificate and key file names.\n"
                );
    }

    PRINT_INFO4 (
                INFO_LEVEL_MINIMAL,
                "Using files:\n"
                "My Public Key File\t: %s\n"
                "My Private Key File\t: %s\n"
                "My Certificate File\t: %s\n"
                "CA Certificate File\t: %s\n"
                "\n"
                ,
                pubKeyFileName,
                privKeyFileName,
                certFileName,
                CA_CERT_FILE
            );
    
    // Load the public key from the disk.
    EVP_PKEY* pubKey = loadPublicKeyFromDisk(pubKeyFileName);
    if (pubKey == NULL) {
        PRINT_INFO1 (INFO_LEVEL_FATAL_ERROR,
                    "Unable to load public key from %s.\n",
                    pubKeyFileName
                );
        exit (EXCODE_KEY_LOAD_FAILED);
    }
    
    // Load the private key from the disk.
    EVP_PKEY* privKey = loadPrivateKeyFromDisk(privKeyFileName);

    if (privKey == NULL) {
        PRINT_INFO1 (INFO_LEVEL_FATAL_ERROR,
                    "Unable to load private key from %s.\n",
                    privKeyFileName
                );
        exit (EXCODE_KEY_LOAD_FAILED);
    }
    // Load the certificate from the disk
    Certificate myCert;
    blankCertificate(&myCert);

    loadCertificateFromFile(certFileName, &myCert);
    PRINT_INFO0 ( INFO_LEVEL_MINIMAL,
                "Your signed certificate is:\n"
            );
    printf_certificate(&myCert);

    char testMsg[] = "This is a test message to check the signatures\n";

    // Check if my private and public keys are consistent.
    unsigned char sigBuf [MAX_SIGN_BUF_SIZE];
    uint16_t sigLen;
    sign (testMsg, strlen(testMsg), sigBuf, &sigLen, privKey);
    
    if (verifySignature(testMsg, strlen(testMsg), sigBuf, sigLen, pubKey)) {
        PRINT_INFO0(
                INFO_LEVEL_MINIMAL,
                "With Stored Keys: My Signature Verified Ok.\n"
            );
    } else {
        PRINT_INFO0(
                INFO_LEVEL_MINIMAL,
                "With Stored Keys: My Signature Verification FAILED.\n"
            );
    }


    // Check if my public key obtained from my certificate is consistent with my private key.
    sign (testMsg, strlen(testMsg), sigBuf, &sigLen, privKey);
   
    EVP_PKEY *pubKeyFromCert = getPublicKeyFromCertificate(&myCert);
    if (pubKeyFromCert == NULL) {
        PRINT_INFO0 (
                    INFO_LEVEL_FATAL_ERROR,
                    "Error while reading public key from certificate file.\n"
                );
        return;
    }
    
    if (verifySignature(testMsg, strlen(testMsg), sigBuf, sigLen, pubKeyFromCert)) {
        PRINT_INFO0(
                INFO_LEVEL_MINIMAL,
                "With Public Key from Certificate: My Signature Verified Ok.\n"
            );
    } else {
        PRINT_INFO0(
                INFO_LEVEL_MINIMAL,
                "With Public Key from Certificate: My Signature Verification FAILED.\n"
            );
    }

    // Verify the CA's signature on your certificate.
    EVP_PKEY* caPubKey = initPublicKey (CA_CERT_FILE);
    if (caPubKey == NULL) {
         PRINT_INFO0 (
                    INFO_LEVEL_FATAL_ERROR,
                    "Error while reading CA public key from CA certificate file.\n"
                );
        return;
    
    }
    if (verifyCertificate(&myCert, caPubKey)) {
        PRINT_INFO0 (
                INFO_LEVEL_MINIMAL,
                "CA's signature on the certificate Verified Ok.\n"
            );
    } else {
        PRINT_INFO0 (
                INFO_LEVEL_MINIMAL,
                "CA's signature on the certificate Verification FAILED.\n"
            );
    }
}

/**
  * Retrieve the certificate associated with the specified name and store
  * it in fileName.
  */
void retrieveCertificate ( char *name, char *fileName) {
    PRINT_INFO1 (
                INFO_LEVEL_VERBOSE,
                "Retrieving Certificate with name %s.\n",
                name
            );
    
    cryptoInit();
    
    // open a connection to the name allocation server
    // and send the name request confirmation.
    
    // Read the server host and port number from config file.
    char serverName[MAX_PARAM_LEN];
    int serverPort;
    getServerDetails(serverConfigFile, serverName, &serverPort);
    
  
    // TODO: We are uncleanly reusing a structure.
    CertPasswordMsg passwordMsg;
    strcpy(passwordMsg.name, name);
    strcpy(passwordMsg.password, "\0");
    
    int sockFD = connectToServer (serverName, serverPort); 

    uint8_t requestCode = REQCODE_RETRIEVE_CERTIFICATE;
    if (send(sockFD, (char *)&requestCode, sizeof(requestCode), 0) == -1) {
        perror ("Error Sending Request Code");
    }
    hton (&passwordMsg);
    if (send(sockFD, (char *) &passwordMsg, sizeof(CertPasswordMsg), 0) == -1) {
        perror ("Error Sending Retrieve Request");
    }

    uint8_t replyCode;
    int numBytes = recv(sockFD, (char *) &replyCode, sizeof(replyCode), MSG_WAITALL);
    if (numBytes == -1) {
    
        perror("Error getting reply code");
    
    } else {
        char msgBuf[MAX_INFO_MSG_LEN];

        switch (replyCode) {

            case REQCODE_RETRIEVE_CERTIFICATE_FAILED:
                PRINT_INFO1 (
                        INFO_LEVEL_MINIMAL,
                        "Unable to retrieve certificate for name %s.\n",
                        name
                        
                    );
                        
                
                printMsgFromSocket (sockFD);
            break;

            case REQCODE_RETRIEVE_CERTIFICATE_SUCCESS:
                // The certificate was successfully retrieved
                Certificate retCert;
                blankCertificate (&retCert);

                numBytes = recv (sockFD, (char *) &retCert, sizeof(Certificate), MSG_WAITALL);
                ntoh (&retCert);
                if (numBytes < sizeof(Certificate)) {
                    PRINT_INFO0 (INFO_LEVEL_WARNING,
                                "Received incomplete certificate.\n"
                            );
                }
                
                writeCertificateToFile (fileName, &retCert);

                PRINT_INFO2 (
                            INFO_LEVEL_MINIMAL,
                            "Success!  The certificate for name %s has been retrieved and \n"
                            "stored in the file %s.\n"
                            ,
                            retCert.name,
                            fileName
                        );
                
            break;
           
            case REQCODE_ERROR:
                
                printMsgFromSocket (sockFD);

            break;
          
            default:
                PRINT_INFO1 (
                        INFO_LEVEL_WARNING,
                        "Unexpected/Unknown reply code %d received.\n",
                        replyCode
                    );
            break;
        }
    }
    nw_close(sockFD);
}


#ifndef _WIN32
int main (int argc, char** argv) {
#else 
int nameAllocatorClientMain (int argc, char **argv) {

	nw_init();
#endif

    // Set the default pubkey, privkey and certificate filenames
    strcpy (pubKeyFileName, "my.pubkey");
    strcpy (privKeyFileName, "my.privkey");
    strcpy (certFileName, "my.cert");

    strcpy (serverConfigFile, DEFAULT_SERVER_CONFIG_FILE);

    opterr = 1; //getopt parameter

    int getCertFlag = 0;    //flag to note if the -g option was specified
    char certName[CERT_MAX_NAME_LEN];   // the argument associated with the -g option
    
    char fileName[MAX_FILENAME_LEN]; //to store argument associated with -f option if any
    int fileFlag = 0;   //flag to not if the -f option was specified.
    
    char choice;
    while ((choice = getopt (argc, argv, "hr:c:t:g:f:C:")) != -1) {
       
        switch (choice) {
            
            case 'h':
                usage();
                exit(EXCODE_NORMAL_TERMINATION);
            break;

            case 'r':
                //Request a new name
                makeNewNameRequest(strdup(optarg));             
            break;

            case 'c':
                // Confirm name request by sending password obtained from email.
                makeNameRequestConfirmation(strdup(optarg));
            break;

            case 'C':
                strcpy (serverConfigFile, optarg);
            break;

            case 't':
                // Test the stored public key, private key and certificate
                testKeys(optarg);
                exit(EXCODE_NORMAL_TERMINATION);
            break;

            case 'g':
                getCertFlag = 1;
                strcpy (certName, optarg);
            break;
            
            case 'f':
                fileFlag = 1;
                strcpy(fileName, optarg);
            break;
            
            default:
                usage();
                exit(EXCODE_NORMAL_TERMINATION);
            break;
        }
    }

    if ( getCertFlag ) {
        
        if ( ! fileFlag) {
            PRINT_INFO0 (
                    INFO_LEVEL_FATAL_ERROR,
                    "You must specify a filename to store the retrieved certificate "
                    "by using the -f option.\n"
                );
           exit (EXCODE_INVALID_CMDLINE_ARGS) ;
        }
        
        retrieveCertificate (certName, fileName);
        
    }
    
}


