/**
 *  \file nameAllocatorServer.cpp
 *  \brief This file contains the code of the certificate authority.
*/

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

    #include "../lib/cryptoHeaders.h"
    #include "../lib/cryptoFunctions.h"
    #include "../lib/certificate.h"
    #include "../lib/securityCommon.h"
    #include "../lib/utils.h"
    #include "mailFunctions.h"
    #include "nameRegPolicy.h"
    #include "db.h"

#include "../../debug.h"
#include "../../common.h"

#define EMAIL_START_MSG_FILE    "emailStartMsg.txt"
#define EMAIL_END_MSG_FILE      "emailEndMsg.txt"

char emailStartStr[3000];
char emailEndStr[3000];

/** The default port to listen on */
#define DEFAULT_PORT_NUM 12345

/** Allow all name registration requests.
  */
#define NAME_REG_POLICY_ALLOW_ALL 1

/** The name should match the specified email address.
  */
#define NAME_REG_POLICY_NAME_EMAIL_MATCH 2


/** The policy to be used while registering names.
  * For example, should the requested name and email match?
  */
int nameRegistrationPolicy = NAME_REG_POLICY_NAME_EMAIL_MATCH;


/** The private key used to sign the certificates */
EVP_PKEY* masterPrivateKey;

/** File containing info about the database to use */
#define DB_CONFIG_FILE "db.config"

int initTCPSocket(int serverPortNum) {
    
    int sock;
    struct sockaddr_in sin;
    int val=1;
    
    if((sock=socket(AF_INET,SOCK_STREAM,0))<0) {
        PRINT_INFO(
                INFO_LEVEL_FATAL_ERROR,
                "FATAL ERRROR: Could not initialize TCP Listening socket.\n"
            );
        EXIT_ON_ERROR;
    }
    
    memset(&sin,0,sizeof(sin));
    sin.sin_addr.s_addr=INADDR_ANY;
    sin.sin_family=AF_INET;
    sin.sin_port=htons(serverPortNum);
    setsockopt(sock,SOL_SOCKET,SO_REUSEADDR, &val,sizeof(val));
    
    if(bind(sock,(struct sockaddr *)&sin, sizeof(sin))<0) {
        PRINT_INFO (INFO_LEVEL_FATAL_ERROR,
                    "FATAL ERROR: Could not bind to %s:%d\n",
                    inet_ntoa(sin.sin_addr),
                    ntohs(sin.sin_port)
                );
        EXIT_ON_ERROR;
    }
    
    PRINT_INFO (INFO_LEVEL_MINIMAL,
                    "Socket bound to %s:%d\n",
                    inet_ntoa(sin.sin_addr),
                    ntohs(sin.sin_port)
                );
        
    
    listen(sock,6);  

    return(sock);
  }

/**
  * Sign the specified certificate with the master
  * private key of the server.
  * The signature is returned in sigBuf.
  * The length of the signature is put in sigLen
  */
void signCertificate(Certificate *cert, unsigned char sigBuf[MAX_SIGN_BUF_SIZE], uint16_t *sigLen) {
    

   // Bring together the data to be signed.
   char dataToBeSigned[MAX_DATA_TO_SIGN_SIZE];
  
   int dataLen;
   getDataToSign(cert, dataToBeSigned, &dataLen);
   
   //Sign it
   sign(dataToBeSigned, dataLen, sigBuf, sigLen, masterPrivateKey);
   
}

/**
  * Send a message saying that requested name is not available.
  */
void sendNameUnavailableMsg (int reqSocket) {
    
    uint8_t msgCode = REQCODE_NAME_UNAVAILABLE;
    if (send (reqSocket, &msgCode, sizeof(msgCode), 0) == 0) {
        perror("Error sendNameUnavailableMsg");   
    }
}

/**
  * Send a message saying that the name was previously verified.
  */
void sendNameAlreadyVerified (int reqSocket) {
    
    uint8_t msgCode = REQCODE_NAME_VERIFY_DUPLICATE;
    if (send (reqSocket, &msgCode, sizeof(msgCode), 0) == 0) {
        perror("Error sendNameAlreadyVerified");   
    }
    
    char tmpMsg[MAX_INFO_MSG_LEN];
    strcpy(tmpMsg, "The name you are trying to verify was previously verified.  Cannot sign it again.  Request just the certificate.");
    
    if (send (reqSocket, tmpMsg, strlen(tmpMsg) + 1, 0) == 0) {
        perror("Error sendNameAlreadyVerified");   
    }
    
}


void sendMsgToClient (int reqSocket, uint8_t msgCode, char *msg) {
    
    if (send (reqSocket, &msgCode, sizeof(msgCode), 0) == 0) {
        perror("Error sendMsgToClient");   
    }
    
    if (send (reqSocket, msg, strlen(msg) + 1, 0) == 0) {
        perror("Error sendMsgToClient");   
    }
}



/**
  * Reads the contents of the specified file into buf.
  * Returns 1 on success, 0 on error.
  */
int readMsgFromFile (char *fileName, char *buf) {
    
    FILE *fp = fopen ( fileName, "r");
    if (fp == NULL) {
        
        PRINT_INFO (
                    INFO_LEVEL_FATAL_ERROR,
                    "Unable to read message from file %s.\n",
                    fileName
                );
        return 0;
    }
    
    buf[0] = '\0';

    char tmpLine[5000];
    while ((fgets(tmpLine, 5000, fp) != NULL)) {
        
        strcat (buf, tmpLine);
    }

    strcat (buf, "\0");
    
    fclose(fp);

    return 1;
}

void initMailSystem() {

    initMail();
    
    if (! readMsgFromFile(EMAIL_START_MSG_FILE, emailStartStr)) {
        EXIT_ON_ERROR;
    }
    
    if (! readMsgFromFile(EMAIL_END_MSG_FILE, emailEndStr)) {
        EXIT_ON_ERROR;
    }
}

int sendEmailToClient(Certificate *cert, char *confirmPassword, char *revokePassword) {

    
    char body [10000];

    sprintf (
                body,
                "%s\n"
                "Requested Name\t\t\t:\t%s\n"
                "Real Name\t\t\t:\t%s\n"
                "Confirmation Password\t\t:\t%s\n"
                "Name revocation Password\t:\t%s\n"
                "%s\n"
                ,

                emailStartStr,
                cert->name,
                cert->realName,
                confirmPassword,
                revokePassword,
                emailEndStr

            );
  
    // form the subject line
    char subjectLine[1000];
    sprintf (
                subjectLine,
                "UCB OCALA CA: Name request for %s"
                ,
                cert->name
            );
    
    // finally send the email
    return sendMail (cert->emailAddr, subjectLine, body);
    
}


/**
  * Checks if the name request specified in cert is to be allowed.
  * If allowed, the function returns NULL.
  * If disallowed, the reason for rejection is put in the msg buffer.
  * A pointer to the msg buffer is returned.
  */
char *doPolicyCheck( Certificate *cert, char *msg) {

    switch (nameRegistrationPolicy) {
        
        case NAME_REG_POLICY_NAME_EMAIL_MATCH:
            return doNameEmailMatchCheck (cert->name, cert->emailAddr, msg);
        break;

        case NAME_REG_POLICY_ALLOW_ALL:
            // Allow all policy
            return NULL;
        break;
            
    }
}

/**
  * This is a new request for a certificate.
  * as against recontact the NAME with the emailed password.
  */
void processNewCertRequest(int reqSocket) {
    
    int numBytes;
    Certificate cert;
    blankCertificate ( &cert);

    if ((numBytes=recv(reqSocket, &cert, sizeof(Certificate), MSG_WAITALL)) == -1) {
        perror("Receive Error");
    }

    ntoh (&cert);

    //sanitize the received request.
    cert.name[CERT_MAX_NAME_LEN - 1] = '\0';
    cert.realName[CERT_MAX_REAL_NAME_LEN - 1] = '\0';
    cert.emailAddr[CERT_MAX_EMAIL_LEN - 1] = '\0';
   
    stripWhiteSpace(cert.name);
    stripWhiteSpace(cert.emailAddr);
    stripWhiteSpace(cert.realName);
    
    printf_certificate(&cert);
   
    //TODO Check if request is wellformed.

    int checkVar = checkIfNameExists(cert.name, cert.emailAddr);
    if (checkVar == NAME_EXISTS_EMAIL_DIFF) {
        PRINT_DEBUG (DEBUG_LEVEL_VERBOSE,
               "Sending Name Unavailable message for name %s.\n",
               cert.name
            );
        sendNameUnavailableMsg(reqSocket);
        return;

    } else if (checkVar == -1) {
         PRINT_DEBUG (DEBUG_LEVEL_VERBOSE,
               "Error occurred while checking if name %s exists.\n",
               cert.name
            );
        sendMsgToClient(
                reqSocket, 
                REQCODE_ERROR,
                "Internal database error 1: Unable to fulfil request."
            );
        return;
    
    } else {
        // Name has not already been taken or taken with same email address..
        
        char policyCheckMsg[MAX_INFO_MSG_LEN];

        if (doPolicyCheck(&cert, policyCheckMsg) == NULL) {

            // This request is allowed under the current
            // name registration policy
    
            //generate random email and revoke password
            char newEmailPassword[CERT_EMAIL_NONCE_LEN];
            char newRevokePassword[CERT_REVOKE_PASSWORD_LEN];

            generateRandomString(newEmailPassword, CERT_EMAIL_NONCE_LEN);
            generateRandomString(newRevokePassword, CERT_REVOKE_PASSWORD_LEN);
	
            // Add to database
            int retVal = addNewNameRecord (  
                    cert.name,
                    cert.realName,
                    cert.emailAddr,
                    cert.publicKeyBuf,
                    cert.publicKeyLen,
                    newEmailPassword,
                    newRevokePassword,
                    0
                );

            if (retVal) {
                //successfully inserted record into the database
                PRINT_INFO (
                            INFO_LEVEL_VERBOSE,
                            "emailPassword: %s\n",
                            newEmailPassword
                        );

		if (checkVar == NAME_NO_EXIST) {
                	sendMsgToClient (reqSocket,
                                REQCODE_NAME_AVAILABLE,
                                "You will shortly receive an email at the address you provided. "
                                "The name will be registered only after you execute the instructions given in the email."
                            );
		} else {
		 	sendMsgToClient (reqSocket,
                                REQCODE_NAME_REREQUEST,
                                "You have re-requested the same name.\n"
				"You will shortly receive an email at the address you provided. "
                                "The name will be registered only after you execute the instructions given in the email."
                            );

		}
            
                // Send the email with the confirmation and revokation password to the client.
                sendEmailToClient(&cert, newEmailPassword, newRevokePassword);
            
            } else {
                sendMsgToClient (
                                reqSocket,
                                REQCODE_ERROR,
                                "Internal Database Error 3: Unable to fulfill request."
                    );
            }
        
        } else {
            
            // The current name registration policy forbids this
            // name request
            sendMsgToClient (reqSocket, REQCODE_POLICY_DISALLOW, policyCheckMsg);
            
        }
    }
}

/**
  * The password emailed to the user during the name request is now supplied.
  * If the password matches, sign the certificate and send the signature to the client.
  */
void processNewCertPassword (int reqSocket) {

    CertPasswordMsg passwordMsg;
    
    int numBytes = recv(reqSocket , &passwordMsg, sizeof(passwordMsg), MSG_WAITALL);
    ntoh (&passwordMsg);

    passwordMsg.name[CERT_MAX_NAME_LEN - 1] = '\0'; //Sanitize
    passwordMsg.password[CERT_EMAIL_NONCE_LEN - 1] = '\0';  //Sanitize

    stripWhiteSpace(passwordMsg.name);
    
    int checkVal = checkIfNameVerified(passwordMsg.name);

    if (checkVal == 1) {
        //The name was already verified earlier.
        sendNameAlreadyVerified(reqSocket);
        
    } else if (checkVal == -1) {
        
        sendMsgToClient(
                reqSocket, 
                REQCODE_ERROR,
                "Internal Database Error 2: Unable to fulfill request."
            );
        return;
        
    } else {
     
        int checkPassVal = checkNamePassword(passwordMsg.name, passwordMsg.password);
        if (checkPassVal == 1) {
            // Password is correct.  So sign the certificate and send it to the client.
        
            Certificate cert;
            blankCertificate ( &cert);

            loadCertificateFromDB(passwordMsg.name, &cert);

            PRINT_INFO (INFO_LEVEL_VERBOSE,
                "Signing Certificate:\n"
                );
            printf_certificate(&cert);

            signCertificate(&cert, cert.sign, &cert.signLen);
      
            int retVal = setNameSigned(passwordMsg.name, cert.sign, cert.signLen);

            uint8_t msgCode = REQCODE_SIGNED;
            if (send (reqSocket, &msgCode, sizeof(msgCode), 0) == 0) {
                perror("Error sendReqCodeSigned");   
            }
       
            hton (&cert);
            if (send (reqSocket, &cert, sizeof(Certificate), 0) == 0) {
                perror("Error sendReqCodeSigned1");   

            }

        } else if (checkPassVal == -2) {
            sendMsgToClient (
                    reqSocket, 
                    REQCODE_UNKNOWN_NAME, 
                    "The name you are trying to confirm was never requested.\n"
                    "First request the name, before attempting to confirm it."
                );
            
        } else if (checkPassVal == 0) {

            sendMsgToClient (
                    reqSocket, 
                    REQCODE_INVALID_PASSWORD, 
                    "The password you supplied for the name is incorrect.\n"
                );
        } else {
            
            sendMsgToClient (
                    reqSocket, 
                    REQCODE_ERROR, 
                    "Internal Error: Unable to fulfil request.\n"
                );

        }
    }
}

/**
  * Retrieve a previously signed certificate.
  */
void retrieveCertificate (int reqSocket) {

    //TODO: Unclean reuse of a structure
    CertPasswordMsg passwordMsg;
    
    int numBytes = recv(reqSocket , &passwordMsg, sizeof(passwordMsg), MSG_WAITALL);
    ntoh (&passwordMsg);

    passwordMsg.name[CERT_MAX_NAME_LEN - 1] = '\0'; //Sanitize
    passwordMsg.password[CERT_EMAIL_NONCE_LEN - 1] = '\0';  //Sanitize

    stripWhiteSpace(passwordMsg.name);

    int checkVal = checkIfNameVerified(passwordMsg.name);

    if (checkVal == 1) {
        //A verified certificate with this name exists.
        Certificate cert;
        blankCertificate ( &cert);

        loadCertificateFromDB(passwordMsg.name, &cert);
     
        uint8_t msgCode = REQCODE_RETRIEVE_CERTIFICATE_SUCCESS;
        if (send (reqSocket, &msgCode, sizeof(msgCode), 0) == 0) {
            perror("Error sending REQCODE_RETRIEVE_CERTIFICATE_SUCCESS");   
        }
       
        hton (&cert);
        if (send (reqSocket, &cert, sizeof(Certificate), 0) == 0) {
            perror("Error sending REQCODE_RETRIEVE_CERTIFICATE_SUCCESS1");   
         }
   
    } else if (checkVal == -1) {
        
        sendMsgToClient(
                reqSocket, 
                REQCODE_ERROR,
                "Internal Database Error 2: Unable to fulfill request."
            );
        return;
        
    } else {
     
        sendMsgToClient (
                    reqSocket, 
                    REQCODE_RETRIEVE_CERTIFICATE_FAILED, 
                    "No certificate exists with the requested name."
            );
    }
}
/** Process the request from the client. 
  * This is where bulk of the action takes place.
  */
void processRequest (int reqSocket, struct sockaddr_in *remoteAddr) {
  
    int numBytes;
    
   
    //Read the request code
    uint8_t requestCode;
    if ((numBytes=recv(reqSocket, &requestCode, sizeof(requestCode), MSG_WAITALL)) == -1) {
        perror("Request Code Receive Error");
    }

    switch (requestCode) {

        case REQCODE_NEW_REQUEST:

            processNewCertRequest(reqSocket);
        break;

        case REQCODE_NEW_REQUEST_CONFIRM:
            //the password for a certificate request is now supplied.
            processNewCertPassword(reqSocket);
        break;

        case REQCODE_RETRIEVE_CERTIFICATE:
            // Retrieve an already existing certificate
            retrieveCertificate(reqSocket);
        break;
        
        default:
            PRINT_INFO (
                        INFO_LEVEL_WARNING,
                        "Received Unknown Request Code %d\n",
                        requestCode
                    );
        break;
    }
    
    
}


void usage() {
    PRINT_INFO (INFO_LEVEL_MINIMAL,
                "nameAllocatorServer\n"
                "\t\t-p <port number> The port number on which the server should listen.\n"
                "\n"
                "\t\t-y <policy number> The name registration policy\n"
                "\t\t\t\t\t1 = ALLOW ALL\n"
                "\t\t\t\t\t2 = NAME EMAIL MATCH\n"
                "\n"
                "\t\t-h \t\t\tPrint thhis help message.\n"
            );
}

/**
  * Reads the database details from
  * file DB_CONFIG_FILE and initializes the database using
  * that information.
  */
void initDB() {
    
    FILE *fp = fopen(DB_CONFIG_FILE, "r");
    if (fp == NULL) {
        PRINT_INFO (INFO_LEVEL_FATAL_ERROR,
                "Unable to open database config file %s.\n",
                DB_CONFIG_FILE
            );
        EXIT_ON_ERROR;
    }

    char dbName [MAX_PARAM_LEN];
    char hostName [MAX_PARAM_LEN];
    char userName [MAX_PARAM_LEN];
    char password [MAX_PARAM_LEN];
    
    fgets(dbName, MAX_PARAM_LEN, fp);
    fgets(hostName,MAX_PARAM_LEN, fp);
    fgets(userName, MAX_PARAM_LEN, fp);
    fgets(password, MAX_PARAM_LEN, fp);
    fclose(fp);

    //get rid of the newline characters at the end
    stripNL(dbName);
    stripNL(hostName);
    stripNL(userName);
    stripNL(password);

    initDBConnection(dbName, hostName, userName, password);
    
}

int main(int argc, char **argv) {

    opterr = 1; //getopt parameter
    int serverPortNum = DEFAULT_PORT_NUM;

    char choice;
    while ((choice = getopt (argc, argv, "hp:y:")) != -1) {
    
        switch (choice) {
            
            case 'h':
                usage();
                exit(0);
            break;

            case 'p':
                //Port number
                serverPortNum = atoi(optarg);
            break;

            case 'y':
                // Name registraton policy
                nameRegistrationPolicy = atoi(optarg);
            break;

        }
    }

    cryptoInit();
    initDB();
    initMailSystem();

    /** Load the private key with which the certificates are to be signed.*/
    masterPrivateKey = initPrivateKey(CA_PRIVATE_KEY_FILE);
    
    int listenSocket = initTCPSocket(serverPortNum);
  
    while (1) {
        // Loop infinitely waiting for connections.
        struct sockaddr_in remoteAddr;
        socklen_t remoteAddrLen;
        
        int reqSocket = accept (listenSocket, (sockaddr *)&remoteAddr, &remoteAddrLen);
        
        if (reqSocket < 0) {
            PRINT_INFO (INFO_LEVEL_WARNING,
                        "Error accepting connection from %s:%d.\n",
                        inet_ntoa(remoteAddr.sin_addr),
                        ntohs(remoteAddr.sin_port)
                    );
            continue;
        }

        PRINT_INFO (    INFO_LEVEL_VERBOSE,
                        "Accepted connection from %s:%d.\n",
                        inet_ntoa(remoteAddr.sin_addr),
                        ntohs(remoteAddr.sin_port)
                    );
        
        pid_t pid;
        
        if ((pid = fork())) {
            //parent process
        } else {
            //child process
            srand (time (NULL)); // each child requires an srand
            processRequest(reqSocket, &remoteAddr);
            shutdown(reqSocket, SHUT_RDWR);
            close(reqSocket);
            pthread_exit(NULL);
        }
         
    }



}
