/**
 *  \file PathInfo.cpp
 *  \brief Functions to manipulate PathInfo objects.
 *  A PathInfo object represents an end to end path
 * consisting of one or more HopInfo objects.
*/

#include <string>
#include <string.h>

#include <sys/types.h>

#if defined(_WIN32)
extern "C" {
	#include <regex.h>
}
#else 
	#include <regex.h>
#endif


using namespace std;

#include "../common.h"
#include "../Proxy.h"
#include "PathInfo.h"
#include "HopInfo.h"
#include "../oci/OCIPacketBlock.h"
#include "../utils/StringTokenizer.h"
#include "../utils/InetFns.h"
#include "../utils/utilFunctions.h"

PathInfo::PathInfo () {

    // dummy constructor
    currHop = -1;
}

/** Constructs a PathInfo object using the details given in path.
  */
PathInfo::PathInfo (const PathInfo_pkt* path) {

    currHop = path->currHop;
    
    for (int hopIndex = 0; hopIndex < path->numHops; hopIndex++) {
        HopInfo *currHop = new HopInfo (path->hops[hopIndex].ocdId, path->hops[hopIndex].endPointName);
        addHop (currHop);
    }
}

/**
  * Creates a copy of the specified path.
  * Not all fields are copied (for eg: patternList).
  */
PathInfo::PathInfo (PathInfo * path) {

    securityInfo = path->securityInfo;
    compressionInfo = path->compressionInfo;

    currHop = path->currHop;
    list<HopInfo *>::iterator itr = path->hopList.begin();
    while ( itr != path->hopList.end()) {
        hopList.push_back (new HopInfo ( (*itr)));
        itr++;
    }
}

/** 
  * Constructs a new path info object from the specified journal str.
  */
PathInfo* PathInfo::constructPathInfo (char *journalStr, Proxy * proxy) {

    PathInfo *newPathInfo = new PathInfo();    
    StringTokenizer st1 = StringTokenizer (journalStr, ":");

    newPathInfo->currHop = st1.nextIntToken();
    newPathInfo->securityInfo.protocol = st1.nextIntToken();
    
    while (st1.hasMoreTokens()) {

        StringTokenizer st2 = StringTokenizer (st1.nextToken(), "&");
        string tmp_ocdId = st2.nextToken();
        string tmp_endPointName = st2.nextToken();
        
        HopInfo* tmpHop = new HopInfo(tmp_ocdId.c_str(), tmp_endPointName.c_str());
        tmpHop->ocd = proxy->getOCDWithId (tmpHop->ocdId);

        printf ("\n\n\n\n\t\t\tConstructing hop = |%s|%s|\n",
                tmp_ocdId.c_str(), tmp_endPointName.c_str());

        newPathInfo->addHop (tmpHop);
    }

    return newPathInfo;
}

/** PathInfo destructor */
PathInfo::~PathInfo () {
    list<HopInfo *>::iterator itr = hopList.begin();
    while ( itr != hopList.end()) {
        delete (*itr);
        itr++;
    }

    list<regex_t*>::iterator regexItr =  patternList.begin();
    while (regexItr != patternList.end()) {
        delete (*regexItr);
		regexItr++;
    }
    patternList.clear();

    
}

/**
  * Adds the specified destination port (in network byte order)
  * as one of the match criteria for this PathInfo object.
  */
void PathInfo::addDstPortMatch (port_protocol_n_t dstPort) {

    dstPortList.push_back (dstPort);
}

/**
  * Adds the specified destination IP address as one of the match 
  * criteria for this PathInfo object.
  * @param ipAddrStr The destination IP address to be added.  The address should
  * be a string of the form A.B.C.D.
  */
void PathInfo::addDstIPAddrMatch (const char* ipAddrStr) {
    
    ipaddr_n_t tmpIPAddr;
    
    // convert the ip address string to ipaddr_n_t
    if (!convert_ipaddr_str2int (ipAddrStr, &tmpIPAddr)) {
        PRINT_INFO1 (
                    INFO_LEVEL_WARNING,
                    "PathInfo: Unable to add destination IP address match for %s.\n",
                    ipAddrStr
                );
    } else {
        dstIPAddrList.push_back (tmpIPAddr);
    }
    
}


/**
  * Adds the specified source port (in network byte order)
  * as one of the match criteria for this PathInfo object.
  */
void PathInfo::addSrcPortMatch (port_protocol_n_t srcPort) {

    srcPortList.push_back (srcPort);
}

/**
  * Adds the specified source IP address as one of the match 
  * criteria for this PathInfo object.
  * @param ipAddrStr The destination IP address to be added.  The address should
  * be a string of the form A.B.C.D.
  */
void PathInfo::addSrcIPAddrMatch (const char* ipAddrStr) {
    
    ipaddr_n_t tmpIPAddr;
    
    // convert the ip address string to ipaddr_n_t
    if (!convert_ipaddr_str2int (ipAddrStr, &tmpIPAddr)) {
        PRINT_INFO1 (
                    INFO_LEVEL_WARNING,
                    "PathInfo: Unable to add source IP address match for %s.\n",
                    ipAddrStr
                );
    } else {
        srcIPAddrList.push_back (tmpIPAddr);
    }
    
}




/**
  * Add the specified URL pattern to the list of patterns associated
  * with this PathInfo object.
  */
void PathInfo::addURLPattern (string pattern) {

    char tmpRule[MAX_LEGACY_URL_RULE_LEN];
    const char *tmpPattern = pattern.c_str();
    int j = 0;
    
    for(int k=0 ; k<(int)strlen(tmpPattern) ; k++) {
        if ( tmpPattern[k] == '.') {
	        tmpRule[j++] = '\\';
	        tmpRule[j++] = '.';
        
	    } else if ( tmpPattern[k] == '*') {
	        tmpRule[j++] = '.';
	        tmpRule[j++] = '*';
        
	    } else {
	        tmpRule[j++] = tmpPattern[k];
	    }
    }

    tmpRule[j] = 0;

   
    regex_t* rex = new regex_t;
    patternList.push_back (rex);

    int retVal = regcomp( rex, tmpRule, REG_EXTENDED | REG_ICASE);
    
    if (retVal) {
    
	    PRINT_INFO1 (INFO_LEVEL_WARNING, 
	 	    "The URL pattern %s is not a valid regular expression.\n",
		    pattern.c_str()
		    );
	    char tmpBuf[MAX_PRINT_BUF];
	    regerror(retVal, rex, tmpBuf, MAX_PRINT_BUF);
	    PRINT_INFO1 (INFO_LEVEL_WARNING,
						"%s\n", tmpBuf
					);
    }
    //char tmpBuf [MAX_PRINT_BUF];
    //PRINT_DEBUG1 (DEBUG_LEVEL_SUPER, "URL Pattern created:\n%s\n", this->toString(tmpBuf));

   
    char tmpPattern2 [MAX_PARAM_LEN];
    strcpy (tmpPattern2, tmpPattern);
    stripWS (tmpPattern2);
    
    if (strCountCh (tmpPattern2, '*') == 1) {
        // There is only one * in the pattern
        
        if ( tmpPattern2 [0] == '*') {
            // Add this pattern to the suffix list if this pattern is of the form *foo.
            patternSuffixList.push_back (string ((tmpPattern2 + 1)));
        } else if (tmpPattern2[strlen(tmpPattern2) - 1] == '*') {
            tmpPattern2[strlen(tmpPattern2) - 1] = '\0';
            patternPrefixList.push_back (string (tmpPattern2));
        }
    }
}

/**
  * Add the specified hop as the next hop in the path.
  */
void PathInfo::addHop (HopInfo *hop) {

    hopList.push_back (hop);
}

/**
  * Retrieve the the specified hop.
  * Hop numbers start from 0.
  */
HopInfo* PathInfo::getHop (int hopNum) {

    int count = 0;
    list<HopInfo*>::iterator itr = hopList.begin(); 
    
    while (itr != hopList.end()) {
        if (hopNum == count) {
            return (HopInfo *) (*itr);
        }
        itr++;
        count++;
    }

    return NULL;
}

/**
  * Returns the name of the last hop of the path.
  */
string PathInfo::getEndPointName () {

    list<HopInfo*>::iterator itr = hopList.end();
    itr--;

    HopInfo * currHop = *itr;
    return currHop->hopEndPointName;
}

/**
  * Get the number of hops in this path.
  */
int PathInfo::getHopCount() {

    return hopList.size();
}

/**
  * Get the number of hops to be authenticated.
  */
int PathInfo::getAuthHopCount() {

    // calculate the number of hops that don't need to be authenticated.
    int noAuth = 0;
    list<HopInfo*>::iterator itr = hopList.begin();
    while (itr != hopList.end()) {

        if ( (*itr)->avoidAuthFlag) {
            noAuth++;
        }
        itr++;
    }
    return hopList.size() - noAuth;
}

/**
  * Construct an OCI path setup message based on this path.
  * @param url The name of the final endpoint
  */
OCIPathSetupBlock* PathInfo::getPathSetupBlock (int * blockSize, const string url) {

    int tmpNumHops = getHopCount();
    
    *blockSize = sizeof (OCIPathSetupBlock) + tmpNumHops * sizeof (HopInfo_pkt);
    OCIPathSetupBlock *block = allocateOCIPathSetupBlock (tmpNumHops);

    block->header.version = PROXY_VERSION;
    block->header.blockType = OCI_PATH_SETUP_BLOCK;
    block->header.nextBlock = 0;
   
    block->path.numHops = tmpNumHops;
    block->path.currHop = currHop + 1;
    //PRINT_DEBUG1 (DEBUG_LEVEL_SUPER, "block->path.currHop = %d.\n", block->path.currHop);

    int hopIndex = 0;
    list<HopInfo*>::iterator itr = hopList.begin();
    while (itr != hopList.end()) {

        HopInfo* currHop = (*itr);
        strcpy(block->path.hops[hopIndex].ocdId, currHop->ocdId.c_str());
        strcpy(block->path.hops[hopIndex].endPointName, currHop->hopEndPointName.c_str());
        
        itr++;
        hopIndex++;
    }

    /** Put the last hop end point name as the actual URL */
    HopInfo *lastHop = * (--itr);
    strcpy (block->path.hops[--hopIndex].endPointName, url.c_str());

    return block;
}

/**
  * Returns true if this path is equal to the specified path x.
  */
bool PathInfo::equals (PathInfo * x) {

    if (this->currHop != x->currHop) {
        printf ("\n\t\t\tcurrHop %d %d \n", this->currHop, x->currHop);
        return false;
    }

    if (this->getHopCount() != x->getHopCount()) {
        
        printf ("\n\t\t\thopCount %d %d \n", this->getHopCount(), x->getHopCount());
        return false;
    }

    list<HopInfo*>::iterator itr = this->hopList.begin();
    list<HopInfo*>::iterator itrX = x->hopList.begin();
    
    while ( itr != this->hopList.end()) {

        if ( ! (*itr)->equals( *itrX)) {

            return false;
        }
        itr++;
        itrX++;
    }

    return true;
}

/**
  * Dumps a string version of this pathInfo object for the journal.
  */
void PathInfo::getJournalString (char *buf) {

    sprintf( buf, "%d:%d", currHop, securityInfo.protocol);

    int numHops = hopList.size();

    list<HopInfo *>::iterator itr = hopList.begin();
    while (itr != hopList.end()) {
        HopInfo *currHop = (*itr);
        strcat (buf, ":");
        strcat (buf, currHop->ocdId.c_str());
        strcat (buf, "&");
        strcat (buf,  currHop->hopEndPointName.c_str());
        itr++;
    }
}
