/**
 *  \file PathInfo.h
 *  \brief Declares an end to end path specification.  Traffic matching the specific patterns in this 
 *  class will use the specified end to end path.
*/
#ifndef PATH_INFO_H
#define PATH_INFO_H

#include <list>

#if defined(_WIN32)
extern "C" {
	#include <regex.h>
}
#else 
	#include <regex.h>
#endif

#include <string>

#include "SecurityInfo.h"
#include "CompressionInfo.h"
#include "../oci/OCIPacketBlock.h"
#include "../common.h"

class Proxy;
class HopInfo;

using namespace std;

/** 
  * This class represents a PathInfo object contained in the configuration
  * file.
  */
class PathInfo {

private:

    /** The list of URL patterns for which this PathInfo is applicable */
    list<regex_t *> patternList;

    /** List of URL patterns which are of the form *foo or foo*.
      */
    list<string> patternPrefixList;
    list<string> patternSuffixList;
    

    /** The list of destination port numbers for which this PathInfo is applicable. 
     *  Each port number is in network byte order
     */
    list<port_protocol_n_t> dstPortList;

    /** The list of destination IP address for which this PathInfo is applicable. 
      * Each IP address is in network byte order.
      */
    list<ipaddr_n_t> dstIPAddrList;
   
    /** The list of source port numbers for which this PathInfo is applicable. 
     *  Each port number is in network byte order
     */
    list<port_protocol_n_t> srcPortList;

    /** The list of source IP address for which this PathInfo is applicable. 
      * Each IP address is in network byte order.
      */
    list<ipaddr_n_t> srcIPAddrList;
    

    
    /** The security properties of this path */
    SecurityInfo securityInfo;

    /** The compression properties of this path */
    CompressionInfo compressionInfo;

    /** The list of hops in this path */
    list<HopInfo *> hopList;

    /** Index to the current hop */
    int currHop;


public:

    void addURLPattern (string pattern);
    void addDstPortMatch (port_protocol_n_t dstPort);
    void addDstIPAddrMatch (const char *ipAddrStr);
    void addSrcPortMatch (port_protocol_n_t srcPort);
    void addSrcIPAddrMatch (const char *ipAddrStr);
    
    void addHop (HopInfo *);

    PathInfo();
    PathInfo (const PathInfo_pkt *path);
    PathInfo (PathInfo *path);
    ~PathInfo();
    
    HopInfo *getHop (int hopNum);
    int getAuthHopCount();
    int getHopCount();

    OCIPathSetupBlock *getPathSetupBlock (int *blockSize, const string finalEndPointName);

    bool equals (PathInfo *);

    string getEndPointName();

    void getJournalString (char *);
    static PathInfo* constructPathInfo (char *journalStr, Proxy *);

    friend class PathInfoConfigFile;
    friend class URLResolver;
    friend class OCIState;
    friend class CryptoState;
    friend class OCI;
    
};

#endif
