/**
 *  \file PathInfoConfigFile.cpp
 *  \brief Functions to parse the PathInfo objects from
 *  outbound.xml configuration file.
*/

#include <string.h>
#include <list>
#include <libxml/parser.h>
#include <libxml/tree.h>

#include "../debug.h"
#include "../common.h"
#include "PathInfoConfigFile.h"
#include "PathInfo.h"
#include "HopInfo.h"
#include "SecurityInfo.h"
#include "CompressionInfo.h"

#include <assert.h>

/** Loads PathInfo objects from the pathInfoFile to the specified list.
  * @param pathInfoFileName The file to parse
  * @param pathInfoList The list to be filled with the PathInfo objects loaded
  * from the config file.
  */
void PathInfoConfigFile::loadPathInfo (
                    const char *pathInfoFileName,
                    list <PathInfo *> * pathInfoList
                ) {

    xmlDoc *doc = NULL;

    doc = xmlReadFile(pathInfoFileName, NULL, 0);

    if (doc == NULL) {
        PRINT_INFO1 (INFO_LEVEL_FATAL_ERROR,
                        "Unable to read PathInfo objects from file %s.\n", pathInfoFileName);
        EXIT_ON_ERROR;
    }

    /*Get the root element node */
    xmlNode *rootElement = xmlDocGetRootElement(doc);

    if ( strcmp( (char *) rootElement->name, PATH_INFO_ROOT_ELEMENT) != 0) {
        
        PRINT_INFO3 (INFO_LEVEL_FATAL_ERROR,
                        "Invalid root tag in file %s\n.  Expected %s, found %s.\n",
                        pathInfoFileName, PATH_INFO_ROOT_ELEMENT, rootElement->name
                    );
        EXIT_ON_ERROR;

    }

    // Go over all the PathInfo elements.
    for (xmlNode *currPathInfoNode = rootElement->children; currPathInfoNode; currPathInfoNode = currPathInfoNode->next) {

        if (strcmp( (char *) currPathInfoNode->name, PATH_INFO_TAG_PATH_INFO) == 0) {

            PathInfo *pathInfo = new PathInfo;
            pathInfoList->push_back (pathInfo);

            // go through all the subelements of PathInfo
            for (xmlNode *childNode = currPathInfoNode->children; childNode; childNode = childNode->next) {
            
                xmlChar *prop = NULL;
            
                if ( strcmp ((char *) childNode->name, PATH_INFO_TAG_MATCH) == 0) {
                    
                    if ( (prop = xmlGetProp (childNode, (xmlChar *) PATH_INFO_ATTRIB_MATCH_URL_PATTERN)) != NULL) {
                   
                        // A url pattern was specified.
                        string pattern = (char *) prop;
                        pathInfo->addURLPattern (pattern);
                        xmlFree (prop);
                    
                    } else if ( (prop = xmlGetProp (childNode, (xmlChar *) PATH_INFO_ATTRIB_MATCH_DST_PORT)) != NULL) {
                        // A destination port number was specified in the Match element
                        
                        uint16_t tmpProtocol = PROTOCOL_TCP;
                        
                        port_protocol_n_t tmpPortProtocol;
                        tmpPortProtocol.port = htons(atoi ((char *) prop));
                        
                        if ( (prop = xmlGetProp (childNode, (xmlChar *) PATH_INFO_ATTRIB_MATCH_PROTOCOL)) != NULL) {
                            if (strcasecmp ((char *) prop, "udp") == 0) {
                                tmpProtocol = PROTOCOL_UDP;
                            }
                        }
                        
                        tmpPortProtocol.protocol = tmpProtocol;
                        
                        pathInfo->addDstPortMatch (tmpPortProtocol);

                        xmlFree (prop);

                     } else if ( (prop = xmlGetProp (childNode, (xmlChar *) PATH_INFO_ATTRIB_MATCH_SRC_PORT)) != NULL) {
                        // A source port number was specified in the Match element
                        
                        uint16_t tmpProtocol = PROTOCOL_TCP;
                        
                        port_protocol_n_t tmpPortProtocol;
                        tmpPortProtocol.port = htons(atoi ((char *) prop));
                        
                        if ( (prop = xmlGetProp (childNode, (xmlChar *) PATH_INFO_ATTRIB_MATCH_PROTOCOL)) != NULL) {
                            if (strcasecmp ((char *) prop, "udp") == 0) {
                                tmpProtocol = PROTOCOL_UDP;
                            }
                        }
                        
                        tmpPortProtocol.protocol = tmpProtocol;
                        
                        pathInfo->addSrcPortMatch (tmpPortProtocol);
                  
                        xmlFree (prop);
                    } else if ( (prop = xmlGetProp (childNode, (xmlChar *) PATH_INFO_ATTRIB_MATCH_DST_IPADDR)) != NULL) {
                        
                        // A destination IP address was specified in the Match element
                        pathInfo->addDstIPAddrMatch ((char *) prop);

                        PRINT_DEBUG1 (
                                        DEBUG_LEVEL_MINIMAL,
                                        "Added destination IP address match %s.\n",
                                        (char *) prop
                                    );

                        xmlFree (prop);
                   } else if ( (prop = xmlGetProp (childNode, (xmlChar *) PATH_INFO_ATTRIB_MATCH_SRC_IPADDR)) != NULL) {
                        
                        // A source IP address was specified in the Match element
                        pathInfo->addSrcIPAddrMatch ((char *) prop);

                        PRINT_DEBUG1 (
                                        DEBUG_LEVEL_MINIMAL,
                                        "Added source IP address match %s.\n",
                                        (char *) prop
                                    );

                        xmlFree (prop);

                    } else {
                            // No match criterion was specified in the Match element
                        
                            PRINT_INFO3 (
                                    INFO_LEVEL_WARNING,
                                    "Required attribute %s or %s not found in element %s.\n",
                                    PATH_INFO_ATTRIB_MATCH_DST_PORT,
                                    PATH_INFO_ATTRIB_MATCH_URL_PATTERN,
                                    childNode->name
                                );
                    }
                } else if ( strcmp ( (char *) childNode->name, PATH_INFO_TAG_SECURITY) == 0) {
                    prop = xmlGetProp (childNode, (xmlChar *) PATH_INFO_ATTRIB_SECURITY_PROTOCOL);        
                    if (prop != NULL) {

                        if ( (strcmp ( (char *) prop, SECURITY_PROTOCOL_CUSTOM_SSL_TXT) == 0)) {
                        
                            pathInfo->securityInfo.protocol = SECURITY_PROTOCOL_CUSTOM_SSL ;
                    
                        } else if ( (strcmp ( (char *) prop, SECURITY_PROTOCOL_SSL_TXT) == 0)) {
                        
                            pathInfo->securityInfo.protocol = SECURITY_PROTOCOL_SSL ;
                    
                        }  else {

                            pathInfo->securityInfo.protocol = SECURITY_PROTOCOL_UNKNOWN;

                        }

                        xmlFree (prop);
                    } else {
                        PRINT_INFO2 (
                                INFO_LEVEL_WARNING,
                                "Required attribute %s not found in element %s.\n",
                                PATH_INFO_ATTRIB_SECURITY_PROTOCOL,
                                childNode->name
                            );
                    }

                    prop = xmlGetProp (childNode, (xmlChar *) PATH_INFO_ATTRIB_SECURITY_PROTOCOL);        

                    if (prop != NULL) {

                        if ( (strcmp ( (char *) prop, SECURITY_PROTOCOL_CUSTOM_SSL_TXT) == 0)) {
                        
                            pathInfo->securityInfo.protocol = SECURITY_PROTOCOL_CUSTOM_SSL ;
                    
                        } else if ( (strcmp ( (char *) prop, SECURITY_PROTOCOL_SSL_TXT) == 0)) {
                        
                            pathInfo->securityInfo.protocol = SECURITY_PROTOCOL_SSL ;
                    
                        }  else {

                            pathInfo->securityInfo.protocol = SECURITY_PROTOCOL_UNKNOWN;

                        }

                        xmlFree (prop);
                    } else {
                        PRINT_INFO2 (
                                INFO_LEVEL_WARNING,
                                "Required attribute %s not found in element %s.\n",
                                PATH_INFO_ATTRIB_SECURITY_PROTOCOL,
                                childNode->name
                            );
                    }

 
                } else if ( strcmp ( (char *) childNode->name, PATH_INFO_TAG_COMPRESSION) == 0) {
                    prop = xmlGetProp (childNode, (xmlChar *) PATH_INFO_ATTRIB_COMPRESSION_ALGO);        
                    if (prop != NULL) {

                        if ( strcmp ((char *) prop, COMPRESSION_ALGO_ZLIB_TXT) == 0) {
                            pathInfo->compressionInfo.algorithm = COMPRESSION_ALGO_ZLIB;
                        } else {
                            pathInfo->compressionInfo.algorithm = COMPRESSION_ALGO_UNKNOWN;
                        }
                    
                        xmlFree (prop);
                    } else {
                        PRINT_INFO2 (
                            INFO_LEVEL_WARNING,
                            "Required attribute %s not found in element %s.\n",
                            PATH_INFO_ATTRIB_COMPRESSION_ALGO,
                            childNode->name
                        );
                    }   
    
                    prop = xmlGetProp (childNode, (xmlChar *) PATH_INFO_ATTRIB_COMPRESSION_LEVEL);        
                    if (prop != NULL) {

                        pathInfo->compressionInfo.compressionLevel = atoi ((char *) prop);
                        xmlFree (prop);
                    } else {
                        PRINT_INFO2 (
                            INFO_LEVEL_WARNING,
                            "Required attribute %s not found in element %s.\n",
                            PATH_INFO_ATTRIB_COMPRESSION_LEVEL,
                            childNode->name
                        );
                    }

                } else if ( strcmp ((char *) childNode->name, PATH_INFO_TAG_HOP) == 0 ) {
                    HopInfo* tmpHopInfo = new HopInfo ();
                    pathInfo->addHop(tmpHopInfo);

                    prop = xmlGetProp (childNode, (xmlChar *) PATH_INFO_ATTRIB_HOP_OVERLAY_ID);        
                    if (prop != NULL) {
                    
                        tmpHopInfo->ocdId = (char *) prop;
                        xmlFree(prop);

                    } else {
                        PRINT_INFO2 (
                            INFO_LEVEL_WARNING,
                            "Required attribute %s not found in element %s.\n",
                            PATH_INFO_ATTRIB_HOP_OVERLAY_ID,
                            childNode->name
                        );
                    }
                 
                    prop = xmlGetProp (childNode, (xmlChar *) PATH_INFO_ATTRIB_HOP_ENDPOINT_NAME);        
                    if (prop != NULL) {
                    
                        tmpHopInfo->hopEndPointName = (char *) prop;
                
                        xmlFree(prop);
                    } 

                    // Should the current hop be authenticated?
                    // This field is currently used only to indicate that a particular hop
                    // need not be authenticated, when we are having end to end authentication
                    // (i.e. when security for the whole path is enabled.)
                    // Ideally, it could be used to specific hop by hop authentication.
                    tmpHopInfo->avoidAuthFlag = false;
                    prop = xmlGetProp (childNode, (xmlChar *) PATH_INFO_ATTRIB_HOP_AUTHENTICATE);        
                    if (prop != NULL) {
                    
                        if (strcasecmp ((char *) prop, "no") == 0) {
                            tmpHopInfo->avoidAuthFlag = true;
                        }
                        xmlFree(prop);
                    } 
 
                } else if ( strcmp ( (char *) childNode->name, PATH_INFO_TAG_FAKE_NEGOTIATION) == 0) {
                    prop = xmlGetProp (childNode, (xmlChar *) PATH_INFO_ATTRIB_FAKE_NEGOTIATION_PREF);        
                    if (prop != NULL) {
                        
                        xmlFree(prop);
                    }

                } else if ( strcmp ( (char *) childNode->name, "text") == 0) {
                    //do nothing as of now.
                } else {
            
                    PRINT_INFO2 ( INFO_LEVEL_WARNING,
                                "Unknown subelement %s in %s.\n",
                                childNode->name,
                                PATH_INFO_TAG_PATH_INFO
                            );

                }
            }

        }
    }
                                
    /*free the document */
    xmlFreeDoc(doc);

    /*
     * Free the global variables that may
     * have been allocated by the parser.
     */
    //xmlCleanupParser();


}

