/**
 *  \file URLResolver.cpp
 *  \brief This file contains functions which resolve URLS
 *  to PathInfo objects.
*/ 

#include <string.h>
#ifdef __APPLE__
#include <sys/types.h>
#endif

#include <assert.h>

#if defined(_WIN32)
extern "C" {
	#include <regex.h>
}
#else 
	#include <regex.h>
#endif

#include "URLResolver.h"
#include "PathInfo.h"
#include "HopInfo.h"
#include "PathInfoConfigFile.h"

#include "../ocd/OCD.h"
#include "../ConfigFile.h"
#include "../Proxy.h"
#include "../Tunnel.h"
#include "../IPPacket.h"

#include "../utils/InetFns.h"
#include "../utils/utilFunctions.h"


/** Initialize the URL resolvers.  The url suffix to OCD
  * mappings are loaded from the specified list of overlay networks.
  * The mappings are created on a first-come-first-served basis, i.e.,
  * if an OCD occuring earlier in the overlayList associates
  * itself with a url prefix z, no other overlay network coming later
  * in the list can take up that prefix.  The order of overlay networks
  * is the same as that in the proxy configuration file.
  *
  * @param pointer to the Proxy object.
  * @param pathInfoFileName The file containing the various path info objects.
  */
URLResolver::URLResolver(Proxy *proxy, char *pathInfoFileName ) {

    this->proxy = proxy;
    this->pathInfoFileName = pathInfoFileName;
    
    // Load URL Pattern --> OCD entries from the configFile
    // Note that destination ip addresses and port numbers can also be part of the
    // pattern, not just plain url regular expressions.
    PathInfoConfigFile::loadPathInfo (pathInfoFileName, &this->pathInfoList);

    // In the configuration file, OCDs are identified by just the ocd id. 
    // Get the pointer to the actual OCD module corresponding to the ocdIds of 
    // the loaded PathInfo objects.
    resolvePathInfo();

    // Tell the proxy about which all dst/src ip/port number packets should be intercepted.    
    initPktCaptureRules();

    doPeriodicReload();
}

/**
  * This function is periodically called to reload the configuration file.
  */
void URLResolver::doPeriodicReload() {

    reloadConfigFile();

    // Schedule the next reload
    TObjectFunctor<URLResolver> *reloadFunctor = 
                        new TObjectFunctor<URLResolver> (this, &URLResolver::doPeriodicReload);
    TimerEvent* reloadTimer = TimerEvent::create(
                            10 ,    //seconds
                            0,    //microseconds
                            (TFunctor *) reloadFunctor,
                            true
                        );
    
    proxy->addTimerEvent(reloadTimer);
}

/**
  * Reload the URL resolver configuration file.
  */
void URLResolver::reloadConfigFile() {
    PRINT_DEBUG0 (DEBUG_LEVEL_VERBOSE,
                "Reloading URLResolver config file.\n"
            );
   
    // Delete all the items in the current list.
    list<PathInfo*>::iterator oldItr= this->pathInfoList.begin();
    while (oldItr != this->pathInfoList.end()) {
        delete (*oldItr);
        oldItr++;
    }
    this->pathInfoList.clear();

    // Read in the new list.
    PathInfoConfigFile::loadPathInfo (pathInfoFileName.c_str(), &this->pathInfoList);
    
    resolvePathInfo();
}

void URLResolver::addMapping (char *suffix, OCD *ocd) {
    
    //TODO check if suffix already exists.
    string tmpSuffix (suffix);
    map_urlSuffix_OCD.insert (
                pair<string, OCD *> ( *(new string (tmpSuffix)), ocd)
            );
}

/**
 * This functions sets the overlayNetwork pointers of the Hops of different
 * PathInfo objects using the Hop's overlayId field.
 */
void URLResolver::resolvePathInfo() {
    
    list<PathInfo*>::iterator itr = pathInfoList.begin();
    while (itr != pathInfoList.end()) {

	    PathInfo* currPathInfo = (PathInfo *) (*itr);

        list<HopInfo *>::iterator hopItr = currPathInfo->hopList.begin();
        while (hopItr != currPathInfo->hopList.end()) {
            HopInfo *currHopInfo = (HopInfo *) (*hopItr);
            currHopInfo->ocd = proxy->getOCDWithId(currHopInfo->ocdId);
			
			// first hop ocd of a path should be active in this node.
			// Else pring warning and exit.
			if (hopItr == currPathInfo->hopList.begin() && currHopInfo->ocd == NULL) {
				
				PRINT_INFO1 (
								INFO_LEVEL_FATAL_ERROR,
								"OCD of unknown id %s specified for first hop of path. Exiting\n",
								currHopInfo->ocdId.c_str()
							);
				EXIT_ON_ERROR;
							
			}
            hopItr++;
        }
        
	    PRINT_DEBUG0 (DEBUG_LEVEL_VERBOSE, "Loaded Path Info\n");
	    itr ++;
    }
}


/**
  * This function returns the OCD associated with the 
  * specified suffix.  If no overlay network is associated with
  * the specified suffix, <code>NULL</code> is returned.
  */
OCD * URLResolver::lookupURLSuffix(string suffix) {

    map<string, OCD*>::iterator itr = map_urlSuffix_OCD.find(suffix);
    
    if (itr != map_urlSuffix_OCD.end()) {
        return (OCD*)itr->second;
        
    } else {
            
        return NULL;
    }
}

/**
  * This funcion decides which OCD should handle
  * the specified URL.  If no overlay network that is willing/capable
  * of handling the specified URL is found, NULL is returned.
  * @param url The url for which the decision is to be made.
  * @return The overlay which is to handle this URL.
  */
PathInfo* URLResolver::resolveURL (const URL url) {

    PathInfo* retPathInfo = matchPrefixSuffix (url);
    return retPathInfo;

    // For now, till we fix memory leak bug in regexec(), use only prefix suffix matching
    // TODO
    /*
    if ( retPathInfo != NULL) {
        PRINT_DEBUG1 (DEBUG_LEVEL_VERBOSE,
                    "URLResolver::resolveURL - prefix/suffix match worked for %s.\n",
                    url.c_str()
                );
        return retPathInfo;
    } else {
        return matchURLPattern(url);
    }
    */
}

/**
  * This function determines the path to be taken by the specified
  * packet.  If no rules match the packet, NULL is returned.
  */
PathInfo* URLResolver::resolvePacket (IPPacket* pkt) {
    
    // go through all the PathInfo objects
    list<PathInfo*>::iterator itr = pathInfoList.begin();
    while (itr != pathInfoList.end()) {
        PathInfo* tmpPathInfo = *itr;
        
        // go through all the destination IP matches of the current PathInfo object
        list<ipaddr_n_t>::iterator dstIPItr = tmpPathInfo->dstIPAddrList.begin();
        while (dstIPItr != tmpPathInfo->dstIPAddrList.end()) {
           
            if (pkt->getDestIPAddress() == *dstIPItr) {
                return tmpPathInfo;
            }
            dstIPItr++;
        }
        
        // go through all the destination port matches of the current PathInfo object
        list<port_protocol_n_t>::iterator dstPortItr = tmpPathInfo->dstPortList.begin();
        while (dstPortItr != tmpPathInfo->dstPortList.end()) {
          
            port_protocol_n_t tmpPortProtocol = *dstPortItr;
            
            if (pkt->getDestPortNum() == tmpPortProtocol.port && 
                                    pkt->getProtocolNum() == tmpPortProtocol.protocol) {

                return tmpPathInfo;
            }
            dstPortItr++;
        }

        // go through all the source IP matches of the current PathInfo object
        list<ipaddr_n_t>::iterator srcIPItr = tmpPathInfo->srcIPAddrList.begin();
        while (srcIPItr != tmpPathInfo->srcIPAddrList.end()) {
           
            if (pkt->getSrcIPAddress() == *srcIPItr) {
                return tmpPathInfo;
            }
            srcIPItr++;
        }

        // go through all the source port matches of the current PathInfo object
        list<port_protocol_n_t>::iterator srcPortItr = tmpPathInfo->srcPortList.begin();
        while (srcPortItr != tmpPathInfo->srcPortList.end()) {
           
            port_protocol_n_t tmpPortProtocol = *srcPortItr;
            if (pkt->getSrcPortNum() == tmpPortProtocol.port && 
                                    pkt->getProtocolNum() == tmpPortProtocol.protocol) {
                return tmpPathInfo;
            }
            srcPortItr++;
        }
        
        itr++;

    }

    // No match found.
    return NULL;
}

/**
 * This function tries to match the url against the suffix and prefix patterns.
 * @return PathInfo associated with the URL or NULL if no match found.
 */
PathInfo* URLResolver::matchPrefixSuffix(const URL url) {

    list<PathInfo *>::iterator itr = pathInfoList.begin();

    while (itr != pathInfoList.end()) {

	    PathInfo* currPathInfo = (PathInfo *) (*itr);

        // Check if the first Hop OCD is active.  If not, do not even consider this PathInfo
        OCD* firstHopOCD = currPathInfo->getHop(0)->ocd;
        assert (firstHopOCD != NULL);
        if (firstHopOCD->isActive()) {

            list<string>::iterator prefixItr = currPathInfo->patternPrefixList.begin();
            while (prefixItr != currPathInfo->patternPrefixList.end()) {
                string currStr = *prefixItr;
                if ( isPrefix (url.c_str(), currStr.c_str()) ) {
                    /*
                    printf ("\n\n\n\n&&&&&&&&&&&&&& PREFIX MATCH %s: %s\n",
                            url.c_str(),
                            currStr.c_str()
                        );
                    */
                    return currPathInfo;
                }
                prefixItr++;
            }
        
            list<string>::iterator suffixItr = currPathInfo->patternSuffixList.begin();
            while (suffixItr != currPathInfo->patternSuffixList.end()) {
                string currStr = *suffixItr;
                if ( isSuffix (url.c_str(), currStr.c_str()) ) {
                    /*
                    printf ("\n\n\n\n&&&&&&&&&&&&&& SUFFIX MATCH %s: %s\n",
                            url.c_str(),
                            currStr.c_str()
                        );
                    */
                    return currPathInfo;
                }
                suffixItr++;
            }
        } else {
            PRINT_DEBUG0 ( DEBUG_LEVEL_MINIMAL,
                        "Ignoring PathInfo in matchPrefixSuffix because the first hop OCD is not active.\n"
                    );
        }
        itr ++;
    }
    return NULL;
}
/**
 * This functions does regular expression matching of the 
 * specified url against the various url pattern->overlaynetwork
 * mappings loaded from the config file. 
 * @return PathInfo associated with the URL or NULL if no match found.
 */
PathInfo* URLResolver::matchURLPattern(const URL url) {

    list<PathInfo *>::iterator itr = pathInfoList.begin();

    while (itr != pathInfoList.end()) {

	    PathInfo* currPathInfo = (PathInfo *) (*itr);
	
        list<regex_t *>::iterator rexItr = currPathInfo->patternList.begin();
        while (rexItr != currPathInfo->patternList.end()) {

            regex_t *currPattern = (*rexItr);
            
   	        regmatch_t *matched = new regmatch_t;
    	
	        if ( !regexec(currPattern , url.c_str() , 1 , matched , 0) 
		            && matched[0].rm_so == 0 
		            && matched[0].rm_eo == (int) strlen(url.c_str()) ) {
    
	            //Match
                PRINT_DEBUG1 (
                            DEBUG_LEVEL_VERBOSE,
                            "\t\t\tMatched URL %s.\n",
                            url.c_str()
                        );

                delete(matched);
	            return currPathInfo;
    	    }

            delete(matched);
            rexItr++;
        }
        
        itr ++;
    }
    return NULL;
}

/**
  * This function goes through all the PathInfo objects looking for Matches involving
  * destination IP addresses and port numbers.  If found, the proxy is passed the 
  * information about the rules.  The proxy will initialize the tunnel with these rules
  * (not necessarily immediately)
  */
void URLResolver::initPktCaptureRules() {
   
    // go through all the PathInfo objects
    list<PathInfo*>::iterator itr = pathInfoList.begin();
    while (itr != pathInfoList.end()) {
        PathInfo* tmpPathInfo = *itr;
        
        // go through all the destination IP matches of the current PathInfo object
        list<ipaddr_n_t>::iterator dstIPItr = tmpPathInfo->dstIPAddrList.begin();
        while (dstIPItr != tmpPathInfo->dstIPAddrList.end()) {
            
            proxy->addDstIPMatch (*dstIPItr);
            dstIPItr++;
        }
        
        // go through all the destination port matches of the current PathInfo object
        list<port_protocol_n_t>::iterator dstPortItr = tmpPathInfo->dstPortList.begin();
        while (dstPortItr != tmpPathInfo->dstPortList.end()) {
           
            proxy->addDstPortMatch (*dstPortItr);
            dstPortItr++;
        }

        // go through all the source IP matches of the current PathInfo object
        list<ipaddr_n_t>::iterator srcIPItr = tmpPathInfo->srcIPAddrList.begin();
        while (srcIPItr != tmpPathInfo->srcIPAddrList.end()) {
            
            proxy->addSrcIPMatch (*srcIPItr);
            srcIPItr++;
        }

        // go through all the source port matches of the current PathInfo object
        list<port_protocol_n_t>::iterator srcPortItr = tmpPathInfo->srcPortList.begin();
        while (srcPortItr != tmpPathInfo->srcPortList.end()) {
            
            proxy->addSrcPortMatch (*srcPortItr);
            srcPortItr++;
        }
        
        itr++;

    }
}
