/**
 *  \file URLResolver.h
 *  \brief The declaration of the URLResolver class.
*/

#ifndef URL_RESOLVER_H
#define URL_RESOLVER_H 1

#include <map>
#include <string>
#include <list>

#include "../common.h"

class OCD;
class ConfigFile;
class URLPattern;
class Proxy;
class PathInfo;
class IPPacket;

using namespace std;

/** This class resolves the URLs of DNS requests captured from
  * legacy applications.  For example, the legacy application
  * sends out a request for www.abc.com.i3.  This class decides 
  * which of the loaded OCDs should handle this request.
  * In this case, it may seem that decision is obvious: The i3 overlay
  * network needs to handle the request.  However, even if the decision
  * is obvious, someone needs to make it.  Things get complicated when
  * multiple i3 overlays are loaded in the same proxy.
  * <br>
  * Let us consider another example:  Suppose the url requested is just
  * www.abc.com.  Now which overlay should handle this packet?  Should 
  * it handled by an overlay at all; or just let it handled in the 
  * &quot;normal&quot; way?  This decision is also made by this object.
  *
  * The URLResolver now also chooses paths on the basis of destination IP addresses
  * and destination port numbers.  So the name URLResolver is no longer appropriate.
  * It should probably be called PathResolver or something like that.
  */
class URLResolver {
private:

    /** Pointer to the proxy */
    Proxy* proxy;
    
    /** Hash table containing urlSuffix to OCD mapings. */
    map<string, OCD*> map_urlSuffix_OCD;

    /** List of PathInfo objects */
    list<PathInfo *> pathInfoList;

    void initPktCaptureRules();

    /** The configuration file from which the URLResolver is initialized.
      */
    string pathInfoFileName;
    
public:
    URLResolver(Proxy *, char* pathInfoFileName);
    void addSuffixes(OCD *);
    PathInfo* resolveURL(const URL url);
    PathInfo* resolvePacket(IPPacket* pkt);
    
    OCD* lookupURLSuffix(string suffix);
    void resolvePathInfo();
    PathInfo* matchURLPattern(const URL url);
    PathInfo* matchPrefixSuffix(const URL url);
    void addMapping (char *suffix, OCD* ocd);

    void reloadConfigFile();
    void doPeriodicReload();
};


#endif

