/**
 *  \file InetFns.cpp
 *  \brief Funtions to lookup and manipulate network addresses.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <sys/types.h>

#if !defined (_WIN32)
	#include <netinet/in.h>
	#include <arpa/inet.h>
	#include <sys/utsname.h>
	#include <netdb.h>
	#include <sys/ioctl.h>
	#include <sys/sysctl.h>
	#include <net/if.h>
	#include <unistd.h>
#endif


#ifdef _WIN32
extern "C" {
	#include "netwrap.h"
}
	#include <Iphlpapi.h>
#endif

#ifdef __APPLE__
#include <ifaddrs.h>
#endif

#include "../debug.h"
#include "InetFns.h"
#include "utilFunctions.h"

#define MAX_NUM_INTERFACES	3
#define IFNAME_LEN		256



/***************************************************************************
 * 
 * Purpose: Get the IP address of an arbitrary machine
 *	given the name of the machine
 * 
 * Return: As an unsigned long in network byte order
 *
 **************************************************************************/
uint32_t name_to_addr(const char *name)
{
    int i;
    struct hostent *hptr = gethostbyname(name);
    if (!hptr) {
	PRINT_DEBUG1(DEBUG_LEVEL_MINIMAL, "gethostbyname(%s) failed", name);
    }
    else {
	for (i = 0; i < hptr->h_length/sizeof(uint32_t); i++) {
	    uint32_t addr = *((uint32_t *) hptr->h_addr_list[i]);
	    if (inet_addr(TRIVIAL_LOCAL_ADDR) != addr)
		return addr;
	}
    }
    return 0;
}

#ifdef __APPLE__
uint32_t getLocalAddress_BSD (bool *foundFlag) {

 	*foundFlag = false;
	uint32_t retAddr = 0;

	struct ifaddrs * ifap;
	if (getifaddrs (&ifap) != 0) {
		PRINT_DEBUG1 ( DEBUG_LEVEL_MINIMAL, "Unable to get list of interfaces: %s\n", strerror (errno));
		*foundFlag = false;
		return false;
	}

	struct ifaddrs* currInterface = ifap;
	while (currInterface != NULL) {
		struct sockaddr_in *tmpSock = (struct sockaddr_in *) currInterface->ifa_addr;
		if (currInterface->ifa_addr->sa_family == AF_INET) {
			PRINT_DEBUG2 (DEBUG_LEVEL_VERY_VERBOSE, "%s %s - ", currInterface->ifa_name, inet_ntoa (tmpSock->sin_addr));
			if ( currInterface->ifa_flags && IFF_UP) {
				PRINT_DEBUG0 (DEBUG_LEVEL_VERY_VERBOSE, " UP\n");
				if ( !isPrefix (currInterface->ifa_name, "lo") && !isPrefix (currInterface->ifa_name, "tun")) {
					*foundFlag = true;
					retAddr = tmpSock->sin_addr.s_addr;
					break;	
				}
			} else {
				PRINT_DEBUG0 (DEBUG_LEVEL_VERY_VERBOSE, " DOWN\n");
			}
		}
		currInterface = currInterface->ifa_next;
	}

	freeifaddrs (ifap);
		
	return retAddr;
} 
#endif // __APPLE__

#if !defined (__APPLE__) && !defined (_WIN32)
uint32_t getLocalAddress_linux (bool *foundFlag) {
    *foundFlag = false;
    
    int tempfd = socket(AF_INET, SOCK_DGRAM, 0);
    if (tempfd == -1) {
        PRINT_DEBUG1 ( DEBUG_LEVEL_MINIMAL, "Unable to open temporary socket for doing ioctl: %s.\n", strerror (errno));
        return 0;
    }
    
    // Go through the list of all interfaces present in this system.
    char tmpBuf [10000];
    struct ifconf tmp_ifconf;;
    tmp_ifconf.ifc_req = (struct ifreq *) tmpBuf; // The array of ifreqs is copied here
    tmp_ifconf.ifc_len = 10000; // available space in the array of ifreqs (in bytes)
    
    if (-1 != ioctl(tempfd, SIOCGIFCONF, (char *) &tmp_ifconf)) { // Get the list of interfaces.
        
        if (tmp_ifconf.ifc_len == 10000) {
            // Overflow
            PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL, "Error: buffer of ifreqs passed to SIOCGIFCONF too small.\n");
            close (tempfd);
            return 0;
        }
        int numInterfaces = tmp_ifconf.ifc_len / sizeof (struct ifreq);
        
        PRINT_DEBUG1 (DEBUG_LEVEL_VERY_VERBOSE, "Found %d interfaces:\n", numInterfaces);
        
        struct ifreq *curr_ifreq = (struct ifreq *) tmpBuf; // Point to the first interface.

        for (int interfaceCount = 0; interfaceCount < numInterfaces; interfaceCount ++) {
            // Go through each interface and check if it is up.
            struct sockaddr_in* tmpSockAddr = (struct sockaddr_in *) &(curr_ifreq->ifr_addr);
            PRINT_DEBUG2 (DEBUG_LEVEL_VERY_VERBOSE, "%s:%s - ", curr_ifreq->ifr_name, inet_ntoa(tmpSockAddr->sin_addr));
    
            // Check if the interface is up or not
            struct ifreq ifr;		
            strcpy(ifr.ifr_name, curr_ifreq->ifr_name);
            if (-1 != ioctl(tempfd, SIOCGIFFLAGS, (char *)&ifr)) {
                if (0 != (ifr.ifr_flags & IFF_UP)) {
                    PRINT_DEBUG0 (DEBUG_LEVEL_VERY_VERBOSE, " UP \n");
                    if ( ((strcmp (curr_ifreq->ifr_name, "lo") != 0)) && ! isPrefix (curr_ifreq->ifr_name, "tun")) {
                        // Not the loopback or tun interface.
                        close (tempfd);
                        *foundFlag = true;
                        return tmpSockAddr->sin_addr.s_addr;
                    }
               } else {
                    PRINT_DEBUG0 (DEBUG_LEVEL_VERY_VERBOSE, " DOWN \n");
                }
            } else {
                PRINT_DEBUG0 (DEBUG_LEVEL_VERY_VERBOSE, "\n");
            }
       
           
            curr_ifreq++; // Increment the pointer to the next interface
        }
    } else {
        PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL, "ioctl failed.\n");
    }
    
    close(tempfd); 

    *foundFlag = false;
    return 0;
    
}
#endif //Linux


#ifdef _WIN32
uint32_t getLocalAddress_windows (bool *foundFlag) {
    uint32_t addr;
    addr = 0;
    *foundFlag = false;
    
    PIP_ADAPTER_INFO pAdapterInfo;
    ULONG OutBufLen = 0;
    
    GetAdaptersInfo(NULL, &OutBufLen);
    pAdapterInfo = (PIP_ADAPTER_INFO)malloc(OutBufLen);
    GetAdaptersInfo(pAdapterInfo, &OutBufLen);
    
    PIP_ADAPTER_INFO pai = pAdapterInfo;
    while (pai) {
        if (pai->GatewayList.IpAddress.String[0] != 0) {
	        addr = inet_addr(pai->IpAddressList.IpAddress.String);
            *foundFlag = true;
        }
      pai = pai->Next;
    }
    free(pAdapterInfo);

    return addr;

}
#endif  //windows




/**
  * Returns the IP address of the first non loopback and non "tun" interface that is UP.
  * The value of foundFlag is set to true. 
  * If no such active interface is found, the value of foundflag is set to false.
  * The return value of the function is then undefined.
  * This function works in Linux, Mac and Windows.
  * @author dilip@cs.berkeley.edu
  */
uint32_t getLocalAddress (bool *foundFlag) {

#if defined (_WIN32) 
    return getLocalAddress_windows (foundFlag);
#elif defined (__APPLE__)
    return getLocalAddress_BSD (foundFlag);
#else
    return getLocalAddress_linux (foundFlag);
#endif
}


/***************************************************************************
 * 
 * Purpose: Get IP address of local machine
 * 
 * Return: As an unsigned long in network byte order
 *
 **************************************************************************/
uint32_t get_local_addr(void)
{

    bool foundFlag;
    uint32_t addr = getLocalAddress (&foundFlag);
    
    if (foundFlag) {
        return addr;
    } else {
        /* This is hopeless, return TRIVIAL_IP */
        return(inet_addr(TRIVIAL_LOCAL_ADDR));
    }
}

/**
  * Converts the specified ip address string (in A.B.C.D notation)
  * to an 32 bit integer in network byte order and places it into the
  * memory pointed to by ipAddrInt32
  * @return true if the conversion succeeded, else returns false.
  */
bool convert_ipaddr_str2int (const char* ipAddrStr, ipaddr_n_t* ipAddrInt) {

    struct in_addr tmpAddr;
    if (inet_aton (ipAddrStr, &tmpAddr) == 0) {
        // error in conversion
        PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
                    "convert_ipaddr_str2int32(): Invalid ip address string : %s.\n",
                    ipAddrStr
                );
        return false;
    
    } else {
        *ipAddrInt = tmpAddr.s_addr;
        return true;
    }
}

/**
  * Converts the specified ip addr in network byte order to a string of the 
  * form A.B.C.D and the result is placed in ipAddrStr.
  * @return true if the conversion is successful, else returns false.
  */
bool convert_ipaddr_int2str (ipaddr_n_t ipAddrInt, char* ipAddrStr) {
    
    struct in_addr tmpIPAddr;
    tmpIPAddr.s_addr = ipAddrInt;
    strcpy (ipAddrStr, inet_ntoa (tmpIPAddr));
    
    return true;
}

#ifdef _WIN32 
/**
 * Gets the name of the specified adapter as shown in the Network Connections window.
 */
bool getHumanReadableInterfaceName (char *adapterName, char* humanReadableName, DWORD* humanReadableNameLen) {
	char regPath [1000];
	sprintf (regPath, "HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Control\\Network\\{4D36E972-E325-11CE-BFC1-08002BE10318}\\%s\\Connection", adapterName);
	return readRegValue (regPath, "Name", humanReadableName, humanReadableNameLen);
}
#endif
