/**
 *  \file utilFunctions.cpp
 *  \brief general purpose utility function definitions.
*/

#include <stddef.h>
#include <ctype.h>

#if defined (_WIN32)
extern "C" {
	#include "netwrap.h"
}
#include <iphlpapi.h>

#else
	#include <arpa/inet.h>
#endif



#include "utilFunctions.h"
#include "string.h"
#include "StringTokenizer.h"

#include "InetFns.h"

#include "../common.h"

/**
  * Returns true if the specified string consists only of spaces and tabs.
  */
bool isBlank (const char *str) {

    int len = strlen (str);
    for (int i = 0; i < len; i++) {
        if (str[i] != ' ' && str[i] != '\t') {
            return false;
        }
    }
    return true;
}

/**
  * This function just strips the whitespace in the specified string.
  * @param str The string to stripped of whitespace.
  */
void stripWS(char* str) {
    char tstr[MAX_PARAM_LEN];
  
    int lindex = 0;
    while ( str[lindex] != 0 && isspace(str[lindex]))
        lindex++;

    int rindex = strlen(str)-1;
    while ( rindex >= 0 && isspace(str[rindex]) )
        rindex--;

    if ( lindex == strlen(str)) {
        strcpy(str,"");
        return;
    }

    str[rindex+1] = 0;
    strcpy(tstr,str+lindex);
    strcpy(str,tstr);
}



/**
  * Strips the specified suffix from str if it exists.
  * @return true if suffix was found and removed.  Else returns false.
  */
bool stripSuffix (char *str, const char *suffix) {
    
    int retVal = isSuffix (str, suffix);
    
    if (retVal == 1) {
        // suffix present; now remove it.
        int suffixLen = strlen (suffix);
        int strLen = strlen (str);
        str [strLen - suffixLen] = '\0';
        return true;
    }
    return false;
}

/**
  * This function converts the given character array string to lower
  * case and places the result in the destionation argument.
  * @param src The character array string to be converted to lower case.
  * @param dst The dstination array to hold the converted string.
  */
void toLowerCase(const char *src, char *dst) {

    int i, len;
    len = strlen(src);
    for(i=0; i < len; i++) {
        dst[i] = tolower(src[i]);
    }
    dst[i] = '\0';
}

/**
  * This function converts the given character array string to lower
  * case and returns the resultant string.
  * @param src The character array string to be converted to lower case.
  */
string toLowerCase(string src) {

    int i, len;
	const char *tmpSrc = src.c_str();

	len = strlen(tmpSrc);
	char *dst = new char[len + 1];
	
    for(i=0; i < len; i++) {
        dst[i] = tolower(tmpSrc[i]);
    }
    dst[i] = '\0';

	string retStr = dst;

	delete dst;

	return retStr;

}

#ifndef _WIN32
/**
  * Changes all the characters in the specified string to lower case.
  * @param str The string to be converted to lowercase.
  * This lowercase operation is done in place; there is no separate 
  * destination string.
  */
void strlwr(char *str) {
  
    uint i, len;

    len = strlen(str);
    
    for(i=0; i<len; i++) {
    
        str[i] = tolower(str[i]);
    }
}
#endif



/**
  * Returns the IP address of this host.
  * Probes eth, etc.
  * @param addressFound This flag is set to true by the function if a valid address
  * could be found.  If this flag is false, the return value of the function 
  * is junk.
  */
struct in_addr get_my_addr(bool* addressFound) {
  struct in_addr ia;
  ia.s_addr = 0;

  *addressFound = false;
  
#ifdef _WIN32
  PIP_ADAPTER_INFO pAdapterInfo;
  ULONG OutBufLen = 0;

  GetAdaptersInfo(NULL, &OutBufLen);
  pAdapterInfo = (PIP_ADAPTER_INFO)malloc(OutBufLen);
  GetAdaptersInfo(pAdapterInfo, &OutBufLen);

  PIP_ADAPTER_INFO pai = pAdapterInfo;
  while (pai) {
    if (pai->GatewayList.IpAddress.String[0] != 0) {
      ia.s_addr = ntohl(inet_addr(pai->IpAddressList.IpAddress.String));
      //printf("\n\n\n\t\t\t--->Address obtained\n\n");
      *addressFound = true;
      break;
    }
    pai = pai->Next;
  }
  free(pAdapterInfo);
#else

  ia.s_addr = ntohl(get_local_addr());
  if (ia.s_addr == ntohl (inet_addr(TRIVIAL_LOCAL_ADDR))) {
    // We are not interested in the trivial local loopback address.
    *addressFound = false;
  } else {
    *addressFound = true;
  }
#endif
  return ia;

}

/**
  * Used for calculating the checksum of ip-headers
  */
unsigned short in_cksum (unsigned short *addr, int len) {
    register int nleft = len;
    register unsigned short *w = addr;
    register int sum = 0;
    unsigned short answer = 0;

    while (nleft > 1) {
        sum += *w++;
        nleft -= 2;
    }

    if (nleft == 1) {
        *(u_char *) (&answer) = *(u_char *) w;
        sum += answer;
    }

    sum = (sum >> 16) + (sum & 0xffff);
    sum += (sum >> 16);
    answer = ~sum;

    return (answer);
}

/**
  * Returns 1 if prefix is a prefix of masterStr.
  * Else return 0.
  */
int isPrefix ( const char* masterStr, const char* prefix) {
    int masterStrLen = strlen (masterStr);
    int prefixLen = strlen (prefix);

    if (prefixLen > masterStrLen) {
        return 0;
    }

    for (int i = 0; i < prefixLen; i ++) {
        if ( masterStr[i] != prefix[i]) {
            return 0;
        }
    }

    return 1;
}

/**
  * Returns 1 if suffix is a suffix of masterStr.
  * Else returns 0.
  */
int isSuffix ( const char* masterStr, const char* suffix) {

    int masterStrLen = strlen(masterStr);
    int suffixLen = strlen(suffix);

    if (suffixLen > masterStrLen) {
        return 0;
    }

    int i = masterStrLen - 1;
    int j = suffixLen - 1;
    
    while (i >= 0 && j >= 0) {
   
        if (tolower(masterStr[i]) != tolower(suffix[j])) {
            return 0;
        }
        i--;
        j--;
    }

    return 1;
}


/** Generates a random string and places it in buf.
  * A pointer to buf is returned.
  */
char *generateRandomStr (char *buf) {

    buf[0] = '\0';
    char numStr[100];
    
    sprintf (numStr, "%d", rand());
    strcat (buf, numStr);

    return buf;
}

/**
  * Removes any \n s at the end of the given string.
  */
void chomp (char *s) {
    int len = strlen (s);
    for (int i = len - 1; i >= 0; i--) {
        if ( s[i] == '\n' || s[i] == '\r')  {
            s[i] = '\0';
        } else {
            break;
        }
    }
}

/**
  * Counts the number of occurrences of ch in str.
  */
int strCountCh (const char *str, char ch) {

    int tmpCount = 0;
    for (int i = 0; i < strlen (str); i++) {
        if (str[i] == ch) {
            tmpCount++;
        }
    }
    return tmpCount;
}


/**
  * This function returns true if this machine is big endian.
  */
bool isBigEndian () {

    uint16_t testNum = 0x0102;
    uint8_t* ptr1 = (uint8_t *) &testNum;
    uint8_t* ptr2 = ptr1 + 1;
    if (*ptr1 == 0x01 && *ptr2 == 0x02) {
        return true;
    } else {
        return false;
    }
}

/** This structure is used by ntoh64 and hton64 */
typedef struct {
    uint8_t a[8];
} uint64_struct;

/** Converts the 64 bit number from network byte order to host byte order. */
uint64_t ntoh64 (uint64_t num) {

    // If this machine is big-endian, i.e. in network byte order,
    // we don't need to do anything.
    if (isBigEndian()) {
        return num;
    } else {
        // This machine is little-endian.
        // We need to convert the num to little endian.

        uint64_struct* tmpNum = (uint64_struct *) &num;

        uint64_struct retNum;
    
        for (int index = 0; index < 8; index++) {
            retNum.a[index] = tmpNum->a[7-index];
        }

        uint64_t ret64 = * ( (uint64_t *) &retNum);
        return ret64;
    }
}

/** Converts the 64 bit number from host byte order to network byte order. */
uint64_t hton64 (uint64_t num) {

    // If this machine is already big-endian, i.e. in network byte order,
    // we don't need to do anything.
    if (isBigEndian()) {
        return num;
    } else {
        // This machine is little endian.
        // We need to convert the num to network byte order, i.e. big endian.

        uint64_struct* tmpNum = (uint64_struct *) &num;

        uint64_struct retNum;
  
        for (int index = 0; index < 8; index++) {
            retNum.a[index] = tmpNum->a[7-index];
        }

        uint64_t ret64 = * ( (uint64_t *) &retNum);
        return ret64;
    }
}

#ifdef _WIN32
LPCSTR GetErrorMsgStr (LPCSTR lpMsgBuf) {
	FormatMessage(
		      FORMAT_MESSAGE_FROM_SYSTEM |
		      FORMAT_MESSAGE_IGNORE_INSERTS,
		      NULL, GetLastError(), 0, (LPSTR) & lpMsgBuf, 0,
		      NULL);
	return lpMsgBuf;
}
#endif


#ifdef _WIN32
bool readRegValue (const char *regPath, const char* valName, char* valStr, DWORD* valStrLen) {

	stack<HKEY> keyStack;
	StringTokenizer st (regPath, "\\");

	HKEY currParentKey;
	string currNode;
	if (st.hasMoreTokens()) {
		currNode = st.nextToken();
		if (currNode == string ("HKEY_LOCAL_MACHINE")) {
			currParentKey = HKEY_LOCAL_MACHINE;	
		} else {
			return FALSE;
		}
	}
	BOOL okFlag = true;
	HRESULT res;
	while (st.hasMoreTokens()) {
		HKEY currChildKey;
		currNode = st.nextToken();
		res = RegOpenKeyEx( 
							 currParentKey,
							 currNode.c_str(),
							 0, /* Reserved must be 0 */
							 KEY_READ,
							 &currChildKey
						);
		if (res != ERROR_SUCCESS) {
			okFlag = false;
			break;
		} else {
			keyStack.push (currChildKey);
			currParentKey = currChildKey;
		}
	}

	if (okFlag) {

		DWORD valType;
		res	= RegQueryValueEx(
						currParentKey,
						valName,
						NULL,
						&valType,
						(BYTE*) valStr,
						valStrLen
					);
		if (res != ERROR_SUCCESS) {
			okFlag = false;
		}
	}

	while ( ! keyStack.empty()) {
		HKEY tmpKey = keyStack.top();
		keyStack.pop();
		RegCloseKey (tmpKey);
	}
	return okFlag;
}

/**
 * Makes the specified key (named keyName)
 * accessible for read/write by Everyone.
 */
BOOL MakeRegKeyAccessible (HKEY key, const char *keyName) {

	PSECURITY_DESCRIPTOR pPipeSD = NULL;

	// security inits

	// alloc & init SD
	if ( ! ( pPipeSD = ( PSECURITY_DESCRIPTOR) 
		( malloc ( SECURITY_DESCRIPTOR_MIN_LENGTH)) ) ) {
		PRINT_INFO1 ( INFO_LEVEL_WARNING,
						"Unable to malloc Security Descriptor for %s.\n", 
						keyName);

		return FALSE;
	}

	if ( ! InitializeSecurityDescriptor ( pPipeSD, SECURITY_DESCRIPTOR_REVISION) ) {
		PRINT_INFO1 ( INFO_LEVEL_WARNING,
						"Unable to initialize Security Descriptor for %s.\n", 
						keyName);

		return FALSE;

	}

	// set NULL DACL on the SD
	if ( ! SetSecurityDescriptorDacl ( pPipeSD, TRUE, ( PACL) NULL, FALSE) ) {
		PRINT_INFO1 ( INFO_LEVEL_WARNING,
						"Unable to set the NULL DACL for Security Descriptor for %s.\n", 
						keyName);
		return FALSE;

	}
  
    HRESULT res =  RegSetKeySecurity(
								key,
								DACL_SECURITY_INFORMATION,
								pPipeSD								
					);

	if (res != ERROR_SUCCESS) {
		PRINT_INFO1 ( INFO_LEVEL_WARNING,
						"Unable to make %s accessible to all users.\n", 
						keyName
					);
	
		return FALSE;
	}

	// We are successful!
	return TRUE;
}

#endif

