/**
 * \file This file contains functions to start, stop, suspend 
 * the proxy service.
 */
#include <afx.h>
#include <afxwin.h>
#include <afxext.h>
#include <afxdtctl.h>		// MFC support for Internet Explorer 4 Common Controls
#ifndef _AFX_NO_AFXCMN_SUPPORT
#include <afxcmn.h>			// MFC support for Windows Common Controls
#endif // _AFX_NO_AFXCMN_SUPPORT

//#include <windowsx.h>
//#include <commctrl.h>
//#include <objbase.h>
//#include <shlobj.h>

// Header file for the recursive registry delete function SHDeleteKey
#include <shlwapi.h>


#include "debug.h"
#include "utilFunctions.h"
#include "StringTokenizer.h"

#include <string>
using namespace std;

#include "GenFunctions.h"
#include "ProxyService.h"
#include "ProxyMonitorApp.h"
#include "OCDStatus.h"
#include "RegistryParams.h"

/** reference to the main application */
extern ProxyMonitorApp theApp;


ProxyService::ProxyService() {

}

/**
 * Starts the proxy service.
 * @param coldStartFlag This flag is set to true to indicate that
 * the proxy should do a cold start.
 * @return TRUE if successful.  Else returns false.
 */
BOOL ProxyService::StartProxy (BOOL coldStartFlag) {

    BOOL            retValue;
    SC_HANDLE       hService, hSCM;
    SERVICE_STATUS  ss;
    int             retcode;

    DWORD dwStartTime = GetTickCount();

	// Open the service manager
    hSCM = OpenSCManager("\\\\.", NULL, SC_MANAGER_CONNECT);
    if (!hSCM) {
		ShowErrorMessage (NULL);
		return FALSE;
    }

	// Try to open the service
    hService = OpenService(hSCM, OCALA_SERVICE_NAME,
						   SERVICE_QUERY_STATUS | SERVICE_START |
							SERVICE_STOP | SERVICE_USER_DEFINED_CONTROL);

	
    if (hService != NULL) {
		// The proxy service was successfully opened.  
		retValue = FALSE;
		theApp.BeginWaitCursor();

		if (coldStartFlag) {
			// Do a cold start of the proxy.	
			char *svcargs[3] = { "-c" };
			retcode = StartService(hService, 1, (LPCTSTR *) & svcargs);
	    
		} else {
			// Do a normal start
			retcode = StartService(hService, 0, NULL);
		}
			
		if (retcode) {
			// Start was successful
			// Wait for the proxy to completely start running or
			// till we wait for too much time.
			while (QueryServiceStatus(hService, &ss)) {
				if ( ss.dwCurrentState == SERVICE_RUNNING ) {
					retValue = TRUE;
					break;
				}
	      
				if ( GetTickCount() - dwStartTime > SERVICE_OP_TIMEOUT) {
					// We waited too long for the service status 
					// to change.
					break;
				}
				// Sleep for some time and then check if the service
				// status has changed to running.
				Sleep( ss.dwWaitHint );
			}
		} else {
			ShowErrorMessage(NULL);
			retValue = FALSE;
		}
		CloseServiceHandle(hService);
		CloseServiceHandle(hSCM);

		theApp.RestoreWaitCursor();
		return retValue;
	
	} else {
		// Unable to open proxy service.
		ShowErrorMessage (NULL);
	    CloseServiceHandle(hSCM);
		retValue = FALSE;
		return retValue;

	}
}


/**
 * Checks if the proxy service is running.
 * @return True if the proxy service is running. Else false.
 */
BOOL ProxyService::IsRunning () {

	return	IsRunningStatusCode() &&
			(IsRunningAsService() || IsRunningAsExec())
		;
}

/**
 * Is the Proxy running according the status code in the registry?
 */
BOOL ProxyService::IsRunningStatusCode () {
	// What should be returned
	BOOL retCode = TRUE;

	LONG res;
	DWORD valType;
	DWORD stringValSize = MAX_REGISTRY_STR_VALUE_LEN;
	DWORD dwordVal;
	DWORD dwordValSize = sizeof (DWORD);

	// Open the Software Key
	HKEY keySoftware;
	res = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
							 "Software",
							 0, /* Reserved must be 0 */
							 KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS,
							 &keySoftware
	
					);
	if (res != ERROR_SUCCESS) {
		PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL,
						"Unable to open the HKEY_LOCAL_MACHINE/Software key.\n"
					);
		ShowErrorMessage (NULL);
		return FALSE;
	}

	// Open the OCALA key.
	HKEY keyOCALA;	
	res = RegOpenKeyEx(		keySoftware,
							 REGISTRY_APP_NAME,
							 0, /* Reserved must be 0 */
							 KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS,
							 &keyOCALA
						);

	if (res != ERROR_SUCCESS) {

		PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
						"Unable to open the HKEY_LOCAL_MACHINE/Software key/%s key.  OCALA probably never started.\n",
						REGISTRY_APP_NAME
					);

		//ShowErrorMessage (NULL);
		RegCloseKey (keySoftware);
		return FALSE;
	}

	// Get the status code
	res = RegQueryValueEx(	
							keyOCALA,
							REGISTRY_STATUS_CODE,
							NULL,
							&valType,
							(BYTE *) &dwordVal,
							&dwordValSize
							);

	if (res != ERROR_SUCCESS) {
		retCode = FALSE;
	} else {
		if (dwordVal == 1) {
			retCode = TRUE;
		} else {
			retCode = FALSE;
		}
	}

	RegCloseKey (keyOCALA);
	RegCloseKey (keySoftware);
	
	return retCode;
}

BOOL ProxyService::IsRunningAsExec() {
	// If the OCALAproxy executable is running, we are done.
	HWND testWnd = FindWindow (_T ("OCALAProxy"), NULL);
	if (testWnd != NULL) {
		return TRUE;
	} else {
		return FALSE;
	}
}

BOOL ProxyService::IsRunningAsService () {
	
    SC_HANDLE       hService, hSCM;
    SERVICE_STATUS  ss;

    DWORD dwStartTime = GetTickCount();

	// Open the service manager
    hSCM = OpenSCManager("\\\\.", NULL, SC_MANAGER_CONNECT);
    if (!hSCM) {
		ShowErrorMessage (NULL);
		return FALSE;
    }

	// Try to open the service
    hService = OpenService(hSCM, OCALA_SERVICE_NAME,
							SERVICE_QUERY_STATUS
							);
	
    if (hService != NULL) {
		// The proxy service was successfully opened.  
		
		// Get the service status
		QueryServiceStatus(hService, &ss);
		if ( ss.dwCurrentState == SERVICE_RUNNING ) {
			return TRUE;
		} else {
			return FALSE;
		}
	
	} else {
		return FALSE;
	}
	return FALSE;
}



/**
 * Stops the proxy service.
 * @return TRUE if successful.  Else returns false.
 */
BOOL ProxyService::StopProxy () {

	if (IsRunningAsService()) {

		BOOL            retValue;
		SC_HANDLE       hService, hSCM;
		SERVICE_STATUS  ss;

		DWORD dwStartTime = GetTickCount();

		// Open the service manager
		hSCM = OpenSCManager("\\\\.", NULL, SC_MANAGER_CONNECT);
		if (!hSCM) {
			ShowErrorMessage (NULL);
			return FALSE;
		}

		// Try to open the service
		hService = OpenService(hSCM, OCALA_SERVICE_NAME,
						   SERVICE_QUERY_STATUS | SERVICE_START |
							SERVICE_STOP | SERVICE_USER_DEFINED_CONTROL);

	
		if (hService != NULL) {
			// The proxy service was successfully opened.  
			retValue = FALSE;

			theApp.BeginWaitCursor();

			if (!ControlService(hService, SERVICE_CONTROL_STOP, &ss)) {
				// Unable to control the service and set it to stop.		
				ShowErrorMessage(NULL);

			} else {
				// Service accepted the command to stop.
				// Now we wait for it to stop.
				while (QueryServiceStatus(hService, &ss)) {
					if ( ss.dwCurrentState == SERVICE_STOPPED ) {
						// Service has finally stopped.
						retValue = TRUE;
						break;
					}	

					if ( GetTickCount() - dwStartTime > SERVICE_OP_TIMEOUT ) {
						// Tired of waiting for the service to stop.
						break;
					}

					Sleep( ss.dwWaitHint );
				}
			}

			CloseServiceHandle(hService);
			CloseServiceHandle(hSCM);

			theApp.RestoreWaitCursor();
			return retValue;
	
	
		} else {
			// Unable to open proxy service.
			ShowErrorMessage (NULL);
			CloseServiceHandle(hSCM);
			retValue = FALSE;
			return retValue;

		}

	} else if (IsRunningAsExec ()) {
		HWND tmpWnd = ::FindWindow( _T("OCALAProxy"), NULL);
		if (tmpWnd != NULL) {
			SendMessage (tmpWnd, WM_DESTROY, 0, 0 );
			return TRUE;
		} else {
			PRINT_DEBUG0 ( DEBUG_LEVEL_MINIMAL, "Unable to find OCALAProxy window.\n");
			return FALSE;
		}
	}

	PRINT_DEBUG0 ( DEBUG_LEVEL_MINIMAL, "Error: This code should not have been reached.\n" );

}


/**
 * This function reads the registry and extracts the status
 * of the proxy and the various loaded OCDs.
 * After reading the values, it clears the value.  If the OC-Ds are active,
 * they will fill the value again.
 * @param ocdStatusList The details about the OCDs are put into this list.
 * @param proxyStatusMsg Any status message general to the proxy is dumped 
 * into this variable.
 * @param proxyStatusCode The current status code of the proxy is put into
 * this variable.
 * @return TRUE if the status information was successfully read in from the
 * registry.  Else returns FALSE.
 */
BOOL ProxyService::RefreshProxyStatus (
					 DWORD* proxyStatusCode, 
					 string* proxyStatusMsg, 
					 list<OCDStatus*>* ocdStatusList) {

	// What should be returned
	BOOL retCode = TRUE;

	LONG res;
	DWORD valType;
	char stringVal [MAX_REGISTRY_STR_VALUE_LEN];
	DWORD stringValSize = MAX_REGISTRY_STR_VALUE_LEN;
	DWORD dwordVal;
	DWORD dwordValSize = sizeof (DWORD);

	// Open the Software Key
	HKEY keySoftware;
	res = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
							 "Software",
							 0, /* Reserved must be 0 */
							 KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS,
							 &keySoftware
	
					);
	if (res != ERROR_SUCCESS) {
		PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL,
						"Unable to open the HKEY_LOCAL_MACHINE/Software key.\n"
					);
		ShowErrorMessage (NULL);
		if (! this->IsRunning()) {
			// Proxy is not running.  Set all OCDs to their no service state.
			blankOCDStatus();		
		} 

		return FALSE;
	}

	// Open the OCALA key.
	HKEY keyOCALA;	
	res = RegOpenKeyEx(		keySoftware,
							 REGISTRY_APP_NAME,
							 0, /* Reserved must be 0 */
							 KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS,
							 &keyOCALA
						);

	if (res != ERROR_SUCCESS) {

		PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
						"Unable to open the HKEY_LOCAL_MACHINE/Software key/%s key.  OCALA probably never started.\n",
						REGISTRY_APP_NAME
					);

		//ShowErrorMessage (NULL);
		RegCloseKey (keySoftware);
		if (! this->IsRunning()) {
			// Proxy is not running.  Set all OCDs to their no service state.
			blankOCDStatus();		
		} 
		return FALSE;
	}

	// Get the proxy's statusCode and statusMsg values.
	// Get the status msg
	res = RegQueryValueEx(
						keyOCALA,
						REGISTRY_STATUS_MSG,
						NULL,
						&valType,
						(BYTE*) stringVal,
						&stringValSize
					);

	if (res != ERROR_SUCCESS) {
		//ShowErrorMessage (NULL);
		*proxyStatusMsg = "Inactive";
	} else {
		*proxyStatusMsg = stringVal;
	}

	// Get the status code
	res = RegQueryValueEx(	
							keyOCALA,
							REGISTRY_STATUS_CODE,
							NULL,
							&valType,
							(BYTE *) &dwordVal,
							&dwordValSize
							);
	if (res != ERROR_SUCCESS) {
		//ShowErrorMessage (NULL);
		*proxyStatusCode = -1;
	} else {
		*proxyStatusCode = dwordVal;
	}

/*	DILIP: May 2: while changing to menu based status
	if (! this->IsRunning()) {
		// Proxy is not running.  Set all OCDs to their no service state.
		blankOCDStatus();
		return TRUE;
	} 


	// Proxy is running
*/
	

	// All the individual OCDs are under the key OCD.
	// Now open that key.
	HKEY keyOCDs;	
	res = RegOpenKeyEx( keyOCALA,
							"OCDs",
							 0, /* Reserved must be 0 */
							 KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS,
							 &keyOCDs
						);

	if (res == ERROR_SUCCESS) {
			
		// Clear any existing status info contained in the list.
		ocdStatusList->clear();

		int currOCDIndex = 0;
		char ocdIdBuf[MAX_REGISTRY_STR_VALUE_LEN];
		DWORD ocdIdBufLen = MAX_REGISTRY_STR_VALUE_LEN;
		while (TRUE) {
			// Note: The name of the sub key is the ocd's id.
			ocdIdBufLen = MAX_REGISTRY_STR_VALUE_LEN;
			res = RegEnumKeyEx(
							keyOCDs,
							currOCDIndex,
							ocdIdBuf,
							&ocdIdBufLen,
							NULL,
							NULL,
							NULL,
							NULL
					);
			if (res != ERROR_SUCCESS && res != ERROR_NO_MORE_ITEMS) {
				// Some error occurred, so we get out of the loop.		
				PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL,
						"Error while iterating through the OCDs.\n"						
					);

				ShowErrorMessage(NULL);
				retCode = FALSE;
				break;
			} else if (res == ERROR_NO_MORE_ITEMS) {
				// We are done enumerating all OCDs. 
				// So we exit out of the loop.
				break;
			}
			currOCDIndex++;

		
			// Open the key associated with the current OCD.
			HKEY keyCurrOCD;	
			res = RegOpenKeyEx( keyOCDs,
							 ocdIdBuf,
							 0, /* Reserved must be 0 */
							 KEY_ALL_ACCESS,
							 &keyCurrOCD
						);

			if (res != ERROR_SUCCESS) {
				PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
						"Error while opening key associated with key of ocd %s.\n",
						ocdIdBuf
					);

				ShowErrorMessage (NULL);
				continue;
			}


			string currOCDId = ocdIdBuf;

			// Read in all the data associated with the OCD.
			// Get the status msg
			stringValSize = MAX_REGISTRY_STR_VALUE_LEN;
			res = RegQueryValueEx(
						keyCurrOCD,
						REGISTRY_STATUS_MSG,
						NULL,
						&valType,
						(BYTE*) stringVal,
						&stringValSize
					);

			string currOCDStatusMsg;
			if (res != ERROR_SUCCESS) {
				currOCDStatusMsg = "Inactive";
			} else {
				currOCDStatusMsg = stringVal;
			}
			// Reset the status msg. It should be set again, if the
			// OC-D is active
			/* DILIP: If the proxy is running, the OCD is supposed to be working too.
			res = RegSetValueEx(
							keyCurrOCD,
							REGISTRY_STATUS_MSG,
							0,
							REG_SZ,
							(BYTE*) "Inactive",
							9
						);
			*/

			// Get the status code
			res	= RegQueryValueEx(
						keyCurrOCD,
						REGISTRY_STATUS_CODE,
						NULL,
						&valType,
						(BYTE*) &dwordVal,
						&dwordValSize
					);

			DWORD currOCDStatusCode;
			if (res != ERROR_SUCCESS) {
				currOCDStatusCode = -1;
			} else {
				currOCDStatusCode = dwordVal;
			}
/**
			// Reset the status code.  It will be set again if the OCD is active
			DWORD keyVal = 0;
			res = RegSetValueEx(
							keyCurrOCD,
							REGISTRY_STATUS_CODE,
							0,
							REG_DWORD,
							(BYTE*) &keyVal,
							sizeof (DWORD)
						);
			if (res != ERROR_SUCCESS) {
				PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL,
						"Error while resetting status code in registry.\n"
						);
			}
*/

			// Get the description
			stringValSize = MAX_REGISTRY_STR_VALUE_LEN;
			res	= RegQueryValueEx(
						keyCurrOCD,
						REGISTRY_OCD_DESCRIPTION,
						NULL,
						&valType,
						(BYTE*) stringVal,
						&stringValSize
					);

			string currOCDDescription;
			if (res != ERROR_SUCCESS) {
				currOCDDescription = "Unknown";
			} else {
				currOCDDescription = stringVal;
			}

			// Get the OCD type
			stringValSize = MAX_REGISTRY_STR_VALUE_LEN;
			res	= RegQueryValueEx(
						keyCurrOCD,
						REGISTRY_OCD_TYPE,
						NULL,
						&valType,
						(BYTE*) stringVal,
						&stringValSize
					);

			string currOCDType;
			if (res != ERROR_SUCCESS) {
				currOCDType = ERROR_MSG_READING_REGISTRY;
			} else {
				currOCDType = stringVal;
			}


			// Get the comma separate icon list
			stringValSize = MAX_REGISTRY_STR_VALUE_LEN;
			res	= RegQueryValueEx(
						keyCurrOCD,
						REGISTRY_OCD_ICON_LIST,
						NULL,
						&valType,
						(BYTE*) stringVal,
						&stringValSize
					);

			string currOCDIconList;
			if (res != ERROR_SUCCESS) {
				currOCDIconList = ERROR_MSG_READING_REGISTRY;
			} else {
				currOCDIconList = stringVal;
			}

			stringValSize = MAX_REGISTRY_STR_VALUE_LEN;
			res	= RegQueryValueEx(
						keyCurrOCD,
						REGISTRY_OCD_CURR_ICON_NAME,
						NULL,
						&valType,
						(BYTE*) stringVal,
						&stringValSize
					);
			

			string currOCDIconName;
			if (res != ERROR_SUCCESS) {
				currOCDIconName = ERROR_MSG_READING_REGISTRY;
			} else {
				currOCDIconName = string(stringVal);
			}
/*
			// Reset the icon to the disconnected icon.
			sprintf (stringVal, "%s_disconnect.ico", toLowerCase(currOCDType).c_str());
			// If the OC-D is active it will be set to active again very soon.
			res = RegSetValueEx(
							keyCurrOCD,
							REGISTRY_OCD_CURR_ICON_NAME,
							0,
							REG_SZ,
							(BYTE*) stringVal,
							sizeof (stringVal) + 1
						);
	
*/


			// Get the balloon msg if any and then reset it to ""
			// Do this only if the proxy is running.
			string currOCDBalloonMsg = "";
			if (IsRunning()) {
				stringValSize = MAX_REGISTRY_STR_VALUE_LEN;
				res	= RegQueryValueEx(
						keyCurrOCD,
						REGISTRY_OCD_BALLOON_MSG,
						NULL,
						&valType,
						(BYTE*) stringVal,
						&stringValSize
					);

				if (res == ERROR_SUCCESS) {
					StringTokenizer st (stringVal, ":");
					
					if (st.countTokens() >= 2) {
						//Timestamp present.
						string tmpTimestampStr = st.nextToken();
						string tmpMsg = st.nextToken();
						time_t tmpTimestamp = (time_t) strtol (tmpTimestampStr.c_str(), NULL, 10);
						if (errno != EINVAL && errno != ERANGE) {
							time_t tmpCurrTime = time (NULL);
							if (tmpCurrTime - tmpTimestamp < 60) { // 1 minute
								// show the message only if the balloon message was 
								// set less than 1 minute ago.
								currOCDBalloonMsg = tmpMsg; //don't show the timestamp part
							}
						} 
					} else {
						//No timestamp.  Just go ahead display the message
						currOCDBalloonMsg = stringVal;
					}
				}
				// Reset the balloon msg so that we display the balloon
				// only once.
				res = RegSetValueEx(
							keyCurrOCD,
							REGISTRY_OCD_BALLOON_MSG,
							0,
							REG_SZ,
							(BYTE*) "",
							1
						);
	
			} else {
				// We definitely don't want to show any leftover messages
				// from a previous run when the proxy is inactive.
				currOCDBalloonMsg = "";
			}
			RegCloseKey (keyCurrOCD);

			OCDStatus* currOCDStatus;

			// Check if an ocd with the specified id already exists in the map.
			currOCDStatus = getOCDStatus (currOCDId);
			if (currOCDStatus == NULL) {
				// This is the first time an OCD with this id is being seen.
				// Create a new OCDStatus Object for this ocd.
				currOCDStatus = new OCDStatus(
													currOCDId,
													currOCDType, 
													currOCDDescription,
													currOCDStatusMsg,
													currOCDStatusCode,
													currOCDIconList,
													currOCDIconName
											);
				addOCDStatus (currOCDStatus);
			} else {
				// Just update the existing OCDStatus object.
				currOCDStatus->update (
													currOCDDescription,
													currOCDStatusMsg,
													currOCDStatusCode,
													currOCDIconList,
													currOCDIconName,
													true
										);
			}
			if (currOCDBalloonMsg != "") {
				currOCDStatus->showBalloon (currOCDBalloonMsg);
			}

			ocdStatusList->push_back (currOCDStatus);

  		} // while (TRUE)
		
		RegCloseKey (keyOCDs);
	}
	

	RegCloseKey (keyOCALA);
	RegCloseKey (keySoftware);
	

	
	return retCode;
}

/** 
 * Adds the specified OCDStatus object to the map of OCDStatus objects
 * keyed by the ocd id.
 * @return true if the OCDStatus object was successfully added.
 * Returns false if the object was not added, probably because an
 * object of the same id already existed.
 */
bool ProxyService::addOCDStatus (OCDStatus *ocdStatus) {
	if (getOCDStatus (ocdStatus->id) != NULL) {
		// already exists
		return false;
	} else {

		map_OCDStatus_id.insert (pair<string, OCDStatus*>(ocdStatus->id, ocdStatus));
		return true;
	}
}

/**
 * Returns the OCDStatus object with the specified id if it exists.
 * Else returns NULL.
 */
OCDStatus* ProxyService::getOCDStatus (string ocdID) {
	map<string, OCDStatus*>::iterator itr = map_OCDStatus_id.find (ocdID);
	if (itr != map_OCDStatus_id.end()) {
		return itr->second;
	} else {
		// Not found
		return NULL;
	}
}

/** Destructor */
ProxyService::~ProxyService() {

	cleanUp();

}

void ProxyService::cleanUp() {
	// Destroy all the OCDStatus objects.
	map<string, OCDStatus*>::iterator itr = map_OCDStatus_id.begin();
	while (itr != map_OCDStatus_id.end()) {
		delete itr->second;
		itr++;
	}
	map_OCDStatus_id.clear();
}

/** Blank the status of all OC-Ds.  Set their icon to disconnect. */
void ProxyService::blankOCDStatus() {
	map<string, OCDStatus*>::iterator itr = map_OCDStatus_id.begin();
	while (itr != map_OCDStatus_id.end()) {
		OCDStatus* tmpOCDStatus = itr->second;
		tmpOCDStatus->statusCode = 0;	//inactive
		tmpOCDStatus->update (
								string(""), 
								string ("Not active"), 
								-1, 
								string (""),
								toLowerCase(tmpOCDStatus->type) + string("_nosrv.ico"),
								true
							);
		itr++;
	}
}