/**
 *  \file win-main.cpp
 *  \brief This is the file containing the starting point when the proxy
 *  is used in Windows.  The progrm also has command line options for 
 *  installing/uninstalling/starting/stopping the proxy as a Windows service.
 */


#include "../common.h"
#include "../debug.h"
#include "../generic_proxy.h"
#include "../Proxy.h"
#include "../Tunnel.h"

#include "resource.h"

#include <windows.h>
#include <winsvc.h>
#include <windowsx.h>
#include <shellapi.h>
#include <sys/stat.h>
#include "win-main.h"
#include "win-service.h"


#include <pbt.h>
#include "resource.h"

void startGenericProxy_Windows();

HWND hwndMain;	//Main window handle

//#include <stdio.h>
//FILE* debugFD;

/** This should be the same as SERVICE_NAME defined in win-service.cpp.
 * For some vague reason, I can't use an extern here.
 */
char serviceName1[255] = GENERIC_PROXY;  

WCHAR **global_argv;
int global_argc;

/** This function exists only in order to resume the
 * proxy in a new thread.
 */
void * resume_generic_proxy(void *dummyArgs) {

    resumeGenericProxy();
	return NULL;
}

void usage(char *progName) {

    printf("\n%s options\n\n", progName);

    printf("Options when the proxy is used as a windows service\n");
    printf("\t-I Install the proxy as a windows service\n");
    printf("\t-R Uninstall the proxy windows service\n");
    printf("\t-S Normal Start the proxy windows service (Tries to reconstruct state from previous run)\n");
    printf("\t-C Cold Start the proxy windows service (Start afresh)\n");
    printf("\t-E Stop the running proxy windows service\n");
    
    printf("Options when proxy is invoked as a normal program\n");
    printf("\t-c Cold Start the proxy (Start afresh)\n");
    printf("\tIn the default starting mode, the proxy tries to reconstruct state from its last run.\n");
    

}

/**
 * This is the function which does all the work when the proxy is invoked
 * from the command line.
 * Based on the command line arguments, it can install/uninstall/start/stop the
 * proxy as a windows system service or start it as a plain command line 
 * program.
 * For want of a better name, this all important function is just called main1.
 */
int main1 (int argc, char *argv[]) {
    
    int i;
    int rc;
    char *error_buf;

    if (argc > 1) {
	
	for (i = 1; i < argc; i++) {
	    if (strcmp(argv[i], "-h") == 0) {
		usage(argv[0]);
		exit(0);
	    
	    } else if (strcmp(argv[i], "-I") == 0) {
		//INSTALL the proxy as a windows system service
		//
		char args[MAX_CMD_LINE_ARG_LEN] = "";
		if (argc > i + 1)
		    strcpy(args, argv[i+1]);
		
		if ((rc = proxy_install(args))) {
		    
		    if (rc != -1) {
			FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |
				FORMAT_MESSAGE_IGNORE_INSERTS,NULL,rc,
				MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),(LPTSTR) &error_buf,0,NULL);
	    
			printf("%s\n", error_buf);
			MessageBox(hwndMain, error_buf, "Unable to install proxy service", MB_ICONERROR);
		    }
	    
		} else {
		    printf("generic proxy has been successfully installed as a service.\n");
		}
		exit(0);
	
		
	    } else if (strcmp(argv[i], "-R") == 0) {

		//UNINSTALL the windows proxy system service.
		if ((rc = proxy_remove())) {
		    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |
			FORMAT_MESSAGE_IGNORE_INSERTS,NULL,rc,
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),(LPTSTR) &error_buf,0,NULL);
		    printf("%s\n", error_buf);
		    MessageBox(hwndMain, error_buf, "Unable to Uninstall proxy service", MB_ICONERROR);
		} else {
		    printf("generic proxy service has been successfully uninstalled.\n");
		}
		exit(0);
	    
		
	    } else if (strcmp(argv[i], "-S") == 0) {

		//START the proxy windows system service 
		//This is a normal start as against a cold start.
		if ((rc = proxy_start(0))) {
		    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |
			FORMAT_MESSAGE_IGNORE_INSERTS,NULL,rc,
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),(LPTSTR) &error_buf,0,NULL);
		    printf("%s\n", error_buf);
		    MessageBox(hwndMain, error_buf, "Unable to start proxy service", MB_ICONERROR);
		}
		exit(0);
	    
		
	    } else if (strcmp(argv[i], "-C") == 0) {
		    
		//START the proxy windows system service 
		//This is a cold start.
		if ((rc = proxy_start(1))) {
		    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |
			FORMAT_MESSAGE_IGNORE_INSERTS,NULL,rc,
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),(LPTSTR) &error_buf,0,NULL);
		    printf("%s\n", error_buf);
		    MessageBox(hwndMain, error_buf, "Unable to cold start proxy service", MB_ICONERROR);
		}
		exit(0);
	
	    } else if (strcmp(argv[i], "-E") == 0) {

		// STOP the running windows proxy service.  
		// This is NOT uninstall.
		if ((rc = proxy_stop())) {
		     FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |
			FORMAT_MESSAGE_IGNORE_INSERTS,NULL,rc,
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),(LPTSTR) &error_buf,0,NULL);
		    printf("%s\n", error_buf);
		    MessageBox(hwndMain, error_buf, "Unable to stop proxy service", MB_ICONERROR);
	  
		}
		exit(0);
      
	    } else if (strcmp(argv[i], "-X") == 0) {
		// This is called when the proxy service is started by the OS.
		// This is NOT to be invoked from the command line.
		SERVICE_TABLE_ENTRY ent[] = { { serviceName1, ServiceMain }, { 0, 0 }, };
		StartServiceCtrlDispatcher (ent);
		exit(0);
	    }
	    i++;
	}
    }
 
	// Proxy is not active yet
	setProxyStatusCode (0);
    
    //We reach this point only when the proxy is being started as a console application.
    //i.e. not as a service.
    init_generic_proxy(argc, argv);
	
	// Proxy is now active.
	setProxyStatusCode (1);
    
	startGenericProxy_Windows();
    exit(0);
}

/**
 * This is a dummy function used only for invoking main1 in
 * a separate thread.
 */
void* main_generic_proxy(void *dummy) {
    
    char **mb_argv = (char **) malloc(global_argc * sizeof(char *));
    for (int i = 0 ; i < global_argc; i++ ) {
	mb_argv[i] = (char *) malloc(sizeof(char) * 1000);
	wcstombs(mb_argv[i], global_argv[i], 1000);
	
    } 
	main1(global_argc, mb_argv);

	return NULL;
}




// Callback function
LRESULT CALLBACK MainWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);

/**
 * This is the entry point whenever the proxy is invoked from the command line:
 * be it run the proxy or to install/uninstall/start/stop it as service.
 * Please note that when starting the proxy as a service, ServiceMain is the place
 * from where the actual proxy code start executing.  Only the notification to start
 * happens from here.
 */
int WINAPI
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, INT nCmdShow)
{

	debugFD = fopen ("debug_OCALA.txt", "w");
	if (debugFD == NULL) {
		exit (-1);
	}

	MSG msg; // MSG structure to store messages
	WNDCLASSEX wcx; // WINDOW class information

	// Initialize the struct to zero
	ZeroMemory(&wcx,sizeof(WNDCLASSEX));
	wcx.cbSize = sizeof(WNDCLASSEX); // Window size. Must always be sizeof(WNDCLASSEX)
	wcx.style = CS_HREDRAW|CS_VREDRAW |CS_DBLCLKS ; // Class styles
	wcx.lpfnWndProc = (WNDPROC)MainWndProc; // Pointer to the callback procedure
	wcx.cbClsExtra = 0; // Extra byte to allocate following the wndclassex structure
	wcx.cbWndExtra = 0; // Extra byte to allocate following an instance of the structure
	wcx.hInstance = hInstance; // Instance of the application
	wcx.hIcon = LoadIcon (hInstance, MAKEINTRESOURCE(ID_MAINICON)); // Class Icon
	wcx.hCursor = LoadCursor(NULL, IDC_ARROW); // Class Cursor
	wcx.hbrBackground = (HBRUSH)(COLOR_WINDOW); // Background brush
	wcx.lpszMenuName = NULL; // Menu resource
	wcx.lpszClassName = "OCALAProxy"; // Name of this class
	wcx.hIconSm = NULL; // Small icon for this class

	// Register this window class with MS-Windows
	if (!RegisterClassEx(&wcx))
		return 0;

	// Create the window
	hwndMain = CreateWindowEx(0, //Extended window style
				"OCALAProxy", // Window class name
				"Generic Overlay Proxy", // Window title
				WS_OVERLAPPEDWINDOW, // Window style
				CW_USEDEFAULT,CW_USEDEFAULT, // (x,y) pos of the window
				CW_USEDEFAULT,CW_USEDEFAULT, // Width and height of the window
				HWND_DESKTOP, // HWND of the parent window (can be null also)
				NULL, // Handle to menu
				hInstance, // Handle to application instance
				NULL); // Pointer to window creation data

	// Check if window creation was successful
	if (!hwndMain)
		return 0;

	// Make the window visible
	//ShowWindow(hwndMain,SW_SHOW);


	global_argv = CommandLineToArgvW(GetCommandLineW(), &global_argc);
	//global_argv = (char *) lpCmdLine;
	//printf("\n\n\t\t\t-->%s\n\n", (char *) lpCmdLine);
	/*
	for (int i=0;i < global_argc;i++) {
		printf("\n\t\t--->%d %s\n",i, global_argv[i]);
	}*/
  	pthread_t mainProxyThread;
  	int mainProxyThrdRetVal = pthread_create(
                            		&mainProxyThread, NULL, 
            				main_generic_proxy, (void *) NULL
                                	);
  	if (mainProxyThrdRetVal != 0) {
        	//some error occurred during the creation of the resume proxy thread.
        	perror("Error while creating main proxy thread.");
  	}


	// Process messages coming to this window
	while (GetMessage(&msg,NULL,0,0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	// return value to the system
	return msg.wParam;	
}

/** 
 * This function handles windows events.  It is used only when invoked 
 * the proxy is being run from the command line.
 */
LRESULT CALLBACK MainWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam) {
	
    switch (msg) {

		case WM_QUERYENDSESSION:
		case WM_ENDSESSION:
		case WM_CLOSE:
			restoreNameServerSettings();
			return TRUE;
		break;

	    case WM_DESTROY:
		
		// User closed the window
			PRINT_DEBUG0 (DEBUG_LEVEL_SUPER, "WM_DESTROY\n\n");
						restoreNameServerSettings();
			PostQuitMessage(0);
	    break;
	
	    case WM_POWERBROADCAST:
		switch (wParam) {
			case (PBT_APMQUERYSUSPEND):
				Proxy::proxy->stopProxy();
				//MessageBox(hwnd, "Request for permission to suspend.", "Info", MB_OK);
				//return TRUE allow suspend, return BROADCAST_QUERY_DENY to not allow.
			break;
			
			case (PBT_APMRESUMESUSPEND):
				if (Proxy::proxy != NULL && !Proxy::proxy->stopProxyFlag) {
						// There is currently a proxy which is not stopped.
						PRINT_DEBUG0 (
										DEBUG_LEVEL_WARNING,
										"Avoiding double resume.\n"
								);
					} else {

  		    			pthread_t resumeProxyThread;

						int resumeProxyThrdRetVal = pthread_create(
                                    			&resumeProxyThread, NULL, 
                                    			resume_generic_proxy, (void *) NULL
                                		);
  		    			if (resumeProxyThrdRetVal != 0) {
        					//some error occurred during the creation of the resume proxy thread.
        					perror("Error while creating resume proxy thread.");
  		    			}
					}
		    
			break;
		}
	    break;

   	    default:
		// Call the default window handler
		return DefWindowProc(hwnd,msg,wParam,lParam);
    }
    return NO_ERROR;
}
