/**
 *  \file win-service.cpp
 *  \brief  Code to make the proxy run as a Windows Service.  The proxy starts
 *  from the ServiceMain function defined in this file when invoked as a Windows service.
 */

#include "../common.h"
#include "../debug.h"
#include "../generic_proxy.h"
#include "../ocd/OCD.h"
#include "../Proxy.h"

#include <pbt.h>
#include "resource.h"

#include <windows.h>
#include <winsvc.h>
#include <windowsx.h>
#include <shellapi.h>
#include <sys/stat.h>

#include <direct.h>

#include "win-service.h"
#include "win-main.h"

void startGenericProxy_Windows();

extern HWND hwndMain;

CHAR            szKey[MAX_PATH];


SERVICE_STATUS g_srv_status = {
  SERVICE_WIN32_OWN_PROCESS,
  SERVICE_START_PENDING,
  SERVICE_ACCEPT_STOP,
  NO_ERROR,
  NO_ERROR,
  0,
  0
};


char serviceName[255] = GENERIC_PROXY;
char displayName[255] = GENERIC_PROXY;

SERVICE_STATUS_HANDLE g_srv_status_handle;

/**
 *  Service event handler.
 *  This is used only when the proxy is started as a service, 
 *  i.e. the proxy is not being run from the command line.
 */
DWORD WINAPI HandlerEx ( DWORD dwControl, DWORD dwEventType, 
		LPVOID lpEventData, LPVOID lpContext) {
  
    switch (dwControl) {
	case SERVICE_CONTROL_STOP:
	    g_srv_status.dwCurrentState = SERVICE_STOP_PENDING;
	    g_srv_status.dwWin32ExitCode = 0;
	    g_srv_status.dwCheckPoint = 0;
	    g_srv_status.dwWaitHint = 0;
	break;
    
	case SERVICE_CONTROL_INTERROGATE:
	break;
    
	default:
	break;
    

	}

    SetServiceStatus (g_srv_status_handle, &g_srv_status);

	return NO_ERROR;
}

void strip_filename(char *filename)
{
  int i, len;

  len = strlen(filename);

  // strip off filename and get path
  for (i = len - 1; i >=0; i--) {
    if (filename[i] != '\\') {
      filename[i] = 0;
    } else {
      break;
    }
  }

  i = 0;
  while (filename[i] != 0) {
    if (filename[i] == '\\')
      filename[i] = '/';
    i++;
  }
}


/**
 * Go through all the loaded overlays and insert into the Windows registery
 * (overlay->winRegistryKey, overlay->status) pairs.
 */
void updateOverlayStatus() {

	if (Proxy::proxy == NULL) { 
        return;
    }

	list<OCD*>::iterator itr = Proxy::proxy->OCD_List.begin();

	while (itr != Proxy::proxy->OCD_List.end()) {
        OCD *tmpOverlay = *itr;
	update_status(tmpOverlay->getWinRegistryKey(), tmpOverlay->getStatus());
        itr++;
    }

}

void init_reg()
{
  CHAR svcPath[MAX_PATH];
  BYTE szImagePath[MAX_PATH];
  CHAR szBuf[MAX_PATH];
  HKEY hKey, hSubKey;
  DWORD retCode, rv, dwKeyType;
  DWORD dwBufLen = MAX_PATH;
  DWORD status = 0;
  int i;

  retCode = RegOpenKeyEx(HKEY_LOCAL_MACHINE, "System\\CurrentControlSet\\Services\\",
			 0, KEY_READ, &hKey);
  if (retCode != ERROR_SUCCESS)
    return;

  for (i = 0, retCode = ERROR_SUCCESS; retCode == ERROR_SUCCESS; i++) {
    retCode = RegEnumKey(hKey, i, svcPath, MAX_PATH);
    if (retCode == ERROR_SUCCESS) {
      lstrcpy(szKey, "System\\CurrentControlSet\\Services\\");
      lstrcat(szKey, svcPath);
      
      if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, szKey, 0,
		       KEY_QUERY_VALUE | KEY_SET_VALUE, &hSubKey) == ERROR_SUCCESS) {
	dwBufLen = MAX_PATH;
	rv = RegQueryValueEx(hSubKey, "ImagePath", NULL,
			     &dwKeyType, szImagePath, &dwBufLen);
	
	if (rv == ERROR_SUCCESS
	    && (dwKeyType == REG_SZ
		|| dwKeyType == REG_EXPAND_SZ) && dwBufLen) {
	  lstrcpy(szBuf, (const CHAR *)szImagePath);
	  CharLower(szBuf);
	  if (strstr(szBuf, "\\OCALA_proxy") != NULL) {
	    RegSetValueEx(hSubKey, "ProxyStatus", 0,
			  REG_DWORD, (BYTE *) &status, sizeof(status));
	    updateOverlayStatus();
	    break;
	  }
	}
	RegCloseKey(hSubKey);
      }
    }
  }
  RegCloseKey(hKey);
  return;
}

/**
 * Puts the value status under key statusStr
 * in the windows registry.
 */
void update_status(const CHAR *statusStr, DWORD status)
{
  HKEY hKey;

  if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, szKey, 0,
		   KEY_SET_VALUE, &hKey) == ERROR_SUCCESS) {
    RegSetValueEx(hKey, statusStr, 0,
		  REG_DWORD, (BYTE *) &status, sizeof(status));
    RegCloseKey(hKey);
  }
  return;
}

/**
 * This function is called in a separate thread.  It keeps looping
 * and updating the state of the proxy in the windows registry
 * till the service status is set to SERVICE_STOPPED (externally).
 */
void* statusUpdater(void *args) {
    while (g_srv_status.dwCurrentState != SERVICE_STOPPED) {
    
	switch (g_srv_status.dwCurrentState) {
	    case SERVICE_STOP_PENDING:
		g_srv_status.dwCurrentState = SERVICE_STOPPED;
		SetServiceStatus (g_srv_status_handle, &g_srv_status);
	    break;
        
	    default:
		Sleep(1000);
		updateOverlayStatus();
	    break;
	}
       
    }
	Proxy::proxy->stopProxy();
    //The proxy is stopped now.  We need to kill the service in order to 
    //stop it completeley.
    exit(-1);
}

/**
 * This is the starting point when the proxy is run as a windows service
 * and not from the command line.
 */
void WINAPI ServiceMain (DWORD ac, char **av)
{
  char path[MAX_PATH];
  DWORD status = 0;

  g_srv_status_handle = RegisterServiceCtrlHandlerEx (serviceName, HandlerEx, NULL);
  if (!g_srv_status_handle)
    return;
  g_srv_status.dwCurrentState = SERVICE_START_PENDING;
  g_srv_status.dwCheckPoint = 0;
  g_srv_status.dwWaitHint = 1000;
  g_srv_status.dwControlsAccepted = SERVICE_ACCEPT_POWEREVENT|SERVICE_ACCEPT_SHUTDOWN|SERVICE_ACCEPT_PAUSE_CONTINUE|SERVICE_ACCEPT_STOP;

  SetServiceStatus (g_srv_status_handle, &g_srv_status);

  /* initialization process */
  
  if (!GetModuleFileName (0, path, MAX_PATH)) {
	  return;
  }
  
  strip_filename(path);
  _chdir(path);
  freopen("proxy_error.log", "w", stderr);
  
  freopen("proxy.log", "w", stdout);
  
  init_reg();
  //This is the entry point to the generic proxy code when it is being
  //executed as a service.
  
  // Proxy is not active yet
  setProxyStatusCode (0);

  // Initialize the proxy.  
  // The tunnel and overlay listening threads have not been started yet.
  init_generic_proxy(ac, av);
  	
  // Proxy is now active.
  setProxyStatusCode (1);
  
  /* notify that the service has been started */
  g_srv_status.dwCurrentState = SERVICE_RUNNING;
  g_srv_status.dwCheckPoint = 0;
  g_srv_status.dwWaitHint = 0;
  SetServiceStatus (g_srv_status_handle, &g_srv_status);

  pthread_t statusUpdateThread;
  // Start the periodic registry status updater in a new thread
  // The status updater thread is also responsible for terminating
  // the application when the service state changes to SERVICE_STOPPED.
  int statusUpdateThrdRetVal = pthread_create(
                                    &statusUpdateThread, NULL, 
                                    statusUpdater, (void *) NULL
                                );
  if (statusUpdateThrdRetVal != 0) {
        //some error occurred during the creation of the status update thread.
        perror("Error while creating status update thread.");
  }


  // Start the listening threads.  This is a blocking call.
  startGenericProxy_Windows();
  
}

DWORD proxy_install (char *args)
{
  char path[MAX_PATH];
  char wrapper_path[MAX_PATH];
  char ImagePath[MAX_PATH];
//  struct stat buf;
  SC_HANDLE scm = 0;
  SC_HANDLE srv = 0;
  int rc = 0;
  if (!GetModuleFileName (0, path, MAX_PATH))
    return GetLastError ();
  strcpy(wrapper_path, path);

  // Get the path name of the current executable, i.e. no filename
  char* lastSlash = strrchr (wrapper_path, '\\');
  if (lastSlash == NULL) {
    MessageBox(hwndMain, "Unable to extract just the path from the full path of the current executable.",
		    	"Unable to install proxy service",
			MB_ICONERROR);
	return -1;
  }

  // Construct the full path to the wrapper batch file which will invoke the proxy.
  strcpy (lastSlash + 1, "OCALA_proxy_run.bat");

  struct stat tmpStat;
  if (stat(wrapper_path, &tmpStat) != 0) {
    printf("OCALA_proxy_run.bat does not exist in the same directory\n");
    return -1;
  }
  
  sprintf(ImagePath, "\"%s\" \"%s\" -X %s", wrapper_path, path, args);
  //sprintf(ImagePath, "\"%s\" -X %s", path, args);
  //printf("\"%s\" -X %s\n", path, args);
  
  scm = OpenSCManager (0, 0, SC_MANAGER_ALL_ACCESS);
  if (!scm)
    return GetLastError ();
  srv = CreateService (scm, serviceName, displayName, SERVICE_ALL_ACCESS,
                       SERVICE_WIN32_OWN_PROCESS, SERVICE_DEMAND_START, SERVICE_ERROR_NORMAL,
                       ImagePath, 0, 0, 0, 0, 0);

  if (!srv) {
		rc = GetLastError ();
		return rc;
  }
    
  char tmpDescStr[] = "Berkeley OCALA: Proxy for new network architectures";
  SERVICE_DESCRIPTION tmpServiceDesc;
  tmpServiceDesc.lpDescription = tmpDescStr;

  // Set the description of the OCALA service
  if ( ! ChangeServiceConfig2(
								srv,
								SERVICE_CONFIG_DESCRIPTION,
								&tmpServiceDesc
		) ) {

        PRINT_INFO0 (INFO_LEVEL_WARNING,
				"Unable to set the description of the service."
			);
        rc = GetLastError();				
  }

  CloseServiceHandle (srv);
  CloseServiceHandle (scm);
  return rc;
}

/**
 * This function removes the proxy windows service,
 * i.e. uninstalls the service.  Not stop a running service.
 */
DWORD proxy_remove ()
{
  SC_HANDLE scm = 0;
  SC_HANDLE srv = 0;
  int rc = 0;
  scm = OpenSCManager (0, 0, SC_MANAGER_ALL_ACCESS);

  if (!scm)
    return GetLastError ();
  srv = OpenService (scm, serviceName, DELETE);
  if (!srv)
    rc = GetLastError ();
  else
    {
      if (!DeleteService (srv))
        rc = GetLastError ();
      CloseServiceHandle (srv);
    }
  CloseServiceHandle (scm);
  return rc;
}

/**
 * This function starts the proxy service.
 */
DWORD proxy_start (BOOL cold_start)
{
  SC_HANDLE scm = 0;
  SC_HANDLE srv = 0;
  SERVICE_STATUS st;
  memset (&st, 0, sizeof (st));
  int rc = 0;
  if (!(scm = OpenSCManager (0, 0, SC_MANAGER_ALL_ACCESS)))
    return GetLastError ();

  if (!(srv = OpenService (scm, serviceName, SERVICE_START|SERVICE_QUERY_STATUS))) {
 
	char *error_buf;
	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |
			  FORMAT_MESSAGE_IGNORE_INSERTS,NULL, GetLastError(),
			  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),(LPTSTR) &error_buf,0,NULL);
	printf("%s\n", error_buf);
	    
    	return GetLastError ();
  }
  if (cold_start) {
    char *args[3] = {"-c"};
    if (!StartService(srv, 1, (LPCTSTR*) &args)) {
	   char *error_buf;
	    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |
			  FORMAT_MESSAGE_IGNORE_INSERTS,NULL, GetLastError(),
			  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),(LPTSTR) &error_buf,0,NULL);
	    printf("%s\n", error_buf);
      return GetLastError();
    }
  } else if (!StartService (srv, 0, 0)) {
	   char *error_buf;
	    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |
			  FORMAT_MESSAGE_IGNORE_INSERTS,NULL, GetLastError(),
			  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),(LPTSTR) &error_buf,0,NULL);
	    printf("%s\n", error_buf);
	  
    return GetLastError();
  }
  
  if (!QueryServiceStatus (srv, &st))
    return GetLastError ();

  {
    DWORD old;
    while (st.dwCurrentState == SERVICE_START_PENDING)
      {
	old = st.dwCheckPoint;
	Sleep (st.dwWaitHint);
	if (!QueryServiceStatus (srv, &st))
	  {
	    rc = GetLastError ();
	    break;
	  }
      }
    if (rc) {
      ;
    } else if (st.dwCurrentState == SERVICE_RUNNING) {
    	PRINT_INFO0 (INFO_LEVEL_MINIMAL,
		"Generic proxy service has been successfully started\n"
		);
    } else {
	PRINT_INFO1 (INFO_LEVEL_WARNING,
		"Unable to start generic proxy service.  Status is %d",
		(int)st.dwCurrentState
	);
    }
  }
  if (srv)
    CloseServiceHandle (srv);
  if (scm)
    CloseServiceHandle (scm);
  return rc;
}

/**
 * This function stops the running service.  Different from uninstalling the service
 * in proxy_remove().
 */
DWORD proxy_stop ()
{
  SC_HANDLE scm = 0;
  SC_HANDLE srv = 0;
  SERVICE_STATUS st;
  memset (&st, 0, sizeof (st));
  int rc = 0;
  
  if (!(scm = OpenSCManager (0, 0, SC_MANAGER_ALL_ACCESS)))
    rc = GetLastError ();
  else if (!(srv = OpenService (scm, serviceName, SERVICE_STOP)))
    rc = GetLastError ();
  else if (!ControlService (srv, SERVICE_CONTROL_STOP, &st))
    rc = GetLastError ();
  else
    printf ("generic proxy service has been stopped\n");
  
  
  if (srv)
    CloseServiceHandle (srv);
  if (scm)
    CloseServiceHandle (scm);
 
  return rc;
}


void resumeProxy() {

	pthread_t resumeProxyThread;
	int resumeProxyThrdRetVal = pthread_create(
        		&resumeProxyThread, NULL, 
                        resume_generic_proxy, (void *) NULL
                        );
  	if (resumeProxyThrdRetVal != 0) {
        	//some error occurred during the creation of the resume proxy thread.
        	perror("Error while creating proxy resume thread.");
  	}
}
		      


