#include <assert.h>
#ifndef _WIN32
#include <arpa/inet.h>
#endif

#include "RSPServerFinder.h"
#include "RSPDNSReplyCallback.h"
#include "RSPTunnelOCD.h"
#include "RSPTunnelOCD_debug.h"

/**
  * @param rspDomainList list of OASIS service domains associated with the RSP.
  * @param resolvePeriod How often (in seconds) should we resolve each of the domains.
  * Default value is RSP_SERVER_FINDER_DEFAULT_RESOLVE_PERIOD.
  * If resolvePeriod is negative, no periodic resolution is done.
  */
RSPServerFinder::RSPServerFinder (list<string> rspDomainList, RSPTunnelOCD* ocd, int32_t resolvePeriod) {

    assert (ocd != NULL);
    
    this->rspDomainList = rspDomainList;
    this->ocd = ocd;
    this->resolvePeriod = resolvePeriod;

    // schedule the first periodic resolve
    doPeriodicResolve();
}

/**
  * This function is called periodically to resolve the various RSP
  * domains.
  * This function resolves all domains and then schedules itself again for later.
  */
void RSPServerFinder::doPeriodicResolve() {
    
    resolveAllDomains();

    // schedule the next check
    TObjectFunctor<RSPServerFinder> *resolveFunctor  = 
                new TObjectFunctor<RSPServerFinder> (this, &RSPServerFinder::doPeriodicResolve);

    TimerEvent* resolveTimer = TimerEvent::create (resolvePeriod, 0, (TFunctor *) resolveFunctor, true /* automatically free cbk when TimerEvent is destroyed*/);
    ocd->addTimerEvent(resolveTimer);

#ifdef _WIN32
	ocd->updateProxyMonitorStatus();
#endif
}

/**
  * This function makes DNS requests for all domains in the rspDomainList.
  * The resolution is not synchronous.  The replies come in later through receivedDNSReply()
  * callbacks.
  */
void RSPServerFinder::resolveAllDomains () {
   
    list<string>::iterator domainItr = rspDomainList.begin();
    while (domainItr != rspDomainList.end()) {
        string currDomain = *domainItr;

        // Make a DNS request for currDomain and register a callback.
        RSPDNSReplyCallback* replyCbk = RSPDNSReplyCallback::create (currDomain, this);
        // replyCbk is automatically deleted after the callback is invoked.
        // Don't keep a pointer to replyCbk anywhere in this object.

    	RSPTunnelOCD_PRINT_DEBUG1 ( DEBUG_LEVEL_VERBOSE, "Making DNS request for %s\n", currDomain.c_str());
        if ( ! ocd->makeDNSRequest (replyCbk, currDomain.c_str())) {
            PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
                            "Unable to send DNS request for %s.\n",
                            currDomain.c_str()
                        );
        }

        domainItr++;
    }
}

/**
  * Add a mapping between the specified domainName and ip.
  * If an existing mapping exists for domainName, it is replaced.
  * @param ip host byte order.
  */
void RSPServerFinder::addMapping (string domainName, uint32_t ip) {
    RSPTunnelOCD_PRINT_DEBUG1 ( DEBUG_LEVEL_VERY_VERBOSE, "RSPServerFinder currently has %d mappings.\n", map_domain_ip.size());

	uint32_t tmpOldIP;
	bool tmpFlag = lookupDomain (domainName, &tmpOldIP);
	if (!tmpFlag || tmpOldIP != ip) {
		// Display message only if this is a new mapping or mapping has changed.
		struct in_addr tmpIP;
		tmpIP.s_addr = htonl (ip);

		RSPTunnelOCD_PRINT_DEBUG3 (DEBUG_LEVEL_MINIMAL,
                                "RSP Gateway mapping: %s --> %x (%s)\n",
                                domainName.c_str(), ip, inet_ntoa (tmpIP)
                            );
	}

	// Erase old mapping, then insert new one.
	map_domain_ip.erase (domainName);
    map_domain_ip.insert(pair<string,uint32_t>(domainName, ip));

}

#ifdef _WIN32

string RSPServerFinder::getMappingStatusStr() {

	string tmpStatusMsg;
	if (findNumValidMappings() > 0) {

		tmpStatusMsg = "The following OCALA Secure Connection gateways are known:\n";
		map<string, uint32_t>::iterator itr = map_domain_ip.begin();
		while (itr != map_domain_ip.end()) {

			struct in_addr tmpIP;
			tmpIP.s_addr = htonl (itr->second);
			char tmpAddrStr [100];
			sprintf (tmpAddrStr, " --> %s", inet_ntoa (tmpIP));
			tmpStatusMsg += itr->first + string(tmpAddrStr) + string("\n");
			itr++;
		}
	} else {
		tmpStatusMsg = "No OCALA Secure Connection Gateways known";
	}

	return tmpStatusMsg;
}
#endif

/**
  * Looks up the IP address of the specified domain name.
  * The IP  (in host byte order) is placed in retIP.
  * @return true if a mapping for domainName was found and the ip was not 0.  Else returns false, retIP is set to 0.
  */
bool RSPServerFinder::lookupDomain (string domainName, uint32_t* retIP) {
    map<string, uint32_t>::iterator itr = map_domain_ip.find (domainName);
    if (itr == map_domain_ip.end()) {
        // No mapping for the specified domain.
        return false;
    }

    *retIP = itr->second;
    if (*retIP == 0) {
        return false;
    } else {
        return true;
    }
}

/**
 * Returns the number of gateways which were correctly resolved.
 */
int RSPServerFinder::findNumValidMappings () {
	int retVal = 0;
	map<string, uint32_t>::iterator itr = map_domain_ip.begin();
	while (itr != map_domain_ip.end()) {
		if (itr->second != 0) {
			retVal ++;
		}
		itr++;
	}
	return retVal;
}
