/**
 *  \file RSPTunnelOCD.cpp
 *  \brief The functions associated with the RSPTunnelOCD are defined here.
 */
#ifndef _WIN32
#	include <sys/types.h>
#	include <sys/socket.h>
#else
#	include "netwrap.h"
#endif

#include <errno.h>
#include <assert.h>

#include "OCD_include.h"

#include "RSPTunnelOCD.h"
#include "RSPTunnelOCDContext.h"
#include "RSPTunnelOCDStateInfo.h"
#include "RSPTunnelOCD_debug.h"
#include "RSPServerFinder.h"

RSPTunnelOCD::RSPTunnelOCD(
        string type, string id, string desc, 
        string fileName, 
        string urlSuffixes, bool handleLegacyURLs
    ) :OCD(    type, id, desc, fileName, 
                urlSuffixes, handleLegacyURLs
            ) {

    // Constructs the RSPTunnelOCDSpecific code.
    // The bulk of RSPTunnelOCD specific processing is handled by the context
    // module.  
    this->context = new RSPTunnelOCDContext(this, fileName);
   
}

/**
  * Construct a new RSPTunnelOCD with fields taken from the
  * specified overlay network.
  */
RSPTunnelOCD::RSPTunnelOCD(OCD* o): OCD() {


    this->type = o->type;
    this->id = o->id;
    this->description = o->description;
    this->configFileName = o->configFileName;
    this->urlSuffixes = o->urlSuffixes;
    this->handleLegacyURLs = o->handleLegacyURLs;

    this->configPropertyMap = o->configPropertyMap;
    this->myNameList = o->myNameList;
	this->context = NULL;
}



/**
  * Initializes the RSPTunnelOCD 
  * @param proxy A pointer to the proxy of which this overlay is a part.
  * @param oci  The OCI associated with this OCD.
  * @param recoverFromCrash This parameter is set to true by the proxy if it is 
  * is recovering from a crash. The OCD is expected to reconstruct its
  * state from the journal.  This flag is usually simply passed on to
  * the OCD's JournalManager which is instantiated in the init() function .
  */
void RSPTunnelOCD::init(Proxy *prxy, OCI* oci, bool recoverFromCrash) {
 
    this->proxy = prxy;
    this->oci = oci;


    // Initializes the RSPTunnelOCD specific processing.
    // A socket is opened for the RSPTunnelOCD to listen on.
    this->context->init();
    
    // If you need to do crash recovery, you need to instantiate a JournalManager for the OCD.
    // Since this is an advanced topic and not necessary for basic operation of the OCD, we are not
    // describing it here.  Please refer the I3OverlayNetwork::init() for an example of how to use
    // the JournalManager.

#ifdef _WIN32
	// Set the list of icons associated with the rspTunnelOCD
    setMonitor_IconList (string("rsptunnelocd_main.ico,rsptunnelocd_nosrv.ico,rsptunnelocd_disconnect.ico,rsptunnelocd_stop.ico,rsptunnelocd_run.ico"));
	setMonitor_CurrentDescription (description);
#endif
}


/**
  * Gives a string representation (pretty printed) of the RSPTunnelOCD.
  * @param tmpStr a character buffer in which the string is stored.
  * Sufficient space should be allocated for tmpStr.
  * @return pointer to tmpStr
  */
char* RSPTunnelOCD::toString(char *tmpStr) {

    sprintf( tmpStr,
            "Sample Overlay Network\n"
            "\ttype\t\t:\t%s\n"
            "\tid\t\t:\t%s\n"
            "\tdescription\t:\t%s\n"
            "\tconfig file\t:\t%s\n",
            type.c_str(),
            id.c_str(),
            description.c_str(),
            configFileName.c_str()
           );
    return tmpStr;
}

/**
  * Setup up a tunnel over the RSPTunnelOCD to the specified name.
  * If a tunnel already exists, reuse it if it makes sense to do so in the specific overlay.
  * @param ociHandle The oci to be notified through a callback when the tunnel setup succeeds
  * or fails.
  * @parentLayer The layer which is setting up this new layer.
  * @return a handle to the tunnel.
  */
OCDHandle RSPTunnelOCD::setup (
                            const string endPointName, 
                            int endPointNameType,
                            char *optionBlocks, 
                            int optionLen, 
                            OCIHandle ociHandle,
                            OCI* oci) {

    RSPTunnelOCDStateInfo* ocdState = context->setupTunnel (endPointName, ociHandle, oci);

    // TODO Warn the user
    if (ocdState != NULL) {
		return ocdState->myHandle;
    } else {
		if  (context->gatewayDownPolicy == RSP_GATEWAY_DOWN_POLICY_NODROP) {
			// OCI will handle this connection as a legacy request.
			return NULL_HANDLE;
		} else {
			// Tell the OCI to drop the connection
        	return DROP_OCD_HANDLE;
		}
    }

}
    
/**
  * This function is called when some other layer wants to send anything via RSPTunnelOCD.
  */
void RSPTunnelOCD::send ( OCDHandle handle, char *data, int dataLen) {

    RSPTunnelOCDStateInfo *ocdState = (RSPTunnelOCDStateInfo *) getStateInfo (handle);

    if (ocdState == NULL) {
        RSPTunnelOCD_PRINT_DEBUG1 (
                        DEBUG_LEVEL_MINIMAL,
                        "Trying to send through unknown OCD with handle %d\n",
                        handle
                    );
        return;
    } 
  
    RSPTunnelOCD_PRINT_DEBUG2 (DEBUG_LEVEL_VERY_VERBOSE,
                "Sending data packet of size %d to %s.\n",
                dataLen, ocdState->endPointName
            );

    // Send the dataLen bytes contained in dataLen through the specific OCD.
    context->sendDataPacket (ocdState, data, dataLen);
}


/**
  * Initializes the state necessary for inbound names associated with this 
  * RSPTunnelOCD.
  */
void RSPTunnelOCD::initMyNames () {

    // Iterate over the incoming names assocated with this OCD 
    list<MyName *>::iterator itr = myNameList.begin();
    while (itr != myNameList.end()) {

        MyName *currName = (*itr);

	    // The OCD specific properties associated with the name
   	    // can be retrieved by invoking getProperty on the MyName object.
        //char *sampleProp = currName->getProperty (SAMPLE_PROPERTY_NAME); 

	    // OCD specific inbound name setup steps.

    	itr++;
    }
}

/**
 * Destroy all mappings/state associated with the specified ocdState
 */
void RSPTunnelOCD::destroyOCDState (RSPTunnelOCDStateInfo *ocdState) {
	assert (ocdState != NULL);
	this->context->destroyOCDState (ocdState);
}

#ifdef _WIN32
void RSPTunnelOCD::updateProxyMonitorStatus () {

	if ( ! isActive () || context == NULL || context->rspServerFinder == NULL) {
		setMonitor_CurrentStatusMsg (string ("Inactive"));
		setMonitor_CurrentIcon (string("rsptunnelocd_nosrv.ico"));
		setMonitor_CurrentStatusCode (0);
		return;
	}

	if (context->findNumValidMappings() > 0) {
		
		setMonitor_CurrentIcon (string("rsptunnelocd_run.ico"));
   		setMonitor_CurrentStatusMsg (context->rspServerFinder->getMappingStatusStr());
		setMonitor_CurrentStatusCode (1);
	} else {
		setMonitor_CurrentIcon (string("rsptunnelocd_disconnect.ico"));
		setMonitor_CurrentStatusMsg (context->rspServerFinder->getMappingStatusStr());
		setMonitor_CurrentStatusCode (0);
	}
}
#endif

// Register the OCD with the proxy when the code is dynamically loaded
// The following line sets the name of the OCD's main class
#define MY_OCD_CLASS_NAME RSPTunnelOCD
// The following line sets the type name of the OCD
#define MY_OCD_TYPE_STR "RSPTunnelOCD"
#include "OCDRegister.h"
