/**
 *  \file RSPTunnelOCD.cpp
 *  \brief The functions associated with the RSPTunnelOCD are defined here.
 */
#ifndef _WIN32
#	include <sys/types.h>
#	include <sys/socket.h>
#	include <netinet/in.h>
#	include <arpa/inet.h>
#	include <netdb.h>
#	include <errno.h>
#	include <strings.h>
#endif


extern "C" {
#	include "netwrap.h"
}
#include "OCD_include.h"
#include "Proxy.h"

#include "RSPTunnelOCD.h"
#include "RSPTunnelOCDContext.h"
#include "RSPTunnelOCDStateInfo.h"
#include "RSPTunnelOCDPacket.h"
#include "RSPTunnelOCD_debug.h"
#include "RSPServerFinder.h"

#include "StringTokenizer.h"


/**
  * Construct a RSPTunnelOCDContext object.
  * Bulk of the RSPTunnelOCD specific work takes place in this module.
  * @param ocd The RSPTunnelOCD associated with this context.
  * @param fileName The RSPTunnelOCD specific configuration file.
  */
RSPTunnelOCDContext::RSPTunnelOCDContext ( RSPTunnelOCD* ocd, string fileName ) {

    this->ocd = ocd;
    this->configFileName = fileName;
	this->rspServerFinder = NULL;

}

/**
  * Initializes the RSPTunnelOCDContext.
  * @param proxy A pointer to the proxy of which this overlay is a part.
  * @param oci  The OCI associated with this OCD.
  * @param recoverFromCrash This parameter is set to true by the proxy if it is 
  * is recovering from a crash. The OCD is expected to reconstruct its
  * state from the journal.  This flag is usually simply passed on to
  * the OCD's JournalManager which is instantiated in the init() function .
  */
void RSPTunnelOCDContext::init() {

	// The following code is given just to show how to use the ConfigFile class.
	// The configuration file associated with your overlay need not even be in XML and 
	// you can parse it in whatever way you like.  If the configuration file is in XML,
	// you may want to use the ConfigFile class or subclass it like in I3ConfigFile. 
	// Please have a look at varRSPTunnelOCD.xml in order to better understand this example.
	ConfigFile myConfigFile (configFileName, "RSPTunnelOCDConfigFile");

    // Read the port on which the RSPTunnelOCD should listen.
    myConfigFile.readUShortAttrib ("//RSPTunnelOCDConfigFile", "myPortNum", 
                        &myPortNum,
                        true /* This is a mandatory argument.  Proxy will abort if this attribute is not found. */
                    );
    RSPTunnelOCD_PRINT_INFO1 (
            INFO_LEVEL_MINIMAL,
            "Port to listen on = %d.\n",
            myPortNum
        );

    // init the RSP list
    char rspDomainListStr[MAX_PARAM_LEN];
    rspDomainListStr[0] = '\0';
    myConfigFile.readStringAttrib ("//RSPTunnelOCDConfigFile", "rspDomainList", 
                        rspDomainListStr,
                        true /* This is a mandatory argument.  Proxy will abort if this attribute is not found. */
                    );
    if ( ! isBlank (rspDomainListStr)) {
        string tmpRSPDomainListStr = rspDomainListStr;
        StringTokenizer rspDomainST (tmpRSPDomainListStr, string(","));
        list<string> rspDomainList;
        while (rspDomainST.hasMoreTokens()) {
            rspDomainList.push_back(rspDomainST.nextToken());
        }
        this->rspServerFinder = new RSPServerFinder (rspDomainList, ocd);
    } else {
        PRINT_DEBUG0 (DEBUG_LEVEL_MINIMAL,
                    "No RSP domains specified.\n"
                );
        this->rspServerFinder = NULL;
    }


    char gatewayDownPolicyStr[MAX_PARAM_LEN];
    gatewayDownPolicyStr[0] = '\0';
    myConfigFile.readStringAttrib ("//RSPTunnelOCDConfigFile", "gatewayDownPolicy",
                                    gatewayDownPolicyStr,
                                    false
                                );
    if ( strcasecmp (gatewayDownPolicyStr, "nodrop") == 0) {
        this->gatewayDownPolicy = RSP_GATEWAY_DOWN_POLICY_NODROP;
    } else {
        this->gatewayDownPolicy = RSP_GATEWAY_DOWN_POLICY_DROP;
    }
    

    // Initialize listening socket

    
#ifdef _WIN32 
	if ( nw_init() != 0) {
		RSPTunnelOCD_PRINT_INFO1 (
								INFO_LEVEL_FATAL_ERROR,
								"Unable to initialize network for RSPTunnelOCD %s.\n",
								ocd->id.c_str()
							);
		EXIT_ON_ERROR;
		// Usually, it is a bad idea to kill the proxy. 
		// Other OCDs which may be working will also be killed.
		// Ideally, you should implement an OCD specific retry mechanism.
		// However, for this particular error, retry may not be fruitful.
	}
#endif

    // Open the socket on which RSPTunnelOCD listens for packets.
    mySockFD =  socket (AF_INET, SOCK_DGRAM, 0); 
    
    // bind the socket to myPortNum
    struct sockaddr_in my_addr;    // my address information
    my_addr.sin_family = AF_INET;         // host byte order
    my_addr.sin_port = htons (myPortNum);     // short, network byte order
    my_addr.sin_addr.s_addr = INADDR_ANY; // automatically fill with my IP
    memset(&(my_addr.sin_zero), '\0', 8); // zero the rest of the struct

    if (bind(mySockFD, (struct sockaddr *)&my_addr,
			sizeof(struct sockaddr)) == -1) {
        RSPTunnelOCD_PRINT_INFO1 (
                    INFO_LEVEL_FATAL_ERROR,
                    "RSPTunnelOCD: Unable to initialize: %s.\n",
                    strerror (errno)
                );
        EXIT_ON_ERROR;

        // Note that this will exit the whole proxy.  Even other OCDs
        // that are plugged into the proxy will be terminated.
        // This is not a polite thing to do.
        
    }

#ifndef _WIN32
    int yes = 1;
    if (setsockopt(mySockFD, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1) {
        PRINT_DEBUG1(
                        DEBUG_LEVEL_WARNING,
                        "Unable to set SO_REUSEADDR socket option for RSPTunnelOCD socket: %s\n",
                        strerror (errno)
                    );
    }
#endif


    // We have bound the socket to the myPortNum.
    // Now, add a callback to be invoked when a packet is received 
    // on the socket.
    TObjectFunctor<RSPTunnelOCDContext> *fdfunc = new TObjectFunctor<RSPTunnelOCDContext>(this, &RSPTunnelOCDContext::receivedPacketFromOverlay);
    ocd->addFDCallback(mySockFD, fdfunc, true);

    RSPTunnelOCD_PRINT_INFO1 ( INFO_LEVEL_VERBOSE,
                "RSPTunnelOCD started listening on port %d.\n",
                myPortNum
            );

    // If you need to do crash recovery, you need to instantiate a JournalManager for the OCD.
    // Since this is an advanced topic and not necessary for basic operation of the OCD, we are not
    // describing it here.  Please refer the I3OverlayNetwork::init() for an example of how to use
    // the JournalManager.



}

/**
 * This function is called when a packet is received over the RSPTunnelOCD architecture.
 */
void RSPTunnelOCDContext::receivedPacketFromOverlay() {
	
    // Read the packet from the socket on which RSPTunnelOCD was listening.
    char recvBuf [ MAX_PACKET_BUF_LEN];
    struct sockaddr_in their_addr; 
    socklen_t addr_len = sizeof(struct sockaddr);
    int recvLen = recvfrom (    mySockFD, 
                            recvBuf, MAX_PACKET_BUF_LEN - 1 , 0,
			                (struct sockaddr *) &their_addr, &addr_len
                        );
    if (recvLen == -1) {
        RSPTunnelOCD_PRINT_DEBUG2 ( DEBUG_LEVEL_WARNING,
                        "Error while receiving packet from %s: %s.\n",
                        inet_ntoa (their_addr.sin_addr),
                        strerror (errno)
                    );
        return;  
    
    } 
        
    RSPTunnelOCD_PRINT_DEBUG2 ( DEBUG_LEVEL_VERY_VERBOSE,
                        "Received packet from %s:%d.\n",
                        inet_ntoa (their_addr.sin_addr), ntohs (their_addr.sin_port)
                    );



    // Cast the received packet into a RSPTunnelOCDPacketHeader so that we can retrieve the blockType.
    RSPTunnelOCDPacketHeader* recvPacketHeader = (RSPTunnelOCDPacketHeader *) recvBuf;

    // Depending on the type of the packet process it appropriately.
    switch (recvPacketHeader->blockType) {

        case RSPTunnelOCD_PKT_SETUP:
            processSetupPacket(recvBuf, recvLen, their_addr);
        break;

        case RSPTunnelOCD_PKT_SETUP_OK:
            processSetupOKPacket(recvBuf, recvLen, their_addr);
        break;

        case RSPTunnelOCD_PKT_SETUP_FAIL:
            processSetupFailPacket(recvBuf, recvLen, their_addr);
        break;

        case RSPTunnelOCD_PKT_DATA:

            RSPTunnelOCD_PRINT_DEBUG0 ( DEBUG_LEVEL_VERBOSE, "RSPTunnelOCD_PKT_DATA\n");
            processDataPacket(recvBuf, recvLen, their_addr);
        break;

        default:
            RSPTunnelOCD_PRINT_DEBUG1 ( DEBUG_LEVEL_WARNING,
                            "Packet of unknown type %d received.\n",
                            recvPacketHeader->blockType
                        );
        break;
    }

}

/**
  * This function is called when a setup packet is received by the 
  * RSPTunnelOCD.
  * @param recvBuf The entire contents of the received packet.
  * @param recvLen The length of the received packet.
  * @param fromAddr The IP address from which the packet was received.
  * Note that some of the parameters are not used by RSPTunnelOCD as the setup packet is very simple
  * and contains no other information, except the fact that it is a setup packet.
  * However, for more complicated OCDs, the setup packet might contain other information.
  * For example, an i3 setup packet (private trigger request) contains the private trigger id 
  * of the remote host.
  */
void RSPTunnelOCDContext::processSetupPacket(const char* recvBuf, int recvLen, struct sockaddr_in fromAddr) {
    
    // In the RSPTunnelOCD, we use the source IP address and port of the packet as an internal unique identifier
    // for a tunnel.  In the case of i3, the private trigger ids play this role.

    // Now, let us check if there already is state associated with this remote with fromAddr.
    // For this we use some OC-D specific data structure.  In this case, it is a simple hash table.
    // We use the remote port number to further filter the retrieved ocdStates.
    RSPTunnelOCDStateInfo* currOCDState = lookupRemoteHost (fromAddr.sin_addr,  fromAddr.sin_port);

    if (currOCDState == NULL) {

        RSPTunnelOCD_PRINT_DEBUG2 ( DEBUG_LEVEL_VERBOSE,
                    "Creating RSPTunnelOCDState for fromAddress %s:%d.\n",
                    inet_ntoa (fromAddr.sin_addr), ntohs (fromAddr.sin_port)
                );

        // This is the first time we are seeing this fromAddr.
        // So, let us now create a new RSPTunnelOCDState object for this tunnel.

        // First we need a new unique OCD handle for this OCDState.
        // This OCD Handle is used by the OCI layer to identify the tunnel.
        OCDHandle newOCDHandle = ocd->getNewOCDHandle();
        
        currOCDState = new RSPTunnelOCDStateInfo (newOCDHandle, ocd->oci, ocd, "inbound");
        // Register this OCDState with the OCD.
        // This step is common for all OCDs, not just RSPTunnelOCD.
        ocd->putStateInfo (currOCDState->myHandle, currOCDState);
       
        // Set the remote address and port of this OCD State.
        currOCDState->remoteAddress = fromAddr.sin_addr;
        currOCDState->remotePort = fromAddr.sin_port;
        
        // Add the newly created OCDState to the remote host address --> RSPTunnelOCDState hash table,
        // so that we can retrieve it the next time we look for fromAddr.
        // This is RSPTunnelOCD specific.
        addRemoteHost (currOCDState);

    }

    // This OCD is now ready to relay data.
    currOCDState->status = OCD_STATE_STATUS_READY;
    
    // Now currOCDState points to a valid OCDState, either newly created
    // or prexisting.
    sendSetupReplyOK (currOCDState);
    
	// Process the received packet.  If the packet is not an OCD specific control packet,
	// hand off the packet to the OCI after stripping of the OCD specific headers.
        // oci->receive ( strippedData, this, handle);
        // handle is the handle of the ocd state corresponding to the tunnel on which the packet was received.

	// How to get the OCDState corresponding to this packet?
	// It is the OCD's responsibility to parse the packet, extract some unique identifier X out of it and then 
	// to use X to retrieve the OCDState.  For example, in the case of i3, the private trigger id contained in 
	// the packet is extracted and used to retrieve the OCDState corresponding to the connection for which 
	// this packet was intended.  It is the i3 OCD module's responsibility to maintain the 
	// private trigger id ---> OCDState mappings, as this OCD specific. 
	// So there is no OCD independent way to obtain a pointer to the OCDState associated with a received packet.

}

/**
  * Process the received data packet.
  * Just strip off the RSPTunnelOCD headers and pass it to the OC-I layer.
  */
void RSPTunnelOCDContext::processDataPacket(char* recvBuf, int recvLen, struct sockaddr_in remoteAddress) {

    // Lookup the OCDState corresponding to this packet.
    RSPTunnelOCDStateInfo* currOCDState = lookupRemoteHost (remoteAddress.sin_addr, remoteAddress.sin_port);
    if (currOCDState == NULL) {
        RSPTunnelOCD_PRINT_DEBUG2 (DEBUG_LEVEL_WARNING,
                    "Unable to lookup for remote host %s:%d.\n",
                    inet_ntoa (remoteAddress.sin_addr), ntohs (remoteAddress.sin_port)
                );
        return;
    }

    RSPTunnelOCD_PRINT_DEBUG2 (DEBUG_LEVEL_VERY_VERBOSE,
                "Received Data packet from %s:%d.\n",
                inet_ntoa (remoteAddress.sin_addr), ntohs (remoteAddress.sin_port)
            );
    
    // Appropriately cast the data.
    RSPTunnelOCDPacketData* tmpDataPkt = (RSPTunnelOCDPacketData *) recvBuf;
   

    // Pass this packet to the OCI layer for further processing.
    // Note that the OCI layer may modify the contents of tmpDataPkt->payload.
    // However, it will not store a pointer to to tmpDataPkt->payload for future use,
    // nor will it try to free the memory.  Basically, it just uses the data 
    // pointed to by tmpDataPkt->payload, may modifiy it and that's it.
    ocd->oci->receive (tmpDataPkt->payload, ocd, currOCDState->myHandle);
}



/**
  * Set up a tunnel over RSPTunnelOCD to the URL endPointName.
  * @param ocdState The OCD State associated with this tunnel.
  * @return false if something goes wrong while trying to set up the connection.
  * True if nothing went wrong - note that setup has not completed yet though.
  */
RSPTunnelOCDStateInfo* RSPTunnelOCDContext::setupTunnel(string endPointName, OCIHandle ociHandle, OCI* oci) {
  
    if (this->rspServerFinder == NULL) {
        PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
                    "Unable to setup tunnel for %s as rspServerFinder is not initialized.\n",
                    endPointName.c_str()
                );
        return NULL;
    }   

    char tmpEndPointName [MAX_URL_LEN];
    strncpy (tmpEndPointName, endPointName.c_str(), MAX_URL_LEN);
    stripSuffix (tmpEndPointName, ".rspTunnel");
 
    // host byte order
    uint32_t domainIP;
    
    // Find the current IP address associated with the domain specified as the end point in the ocdState.
    if ( ! rspServerFinder->lookupDomain (string(tmpEndPointName), &domainIP)) {

        PRINT_DEBUG1 (DEBUG_LEVEL_MINIMAL,
                    "Unable to setup tunnel as rspServerFinder could not find domain %s.\n",
                    tmpEndPointName
                );
        return NULL;
    }

    // Check if a tunnel exists to the resolved domainIP.
    RSPTunnelOCDStateInfo* ocdState =  lookupRemoteHost (htonl (domainIP), htons(myPortNum));
    if (ocdState == NULL) {
        // No tunnel exists,  create a new one.
 
        // First get a new unique handle.
	    OCDHandle ocdHandle = ocd->getNewOCDHandle();
                    
        // Now create the OCD State for this new tunnel
        ocdState = new RSPTunnelOCDStateInfo (ocdHandle, oci, ocd, endPointName);
	        
        //add to the list of ocd states present in this OCD.
        ocd->addOCDState (ocdState);
		
        RSPTunnelOCD_PRINT_DEBUG2 (
                    DEBUG_LEVEL_VERBOSE, 
                    "Creating a new RSPTunnel (handle %d, endpointName  = %s)\n", 
                    ocdHandle,
                    ocdState->endPointName
                );

        // Add the specified ociHandle to the list of OCIStates waiting for this OCD setup to complete
        ocdState->addParentHandle (ociHandle);

	    // Add the OCDState of the newly created tunnel to the handle-->state hash table
        ocd->putStateInfo (ocdHandle, ocdState);


       	// NOTE: optionBlocks are currently not used.
        ocdState->setupOptionBlocks = NULL;
        ocdState->setupOptionLen = 0;

        ocdState->remoteAddress.s_addr = htonl (domainIP);
        ocdState->remotePort = ntohs(myPortNum);

        // Add this OCDState to the hash table keyed by the remote address.
        addRemoteHost (ocdState);

        // Send the setup packet to the remote end point.
        sendSetupPacket (ocdState);

        // Processing continues when a tunnel setup reply is received.
    
    } else {

        // Tunnel the specified domain IP alreay exists.
        // So we will just reuse it, if active.
        if (ocdState->status == OCD_STATE_STATUS_READY) {
                
            // Add the specified ociHandle to the list of OCIStates waiting for this OCD setup to complete
            ocdState->addParentHandle (ociHandle);

			// This OCD State is already ready.
            // We are ready to call notifySetupCompleted().  We need to call it asynchronously
            // after returning from this function.
            // So we set up a timer event.
       
	   		ocdState->setFakeSetupTimer ();

            

        } else {
                
			// Tunnel existed, but no longer active
            RSPTunnelOCD_PRINT_DEBUG2 (
                    DEBUG_LEVEL_VERBOSE, 
                    "In RSPTunnelOCD setup, refreshing tunnel handle %d, endpointName  = %s\n", 
                    ocdState->myHandle,
                    ocdState->endPointName
                );

			// Add the specified ociHandle to the list of OCIStates waiting for this OCD setup to complete
            ocdState->addParentHandle (ociHandle);

	        // Send the setup packet to the remote end point.
		    sendSetupPacket (ocdState);

			// Processing continues when a tunnel setup reply is received.

        }
   }
        
    return ocdState;
}




/** Send a setup packet to the remote end point of the specified ocdState.
  * @return true if the sending was successful, else return false.
  */
bool RSPTunnelOCDContext::sendSetupPacket (RSPTunnelOCDStateInfo* ocdState) {

    RSPTunnelOCD_PRINT_DEBUG2 (DEBUG_LEVEL_VERBOSE,
                "Sending setup packet to %s:%d.\n",
                inet_ntoa (ocdState->remoteAddress), ntohs (ocdState->remotePort)
            );

    char sendBuf[MAX_PACKET_BUF_LEN];
    
    // Construct the setup packet
    RSPTunnelOCDPacketHeader *sendPkt = (RSPTunnelOCDPacketHeader *) sendBuf;
    sendPkt->blockType = RSPTunnelOCD_PKT_SETUP;
    int pktSize = sizeof (RSPTunnelOCDPacketHeader);
    
    
    // Send out the packet over the RSPTunnelOCD's socket.
    int numSent = sendPacket (ocdState, sendBuf, pktSize);

    if (numSent < pktSize) {
        RSPTunnelOCD_PRINT_DEBUG3(DEBUG_LEVEL_WARNING,
                    "Error while sending setup packet. (Only %d of %d bytes sent.): %s\n",
                    numSent, pktSize, strerror (errno)
                );
        return false;
    }

    return true;
    
}

/** Send a setup reply OK packet to the remote end point of the specified ocdState.
  * @return true if the sending was successful, else return false.
  */
bool RSPTunnelOCDContext::sendSetupReplyOK (RSPTunnelOCDStateInfo* ocdState) {

    RSPTunnelOCD_PRINT_DEBUG2 (DEBUG_LEVEL_VERBOSE,
                "Sending setup reply packet to %s:%d.\n",
                inet_ntoa (ocdState->remoteAddress), ntohs (ocdState->remotePort)
            );

    char sendBuf[MAX_PACKET_BUF_LEN];
    
    // Construct the setup reply ok packet
    RSPTunnelOCDPacketHeader *sendPkt = (RSPTunnelOCDPacketHeader *) sendBuf;
    sendPkt->blockType = RSPTunnelOCD_PKT_SETUP_OK;
    int pktSize = sizeof (RSPTunnelOCDPacketHeader);
   
    int numSent = sendPacket (ocdState, sendBuf, pktSize);
    if (numSent < pktSize) {
        RSPTunnelOCD_PRINT_DEBUG3 (DEBUG_LEVEL_WARNING,
                    "Error while sending setup reply OK packet. (Only %d of %d bytes sent.): %s\n",
                    numSent, pktSize, strerror (errno)
                );
        return false;
    }

    return true;
    
}


/** Add the specified ocd state to the hash table of remote Address-->ocdstate mappings.
  * Note that we using only the remote IP address.  We use the port number to do a linear
  * search of the multiple entries retrieved from the multimap lookup.
  */
void RSPTunnelOCDContext::addRemoteHost (RSPTunnelOCDStateInfo* currOCDState) {
    
    map_fromAddr2OCDState.insert ( pair<uint32_t, RSPTunnelOCDStateInfo *>(currOCDState->remoteAddress.s_addr, currOCDState));

}


/**
  * Looks up the OCDState associated with the specified remote address and port number (in network byte order).
  * @return NULL if not found.
  */
RSPTunnelOCDStateInfo* RSPTunnelOCDContext::lookupRemoteHost (struct in_addr remoteAddress, uint16_t remotePort) {

    return lookupRemoteHost (remoteAddress.s_addr, remotePort);

}

/**
  * @param remoteIP network byte order.
  * @param remotePort network byte order.
  */
RSPTunnelOCDStateInfo* RSPTunnelOCDContext::lookupRemoteHost (uint32_t remoteIP, uint16_t remotePort) {
    
    multimap<uint32_t, RSPTunnelOCDStateInfo*>::iterator itr = map_fromAddr2OCDState.find (remoteIP);
    
    while (itr != map_fromAddr2OCDState.end()) {
        RSPTunnelOCDStateInfo* tmpOCD = itr->second;
        if (tmpOCD->remotePort == remotePort) {
            return tmpOCD;
        }
        itr++;
    }
    
    return NULL;
}

/**
  * This function is called when a setup failed packet is received from the specified remote host.
  * We just invoke the notifySetupFailed function on the ocdState associated with the remote host.
  */
void RSPTunnelOCDContext::processSetupFailPacket(const char* recvBuf, int recvLen, struct sockaddr_in remoteAddress) {

    // Lookup the OCDState associated with the remote host.
    RSPTunnelOCDStateInfo* currOCDState = lookupRemoteHost (remoteAddress.sin_addr, remoteAddress.sin_port);
    if (currOCDState == NULL) {
        RSPTunnelOCD_PRINT_DEBUG2 (DEBUG_LEVEL_WARNING,
                    "Unable to lookup remote host %s:%d while processing SetupFail packet.\n",
                    inet_ntoa (remoteAddress.sin_addr), ntohs (remoteAddress.sin_port)
                );
        return;
    }

    RSPTunnelOCD_PRINT_DEBUG2 (DEBUG_LEVEL_VERBOSE,
                "Received SetupFail packet from %s:%d.\n",
                inet_ntoa (remoteAddress.sin_addr), ntohs (remoteAddress.sin_port)
            );
    
    // Notify the OCI that the setup for this tunnel failed.
    currOCDState->notifySetupFailure();
}

/**
  * This function is called when a setup ok packet is received from the specified remote host.
  * We just invoke the notifySetupSuccess function on the ocdState associated with the remote host.
  */
void RSPTunnelOCDContext::processSetupOKPacket(const char* recvBuf, int recvLen, struct sockaddr_in remoteAddress) {

    // Lookup the OCDState associated with the remote host.
    RSPTunnelOCDStateInfo* currOCDState = lookupRemoteHost (remoteAddress.sin_addr, remoteAddress.sin_port);
    if (currOCDState == NULL) {
        RSPTunnelOCD_PRINT_DEBUG2 (DEBUG_LEVEL_WARNING,
                    "Unable to lookup remote host %s:%d while processing SetupOK packet.\n",
                    inet_ntoa (remoteAddress.sin_addr), ntohs (remoteAddress.sin_port)
                );
        return;
    }

    RSPTunnelOCD_PRINT_DEBUG2 (DEBUG_LEVEL_VERBOSE,
                "Received SetupOK packet from %s:%d.\n",
                inet_ntoa (remoteAddress.sin_addr), ntohs (remoteAddress.sin_port)
            );

    currOCDState->status = OCD_STATE_STATUS_READY;

    // Notify the OCI that the setup for this tunnel succeeded.
    currOCDState->notifySetupSuccess();
}


/**
  * Wrap the specified data in a RSPTunnelOCD DATA packet and send it across to the
  * remote end point as specified in ocdState.
  */
bool RSPTunnelOCDContext::sendDataPacket (RSPTunnelOCDStateInfo* ocdState, char* buf, int bufLen) {

    char tmpBuf[MAX_PACKET_BUF_LEN];
    RSPTunnelOCDPacketData* dataPkt = (RSPTunnelOCDPacketData *) tmpBuf;
    dataPkt->header.blockType = RSPTunnelOCD_PKT_DATA;
    memcpy (dataPkt->payload, buf, bufLen);

    int totalLen = sizeof (RSPTunnelOCDPacketData) + bufLen;

    RSPTunnelOCD_PRINT_DEBUG3 (DEBUG_LEVEL_VERBOSE,
                "RSPTunnelOCDContext: Sending data packet of length %d  to %s:%d.\n",
                totalLen,
                inet_ntoa (ocdState->remoteAddress), ntohs (ocdState->remotePort)
            );
    
    // Actually send the whole packet on the RSPTunnelOCD wire
    return (sendPacket (ocdState, tmpBuf, totalLen) == totalLen);
}


/**
  * Send the specified data through the specific OCDState/Tunnel.
  * @return The number of bytes sent.
  */
int RSPTunnelOCDContext::sendPacket (RSPTunnelOCDStateInfo* ocdState, char* buf, int bufLen) {
    struct sockaddr_in tmpDestAddr;
    tmpDestAddr.sin_addr = ocdState->remoteAddress;
    tmpDestAddr.sin_port = ocdState->remotePort;
   
    tmpDestAddr.sin_family = AF_INET;
    memset(&(tmpDestAddr.sin_zero), '\0', 8);  // zero the rest of the struct
    
    // Send out the packet over the RSPTunnelOCD's socket.
    int numSent = sendto (mySockFD, buf, bufLen, 0, (struct sockaddr *)& (tmpDestAddr), sizeof (struct sockaddr_in));

    return numSent;

}

int RSPTunnelOCDContext::findNumValidMappings() {
	if (rspServerFinder != NULL) {
		return rspServerFinder->findNumValidMappings();
	} else {
		return 0;
	}
}

/**
 * Removes all mappings/state associated with ocdState 
 */
void RSPTunnelOCDContext::destroyOCDState (RSPTunnelOCDStateInfo* ocdState) {

    map_fromAddr2OCDState.erase (ocdState->remoteAddress.s_addr);
}
