/**
 *  \file  RSPTunnelOCDContext.h
 *  \brief This is where the bulk of the RSPTunnelOCDSpecific processing takes place.
 */
#ifndef RSPTunnelOCD_CONTEXT_H
#define RSPTunnelOCD_CONTEXT_H 1

#ifndef _WIN32
#	include <sys/types.h>
#	include <sys/socket.h>
#else
#	include "netwrap.h"
#endif

#include <errno.h>

#include <map>

using namespace std;

class RSPTunnelOCD;
class RSPTunnelOCDStateInfo;
class RSPServerFinder;
class OCI;

#define RSP_GATEWAY_DOWN_POLICY_NODROP  100
#define RSP_GATEWAY_DOWN_POLICY_DROP  200


/**
  * This class contains the RSPTunnelOCD specific protocol and processing.
  * The RSPTunnelOCD class just acts as an interface to the OC-I, the bulk of the 
  * work happens here.
  */
class RSPTunnelOCDContext {

    /** The sample OCD listens for data on this connection. */
    int mySockFD;
   
    /** The port on which the varRSPTunnelOCD listens. 
     *  The port number is in host byte order.
     */
    uint16_t myPortNum;
  
    /** What should happens to new connections that are to use this RSPTunnel
      * if all the gateways are down?
      * RSP_GATEWAY_DOWN_POLICY_DROP, RSP_GATEWAY_DOWN_POLICY_NODROP
      */
    uint16_t gatewayDownPolicy;
    
    /** fromAddr --> RSPTunnelOCDState hash table.
      * The hash table is keyed by the IP address (network byte order).
      * There can be many OCDState entries with the same from IP address,
      * but with different port numbers.  We linearly go through all the
      * OCDStates with the same from IP while searching for a particular 
      * from port number.
     */
    multimap<uint32_t, RSPTunnelOCDStateInfo*> map_fromAddr2OCDState;


    /** Pointer to the OCD associated with this context. */
    RSPTunnelOCD* ocd;
    
    /** Used to find the IP address of the closest alive remote server proxy. */
    RSPServerFinder* rspServerFinder;
   
    /** The configuration file used for this OCD module */
    string configFileName;
    
    RSPTunnelOCDContext ( RSPTunnelOCD* varRSPTunnelOCD, string fileName );

    void receivedPacketFromOverlay();
    bool sendSetupReplyOK (RSPTunnelOCDStateInfo* ocdState);
    bool sendSetupPacket (RSPTunnelOCDStateInfo* ocdState);
    void processSetupPacket(const char* recvBuf, int recvLen, struct sockaddr_in fromAddr);
    void processDataPacket(char* recvBuf, int recvLen, struct sockaddr_in remoteAddress);
    void processSetupOKPacket(const char* recvBuf, int recvLen, struct sockaddr_in remoteAddress);
    void processSetupFailPacket(const char* recvBuf, int recvLen, struct sockaddr_in remoteAddress);

    void init();

    void addRemoteHost (RSPTunnelOCDStateInfo* currOCDState);
    RSPTunnelOCDStateInfo* lookupRemoteHost (struct in_addr remoteAddress, uint16_t remotePort);
    RSPTunnelOCDStateInfo* lookupRemoteHost (uint32_t remoteIP, uint16_t remotePort);

    int sendPacket (RSPTunnelOCDStateInfo* ocdState, char* buf, int bufLen);
    bool sendDataPacket (RSPTunnelOCDStateInfo* ocdState, char* buf, int bufLen);

    RSPTunnelOCDStateInfo* setupTunnel(string endPointName, OCIHandle ociHandle, OCI* oci);

	void destroyOCDState (RSPTunnelOCDStateInfo* ocdState);

	int findNumValidMappings();
    friend class RSPTunnelOCD;
    friend class RSPTunnelOCDStateInfo;
};

#endif
