#include <sys/types.h>
#include <sys/socket.h>
#include <assert.h>

#include "OCD_include.h"

#include "HipOCD.h"
#include "HipOCDStateInfo.h"

HipOCD::HipOCD(
	       string type, string id, string desc, 
	       string fileName, 
	       string urlSuffixes, bool handleLegacyURLs
	       ) :OCD(    type, id, desc, fileName, 
			  urlSuffixes, handleLegacyURLs
            ) {
}

/**
  * Construct a new HipOCD with fields taken from the
  * specified overlay network.
  */
HipOCD::HipOCD(OCD* o): OCD() {
  this->type = o->type;
  this->id = o->id;
  this->description = o->description;
  this->urlSuffixes = o->urlSuffixes;
  this->handleLegacyURLs = o->handleLegacyURLs;
  
  this->configPropertyMap = o->configPropertyMap;
  this->myNameList = o->myNameList;
}



/**
  * Initializes the HipOCD 
  * @param proxy A pointer to the proxy of which this overlay is a part.
  * @param oci  The OCI associated with this OCD.
  */
void HipOCD::init(Proxy *prxy, OCI* oci, bool recoverFromCrash) {
  socklen_t l;
  this->proxy = prxy;
  this->oci = oci;
  sockaddr_in my_addr;
     
  // read any OCD specific configuration files.
  //char tmpIPBuf[MAX_PARAM_LEN];
  //ConfigFile configFile = ConfigFile(configFileName, HIP_CFG_TAG);
  //configFile.readStringParam("//HIPConfigFile/ProxyIPAddress", tmpIPBuf, true);
  //configFile.readStringAttrib("//HIPConfigFile", "ProxyIPAddress", tmpIPBuf, true);
  //bzero((char *) &proxy_addr, sizeof(proxy_addr));
  //proxy_addr.sin_family = AF_INET;
  //proxy_addr.sin_addr.s_addr = inet_addr(t;
  //proxy_addr.sin_port = htons(HIP_PORTNO);
  
  sockFD =  socket (AF_INET, SOCK_DGRAM, 0); // the socket on which the OCD receives packets
  
  bzero((char *) &my_addr, sizeof(my_addr));
  my_addr.sin_family = AF_INET;
  my_addr.sin_addr.s_addr = INADDR_ANY;
  my_addr.sin_port = htons(HIP_PORTNO);
  
  // bind the socket
  if (bind(sockFD, (struct sockaddr *) &my_addr,
	   sizeof(my_addr)) < 0) {
    PRINT_INFO(
	       INFO_LEVEL_FATAL_ERROR,
	       "ERROR binding");	      
    EXIT_ON_ERROR;
  }             
    
  // The OCD listens for packets on some socket, say sockFD.
  
  // Add a callback to be invoked when a packet is received on this socket.
  TObjectFunctor<HipOCD> *fdfunc = new TObjectFunctor<HipOCD>(this, &HipOCD::receivedPacketFromOverlay);
  addFDCallback(sockFD, fdfunc, true);
  //printf("Initialized HipOCD, proxy ip=%ul\n", proxy_addr.sin_addr.s_addr);
}


/**
  * Gives a string representation (pretty printed) of the HipOCD.
  * @param tmpStr a character buffer in which the string is stored.
  * Sufficient space should be allocated for tmpStr.
  * @return pointer to tmpStr
  */
char* HipOCD::toString(char *tmpStr) {

    sprintf( tmpStr,
            "Hip Overlay Network\n"
            "\ttype\t\t:\t%s\n"
            "\tid\t\t:\t%s\n"
            "\tdescription\t:\t%s\n"
            "\tconfig file\t:\t%s\n",
            type.c_str(),
            id.c_str(),
            description.c_str(),
            configFileName.c_str()
           );
    return tmpStr;
}

/**
  * Setup up a tunnel over the hipOCD to the specified name.
  * If a tunnel already exists, reuse it if it makes sense to do so in the specific overlay.
  * @param ociHandle The oci to be notified through a callback when the tunnel setup succeeds
  * or fails.
  * @parentLayer The layer which is setting up this new layer.
  * @return a handle to the tunnel.
  */
OCDHandle HipOCD::setup (
                            const string endPointName, 
                            int endPointNameType,
                            char *optionBlocks, 
                            int optionLen, 
                            OCIHandle ociHandle,
                            OCI* oci) {
   
  //printf("\nHipOCD setup\n");
  
  if (endPointNameType == OCI_NAME) {
    // The end point specified in the parameter endPointName is 
    // the name of an end point.  It should be resolved to
    // an OCD specific ID in an OCD specific manner.
    
    // Check if there already exists a tunnel to the end point
    HipOCDStateInfo *ocdState = (HipOCDStateInfo *) lookupEndPoint (endPointName);
    OCDHandle ocdHandle = NULL_HANDLE;
    
    if (ocdState == NULL) {
      char cmd[MAX_PRINT_BUF];
      // No tunnel exists,  create a new one.
      
      // First get a new unique handle.
      ocdHandle = getNewOCDHandle();
      
      // Now create the OCD State for this new tunnel
      ocdState = new HipOCDStateInfo (ocdHandle, oci, this);
      
      //add to the list of ocd states present in this OCD.
      addOCDState (ocdState);
      
      PRINT_DEBUG (
		   DEBUG_LEVEL_SUPER, 
		   "In HipOCD setup, creating a new tunnel handle %d, endpointName  = %s\n", 
		   ocdHandle,
		   ocdState->endPointName
                );
      
      // Add the specified ociHandle to the list of OCIStates waiting for this OCD setup to complete
      ocdState->addParentHandle (ociHandle);

      // Add the OCDState of the newly created tunnel to the handle-->state hash table
      putStateInfo (ocdHandle, ocdState);
      
      // NOTE: optionBlocks are currently not used.
      ocdState->setupOptionBlocks = optionBlocks;
      ocdState->setupOptionLen = optionLen;

      // Strip off .hip suffix before trying to resolve URL to IP
      char reqURL[MAX_URL_LEN];
      strcpy (reqURL, endPointName.c_str());
      stripSuffix (reqURL, ".hip");   //TODO

      if (proxy->makeDNSRequest (ocdState, reqURL)) {
	
	//DNS successfully sent to the name server of the machine on which the proxy is running
	
      } else {
	PRINT_INFO (
		    INFO_LEVEL_WARNING,
		    "Unable to send DNS requested to the name server of this machine.\n"
		    );
      }
    } else {
      
      // Tunnel to end point already exists.  Reuse it if it makes sense to do
      // so in the current overlay.
      // sampleOCDReuseTunnel(ocdState);
      
      ocdHandle = ocdState->myHandle;    
    }
    
    return ocdHandle;
    
  } else {
    
    // The endPointName specified in the setup call is an i3 id or i3 stack.
    // There is no need for resolution.
    // Steps similar to if part.
  }
}
    
/**
  * This function is called when some other layer wants to send anything via HIP.
  */
void HipOCD::send ( OCDHandle handle, char *data, int dataLen) {
  //char tmp[MAX_PACKET_SIZE];
  HipOCDStateInfo *ocdState = (HipOCDStateInfo *) getStateInfo (handle);
  sockaddr_in peer_addr;

  bzero((char *) &peer_addr, sizeof(peer_addr));
  peer_addr.sin_family = AF_INET;
  peer_addr.sin_port = htons(HIP_PORTNO);
  peer_addr.sin_addr = ocdState->peer_addr;
  
  if (ocdState == NULL) {
    PRINT_DEBUG (
		 DEBUG_LEVEL_MINIMAL,
		 "Trying to send through unknown OCD with handle %d\n",
		 handle
		 );
    return;
  } 
  //memcpy(tmp, &ocdState->peer_addr, sizeof(ocdState->peer_addr));
  //memcpy(tmp+sizeof(ocdState->peer_addr), data, dataLen);

  // Send the dataLen bytes contained in dataLen through the specific OCD.
  if (sendto(sockFD, data, dataLen, 0, (sockaddr *) &peer_addr, 
	     sizeof(peer_addr)) < 0) {
    PRINT_INFO(
	       INFO_LEVEL_FATAL_ERROR,
	       "ERROR sending");	      
	      EXIT_ON_ERROR;
  }
}


/**
  * Initializes the state necessary for inbound names associated with this 
  * Hip OCD.
  */
void HipOCD::initMyNames () {

  // Iterate over the incoming names assocated with this OCD 
  list<MyName *>::iterator itr = myNameList.begin();
  while (itr != myNameList.end()) {
    
    MyName *currName = (*itr);
    
    // The OCD specific properties associated with the name
    // can be retrieved by invoking getProperty on the MyName object.
    //char *sampleProp = currName->getProperty (SAMPLE_PROPERTY_NAME); 
    
    // OCD specific inbound name setup steps.
    
    itr++;
  }
}

/**
 * This function is called when a packet is received over the hipOCD 
 * architecture.
 */
void HipOCD::receivedPacketFromOverlay() {
  char buf[MAX_PACKET_SIZE];
  int n;
  socklen_t l;
  char cmd[MAX_PRINT_BUF];
  sockaddr_in peer_addr;
  char peer_addr_a[MAX_PRINT_BUF]; 

  // Read the packet from the socket on which HipOCD was expecting the packet,
  // i.e. the socket on which the packet was expected.
  l = sizeof(peer_addr);
  n = recvfrom(sockFD, buf, sizeof(buf), 0, (sockaddr*) &peer_addr, &l);

  //memcpy(&peer_addr, buf, sizeof(peer_addr));
  //strcpy(proxy_addr_a, inet_ntoa(proxy_addr.sin_addr));	

  strcpy(peer_addr_a, inet_ntoa(peer_addr.sin_addr));

  printf("\nHipOCD receive, peer=%s\n", peer_addr_a);

  HipOCDStateInfo *ocdState = 
	(HipOCDStateInfo *) lookupEndPoint(peer_addr_a);

  if (ocdState == NULL) {
        //This remote sender is unknown, i.e. we haven't seen him before,
        //So create a new OCDState.

        // Create new OCD state
        OCDHandle ocdHandle = getNewOCDHandle();
        ocdState = new HipOCDStateInfo (ocdHandle, oci, this);
        putStateInfo (ocdHandle, ocdState);

        ocdState->peer_addr = peer_addr.sin_addr;
        ocdState->hipOCD = this;

	// add newly created state to hash table keyed by end point name.
	putEndPoint  (peer_addr_a, ocdState);

      	// set the name of the end point of the tunnel.
      	strcpy (ocdState->endPointName, peer_addr_a);

        //addRemoteHandle (ocdState->remoteHandle, ocdState);

	snprintf(cmd, sizeof(cmd),
		 "echo \"spdadd %s %s any -P out ipsec hip/transport//require;\" >/tmp/hipocd ", 		     
		 inet_ntoa(proxy->get_my_addr_nf()),
		 peer_addr_a);	
	printf(cmd);
	system(cmd);
	system(SET_KEY);
  }

  assert (ocdState != NULL);

  // Process the received packet.  If the packet is not an OCD specific control packet,
  // hand off the packet to the OCI after stripping of the OCD specific headers.
  oci->receive (buf, this, ocdState->myHandle);
  // handle is the handle of the ocd state corresponding to the tunnel on which the packet was received.
}

// Register the OCD with the proxy when the code is dynamically loaded
// The following line sets the name of the OCD's main class
#define MY_OCD_CLASS_NAME HipOCD
// The following line sets the type name of the OCD
#define MY_OCD_TYPE_STR "HIP"
#include "OCDRegister.h"
