/* Purpose: To perform indirect measurements with i3.
 * Assumes that NEWS_INSTRUMENT is turned on.
 * Hacked so that chord routing is not used.
 *
 * - RTT: computed and validated locally
 * - loss: computed locally, validated using server logs
 * - bandwidth: computed locally, validated using server measurements
 */

#include <stdio.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

extern "C" {
#include "i3.h"
#include "i3_id.h"
#include "i3_addr.h"
#include "i3_client_api.h"
#include "i3_utils.h"
#include "utils.h"
#include "gen_utils.h"
}
    
using namespace std;

#include <list>
#include <map>
#include <string>
#include <vector>

// List of IDs associated with a pair of nodes
typedef struct IDList {
    ID id1a;
    ID id2;
    ID id1b;
} IDList;

// i3 server's add and port
class i3_server {
    public:
	uint32_t addr;
	uint16_t port;

	void Print() {
	    struct in_addr ia; ia.s_addr = htonl(addr);
	    printf("%s:%d", inet_ntoa(ia), port);
	}
};


// i3 server list structure
list<i3_server> i3_server_list;

// list of IDs for each measurement pair
list<IDList> measurement_id_list;

/*********************************************************************
 * Purpose (helper function): Tokenize a string based on a delimiter
 ********************************************************************/
void Tokenize(const string& str, vector<string>& tokens,
        const string& delimiters)
{
    // Skip delimiters at beginning.
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos)
    {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, lastPos);
    }
}

/*********************************************************************
 * Read list of servers from the server_list file
 * Each line has entry of the form IP:port 
 ********************************************************************/
#define CHARS_PER_NODE 25
int read_server_list(char *server_list)
{
    FILE *fp = fopen(server_list, "r");
    static char buf[CHARS_PER_NODE];
    i3_server i3s;

    while (!feof(fp)) {
	if (fgets(buf, CHARS_PER_NODE, fp) == 0) break;

	vector<string> tokens;
        string str(buf);
        Tokenize(str, tokens, ":");
	
	vector<string>::iterator iter = tokens.begin();
        i3s.addr = ntohl(inet_addr(iter->c_str()));
        iter++;
        i3s.port = (atoi(iter->c_str()));

	i3_server_list.push_back(i3s);
    }

    fclose(fp);

    list<i3_server>::iterator iter;
    for (iter = i3_server_list.begin(); iter != i3_server_list.end(); iter++) {
	(*iter).Print(); printf("\n");
    }
}


/*********************************************************************
 * Purpose: Pack/unpack i3-server in the prefix of an ID
 ********************************************************************/
void pack_server2id(ID *id, i3_server *i3s)
{
    i3_addr a; uint16_t length;
    a.type = I3_ADDR_TYPE_IPv4;
    a.t.v4.addr.s_addr = i3s->addr;
    a.t.v4.port = i3s->port;
    pack_i3_addr((char *)id->x, &a, &length);
}

void unpack_id2server(ID *id, i3_server *i3s)
{
    uint16_t length;
    i3_addr *a = unpack_i3_addr((char *)id->x, &length);
    assert(I3_ADDR_TYPE_IPv4 == a->type);
    i3s->addr = a->t.v4.addr.s_addr;
    i3s->port = a->t.v4.port;
    free_i3_addr(a);
}

/*********************************************************************
 * Purpose: Callback when packet is received for an id
 ********************************************************************/
void receive_payload(cl_trigger *t, void* data, void *fun_ctx)
{
  long seq;
  cl_buf *clb = (cl_buf *) data;

  if (clb->data_len) {
    seq = ntohl(*((long *)(clb->data)));
    printf("RECV "); printf_i3_id(&(t->t->id), 0);
    printf(" %ld %Ld\n", seq, wall_time());
  }
}
void constraint_failed(cl_trigger *t, void *data, void *fun_ctx)
{
  printf("Trigger constraint failed\n");
}
void trigger_inserted(cl_trigger *t, void *data, void *fun_ctx)
{
  printf("Trigger inserted\n");
}
void trigger_failure(cl_trigger *t, void *data, void *fun_ctx)
{
  printf("Trigger failed\n");
}


/*********************************************************************
 * Purpose: Form IDs for measurement
 ********************************************************************/
void insert_ids()
{   
    list<i3_server>::iterator iter1;
    list<i3_server>::iterator iter2;
    Key key;
    
    for (iter1 = i3_server_list.begin(); iter1 != i3_server_list.end(); iter1++) {
	//for (iter2 = iter1; iter2 != i3_server_list.end(); iter2++) {
	for (iter2 = i3_server_list.begin(); iter2 != i3_server_list.end(); iter2++) {
	    if (iter1 == iter2) continue;

	    IDList idl;
	    cl_trigger *t1, *t2, *t3, *t4, *t5;

	    set_random_id(&(idl.id1a)); pack_server2id(&(idl.id1a), &(*iter1));
	    set_random_id(&(idl.id2)); pack_server2id(&(idl.id2), &(*iter2));
	    set_random_id(&(idl.id1b)); pack_server2id(&(idl.id1b), &(*iter1));
    
	    // (id1, id2); r-constrained; id2 will be updated
	    t1 = cl_create_trigger_id(&(idl.id1a), ID_LEN_BITS, 
		    &(idl.id2), CL_TRIGGER_CFLAG_R_CONSTRAINT);
    
	    // (id1, R)
	    t2 = cl_create_trigger(&(idl.id1a), ID_LEN_BITS, 
		    &key, CL_TRIGGER_CFLAG_R_CONSTRAINT);
	    
	    // (id2, R)
	    t3 = cl_create_trigger(&(idl.id2), ID_LEN_BITS, 
		    &key, CL_TRIGGER_CFLAG_R_CONSTRAINT);
	    
	    // (id2, id1')
	    t4 = cl_create_trigger_id(&(idl.id2), ID_LEN_BITS, 
		    &(idl.id1b), CL_TRIGGER_CFLAG_R_CONSTRAINT);

	    // (id1', R)
	    t5 = cl_create_trigger(&(idl.id1b), ID_LEN_BITS, 
		    &key, CL_TRIGGER_CFLAG_R_CONSTRAINT);

	    // print IDs to log file
	    printf("IDL ");
	    (*iter1).Print(); printf(" ");
	    (*iter2).Print(); printf(" ");
	    printf_i3_id(&(idl.id1a), 0); printf(" ");
	    printf_i3_id(&(idl.id2), 0); printf(" ");
	    printf_i3_id(&(idl.id1b), 0); printf("\n");

	    measurement_id_list.push_back(idl);
  
	    // insert triggers
	    cl_register_trigger_callback(t1, CL_CBK_RECEIVE_PAYLOAD, 
			       receive_payload, NULL);
	    cl_register_trigger_callback(t2, CL_CBK_RECEIVE_PAYLOAD, 
			       receive_payload, NULL);
	    cl_register_trigger_callback(t3, CL_CBK_RECEIVE_PAYLOAD, 
			       receive_payload, NULL);
	    cl_register_trigger_callback(t4, CL_CBK_RECEIVE_PAYLOAD, 
			       receive_payload, NULL);
	    cl_register_trigger_callback(t5, CL_CBK_RECEIVE_PAYLOAD, 
			       receive_payload, NULL);
	    cl_insert_trigger(t1, 0);
	    cl_insert_trigger(t2, 0);
	    cl_insert_trigger(t3, 0);
	    cl_insert_trigger(t4, 0);
	    cl_insert_trigger(t5, 0);
	}
    }
}

/*********************************************************************
 * Purpose: Send packet to an ID by rotation
 ********************************************************************/
#define MEASUREMENT_PERIOD 1 * 1000000ULL
#define MY_PKT_SIZE 2000
void send_measurement()
{
    static int seq = 0;
    static cl_buf  *clb = cl_alloc_buf(MY_PKT_SIZE);
    
    static uint64_t last_ping_time = wall_time();
    static list<IDList>::iterator iter = measurement_id_list.begin();
    assert(measurement_id_list.begin() != measurement_id_list.end());

    if (iter == measurement_id_list.end()) {
	++seq;
	iter = measurement_id_list.begin();
    }

    uint64_t curr_time = wall_time();
    int num_send = MIN(measurement_id_list.size(),
	    (curr_time-last_ping_time)/(MEASUREMENT_PERIOD/measurement_id_list.size()));
    printf("Num send = %d\n", num_send);
    curr_time = last_ping_time;

    for (int i = 0; i < num_send; i++, iter++) {
	long temp = htonl(seq);
	memcpy(clb->data, &temp, sizeof(long));
	clb->data_len = 100;
	cl_send(&((*iter).id1a), clb, 0);
	printf("SENT ");
	printf_i3_id(&((*iter).id1a), 0);
	printf(" %ld %Ld\n", seq, wall_time());
    }
}

/*********************************************************************
 * Purpose: send_packet for each of the IDs and receive/log replies
 ********************************************************************/
#define TO_SEND_USEC 1000000
#define TO_SEND_SEC  0
int i3_main_loop(char *cfg_file, char *server_list)
{
    // read server list from file
    read_server_list(server_list);
    
    // initialize context
    cl_init(cfg_file);

    // create and insert all necessary ids for measurement
    insert_ids();

    // loop send/receive
    fd_set rset; int rc; ID id;
    struct timeval select_to;
    select_to.tv_usec = TO_SEND_USEC;
    select_to.tv_sec = TO_SEND_SEC;

    FD_ZERO(&rset);
    for (;;) {
	FD_SET(0, &rset); /* just here, to be modified if application
			   * listens on other fds */
	if ((rc = cl_select(0, &rset, NULL, NULL, &select_to)) < 0) {
	    if (errno == EINTR)
		continue;
	    else
		err_sys("select_error\n");
	}

	// send data to an id by rotation
	send_measurement();
	
	// reinitialize the timer
	select_to.tv_usec = TO_SEND_USEC;
	select_to.tv_sec = TO_SEND_SEC;
    }
  
    // destroy context
    cl_exit();
}


int main(int argc, char **argv)
{
    if (argc != 3) {
	printf("%s <dummy_cfg_file.xml> <server_list>\n", argv[0]);
	exit(-1);
    }

    i3_main_loop(argv[1], argv[2]);
    return -1;
}
