/*
 * This program (in conjunction to send_shortcut) illustrates the 
 * use of shortcuts. If the sender requests shortcuts and the receiver
 * allows shortucts then i3 is used as a lookup infrastructure rather
 * than as a forward infrastructure, i.e., after the first packet reaches
 * the receiver, the sender caches the receiver and sends the subsequent
 * packets dirtectly to the receiver. The shortcuts are disabled if (1)
 * the sender is behind a NAT, and (2) if either the sender or the receiver
 * use tcp to connect to i3.
 * Usage:
 * > recv_shortcut <cfg_file.xml>
 * > send_shortcut  <cfg_file.xml>
 */ 

#include <stdio.h>

#include "i3.h"
#include "i3_client_api.h"
#include "utils.h"

/* callbacks */
void receive_payload(cl_trigger *t, void *data, void *fun_ctx)
{
  long seq;

  cl_buf *clb = (cl_buf *) data;
  if (clb->data_len) {
    seq = ntohl(*((long *)(clb->data)));
    printf("recv. seq. #: %ld \n", seq);
  }
}

void constraint_failed(cl_trigger *t, void *data, void *fun_ctx)
{
  printf("Trigger constraint failed\n");
}

void trigger_inserted(cl_trigger *t, void *data, void *fun_ctx)
{
  printf("Trigger inserted\n");
}

void trigger_failure(cl_trigger *t, void *data, void *fun_ctx)
{
  printf("Trigger failed\n");

  /* reinsert trigger */
  cl_insert_trigger(t, 0);
}

int i3_main_loop(char *cfg_file)
{
  ID id;
  cl_trigger *t;
  Key key;
  
  /* initialize context */
  cl_init(cfg_file);

  /* 
   * Create and insert triggers (id, R) respectively.
   * All triggers are r-constrained 
   */
  init_id_fromstr(&id, "this is a test id");

  /* make the trigger right constrained;
   * note that right constrained triggers are subject
   * to eavesdropping and impersonation -- that's why right
   * constrained triggers are appropiate only for private triggers.
   * for public triggers use left constrained triggers
   */ 
  t = cl_create_trigger(&id, ID_LEN_BITS, &key, CL_TRIGGER_CFLAG_R_CONSTRAINT);
  
  /* print trigger's ID */
  printf("trigger's ID: "); print_id(&id);
  /* insert trigger */
  cl_insert_trigger(t, CL_IFLAGS_TRIGGER_ALLOW_SHORTCUT);

  /* associate callbacks with the inserted trigger */
  cl_register_trigger_callback(t, CL_CBK_TRIGGER_CONSTRAINT_FAILED, 
			       constraint_failed, NULL);
  cl_register_trigger_callback(t, CL_CBK_RECEIVE_PAYLOAD, 
			       receive_payload, NULL);
  cl_register_trigger_callback(t, CL_CBK_TRIGGER_INSERTED, 
			       trigger_inserted, NULL);
  cl_register_trigger_callback(t, CL_CBK_TRIGGER_REFRESH_FAILED, 
			       trigger_failure, NULL);

  /* loop forever */
  cl_loop();

  /* remove & destroy trigger */
  cl_destroy_trigger(t);
  
  /* destroy context */
  cl_exit();
}


int main(int argc, char **argv)
{
  if (argc != 2) {
    printf("%s <cfg_file.xml>\n", argv[0]);
    exit(-1);
  }

  i3_main_loop(argv[1]);
  return -1;
}

