#!/usr/bin/perl
# This script is used start i3 servers on a set of PlanetLab nodes.
# Usage:
# start_i3_servers.pl
#                   The config directory is specified via the "conf_dir" attribute in deployment.txt.
#                   The config directory contains the different configuration files
#                   for the different planetlab nodes on which an i3 server is to be started.
#                   I3 servers are started only all planetlab nodes for which cfg files exist.
#                   The cfg files are generated by the gen_conf program.  The file names
#                   follow a specific pattern.
#                   -h <host_file>
#                   If the host_file is not specified, i3 servers are started on all nodes
#                   corresponding to the config files in config_directory.  If the host_file
#                   is specified, i3 servers are started only on the nodes specified in host_file.
#
# This is the sequence of actions performed by the script.
# 1. Copy the i3 server binary to the planetlab node.  The executable is by default called
#    i3serverd.  You can specify a different name; but then you will have to make a similar change
#    in multiple scripts (for automatic monitoring, stopping etc).  It is best to leave it unchanged.
# 2. Kill any currently running i3 server on the planet lab node.
# 3. Start the newly copied i3 server.


use strict 'vars';
package main;

my $i3server_check_script = "i3server_check.pl";

use Getopt::Std;
use Cwd;

$SIG{ALRM} = sub { timeoutSSH(); };

sub timeoutSSH () {
    die "SSH timed out.  Moving on.\n";
}

my $ssh_options = "-o \"ConnectTimeout=20\" -o \"KeepAlive=yes\" -o \"PreferredAuthentications=publickey\" -o \"StrictHostKeyChecking=no\" -n";
my $scp_options = "-o \"ConnectTimeout=20\" -o \"KeepAlive=yes\" -o \"PreferredAuthentications=publickey\" -o \"StrictHostKeyChecking=no\"";


require "read_deployment.pl";

my %deployment_details;
%deployment_details = &get_deployment_details;

# This is the default (and highly recommended) name for the i3 server binary.
my $deployment_binary = $deployment_details{"deployment_binary"};
my $deployment_username = $deployment_details{"deployment_username"};
my $deployment_dir = $deployment_details {"deployment_dir"};
my $conf_dir = $deployment_details {"conf_dir"};

print ("\ndeployment_binary = $deployment_binary\n");
print ("deployment_username=$deployment_username\n");
print ("deployment_dir=$deployment_dir\n");


#get command line options
use vars qw($opt_h);
getopts('h:');


# Read in the host file into hostsToStart array.
my @hostsToStart;
my $hostsToStartFlag = 0;	#We will look at hostsToStart array only if this flag is 1.
if ( -f $opt_h) {
    open HOSTS_TO_START, $opt_h || die "Unable to read in hosts to start from $opt_h.\n";
    @hostsToStart = <HOSTS_TO_START>;
    $hostsToStartFlag = 1;
}

chdir ($conf_dir) || die "Cannot change to conf directory $conf_dir\n";
opendir (CONF_DIR, ".") || die "Cannot open conf directory\n";

my $filename;
# Go through each .cfg file in the specificed config directory
while ($filename = readdir(CONF_DIR)) {
    
    if ($filename =~ "cfg") {
	    
        my @arr;
        @arr = split(/_/, $filename);

	    my $currHost = $arr[1];
	    my $currPort = $arr[2];

        my $startOnThisHostFlag = 1;  #i3 server will be started on currHost only if this flag is 1.
    
	    if ($hostsToStartFlag == 1) {
            $startOnThisHostFlag = 0;
	        # Check if the currHost is in hostsToStart.
            my $tmpHostPort;
            foreach $tmpHostPort (@hostsToStart) {
                chomp $tmpHostPort;
                my $tmpHost, my $tmpPort;
                ($tmpHost, $tmpPort) = split (/ /, $tmpHostPort);
		        if ( ($tmpHost eq $currHost) && ($tmpPort == $currPort)) {
                    $startOnThisHostFlag = 1;
                }
            }
	    }
    

        if ($startOnThisHostFlag == 1) {
            alarm (60);
            eval {
                startI3Server ($currHost, $currPort, $filename);
            };
            alarm (0);
        }
    }
}
close CONF_DIR;

sub startI3Server {
    my $currHost = shift (@_);
    my $currPort = shift (@_);
    my $cfgFileName = shift (@_);

    my $cmd;
    print "\n\n\n###############################################################\n";
    print "Starting i3 on host $currHost at port $currPort\n";
	$cmd = "ssh $ssh_options $deployment_username\@$currHost \"mkdir $deployment_dir; $deployment_dir/kill-process.pl -p $currPort -P $deployment_binary\"";
	system($cmd);
	
    # A planetlab node specific i3 server start script is created and uploaded to the node.
    # This script with the name $i3server_start_script is used to start the i3 server now
    # and also during automatic restarts.
    my $i3server_start_script = "start_i3server.pl";

   
	&generate_i3server_check_script ($currPort, $deployment_binary);
 	system ("chmod +x $i3server_check_script");

    open(OUTFILE,">$i3server_start_script") || die("Cannot Open File $i3server_start_script");
    print OUTFILE "date >> i3server_start.log\n";
	print OUTFILE "./kill-process.pl -p $currPort -P $deployment_binary;\n";
    print OUTFILE "nohup ./$deployment_binary $cfgFileName $currPort 2>&1>/dev/null &\n"; 
        
    close(OUTFILE);
   
    system("chmod +x $i3server_start_script");

	$cmd = "scp $scp_options -C ../$deployment_binary $i3server_start_script $i3server_check_script ../kill-process.pl $cfgFileName $deployment_username\@$currHost:$deployment_dir";
	print "$cmd\n";
	system($cmd);
    
    system("rm -f $i3server_start_script");
    
	$cmd = "ssh $ssh_options $deployment_username\@$currHost \"cd $deployment_dir; ulimit -c 5000; chmod +x ./$i3server_start_script; ./$i3server_start_script &\" &";
	print $cmd."\n";
	system($cmd);
	print "\n";
    print "-----------------------------------------------------------------------\n\n";
	    
    sleep(10);
}

# This subroutine generates the i3server_check.pl script tailormade for this deployment.
# It uses the i3server_check.pl.template file for this purpose.
sub generate_i3server_check_script {

	my $portNum = $_[0];
	my $processName = $_[1];

	open (OUTSCRIPT, ">i3server_check.pl") || die "Unable to open i3server_check.pl for writing.\n";

	print OUTSCRIPT "#!/usr/bin/perl\n";
	print OUTSCRIPT "use strict \'vars\';\n";
	print OUTSCRIPT "my \$portNum=$portNum;\n";
	print OUTSCRIPT "my \$processName=\"$processName\";\n";
	print OUTSCRIPT "my \$deployment_dir=\"$deployment_dir\";\n";

	open (TEMPLATESCRIPT, "../i3server_check.template") || die "Unable to open file.\n";
	my $tmpLine;
	while ($tmpLine = <TEMPLATESCRIPT>) {
		print OUTSCRIPT $tmpLine;
	}
	close (TEMPLATESCRIPT);
	close (OUTSCRIPT);
}
