#ifndef _BROADCAST_MEMBERSHIP_H_
#define _BROADCAST_MEMBERSHIP_H_

/*
 * What do I need for this?
 * - Peers announce themselves via broadcasts.
 * - People cache the list of peers they've heard from and expire them
 *   on some hourly basis or whatever.
 * --> Periodic "Hello, world, I'm alive!" announcements.
 * --> How to handle bootstrapping?
 *  --> A "membership request" message that requests transmission of
 *      the existing membership domain?
 * Shrug.
 */

#include <stdio.h>

#include "ron.h"
#include "membership.h"
#include "flood_protocol.h"
#include "scheduler.h"

#define MEMBERSHIP_NEW 0x1	/* I need to be updated */
#define MEMBERSHIP_UPDATE 0x2	/* I'm just beaconing */

#ifndef UPDATE_FREQUENCY
#define UPDATE_FREQUENCY 180000000 /* Every 3 minutes */
#endif
#ifndef EXPIRE_TIME
#define EXPIRE_TIME 360 /* In seconds!  One hour */
#endif
#define SPECIAL_UPDATE_TIMER 3000000
#define ONEHOP_SECONDS 1 /* One second of age per hop */

#define OP_JOIN  0x00
#define OP_LEAVE 0x01


namespace RON {

struct ron_membership_header {
    u_int8_t type;
    u_int8_t nmembers;
};

struct ron_membership_entry {
    struct in_addr src_addr;
    fixedpoint age; /* How long has this been floating around the net? */
    u_int16_t src_port;
    u_int8_t op;  /* leave, join, etc. */
};

class node_data {
public:
    node_data() { last_update = fixedpoint(0,0); expires = 1; }
    fixedpoint last_update;
    int expires; /* 0 or 1.. 0 if present in peers.txt */
};

class broadcast_membership : public membership, time_callable {
public:
    void configure(Scheduler *s, forw *f);
    broadcast_membership() { infilename = NULL; init(); }
    broadcast_membership(char *filename) {
        infilename = strdup(filename);
        init();
    }

    broadcast_membership(char *filename, flood_protocol *nfp) {
	infilename = filename;
	fp = nfp;
	init();
    }
    virtual ~broadcast_membership() { 
        if (infilename) {
            free(infilename);
        }
    }

    int reload();
    int last_changed() { return 0; }

    void init();
    void dump(FILE *out);
    int recv(pkt *mypkt, node *, int via_ron);
    int timer_callback(void *data);


   /* (SAR) Will need to be overwritten in NS emulation */
    virtual int gettimeofday(struct timeval *tv, struct timezone *tz)
	{
	    return ::gettimeofday(tv, tz);
	}

    /* (SAR) Will need to be overwritten in NS emulation */
    virtual long int random(void)
	{
	    return ::random();
	}

    list<node> *peers() {
	return &peerlist; /* XXX:  This seems stupid */
    }
    int initlist(char *fname, list<node> *uplist);
    int savelist(char *fname, list<node> *uplist);
    list<node> peerlist;
    list<node> peercache;
    int pkt_create(pkt *mypkt);
private:
    Scheduler *sched;
    map<node, node_data> peermap;
    flood_protocol *fp;
    char *infilename;
    int clean_peers();
    int send_announce(list<node> *tolist);
    int awake;
    int special_request_outstanding;
    fixedpoint nowpoint;
};

}

#endif /* _BROADCAST_MEMBERSHIP_H_ */
