/*
 * RON forwarding daemon
 *
 * Author:  David G. Andersen
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <stdarg.h>
#include <ctype.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/time.h>
#include <netinet/udp.h>
#include <string.h>
#include <sys/uio.h>
#include "plumber.h"
#include "rondebug.h"
#include "forw.h"
#include "ron.h"
#include "ronlib.h"
#include "packet.h"

using namespace RON;


/*
 * init opens the socket and binds it to the specified port.
 */

int forw::init(int ronport)
{

	debug = 0xfffffff;

    mylocalhost = new localhost();
    mylocalhost->init(); // (SAR)
    if (!mylocalhost) {
	perror("Could not allocate localhost");
	exit(-1);
    }
    mylocalhost->configure(sched, this);
    mylocalhost->add_addr(myaddr.s_addr);

    port = htons(ronport);

    sock = bound_udp_socket(ronport);

    if (sock == -1) {
	DEBUG_PERROR("forw::init socket creation failed");
	return -1;
    }

#ifdef SO_TIMESTAMP
    int on = 1;
    if (setsockopt(sock, SOL_SOCKET, SO_TIMESTAMP,
		   (const char *)&on, sizeof(on)) < 0) {
	perror("setsockopt for SO_TIMESTAMP");
	exit(-1);
    }
#endif
    if (sched->register_socket(sock, this, NULL)) {
	perror("forw::init could not register RON socket");
	exit(-1);
    }
    DPRINTF(DEBUG_INIT, "forw::init ready with socket %d\n", sock);
    return 0;
}

/*
 * Register a particular packet type and callback object
 */


int
forw::route_packet(pkt *mypkt, const node *dst, int via_ron)
{
    node *mynexthop = NULL;
    if (via_ron) {
	mynexthop = nexthop(dst, mypkt);
    } else {
	mynexthop = new node(*dst);
    }

    if (!mynexthop) {
	DEBUG_PERROR("forw::route_packet could not get next hop for packet\n");
	return -1;
    }
    if (mylocalhost->is_local(mynexthop)) {
	DPRINTF(DEBUG_PACKETS,
		"forw::route_packet local delivery for remote packet: LOOP\n");
	delete mynexthop;
	return -1;
    }

    struct sockaddr_in dsthost;
    
    DPRINTF(DEBUG_PACKETS,
	    "forw::route_packet remote delivery (%d bytes, type %d rtf %d dst %s)\n",
	    mypkt->get_size(), mypkt->get_traffic_type(),
	    mypkt->get_routing_flags(), mynexthop->to_string().c_str());
    mypkt->set_hlim(mypkt->get_hlim() - 1);
    if (mypkt->get_hlim() < 5) {
	DPRINTF(DEBUG_ERRS,
		"forw::route_packet type %d hop limit exceeded (%d)\n"
		"   src %s\n"
		"   dst %s\n"
		" exdst %s\n"
		"  last hop: %s\n",
		mypkt->get_type(), mypkt->get_hlim(),
		mypkt->get_src().to_string().c_str(),
		mypkt->get_dst().to_string().c_str(),
		dst->to_string().c_str(),
		inet_ntoa(mypkt->fromsock.sin_addr));
	delete mynexthop;
	return -1;
    }

    if (*mynexthop == *dst) {
	/* Tell them it's going directly */
	mypkt->set_version(mypkt->get_version() | VERS_DIRECTMASK);
    }

    node rdst;
#if IP_NAT_TABLE
    rdst = ipmap[*mynexthop];
    if (rdst.is_null())
#endif
	rdst = *mynexthop;

    bzero(&dsthost, sizeof(dsthost));
#ifdef RON_HAS_STRUCT_SINLEN
    dsthost.sin_len = sizeof(struct sockaddr_in);
#endif
    dsthost.sin_family = AF_INET;
    dsthost.sin_port = rdst.get_port();
    dsthost.sin_addr = rdst.node_addr;

    int rc = sendto(sock, mypkt->contents, mypkt->size, 0,
		    (struct sockaddr *)&dsthost, sizeof(dsthost));
    if (rc != mypkt->size) {
	DEBUG_PERROR("forw::route_packet failed write for relay");
	delete mynexthop;
	return -1;
    }
    delete mynexthop;
    stats.counter_table["total sent packets"]++;
    stats.counter_table["total sent bytes"] += mypkt->size;
    return 0;
}


/*
 * We have a generic packet.  Do something with it - either
 * forward it or dispatch it locally
 */

int
forw::recv(pkt *packet, node *mynexthop, int via_ron)
{
    int rc;
    node *dst = NULL;

    DPRINTF(DEBUG_PACKETS,
	    "FORW|RECV iteration starts  ----------------\n");

    // (SAR)
    describe_pkt(packet);

    if (!mynexthop) {
	dst = new node(packet->get_dst());
	mynexthop = dst;
    }
    if (mylocalhost->is_local(mynexthop)) {
	rc = dispatch(packet, mynexthop, via_ron);
    } else {
	rc = route_packet(packet, mynexthop, via_ron);
    }
    if (dst) {
	delete dst;
    }

  DPRINTF(DEBUG_PACKETS,
	    "FORW|RECV iteration ends  ----------------\n");

    return rc;

}

/*
 * An event has occured on socket 'sockno' that we've registered
 * with the controlling application.  Handle it.
 */

int forw::socket_callback(int sockno, void *data)
{
  pkt *mypkt;
  node *nexthop;
  int rc;
  
  /* Data has come in on the RON socket.  Handle it */
  if (sockno == sock) {
    mypkt = get_ron_packet();
    if (!mypkt) {
      return -1;
    }
    
    nexthop = new node(mypkt->get_dst());
    rc = recv(mypkt, nexthop, 0);
    delete nexthop;
    delete mypkt;
    return rc;
  } else {
    
    fprintf(stderr, "forw::socket_callback unknown FD %d\n", sockno);
  }
  return 0;
}

/*
 * A packet has come in on our interface;  cope with it.
 */

pkt *forw::get_ron_packet()
{
    pkt *mypkt;
    int rc;
    struct msghdr msg;
    struct iovec iov;
    /* rectime, fromsock in packet */
    mypkt = new pkt();

    if (!mypkt) {
	DEBUG_PERROR("forw::handle_packet could not allocate pkt");
	return NULL;
    }

#ifdef SO_TIMESTAMP
    char ctrl[CMSG_SPACE(sizeof(struct timeval))];
    struct cmsghdr *cmsg = (struct cmsghdr *)&ctrl;
    msg.msg_control = (caddr_t)ctrl;
    msg.msg_controllen = sizeof(ctrl);
#endif
    msg.msg_name = &(mypkt->fromsock);
    msg.msg_namelen = sizeof(mypkt->fromsock);
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;
    iov.iov_base = mypkt->contents;
    iov.iov_len = maxpktsize;

    rc = recvmsg(sock, &msg, 0);

    if (rc == -1) {
	DEBUG_PERROR("forw::handle_packet read failed");
	delete mypkt;
	return NULL;
    }

    if (rc == 0) {
	DEBUG_PERROR("read returned 0");
	delete mypkt;
	return NULL;
    }
    
    int gottime = 0;
#ifdef SO_TIMESTAMP
    if (cmsg->cmsg_level == SOL_SOCKET &&
	cmsg->cmsg_type == SCM_TIMESTAMP &&
	cmsg->cmsg_len == CMSG_LEN(sizeof mypkt->rectime)) {
	/* Copy to avoid alignment problems: */
	memcpy(&mypkt->rectime,CMSG_DATA(cmsg),sizeof(mypkt->rectime));
	gottime = 1;	
    }
#endif
    if (!gottime) {
	gettimeofday(&mypkt->rectime, NULL);
    }

    mypkt->size = rc;
    stats.counter_table["total received packets"]++;
    stats.counter_table["total received bytes"] += mypkt->size;

#ifdef IP_NAT_TABLE
    /* Update our IP table if possible */
    if ((mypkt->size >= sizeof(struct ron_pkt_header))
	&& (mypkt->get_version() & VERS_DIRECTMASK)) {
	node r;
	r.node_addr = mypkt->fromsock.sin_addr;
	r.node_port = mypkt->fromsock.sin_port;
	ipmap[mypkt->get_src()] = r;
	printf("node %s accessible via %s\n",
	       mypkt->get_src().to_string().c_str(),
	       r.to_string().c_str());
	ipmap[r] = mypkt->get_src();
    }
#endif

    return mypkt;
}

node *
forw::nexthop(const node *idst, pkt *mypkt)
{
    node *dst, *rt = NULL;

    dst = new node(*idst);

    if (mylocalhost->is_local(idst)) {
	return dst;
    }

    /*
     * Note that there's some implicit routing policy here:
     * when in doubt, go directly to the destination!
     * If we have no routing table, go direct.
     */
    if (rtable) rt = rtable->lookup(dst, mypkt);

    if (rt) {
	delete dst;
	return rt;
    } else {
	return dst;
    }
}

int
localhost::check_intfs()
{
    if (intfs != NULL) {
	free_intf_descs(intfs);
    }
    intfs = get_intf_descs();
    if (!intfs) {
	DEBUG_PERROR("localhost::check_intfs could not get_intf_descs");
	return -1;
    }
    return 0;
}

int
localhost::timer_callback(void *data) {
    check_intfs();
    sched->register_timer(60000000, this, 0); /* 1 minute */
    return 0;
}

void
localhost::configure(Scheduler *s, forw *newforw) {
    sched = s;
    myforw = newforw;
    sched->register_timer(60000000, this, 0); /* 1 minute */
}

int
localhost::is_local(const node *mynode)
{
    struct intf_desc *desc;

    if (extra_addr && mynode->node_addr.s_addr == extra_addr) { return 1; }
    
    desc = intfs;
    while (desc != NULL) {
	if (desc->intf_addr.s_addr == mynode->node_addr.s_addr) {
	    return 1;
	}
	desc = desc->next;
    }
    return 0;
}

#ifdef _USE_PKT_CACHE_
static char *pkt_cache = NULL;
#endif

// (SAR) Create packet, allocate n bytes
pkt::pkt(unsigned n) // default:  n = maxpktsize 
{
#ifdef _USE_PKT_CACHE_
    if (pkt_cache != NULL) {
	contents = pkt_cache;
	pkt_cache = 0;
    } else
#endif
	contents = new char[n];

    // (SAR) Should set size to 0?
    size = 0;

    if (!contents) {
	DEBUG_PERROR("pkt::pkt could not allocate packet contents");
    }
    set_hlim(DEFAULT_RON_HLIM);
    set_version(RON_VERSION);
    set_traffic_type(TRAFFIC_NORMAL);
    set_seq(0);
    bzero(&rectime, sizeof(rectime));
    bzero(&fromsock, sizeof(fromsock));

}

pkt::~pkt()
{
    if (contents != NULL) {
#ifdef _USE_PKT_CACHE_
	if (pkt_cache == NULL) {
	    pkt_cache = contents;
	} else
#endif
	delete contents;
	contents = NULL;
    }
}
