#ifndef _FORW_H_
#define _FORW_H_

#include <map>
#include <list>
#include <vector>
#include <queue>
#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/param.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <netinet/ip.h>
#include <stream.h>

#include "ron.h"
#include "node.h"
#include "interface.h"
#include "fixedpoint.h"
#include "packet.h"
#include "scheduler.h"
#include "plumber.h"
#include "multipath.h"
#include "ron_stats.h"

namespace RON {

  class ConnectionAware {
  public:
    virtual void connectionTerminated(vcID id) = 0;
  };

class forw;  /* Defined later */
class membership; /* Defined in membership.h */

/*
 * Localhost will tell us if an IP address is local or not.
 * It should be used to store local parameters, like
 * hostnames, IP addresses, and other identifiers
 * (e.g. RON ids, etc.)
 *
 * There's really only one of these per machine.
 */

class localhost : public time_callable {
public:
    localhost() {

	set_timer_name("Local host");

	intfs = NULL;
	extra_addr = 0;
	// Now called outside of the constructor (SAR)
	// init();
    }

    /* (SAR) We really don't want to check those interfaces in NS */
    virtual void init()	{
	check_intfs();
    }
    virtual ~localhost() {
	if (intfs != NULL) {
	    free_intf_descs(intfs);
	}
    }
    void add_addr(u_int32_t e) { extra_addr = e; }
    virtual int is_local(const node *mynode); // (SAR)
    int check_intfs();
    void configure(Scheduler *s, forw *newforw);
    int timer_callback(void *data);

private:
    u_int32_t extra_addr;
    struct intf_desc *intfs;
    forw *myforw;
    Scheduler *sched;
};

class encapsulator {
public:
    virtual pkt *encapsulate(pkt *packet, node *mynode, int via_ron) = 0;
};

class routing_table : public Plumber {
public:
    virtual node *lookup(const node *dst, pkt *mypkt) = 0;

    // (SAR) Multipath
    virtual node *lookup(const node *dst, pkt *mypkt,
			 int mp_policy, double param) {
	return lookup(dst, mypkt, 0, 0);
    }

    virtual u_int32_t get_table_version(const node *dst) {
	return 0;
    }

    virtual void set_remote_version(const node *dst, u_int32_t version,
				    time_t update_time) {
	return;
    }

    /* The packet decoding machinery should be inheritable once written */
    virtual int recv(pkt *mypkt, node *mynode, int via_ron) = 0;
    /*virtual ~routing_table();*/
};


/*
 * A forwarding object.  It has a socket on which it sends
 * and receives packets, a buffer for packets flowing through
 * it, and pointers to its routing objects and whatnot.
 */

class forw : public Plumber, public socket_callable, public ConnectionAware {
public:
    forw(Scheduler *s,
	 struct in_addr n_myaddr, int port,
	 routing_table *nrtable,
	 membership *mm) {
	sched = s;
	rtable = nrtable;
	membership_manager = mm;
	myaddr = n_myaddr;
	//Init becomes virtual now, called outside of 
	// constructor (tnight)
	//init(port);
    }
  
    virtual void connectionTerminated(vcID id) {}

    virtual int init(int ronport);
    int get_socket() { return sock; }
    virtual int socket_callback(int sockno, void *data);
    
    virtual pkt *get_ron_packet();
    virtual node *nexthop(const node *dst, pkt *mypkt);

    virtual int route_packet(pkt *mypkt, const node *dst, int via_ron);

    membership *get_membership() {
	assert(membership_manager != NULL);
	return membership_manager;
    }
    int recv(pkt *packet, node *mynexthop, int via_ron);
    int dispatch_packet(pkt *packet, node *mynexthop, int via_ron);
    struct in_addr get_myaddr() { return myaddr; }

    virtual u_int16_t get_port() { return port; }
    node mynode() {
	node n;
	n.node_addr = get_myaddr();
	n.node_port = get_port();
	return n;
    }
    int is_local(const node *mynode) { return mylocalhost->is_local(mynode); }
#if IP_NAT_TABLE
    map<node, node> ipmap; /* Map nodes to their perceived IP */
    map<node, node> dstmap; /* And IP addresses to a dest RON node */
#endif
    stat_collector stats;
    routing_table *router() { return rtable; }

protected:
    Scheduler *sched;
    struct in_addr myaddr;
    struct in_addr my_inaddr;
    int port;  /* The port on which we receive and send RON datagrams */
    int sock;  /* The socket for the above */
    localhost *mylocalhost;
    routing_table *rtable;
    membership *membership_manager;
};

}

#endif /* _FORW_H_ */
