#ifndef __RON_MP_POLICIES__
#define __RON_MP_POLICIES__

#include <stdlib.h>

#include "ron.h"
#include "fixedpoint.h"
#include "ronlib.h"
#include "node.h"
#include "ron_named.h"

#include "multipath.h"


namespace RON {

    void init_multipath();


/* MultiPath Routing Policy ----------------------- */
class MPPolicy : public Named {

 public:
    // MP_DEFAULT should always be first
    static const enum {MP_DEFAULT, MP_BESTPATH, MP_RANDOM, MP_RROBIN, MP_BESTK, MP_DTMP} policyID;
    static const int NUM_MPPOLICIES = 6;

    // (SAR) Multipath extensions start...
    static MPPolicy *multipathPolicies[MPPolicy::NUM_MPPOLICIES];

    MPPolicy(string s):Named(s)
	{}

    virtual node decide(Rankings& r) = 0;
    virtual node decide(Rankings& r, double param)
	{
	    return decide(r);
	}

};
/* ----------------------- MultiPath Routing Policy */
 
class MPBestPath : public MPPolicy {
 public:
    MPBestPath() : MPPolicy("Best Path")
	{}

    node decide(Rankings& r)
	{
	    node next;
	    double f;

	    if (r.size() == 0)
		return next; // null node

	    r.getRanking(r.size() - 1, next, f);

	    return next;
	}

};

class MPRandom : public MPPolicy {
 public:
    MPRandom() : MPPolicy("Random")
	{}

    node decide(Rankings& r)
	{
	    node next;
	    double f;
	    double drnum;

	    if (r.size() == 0)
		return next; // null node

	    drnum = random()/(double)RAND_MAX * r.size();

	    assert(r.getRanking((int)drnum, next, f) != MPRON_NOTFOUND);

	    return next;
	}

};

// Round-robin
class MPRoundRobin : public MPPolicy {
 public:
    MPRoundRobin() : MPPolicy("Round Robin")
	{
	}

    node decide(Rankings& r)
	{
	    node next;
	    double f;

	    if (r.size() == 0)
		return next; // null node

	    if ((unsigned)r.lastSelected >= r.size())
		r.lastSelected = 0;

	    assert(r.getRanking(r.lastSelected, 
				next, f) != MPRON_NOTFOUND);
	    
	    r.lastSelected++;

	    return next;	    
	}
};


// Round-robin among best k
class MPBestK : public MPPolicy {
 public:
    MPBestK() : MPPolicy("Best K")
	{
	}

    node decide(Rankings& r)
	{
	    return decide(r, defaultK);
	}

    node decide(Rankings& r, double k)
	{
	    node next;
	    double f;
	    int intK = (int)k;

	    if (r.size() == 0)
		return next; // null node


	    r.lastSelected = ((unsigned)intK > r.size() ? 
			       r.lastSelected % r.size() : 
			       r.lastSelected % intK); 

	    assert(r.getRanking(r.size() - r.lastSelected - 1, 
				 next, f) != MPRON_NOTFOUND);
	    r.lastSelected++;

	    return next;	    
	}

 private:
    static const double defaultK = 3.0;
};

// DTMP:  Round-robin among all within a constant
// of the best...(k here can be passed as a constant) 
class MPDTMP : public MPPolicy {
 public:
    MPDTMP() : MPPolicy("DTMP")
	{
	}

    node decide(Rankings& r)
	{
	    return decide(r, defaultK);
	}

    node decide(Rankings& r, double k)
	{
	    node next;
	    double f, best;
	    int lowestWithinFactor, nBest;

	    if (r.size() == 0)
		return next; // null node

	    // Get the best
	    lowestWithinFactor = r.size() - 1;
	    assert(r.getRanking(lowestWithinFactor, next, best) 
		   != MPRON_NOTFOUND);

	    if (r.size() > 1)
		// Get the count of all entries that fall into the range
		for (int i = r.size() - 2; i >= 0; i--)
		    {
			r.getRanking(i, next, f);

			if (f * k < best)
			    {
				lowestWithinFactor = i + 1;
				break;
			    }

		    }
	    nBest = r.size() - lowestWithinFactor;

	    r.lastSelected = r.lastSelected % nBest;

	    assert(r.getRanking(r.size() - r.lastSelected - 1, 
				 next, f) != MPRON_NOTFOUND);
	    r.lastSelected++;

	    return next;	    
	}

 private:
    static const double defaultK = 3.0;
};


}

#endif
