/*
 *  Multipath routing extensions to RON.
 *
 *
 *
 *  ------------------------ Stanislav Rost [Dec 2001]
 */
#ifndef __RON_MULTIPATH__
#define __RON_MULTIPATH__

#include <sys/types.h>

#include "ron.h"
#include "fixedpoint.h"
#include "ronlib.h"
#include "node.h"
#include "ron_named.h"

#include <algorithm>
#include <vector>
#include <map>
#include <hash_map.h>
#include <assert.h>

namespace RON {

    // Return codes
    enum {
	MPRON_NOTFOUND,   // Entry not found
	MPRON_FOUND      // Entry found
    };
 
/* Virtual Curcuits ------------------------------- */
/*
 * Simple table, such that given a virtual ID resolves nextHop
 */
typedef uint32_t vcID;

class VCTable
{
 public:

    VCTable(int hashSize = 32);

    void add(vcID id, const node& dest);

    int remove(vcID id);
    int remove(node dest); // remove all VCs for that destination

    int lookup(vcID id, node& result) const;
    
 private:
    hash_map<vcID, node> *map_;
};
 
/* ------------------------------- Virtual Curcuits */
 
/* Rankings --------------------------------------- */
/*
 *
 * Contains nextHops sorted according to a particular metric
 * 
 */
 
// Normally, sorted by the double value
// of the metric (to obtain ranking)
 
// For insertions and lookups of Ranking rows on nodes,
// we will also provide a StrictWeakOrdering for heapSort
 class rankingRow {
     friend int CompareByMidhops(const rankingRow *left, const rankingRow *right);
     friend int CompareByMetrics(const rankingRow *left, const rankingRow *right);
 private:
     
     // Used to stabilize heapSort
     static unsigned stabilizerCounter;	
     unsigned stabilizer;
     
 public:
     double metric;
     node midHop;

     rankingRow(const node& mh, double& m) :
	 stabilizer(stabilizerCounter++),
	 metric(m),
	 midHop(mh)
	 {
	     midHop = mh;
	     metric = m;
	 }

     rankingRow& operator=(const rankingRow& b)
	 {
	     metric = b.metric;
	     midHop = b.midHop;

	     return *this;
	 }

     const int operator==(const node& right) const {
	 return midHop == right;
     }
     const int operator<(const node& right) const {
	 return midHop < right;
     }
     
     const int operator==(const rankingRow& right) const {
	 if (metric == right.metric)
	     return (stabilizer == right.stabilizer);
	 else
	     return 0;
     }
     const int operator<(const rankingRow& right) const {
	 if (metric == right.metric)
	     return (stabilizer < right.stabilizer);
	 else
	     return metric < right.metric;
     }
 };

class Rankings {
 public:

    Rankings(); 
    ~Rankings();

    // Also inserts if midHop is not there
    int update(const node& midHop, double& m);

    int remove(node &midHop);

    size_t size() const
	{
	    assert(heap.size() == index.size());

	    return heap.size();
	}

    int getRanking(int index, node& midHop, double& value); 

    inline void reSort();

    void print(FILE *out);

    int lastSelected; // useful for Round-Robin etc...

 private:
    int needsReSort;
    vector<rankingRow *> heap;
    map<node, rankingRow *> index;
};


/* --------------------------------------- Rankings */

/* StatTable -------------------------------------- */
/*
 *  Will need a table of rankings for a particular destination
 *  (Rankings store goodness values for various midhops towards
 *   those destinations).
 *
 */

/* --------------------------------------- StatTable */

}

#endif
