#include "ron.h"
#include "forw.h"
#include "membership.h"
#include "ping_protocol.h"
#include "rondebug.h"
#include <list>

#define TIMER_INTERVAL 100000
#define MIN_PING_TIMEOUT 1.0
#define PING_TIMEOUT 3
#define PING_KEEP    10  /* Keep the record for 10 seconds just in case */

using namespace RON;

static fixedpoint
ping_timeout(peer_contact *peer)
{
    fixedpoint r;

    if (peer->lat_avg != fixedpoint(0.0)) {
	r = peer->lat_avg + fixedpoint(3.0) * peer->lat_md;
	r = r + fixedpoint(0.100); /* Add an extra 100 ms to be paranoid */
	r = r * fixedpoint(2.0); /* And be really paranoid */
	if (peer->failures > 4) {
	    r = r + fixedpoint(1.0); /* Be gentle when actual failures
					have occurred */
	}
	if (r > fixedpoint(PING_INTERVAL))
	    r.set(PING_INTERVAL);
	if (r < MIN_PING_TIMEOUT)
	    r.set(MIN_PING_TIMEOUT);
    } else {
	r.set(PING_TIMEOUT);
    }
#if 0
    DPRINTF(DEBUG_PING, "timeout for %s is %s\n",
	    peer->peer.to_string().c_str(),
	    r.to_string().c_str());
#endif
    return r;
}

void
ping_protocol::configure(Scheduler *s, forw *f) {
    sched = s;
    myforw = f;
    f->register_type(RON_PING, this, NULL);
    f->register_type(RON_MAYBEDOWN, this, NULL);
    s->register_timer(TIMER_INTERVAL, this, 0);
    DPRINTF(DEBUG_INIT|DEBUG_PING,
	    "ping_protocol::configure with forw %p\n", f);
}

int
ping_protocol::timer_callback(void *data)
{
    reconcile_peers();
    send_pings();
    clean_pings();
    sched->register_timer(TIMER_INTERVAL+(random() % 50000), this, 0);
    return 0;
}

void
ping_protocol::reconcile_peers()
{
    list<node>::const_iterator iter, iter2;
    list<node> *npeers;

    assert (myforw != NULL);
    assert (myforw->get_membership() != NULL);
    npeers = myforw->get_membership()->peers();
    if (!npeers) {
	DEBUG_PERROR("ping_protocol::reconcile_peers no peers!\n");
	return;
    }

    /* Clean out expired entries */
    list<peer_contact>::iterator c;
    list<peer_contact> kills;

    for (c = contacts.begin(); c != contacts.end(); c++) {
	int found = 0;
	/* See if that peer is still in the master list */
	for (iter = npeers->begin(); iter != npeers->end(); iter++) {
	    if ((*iter) == (*c).peer) {
		found = 1;
		break;
	    }
	}
	if (!found) {
	    kills.push_back(*c);
	}
    }
    
    for (c = kills.begin(); c != kills.end(); c++) {
	DPRINTF(DEBUG_PING, "ping_protocol::reconcile_peers removing %s\n",
		(*c).peer.to_string().c_str());
	contacts.remove(*c);
    }
    kills.clear();


    /* This is n^2, but I don't care for now.  2500 cycles won't kill me. */
    for (iter = npeers->begin(); iter != npeers->end(); iter++) {
	int iter_found = 0;
#if 0 /* Too verbose for normal use */
	    DPRINTF(DEBUG_PING, "ping_protocol::reconcile_peers considering %s\n", (*iter).to_string().c_str());
#endif


	if (myforw->is_local(&*iter)) {
	    DPRINTF(DEBUG_PING, "ping_protocol::reconcile_peers LOCAL\n");

	    continue;
	}
	for (iter2 = peers.begin(); iter2 != peers.end(); iter2++) {
#if 0 /* Too verbose for normal use */
	    DPRINTF(DEBUG_PING, "ping_protocol::reconcile_peers comparing (%s) to (%s)\n", (*iter).to_string().c_str(), (*iter2).to_string().c_str());
#endif

	    if (*iter == *iter2) {
		iter_found = 1;
		break;
	    }
	}
	if (!iter_found) {
	    
	    peers.push_back(*iter);
	    peer_contact pc;
	    pc.peer = *iter;
	    fixedpoint last = nowpoint + fixedpoint(0, random() % 3000000);
	    pc.last_ping = last;
	    pc.failures = 0;

	    DPRINTF(DEBUG_PING, "ping_protocol::reconcile_peers pushing %s into contacts\n", pc.peer.to_string().c_str());

	    contacts.push_back(pc);
	}
    }
}

void
ping_protocol::send_pings()
{
    list<peer_contact>::iterator iter;
    struct timeval now;

    gettimeofday(&now, NULL);
    nowpoint.set(now.tv_sec, now.tv_usec);

#if 0 /* Too verbose for normal use */
    DPRINTF(DEBUG_PING,
	"ping_protocol::send_pings at %ld,%ld\n", 
 	    now.tv_sec, now.tv_usec);
#endif

    node longest_peer;
    fixedpoint longest_time = 0;

    for (iter = contacts.begin(); iter != contacts.end(); iter++) {
	fixedpoint delta = (nowpoint - (*iter).last_ping);
	if (delta > longest_time) {
	    longest_time = delta;
	    longest_peer = (*iter).peer;
	}
    }
    if (longest_time > fixedpoint(PING_INTERVAL)) {
	for (iter = contacts.begin(); iter != contacts.end(); iter++) {
	    if ((*iter).peer == longest_peer) {
		fixedpoint last;
		/* Jitter it */
		last = nowpoint + fixedpoint(0, random() % 3000000);
		(*iter).last_ping = last;
	    }
	}
	send_ping(&longest_peer, 0, TYPE_FIRST, 0, fixedpoint(0,0));
	(*iter).last_ping = nowpoint;
    }
}

int
ping_protocol::send_ping(node *mynode, u_int64_t id, u_int8_t type,
			 int via_ron, fixedpoint proctime)
{
    pkt *packet;
    struct ron_ping_header *ph;
    
    DPRINTF(DEBUG_PING, "ping_protocol::send_ping to %s\n", mynode->to_string().c_str());


    packet = new pkt();
    if (!packet) {
	DEBUG_PERROR("ping_protocol::send_ping could not allocate packet");
	return ENOMEM;
    }

    if (!mynode) {
	delete packet;
	return -1;
    }
    
    packet->set_datalen(sizeof(struct ron_ping_header));
    ph = (struct ron_ping_header *)packet->data();

    if (!id) {
	id = (u_int64_t)random() << 32 | random();
    }
    ph->id1 = htonl(id >> 32);
    ph->id2 = htonl(id & 0xffffffff);
    ph->type = type;
    ph->proctime = proctime.to_wire();
    
    u_int32_t router_vers = myforw->router()->get_table_version(mynode);
    ph->rtable_version = htonl(router_vers);

    packet->ron_hdr()->dst_addr = mynode->node_addr;
    packet->ron_hdr()->dst_port = mynode->node_port;
    packet->ron_hdr()->src_addr = myforw->get_myaddr();
    packet->ron_hdr()->src_port = myforw->get_port();
    packet->set_type(RON_PING);

    DPRINTF(DEBUG_PING,
	    "ping_protocol::send_ping send type %d id 0x%qx to %s\n",
	    type, id, mynode->to_string().c_str());
    int rc = dispatch(packet, mynode, 0);
    
    if (type == TYPE_FIRST || type == TYPE_SECOND) {
	struct ping_record prc;
	prc.id = id;
	prc.time_sent = nowpoint;
	prc.type = type;
	prc.dst = *mynode;
	prc.marked_failed = 0;

	/* We sent them a ping;  don't send another for a while */
	list<peer_contact>::iterator peer;
	for (peer = contacts.begin(); peer != contacts.end(); peer++) {
	    if ((*peer).peer == *mynode) {
		(*peer).last_ping = nowpoint;
		break;
	    }
	}
	if (peer != contacts.end()) {
	    prc.con = &(*peer);
	    record.push_back(prc);

	} else {
	    DPRINTF(DEBUG_PING, "ping_protocol::send_ping no contact for %s\n",
		    prc.dst.to_string().c_str());
	}
	// (SAR) Move to location above ...this was causing
	// segfaults in send_pings() since these situations may occur in 
	// start-up phases
	//	record.push_back(prc);
    }
    
    myforw->stats.counter_table["sent pings"]++;
    delete packet;
    
    return rc;
    
}

/*
 * Cleanup outstanding ping packets and record them as losses
 */

void
ping_protocol::clean_pings()
{
    list<ping_record>::iterator iter;
    list<ping_record> kills;

    struct timeval now;
    gettimeofday(&now, NULL);
    nowpoint.set(now.tv_sec, now.tv_usec);

#if 0 /* Too verbose for normal use */
    DPRINTF(DEBUG_PING,
 	    "ping_protocol::clean_pings at %ld,%ld\n", 
 	    now.tv_sec, now.tv_usec);
#endif

    for (iter = record.begin(); iter != record.end(); iter++) {
	if ((!((*iter).marked_failed)) &&
	    nowpoint - (*iter).time_sent > ping_timeout((*iter).con)) {
	    /* Exceeded the timeout for that computer.  Ping them,
	     * but don't eliminate the record, necessarily,
	     * in case the ping just took a LONG time */
	    
	    pdbc->insert((*iter).dst, PDB_REACH, fixedpoint(0,0));
	    (*iter).marked_failed = 1;
	    /* The ping got lost.  Send another for outage detection
	     * purposes */
	    DPRINTF(DEBUG_PING,
		    "ping_protocol::clean_pings LOSS id 0x%qx\n",
		    (*iter).id);
	    maybedown(&((*iter).dst));
	}
	if (nowpoint - (*iter).time_sent > fixedpoint(PING_KEEP)) {
	    kills.push_back(*iter);
	}
    }

    for (iter = kills.begin(); iter != kills.end(); iter++) {
	DPRINTF(DEBUG_PING, "ping_protocol::clean_pings removing %s\n",
		(*iter).dst.to_string().c_str());

	record.remove(*iter);
    }
    kills.clear();
}

int
ping_protocol::recv(pkt *packet, node *mynode, int via_ron) {
    struct ron_ping_header *fh;

    if (packet->get_type() == RON_MAYBEDOWN) {
	struct ron_maybedown *m;
	if (packet->datalen() < sizeof(struct ron_maybedown)) {
	    DPRINTF(DEBUG_PING|DEBUG_ERRS,
		    "ping_protocol::recv runt maybedown (%d)\n",
		    packet->datalen());
	    return -1;
	}
	m = (struct ron_maybedown *)packet->data();
	node n;
	n.node_addr = m->node_addr;
	n.node_port = m->node_port;
	maybedown(&n);
	return 0;
    }

    if (packet->datalen() < sizeof(struct ron_ping_header)) {
	DPRINTF(DEBUG_PING|DEBUG_ERRS,
		"ping_protocol::recv runt packet (%d bytes)\n",
		packet->datalen());
	return -1;
    }

    struct timeval now;
    gettimeofday(&now, NULL);
    nowpoint.set(now.tv_sec, now.tv_usec);

    DPRINTF(DEBUG_INIT|DEBUG_PING,
	    "ping_protocol::recv at %ld,%ld\n", 
	    now.tv_sec, now.tv_usec);

    fh = (struct ron_ping_header *)packet->data();

    u_int64_t id;

    id = ((u_int64_t)ntohl(fh->id1) << 32) | htonl(fh->id2);

    /* Check to see if we've seen it in the cache before */

    /*printf("ping_protocol::recv Got packet id 0x%qx\n", id);*/

    list<ping_record>::iterator iter;
    ping_record *pr = NULL;
    
    for (iter = record.begin(); iter != record.end(); iter++) {
	if ((*iter).id == id) {
	    pr = &(*iter);
	    break;
	}
    }
    myforw->stats.counter_table["received pings"]++;

    node dst;
    dst.node_addr = packet->ron_hdr()->src_addr;
    dst.node_port = packet->ron_hdr()->src_port;
    u_int32_t rtable_version = ntohl(fh->rtable_version);
    myforw->router()->set_remote_version(&dst, rtable_version, now.tv_sec);
    
    fixedpoint rectime, proctime;
    rectime.set(packet->rectime.tv_sec, packet->rectime.tv_usec);
    if (rectime == 0) {
	DPRINTF(DEBUG_PING, "ping_protocol using inaccurate time\n");
	rectime = nowpoint;
    }
    proctime.set(nowpoint - rectime);

    if (pr) {
	fixedpoint rtt;
	rtt = rectime - pr->time_sent - fixedpoint(fh->proctime);
	DPRINTF(DEBUG_PING, "ping_protocol: %s 0x%qx  %d  RTT: %s\n",
		dst.to_string().c_str(),
		id, fh->type, rtt.to_string().c_str());
	record.remove(*pr);
	if (fh->type == TYPE_SECOND) {
	    DPRINTF(DEBUG_PING, " ... sending it back to the originator\n");
	    send_ping(&dst, id, TYPE_THIRD, 0, proctime);
	}
	pdbc->insert((*iter).dst, PDB_REACH, fixedpoint(1,0));
	pdbc->insert((*iter).dst, PDB_LAT, rtt);

	list<peer_contact>::iterator i;
	for (i = contacts.begin(); i != contacts.end(); i++) {
	    if (dst == (*i).peer) {
		/* Update the latency and deviation estimates, a'la TCP */
		(*i).lat_avg = (*i).lat_avg * fixedpoint(PDB_Q) +
		    fixedpoint(1.0 - PDB_Q) * rtt;
		fixedpoint err;
		if ((*i).lat_avg > rtt) {
		    err = (*i).lat_avg - rtt;
		} else {
		    err = rtt - (*i).lat_avg;
		}
		(*i).lat_md = (*i).lat_md * fixedpoint(PDB_Q) +
		    fixedpoint(1.0 - PDB_Q) * err;
		DPRINTF(DEBUG_PING, "ping_protocol host %s RTT %s ERR %s lat: %s dev: %s\n",
			(*iter).dst.to_string().c_str(),
			rtt.to_string().c_str(),
			err.to_string().c_str(),
			(*i).lat_avg.to_string().c_str(),
			(*i).lat_md.to_string().c_str());
		(*i).failures = 0;
		break;
	    }
	}
    } else {
	/*printf("New packet id 0x%qx\n", id);*/
	if (fh->type == TYPE_FIRST) {
	    DPRINTF(DEBUG_PING, "ping_protocol::recv sending TYPE_SECOND\n");

	    /* Update their 'last contacted' field.. */
	    send_ping(&dst, id, TYPE_SECOND, 0, proctime);
	} else {
	    DPRINTF(DEBUG_ERRS, "ping_protocol::recv unknown packet id 0x%qx\n", id);
	}
    }
    return 0;
}

/*
 * Notification that a node may be down.
 * Trigger the "fast probes" mode
 */
void
ping_protocol::maybedown(node *downnode)
{
    int found = 0;

    DPRINTF(DEBUG_PING, "ping_protocol::maybedown triggered for %s\n",
	    downnode->to_string().c_str());
    /* Check to see if we have an outstanding ping */
    
    list<ping_record>::iterator rec;
    for (rec = record.begin(); rec != record.end(); rec++) {
	if ((*rec).dst == *downnode && !(*rec).marked_failed) {
	    found = 1;
	}
    }
    if (found) {
	DPRINTF(DEBUG_PING, "ping_protocol::maybedown outstanding ping\n");
    }

    if (!found) {
	list<peer_contact>::iterator i;
	for (i = contacts.begin(); i != contacts.end(); i++) {
	    if (*downnode == (*i).peer) {
		if ((*i).failures < 4) {
		    DPRINTF(DEBUG_PING,
			    "ping_protocol::maybedown sending outage detector\n");
		    send_ping(downnode, 0, TYPE_FIRST, 0,
			      fixedpoint(0,0));
		}
		(*i).failures++;

		break;
	    }
	}
    }
}
