#include "broadcast_membership.h"
#include "rondebug.h"
#include "ronlib.h"
#include "errno.h"

using namespace RON;

void
broadcast_membership::configure(Scheduler *s, forw *f) {
    sched = s;
    myforw = f;
    f->register_type(RON_MEMBERSHIP, this, NULL);
    

    TObjectFunctor<broadcast_membership> *functor  = 
                        new TObjectFunctor<broadcast_membership>(
                                                this, 
                                                &broadcast_membership::timer_callback,
                                                0
                                            );
    
    TimerEvent *timer_event = TimerEvent::create(
                                10,
                                0, 
                                (TFunctor *) functor, 
                                "Broadcast membership timer - configure",
                                true
                            );
    
    ronOverlay->addTimerEvent(timer_event);


    
    DPRINTF(DEBUG_INIT|DEBUG_MEMBERSHIP,
	    "broadcast_membership::configure with forw %p and sched %p\n", f, s);
    init();
}

void
broadcast_membership::init()
{
    struct timeval tnow;
    gettimeofday(&tnow, NULL);
    nowpoint.set(tnow.tv_sec, tnow.tv_usec);


    peerlist.clear();
    peercache.clear();

    // (SAR)  Should probably read from 
    // infile supplied to the constructor.
    if (infilename)
      initlist(infilename, &peercache);
    else
      initlist("peerinit.txt", &peercache); 
}

int
broadcast_membership::reload()
{
    /* Is this really what we want to do? */
    init();
    return 0;
}

void
broadcast_membership::dump(FILE *out)
{
    list<node>::const_iterator iter;

    if (!out) { out = stdout; }

    for (iter = peerlist.begin(); iter != peerlist.end(); iter++) {
	fprintf(out, "%s\n", (*iter).to_string().c_str());
    }
}

int
broadcast_membership::savelist(char *fname, list<node> *uplist)
{
    FILE *outfile;

    if (!fname) {
	outfile = stdout;
    } else {
	outfile = fopen(fname, "w");
	if (!outfile) {
	    DPRINTF(DEBUG_ERRS,
		    "broadcast_membership::savelist could not open ");
	    DEBUG_PERROR(fname);
	    return -1;
	}
    }
    list<node>::const_iterator iter;
    for (iter = uplist->begin(); iter != uplist->end(); iter++) {
	DPRINTF(DEBUG_MEMBERSHIP, "broadcast_membership::savelist %s\n",
		(*iter).to_string().c_str());
	if (myforw->is_local(&(*iter))) {
	    continue;
	}
	fprintf(outfile, "%s\n", (*iter).to_string().c_str());
    }
    if (fname) {
	fclose(outfile);
    }
    return 0;
}


int broadcast_membership::initlist(char * fname, list<node> *uplist)
{
    char inbuf[256];
    FILE *infile;

    if (!fname) {
	DPRINTF(DEBUG_ERRS, "broadcast_membership::initlist null fname\n");
	return -1;
    }

    infile = fopen(fname, "r");
    if (!infile) {
	DPRINTF(DEBUG_ERRS,
		"broadcast_membership::initlist could not open input file ");
	DEBUG_PERROR(fname);
	return -1;
    }

    uplist->clear();

    while ((fgets(inbuf, sizeof(inbuf)-1, infile)) != NULL) {
	char *comment;
	char dst[256];
	int port;
	
	struct in_addr dstaddr;

	comment = strchr(inbuf, '#');
	if (comment) *comment = '\0';
	/* XXX:  note that this parsing can overflow.  shrug. */

	comment = strchr(inbuf, '\n');
	if (comment) *comment = '\0';
	comment = strchr(inbuf, '\r');
	if (comment) *comment = '\0';
	if (strlen(inbuf) == 0) {
	    continue; /* avoid some error messages... */
	}
	bzero(dst, sizeof(dst));
	strncpy(dst, inbuf, strlen(inbuf));
	char *d, *p;
	d = strchr(dst, ':');
	if (d) {
	    p = d + 1;
	    *d = 0;
	    port = atoi(p);
	    if (!nametoip(dst, &dstaddr)) {
		uplist->push_back(node(dstaddr, htons(port)));
		DPRINTF(DEBUG_MEMBERSHIP,
			"broadcast_membership::initlist node %s\n",
			node(dstaddr, htons(port)).to_string().c_str());
	    } else {
		fprintf(stderr, "Bad peer line: %s\n", inbuf);
	    }
	} else {
	    fprintf(stderr, "Bad peer line: %s\n", inbuf);
	}
    }
    DPRINTF(DEBUG_MEMBERSHIP, "end of nodes\n");
    fclose(infile);
    return 0;
}

int
broadcast_membership::clean_peers()
{
    struct timeval tnow;
    map<node, node_data>::const_iterator i;
    list<node> kills;
    list<node>::const_iterator l;

#ifdef DEBUG
    if (debug & DEBUG_MEMBERSHIP) {
	fprintf(stderr, "\nbroadcast_membership::clean_peers start"
		        "\n---------------------------------------\n");
	dump(stderr);
    }
#endif
    int nkills = 0;
    
    gettimeofday(&tnow, NULL);
    nowpoint.set(tnow.tv_sec, tnow.tv_usec);

    for (i = peermap.begin(); i != peermap.end(); i++) {
	if ((*i).second.expires &&
	    ((*i).second.last_update + fixedpoint(EXPIRE_TIME, 0)) < nowpoint)
	{
	    kills.push_back((*i).first);
	}
    }
    for (l = kills.begin(); l != kills.end(); l++) {
	DPRINTF(DEBUG_MEMBERSHIP, " -> kill %s\n", l->to_string().c_str());
	peermap.erase(*l);
	nkills++;
    }
    kills.clear();
    peerlist.clear();
    for (i = peermap.begin(); i != peermap.end(); i++) {
	peerlist.push_back((*i).first);
    }
#ifdef DEBUG
    if (debug & DEBUG_MEMBERSHIP) {
	fprintf(stderr, "\n+++++++++++++++++++++++++++++++++++++\n");
	dump(stderr);
	fprintf(stderr, "\n");
    }
#endif
    return nkills;
}

void
broadcast_membership::timer_callback(void *data)
{
    struct timeval tnow;
    static fixedpoint last_time;
    gettimeofday(&tnow, NULL);
    nowpoint.set(tnow.tv_sec, tnow.tv_usec);

    clean_peers();

    /* Normal requests have a 0 in the data field;  special out-of-band
     * delayed announcement requests have a 1.  Only request another
     * callback for normal timers.. */
    if ((int)data == 0) {
	int nextun = UPDATE_FREQUENCY + (random() % (UPDATE_FREQUENCY/3));


        TObjectFunctor<broadcast_membership> *functor  = 
                        new TObjectFunctor<broadcast_membership>(
                                                this, 
                                                &broadcast_membership::timer_callback,
                                                0
                                            );
    
        TimerEvent *timer_event = TimerEvent::create(
                                nextun/1000,
                                0, 
                                (TFunctor *) functor, 
                                "Broadcast membership timer",
                                true
                            );
    
        ronOverlay->addTimerEvent(timer_event);


    }
    special_request_outstanding = 0;
    last_time = nowpoint;

    if (peerlist.size()) {
	send_announce(NULL);
	// (SAR) Had to change so savelist file name prefixed by its IP address
	char fname[30];

	// (SAR) Not likely to overflow... ... until ipv6 -- dga
#ifdef PERSONALIZED_PEERCACHE
	sprintf(fname, "%s-peercache.txt", inet_ntoa(myforw->get_myaddr()));
#else
	sprintf(fname, "peercache.txt");
#endif
	savelist(fname, &peerlist);
    } else {
	send_announce(&peercache);
    }

}

int
broadcast_membership::recv(pkt *mypkt, node *, int via_ron)
{
    struct ron_membership_header *mh;
    struct ron_membership_entry *me;
    struct timeval tnow;
    int npeers;

    int new_peer = 0;

    gettimeofday(&tnow, NULL);
    nowpoint.set(tnow.tv_sec, tnow.tv_usec);

    DPRINTF(DEBUG_MEMBERSHIP|DEBUG_PACKETS,
	    "broadcast_membership::recv got a packet type 0x%x\n",
	    mypkt->get_type());

    mh = (struct ron_membership_header *)mypkt->data();
    if (mypkt->datalen() < sizeof(struct ron_membership_header)) {
	DPRINTF(DEBUG_MEMBERSHIP|DEBUG_ERRS,
		"broadcast_membership::recv runt packet (%d)\n",
		mypkt->datalen());
	return -1;
    }

    npeers = mh->nmembers;
    unsigned int expected_size = sizeof(struct ron_membership_header) +
	npeers * sizeof(struct ron_membership_entry);
    
    if (mypkt->datalen() != expected_size) {
	DPRINTF(DEBUG_MEMBERSHIP|DEBUG_ERRS,
		"broadcast_membership::recv malformed packet "
		"(got %d bytes, expected %d)\n",
		mypkt->datalen(), expected_size);
	return -1;
    }
    if (mh->type == MEMBERSHIP_NEW) {
	new_peer = 1;
    }
    me = (struct ron_membership_entry *)((char *)mypkt->data() +
					 sizeof(struct ron_membership_header));
    DPRINTF(DEBUG_MEMBERSHIP,
            "broadcast_membership::recv packet has %d peers\n", npeers);

    for (int i = 0; i < npeers; i++, me++) {
	node n;
	n.node_addr = me->src_addr;
	n.node_port = me->src_port;
	/* Find 'n' in the peermap */
	/* If it's not there, insert it */
	/* If it is there, update its expiration time if necessary */
	if (myforw->is_local(&n)) { continue; }
	if (peermap[n].last_update == fixedpoint(0,0)) {
	    new_peer = 1;
	}
	if (peermap[n].last_update < (nowpoint - me->age)) {
	    peermap[n].last_update = (nowpoint - me->age) -
		fixedpoint(ONEHOP_SECONDS, 0);
	    DPRINTF(DEBUG_MEMBERSHIP,
		    "broadcast_membership::recv updated node %s\n",
		    node(me->src_addr, me->src_port).to_string().c_str());
	}
    }

    if (new_peer) {
	    if (!special_request_outstanding) {
            TObjectFunctor<broadcast_membership> *functor  = 
                        new TObjectFunctor<broadcast_membership>(
                                                this, 
                                                &broadcast_membership::timer_callback,
                                                (void *)1
                                            );
    
            TimerEvent *timer_event = TimerEvent::create (
                                SPECIAL_UPDATE_TIMER /1000,
                                0, 
                                (TFunctor *) functor, 
                                "Broadcast membership - SPECIAL_UPDATE_TIMER",
                                true
                            );
    
            ronOverlay->addTimerEvent(timer_event);


	        special_request_outstanding = 1;
	    }
    }

    clean_peers();

    // (SAR) Had to change so savelist file name prefixed by its IP address
    char fname[30];
    
    // (SAR) Not likely to overflow...
#ifdef PERSONALIZED_PEERCACHE
	sprintf(fname, "%s-peercache.txt", inet_ntoa(myforw->get_myaddr()));
#else
	sprintf(fname, "peercache.txt");
#endif
    
    savelist(fname, &peerlist);
    return 0;
}

int
broadcast_membership::pkt_create(pkt *mypkt)
{
    struct ron_membership_header *mh;
    struct ron_membership_entry *me;
    int npeers;

    struct timeval tnow;
    gettimeofday(&tnow, NULL);
    nowpoint.set(tnow.tv_sec, tnow.tv_usec);

    npeers = peerlist.size() + 1; /* Announce myself too */

    DPRINTF(DEBUG_MEMBERSHIP,
	    "broadcast_membership::pkt_create starting with %d members\n",
	    npeers);

    mh = (struct ron_membership_header *)mypkt->data();
    mh->nmembers = npeers;
    if (npeers <= 1) {
	mh->type = MEMBERSHIP_NEW;
    } else {
	mh->type = MEMBERSHIP_UPDATE;
    }

    /* XXX:  Check for size overflow. */
    me = (struct ron_membership_entry *)((char *)mypkt->data() +
					 sizeof(struct ron_membership_header));
    map<node, node_data>::const_iterator iter;

    /* First, stick me on the list and give me a zero age */

    me->src_addr = myforw->get_myaddr();
    me->src_port = myforw->get_port();
    me->op = OP_JOIN;
    me->age.set(0); /* Totally happily fresh */
    me++;

    /* And then announce people I know about */
    for (iter = peermap.begin(); iter != peermap.end(); iter++) {
	me->src_addr = (*iter).first.node_addr;
	me->src_port = (*iter).first.node_port;
	me->age = nowpoint - (*iter).second.last_update;
	DPRINTF(DEBUG_MEMBERSHIP,
                "   sending node %s (age %s)\n",
                iter->first.to_string().c_str(),
                me->age.to_string().c_str());
	me++;
    }

    mypkt->set_datalen(sizeof(struct ron_membership_entry)*npeers +
		       sizeof(struct ron_membership_header));
    mypkt->ron_hdr()->src_addr = myforw->get_myaddr();
    mypkt->ron_hdr()->src_port = myforw->get_port();
    mypkt->set_type(RON_MEMBERSHIP);
    return 0;
}

/* If dst is null, send to everyone */
int
broadcast_membership::send_announce(list<node> *tolist) {
    DPRINTF(DEBUG_MEMBERSHIP,
	    "broadcast_membership::send_announce starting\n");
    pkt *mypkt = new pkt();
    if (!mypkt) {
	DPRINTF(DEBUG_ERRS | DEBUG_MEMBERSHIP,
		"broadcast_membership::send_announce alloc failed\n");
	return ENOMEM;
    }
    if (pkt_create(mypkt)) {
	DPRINTF(DEBUG_ERRS | DEBUG_MEMBERSHIP,
		"broadcast_membership::send_announce pkt_create failed\n");
	delete mypkt;
	return -1;
    }

    if (tolist) {
	list<node>::const_iterator i;
	for (i = tolist->begin(); i != tolist->end(); i++) {
	    node dst = (*i);
	    mypkt->set_dst(dst);
	    mypkt->set_hlim(DEFAULT_RON_HLIM);
	    DPRINTF(DEBUG_MEMBERSHIP,
		    "broadcast_membership::send_announce direct to %s\n",
		    dst.to_string().c_str());
	    dispatch(mypkt, &dst, 0);
	}
    } else {
	pkt *flooder;
	mypkt->set_dst(myforw->mynode());
	flooder = fp->encapsulate(mypkt); /* Give it to the flood protocol */
	if (!flooder) {
	    DPRINTF(DEBUG_ERRS | DEBUG_MEMBERSHIP,
		    "broadcast_membership::send_announce encapsulate failed\n");

	    delete mypkt;
	    return -1;
	}
	dispatch(flooder, NULL, 0);
	delete flooder;
    }

    delete mypkt;
    return 0;
}

#ifdef TEST_BROADCAST_MEMBERSHIP
int main() {
    pkt mypkt;
    debug = 0xfffffff;
    broadcast_membership *bm = new broadcast_membership();
    printf("initlisting\n");
    bm->initlist("peers.txt", &(bm->peercache));
    bm->savelist("peercache.txt", &(bm->peercache));
    bm->initlist("peercache.txt", &(bm->peerlist));
    bm->pkt_create(&mypkt);
    delete bm;
    printf("Done\n");
}
#endif 
