#include "ron.h"
#include "forw.h"
#include "membership.h"
#include "flood_protocol.h"
#include "rondebug.h"
#include <list>

using namespace RON;

void
flood_protocol::configure(Scheduler *s, forw *f) {
    sched = s;
    myforw = f;
    f->register_type(RON_FLOOD, this, NULL);
}

inline int
flood_protocol::check_cache(pkt *packet)
{
    struct ron_flooding_header *fh;
    fh = (struct ron_flooding_header *)packet->data();
    u_int64_t id;

    id = ((u_int64_t)ntohl(fh->id1) << 32) | htonl(fh->id2);

    DPRINTF(DEBUG_FLOOD, "flood_protocol::check_cache id 0x%qx\n", id);

    map<u_int64_t, unsigned int>::const_iterator p;

    p = cache1.find(id);
    if (p != cache1.end()) {
	return cache1[id];
    }
    p = cache2.find(id);
    if (p != cache2.end()) {
	return cache2[id];
    }
    return 0;
}

inline int
flood_protocol::add_to_cache(pkt *packet)
{
    struct ron_flooding_header *fh;
    fh = (struct ron_flooding_header *)packet->data();
    u_int64_t id;

    id = ((u_int64_t)ntohl(fh->id1) << 32) | htonl(fh->id2);
    DPRINTF(DEBUG_FLOOD, "flood_protocol::add_to_cache id 0x%qx\n", id);
    if (current_cache == 1) {
	cache1[id]++;
	if (cache1.size() >= max_cache_size) {
	    DPRINTF(DEBUG_FLOOD, "flood_protocol::add_to_cache clear 2\n");
	    current_cache = 2;
	    cache2.clear();
	}
    } else {
	cache2[id]++;
	if (cache2.size() >= max_cache_size) {
	    DPRINTF(DEBUG_FLOOD, "flood_protocol::add_to_cache clear 1\n");
	    current_cache = 1;
	    cache1.clear();
	}
    }
    return 0;
}

/* XXX:  This is not a well-thought-out interface for encapsulate */

pkt *
flood_protocol::encapsulate(pkt *packet) {
    pkt *ret = new pkt();
    if (!ret) {
	DEBUG_PERROR("Could not allocate pkt");
	return NULL;
    }
    struct ron_flooding_header *fh;
    fh = (struct ron_flooding_header *)ret->data();

    /* Check the length to ensure that we can fit it in the encapsulation */

    int found = 1;

    /* Spin until we get a unique-ish ID.  Shouldn't be necessary
     * with a 64-bit namespace, but a little paranoia never hurt
     * anyone */
    
    while (found) {
	fh->id1 = random();
	fh->id2 = random();
	if (!check_cache(packet)) {
	    found = 0;
	}
    }

    *ret->ron_hdr() = *packet->ron_hdr();
    
    fh->new_type = packet->ron_hdr()->type;
    fh->src_addr = packet->ron_hdr()->src_addr;
    fh->src_port = packet->ron_hdr()->src_port;

    ret->set_type(RON_FLOOD);
    ret->set_hlim(DEFAULT_RON_HLIM);
    ret->set_datalen(packet->datalen() + sizeof(struct ron_flooding_header));

    memcpy(ret->data() + sizeof(struct ron_flooding_header),
	   packet->data(),
	   packet->datalen());
    return ret;
}

int
flood_protocol::recv(pkt *packet, node *mynode, int via_ron) {
    struct ron_flooding_header *fh;

    if (packet->datalen() < sizeof(struct ron_flooding_header)) {
	DPRINTF(DEBUG_FLOOD|DEBUG_ERRS,
		"flood_protocol::recv runt packet (%d bytes)\n",
		packet->datalen());
	return -1;
    }

    assert(myforw != NULL);
    /* If we have no peers, exit early */
    if (!myforw->get_membership()) {
	DPRINTF(DEBUG_ERRS|DEBUG_MEMBERSHIP,
		"flood_protocol::recv no membership\n");
	return -1;
    }
    if (!myforw->get_membership()->peers()) {
	DPRINTF(DEBUG_ERRS|DEBUG_MEMBERSHIP,
		"flood_protocol::recv no peers\n");
	return 0;
    }
    

    fh = (struct ron_flooding_header *)packet->data();

    /* Check to see if we've seen it in the cache before */

    int count;
    if ((count = check_cache(packet))) {
	
	DPRINTF(DEBUG_FLOOD,
		"flood_protocol::recv duplicate packet 0x%lx%lx: %d\n",
		ntohl(fh->id1), ntohl(fh->id2), count);
	return 0;
    }
		
    /* Add to cache */
    add_to_cache(packet); /* Increment the count */
    
    /* Cache these;  we'll avoid sending back to them for politeness */
    node original_sender, our_sender;
    
    original_sender.node_addr = fh->src_addr;
    original_sender.node_port = fh->src_port;
    our_sender.node_addr = packet->ron_hdr()->src_addr;
    our_sender.node_port = packet->ron_hdr()->src_port;
    /*
     * Take the existing packet and send it to the local delivery
     * mechanism, with its new type header
     */

    pkt *newpkt;
    newpkt = new pkt();
    *newpkt->ron_hdr() = *packet->ron_hdr();
    memcpy(newpkt->data(),
	   packet->data() + sizeof(struct ron_flooding_header),
	   packet->datalen() - sizeof(struct ron_flooding_header));

    newpkt->ron_hdr()->type = fh->new_type;
    newpkt->ron_hdr()->src_addr = fh->src_addr;
    newpkt->ron_hdr()->src_port = fh->src_port;
    newpkt->set_hlim(DEFAULT_RON_HLIM);
    newpkt->set_datalen(packet->datalen() - sizeof(struct ron_flooding_header));

    DPRINTF(DEBUG_FLOOD,
	    "flood_protocol::recv %s sent us type %d from %s\n",
	    packet->get_src().to_string().c_str(),
	    newpkt->get_type(),
	    newpkt->get_src().to_string().c_str());
    dispatch(newpkt, NULL, 1);

    /* Now send the packet to the other peers */

    *newpkt->ron_hdr() = *packet->ron_hdr();
    memcpy(newpkt->data(),
	   packet->data(),
	   packet->datalen());
    newpkt->set_datalen(packet->datalen());
    newpkt->ron_hdr()->src_addr = myforw->get_myaddr();
    newpkt->ron_hdr()->src_port = myforw->get_port();

    list<node> tmp;
    list<node>::const_iterator iter;

    /*
     * Store a temporary copy of the list here, because some of
     * our forwarding may modify it
     */
    tmp = *(myforw->get_membership()->peers());
    
    for (iter = tmp.begin(); iter != tmp.end(); iter++) {
	node dst = *iter;
	if (myforw->is_local(&dst)
	    || dst == original_sender
	    || dst == our_sender) {
	    continue;
	}
	/* The forwarder modifies the hop limit */
	newpkt->set_hlim(DEFAULT_RON_HLIM);
	newpkt->ron_hdr()->dst_addr = (*iter).node_addr;
	newpkt->ron_hdr()->dst_port = (*iter).node_port;
	DPRINTF(DEBUG_FLOOD,
	    "flood_protocol::recv retrans to %s\n",
		dst.to_string().c_str());
	/* Tiny heuristic:  Don't send back to the original sender,
	   or the person who sent to us: original_sender, our_sender */


	if (dispatch(newpkt, &dst, 0)) {
	    fprintf(stderr, "flood_protocol::recv error in retransmission to ");
	    perror(dst.to_string().c_str());
	    
	}
    }
    DPRINTF(DEBUG_FLOOD, "flood_protocol::recv complete\n");

    delete newpkt;
    /* Set up the header */
    return 0;
}
