#include "latency_router.h"
#include "rondebug.h"
#include "ronlib.h"
#include "membership.h"
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <math.h>
#include <time.h>

using namespace RON;

#define CALLBACK_INTERVAL 1400000
#define DIRECT_BONUS      1.05
#define HYSTERESIS_BONUS  1.05
#define DIRECT_BW_BONUS   1.50

#define Q_PDB_LATENCY 0
#define Q_PDB_LOSS    1
#define Q_PDB_REACH   2

void
RON::print_rement_table(map<rement, fixedpoint> *table, char *name)
{
    map<rement, fixedpoint>::iterator iter;

    DPRINTF(DEBUG_ROUTING, "\n+---------- %-5.5s ----------+\n", name);

    for (iter = table->begin(); iter != table->end(); ++iter) {
	DPRINTF(DEBUG_ROUTING, "%s = %f \n",
		iter->first.to_string().c_str(), iter->second.to_double());
    }
    
    DPRINTF(DEBUG_ROUTING, "+---------------------------+\n");
}

void
latency_router::configure(Scheduler *s, forw *f)
{
    sched = s;
    myforw = f;
    f->register_type(RON_ROUTING, this, NULL);


    TObjectFunctor<latency_router> *functor  = 
                        new TObjectFunctor<latency_router>(
                                                this, 
                                                &latency_router::timer_callback,
                                                NULL
                                            );
    
    TimerEvent *timer_event = TimerEvent::create(
                                CALLBACK_INTERVAL/1000,
                                0,
                                (TFunctor *) functor, 
                                "latency router: timer",
                                true
                            );
    
    ronOverlay->addTimerEvent(timer_event);


    
    DPRINTF(DEBUG_INIT|DEBUG_ROUTING,
	    "latency_router::configure with forw %p\n", f);
}

void
latency_router::init()
{

    init_routers();
    init_tables();
    router_ent **routers;

    for (int j = 0; j < num_classifiers; j++) {
	routers = type_routers[j]->routers;
	
	for (int i = 0; i < num_routers; i++) {
	    routers[i]->table.clear();
	}
    }
    
    output_routes = 0;
    while (!table_version || table_version > 2147483648U) {
	table_version = random();
    }

    myforw = NULL;
    link_log = NULL;
}

latency_router::latency_router(pdb_client *npdbc, RONOverlayNetwork * ronOverlay)
{
    this->ronOverlay = ronOverlay;
    pdbc = npdbc;
    init();
}

latency_router::latency_router(unsigned long newmyaddr, 
                               unsigned short newmyport, pdb_client *npdbc, RONOverlayNetwork *ronOverlay)
{
    this->ronOverlay = ronOverlay;
    pdbc = npdbc;
    init();
    mynode.node_addr.s_addr = newmyaddr;
    mynode.node_port = htons(newmyport);
}

void
latency_router::set_visual(int vis) {
    if (vis) {
	//link_log = fopen("link.log", "a");
	
	unlink("/tmp/ron-fifo");
	if (mkfifo("/tmp/ron-fifo", 0644)) {
	    perror("latency_router::set_visual mkfifo");
	    return;
	}
	/*
	 * This is kind of a nasty game to get around the fact that
	 * one can't open a fifo O_NONBLOCK | O_WRONLY with FreeBSD.
	 * Open a reader nonblocking, open the writer, then close
	 * the reader.  Ensure that whomever opens us has signals
	 * blocked for SIGPIPE, or it'll kill us.
	 */
	
	int tmp;
	tmp = open("/tmp/ron-fifo", O_RDONLY | O_NONBLOCK);
	if (tmp == -1) {
	    perror("latency_router::set_visual mkfifo");
	    return;
	}
	output_fileno = open("/tmp/ron-fifo",
			     O_WRONLY, 064);
	
	if (output_fileno == -1) {
	    perror("latency_router::set_visual open fifo");
	    return;
	}
	close(tmp);

	int flags = fcntl(output_fileno, F_GETFL);
	if (flags == -1) {
	    perror("latency_router::set_visual fcntl get flags");
	    return;
	}
	flags |= O_NONBLOCK;
	if (fcntl(output_fileno, F_SETFL, flags)) {
	    perror("latency_router::set_visual fcntl set flags");
	    return;
	}
	output_filep = fdopen(output_fileno, "w");
	if (!output_filep) {
	    perror("latency_router::set_visual fdopen from fileno");
	    return;
	}
	
    }
    output_routes = vis;
    return;
}


void
latency_router::timer_callback(void *data)
{
    static int count;
    count++;

    now_seconds = time(NULL);

    update_local_links();
    rebuild_forwarding_table();
    if (count > 10) count = 0;

    if (count == 2) {
#ifdef DEBUG
 	if (debug & DEBUG_ROUTING) dump(stdout);
#endif
    }
    if (output_routes) dump_visual_routes();
    if (count == 3) {
	transmit_routing_summary();
    }
    if (count == 7 && (link_log != NULL)) {
	dump_linktable(link_log);
    }
	

    TObjectFunctor<latency_router> *functor  = 
                        new TObjectFunctor<latency_router>(
                                                this, 
                                                &latency_router::timer_callback,
                                                NULL
                                            );
    
    TimerEvent *timer_event = TimerEvent::create (
                                CALLBACK_INTERVAL+(random()%100000)/1000,
                                0,
                                (TFunctor *) functor, 
                                "latency router: timer",
                                true
                            );
    
    ronOverlay->addTimerEvent(timer_event);

}

/*
 * Request all of the local link information from the pdb
 */

void
latency_router::update_local_links()
{
    list<node>::const_iterator iter;

    if (!myforw) return;

    rement ent;
    ent.src = mynode;

    for (iter = myforw->get_membership()->peers()->begin();
	 iter != myforw->get_membership()->peers()->end();
	 iter++) {
	if (!myforw->is_local(&*iter)) {
	    fixedpoint result;
	    ent.dst = *iter;
	    if (!pdbc->query(*iter, PDB_REACH, SUMMARIZE_AVG,
			     fixedpoint(100,0), &result)) {
		rem_reach_table[ent] = result;
	    }
	    if (!pdbc->query(*iter, PDB_REACH, SUMMARIZE_MAX,
			     fixedpoint(4,0), &result)) {
		rem_out_table[ent] = result;
	    }
	    if (!pdbc->query(*iter, PDB_LAT, SUMMARIZE_EWMA,
			     fixedpoint(0.9), &result)) {
		rem_lat_table[ent] = result;
	    }
	}
    }
}

int
latency_router::fill_entries(struct ron_routing_entry *re,
			     int use_thresh)
{
    int nroutes = 0;
    list<node>::const_iterator iter;
    
    struct rement ent;
    ent.src = mynode;
    
    for (iter = myforw->get_membership()->peers()->begin();
	 iter != myforw->get_membership()->peers()->end();
	 iter++) {

	assert (!((*iter) == mynode)); /* temporary test */

	ent.dst = (*iter);


	for (int i = 0; i < ntables; i++) {
	    map<rement, fixedpoint>::iterator val;
	    fixedpoint lat = 0, lastlat = 0;

	    val = tables[i].curmap->find(ent);
	    if (val != tables[i].curmap->end()) {
		lat = val->second;
	    }

	    val = tables[i].lastmap->find(ent);
	    if (val != tables[i].lastmap->end()) {
		lastlat = val->second;
	    }

	    double diff, old;
	    diff = (lat - lastlat).to_double();
	    old = lastlat.to_double();
	    
	    // Might be nice to add an absolute difference test as
	    // well, at least for latency.  1ms difference is irrelevant,
	    // which would help optimize local nodes..
	    if (!use_thresh || lastlat == 0 ||
		fabs(diff / old) > tables[i].transmit_thresh) {
		re->dst_addr = (*iter).node_addr;
		re->dst_port = (*iter).node_port;
		re->metric = tables[i].metric;
		re->val = lat.to_wire();
		DPRINTF(DEBUG_ROUTING_MESSAGES,
			"  %d %s: %s\n",
			tables[i].metric, ent.to_string().c_str(),
			lat.to_string().c_str());
	    
		re++;
		nroutes++;
	    }
	}
    }
    return nroutes;
}

void
latency_router::transmit_routing_summary()
{
    pkt *mypkt;
    int npeers = myforw->get_membership()->peers()->size();
    
    if (npeers <= 0) {
	DPRINTF(DEBUG_ROUTING,
		"latency_router::transmit_routing summary found no peers\n");
	return;
    }

    mypkt = new pkt();
    if (!mypkt) {
	DEBUG_PERROR("latency_router::transmit_routing_summary could not create packet");
	return;
    }
    /* Create the packet */
    struct ron_routing_header *rh;
    struct ron_routing_entry *re;
    rh = (struct ron_routing_header *)mypkt->data();
    re = (struct ron_routing_entry *)(mypkt->data() + sizeof(struct ron_routing_header));
    // XXX: Assume they'll all fit in a packet.

    list<node>::iterator iter; /* Make this const again some day */
    unsigned int nroutes = 0;
    int send_full = 0;

    nroutes = fill_entries(re, 1); /* The differential packet */
    
    DPRINTF(DEBUG_ROUTING_MESSAGES,
	    "latency_router::transmit_routing_summary %d routes\n",
	    nroutes);

    mypkt->ron_hdr()->src_addr = myforw->get_myaddr();
    mypkt->ron_hdr()->src_port = myforw->get_port();
    mypkt->set_type(RON_ROUTING);
    // Set the packet data size
    mypkt->set_datalen((nroutes * sizeof(struct ron_routing_entry)) +
		       sizeof(struct ron_routing_header));

    /* Yes, this is slow.  oh well. */
    u_int32_t old_table_version = table_version;
    if (nroutes > 0) {
	last_lat_table = rem_lat_table;
	last_reach_table = rem_reach_table;
	last_out_table = rem_out_table;
	table_version++;
    }

    rh->nroutes = nroutes;
    rh->src_addr = mynode.node_addr;
    rh->src_port = mynode.node_port;
    rh->table_version = htonl(table_version);

    for (iter = myforw->get_membership()->peers()->begin();
	 iter != myforw->get_membership()->peers()->end();
	 iter++) {
	if (!myforw->is_local(&*iter) && is_alive(*iter)) {
	    if (old_table_version != remote_versions[*iter]) {
		send_full = 1;
	    } else if (nroutes > 0) {
		mypkt->set_dst(*iter);
		mypkt->set_hlim(DEFAULT_RON_HLIM);
		dispatch(mypkt, &(*iter), 1);
		myforw->stats.counter_table["sent update entries"] += nroutes;
		myforw->stats.counter_table["sent routing packets"]++;
		myforw->stats.counter_table["partial routing updates"]++;
		DPRINTF(DEBUG_ROUTING_MESSAGES,
			"latency_router::transmit %d diff v. %u to %s\n",
			nroutes, table_version, iter->to_string().c_str());
	    }
	} /* end of local and alive check */
    }
	
    if (send_full) {
	/* Fill with all routes */
	re = (struct ron_routing_entry *)(mypkt->data() + sizeof(struct ron_routing_header));
	nroutes = fill_entries(re, 0);
	rh->nroutes = nroutes;
	mypkt->set_datalen((nroutes * sizeof(struct ron_routing_entry)) +
			   sizeof(struct ron_routing_header));
	
	for (iter = myforw->get_membership()->peers()->begin();
	     iter != myforw->get_membership()->peers()->end();
	     iter++) {
	    if (!myforw->is_local(&*iter) && is_alive(*iter)) {
		if (old_table_version != remote_versions[*iter]) {
		    mypkt->set_dst(*iter);
		    mypkt->set_hlim(DEFAULT_RON_HLIM);
		    dispatch(mypkt, &(*iter), 1);
		    myforw->stats.counter_table["sent update entries"] += nroutes;
		    myforw->stats.counter_table["sent routing packets"]++;
		    myforw->stats.counter_table["full routing updates"]++;
		    DPRINTF(DEBUG_ROUTING_MESSAGES,
			    "latency_router::transmit %d full v. %u to %s\n",
			    nroutes, table_version, iter->to_string().c_str());
		}
	    }
	}
    }
    delete mypkt;

    myforw->stats.counter_table["possible routing packets"] += npeers;

    myforw->stats.counter_table["possible routing entries"] +=
	npeers * npeers * 3;
}

node *
latency_router::lookup(const node *dst, pkt *mypkt)
{
    return lookup(dst, mypkt, MPPolicy::MP_DEFAULT, 0.0);
}

node *
latency_router::lookup(const node *dst, pkt *mypkt, int mp_policy,
		       double mp_param)
{
    u_int16_t routing_flags;
    u_int16_t traffic_type;
    u_int32_t fid;

    router_ent **routers = NULL;
    router_ent *router;
    /* We haven't been initialized yet */
    if (!myforw) return NULL;
    int classno = 0;

    routing_flags = mypkt->get_routing_flags();
    traffic_type = mypkt->get_traffic_type();
    fid = mypkt->get_flowid();

    /* Find the right set of tables based upon traffic type */
    for (int j = 0; j < num_classifiers; j++) {
	if (type_routers[j]->p->traffic_type() == traffic_type) {
	    routers = type_routers[j]->routers;
	    DPRINTF(DEBUG_PACKETS,
		    "latency_router:lookup type %d matched type %d\n",
		    traffic_type, j);
	    classno = j;
	    break;
	}
    }
    if (!routers) {
	DPRINTF(DEBUG_PACKETS, "latency_router::lookup type %d fell through\n",
		traffic_type);
	routers = type_routers[0]->routers;
	classno = 0; 
    }

    classno = 1;	//SOSP_HACK_DILIP

    /* Find the appropriate routing table */
     router = routers[2]; /* latency */

    for (int i = 0; i < num_routers; i++) {
	if (routers[i]->routing_flags == routing_flags) {
	    DPRINTF(DEBUG_PACKETS,
		    "latency_router::lookup flags %d match router %d\n",
		    routing_flags, i);
	    router = routers[i];
	}
    }
    
	router = routers[2]; /* throughput */


//#ifdef 0
    DPRINTF (DEBUG_PACKETS, "Packet to %s classified as type %s and router %s\n",
	   dst->to_string().c_str(),
	   type_routers[classno]->p->to_string().c_str(),
	   router->to_string().c_str());
//#endif

    node *newnode = new node();
    // (SAR) This is not necessary as the "new" operator
    // will throw an exception if it cannot allocate memory...effectively
    // aborting the program
    // Which sucks.  I hate C++ sometimes.  --dga
    //     if (!newnode) {
    // 	DEBUG_PERROR("latency_router::lookup could not allocate newnode");
    // 	return NULL;
    //     }
    
    if (mp_policy == MPPolicy::MP_DEFAULT) {
	map<node, node>::const_iterator p = router->table.find(*dst);
	 
	if (p == router->table.end() || p->second.node_addr.s_addr == 0) {
	    delete newnode;
	    return NULL;
	}
	 
	*newnode = p->second;
    } else {
	if (mp_policy < MPPolicy::NUM_MPPOLICIES) {
	    DPRINTF(DEBUG_MULTIPATH,
		    "latency_router::USING multipath policy %s\n",
		    MPPolicy::multipathPolicies[mp_policy]->get_name().c_str());
	    *newnode = 
		MPPolicy::multipathPolicies[mp_policy]->decide(router->pathRankings[*dst], 
							       mp_param);
	    if (newnode->node_addr.s_addr == 0) { // Going direct
		delete newnode;
		return NULL;
	    }
	} else {
	    DPRINTF(DEBUG_MULTIPATH,
		    "latency_router::lookup uknown multipath policy %d\n",
		    mp_policy);
	    delete newnode;
	    return NULL;
	}
    }
    return newnode;
}

/*
 * Print the forwarding table
 */

void
latency_router::dump(FILE *out)
{
    list<node>::const_iterator iter;

    if (!myforw) return;

    fprintf(out, "%16.16s", "");
    for (int i = 0; i < num_classifiers; i++) {
	fprintf(out, "%-32.32s ", type_routers[i]->p->to_string().c_str());
    }
    fprintf(out, "\n");
    
    for (int category = 0; category < num_routers; category++) {
	int first = 1;
	for (iter = myforw->get_membership()->peers()->begin();
	     iter != myforw->get_membership()->peers()->end();
	     iter++) {
	    node hop;
	    if (first) 
		fprintf(out, "  %4.4s router   ",
			type_routers[0]->routers[category]->to_string().c_str());
	    else fprintf(out, "                ");
	    first = 0;
	    for (int i = 0; i < num_classifiers; i++) {
		char strtmp[128];
		router_ent **routers = type_routers[i]->routers;
		hop = routers[category]->table[(*iter)];

		sprintf(strtmp, "%s: %s ",			
			(*iter).to_string().c_str(),
			hop.is_null() ? "direct" : hop.to_string().c_str());
		fprintf(out, "%-32.32s ", strtmp);
	    }
	    fprintf(out, "\n");
	    
	    // (SAR) Multipath
#ifdef DEBUG
	    if (debug & DEBUG_MULTIPATH)
		type_routers[0]->routers[category]->pathRankings[(*iter)].print(out);
#endif
	}	    
    }
}

/*
 * Return a score representing the path.
 * If 'hop' is null, evaluate the direct path.
 */
/* Calculate based on a simplified TCP throughput equation:
                  1
   X =  ---------------------
         rtt * sqrt(2 * p/3)
*/
/*
 * Note that this throughput equation ignores the effect of losses
 * on throughput, because those values depend on the TCP
 * retransmission timer.  It also assumes a uniform packet size,
 * which is OK for comparison purposes.  The real TCP throughput
 * equation is:
                                  S
   X =  ----------------------------------------------------------
           R*sqrt(2*p/3) + (t_RTO * (3*sqrt(3*p/8) * p * (1+32*p^2)))
*/
/* Where X = throughput, R = rtt, t_RTO is TCP retransmit timer,
 * and p is rho, the packet loss rate between 0.0 and 1.0 */

/* Another issue here is the way by which we combine loss rates.
 * One can take either min(p1, p2) or (p1*p2);  Which of these
 * gives a better answer?  The answer depends if we're causing
 * the packet loss by our own congestive acts or not.
 */

/* Macro-esque function to fill in data for us.
 * Please set now_seconds before calling.. */
void
latency_router::lookup_path(const node * const src,
			    const node * const dst,
			    const node * const hop,
			    fixedpoint &direct_latency,
			    fixedpoint &hop1_latency,
			    fixedpoint &hop2_latency,
			    fixedpoint &direct_reach,
			    fixedpoint &hop1_reach,
			    fixedpoint &hop2_reach,
			    fixedpoint &direct_outage,
			    fixedpoint &hop1_outage,
			    fixedpoint &hop2_outage,
			    int &indirect_timeout)
{
    rement ent;
    ent.src = *src;
    ent.dst = *dst;

    direct_latency = rem_lat_table[ent];
    direct_reach   = rem_reach_table[ent];
    direct_outage  = rem_out_table[ent];

    if (!hop) return;
    ent.dst = *hop;

    hop1_latency = rem_lat_table[ent];
    hop1_reach   = rem_reach_table[ent];
    hop1_outage  = rem_out_table[ent];

    indirect_timeout = now_seconds - rem_update_table[*hop];

    ent.src = *hop;
    ent.dst = *dst;
    
    hop2_latency = rem_lat_table[ent];
    hop2_reach   = rem_reach_table[ent];
    hop2_outage  = rem_out_table[ent];
}

#define TIMEOUT_THRESH 30

double
bw_router_ent::evaluate(const node * const src,
			 const node * const dst,
			 const node * const hop)
{
    // direct, hop1, and hop2 values for latency, reach, and outage
    fixedpoint ld, l1, l2, rd, r1, r2, od, o1, o2;
    int to = 0;
    parent->lookup_path(src, dst, hop, ld, l1, l2, rd, r1, r2, od, o1, o2, to);
    
    DPRINTF(DEBUG_ROUTING_V, "BANDWIDTH ROUTER EVALUATE\n");

    if (hop) {
	ld = l1 + l2;
	rd = r1 * r2;
	od = min(o1, o2);
	
	if (l1 == 0 || l1 == 0) {
	    DPRINTF(DEBUG_ROUTING_V,
		    "Insufficient info in bw_router, l1 = %s l2 = %s\n",
		    l1.to_string().c_str(), l2.to_string().c_str());
	    return 0.0;
	}
    }
    
    if (od == 0 || to > TIMEOUT_THRESH) {
	DPRINTF(DEBUG_ROUTING_V, "Outage 0 in bw_router\n");

	return rd.to_double();
    }

    /* Cap how perfect a link can be */
    double dcd = ld.to_double();
    double drd = rd.to_double();
    
    if (drd > 0.99) { drd = 0.99; }
    double rho = (1.0 - drd);
    double ret;
    ret =  (1.0/(dcd * sqrt(2.0*rho)));

    if (!hop) { ret *= DIRECT_BW_BONUS; }  // Really favor direct..
    return ret;
}

double
lat_router_ent::evaluate(const node * const src,
			 const node * const dst,
			 const node * const hop)
{
    fixedpoint ld, l1, l2, rd, r1, r2, od, o1, o2;
    int to = 0;
    parent->lookup_path(src, dst, hop, ld, l1, l2, rd, r1, r2, od, o1, o2, to);
    
    DPRINTF(DEBUG_ROUTING_V, "LATENCY ROUTER EVALUATE\n");

    if (hop) {
	ld = l1 + l2;
	rd = min(r1, r2);
	od = min(o1, o2);
	
	if (l1 == 0 || l1 == 0) {
	    DPRINTF(DEBUG_ROUTING_V,
		    "Insufficient info in lat_router, l1 = %s l2 = %s\n",
		    l1.to_string().c_str(), l2.to_string().c_str());
	    return 0.0;
	}
    }
    
    if (od == 0 || to > TIMEOUT_THRESH) {
	DPRINTF(DEBUG_ROUTING_V, "Outage 0 in lat_router\n");
	return rd.to_double();
    }

    double ret = (1.0/ld.to_double());
    if (!hop) {
	ret *= 1.02; /* Small advantage by avoiding encapsulation */
    }
    return ret;
}

double
loss_router_ent::evaluate(const node * const src,
			  const node * const dst,
			  const node * const hop)
{
    fixedpoint ld, l1, l2, rd, r1, r2, od, o1, o2;
    int to = 0;
    parent->lookup_path(src, dst, hop, ld, l1, l2, rd, r1, r2, od, o1, o2, to);
    
    DPRINTF(DEBUG_ROUTING_V, "LOSS ROUTER EVALUATE\n");

    if (hop) {
	rd = r1 * r2;
	od = min(o1, o2);
    }
    
    double drd = rd.to_double();

    if (od == 0 || to > TIMEOUT_THRESH) {
	DPRINTF(DEBUG_ROUTING_V, "Outage 0 in loss_router\n");
	return drd / 2;
    }
    return drd;
}

/*
 * Compute the best path from src to dst according to the policy specified
 * by 'classifier' and the routing metric specified by 'router'
 */

node
latency_router::find_best_path(node src, node dst, int cl, int rtr)
{
    node best_peer;
    double best_score;
    list<node>::const_iterator hop_iter;

    /* Try to find the best route to get to them */
    /* First, compute the direct cost */

    router_ent **routers;
    routers = type_routers[cl]->routers;

    best_peer = dst; /* Default is direct */
    best_score = routers[rtr]->evaluate(&src, &dst, NULL);
    best_score = best_score * DIRECT_BONUS;

    /* Hysteresis.  Best previous path gets a small bonus to avoid flapping.
     * Note that we ONLY compute this if src is local, because we don't
     * store the full routing table for other peers.  Source of a bit
     * of small inaccuracy when using this routine for simulation */

    if (src == mynode) {
	if (routers[rtr]->table[dst] == best_peer) {
	    best_score = best_score * HYSTERESIS_BONUS;
	}
	// (SAR) Add to multipath rankings...
	// dga - I don't think this is the right place for this update.
	// from a conceptual standpoint.  I agree that it reduces
	// computation, but it seems odd somehow.
	routers[rtr]->pathRankings[dst].update(best_peer, best_score);
	// (SAR) This printf is more illustrative here, after best_
	//  score adjustments for preferences...
	DPRINTF(DEBUG_ROUTING_V,
		"latency_router::rft type %d examining peer: %s (%.2f direct)\n",
		rtr, dst.to_string().c_str(), best_score);
    }

    /* Now try all of the 1-intermediate routes */
    for (hop_iter = myforw->get_membership()->peers()->begin();
	 hop_iter != myforw->get_membership()->peers()->end();
	 hop_iter++) {
	/* XXX - does peers return our address ??? it better! */
	if ((*hop_iter) == src || (*hop_iter) == dst) { continue; }

	/* Is this path, 1st or 2nd hop, permitted by policy? */
	if (type_routers[cl]->p->permitted(&src, &(*hop_iter)) !=
	    POLICY_PERMITTED) {
	    DPRINTF(DEBUG_POLICY,
		    "  POLICY SKIP %s to %s for type %d\n",
		    src.to_string().c_str(),
		    (*hop_iter).to_string().c_str(),
		    type_routers[cl]->p->traffic_type());
	    continue;
	}
	if (type_routers[cl]->p->permitted(&(*hop_iter), &dst) !=
	    POLICY_PERMITTED) {
	    DPRINTF(DEBUG_POLICY,
		    "  POLICY SKIP %s to %s for type %d\n",
		    (*hop_iter).to_string().c_str(),
		    dst.to_string().c_str(),
		    type_routers[cl]->p->traffic_type());
	    continue;
	}
	double newscore;
	newscore = routers[rtr]->evaluate(&src, &dst, &(*hop_iter));
	/* Again, local only, check hysteresis and multipath */
	if (src == mynode) {
	    if (routers[rtr]->table[dst] == (*hop_iter)) {
		newscore = newscore * HYSTERESIS_BONUS;
	    }
	    // (SAR) Add to the multipath rankings...
	    routers[rtr]->pathRankings[dst].update(*hop_iter, newscore);
	    DPRINTF(DEBUG_ROUTING_V,
		    "   Considering hop through %s: %.5f\n",
		    (*hop_iter).to_string().c_str(), newscore);
	}
	if (newscore > best_score) {
	    best_score = newscore;
	    best_peer = *hop_iter;
	}
    }
    return best_peer;
}

/*
 * Rebuild the forwarding table from the routing table
 */

void
latency_router::rebuild_forwarding_table()
{
    list<node>::const_iterator iter;
    
    DPRINTF(DEBUG_ROUTING,
	    "latency_router::rebuild_forwarding_table LINK STATE TABLES\n");
    
    print_rement_table(&rem_lat_table, "LAT");
    print_rement_table(&rem_reach_table, "REACH");
    print_rement_table(&rem_out_table, "OUT");

    /* Ensure that we've been set up properly */
    assert(myforw != NULL);
    assert(myforw->get_membership() != NULL);
    if (myforw->get_membership()->peers() == NULL) {
	return;
    }

    for (iter = myforw->get_membership()->peers()->begin();
	 iter != myforw->get_membership()->peers()->end();
	 iter++) {

	/* XXX:  Testing for a while, ensure that peers is never violated */
	assert(!((*iter) == mynode));
	
	for (int cl = 0; cl < num_classifiers; cl++) {
	    router_ent **routers;
	    routers = type_routers[cl]->routers;
	    
	    for (int rtr = 0; rtr < num_routers; rtr++) {
		node best_peer = find_best_path(mynode, (*iter), cl, rtr);
		routers[rtr]->table[(*iter)] = best_peer;

		DPRINTF(DEBUG_ROUTING,
			"Router (%d):  %s to %s %s %s\n",
			rtr,
			mynode.to_string().c_str(),
			(*iter).to_string().c_str(),
			best_peer.node_addr.s_addr ? "via" : "direct",
			best_peer.node_addr.s_addr ?
			best_peer.to_string().c_str() : "");
	    }
	}
    }
}

void
latency_router::set_remote_version(const node *dst, u_int32_t version,
				   time_t update_time)
{
    /* XXX:  Could also leak.. */
    remote_versions[*dst] = version;
    rem_update_table[*dst] = update_time;
}

u_int32_t
latency_router::get_table_version(const node *dst)
{
    /* XXX:  Could leak, need to clean up every now and then */
    return my_versions[*dst];
}

int
latency_router::recv(pkt *mypkt, node *mynode, int via_ron)
{
    char *pkt_data;
    struct ron_routing_header *rh;
    struct ron_routing_entry *re;
    
    if (mypkt->datalen() < sizeof(struct ron_routing_header)) {
	DPRINTF(DEBUG_PACKETS|DEBUG_ROUTING|DEBUG_ROUTING_MESSAGES|DEBUG_ERRS,
		"latency_router::recv runt packet (%d data bytes)\n",
		mypkt->datalen());
	return -1;
    }

    pkt_data = mypkt->data();
    rh = (struct ron_routing_header *)pkt_data;
    re = (struct ron_routing_entry *)((char *)pkt_data +
					 sizeof(struct ron_routing_header));

    if (mypkt->datalen() - sizeof(struct ron_routing_header) <
	rh->nroutes * sizeof(struct ron_routing_entry)) {
	DPRINTF(DEBUG_PACKETS|DEBUG_ROUTING|DEBUG_ROUTING_MESSAGES|DEBUG_ERRS,
		"latency_router::recv runt packet (%d data bytes)\n",
		mypkt->datalen());
	return -1;
    }

    u_int32_t newvers = ntohl(rh->table_version);

    DPRINTF(DEBUG_PACKETS|DEBUG_ROUTING|DEBUG_ROUTING_MESSAGES,
	    "latency_router::recv version %u with %d routes\n",
	    newvers, rh->nroutes);
    
    myforw->stats.counter_table["received routing packets"]++;
    myforw->stats.counter_table["received routing entries"] += rh->nroutes;

    my_versions[mypkt->get_src()] = newvers;

    for (int i = 0; i < rh->nroutes; i++) {
	fixedpoint val;
	val.set(re->val);
	rement ent;
	ent.src.node_addr = rh->src_addr;
	ent.src.node_port = rh->src_port;
	ent.dst.node_addr = re->dst_addr;
	ent.dst.node_port = re->dst_port;

	if (re->metric == RON_METRIC_LATENCY) {
	    DPRINTF(DEBUG_ROUTING|DEBUG_ROUTING_MESSAGES,
		    "  lat %s: %s\n",
		    ent.to_string().c_str(), val.to_string().c_str());
	    rem_lat_table[ent] = val;
	    
	} else if (re->metric == RON_METRIC_REACH) {
	    DPRINTF(DEBUG_ROUTING|DEBUG_ROUTING_MESSAGES,
		    "  rea %s: %s\n",
		    ent.to_string().c_str(), val.to_string().c_str());
	    rem_reach_table[ent] = val;
	    
	} else if (re->metric == RON_METRIC_OUTAGE) {
	    DPRINTF(DEBUG_ROUTING|DEBUG_ROUTING_MESSAGES,
		    "  out %s: %s\n",
		    ent.to_string().c_str(), val.to_string().c_str());
	    rem_out_table[ent] = val;
	    
	} else {
	    DPRINTF(DEBUG_ROUTING|DEBUG_ROUTING_MESSAGES|DEBUG_ERRS,
		    "latency_router::receive unknown metric %d\n",
		    re->metric);
	}

	re++;
    }

    /*rebuild_forwarding_table();*/
    return 0;
}

void
latency_router::dump_linktable(FILE *outfile) {
    list<node> *peers;
    list<node>::const_iterator src, dst;
    
    peers = myforw->get_membership()->peers();
    if (!peers) return;

    fprintf(outfile, "\n%d\n", (int)time(NULL));

    for (int iter = 0; iter <= 2; iter++) {
	for (src = peers->begin(); src != peers->end(); src++) {
	    for (dst = peers->begin(); dst != peers->end(); dst++) {
		map<rement, fixedpoint> *itable = NULL; /* suppress warning */
		map<rement, fixedpoint>::const_iterator p;
		fixedpoint val;
		char *desc;
		rement ent;
		
		ent.src = *src;
		ent.dst = *dst;
		
		desc = "error";
		
		switch(iter) {
		case 0: itable = &rem_lat_table; desc = "lat"; break;
		case 1: itable = &rem_reach_table; desc = "rea"; break;
		case 2: itable = &rem_out_table; desc = "out"; break;
		}
		p = itable->find(ent);
		if (p == itable->end()) {
		    ;
		} else {
		    val = p->second;
		    fprintf(outfile,
			    "%s %s-%s: %s\n",
			    desc,
			    (*src).to_string().c_str(),
			    (*dst).to_string().c_str(),
			    val.to_string().c_str()
			    );
		}
	    }
	}
    }
}

void
latency_router::dump_visual_routes()
{
    list<node> allnodes;
    list<node>::const_iterator src, dst;

    int cl;

    /* XXX:  This should be selectable, or we should print 'em all out */
    cl = 0;
    router_ent **routers;
    routers = type_routers[cl]->routers;

    DPRINTF(DEBUG_ROUTING,
	    "latency_router::dump_visual_routes starting\n");

    /* Deal with _all_ nodes.  peers() is only other nodes.. */
    allnodes = *(myforw->get_membership()->peers());
    allnodes.push_back(mynode);
    int age = 0;

    fprintf(output_filep, "dvr s\n");

    for (src = allnodes.begin(); src != allnodes.end(); src++) {
	fprintf(output_filep, "n %s s\n", (*src).to_string().c_str());

	age = now_seconds - rem_update_table[*src];
	for (dst = allnodes.begin(); dst != allnodes.end(); dst++) {
	    fixedpoint lat, loss, out;
	    rement ent;
	    
	    /* We know how to get to ourselves... */
	    if ((*dst) == (*src))
		continue;

	    /* Try to find the best route to get to them */
	    /* And, for bookkeeping, output the direct cost and latency */
	    ent.src = (*src);
	    ent.dst = (*dst);
	    lat = rem_lat_table[ent];
	    loss = rem_reach_table[ent];
	    out = rem_out_table[ent];

	    /* Print out all of the link state information:
	     * latency (ms), loss %, and outage state */
	    fprintf(output_filep, "e %s %d %.2f %d %d\n",
		    (*dst).to_string().c_str(),
		    (lat*1000).ones,
		    (loss.to_double() * 100),
		    out.ones,
		    age);
	    for (int rtr = 0; rtr < num_routers; rtr++) {
		node best_peer = find_best_path((*src), (*dst), cl, rtr);
		
            if (!(best_peer == (*dst))) {
		        fprintf(output_filep, "r %s %s %s\n",
			        routers[rtr]->to_string().c_str(),
			        (*dst).to_string().c_str(),
			        best_peer.to_string().c_str());
		    } 
	    }
	}
	fprintf(output_filep, "n %s e\n", (*src).to_string().c_str());
    }
    fprintf(output_filep, "dvr e\n");
    if (ferror(output_filep)) {
	DEBUG_PERROR("Could not complete write to dvr fifo");
	clearerr(output_filep);
    }
    fflush(output_filep);
}
