#ifndef _LATENCY_ROUTER_H_
#define _LATENCY_ROUTER_H_

/*
 * A misnamed optimizing router.  Optimizes multiple metrics
 * along direct or one-hop indirect paths.
 *
 * The "direct" path best case is stored in the routing table
 * as the destination node's address, for ease of representation.
 */

#include "forw.h"
#include "../pdb/pdb_client.h"
#include <list>
#include "policy_classifier.h"
#include "scheduler.h"
#include <time.h>

#include "multipath.h"
#include "mp_policies.h"

#include "../../ronOCD/RONOverlayNetwork.h"

#define LATENCY_THRESH .05  /* 10% difference before we scream */
#define LOSS_THRESH    .02 /* We'll notice a 2% difference in loss rate */
#define OUTAGE_THRESH  .1  /* doesn't matter */
#define MAX_SKIP_INTERVAL  6

namespace RON {

  

class rement {
public:
    rement() { }
    rement(const rement& n) { src = n.src; dst = n.dst; }
    const rement& operator=(const rement& right) {
	src = right.src; dst = right.dst;
	return right;
    }
    bool operator==(const rement& right) {
	if (right.src == src) {
	    return (right.dst == dst);
	} 
	return (right.src == src);
    }
    bool operator<(const rement& right) const {
	if (right.src == src) {
	    return (dst < right.dst);
	} else {
	    return (src < right.src);
	}
    }

    string to_string() const {
	char mystring[128];
	bzero(mystring, sizeof(mystring));
	sprintf(mystring, " (%s,%s)",
		src.to_string().c_str(), dst.to_string().c_str());
	string s1(mystring);
	return s1;
    }

    node src;
    node dst;
};

 

class latency_router;

class router_ent {
public:
    u_int16_t routing_flags;
    map<node, node> table;

    // (SAR) Multipath extensions start...

    // pathRankings resolves a destination node 
    // into the "goodness"-based Rankings of all middle
    // hops to be taken towards this node...
    map<node, Rankings> pathRankings;

    // (SAR) Multipath extensions end...

    latency_router *parent;
    virtual string to_string() = 0;
    virtual ~router_ent() { }
    virtual double evaluate(const node * const src,
			    const node * const dst,
			    const node * const hop) = 0;
};

class lat_router_ent : public router_ent {
public:
    lat_router_ent(latency_router *npar) {
	routing_flags = RON_ROUTE_LAT;
	parent = npar;
    }
    virtual ~lat_router_ent() { }
    double evaluate(const node * const src,
		    const node * const dst,
		    const node * const hop);
    virtual string to_string() { return string("lat"); }
};

class bw_router_ent : public router_ent {
public:
    bw_router_ent(latency_router *npar) {
	routing_flags = RON_ROUTE_BW;
	parent = npar;
    }
    virtual ~bw_router_ent() { }
    double evaluate(const node * const src,
		    const node * const dst,
		    const node * const hop);
    virtual string to_string() { return string("bw"); }
};

class loss_router_ent : public router_ent {
public:
    loss_router_ent(latency_router *npar) {
	routing_flags = RON_ROUTE_LOSS;
	parent = npar;
    }
    virtual ~loss_router_ent() { }
    double evaluate(const node * const src,
		    const node * const dst,
		    const node * const hop);
    virtual string to_string() { return string("loss"); }
};

void print_rement_table(map<rement, fixedpoint> *table, char *name);

/* Describe the different metrics we know about */
struct table_desc {
    map<rement, fixedpoint> *curmap;
    map<rement, fixedpoint> *lastmap;
    double transmit_thresh;
    u_int8_t metric;
};

class latency_router : public routing_table {
    RONOverlayNetwork* ronOverlay;
    
public:
    latency_router(pdb_client *npdbc, RONOverlayNetwork *ronOverlay);
    latency_router(unsigned long newmyaddr, unsigned short newmyport,
                   pdb_client *npdbc, RONOverlayNetwork *ronOverlay);

    virtual ~latency_router() { }
    

    /* This whole batch of shit should be fixed after the SOSP
     * deadline */
    static const int num_routers = 3;
    class type_router {
    public:
	policy_classifier *p;
	router_ent *routers[num_routers];
	type_router(latency_router *par) {
	    routers[0] = ( new lat_router_ent(par) );
	    routers[1] = ( new loss_router_ent(par) );
	    routers[2] = ( new bw_router_ent(par) );
	}
    };

    static const int num_classifiers = 2;
    type_router *type_routers[num_classifiers];
    void init_routers() {
	type_routers[0] = new type_router(this);
	type_routers[0]->p = new vbns_classifier;
	/* Default:  Accept all! */
	type_routers[1] = new type_router(this);
	type_routers[1]->p = new policy_classifier;	

	// (SAR)
	init_multipath();

    }

    // (SAR) Multipath extensions end...


    /* Keep a map of (src,dst) pairs to their latencies */
    map<rement, fixedpoint> rem_lat_table;
    map<rement, fixedpoint> rem_reach_table;
    map<rement, fixedpoint> rem_out_table;
    map<node, time_t> rem_update_table;
    list<unsigned long> peers;

    /* And store the previous things we transmitted */
    map<rement, fixedpoint> last_lat_table;
    map<rement, fixedpoint> last_reach_table;
    map<rement, fixedpoint> last_out_table;

    int ntables;
    struct table_desc *tables;
    void init_tables() {
	    ntables = 3;
	    tables = (struct table_desc *)malloc(sizeof(struct table_desc) * ntables);
	    tables[0].curmap = &rem_lat_table;
	    tables[0].lastmap = &last_lat_table;
	    tables[0].transmit_thresh = LATENCY_THRESH;
	    tables[0].metric = RON_METRIC_LATENCY;

	    tables[1].curmap = &rem_reach_table;
	    tables[1].lastmap = &last_reach_table;
	    tables[1].transmit_thresh = LOSS_THRESH;
	    tables[1].metric = RON_METRIC_REACH;

	    tables[2].curmap = &rem_out_table;
	    tables[2].lastmap = &last_out_table;
	    tables[2].transmit_thresh = OUTAGE_THRESH;
	    tables[2].metric = RON_METRIC_OUTAGE;
    }
    u_int32_t table_version;

    // (SAR) Could have been done with default args,
    // but not in this case
    node *lookup(const node *dst, pkt *mypkt);  
    node *lookup(const node *dst, pkt *mypkt, int mp_policy, double param);
    void configure(Scheduler *s, forw *f);
    int recv(pkt *mypkt, node *mynode, int via_ron);
    int reload();		// Reload routes.
    void dump(FILE *out);
    void dump_visual_routes();
    void timer_callback(void *data);
    void set_visual(int vis);

    /* (SAR) Overwritten in NS emulation */
    virtual long int random(void) {
	return ::random();
    }

    u_int32_t get_table_version(const node *dst);
    void set_remote_version(const node *dst, u_int32_t version,
			    time_t update_time);

    void lookup_path(const node * const src,
		     const node * const dst,
		     const node * const hop,
		     fixedpoint &direct_latency,
		     fixedpoint &hop1_latency,
		     fixedpoint &hop2_latency,
		     fixedpoint &direct_reach,
		     fixedpoint &hop1_reach,
		     fixedpoint &hop2_reach,
		     fixedpoint &direct_outage,
		     fixedpoint &hop1_outage,
		     fixedpoint &hop2_outage,
		     int &indirect_timeout);

 protected:
    node mynode;
    
 private:
    Scheduler *sched;
    forw *myforw;
    pdb_client *pdbc;
    int skipped_intervals;
    node find_best_path(node src, node dst, int cl, int rtr);
    void rebuild_forwarding_table();
    void init();
    void update_local_links();
    void dump_linktable(FILE *outfile);
    void transmit_routing_summary();
    char *infilename;
    double evaluate(const node * const src, // God have mercy on my
		    const node * const dst, // const soul const.
		    const node * const hop);
    int fill_entries(struct ron_routing_entry *re, int use_thresh);

    int output_routes; /* Send route output to monitor program */
    int output_fileno; /* Output file node desc. */
    FILE *output_filep;
    FILE *link_log;
    map<node, u_int32_t> remote_versions;  /* Their version of my table */
    map<node, u_int32_t> my_versions;      /* My version of their table */
    time_t now_seconds;

    int is_alive(const node n) {
	struct rement ent;
	return 1;
	ent.src = mynode;
	ent.dst = n;
	map<rement, fixedpoint>::const_iterator i;
	i = rem_out_table.find(ent);
	if (i == rem_out_table.end() ||
	    i->second == 0) {
	    return 0;
	} else {
	    return 1;
	}
    }
};

}

#endif /* _LATENCY_ROUTER_H_ */
