#include "ron.h"
#include "forw.h"
#include "membership.h"
#include "mesh_protocol.h"
#include "rondebug.h"
#include <list>

using namespace RON;

mesh_protocol::mesh_protocol():Plumber() {


}

void
mesh_protocol::configure(Scheduler *s, forw *f) {
    sched = s;
    myforw = f;
    f->register_type(RON_MESH, this, NULL);
}

int mesh_protocol::socket_callback (int socknot, void *data) {

    printf ("\n\n\nError: Called mesh_protocol socket_callback!!! EMPTY function.\n");

}

int
mesh_protocol::transmit(pkt *packet, node *mynode, int via_ron) {
    int rc;
    node *dst = NULL;

    /* It takes three to tango */
    if (myforw->get_membership()->peers()->size() < 2) {
	DPRINTF(DEBUG_MESH, "mesh_protocol::transmit too few peers\n");
	return dispatch(packet, mynode, via_ron);
    }
     
    /* This duplicates some code from forw.  Hrm. */
    if (!mynode) {
	dst = new node(packet->get_dst());
	mynode = dst;
    }
    if (myforw->is_local(mynode)) {
	rc = dispatch(packet, mynode, via_ron);
	if (dst) delete dst;
	DPRINTF(DEBUG_MESH, "mesh_protocol::transmit local destination\n");
	return rc;
    }

    if (!packet->get_seq()) {
	packet->set_seq(get_next_seqno(packet));
    } else {
	DPRINTF(DEBUG_MESH, "mesh_protocol::transmit pkt has seq %d already\n",
		packet->get_seq());
    }
    DPRINTF(DEBUG_MESH,
	    "Packet sequence number now:  %d\n", packet->get_seq());

    pkt *ret = new pkt();
    if (!ret) {
	DEBUG_PERROR("Could not allocate pkt");
	return NULL;
    }
    struct ron_meshing_header *fh;
    fh = (struct ron_meshing_header *)ret->data();

    *ret->ron_hdr() = *packet->ron_hdr();
    DPRINTF(DEBUG_MESH, "mpt:  seq in copy %d\n", ret->get_seq());
    
    fh->new_type = packet->ron_hdr()->type;
    ret->set_type(RON_MESH);
    ret->set_datalen(packet->datalen() + sizeof(struct ron_meshing_header));

    memcpy(ret->data() + sizeof(struct ron_meshing_header),
	   packet->data(),
	   packet->datalen());


    /* Now, send one out the default RON channel, and one out
     * a random different host.  Send them off to to
     * our dispatch function with precomputed next hops */

    node *hop;
    hop = myforw->nexthop(mynode, packet);
    DPRINTF(DEBUG_MESH, "mesh_protocol::transmit 1 to %s\n",
	    hop->to_string().c_str());
    rc = dispatch(ret, hop, 1);
    /* Pick a random node here! */
    /* Dispatch again! */
    list<node> *peers = myforw->get_membership()->peers();
    list<node>::const_iterator peer;
    int hopidx = 0;
    int newidx = 0;
    int npeers = peers->size();

    for (peer = peers->begin();  peer != peers->end(); peer++) {
	if (*peer == *hop) { break; }
	hopidx++;
    }

    do {
	newidx = random() % npeers;
    } while (newidx == hopidx);
    
    hopidx = 0;
    for (peer = peers->begin();  peer != peers->end(); peer++) {
	if (hopidx == newidx) {
	    *hop = *peer;
	    break;
	}
	hopidx++;
    }
    DPRINTF(DEBUG_MESH, "mesh_protocol::transmit 2 to %s\n",
	    hop->to_string().c_str());
    rc = dispatch(ret, hop, 0);
    delete(hop);
    delete(ret);
    return rc;
}

int
mesh_protocol::recv(pkt *packet, node *mynode, int via_ron) {
    struct ron_meshing_header *mh;

    /* This makes for wacky implicit semantics.  Are we sure this
     * is what we want?  I don't like it.  I favor an explicit
     * "outbound" vs. "inbound" mechanism.  But this hack works for
     * now until I can separate them. */
    DPRINTF(DEBUG_MESH, "mesh_protocol::recv got packet type %d\n",
	    packet->get_type());
    if (packet->get_type() != RON_MESH) {
	return transmit(packet, mynode, via_ron);
    }
    
    if (packet->datalen() < sizeof(struct ron_meshing_header)) {
	DPRINTF(DEBUG_MESH|DEBUG_ERRS,
		"mesh_protocol::recv runt packet (%d bytes)\n",
		packet->datalen());
	return -1;
    }

    mh = (struct ron_meshing_header *)packet->data();

    /* Check to see if we've seen it in the in_cache before */

    map<flowid, seq_store>::const_iterator res;
    seq_store s;
    flowid fid(packet);
    u_int32_t pseq = packet->get_seq();

    res = in_cache.find(fid);
    if (res != in_cache.end()) {
	/* Seen the flow before.  Deal */
	s = in_cache[fid];
	for (int i = 0; i < NUMSEQS; i++) {
	    if (s.seqs[i] == pseq) {
		DPRINTF(DEBUG_MESH, "Seen %s seq %d before!\n",
		       fid.to_string().c_str(), pseq);
		return 0;
	    }
	}
	s.seqs[s.nextseq] = pseq;
	s.incr();
    } else {
	s.seqs[0] = pseq;
	s.nextseq = 1;
    }
    in_cache[fid] = s;
    /* terribly inefficient, should do that all by reference */

    /*
     * Take the existing packet and send it to the local delivery
     * mechanism, with its new type header
     */

    pkt *newpkt;
    newpkt = new pkt();
    *newpkt->ron_hdr() = *packet->ron_hdr();
    memcpy(newpkt->data(),
	   packet->data() + sizeof(struct ron_meshing_header),
	   packet->datalen() - sizeof(struct ron_meshing_header));

    newpkt->ron_hdr()->type = mh->new_type;
    newpkt->set_datalen(packet->datalen() - sizeof(struct ron_meshing_header));

    DPRINTF(DEBUG_MESH,
	    "mesh_protocol::recv %s sent us type %d\n",
	    packet->get_src().to_string().c_str(),
	    newpkt->get_type());
    dispatch(newpkt, NULL, 1);

    delete newpkt;
    return 0;
}

u_int32_t
mesh_protocol::get_next_seqno(pkt *packet)
{
    map<flowid, u_int32_t>::const_iterator res;
    flowid fid(packet);
    int nseq;

    res = out_cache.find(fid);
    if (res != out_cache.end()) {
	/* Seen the flow before.  Deal */
	nseq = out_cache[fid];
    } else {
	nseq = 1;
    }
    out_cache[fid] = nseq + 1;
    DPRINTF(DEBUG_MESH, "Next seqno for %s:  %d\n",
	   fid.to_string().c_str(), nseq);
    return nseq;
}

