#include "ron.h"
#include "forw.h"
#include "membership.h"
#include "mesh_tester.h"
#include "rondebug.h"
#include <list>
#include <fcntl.h>

using namespace RON;

#define TIMER_INTERVAL 600000

struct ron_meshping_header {
    u_int32_t id1;
    u_int32_t id2;
    u_int8_t  seqno;
    u_int8_t  probetype;
    u_int8_t  type;
    u_int8_t  padding;
};
#define TYPE_FIRST      0x0
#define TYPE_SECOND     0x1

void
mesh_tester::configure(Scheduler *s, forw *f) {
    sched = s;
    myforw = f;
    f->register_type(RON_MESHTEST, this, NULL);
    s->register_timer(TIMER_INTERVAL, this, 0);
    DPRINTF(DEBUG_INIT|DEBUG_MESHTEST,
	    "mesh_tester::configure with forw %p\n", f);
}

/* May as well have a good source of randomness for our packet ids */
/* XXX:  Note that this is not going to virtualize well for ns.
 * sorry.  - dga */
void
mesh_tester::get_random(u_int64_t *q)
{
    int rdev;
    int got = 0;

    if (devrand == -1 && !randinit) {
	devrand = open("/dev/urandom", O_RDONLY);
    }
    rdev = devrand;
    if (rdev != -1) {
	int rc;
	rc = read(rdev, q, sizeof(u_int64_t));
	if (rc == sizeof(u_int64_t)) {
	    got = 1;
	}
    }
    if (!got) {
	/* Hope this is enough - functionality is more important
	 * than uniqueness;  we can post process and time correlate
	 * to eliminate most false similarities if any occur
	 */
	if (!randinit) {
	    struct timeval tn;
	    gettimeofday(&tn, NULL);
	    srandom(tn.tv_sec + tn.tv_usec + getpid());
	}
	*q = (u_int64_t)random() << 32 | random();
    }
    randinit = 1;
}

node *
mesh_tester::pick_random_node(node *skip) {
    /* Pick a random node here! */

    list<node> *peers = myforw->get_membership()->peers();
    list<node>::const_iterator peer;
    node *dst = NULL;
    
    int hopidx = 0;
    int newidx = 0;
    int npeers = peers->size();

    if (!npeers) { return NULL; }

    for (peer = peers->begin();  peer != peers->end(); peer++) {
	if (skip && *peer == *skip) { break; }
	hopidx++;
    }

    do {
	newidx = random() % npeers;
    } while (newidx == hopidx);
    
    hopidx = 0;
    for (peer = peers->begin();  peer != peers->end(); peer++) {
	if (hopidx == newidx) {
	    dst = new node(*peer);
	    break;
	}
	hopidx++;
    }
    return dst;
}

/* With all of these, if the two are the same, pick an alternate
 * random node */

#define MESH_DIRECT       0
#define MESH_RAND         1
#define MESH_LAT          2
#define MESH_LOSS         3
#define MESH_DIRECTDIRECT 4
#define MESH_DIRECTRAND   5
#define MESH_DIRECTLAT    6
#define MESH_DIRECTLOSS   7
#define MESH_RANDRAND     8
#define MESH_RANDLAT      9
#define MESH_RANDLOSS     10
#define MESH_LATLOSS      11

#define MESH_MAXPROBE     11
#define TWOPROBE_START 4

int probes_to_use[] = { MESH_LOSS, MESH_DIRECTRAND,
			MESH_LATLOSS, -1 };

#define DELAY_NONE 0 /* XXX - delays not implemented yet */
#define DELAY_10ms 1
#define DELAY_20ms 2

static char *mesh_names[] = {
    "direct",
    "rand",
    "lat",
    "loss",
    "direct direct",
    "direct rand",
    "direct lat",
    "direct loss",
    "rand rand",
    "rand lat",
    "rand loss",
    "lat loss"
};

int
mesh_tester::send_ping(node *dst, int probetype, int type, u_int64_t id)
{
    node *intone = NULL;
    node *inttwo = NULL;
    pkt *mypkt;
    node mynode = myforw->mynode();
    int rc;

    mypkt = new pkt();

    mypkt->set_src(mynode);
    mypkt->set_dst(*dst);
    mypkt->set_type(RON_MESHTEST);

    switch(probetype) {
    case MESH_DIRECT:
	intone = new node(*dst);
	break;
    case MESH_RAND:
	intone = pick_random_node(NULL);
	break;
    case MESH_LAT:
	mypkt->set_routing_flags(RON_ROUTE_LAT);
	intone = myforw->nexthop(dst, mypkt);
	break;
    case MESH_LOSS:
	mypkt->set_routing_flags(RON_ROUTE_LOSS);
	intone = myforw->nexthop(dst, mypkt);
	break;
    case MESH_DIRECTDIRECT:
	intone = new node(*dst);
	inttwo = new node(*dst);
	break;
    case MESH_DIRECTRAND:
	intone = new node(*dst);
	inttwo = pick_random_node(intone);
	break;
    case MESH_DIRECTLAT:
	intone = new node(*dst);
	mypkt->set_routing_flags(RON_ROUTE_LAT);
	inttwo = myforw->nexthop(dst, mypkt);
	break;
    case MESH_DIRECTLOSS:
	intone = new node(*dst);
	mypkt->set_routing_flags(RON_ROUTE_LOSS);
	inttwo = myforw->nexthop(dst, mypkt);
	break;
    case MESH_RANDRAND:
	intone = pick_random_node(NULL);
	inttwo = pick_random_node(intone);
	break;
    case MESH_RANDLAT:
	intone = pick_random_node(NULL);
	mypkt->set_routing_flags(RON_ROUTE_LAT);
	inttwo = myforw->nexthop(dst, mypkt);
	break;
    case MESH_RANDLOSS:
	intone = pick_random_node(NULL);
	mypkt->set_routing_flags(RON_ROUTE_LOSS);
	inttwo = myforw->nexthop(dst, mypkt);
	break;
    case MESH_LATLOSS:
	mypkt->set_routing_flags(RON_ROUTE_LAT);
	intone = myforw->nexthop(dst, mypkt);
	mypkt->set_routing_flags(RON_ROUTE_LOSS);
	inttwo = myforw->nexthop(dst, mypkt);
	break;
    default:
	DPRINTF(DEBUG_MESHTEST|DEBUG_ERRS,
		"mesh_tester invalid probetype on transmit\n");
	exit(-1);
    }
    if ( !intone || (probetype >= TWOPROBE_START && !inttwo)) {
	DPRINTF(DEBUG_MESHTEST|DEBUG_ERRS,
		"mesh_tester could not allocate node\n");
	if (intone) delete intone;
	if (inttwo) delete inttwo;
	return -1;
    }
    if ( probetype != MESH_DIRECTDIRECT &&
	 intone && inttwo && 
	 *intone == *inttwo ) {
	delete inttwo;
	inttwo = pick_random_node(intone);
	if (!inttwo) {
	    DPRINTF(DEBUG_MESHTEST|DEBUG_ERRS,
		    "mesh_tester could not allocate node\n");
	    if (intone) delete intone;
	    return -1;
	}
    }

    DPRINTF(DEBUG_MESHTEST, "%-20s  %s  %s  %s\n",
	   mesh_names[probetype],
	   dst->to_string().c_str(),
	   intone ? intone->to_string().c_str() : "(none)",
	   inttwo ? inttwo->to_string().c_str() : "(none)");

    /* We now have one or two intermediate nodes and a destination.
     * send two packets to them.  Do you think it matters if we
     * send one first or the other?  We can randomize those later
     * if we choose. */

    /* Fill in the packet */

    char *data;
    struct ron_meshping_header *mph;

    data = mypkt->data();
    unsigned int datalen = sizeof(struct ron_meshping_header);
    mypkt->set_datalen(datalen);
    mypkt->set_src(mynode);
    mph = (struct ron_meshping_header *)data;

    
    mph->probetype = probetype;
    mph->type = type;
    mph->id1 = htonl(id >> 32);
    mph->id2 = htonl(id & 0xffffffff);
    
    for (int i = 0; i <= 1; i++) {
	node *inter;

	if (!i) {
	    inter = intone;
	} else {
	    inter = inttwo;
	    if (!inttwo) {
		break;
	    }
	}

	mph->seqno = i;
	mypkt->set_hlim(DEFAULT_RON_HLIM);

	/* Send it!  via_ron = 0 means send it directly to 'inter' */
	dispatch(mypkt, inter, 0);

	struct timeval time_sent;
	gettimeofday(&time_sent, NULL);
	double linetime;
	linetime = time_sent.tv_sec + ((double)time_sent.tv_usec)/(1000000.0);
	char outbuf[128];
	snprintf(outbuf, 127, "%f S 0x%qx %d %d %d %s %s %s %d\n",
		 linetime,
		 id, mph->seqno, type, datalen,
		 mynode.to_string().c_str(),
		 dst->to_string().c_str(),
		 inter->to_string().c_str(),
		 probetype);
	outbuf[127] = '\0';
	rc = write(STDOUT_FILENO, outbuf, strlen(outbuf));
    }
    
    if (intone) delete intone;
    if (inttwo) delete inttwo;
    if (mypkt) delete mypkt;
    return 0;
}

int
mesh_tester::timer_callback(void *cbdata)
{
    static int probeidx = -1;
    int probetype;
    /* Rotate through the ping types and send one using transmit */
    node *dst = NULL;

    probeidx++;
    if (probes_to_use[probeidx] == -1) {
	probeidx = 0;
    }
    probetype = probes_to_use[probeidx];
    
    /* They keep calling me */
    sched->register_timer(TIMER_INTERVAL+(random() % TIMER_INTERVAL), this, 0);

    /* It takes three to tango */
    if (myforw->get_membership()->peers()->size() < 2) {
	return 1;
    }

    /* Pick a random destination */
    dst = pick_random_node(NULL);
    
    if (!dst) {
	DPRINTF(DEBUG_MESHTEST|DEBUG_ERRS,
		"mesh_tester::timer_callback could not choose a dest\n");
	return -1;
    }

    u_int64_t id = 0;
    get_random(&id);

    DPRINTF(DEBUG_MESHTEST, "DEST:  %s  Probe type:  %d\n",
	    dst->to_string().c_str(), probetype);

    send_ping(dst, probetype, TYPE_FIRST, id);
    if (dst) delete dst;

    return 0;
}

int
mesh_tester::transmit(pkt *packet, node *mynode, int via_ron)
{
    DPRINTF(DEBUG_ALL, "mesh_tester::transmit -- we don't transmit!\n");
    return -1;
}

int
mesh_tester::recv(pkt *packet, node *mynode, int via_ron) {
    char *data;
    struct ron_meshping_header *mph;

    data = packet->data();

    /* This makes for wacky implicit semantics.  Are we sure this
     * is what we want?  I don't like it.  I favor an explicit
     * "outbound" vs. "inbound" mechanism.  But this hack works for
     * now until I can separate them. */
    DPRINTF(DEBUG_MESH, "mesh_tester::recv got packet type %d\n",
	    packet->get_type());
    if (packet->get_type() != RON_MESHTEST) {
	return -1;
    }
    
    if (packet->datalen() < sizeof(struct ron_meshping_header)) {
	DPRINTF(DEBUG_MESH|DEBUG_ERRS,
		"mesh_tester::recv runt packet (%d bytes)\n",
		packet->datalen());
	return -1;
    }

    DPRINTF(DEBUG_MESH,
	    "mesh_tester::recv %s sent us type %d\n",
	    packet->get_src().to_string().c_str(),
	    packet->get_type());

    mph = (struct ron_meshping_header *)packet->data();
    int datalen = packet->datalen();

    u_int64_t id;
    u_int8_t type;
    u_int8_t probetype;

    probetype = mph->probetype;
    type = mph->type;
    id = ntohl(mph->id1);
    id = id << 32;
    id |= ntohl(mph->id2);

    struct timeval time_sent;
    time_sent = packet->rectime;
    if (!time_sent.tv_sec) {
	gettimeofday(&time_sent, NULL);
    }
    double linetime;
    linetime = time_sent.tv_sec + ((double)time_sent.tv_usec)/(1000000.0);
    char outbuf[128];
    snprintf(outbuf, 127, "%f R 0x%qx %d %d %d %s %s %d\n",
	     linetime,
	     id, (int)mph->seqno, (int)type, datalen,
	     mynode->to_string().c_str(),
	     packet->get_src().to_string().c_str(),
	     probetype);
    int rc = write(STDOUT_FILENO, outbuf, strlen(outbuf));

    /* Check our cache to see if we've seen this guy before.
       If not, send a reply using the same mechanism */
#if 0
    if (type == TYPE_FIRST) {
	/* Send a TYPE_SECOND back to the originator if not seen before */
	map<u_int64_t, u_int32_t>::const_iterator res;
	res = cache1.find(id);
	if (res == cache1.end()) {
	    res = cache2.find(id);
	}
	if (res != cache2.end()) {
	    DPRINTF(DEBUG_MESHTEST, "Saw %llu before\n", id);
	} else {
	    if (current_cache == 0) {
		cache1[id] = 1;
		if (cache1.size() >= PCACHESIZE) {
		    current_cache = 1;
		    cache2.clear();
		}
	    } else {
		cache2[id] = 1;
		if (cache2.size() >= PCACHESIZE) {
		    current_cache = 0;
		    cache1.clear();
		}
	    }

	    node dst = packet->get_src();
	    send_ping(&dst, probetype, TYPE_SECOND, id);
	}
    }
#endif
    return 0;
}


