/*
 *  Multipath routing extensions to RON.
 *
 *
 *
 *  ------------------------ Stanislav Rost [Dec 2001]
 */

#include "latency_router.h"
#include "rondebug.h"
#include "ronlib.h"
#include "membership.h"
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <math.h>
#include <time.h>

#include "multipath.h"
#include "mp_policies.h"


namespace RON {

    unsigned rankingRow::stabilizerCounter = 0;

    int CompareByMetrics(const rankingRow *left, const rankingRow *right)
    {
	if (left->metric == right->metric)
	     return (left->stabilizer < right->stabilizer);
	 else
	     return left->metric < right->metric;
    }

    int CompareByMidhops(const rankingRow *left, const rankingRow *right)
    {
	return (left->midHop < right->midHop);
    }


/* Virtual Curcuits ------------------------------- */
/*
 * Simple table, such that given a virtual ID resolves nextHop
 */

VCTable::VCTable(int hashSize = 32) : map_(new hash_map<vcID, node>(hashSize))
{}

void VCTable::add(vcID id, const node& dest)
	{
	    (*map_)[id] = dest;
	}

int VCTable::remove(vcID id)
{
    hash_map<vcID, node>::iterator iter = map_->find(id);
    
    if (iter == map_->end())
	return MPRON_NOTFOUND;
    else
	{
	    map_->erase(id);
	    return MPRON_FOUND;
	}
}

int VCTable::remove(node dest)
{
    hash_map<vcID, node>::iterator iter;
    
    for (iter = map_->begin(); iter != map_->end(); ++iter)
	{
	    if (iter->second == dest)
		{
		    map_->erase(iter);
		    return MPRON_FOUND;
		}
	}
    
    return MPRON_NOTFOUND;

    
}

int VCTable::lookup(vcID id, node& result) const
{
    hash_map<vcID, node>::iterator iter = map_->find(id);
    
    if (iter == map_->end())
	return MPRON_NOTFOUND;
    else
	{
	    result = iter->second;
	    return MPRON_FOUND;
	}
}

/* ------------------------------- Virtual Curcuits */

/* Rankings --------------------------------------- */
/*
 *
 * Contains nextHops sorted according to a particular metric
 * 
 */

Rankings::Rankings() : lastSelected(0), needsReSort(0)
{
}

Rankings::~Rankings()
{
    for (map<node,rankingRow*>::iterator iter = index.begin();
	 iter != index.end();
	 ++iter)
	{
	    delete iter->second;
	}

    // index.clear and heap.clear will be called implicitly
}

void Rankings::print(FILE *out)
{
    reSort();

    fprintf(out,"Rankings:\n");
    fprintf(out,"\n--------------------------\n");    
    for (vector<rankingRow*>::iterator iter = heap.begin();
	 iter != heap.end();
	 ++iter)
	{
	    fprintf(out,"[Midhop: %s, Goodness: %f]\n",
		    (*iter)->midHop.to_string().c_str(),
		    (*iter)->metric);
	}
    fprintf(out,"--------------------------\n");

    
}

void Rankings::reSort()
{
    if (needsReSort)
	{
	    make_heap(heap.begin(), heap.end(), CompareByMetrics);    
 	    sort_heap(heap.begin(), heap.end(), CompareByMetrics);    

	    needsReSort = 0;
	}
}

int Rankings::update(const node& midHop, double& m)
{

    map<node,rankingRow*>::iterator iter = index.find(midHop);
	    
    DPRINTF(DEBUG_MULTIPATH, "Rankings::update(%s, %f)\n",
	    midHop.to_string().c_str(), m);
    
    if (iter == index.end())
	{
	    rankingRow *newRow = new rankingRow(midHop, m);
	    
 	    heap.push_back(newRow);
	    needsReSort = 1;
 	    
	    index[newRow->midHop] = newRow;

	    return MPRON_FOUND;

	}
    else
	{
	    rankingRow *newRow = iter->second;

	    if (newRow->metric != m)
		{
		    newRow->metric = m;

		    needsReSort = 1;
		}

	    return MPRON_FOUND;
	}
		
}

int Rankings::remove(node& midHop)
{
    map<node,rankingRow*>::iterator iter = index.find(midHop);
	    
    if (iter == index.end())
	{
	    return MPRON_NOTFOUND;
	}
    else
	{
	    vector<rankingRow *>::iterator iter2;
	    
	    for (iter2 = heap.begin(); iter2 != heap.end(); ++iter2)
		{
		    if (**iter2 == midHop)
			{
			    heap.erase(iter2);
			    break;
			}
		}

	    index.erase(iter);

	    needsReSort = 1;
	    
	    return MPRON_FOUND;
	}
}

int Rankings::getRanking(int index, node& midHop, double& value)
{
    int size = heap.size();

    if (index < size && index>=0 )
	{
	    reSort();  // Will only sort if necessary

	    rankingRow *result = heap[index];

	    midHop = result->midHop;
	    value = result->metric;

	    return MPRON_FOUND;

	}
    else
	{
	    return MPRON_NOTFOUND;
	}
}
    

/* --------------------------------------- Rankings */

MPPolicy *MPPolicy::multipathPolicies[MPPolicy::NUM_MPPOLICIES];

 void init_multipath()
     {
	 srandom(time(NULL));

	 MPPolicy::multipathPolicies[MPPolicy::MP_BESTPATH] = ( new MPBestPath() );
	 MPPolicy::multipathPolicies[MPPolicy::MP_RANDOM] = ( new MPRandom() );
	 MPPolicy::multipathPolicies[MPPolicy::MP_BESTK] = ( new MPBestK() );
	 MPPolicy::multipathPolicies[MPPolicy::MP_DTMP] = ( new MPDTMP() );
	 MPPolicy::multipathPolicies[MPPolicy::MP_RROBIN] = ( new MPRoundRobin() );
     }

}
// Close namespace


#if 0
// Lame test suite
int main(void)
{
    char *descr[] = {"NOT FOUND","FOUND"};

    // Virtual circuits
    RON::VCTable vct;
    RON::node n;
    RON::vcID vcid;

    n.node_addr.s_addr = 3;
    vcid = 15;

    vct.add(vcid, n);

    n.node_addr.s_addr = 17;
    vcid = 12;

    vct.add(vcid, n);

    printf("Lookup(vcid = 3) = %s, %s\n", descr[vct.lookup(3, n)], n.to_string().c_str());
    

    printf("Lookup(vcid = 15) = %s, %s\n", descr[vct.lookup(15, n)], n.to_string().c_str());
    

    printf("Lookup(vcid = 12) = %s, %s\n", descr[vct.lookup(12, n)],  n.to_string().c_str());
    

    vct.remove(12);

    printf("After remove:  Lookup(vcid = 12) = %s, %s\n", descr[vct.lookup(12, n)],  n.to_string().c_str());
    

    n.node_addr.s_addr = 3;

    vct.remove(n);

    printf("After remove:  Lookup(vcid = 15) = %s, %s\n", descr[vct.lookup(15, n)],  n.to_string().c_str());
    

    // Rankings
    RON::Rankings rt;
    double f;

    for (int i = 0; i < 240; i++)
	{	    
	    n.node_addr.s_addr = i;
	    double rnum = random() / (double)RAND_MAX;

	    printf("Inserting [%s,%f]\n", n.to_string().c_str(), rnum);

	    rt.update(n, rnum);
	    
	}

    printf("\n\nRankings size is %d\n\n", rt.size());

    for (unsigned i=0; i < rt.size(); i++)
	{
	    printf("Lookup %d 'th in rankings: %s, %s\n", i, descr[rt.getRanking(i, n, f)],  n.to_string().c_str());
	    
	    printf("Fixepoint val returned:  %f\n", f);
	}

    for (int i = 0; i < 240; i++)
	{	    
	    n.node_addr.s_addr = i;
	    double rnum = random() / (double)RAND_MAX;

	    printf("Update [%s,%f]\n", n.to_string().c_str(), rnum);

	    rt.update(n, rnum);
	    
	}    

    printf("\n\nRankings size is %d\n\n", rt.size());

    for (unsigned i=0; i < rt.size(); i++)
	{
	    printf("Lookup %d 'th in rankings: %s, %s\n", i, descr[rt.getRanking(i, n, f)],  n.to_string().c_str());
	    
	    printf("Fixepoint val returned:  %f\n", f);
	}

    for (unsigned i=0; i < rt.size(); i++)
	{
	    printf("Rankings(%d): %s", i, descr[rt.getRanking(i, n, f)]);
	    printf(" %s, %f\n", n.to_string().c_str(), f);
	    
	}

    // Multipath policies
    // Bestpath
    RON::MPPolicy *mp1 = new RON::MPBestPath();

    printf("BESTPATH: %s\n", mp1->decide(rt).to_string().c_str());


    // BestK
    RON::MPPolicy *mp2 = new RON::MPBestK();

    for (int i=0;i<10;i++)
	{
	    printf("BBESTK(3): %s\n", mp2->decide(rt, 3.0).to_string().c_str());
	}

    // DTMP
    RON::MPPolicy *mp3 = new RON::MPDTMP();

    for (int i=0;i<10;i++)
	{
	    printf("DTMP(1.05): %s\n", mp3->decide(rt, 1.05).to_string().c_str());
	}
}

#endif
