#include "scheduler.h"
#include <errno.h>

using namespace RON;

/*
 * A basic, trivial dispatching scheduler.  Provides the event
 * loop functionality for RON.
 */

Scheduler::Scheduler(Proxy* proxy)
{
    this->proxy = proxy;
    
    socket_callback_table.clear();
}

/*
 * Register a particular packet type and callback object
 */

int
Scheduler::register_socket(int sockno, socket_callable *callback, void *data)
{
    DPRINTF(DEBUG_DISPATCH,
	    "Scheduler::register_socket callback %p for socket %d\n",
	    callback, sockno);
    socket_callback_table[sockno] = callback;
    socket_list.push_back(sockno);
    return 0;
}

int
Scheduler::register_timer(u_int32_t msec, time_callable *callback, void *data)
{
    struct timeval t;
    time_callback_entry foo;

    gettimeofday(&t, NULL);
    t.tv_sec += (msec / 1000000);
    t.tv_usec += (msec % 1000000);
    while (t.tv_usec > 1000000) {
	t.tv_sec++;
	t.tv_usec -= 1000000;
    }
    foo.set_time(t);
    foo.callback = callback;
    foo.data = data;
    timer_callback_table.push(foo);
#if 0
    DPRINTF(DEBUG_DISPATCH,
	    "Scheduler::register_timer %d msec cb %p (%d ents)\n",
	    msec, callback, timer_callback_table.size());
#endif /* too noisy! */
    return 0;
}

int
Scheduler::main()
{
    fd_set readfds; /* worry about write blocking later */
    int maxfd;
    int rc;

    list<int>::const_iterator p;
    
    FD_ZERO(&readfds);
    maxfd = 0;

    for (p = socket_list.begin(); p != socket_list.end(); p++) {
	FD_SET(*p, &readfds);
	if (*p > maxfd) maxfd = *p;
    }

    maxfd++;

    struct timeval curtime, max_sleeptime;
    fixedpoint now, sleeptime;
    
    gettimeofday(&curtime, NULL);
    now.set(curtime.tv_sec, curtime.tv_usec);
    /* Send off timer events */
    while (timer_callback_table.size() > 0 &&
	   (timer_callback_table.top().exptime <= now)) {
	time_callback_entry ent = timer_callback_table.top();
	timer_callback_table.pop();
	ent.callback->timer_callback(ent.data);
    }
    /* Figure out the max sleep time */
    if (timer_callback_table.size() <= 0) {
	sleeptime.set(10000, 0); /* Take a long nap.  Doesn't matter. :) */
    } else {
	sleeptime = timer_callback_table.top().exptime - now;
    }

    max_sleeptime.tv_sec = sleeptime.ones;
    max_sleeptime.tv_usec = sleeptime.micros;
    
    rc = select(maxfd, &readfds, NULL, NULL, &max_sleeptime);
    if (rc < 0) {
	/* XXX:  Should retry on interrupted. */
	if (errno != EINTR) {
	    DEBUG_PERROR("scheduler::dispatch select failed");
	}
	return -1;
    }
    /* rc == 0 simply means that our time expired */
    if (rc == 0) {
	return 0;
    }

    /* Dispatch to any local sockets */
    for (p = socket_list.begin(); p != socket_list.end(); p++) {
	if (FD_ISSET(*p, &readfds)) {
	    socket_callback_table[*p]->socket_callback(*p, NULL);
	}
    }	
    return 0;

}
