#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <stdarg.h>
#include <ctype.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/time.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include "rondebug.h"
#include "forw.h"
#include "static_router.h"
#include "latency_router.h"
#include "ron.h"
#include "ronlib.h"
#include "static_membership.h"
#include "broadcast_membership.h"

#include "ip_conduit.h"
#include "print_protocol.h"
#include "flood_protocol.h"
#include "ping_protocol.h"
#include "../pdb/pdb_client.h" /* XXX */
#include "../pdb/pdb_local_client.h"
#include "scheduler.h"
#include "mesh_protocol.h"
#include "mesh_tester.h"
#include "passive_monitor.h" /* Experimental.. but so is everything. :) */

using namespace RON;

void usage()
{
    fprintf(stderr,
	    "[ip|test]-forw  [-Mvhe] [-d <debug>] [-p <port>] [-m <myip>]\n"
	    );
}

void help()
{
    usage();
    fprintf(stderr,
	    "       -h ......... help (this message)\n"
	    "       -d <debug> . see -d list for a list of debug flags\n"
	    "       -m <myip> .. set the local IP addr if we guess wrong\n"
	    "       -p <port> .. use <port> as the RON communication port\n"
	    "       -v ......... output routes to the viz socket\n"
	    "       -e ......... use mEsh protocol with ip conduit\n"
	    "       -a ......... active conduit (test-forw only)\n"
	    );
}

forw *forw_ptr = NULL;


void sigcaught(int sig)
{
    switch(sig) {
    case SIGUSR1:
	forw_ptr->stats.dump();
	return;
    case SIGUSR2:
	forw_ptr->stats.clear();
	return;
    case SIGINT:
	forw_ptr->stats.dump();
	exit(0);
    default:
	exit(0);
    }
}

int main(int argc, char **argv)
{
    extern char *optarg;
    extern int optind;
    int ch;
    membership *memb;

    int forw_port = DEFAULT_RON_FORWARDER_PORT;
    char myhostname[MAXHOSTNAMELEN];
    int hostnameset = 0;
    int use_mesh_protocol = 0;
    int test_mesh = 0;

    Scheduler *sched;   // DILIP: this will be replaced by the Proxy addTimer and addFDCallback

    forw *myforw;
    unsigned long myaddr;
    struct in_addr myinaddr;
    int do_visual = 0;
    int active_conduit = 0;

    while ((ch = getopt(argc, argv, "aMm:hp:d:ve")) != -1)
	switch (ch) {
	case 'a':
	    active_conduit = 1;
	    break;
	case 'e':
	    use_mesh_protocol = 1;
	    break;
	case 'M':
	    test_mesh = 1;
	    break;
	case 'v':
	    do_visual = 1;
	    break;
	case 'm':
	    if (nametoip(optarg, &myinaddr)) {
		perror(optarg);
		exit(-1);
	    }
	    hostnameset = 1;
	    break;
	case 'p':
	    forw_port = atoi(optarg);
	    break;
	case 'h':
	    help();
	    exit(0);
	case 'd':
	    printf("Setting debug: %s\n", optarg);
	    if (set_debug(optarg)) {
		exit(-1);
	    }
	    break;
	default:
	    usage();
	    exit(-1);
	}
    argc -= optind;
    argv += optind;

    srandom(time(NULL)+getpid());
    signal(SIGINT, sigcaught);
    signal(SIGTERM, sigcaught);
    signal(SIGUSR1, sigcaught);
    signal(SIGUSR2, sigcaught);
    struct sigaction ignme;
    ignme.sa_handler = SIG_IGN;
    sigaction(SIGPIPE, &ignme, NULL);

    DPRINTF(DEBUG_INIT, "test-forw::main starting up\n");

    if (!hostnameset) {
	DPRINTF(DEBUG_INIT, "test-forw::main getting hostname\n");
	if (gethostname(myhostname, sizeof(myhostname)-1)) {
	    perror("gethostname");
	    exit(-1);
	}
	DPRINTF(DEBUG_INIT, "test-forw::main got hostname: %s\n",
		myhostname);
	if (nametoip(myhostname, &myinaddr)) {
	    perror(myhostname);
	    exit(-1);
	}
    }

    sched = new Scheduler();
    if (!sched) {
	perror("could not allocate Scheduler.  Bailing!\n");
	exit(-1);
    }
    
    myaddr = myinaddr.s_addr;
    DPRINTF(DEBUG_INIT, "test-forw::main got address %s\n",
	    inet_ntoa(myinaddr));

    flood_protocol *fp;
    fp = new flood_protocol();

    memb = new broadcast_membership("peerinit.txt", fp); //I think this defines the neighbours of this RON node.

    if (!memb) {
	perror("could not create membership class");
	exit(-1);
    }

    pdb_client *pdbc = new pdb_local_client();
    pdbc->init();

#ifdef IP_FORW
    ip_conduit *conduit = new ip_conduit();
#else
    print_protocol *conduit = new print_protocol();
#endif
    if (!conduit) {
	perror("could not create print protocol class");
	exit(-1);
    }
#if 1 /* Use the latency-based router */
    latency_router *router;
    router = new latency_router(myaddr, forw_port, pdbc);
    if (!router) {
	perror("test-forw::main could not create latency router");
	exit(-1);
    }
    router->set_visual(do_visual);
#else
    static_router *router;
    router = new static_router();
#endif

    myforw = new forw(sched, myinaddr, forw_port, router, memb);
    DPRINTF(DEBUG_INIT, "test-forw::membership %p forwarder %p\n",
	    memb, myforw);
    if (!myforw) {
	perror("could not initialize forwarder");
	exit(-1);
    }
    myforw->init(forw_port);
    forw_ptr = myforw;
    myforw->stats.counter_table["start"] = time(NULL);

    mesh_protocol *mp;
    mp = new mesh_protocol;
#if 0
    passive_monitor *monitor = new passive_monitor();
    if (!monitor) {
	perror("could not initialize passive monitor");
	exit(-1);
    }
#endif
    /* Set up the outbound routing between components.  In this case,
     * it's all trivial - they simply all send their packets
     * back to the forwarder */
    fp->register_default(myforw);
    router->register_default(myforw);
#if 0 /* XXX:  not ready for prime-time, but want to commit print
       * protocol changes... */
    conduit->register_default(monitor);  /* All IP goes through monitor */
    monitor->register_default(myforw);
#else
    conduit->register_default(myforw);
#endif
    if (use_mesh_protocol) {
	conduit->register_type(RON_MESH_DATA, mp, NULL);
    }
    memb->register_default(myforw);
    fp->register_default(myforw);
    mp->register_default(myforw);

    /* Let them all register their callbacks and such.  After
     * 'configure' is called, the system should be ready to run. */
    fp->configure(sched, myforw);
    router->configure(sched, myforw);
    conduit->configure(sched, myforw);
#if 0
    monitor->configure(sched, myforw);
#endif
    memb->configure(sched, myforw);
    mp->configure(sched, myforw);
    pdbc->configure(sched, myforw);

#if 1
    ping_protocol *ping = new ping_protocol(pdbc);
    ping->configure(sched, myforw);
    ping->register_default(myforw);
#endif
#ifndef IP_FORW
    if (active_conduit)
	conduit->set_active(1);
    mesh_tester *mtest = NULL;
    if (test_mesh) {
	mtest = new mesh_tester();
	mtest->configure(sched, myforw);
	mtest->register_default(myforw);
    }
#endif

    while (1) {
	sched->main();
    }
    
    exit(0);
}
