#ifndef _FIXEDPOINT_H_
#define _FIXEDPOINT_H_

#ifdef __cplusplus
using namespace std;
#endif

#include <sys/types.h>
#include <math.h>
#include <stdio.h>
#include <unistd.h>
#include <netinet/in.h>

/* The C and wire representation of a fixedpoint value */
struct fixedpoint_wire {
    int32_t ones;
    u_int32_t micros;
};

/* The rest of this is cplusplus specific... */

#ifdef __cplusplus
#include <string>
#include "ronlib.h"

/*
 * A simple fixed point class - 32 bits of 1s, 32 bits of micros
 * Used in place of floating point values in RON, because it's
 * simple and architecture independent
 */

class fixedpoint {
public:
    int32_t   ones;
    int32_t micros;

    void set(const fixedpoint &s) {
	set(s.ones, s.micros);
    }
    
    void set(int32_t n_ones, u_int32_t n_micros) {
	ones = n_ones;
	micros = n_micros;
	normalize();
    }

    fixedpoint() {  set(0, 0); }

    fixedpoint(int32_t n_ones, u_int32_t n_micros) {
	set(n_ones, n_micros);
    }

    fixedpoint(const fixedpoint &s) {
	set(s.ones, s.micros);
    }
    fixedpoint(const fixedpoint_wire &s) {
	set(ntohl(s.ones), ntohl(s.micros));
    }

    fixedpoint(const int s) {
	set(s, 0);
    }

    fixedpoint(const double s) {
	set(s);
    }

    void set(const fixedpoint_wire &s) {
	set(ntohl(s.ones), ntohl(s.micros));
    }

    void set(const int s) {
	set(s, 0);
    }

    void set(const double s) {
	ones = (int)floor(s);
	micros = (int)((s - floor(s))*1000000);
    }

    fixedpoint operator*(const float s) const {
	fixedpoint result = *this;
	result.ones = (int)(result.ones * s);
	result.micros = (int)(result.micros * s);
	result.micros += (int)(floorf(((result.ones * s - floorf(result.ones*s)) * 1000000)));
	result.normalize();
	return result;
    }
    double to_double() const {
	return ((double) ones + (double) micros / (double) 1000000);
    }

    /* THIS IS SO CHEATING */
    fixedpoint operator*(const fixedpoint &s) const {
	fixedpoint result;
	result.set(to_double() * s.to_double());
	result.normalize();
	return result;
    }

        /* THIS IS SO CHEATING */
    fixedpoint operator/(const fixedpoint &s) const {
	fixedpoint result;
	result.set(to_double() / s.to_double());
	result.normalize();
	return result;
    }

    fixedpoint operator+(const fixedpoint &s) const {
	fixedpoint result = *this;
	if ((result.ones >= 0 && s.ones >= 0) ||
	    (result.ones < 0 && s.ones < 0)) {
	    result.ones += s.ones;
	    result.micros += s.micros;
	} else if (result.ones >= 0 && s.ones < 0) {
	    result = result - (-s);
	} else {
	    result = s - (-result);
	}
	result.normalize();
	return result;
    }

    // (SAR)
    fixedpoint operator+=(const fixedpoint &s) {

	ones += s.ones;
	micros += micros;
	normalize();

	fixedpoint result = *this;
	return result;
    }

    fixedpoint operator-() const {
	fixedpoint result = *this;
	result.ones = -result.ones;
	return result;
    }

    fixedpoint operator-(const fixedpoint &s) const {
	fixedpoint result = *this;
	while (s.micros > result.micros) {
	    result.micros += 1000000;
	    result.ones--;
	}
	result.ones -= s.ones;
	result.micros -= s.micros;
	result.normalize();
	return result;
    }

    int operator==(const fixedpoint &s) const {
	if (micros != s.micros ||
	    ones != s.ones) {
	    return 0;
	}
	return 1;
    }

    /* Need this one too, using it extensively in PDB (SAR) */
    int operator!=(const fixedpoint &s) const {
	if (micros != s.micros ||
	    ones != s.ones) {
	    return 1;
	}
	return 0;
    }


    /* Forgot about these too, trusting the compiler too much here (SAR) */
    int operator>(const fixedpoint &s) const {
	if (ones != s.ones) {
	    return ones > s.ones;
	}
	return micros > s.micros;
    }


    int operator<(const fixedpoint &s) const {
	if (ones != s.ones) {
	    return ones < s.ones;
	}
	return micros < s.micros;
    }

    int operator>=(const fixedpoint &s) const {
	if (ones != s.ones) {
	    return ones >= s.ones;
	}
	return micros >= s.micros;
    }


    int operator<=(const fixedpoint &s) const {
	if (ones != s.ones) {
	    return ones <= s.ones;
	}
	return micros <= s.micros;
    }

    string to_string() const {
	char tmp[64];
	bzero(tmp, sizeof(tmp));
	sprintf(tmp, "%d.%6.6d", ones, micros);
	string s1(tmp);
	return s1;
    }

    struct fixedpoint_wire to_wire() {
	struct fixedpoint_wire a;
	a.ones = htonl(ones);
	a.micros = htonl(micros);
	return a;
    }

private:

    void normalize() {
	while (micros >= 1000000) {
	    if (ones >= 0) {
		ones++;
	    } else {
		ones--;
	    }
	    micros -= 1000000;
	}
    }
    
};

#endif /* __cplusplus */
#endif /* _FIXEDPOINT_H_ */
