#include <stdlib.h>
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "pdb.h"
#include "rondebug.h"

#include <db_cxx.h>

long timediff(const struct timeval *time_now,
              const struct timeval *time_prev)
{
        long usecs_now;
        
        if (time_prev->tv_sec == 0 &&
            time_prev->tv_usec == 0) {
                return 0;
        }
        
        usecs_now = (time_now->tv_sec - time_prev->tv_sec) * 1000000;
        usecs_now += time_now->tv_usec;

        return usecs_now - time_prev->tv_usec;
}

/* If you want the databases to be persistent, give them
 * a non-null filename here. Otherwise, they'll be stored in memory. */

#define PERF_DATABASE NULL /* "perf.db" */
#define TYPE_DATABASE NULL  /* "type.db" */


/*
 * We have the DB sort our samples by timestamp for us
 */

#if DB_VERSION_MAJOR == 3 && DB_VERSION_MINOR < 2
#  define COMPARE_ARGS const DBT *dbt1, const DBT *dbt2
#else
#  define COMPARE_ARGS DB *da_db, const DBT *dbt1, const DBT *dbt2
#endif

extern "C" {
    int pdb_sample_compare(COMPARE_ARGS);
}

/*static*/ int
pdb_sample_compare(COMPARE_ARGS)
{
    struct sampletype *sample1, *sample2;
    
    sample1 = (struct sampletype *)dbt1->data;
    sample2 = (struct sampletype *)dbt2->data;

    /* Sanity checks */
    if (dbt1->size != sizeof(struct sampletype) ||
	dbt2->size != sizeof(struct sampletype)) {
	fprintf(stderr, "pdb_sample_compare:  invalid data sizes %d, %d\n",
		dbt1->size, dbt2->size);
	fprintf(stderr, "Magic:  0x%x  0x%x\n",
		sample1->header.magic,
		sample2->header.magic);
	return 0;
    }

    return timediff((struct timeval *)&sample1->header.timestamp,
		    (struct timeval *)&sample2->header.timestamp);
}

int
pdb::db_init()
{
    int ret;

    perf_db = new Db(NULL, 0);
    if (!perf_db) {
	fprintf(stderr, "Could not create perf_db\n");
	goto err;
    }

    if ((ret = perf_db->set_flags(DB_DUPSORT)) != 0) {
	perf_db->err(ret, "Setting DB_DUPSORT");
	goto err;
    }

    if ((ret = perf_db->set_dup_compare(&pdb_sample_compare)) != 0) {
	perf_db->err(ret, "Setting BT compare");
	goto err;
    }

    if ((ret = perf_db->open(PERF_DATABASE, NULL,
			     DB_BTREE, DB_CREATE, 0664)) != 0) {
	perf_db->err(ret, "Opening perf_database %s",
		     (PERF_DATABASE ? PERF_DATABASE : "(in memory)"));
    }

    type_db = new Db(NULL, 0);
    if (!type_db) {
	fprintf(stderr, "Could not create type_db\n");
	goto err;
    }

    if ((ret = type_db->open(TYPE_DATABASE, NULL,
			     DB_BTREE, DB_CREATE, 0664)) != 0) {
	perf_db->err(ret, "Opening type_database %s",
		     (TYPE_DATABASE ? TYPE_DATABASE : "(in memory)"));
    }

    return 0;

    /* Unrecoverable error in the database loop */
    err:
    pdb_cleanup();
    return -1;
}

int
pdb::add_class(const char *classname, struct class_defn *classdef)
{
    Dbt key, data;
    struct class_defn *searchres;
    int ret;

    if (classname == NULL) {
	return EINVAL;
    }
    
    if ((searchres = find_class(classname)) != NULL) {
	delete searchres;
	DPRINTF(DEBUG_ERRS, "pdb::add_class:  class %s already exists\n",
		classname);
	return EEXIST;
    }
    
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));

    key.set_data((void *)classname);
    key.set_size(strlen(classname));
    data.set_data((void *)classdef);
    data.set_size(sizeof(struct class_defn));

    ret = type_db->put(NULL, &key, &data, 0);
    if (ret != 0) {
	type_db->err(ret, "Error inserting type %s into type_db", classname);
	return -1;
    }
    return 0;
}

class_defn *
pdb::find_class(const char *classname)
{
    Dbt key, data;
    struct class_defn *searchres;
    int ret;

    if (classname == NULL) {
	return NULL;
    }
    
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));

    key.set_data((void *)classname);
    key.set_size(strlen(classname));

    if ((ret = type_db->get(NULL, &key, &data, 0)) != 0) {
	/*type_db->err(ret, "Error searching for type %s", classname);*/
	return NULL;
    }
    if (data.get_size() != sizeof(struct class_defn)) {
	fprintf(stderr, "Size mismatch between returned data %d and cdef %d\n",
		data.get_size(), sizeof(struct class_defn));
	return NULL;
    }

    searchres = new struct class_defn;
    memcpy(searchres, data.get_data(), sizeof(struct class_defn));
    return searchres;
}

/*
 * Delete the class definition
 */

int
pdb::delete_class(const char *classname)
{
    return 0;
}

int
pdb::add_sample(char *classname, struct sampletype *sample)
{
    int retval = 0;
    Dbt key, data;
    int ret;
    struct class_defn *classdef = NULL;
    Dbc *cursor = NULL;
    db_recno_t count;
    
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));

    if ((classdef = find_class(classname)) == NULL) {
	fprintf(stderr, "add_sample:  no such class %s\n", classname);
	retval = EINVAL;
	goto err;
    }

    if (sample->header.type != classdef->type_type) {
	fprintf(stderr, "add_sample:  invalid type for class %s\n",
		classname);
	retval = EINVAL;
	goto err;
    }

    sample->header.magic = 0xDEADBEEF;

    key.set_data((void *)classname);
    key.set_size(strlen(classname));
    data.set_size(sample->header.samplesize);
    data.set_data(sample);

    if ((ret = perf_db->put(NULL, &key, &data, 0)) != 0) {
	perf_db->err(ret, "DB->put to class %s", classname);
	retval = EINVAL;
	goto err;
    }

    /* XXX:  Do some cleanup of the class here to ensure that
     * we retain the desired number of samples in storage */

    if ((ret = perf_db->cursor(NULL, &cursor, 0)) != 0) {
	perf_db->err(ret, "perf_db->cursor");
	goto done;
    }
    if ((ret = cursor->get(&key, &data, DB_SET)) != 0) {
	perf_db->err(ret, "perf_db->cursor get");
	goto done;
    }
    if ((ret = cursor->count(&count, 0)) != 0) {
	perf_db->err(ret, "perf_db->cursor count");
	goto done;
    }
    if (PDB_MAX_ENTRIES && count > PDB_MAX_ENTRIES) {
	int ndel = (count - PDB_MAX_ENTRIES);
	while (ndel-- > 0) {
	    if ((ret = cursor->del(0)) != 0) {
		perf_db->err(ret, "perf_db->cursor del");
		goto done;
	    }
	    if ((ret = cursor->get(&key, &data, DB_NEXT_DUP)) != 0) {
		perf_db->err(ret, "perf_db->cursor getnext");
		goto done;
	    }
	}
    }

 done:
    
    if (cursor) cursor->close();
    delete classdef;
    return 0;

 err:
    if (classdef) delete classdef;
    return retval;
}

Dbc *
pdb::get_samples_res(char *classname)
{
    Dbt key, data;
    Dbc *cursor = NULL;
    struct class_defn *classdef = NULL;
    int ret;

    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));

    if ((classdef = find_class(classname)) == NULL) {
	fprintf(stderr, "get_samples:  no such class %s\n", classname);
	goto err;
    }

    key.set_data((void *)classname);
    key.set_size(strlen(classname));
    
    if ((ret = perf_db->cursor(NULL, &cursor, 0)) != 0) {
	perf_db->err(ret, "perf_db->cursor");
	goto err;
    }

    if ((ret = cursor->get(&key, &data, DB_SET)) != 0) {
	perf_db->err(ret, "cursor->get");
	goto err;
    }
    
    free(classdef);
    return cursor;
    
 err:
    if (classdef) free(classdef);
    return NULL;
}

/*
 * Write out the database contents in an unstructured
 * manner to the supplied output filehandle.
 * Useful for debugging and external monitoring
 */

void
pdb::pdb_dump(char *opts, FILE *outfile)
{
    Dbt key, data, pkey, pdata;
    int ret, iret;
    Dbc *cursor = NULL, *pcursor = NULL;
    char classname[64]; /* XXX */
    struct class_defn *classdef;
    struct sampletype *sample;
    struct stringsampletype *stringsample;

    int classlist = 0;
    int oneclass = 0;

    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));

    /* First get a list of the classes we know about */
#define min(a,b) (a < b ? a : b)
    
    if ((ret = type_db->cursor(NULL, &cursor, 0)) != 0) {
	type_db->err(ret, "pdb::pdb_dump type_db->cursor");
	goto error;
    }

    if (opts && opts[0]) {
	if (!strcmp(opts, "classlist")) {
	    classlist = 1;
	} else {
	    key.set_data((void *)opts);
	    key.set_size(strlen(opts));
	    oneclass = 1;
	}
    }
    
    ret = cursor->get(&key, &data, oneclass ? DB_SET : DB_FIRST);
    while (ret == 0) {
	time_t tm;
	
	classdef = (struct class_defn *)data.get_data();
	bzero(classname, sizeof(classname));
	strncpy(classname, (char *)key.get_data(),
		min(63, key.get_size()));
	printf("%s\n", classname);
	if (!classlist) {
	    memset(&pkey, 0, sizeof(pkey));
	    memset(&pdata, 0, sizeof(pdata));
	    if ((iret = perf_db->cursor(NULL, &pcursor, 0)) != 0) {
		goto error;
	    }
	    iret = pcursor->get(&key, &pdata, DB_SET);
	    while (iret == 0) {
		sample = (struct sampletype *)pdata.get_data();
		stringsample = (struct stringsampletype *)sample;
		tm = sample->header.timestamp.sec;
		char *timestr = ctime(&tm);
		timestr[strlen(timestr)-1] = '\0';
		
		printf("   %s", timestr);

		fixedpoint fixedtmp;
		
		switch(classdef->type_type) {
		case TYPE_FIXED:
		    fixedtmp.set(sample->val.fixedval);
		    printf("   fixed %s\n", fixedtmp.to_string().data());
		    break;
		case TYPE_STRING:
		    printf("   string %s\n", stringsample->val.contents);
		    break;
		default:
		    printf("   Unknown type\n");
		}
		iret = pcursor->get(&key, &pdata, DB_NEXT_DUP);
	    }
	}

	if (pcursor) pcursor->close();
	pcursor = NULL;
	/* And get the next class definition key */
	if (oneclass) {
	    ret = -1;
	} else {
	    ret = cursor->get(&key, &data, DB_NEXT);
	}
    }

 error:
    if (cursor) cursor->close();
    if (pcursor) pcursor->close();
}
