#ifndef _PDB_H_
#define _PDB_H_

#include <assert.h>
#include <iostream.h>
#include <db_cxx.h>
#include <sys/time.h>

#include "fixedpoint.h"

#include "pdb_wire.h" /* Coordinate TYPEs */

#include "pdb_res.h"

/*
 * We should have templated functions that can take datatypes
 * and perform various operations on them, like EWMAs
 */

/*
 * The ultimate goal of the PDB:
 *  - Insert entries of a particular class into the DB
 *  - Respond to queries like "give me the EWMA of the last 20 items
 *    of this class"
 *
 * The above can be viewed more completely as:
 * - Insert entries and tag them as belonging to a particular classs
 * - Retrieve a list of items by a class tag
 * - Apply one of many type-specific operators to that data
 *   to create the answer for the user.
 */

/*
 * To make it easier on us, let's do the intermediate result
 * computation as a continuation style.
 *
 * function_continuation foo;
 * foo.reset();
 * int_ewma_blah(value, foo)
 * int_ewma_blah();
 *
 * Or maybe it should be a class that we instantiate to do the processing
 * for us?  How do we make it generic?  Or do we care?  How do we make
 * it ameniable to future optimization?  I think we make each one a class
 * and let it have an optional static optimization function or something
 * like that.
 * Hm...
 */

/*
 * Structs used by the pdb.
 * Start with fixed, and turn it into a template class later.
 */

#define PDB_MAX_ENTRIES 200  /* Max entries per class */

class pdb {
public:
    /* Define a new class for the database.  Returns 0 on success,
     * an error code on failure.
     */
    int add_class(const char *classname, struct class_defn *classdef);
    /* Returns a pointer to the class_defn if the class exists,
     * NULL otherwise.  This allocates memory, and the class
     * defn returned MUST be freed by the caller */
    class_defn *find_class(const char *classname);

    /*
     * Remove both the class definition AND any associated data
     * from the database.
     *
     * XXX:  Not yet implemented!
     */
    int delete_class(const char *classname);

    /*
     * Add a sample of a particular type to the database
     * Returns 0 on success.
     */
    int add_sample(char *classname, struct sampletype *sample);

    pdb_res<float> *get_samples(char *classname);
    Dbc *get_samples_res(char *classname);

    pdb() {
	perf_db = NULL;
	type_db = NULL;
    }
    ~pdb() {
	pdb_cleanup();
    }
    void pdb_cleanup() {
	if (perf_db != NULL) {
	    perf_db->close(0);
	    delete perf_db;
	    perf_db = NULL;
	}
	if (type_db != NULL) {
	    type_db->close(0);
	    delete type_db;
	    type_db = NULL;
	}
    }
    int db_init();
    void pdb_dump(char *opts, FILE *outfile);
    /*private:*/
    Db *perf_db;
    Db *type_db;
private:
    /* db3 comparison function for btree sorting */
    //static int pdb_sample_compare(const DBT *dbt1, const DBT *dbt2);
};

/*
 * XXX:  From here below is simply playing with different ways to
 * come up with a generic internal representation of the data
 * that we stream out of the database so we can do EWMAish
 * things to it.  Basically, ignore below this point, because
 * Dave is clueless thus far.
 */

class pdb_listitem {
public:
    pdb_listitem() { next = prev = NULL; }
    class pdb_listitem *next, *prev;
};

const int defaultSize = 100;

template <class Type>
class pdb_list {
public:
    pdb_list() {
	cur_index = 0;
	max_index = -1;
	ia = new Type[defaultSize];
	assert( ia != NULL );
    }
    /*~pdb_list() { delete ia; }*/
    Type cur() { return ia[cur_index]; }
    Type next() {
	if (cur_index < max_index) {
	    cur_index++;
	    return ia[cur_index];
	} else {
	    return ia[cur_index];
	}
    }
    int eof() {
	if (cur_index < max_index) { return 0; }
	else return 1;
    }
    void reset() { cur_index = 0; }

    int append(Type a) {
	if (max_index + 1 < defaultSize) {
	    max_index++;
	    ia[max_index] = a;
	    cout << "Item " << max_index << " is now " << a << endl;
	    return 0;
	} else {
	    return -1;
	}
    }
    
private:
    int cur_index;
    int max_index;
    Type *ia;
};

#endif /* _PDB_H_ */
