#include "pdb_client.h"

using namespace RON;

void
pdb_client::init()
{
    /* Init the address */
    
    bzero(&pdbaddr, sizeof(pdbaddr));
#ifdef RON_HAS_STRUCT_SINLEN
    pdbaddr.sin_len = sizeof(pdbaddr);
#endif
    pdbaddr.sin_family = AF_INET;
    pdbaddr.sin_port = htons(9431);
    if (!pdb_host) {
	pdbaddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
    } else {
	if (nametoip(pdb_host, &pdbaddr.sin_addr)) {
	    fprintf(stderr, "pdb_client::init could not resolve ");
	    perror(pdb_host);
	    return;
	}
    }

    /* Grab a socket */

    pdb_sock = bound_udp_socket(0);
    if (pdb_sock == -1) {
	perror("could not create socket");
	exit(-1); /* XXX - do some real error handling here */
    }
}

pdb_client::pdb_client() {
    pdb_host = NULL;
}

pdb_client::pdb_client(char *hostname) {
    pdb_host = hostname;
}

int
pdb_client::query_fixed(char *classname, int summarizer, fixedpoint param,
			pdb_callable *cb)
{
    fixedpoint ret;
    ret.set(0,0);

    pdb_req_record rr;

    rr.queryid = random();
    rr.cb = cb;

    char indata[1500]; /* XXX Blah */
    struct pdb_header *pdbh;
    struct pdb_update *pdbu;
    int nreqs = 1;
    int pktsize, rc;
    
    pdbh = (struct pdb_header *)indata;
    pdbu = (struct pdb_update *)(indata + sizeof(struct pdb_header));

    bzero(indata, sizeof(indata));
    pktsize = sizeof(struct pdb_header) + sizeof(struct pdb_update)*nreqs;

    pdbh->nsamples = htonl(nreqs);

    strncpy((char *)pdbu->classname, classname, sizeof(pdbu->classname)-1);
    pdbu->classname[sizeof(pdbu->classname)-1] = 0; /* Ensure strncpy OK */
    pdbu->classtype = htonl(TYPE_FIXED);
    pdbu->queryid = htonl(rr.queryid);

    struct timeval timenow;
    gettimeofday(&timenow, NULL);
    rr.time_sent.set(timenow.tv_sec, timenow.tv_usec);
	
    pdbu->timestamp.sec = htonl(timenow.tv_sec);
    pdbu->timestamp.usec = htonl(timenow.tv_usec);

    pdbu->querytype = PDB_QUERY;
    pdbu->method.summarize = summarizer;
    pdbu->param = param.to_wire();
    
    rc = sendto(pdb_sock, indata, pktsize, 0,
		(struct sockaddr *)&pdbaddr, sizeof(pdbaddr));
    if (rc <= 0) {
	perror("pdb_client::query_fixed sendto error");
    } else {
	record.push_back(rr);
    }
    
    return rc;
}

int
pdb_client::multiquery_fixed(int numreqs, pdb_onereq *reqlist)
{
    int i;
    char indata[1500]; /* XXX Blah */
    struct pdb_header *pdbh;
    struct pdb_update *pdbu;
    int pktsize, rc;
    pdb_req_record rr;
    
    pdbh = (struct pdb_header *)indata;
    pdbu = (struct pdb_update *)(indata + sizeof(struct pdb_header));

    bzero(indata, sizeof(indata));
    pktsize = sizeof(struct pdb_header) + sizeof(struct pdb_update)*numreqs;
    pdbh->nsamples = htonl(numreqs);

    struct timeval timenow;
    gettimeofday(&timenow, NULL);
    rr.time_sent.set(timenow.tv_sec, timenow.tv_usec);

    for (i = 0; i < numreqs; i++) {
	fixedpoint ret;
	ret.set(0,0);

	rr.queryid = random();
	rr.cb = reqlist[i].cb;

	strncpy((char *)pdbu->classname, reqlist[i].classname,
		sizeof(pdbu->classname)-1);
	pdbu->classname[sizeof(pdbu->classname)-1] = 0; /* Ensure strncpy OK */
	pdbu->classtype = htonl(TYPE_FIXED);
	pdbu->queryid = htonl(rr.queryid);
	
	pdbu->timestamp.sec = htonl(timenow.tv_sec);
	pdbu->timestamp.usec = htonl(timenow.tv_usec);

	pdbu->querytype = PDB_QUERY;
	pdbu->method.summarize = reqlist[i].summarizer;
	pdbu->param = (reqlist[i].param).to_wire();
	/* If we don't actually successfully transmit it, we'll just
	 * clean it up later.  It delays notification a bit, but won't
	 * kill us */
	record.push_back(rr);
	pdbu++;
    }
    rc = sendto(pdb_sock, indata, pktsize, 0,
		(struct sockaddr *)&pdbaddr, sizeof(pdbaddr));
    if (rc <= 0) {
	perror("pdb_client::multiquery_fixed sendto error");
    }
    
    return rc;
}


/*
 * Cleanup outstanding queries and report them as failures
 * XXX:  This is nearly identical to the code in ping_protocol
 *       for timing out queries.  Can we reuse?
 */

void
pdb_client::clean_queries()
{
    list<pdb_req_record>::iterator iter;
    pdb_req_record pr;
    int nuked = 0;

    struct timeval now;
    fixedpoint nowpoint;
    gettimeofday(&now, NULL);
    nowpoint.set(now.tv_sec, now.tv_usec);

    do {
	nuked = 0;
	for (iter = record.begin(); iter != record.end(); iter++) {
	    if (nowpoint - (*iter).time_sent > fixedpoint(6,0)) {
		pr = *iter;
		nuked = 1;
		break;
	    }
	}
	if (nuked) {
	    DPRINTF(DEBUG_REQUESTS, "Never heard back on query 0x%x\n",
		    pr.queryid);
	    pr.cb->pdb_callback(0, NULL, 0, fixedpoint(0,0));
	    record.remove(pr);
	}
    } while (nuked != 0);
}

int
pdb_client::update_fixed(char *classname, fixedpoint value)
{
    char indata[1500]; /* XXX Blah */
    struct pdb_header *pdbh;
    struct pdb_update *pdbu;
    int nreqs = 1;
    int pktsize, rc;
    
    pdbh = (struct pdb_header *)indata;
    pdbu = (struct pdb_update *)(indata + sizeof(struct pdb_header));

    bzero(indata, sizeof(indata));
    pktsize = sizeof(struct pdb_header) + sizeof(struct pdb_update)*nreqs;

    pdbh->nsamples = htonl(nreqs);

    strncpy((char *)pdbu->classname, classname, sizeof(pdbu->classname)-1);
    pdbu->classname[sizeof(pdbu->classname)-1] = 0; /* Ensure strncpy OK */
    pdbu->classtype = htonl(TYPE_FIXED);

    struct timeval timenow;
    gettimeofday(&timenow, NULL);
    
    pdbu->timestamp.sec = htonl(timenow.tv_sec);
    pdbu->timestamp.usec = htonl(timenow.tv_usec);

    pdbu->querytype = PDB_INSERT;
    pdbu->method.overwrite = PDB_APPEND;
    pdbu->data.fixedval.ones = htonl(value.ones);
    pdbu->data.fixedval.micros = htonl(value.micros);

    rc = sendto(pdb_sock, indata, pktsize, 0,
		(struct sockaddr *)&pdbaddr, sizeof(pdbaddr));
    if (rc <= 0) {
	perror("pdb_client::update_fixed error");
    }
    return rc;
}

int
pdb_client::handle_reply(struct pdb_update *pdbu)
{
    /* Look up its callback entry */
    list<pdb_req_record>::iterator iter;
    pdb_req_record pr;
    int found = 0;
    u_int32_t id;

    static int cheap_hack_counter = 0;
    if (cheap_hack_counter > 50) { clean_queries(); cheap_hack_counter = 0; }
    cheap_hack_counter++;

    id = ntohl(pdbu->queryid);

    for (iter = record.begin(); iter != record.end(); iter++) {
	if (id == (*iter).queryid) {
	    pr = *iter;
	    found = 1;
	    break;
	}
    }
    if (!found) {
	DPRINTF(DEBUG_ERRS, "pdb_client::handle_reply unknown query 0x%x\n",
		id);
	return -1;
    }

    if (pdbu->querytype == PDB_NOANSWER) {
	pr.cb->pdb_callback(0, NULL, 0, fixedpoint(0,0));
    } else {
	fixedpoint fp;
	fp.set(pdbu->data.fixedval);
	pr.cb->pdb_callback(1, (char *)pdbu->classname,
			    pdbu->method.summarize, fp);
    }
    record.remove(pr);
    return 0;
}

int
pdb_client::socket_callback(int sockno, void *data)
{
    char indata[1500]; /* XXX Blah */
    struct pdb_header *pdbh;
    struct pdb_update *pdbu;
    struct sockaddr_in pdbaddr;
    int rc;
    unsigned int fromlen;
    
    rc = recvfrom(sockno, indata, sizeof(indata), 0,
		  (struct sockaddr *)&pdbaddr, &fromlen);
    if (rc < 0) {
	perror("pdb_client::read error");
	return -1;
    }
    if (rc == 0) {
	perror("pdb_client::socket_callback zero read");
	return -1;
    }

    fixedpoint fixedtmp;
    pdbh = (struct pdb_header *)indata;
    pdbu = (struct pdb_update *)(indata + sizeof(struct pdb_header));

    int nsamples = ntohl(pdbh->nsamples);
    DPRINTF(DEBUG_REQUESTS,
	    "pdb_client::socket_callback got a packet of %d bytes\n", rc);
    while (nsamples--) {
	switch (pdbu->querytype) {
	case PDB_INSERT:
	    DPRINTF(DEBUG_ERRS,
		    "pdb_client::socket_callback got an insert packet.  wtf?\n");
	    break;
	case PDB_QUERY:
	    DPRINTF(DEBUG_ERRS,
		    "pdb_client::socket_callback got a query packet. boggle\n");
	    break;
	case PDB_ANSWER:
	case PDB_NOANSWER:
	    DPRINTF(DEBUG_REQUESTS, "pdb_client::socket_callback got an answer packet.\n");
	    switch(ntohl(pdbu->classtype)) {
	    case TYPE_FIXED:
		handle_reply(pdbu);
		break;
		fixedtmp.set(pdbu->data.fixedval);
		printf("Data:  %s\n", fixedtmp.to_string().data());
		break;
	    default: /* Including string */
		printf("pdb_client::socket_callback got unknown type answer\n");
		break;
	    }
	    break;
	default:
	    fprintf(stderr, "pdb_client::socket_callback got unknown querytype %d\n",
		    pdbu->querytype);
	}
	pdbu++;
    }
    return 0;
}
