#include "rondebug.h"
#include "pdb_local_client.h"

using namespace RON;


/* RonDbClass:  A circular buffer that provides basic
   pdb-like functionality. */

fixedpoint RonDbClass::getMax(fixedpoint lastN)
{
    fixedpoint retVal(0.0);

    if (lastN.ones == 0)
	lastN.ones = winLen;

    if (lastN.ones > winLen) {
	/* Bad, should not query before inserting */
	if (winLen == 0) {
	    error = true;
	    return fixedpoint(0,0);
	}
	
	retVal = window[0];
	
	for (int i = 0; i < winLen; i++) {
	    if (window[i] > retVal)
		retVal = window[i];
	}
	
    } else {
	retVal = window[fromLast(lastN.ones - 1)];
	
	for (int i = lastN.ones - 1; i >=0; i--) {
	    int pos = fromLast(i);
	    
	    if (window[pos] > retVal)
		retVal = window[pos];
	}
    }
    
    return retVal;
}

fixedpoint RonDbClass::getMin(fixedpoint lastN)
{
    fixedpoint retVal;

    if (lastN.ones == 0)
	lastN.ones = winLen;
    
    if (lastN.ones > winLen) {
	/* Bad, should not query before inserting */
	if (winLen == 0) {
	    error = true;
	    return fixedpoint(0,0);
	}
	
	retVal = window[0];
	
	for (int i = 0; i < winLen; i++) {
	    if (window[i] < retVal)
		retVal = window[i];
	}
	
    } else {
	retVal = window[fromLast(lastN.ones - 1)];
	
	for (int i = lastN.ones - 1; i >=0; i--) {
	    int pos = fromLast(i);
	    
	    if (window[pos] < retVal)
		retVal = window[pos];
	}
    }
    
    return retVal;
}

fixedpoint RonDbClass::getRecent()
{
    /* Bad, should not query before inserting */
    if (winLen == 0) {
	error = true;
	return fixedpoint(0,0);
    } else {
	return window[fromLast(0)];
    }
}

fixedpoint RonDbClass::getFirst()
{
    if (winLen == 0) {
	error = true;
	return fixedpoint(0,0);
    } else {
	return window[winPos];
    }
}

fixedpoint RonDbClass::getEwma(fixedpoint alpha)
{
    
    if (alpha == 0.0)
	alpha.set(0.8);
    
    fixedpoint one(1), 
	       oneminus = one - alpha,
	       estimate;
    
    if (winLen < MAX_HISTORY) {
	/* Bad, should not query before inserting */
	if (winLen == 0) {
	    error = true;
	    return fixedpoint(0,0);
	}
	
	estimate = window[0];
	
	for (int i = 1; i<winLen; i++) {
	    estimate = alpha * estimate + oneminus * window[i];
	}
    } else {
	estimate = window[fromLast(MAX_HISTORY - 1)];
	for (int i = MAX_HISTORY - 2; i >= 0; i--) {
	    estimate = alpha * estimate + oneminus * window[fromLast(i)];
	}
    }
    
    return estimate;
}

fixedpoint RonDbClass::getAvg(fixedpoint param)
{
    fixedpoint retVal(0), count(0), one(1);
    
    if (param.ones == 0) {
	param.ones = winLen;
    }
    
    if (param.ones > winLen) {
	/* Bad, should not query before inserting */
	if (winLen == 0) {
	    error = true;
	    return fixedpoint(0,0);
	}
	
	for (int i = 0; i < winLen; i++) {
	    retVal += window[i];
	}
	
	count = winLen;
	return retVal / count;
    } else {
	for (int i = param.ones - 1; i >= 0; i--) {
	    retVal += window[fromLast(i)];
	    count = count + one;
	}
	return retVal/count;
    }
}


void RonDbClass::sample(fixedpoint f)
{
    if (winLen < MAX_HISTORY) {
	window[winLen++] = f;
    } else {
	window[winPos++] = f;
	
	if (winPos == MAX_HISTORY)
	    winPos = 0;
    }
}

/*
 * PDB completely local client implementation
 */

pdb_local_client::pdb_local_client()
{
}

pdb_local_client::~pdb_local_client()
{
    map<nodetype, RonDbClass *>::iterator i;
    
    for (i = classes.begin(); i != classes.end(); ++i) {
	delete(i->second);
    }
    
}    

int pdb_local_client::insert(node n, probe_name_t probetype, fixedpoint value)
{
    RonDbClass *rdbc = NULL;
    map<nodetype, RonDbClass *>::iterator iter;
    nodetype nt(n, probetype);

    iter = classes.find(nt);
    
    if (iter != classes.end()) {
	rdbc = iter->second;
	assert(rdbc != NULL);
    } else {
	rdbc = new RonDbClass();
	classes[nt] = rdbc;
    }
    
    rdbc->sample(value);
    
    return 0;
}

int pdb_local_client::query(node n, probe_name_t probetype, int summarizer,
	      fixedpoint param, fixedpoint *result)
{
    RonDbClass *rdbc = NULL;
    map<nodetype, RonDbClass *>::iterator iter;
    nodetype nt(n, probetype);

    iter = classes.find(nt);
    
    if (iter == classes.end()) {
	DPRINTF(DEBUG_REQUESTS,
		"pdb_local_client::query non-existent node %s type %d\n",
		n.to_string().c_str(), probetype);
	return -1;
    }
    
    rdbc = iter->second;
    
    if (rdbc == NULL) {
	DPRINTF(DEBUG_REQUESTS,
		"pdb_local_client::query Null class %s type %d\n",
		n.to_string().c_str(), probetype);
	return -1;
    }
    
    switch(summarizer) {
    case SUMMARIZE_EWMA: 
	*result = rdbc->getEwma(param);
	break;
    case SUMMARIZE_MAX:
	*result = rdbc->getMax(param);
	break;
    case SUMMARIZE_MIN: 
	*result = rdbc->getMin(param);
	break;
    case SUMMARIZE_FIRST:
	/* XXX */
	break;
    case SUMMARIZE_RECENT: 
	*result = rdbc->getRecent();
	break;
    case SUMMARIZE_AVG:
	*result = rdbc->getAvg(param); 
	break;
    case SUMMARIZE_INVALID: 
	DPRINTF(DEBUG_REQUESTS,
		"pdb_local_client::query invalid summary requested\n");
	return -1;
	break;
    default:
	DPRINTF(DEBUG_REQUESTS,
		"pdb_local_client::query unknown summary requested: %d\n",
		summarizer);
	return -1;
    }
    
    if (*result == 0.0) {
	if (rdbc->isError()) {
	    /* XXX:  wtf is this? */
	    DPRINTF(DEBUG_REQUESTS, "No EWMA estimator with alpha = %s",
		    param.to_string().c_str());
	    return -1;
	}
    }
    
    return 0;
}

