/*
 * A simple UDP-based performance database.
 */

#include "pdb.h"
#include "ronlib.h"
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/uio.h>
#include <unistd.h>
#include <string>
#include <sys/socket.h>
#include <netinet/in.h>
#include "rondebug.h"


const int nclasses = 100;
const int MAX_CLASS_NAME = 128;

/*
 * Only for numerical datatypes
 */
template <class Type>
fixedpoint pdb_ewma(pdb_res<Type> *sample_list, int *error, fixedpoint q)
{
    fixedpoint ewma, tmp;
    int rc, res;
    Type item;

    *error = 0;

    if (q == 0.0) {
	q.set(0.8);
    }

    rc = sample_list->cur(&item);
    if (rc != 0) {
	*error = rc;
	return ewma;
    }
    ewma.set(item);
    while ((res = sample_list->next(&item)) == 0) {
	tmp.set(item);
	fixedpoint one;
	one.set(1);
	tmp = tmp * (one-q);
	ewma = (ewma * q) + tmp;
    }

    return ewma;
}

template <class Type>
fixedpoint pdb_meandev(pdb_res<Type> *sample_list, int *error, fixedpoint q)
{
    fixedpoint ewma, meandev, tmp;
    int rc, res;
    Type item;

    *error = 0;

    if (q == 0.0) {
	q.set(0.8);
    }

    rc = sample_list->cur(&item);
    if (rc != 0) {
	*error = rc;
	return ewma;
    }
    ewma.set(item);
    meandev.set(0.0);
    while ((res = sample_list->next(&item)) == 0) {
	tmp.set(item);
	fixedpoint one;
	one.set(1);
	tmp = tmp * (one-q);
	ewma = (ewma * q) + tmp;
	fixedpoint newerr;
	newerr.set(0);
	newerr = tmp - ewma;
	if (newerr < 0) { newerr = newerr * -1; }
	meandev = (meandev * q) + (one - q) * newerr;
    }

    return meandev;
}

/* XXX:  This averaging method will break with large numbers. */

#define skip_n()                         \
  do {                                   \
  } while (0)
	

template <class Type>
fixedpoint pdb_avg(pdb_res<Type> *sample_list, int *error, fixedpoint q)
{
    fixedpoint ret, tmp;
    int res, rc;
    Type item;

    int count, skipcount;
    count = sample_list->count();
    if (!count) {
	*error = 1;
	return ret;
    }
    if (q.ones == 0) {
	q.ones = count;
    }
    skipcount = count - q.ones;

    if (skipcount < 1) {
	rc = sample_list->cur(&item);
	if (rc != 0) {
	    *error = rc;
	    return ret;
	}
	ret.set(item);
    } else {
	/* Skip the implicit one */
	skipcount--;
    }
    while (skipcount-- > 0) {
	res = sample_list->next(&item);
    }

    while ((res = sample_list->next(&item)) == 0) {
	tmp.set(item);
	ret = ret + tmp;
    }
    ret = ret / q;
    
    return ret;
}

template <class Type>
Type pdb_recent(pdb_res<Type> *sample_list, int *error, fixedpoint param)
{
    Type recent; /* Assume cast is possible */
    int rc, res;
    Type item;

    *error = 0;

    rc = sample_list->cur(&item);
    if (rc != 0) {
	*error = rc;
	return recent;
    }
    recent = item;
    while ((res = sample_list->next(&item)) == 0) {
	recent = item;
    }

    return recent;
}

template <class Type>
Type pdb_min(pdb_res<Type> *sample_list, int *error, fixedpoint q)
{
    Type target;
    int rc, res;
    Type item;
    
    *error = 0;

    int count, skipcount;
    int set_target = 0;
    count = sample_list->count();
    if (!count) {
	*error = 1;
	return target;
    }
    if (q.ones == 0) {
	q.ones = count;
    }
    skipcount = count - q.ones;

    if (skipcount < 1) {
	rc = sample_list->cur(&target);
	set_target = 1;
	if (rc != 0) {
	    *error = rc;
	    return target;
	}
    } else {
	/* Skip the implicit one */
	skipcount--;
    }
    while (skipcount-- > 0) {
	res = sample_list->next(&item);
    }

    while ((res = sample_list->next(&item)) == 0) {
	if (!set_target) {
	    target = item;
	    set_target = 1;
	}
	if (item < target) {
	    target = item;
	}
    }
    return target;
}

/* An unfortunate copy-pasting */
template <class Type>
Type pdb_max(pdb_res<Type> *sample_list, int *error, fixedpoint q)
{
    Type target;
    int rc, res;
    Type item;
    
    *error = 0;

    int count, skipcount;
    int set_target = 0;
    count = sample_list->count();
    if (!count) {
	*error = 1;
	return target;
    }
    if (q.ones == 0) {
	q.ones = count;
    }
    skipcount = count - q.ones;

    if (skipcount < 1) {
	rc = sample_list->cur(&target);
	set_target = 1;
	if (rc != 0) {
	    *error = rc;
	    return target;
	}
    } else {
	/* Skip the implicit one */
	skipcount--;
    }
    while (skipcount-- > 0) {
	res = sample_list->next(&item);
    }

    while ((res = sample_list->next(&item)) == 0) {
	if (!set_target) {
	    target = item;
	    set_target = 1;
	}
	if (item > target) {
	    target = item;
	}
    }
    return target;
}

void
form_response(struct pdb_update *pdbu, pdb *my_pdb)
{
    struct class_defn *retdef = NULL;
    char *classname;
    int classtype;
    fixedpoint param;

    param.set(pdbu->param);

    classname = (char *)pdbu->classname;
    classtype = ntohl(pdbu->classtype);

    retdef = my_pdb->find_class(classname);
    if (!retdef) {
	DPRINTF(DEBUG_ERRS, "query for class %s not found\n", classname);
	goto send_nak;
    }

    if (retdef->type_type != classtype) {
	DPRINTF(DEBUG_ERRS, "type mismatch in query %s:  wanted %d got %d\n",
	       classname, classtype, retdef->type_type);
	goto send_nak;
    }

    if (retdef->type_type == TYPE_STRING) {
	char item[PDB_MAXSTRING];
	pdb_res<char[PDB_MAXSTRING]> *sample_list = new pdb_res<char[PDB_MAXSTRING]>(classname, my_pdb);
	if (!sample_list) {
	    fprintf(stderr, "pdb_simple: string sample list failed\n");
	    goto send_nak;
	}
	if (!sample_list->succeeded()) {
	    fprintf(stderr, "pdb_simple: string sample list !success\n");
	    delete sample_list;
	    goto send_nak;
	}
	if (sample_list->cur(&item) != 0) {
	    perror("pdb_simple::form_response could not get cur string");
	    delete sample_list;
	    goto send_nak;
	}
	printf("String got an item: %p\n", item);
	fflush(stdout);
	if (item) {
	    printf("Item: %s\n", item);
	}
	printf("Think I got a string, but I'm quitting anyway\n");
	pdbu->classtype = htonl(TYPE_STRING);
	goto send_nak;
    } else if (retdef->type_type == TYPE_FIXED) {
	pdb_res<fixedpoint> *sample_list =
	    new pdb_res<fixedpoint>(classname, my_pdb);
	fixedpoint item;
	int err = 0;
	if (!sample_list) goto send_nak;
	if (!sample_list->succeeded()) { delete sample_list; goto send_nak; }
	switch(pdbu->method.summarize) {
	case SUMMARIZE_MIN:
	    pdbu->data.fixedval = pdb_min(sample_list, &err, param).to_wire();
	    break;
	case SUMMARIZE_MAX:
	    pdbu->data.fixedval = pdb_max(sample_list, &err, param).to_wire();
	    break;
	case SUMMARIZE_EWMA:
	    pdbu->data.fixedval = pdb_ewma(sample_list, &err, param).to_wire();
	    break;
	case SUMMARIZE_MEANDEV:
	    pdbu->data.fixedval = pdb_meandev(sample_list, &err, param).to_wire();
	    break;
	case SUMMARIZE_AVG:
	    pdbu->data.fixedval = pdb_avg(sample_list, &err, param).to_wire();
	    break;
	case SUMMARIZE_RECENT:
	    pdbu->data.fixedval = pdb_recent(sample_list, &err, param).to_wire();
	    break;
	case SUMMARIZE_FIRST:
	    if ((sample_list->cur(&item)) != 0) {
		err = 1;
		perror("pdb_simple::form_response could not get cur");
	    } else {
		pdbu->data.fixedval = item.to_wire();
	    }
	    break;
	default:
	    DPRINTF(DEBUG_ERRS,
		    "pdb_simple::form_response unsupported summarizer %d\n",
		    pdbu->method.summarize);
	    err = 1;
	}
	if (sample_list) delete sample_list;

	if (err) {
	    goto send_nak;
	}
    } else {
	DPRINTF(DEBUG_ERRS, "unsupported type %d\n", retdef->type_type);
	goto send_nak;
    }

    /* Life is good */
    pdbu->querytype = PDB_ANSWER;
    if (retdef) delete retdef;
    return;

 send_nak:
    if (retdef) delete retdef;
    pdbu->querytype = PDB_NOANSWER;
    return;
}

void usage()
{
    fprintf(stderr,
            "pdb_simple [-h] [-d debug_level]\n"
            );
}

void help()
{
    usage();
    fprintf(stderr,
            "             -h ............ help (this message)\n"
            );
}

int
main(int argc, char **argv)
{
    int sock;
    pdb *my_pdb;

    extern char *optarg;
    extern int optind;
    int ch;

    char indata[1500]; /* XXX Blah */

#ifndef max
#define max(a, b) ((a) >= (b) ? (a) : (b))
#endif

    struct sampletype *sample;
    struct stringsampletype *stringsample;
    struct class_defn tdef;

    struct stringsampletype sampledata;

    /* XXX:  This depends on stringsample being always larger... */
    sample = (struct sampletype *)&sampledata;
    stringsample = (struct stringsampletype *)&sampledata;

    sample->header.samplesize = sizeof(struct sampletype);
    
    FILE *logfile = NULL;

    
    while ((ch = getopt(argc, argv, "d:l:h")) != -1)
        switch (ch) {
	case 'd':
	    set_debug(optarg);
	    break;
        case 'h':
            help();
            exit(0);
	case 'l':
	    logfile = fopen(optarg, "a");
	    if (!logfile) {
		perror(optarg);
		exit(-1);
	    }
	    break;
        default:
            usage();
            exit(-1);
        }
    argc -= optind;
    argv += optind;

    my_pdb = new pdb;

    if (!my_pdb) {
	fprintf(stderr, "pdb_simple::main:create pdb failed");
	exit(-1);
    }
    my_pdb->db_init();

    sock = bound_udp_socket(9431);
    if (sock == -1) {
	perror("could not open UDP socket");
	exit(-1);
    }

    DPRINTF(DEBUG_INIT, "pdb_simple::main starting up successfully\n");

    while (1) {
	int rc;
	int skip = 0;
	int outlen = 1500; /* XXX */
	struct pdb_header *pdbh;
	struct pdb_update *pdbu;
	struct sockaddr_in from;
	size_t fromlen = sizeof(from);
	int dosend = 0;

	pdbh = (struct pdb_header *)indata;
	pdbu = (struct pdb_update *)(indata + sizeof(struct pdb_header));

	fd_set readfds;
	int maxfd;
	
	FD_ZERO(&readfds);
#if 0 /* For Utah DEMO ONLY */
	FD_SET(STDIN_FILENO, &readfds);
	maxfd = STDIN_FILENO;
#endif
	maxfd = sock;
	FD_SET(sock, &readfds);
	if (sock > maxfd) maxfd = sock;
	maxfd++;

	rc = select(maxfd, &readfds, NULL, NULL, NULL);
#if 0
	if (FD_ISSET(STDIN_FILENO, &readfds)) {
	    char inbuf[128];
	    char *retpos;
	    fgets(inbuf, 127, stdin); /* Flush the file descriptor */
	    retpos = strchr(inbuf, '\n');
	    if (retpos) *retpos = 0;
	    retpos = strchr(inbuf, '\r');
	    if (retpos) *retpos = 0;
	    my_pdb->pdb_dump(inbuf, stdout);
	}
#endif
	if (!FD_ISSET(sock, &readfds)) {
	    continue;
	}
	    
	rc = recvfrom(sock, indata, sizeof(indata), 0,
		      (struct sockaddr *)&from, &fromlen);
	if (rc < 0) {
	    perror("pdb_simple::recvfrom error");
	    continue;
	}
	if (rc == 0) {
	    perror("pdb_simple::main zero read");
	    continue;
	}
	outlen = rc;
	DPRINTF(DEBUG_REQUESTS,
		"pdb_simple::main got a packet of %d bytes\n", rc);
	int nsamples =  ntohl(pdbh->nsamples);
	if ((rc - sizeof(struct pdb_header)) != (sizeof(struct pdb_update) * nsamples)) {
	    fprintf(stderr, "pdb_simple::main wrong size packet %d samples size %d should be %d\n",
		    nsamples, rc,
		    sizeof(struct pdb_header) + (sizeof(struct pdb_update)*nsamples));
	    continue;
	}
	while (nsamples--) {
	    /* Repoint pdbu at the end of this while loop */
	    switch (pdbu->querytype) {
	    case PDB_INSERT:
		DPRINTF(DEBUG_REQUESTS, "pdb_simple::main got an insert packet\n");
		tdef.type_type = ntohl(pdbu->classtype);
		if (my_pdb->add_class((const char *)pdbu->classname, &tdef) != 0)
		{
		    DPRINTF(DEBUG_REQUESTS, "mypdb->add_class failed\n");
		}
		/* XXX:  This is bogus.  Verify this cruft first */
		sample->header.type = ntohl(pdbu->classtype);
		sample->header.timestamp.sec = ntohl(pdbu->timestamp.sec);
		sample->header.timestamp.usec = ntohl(pdbu->timestamp.usec);
		switch(sample->header.type) {
		case TYPE_FIXED:
		    sample->val.fixedval.set(pdbu->data.fixedval);
		    break;
		case TYPE_STRING:
		    fprintf(stderr, "Contents:  %s\n",
			    pdbu->data.freedata.contents);
		    if (pdbu->data.freedata.datalen > 255) {
			fprintf(stderr, "Error:  Too long freedata %d\n",
				pdbu->data.freedata.datalen);
			skip = 1;
			break;
		    }
		    memcpy(stringsample->val.contents,
			   pdbu->data.freedata.contents,
			   pdbu->data.freedata.datalen);
		    stringsample->header.samplesize = sizeof(struct stringsampletype);
		    break;
		default:
		    DPRINTF(DEBUG_ERRS, "pdb_simple::main unknown type!  Feh\n");
		    skip = 1;
		    break;
		}
		if (skip) { break; }
		if (logfile) {
		    fprintf(logfile, "%d.%d\t%s\t",
			    sample->header.timestamp.sec,
			    sample->header.timestamp.usec,
			    pdbu->classname);
		    fixedpoint tmpfix;
		    
		    switch(sample->header.type) {
		    case TYPE_FIXED:
			tmpfix.set(sample->val.fixedval);
			fprintf(logfile, "%s\n", tmpfix.to_string().data());
			break;
		    case TYPE_STRING:
			fprintf(logfile, "%s\n", stringsample->val.contents);
			break;
		    }
		    fflush(logfile);
		}
		
		if ((my_pdb->add_sample((char *)pdbu->classname, sample)) != 0) {
		    perror("pdb_simple::main add_sample failed");
		    exit(-1);
		}
		DPRINTF(DEBUG_REQUESTS, "pdb_simple::main added a sample for %s\n",
			(char *)pdbu->classname);
		break;
	    case PDB_QUERY:
		DPRINTF(DEBUG_REQUESTS, "pdb_simple::main got a query packet\n");
		form_response(pdbu, my_pdb);
		dosend = 1;
		break;
	    default:
		fprintf(stderr, "pdb_simple::main got unknown querytype %d\n",
			pdbu->querytype);
		break;
	    }
	    pdbu++;
	}
	if (dosend) {
	    sendto(sock, indata, outlen, 0,
		   (struct sockaddr *)&from, fromlen);
	    dosend = 0;
	}
    }

    /* Yeah, yeah, I know - we can't get here.
     * Deal with having a real program structure later;
     * I want to get the database and network interface
     * working for now
     * (famous last words).  -dga, Nov-08-2000.
     */
    
    my_pdb->pdb_cleanup();

    exit(0);

}
