/*
 * A test harness to test the performance database.
 */

#include "pdb.h"
#include <string>

#include <db_cxx.h>

const int nclasses = 100;
const int MAX_CLASS_NAME = 128;

/*
 * Return codes
 */

enum retvals {
    TEST_PASSED,
    MEMORY_ERROR,
    CLASS_ERROR,
    SAMPLE_ERROR
};

/*
 * Test function:  classes
 *
 * Create and destroy sample classes in the database and ensure
 * that everything remains happy
 */

int
test_classes(pdb *my_pdb)
{
    string classnames[nclasses];
    struct class_defn tdef, *retdef;
    int res;
    
    char strbuf[MAX_CLASS_NAME];

    for (int i = 0; i < nclasses; i++) {
	sprintf(strbuf, "class-test-%d", i);
	classnames[i] = strbuf;
    }

    tdef.type_type = TYPE_FIXED;

    /*
     * Insert and immediately test
     */

    for (int i = 0; i < nclasses; i++) {
	if ((res = my_pdb->add_class(classnames[i].data(), &tdef)) != 0) {
	    fprintf(stderr, "pdb_test:test_classes:add_class failed on %s (retval %d)\n",
		    classnames[i].data(), res);
	    return CLASS_ERROR;
	}
	retdef = my_pdb->find_class(classnames[i].data());
	if (!retdef ||
	    retdef->type_type != TYPE_FIXED) {
	    fprintf(stderr, "pdb_test:test_classes:find_class failed on %s\n",
		    classnames[i].data());
	    return CLASS_ERROR;
	}
    }

    /*
     * And then check again to see if we clobbered something
     */

    for (int i = 0; i < nclasses; i++) {
	retdef = my_pdb->find_class(classnames[i].data());
	if (!retdef ||
	    retdef->type_type != TYPE_FIXED) {
	    fprintf(stderr, "pdb_test:test_classes:find_class failed test2 on %s\n",
		    classnames[i].data());
	    return CLASS_ERROR;
	}
    }

    /*
     * Delete them.  This is NOT YET IMPLEMENTED.
     */
#if 0
    for (int i = 0; i < nclasses; i++) {
	/* Fill this in when it's implemented */
	if ((res = my_pdb->delete_class(classnames[i].data())) != 0) {
	    fprintf(stderr, "pdb_test:test_classes:delete class failed on %s res %d\n",
		    classnames[i].data(), res);
	    return CLASS_ERROR;
	}
    }

    /*
     * And then make sure we can't find them afterwords
     */
    for (int i = 0; i < nclasses; i++) {
	retdef = my_pdb->find_class(classnames[i].data());
	if (retdef) {
	    fprintf(stderr, "pdb_test:test_classes:find_class found deleted item %s\n",
		    classnames[i].data());
	    return CLASS_ERROR;
	}
    }
#endif
    
    return TEST_PASSED;
}

/*
 * Create a few classes of data, and insert data in them.
 * Test to see that the data comes out in the right order
 * and with the right values.
 *
 * At present, this is not a very sophisticated test.
 */

int
test_samples(pdb *my_pdb)
{
    struct class_defn tdef;
    struct sampletype sample;
    fixedpoint item;
    int res;


    tdef.type_type = TYPE_FIXED;
    
    my_pdb->add_class("latency", &tdef);
    my_pdb->add_class("bandwidth", &tdef);

    sample.header.type = TYPE_FIXED;
    sample.header.timestamp.sec = time(NULL);
    fixedpoint point(3, 0);
    sample.val.fixedval = point;
    
    for (int i = 0; i < 100; i++) {
	point.set(i, 0);
	sample.val.fixedval = point.to_wire();
	if ((res = my_pdb->add_sample("latency", &sample))) {
	    fprintf(stderr, "pdb_test:test_samples:add_sample latency failed (%d)\n",
		    res);
	    return SAMPLE_ERROR;
	}
	point.set(200-i, 0);
	sample.val.fixedval = point.to_wire();
	if ((res = my_pdb->add_sample("bandwidth", &sample))) {
	    fprintf(stderr, "pdb_test:test_samples:add_sample bandwidth failed (%d)\n",
		    res);
	    return SAMPLE_ERROR;
	}
	sample.header.timestamp.sec += 2;
    }

    pdb_res<fixedpoint> *sample_list = new pdb_res<fixedpoint>("bandwidth", my_pdb);
    
    if (!sample_list) {
	fprintf(stderr, "pdb_test:test_samples:get_samples failed for bandwidth");
	return SAMPLE_ERROR;
    }

    if ((res = sample_list->cur(&item)) != 0) {
	fprintf(stderr, "pdb_test:test_samples:sample_list->cur() failed for bandwidth (%d)\n",
		res);
	return SAMPLE_ERROR;
    }
    point.set(200, 0);
    if (point != item) {
	fprintf(stderr, "pdb_test:test_samples:first sample returned was %s, not %s\n",
		item.to_string().data(),
		point.to_string().data());
	return SAMPLE_ERROR;
    }

    for (int i = 199; i >= 101; i--) {
	if ((res = sample_list->next(&item)) != 0) {
	    fprintf(stderr, "pdb_test:test_samples:sample_list->next() failed for bandwidth (%d)\n",
		    res);
	    return SAMPLE_ERROR;

	}
	point.set(i);
	if (point != item) {
	    fprintf(stderr, "pdb_test:test_samples:next sample returned was %s, not %s\n",
		    item.to_string().data(), point.to_string().data());
	    return SAMPLE_ERROR;
	}
    }

    /* XXX:  Delete the samples here */
	
    return TEST_PASSED;
}

int
main(int argc, char **argv)
{
    int exit_code = 0;
    
    pdb *my_pdb;

    my_pdb = new pdb;

    if (!my_pdb) {
	fprintf(stderr, "pdb_test:main:create pdb failed");
	exit(-1);
    }
    
    my_pdb->db_init();
    if (test_classes(my_pdb) == TEST_PASSED) {
	printf("pdb_test:test_classes:success\n");
    } else {
	printf("pdb_test:test_classes:failure\n");
	exit_code = -1;
    }

    if (test_samples(my_pdb) == TEST_PASSED) {
	printf("pdb_test:test_samples:success\n");
    } else {
	printf("pdb_test:test_samples:failure\n");
	exit_code = -1;
    }

    my_pdb->pdb_cleanup();

    exit(exit_code);

}
