/**
 *  \file RONOCDStateInfo.cpp
 *  \brief This file defines the functions associated with the RON specific OCD state.
 */

#include "OCI.h"
#include "OCDStateInfo.h"

#include "RONOCDStateInfo.h"
#include "RONOverlayNetwork.h"

/**
  * This constructor is used to create a RON Virtual node when a DNS request has been captured.
  * @param dnsPrefStr the DNS pref string which lead to the creation of this VN.
  * @param fromLayer The layer which created this layer.
  * @param overlayToUse The overlay network associated with this virtual node.
  */
RONOCDStateInfo::RONOCDStateInfo (
                                OCDHandle myHandle,
                                OCI *oci,
                                RONOverlayNetwork* overlay
                            )
        : OCDStateInfo (myHandle, oci, overlay) {
    
    this->overlay = overlay;
    this->status = OCD_STATE_STATUS_NOT_READY;

    this->myIPAddr = overlay->proxy->get_my_addr_nf();
}


RONOCDStateInfo::RONOCDStateInfo(OCDHandle myHandle, OCI* oci, RONOverlayNetwork *overlay, struct in_addr realIPAddr)
            :OCDStateInfo (myHandle, oci, overlay) {

    this->overlay = overlay;
    this->realIPAddress = realIPAddr;

    this->status = OCD_STATE_STATUS_NOT_READY;
    this->myIPAddr = overlay->proxy->get_my_addr_nf();
}

    
    
    /** Destructor */
RONOCDStateInfo::~RONOCDStateInfo() {

}

/**
  * This function is called when the reply to a DNS request made
  * by this Virtual node arrives.
  * @param ip The IP address corresponding to the request.  Error if 0.
  */
void RONOCDStateInfo::receivedLegacyDNSReply(uint32_t ip) {
    
    if (ip != 0) {
      
        {
            struct in_addr tmpIPAddr;
            tmpIPAddr.s_addr = ip;

            PRINT_DEBUG ( 
                    DEBUG_LEVEL_SUPER,
                    "RON OCD state with handle %lu received DNS reply: IP = %s.\n",
                    myHandle,
                    inet_ntoa (tmpIPAddr)
                );
        }

        realIPAddress.s_addr = htonl (ip);
    
	    int offset = 0;

	    char tmpPkt [MAX_PACKET_BUF_LEN];

        RONSetupRequestBlock *requestBlk = (RONSetupRequestBlock *) (tmpPkt + offset);
        requestBlk->header.destHandle = NULL_HANDLE; //we don't know this yet.
        requestBlk->header.blockType = RON_BLOCK_SETUP_REQUEST;
        requestBlk->header.nextBlock = sizeof (RONSetupRequestBlock);
        requestBlk->srcHandle = myHandle;
        requestBlk->srcAddr =  myIPAddr;

        printf ("\n\n\n********* %s\n\n\n", inet_ntoa (myIPAddr));
	    offset += sizeof (RONSetupRequestBlock);
        // add any piggy back packets
        // Now pack the RON independent block.
        RONProxyBlock* genericBlock = (RONProxyBlock *) (tmpPkt + offset);
    
        // Set the top level headers
        RONOverlayNetwork::setRONBlock ( (RONBlock *) genericBlock, RON_BLOCK_PROXY, 0 /* last block */);
        int genBlockSize = setupOptionLen; /* TODO FIX vn->packToSendBlocks(genericBlock->genericBlock) ; */
        memcpy (genericBlock->proxyBlock, setupOptionBlocks, setupOptionLen);
	    int tmpPktLen = 0;
        if (genBlockSize == 0) {
            // no generic blocks to be piggybacked
            requestBlk->header.piggybackLen = 0;
	        tmpPktLen = offset;
        } else {
    
            tmpPktLen = offset + (genBlockSize + sizeof (RONProxyBlock));
            requestBlk->header.piggybackLen = genBlockSize + sizeof (RONProxyBlock);
        }
    
        //send the Setup message
        PRINT_DEBUG (DEBUG_LEVEL_SUPER,
            "RON sending setup request to %s\n", inet_ntoa ( realIPAddress));
        
        overlay->conduit->send ( (char *) tmpPkt, tmpPktLen, realIPAddress);
    
    
        //Processing continues when the RON setup reply is received.
    
    } else {
        //Unable to resolve the requested URL
        PRINT_DEBUG (
                DEBUG_LEVEL_MINIMAL,
                "Unable to resolve %s.\n",
                endPointName
            );
        notifySetupFailure();
    }

}

/**
  * Returns a hash of the src IP, src port, dest IP and dest ports.
  */
long RONOCDStateInfo::calculateFlowId (char *ip_hdr) {

    uint8_t ipHeaderLen;
    memcpy (&ipHeaderLen, ip_hdr, 1);
    ipHeaderLen = ipHeaderLen & 0x0F; //the header len is only the last 4 bits

    ipHeaderLen = ipHeaderLen * 4; // ip header len in bytes
    
    struct in_addr srcAddr;
    memcpy (&srcAddr, ip_hdr + 12, sizeof (srcAddr));

    struct in_addr dstAddr;
    memcpy (&dstAddr, ip_hdr + 16, sizeof (dstAddr));

    uint16_t srcPort;
    memcpy (&srcPort, ip_hdr + ipHeaderLen, sizeof (uint16_t));
    
    uint16_t dstPort;
    memcpy (&dstPort, ip_hdr + ipHeaderLen + 2, sizeof (uint16_t));
    
    PRINT_DEBUG (DEBUG_LEVEL_SUPER,
                "IP Header Len = %d bytes, src port = %d dst port = %d\n",
                ipHeaderLen, srcPort, dstPort
            );

    return srcAddr.s_addr + dstAddr.s_addr + srcPort + dstPort; //TODO take hashof this
    
}
