/**
 *  \file RONOverlayNetwork.cpp
 *  \brief The RONOverlayNetwork class simply acts as an intermediary
 *  between the proxy and the actual MIT RON code.
 */
#include "IPPacket.h"
#include "utilFunctions.h"
//#include "SecurityManager.h"
#include "TFunctor.h"

#include "Proxy.h"


#include "OCI.h"
#include "OCD.h"

#include "RONOverlayNetwork.h"
#include "RONOCDStateInfo.h"
#include "ronDebug.h"

#include "ronHeaders.h"
#include "RONPacket.h"

#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>

#include <string>
#include <assert.h>

using namespace RON;
/**
  * Constructor.
  * @param type         The type of the overlay network. For eg: RON
  * @param id           The id of the overlay network. For eg: RON.PlanetLab
  * @param desc         A brief description of the overlay.
  * @param fileName     The name of the overlay specific configuration
  * file
  * @param urlSuffixes  The string of space separated suffixes of URLs to be
  * handled by this overlay network.
  * @param handleLegacyURLs Should this overlay handle legacy URLs like 
  * www.abc.com?
  * @param tunnel The tunnelling (packet capturing device) used for
  * sending packets to the legacy application.
  */
RONOverlayNetwork::RONOverlayNetwork(
        string type, string id, string desc, 
        string fileName, 
        string urlSuffixes, bool handleLegacyURLs
    )
        :OCD(  type, id, desc, fileName, 
                urlSuffixes, handleLegacyURLs) {
  

}

/**
  * Construct a new RONOverlayNetwork with fields taken from the
  * specified overlay network.
  */
RONOverlayNetwork::RONOverlayNetwork(OCD* o) : OCD() {

    this->type = o->type;
    this->id = o->id;
    this->description = o->description;
    this->configFileName = o->configFileName;
    this->urlSuffixes = o->urlSuffixes;
    this->handleLegacyURLs = o->handleLegacyURLs;
}

/**
  * Initializes the RON Overlay interface.  
  * A public trigger for this proxy is inserted in the i3
  * infrastructure.  The process is synchronous, i.e. the function will
  * return only after all initialization is complete.
  * @param proxy A pointer to the proxy of which this overlay is a part.
  */
void RONOverlayNetwork::init(Proxy *prxy, OCI* oci, bool recoverFromCrash) {

    this->proxy = prxy;
    this->oci  = oci;
    
    PRINT_DEBUG(RON_DEBUG_LEVEL_MINIMAL, "Initializing RON Overlay %s...\n", id.c_str());
    
    //TODO: These options should be read in from the config file.
    int forw_port = DEFAULT_RON_FORWARDER_PORT;
    char myhostname[MAXHOSTNAMELEN];
    int hostnameset = 0;
    int use_mesh_protocol = 0;
    int test_mesh = 0;

    unsigned long myaddr;
    bool addressFoundFlag;
    struct in_addr myinaddr = proxy->getCurrentIP(&addressFoundFlag);
    
    if ( ! addressFoundFlag) {
        PRINT_INFO (INFO_LEVEL_FATAL_ERROR,
                "Unable to obtain local IP address for RON.\n"
            );
        EXIT_ON_ERROR;
    }
    
    myaddr = htonl(myinaddr.s_addr);
    struct in_addr myinaddr_nf;
    myinaddr_nf.s_addr = htonl (myinaddr.s_addr);

    flood_protocol *fp;
    fp = new flood_protocol();

    char tmpPeerFileName[1000];
    //parse the configFileName and extract the name of the peer list file and
    //the routing metric to be used.
    parseConfigFile (tmpPeerFileName);
    membership *memb = new broadcast_membership(tmpPeerFileName, fp, this); //TODO read from config file.

    if (!memb) {
        PRINT_INFO (INFO_LEVEL_FATAL_ERROR,
                "Unable to create membership!\n"
            );
        EXIT_ON_ERROR;
    }

    conduit = new ip_conduit(this);

    pdb_client *pdbc = new pdb_local_client();
    pdbc->init();

#if 1 /* Use the latency-based router */
    latency_router *router;
    router = new latency_router(myaddr, forw_port, pdbc, this);
    if (!router) {
	    PRINT_INFO (
                INFO_LEVEL_WARNING, 
                "RON::main could not create latency router");
	    EXIT_ON_ERROR;
    }
#else
    static_router *router;
    router = new static_router();
#endif
   
    Scheduler* sched = new Scheduler(proxy);

    forw* myforw = new forw(sched, myinaddr_nf, forw_port, router, memb, this);
    PRINT_DEBUG (
            RON_DEBUG_LEVEL_VERBOSE,
            "RON::membership %p forwarder %p\n",
	    memb, myforw);
    
    if (!myforw) {
	    perror("could not initialize forwarder");
	    exit(-1);
    }
    
    myforw->init(forw_port);
    forw* forw_ptr = myforw;  //this was declared global variable in original code
    myforw->stats.counter_table["start"] = time(NULL);

    mesh_protocol *mp = new mesh_protocol();

#if 0
    passive_monitor *monitor = new passive_monitor();
    if (!monitor) {
	perror("could not initialize passive monitor");
	exit(-1);
    }
#endif
    /* Set up the outbound routing between components.  In this case,
     * it's all trivial - they simply all send their packets
     * back to the forwarder */
    fp->register_default(myforw);
    router->register_default(myforw);
#if 0 /* XXX:  not ready for prime-time, but want to commit print
       * protocol changes... */
    conduit->register_default(monitor);  /* All IP goes through monitor */
    monitor->register_default(myforw);
#else
    conduit->register_default(myforw);
#endif
    if (use_mesh_protocol) {
	    conduit->register_type(RON_MESH_DATA, mp, NULL);
    }
    
    memb->register_default(myforw);
    fp->register_default(myforw);
    mp->register_default(myforw);

    /* Let them all register their callbacks and such.  After
     * 'configure' is called, the system should be ready to run. */
    fp->configure(sched, myforw);
    router->configure(sched, myforw);
    conduit->configure(sched, myforw);
#if 0
    monitor->configure(sched, myforw);
#endif
    memb->configure(sched, myforw);
    mp->configure(sched, myforw);
    pdbc->configure(sched, myforw);

#if 1
    ping_protocol *ping = new ping_protocol(pdbc, this);
    ping->configure(sched, myforw);
    ping->register_default(myforw);
#endif

#if 0 //Initially ifndef IP_FORW
    bool active_conduit = true;//This was a command line opiton
    if (active_conduit) {
	    conduit->set_active(1);
    }

    mesh_tester *mtest = NULL;
    if (test_mesh) {
	    mtest = new mesh_tester();
	    mtest->configure(sched, myforw);
	    mtest->register_default(myforw);
    }
#endif

}


void RONOverlayNetwork::handleSetupReplyBlock ( RONSetupReplyBlock *blk) {

    RONOCDStateInfo *ronOCDState = (RONOCDStateInfo *) getStateInfo (blk->header.destHandle);
    if (ronOCDState != NULL) {

        ronOCDState->remoteHandle = blk->srcHandle;

        char *optionBlocks = NULL;
            
        int optionLen = 0;
        if (blk->header.piggybackLen > 0) {
               
		    PRINT_DEBUG (DEBUG_LEVEL_SUPER, "PIggback blocks present\n"); 
            //piggyback blocks are present
            RONBlock* tmpRONBlk = (RONBlock *) (((char *) blk) + sizeof (RONSetupReplyBlock));
            if (tmpRONBlk->blockType == RON_BLOCK_PROXY) {
                RONProxyBlock *tmpRONGenericBlock = (RONProxyBlock *) tmpRONBlk;
                optionBlocks = (char *) tmpRONGenericBlock->proxyBlock;
                optionLen = tmpRONGenericBlock->header.nextBlock - sizeof (RONProxyBlock);
            } 
        } 

	    ronOCDState->status = OCD_STATE_STATUS_READY;
        ronOCDState->notifySetupSuccess();

    } else {

        PRINT_DEBUG (RON_DEBUG_LEVEL_MINIMAL,
                "Received setup reply for unknown handle %lu.\n",
                blk->header.destHandle
            );
    }
}

/**
  * This function is called when a tunnel setup request packet is received
  * by the RON OCD.
  */
void RONOverlayNetwork::handleSetupRequestBlock (RONSetupRequestBlock* setupBlk) {

    PRINT_DEBUG (RON_DEBUG_LEVEL_MINIMAL,
            "RON handling Setup Request Block from RON handle %lu\n",
            setupBlk->srcHandle);


    // Check if we have seen this guy before.
    RONOCDStateInfo * ocdState = lookupRemoteVN_remote(setupBlk->srcHandle);

    if (ocdState == NULL) {

        //This remote sender is unknown, i.e. we haven't seen him before,
        //So create a new OCDState.

        // Create new OCD state
        OCDHandle ocdHandle = getNewOCDHandle();
        ocdState = new RONOCDStateInfo (ocdHandle, oci, this);
        putStateInfo (ocdHandle, ocdState);
        
        ocdState->realIPAddress = setupBlk->srcAddr;
        ocdState->overlay = this;
        ocdState->remoteHandle = setupBlk->srcHandle;
        
        addRemoteHandle (ocdState->remoteHandle, ocdState);
 
   }	

    char *optionBlocks = NULL;
    int optionLen = 0;
    if (setupBlk->header.piggybackLen > 0) {
                
        //piggyback blocks are present
        RONBlock* tmpRONBlk = (RONBlock *) (((char *) setupBlk) + sizeof (RONSetupRequestBlock));
        if (tmpRONBlk->blockType == RON_BLOCK_PROXY) {
            RONProxyBlock *tmpRONGenericBlock = (RONProxyBlock *) tmpRONBlk;
            optionBlocks = (char *) tmpRONGenericBlock->proxyBlock;
            optionLen = tmpRONGenericBlock->header.nextBlock - sizeof (RONProxyBlock);
        } 
    } 

    sendSetupReply (ocdState, optionBlocks, optionLen);
}

void RONOverlayNetwork::sendSetupReply (RONOCDStateInfo *ocdState, char *optionBlocks, int optionLen) {

    PRINT_DEBUG (
                DEBUG_LEVEL_VERBOSE,
                "RONOverlay: Sending setup reply for OCDState with handle %lu.\n",
                ocdState->get_myHandle()
            );

    // Now send the tunnel setup success reply
    ocdState->status = OCD_STATE_STATUS_READY;

    int offset = 0;
    char tmpPkt[MAX_PACKET_BUF_LEN];

    RONSetupReplyBlock* replyBlk = (RONSetupReplyBlock *) (tmpPkt + offset) ;

    replyBlk->header.destHandle = ocdState->remoteHandle;
    replyBlk->header.blockType = RON_BLOCK_SETUP_REPLY;
    replyBlk->header.nextBlock = sizeof (RONSetupReplyBlock);
    replyBlk->srcHandle = ocdState->get_myHandle();

    offset = sizeof (RONSetupReplyBlock);

    // Now pack the RON independent block.
    RONProxyBlock* genericBlock = (RONProxyBlock *) (tmpPkt + offset);
    
    // Set the top level headers
    RONOverlayNetwork::setRONBlock ( (RONBlock *) genericBlock, RON_BLOCK_PROXY, 0 /* last block */);

    int genBlockSize = optionLen; /* TODO FIX vn->packToSendBlocks(genericBlock->genericBlock) ; */
    memcpy (genericBlock->proxyBlock, optionBlocks, optionLen);

    int tmpPktLen = 0;
    if (genBlockSize == 0) {
 	    PRINT_DEBUG (DEBUG_LEVEL_SUPER, "NO packets to piggyback on RON SETUP REPLY\n\n\n");
        tmpPktLen = offset;
        replyBlk->header.piggybackLen = 0;
    } else {

        tmpPktLen = offset + (genBlockSize + sizeof(RONProxyBlock));
        replyBlk->header.piggybackLen = genBlockSize + sizeof (RONProxyBlock); // Because this is the last block
    }

    //Send the reply to the requestor.
	PRINT_DEBUG (DEBUG_LEVEL_SUPER, "Sending RON SETUP REPLY\n");
    conduit->send ( (char *) tmpPkt, tmpPktLen, ocdState->realIPAddress);

}

void RONOverlayNetwork::handleProxyBlock ( RONProxyBlock *blk, int blkLen) {

    RONOCDStateInfo* ocdState = (RONOCDStateInfo *) getStateInfo (blk->header.destHandle);
    
    if (ocdState != NULL) {

#ifdef BENCHMARK
	addCheckpoint (EVENT_DATA_OCD_TO_OC_I, 0);
#endif
        ocdState->oci->receive (
                            (char *)blk->proxyBlock, 
                            this,
                            ocdState->get_myHandle()
                );
    } else {
        
        PRINT_DEBUG ( RON_DEBUG_LEVEL_MINIMAL,
                "Proxy block received by RON on unknown handle %lu\n", blk->header.destHandle);
        
    }
    
}

void RONOverlayNetwork::handleFastProxyBlock ( RONFastProxyBlock *blk, int blkLen) {

    RONOCDStateInfo* ocdState = (RONOCDStateInfo *) lookupRemoteHandle (blk->myHandle);
    
    if (ocdState != NULL) {

#ifdef BENCHMARK
	addCheckpoint (EVENT_DATA_OCD_TO_OC_I, 0);
#endif
        ocdState->oci->receive (
                            (char *) blk->proxyBlock, 
                            this,
                            ocdState->myHandle
                );
    } else {
        
        PRINT_DEBUG ( RON_DEBUG_LEVEL_MINIMAL,
                "Fast Proxy block received by RON on unknown handle %lu\n", blk->myHandle);
        
    }
    
}

RONOCDStateInfo * RONOverlayNetwork::lookupRemoteHandle (OCDHandle remoteHandle) {
    map<OCDHandle, RONOCDStateInfo *>::iterator itr = remoteHandleMap.find (remoteHandle);
    
    if (itr != remoteHandleMap.end()) {
        return (RONOCDStateInfo*)itr->second;
        
    } else {
            
        return NULL;
    }


}


/**
  * This function is called when a data packet is received over the RON overlay
  * This data packet needs to be handed over to the legacy application waiting for it.
  * @param ip_hdr The entire contents of the packet, starting with the IP header
  * @param len The length of the data packet
  */
void RONOverlayNetwork::handleRONPacket(char *ip_hdr, int len) {

    PRINT_DEBUG(RON_DEBUG_LEVEL_MINIMAL, "Handling packet received over RON.\n");
    RONBlock *ronPkt = (RONBlock *) ip_hdr;

    switch (ronPkt->blockType) {
        
        case RON_BLOCK_SETUP_REQUEST:
	    PRINT_DEBUG (DEBUG_LEVEL_SUPER, "RON_BLOCK_SETUP_REQUEST\n");
            handleSetupRequestBlock ( (RONSetupRequestBlock *) ip_hdr);
        break;

        case RON_BLOCK_SETUP_REPLY:
	    PRINT_DEBUG (DEBUG_LEVEL_SUPER, "RON_BLOCK_SETUP_REPLY\n");
            handleSetupReplyBlock ( (RONSetupReplyBlock *) ip_hdr);
        break;

        case RON_BLOCK_PROXY:
	    PRINT_DEBUG (DEBUG_LEVEL_SUPER, "\n\n\t\t\tRON_BLOCK_PROXY\n\n");
            //This block is for the proxy.
            handleProxyBlock ( (RONProxyBlock *) ip_hdr, len);
        break;

	case RON_BLOCK_FAST_PROXY:
	    PRINT_DEBUG (DEBUG_LEVEL_SUPER, "\n\n\t\t\tRON_BLOCK_FAST_PROXY\n\n");
	    // This block is a result of enabling fast connection startup
	    handleFastProxyBlock ( (RONFastProxyBlock *) ip_hdr, len);
	break;

        default:
            PRINT_DEBUG (RON_DEBUG_LEVEL_MINIMAL,
                    "Unknown RON packet (type = %d) received.\n",
                    ronPkt->blockType
                );
        break;
    }
  
}

/** TODO KILL
  * We have now obtained the actual IP address of this virtual node. 
  * Change its status to OK, send the fake DNS replies to the legacy applications and wait for data packets. 
  * Add the VN to the necessary Proxy tables.
  */
void RONOverlayNetwork::completedSetup (RONOCDStateInfo *ocdState) {

    PRINT_DEBUG (
                    RON_DEBUG_LEVEL_VERBOSE,
                    "Completed Setup for RON tunnel with handle %lu\n",
                    ocdState->myHandle
            );

    // If any packet is received from this realIPAddress,
    // we should be able to retrieve the VN.
    addRemoteVN_local (ocdState);
        
    ocdState->status = OCD_STATE_STATUS_READY;
    ocdState->notifySetupSuccess();
    

}


/**
  * Gives a string representation (pretty printed) of the RONOverlayNetwork.
  * @param tmpStr a character buffer in which the string is stored.
  * Sufficient space should be allocated for tmpStr.
  * @return pointer to tmpStr
  */
char* RONOverlayNetwork::toString(char *tmpStr) {

    sprintf( tmpStr,
            "RON Overlay Network\n"
            "\ttype\t\t:\t%s\n"
            "\tid\t\t:\t%s\n"
            "\tdescription\t:\t%s\n"
            "\tconfig file\t:\t%s\n",
            type.c_str(),
            id.c_str(),
            description.c_str(),
            configFileName.c_str()
           );
    return tmpStr;
}


/** 
 * This function is called when the proxy is restarted after 
 * a crash.  The journal associated with the overlay (if it exists)
 * is used to restore the state of the overlay.
 */
void RONOverlayNetwork::recoverFromCrash() {

}

/**
 * Returns RON specific status information 
 */
int RONOverlayNetwork::getStatus() {

}

void RONOverlayNetwork::addRemoteVN_local (RONOCDStateInfo *ronOCDState) {
   
    remoteVNMap_local.insert (pair<long, RONOCDStateInfo*> (ronOCDState->realIPAddress.s_addr, ronOCDState)); 

}

void RONOverlayNetwork::addRemoteVN_remote (RONOCDStateInfo *ronOCDState, long flowId) {
    
    remoteVNMap_remote.insert (pair<long, RONOCDStateInfo*> (flowId, ronOCDState));
}

RONOCDStateInfo* RONOverlayNetwork::lookupRemoteVN_local (struct in_addr remoteAddr) {

    map<long, RONOCDStateInfo*>::iterator itr = remoteVNMap_local.find(remoteAddr.s_addr);
    
    if (itr != remoteVNMap_local.end()) {
        return (RONOCDStateInfo*)itr->second;
        
    } else {
            
        return NULL;
    } 
}

RONOCDStateInfo* RONOverlayNetwork::lookupRemoteVN_remote (long flowId) {

    map<long, RONOCDStateInfo*>::iterator itr = remoteVNMap_remote.find(flowId);
    
    if (itr != remoteVNMap_remote.end()) {
        return (RONOCDStateInfo*)itr->second;
        
    } else {
            
        return NULL;
    } 
}

/**
  * Set up a tunnel to the specified end point name over i3.  
  * If a tunnel already exists, reuse it.
  * @param ociHandle The oci to be notified through a callback when the tunnel setup succeeds
  * or fails.
  * @parentLayer The layer which is setting up this new layer.
  * @return a handle to the tunnel.
  */
OCDHandle RONOverlayNetwork::setup (
                            const string endPointName, 
                            int endPointNameType,
                            char *optionBlocks, 
                            int optionLen, 
                            OCIHandle ociHandle,
                            OCI* oci) {

    RONOCDStateInfo *ronOCDState = (RONOCDStateInfo*) lookupEndPoint (endPointName);
   
    
    OCDHandle ronHandle = NULL_HANDLE;
    if (ronOCDState == NULL) {
        // No tunnel exists to the specified end point.  Create a new one.
        ronHandle = getNewOCDHandle();
        ronOCDState = new RONOCDStateInfo (ronHandle, oci, this);

        ronOCDState->addParentHandle (ociHandle);

        putStateInfo (ronHandle, ronOCDState);

        // add newly created state to hash table keyed by end point name.
        putEndPoint  (endPointName, ronOCDState);

        strcpy (ronOCDState->endPointName, endPointName.c_str());

        ronOCDState->setupOptionBlocks = optionBlocks;
        ronOCDState->setupOptionLen = optionLen;

        PRINT_DEBUG (
                DEBUG_LEVEL_SUPER, 
                "In RONOverlayNetwork setup on handle %lu, endpointName  = %s\n", 
                ronHandle,
                ronOCDState->endPointName
            );

        // Make a real DNS query for the URL with the suffix stripped.
        
        // Strip off .ron suffix before trying to resolve URL to IP
        char reqURL[MAX_URL_LEN];
        strcpy (reqURL, ronOCDState->endPointName);
        stripSuffix (reqURL, ".ron");   //TODO
 
        if (proxy->makeDNSRequest (ronOCDState, reqURL)) {
    
            //DNS successfully sent to the name server of the machine on which the proxy is running

        } else {
            PRINT_INFO (
                INFO_LEVEL_WARNING,
                "Unable to send DNS requested to the name server of this machine.\n"
            );
        }

        // Processing continues when the receivedLegacyDNSReply() is called on ronOCDState.
 
    } else {
        ronHandle = ronOCDState->myHandle;
    }

    return ronHandle;
    
}

void RONOverlayNetwork::send ( OCDHandle handle, char *data, int dataLen) {
    
    PRINT_DEBUG (RON_DEBUG_LEVEL_VERBOSE, "RON has been asked to send a packet.\n");
    RONOCDStateInfo *ronOCDState = (RONOCDStateInfo*) getStateInfo (handle);

    assert (ronOCDState != NULL);
    
    // Change the src and dest address of the data packet to reflect
    // the real source (address of local machine) and destination addresses (address of remote machine).

    //TODO FIX proxy->change_addr_fix_checksum (pktContents, proxy->my_addr_nf, ronOCDState->realIPAddress);

    char tmpBuf [MAX_PACKET_BUF_LEN];
    int tmpLen = -1;

    if (ronOCDState->remoteHandle != NULL_HANDLE) {
    
	// pack the RON packet in a RONProxyBlock;
    	RONProxyBlock *proxyBlk = (RONProxyBlock *) tmpBuf;
    	proxyBlk->header.destHandle = ronOCDState->remoteHandle;
    	proxyBlk->header.blockType = RON_BLOCK_PROXY;
    
    	tmpLen = dataLen + sizeof (RONProxyBlock);
    	proxyBlk->header.nextBlock = tmpLen;
    	memcpy (proxyBlk->proxyBlock, data, dataLen);
    
    } else if (proxy->enableRONFastStartup) {
	// We haven't heard the setup reply from the end point yet.
	// We are in fast connection start up mode.
	// pack the RON packet in a RONFastProxyBlock;
    	RONFastProxyBlock *proxyBlk = (RONFastProxyBlock *) tmpBuf;
    	proxyBlk->myHandle = ronOCDState->myHandle;
    	proxyBlk->header.blockType = RON_BLOCK_FAST_PROXY;
    
    	tmpLen = dataLen + sizeof (RONFastProxyBlock);
    	proxyBlk->header.nextBlock = tmpLen;
    	memcpy (proxyBlk->proxyBlock, data, dataLen);
    }

    conduit->send (tmpBuf, tmpLen, ronOCDState->realIPAddress);
}

/** Fill the details of an RONBlock */
void RONOverlayNetwork::setRONBlock (RONBlock *ronBlk, uint8_t ronMsgType, uint16_t blkLen) {
    
    ronBlk->blockType = ronMsgType;
    ronBlk->nextBlock = blkLen;
    
}

void RONOverlayNetwork::addRemoteHandle (OCDHandle handle, RONOCDStateInfo *state) {

    remoteHandleMap.insert ( pair<OCDHandle, RONOCDStateInfo *> (handle, state));
}

/**
  * parse the configFileName and extract the name of the peer list file and
  * the routing metric to be used.
  * @return true if the parsing was successful, else returns false.
  */
bool RONOverlayNetwork::parseConfigFile (char* peerListFileName) {
    FILE* tmpFD = fopen (configFileName.c_str(), "r");
    if (tmpFD == NULL) {
        //Unable to open the configuration file.
        PRINT_INFO (
                        INFO_LEVEL_WARNING,
                        "Unable to open the RON configuration file %s.\n",
                        configFileName.c_str()
                    );
        return false;
    }

    char tmpBuf[1000];
        
    // The first line of the configuration file is the name of the peer list file.
    if (fgets(tmpBuf, 1000, tmpFD) != NULL) {
        strncpy (peerListFileName, tmpBuf, 1000);
        PRINT_DEBUG (DEBUG_LEVEL_SUPER,
                "Using RON peer list %s.\n",
                peerListFileName
            );
    } else {
        PRINT_DEBUG (DEBUG_LEVEL_WARNING,
                "Unable to read RON peer list.\n"
            );
        strcpy (peerListFileName, "error");
    }
    int tmpLen = strlen (peerListFileName);
    if (peerListFileName[tmpLen - 1] == '\n') {
        peerListFileName[tmpLen - 1] = '\0';
    }

    routingMetric = RON_ROUTE_BW; //default
    // The second line of the configuration file is the routing mode - bandwidth or latency.
    if (fgets(tmpBuf, 1000, tmpFD) != NULL) {
         if (strstr (tmpBuf, "bandwidth") != NULL) {
            routingMetric = RON_ROUTE_BW;
            PRINT_DEBUG (DEBUG_LEVEL_SUPER,
                        "Using RON routing metric = RON_ROUTE_BW\n"
                    );
         } else if (strstr (tmpBuf, "latency") != NULL) {
            routingMetric = RON_ROUTE_LAT;
            PRINT_DEBUG (DEBUG_LEVEL_SUPER,
                        "Using RON routing metric = RON_ROUTE_LAT\n"
                    );
         } 
    } else {
        PRINT_DEBUG (DEBUG_LEVEL_WARNING,
                    "Unable to read RON routing metric. Using default\n"
                );
    }

    return true;
}

// Register the OCD with the proxy when the code is dynamically loaded
// The following line sets the name of the OCD's main class
#define MY_OCD_CLASS_NAME RONOverlayNetwork
// The following line sets the type name of the OCD
#define MY_OCD_TYPE_STR "RON"
#include "OCDRegister.h"



