/**
 *  \file RONOverlayNetwork.h
 *  \brief This file defines the RONOverlayNetwork object.  The RONOverlayNetwork
 *  class acts as an interface to the MIT RON code.
 */

#ifndef RON_OVERLAY_NETWORK_H
#define RON_OVERLAY_NETWORK_H 1

#include "OCD.h"
#include "IPPacket.h"

#include <vector>
#include <list>
#include <map>

#include "ip_conduit.h"
#include "RONPacket.h"

class RONOCDStateInfo;

using namespace std;


/**
  * This class encapsulates a RON overlay network to the proxy.
  * It is a subclass of OCD.
  */
class RONOverlayNetwork:public OCD {
public:
    
    RONOverlayNetwork(string type, string id, 
                    string desc, string fileName, 
                    string urlSuffixes, bool handleLegacyURLs
                );
    RONOverlayNetwork(OCD *);


/* OCD API - Start */
    virtual void init(Proxy*, OCI*, bool recoverFromCrash);

    virtual void send ( OCDHandle handle, char* data, int dataLen);
    virtual OCDHandle setup (
                            const string endPointName, 
                            int endPointNameType,
                            char *optionBlocks, 
                            int optionLen, 
                            OCIHandle ociHandle,
                            OCI* oci);

    virtual char* toString(char *tmpStr);
    virtual int getStatus();
/* OCID API - End */


private:

    map<OCDHandle, RONOCDStateInfo *> remoteHandleMap;

    /**
      * Hashtable of remote (src) IP address to RON VN mappings 
      * Entries are made to this map only when a packet belonging to a new
      * flow is received over RON.
     */
    map<long, RONOCDStateInfo*> remoteVNMap_remote;
    
    /**
      * Hashtable of remote (src) IP addresses to RON VN mappings.
      * Entries are made to this table only when a new connection
      * is being set up to the remote end point.
      */
    map<long, RONOCDStateInfo*> remoteVNMap_local;
    
    /** 
      * This is a temporary character array needed while printing out stuff to screen.
      */
    char tmpBuf[MAX_PRINT_BUF];


    void handleFastProxyBlock ( RONFastProxyBlock *blk, int blkLen);
    RONOCDStateInfo * lookupRemoteHandle (OCDHandle remoteHandle);
    void addRemoteHandle (OCDHandle handle, RONOCDStateInfo *state);

    /** The IP conduit used to communicate with the code written by Dave Anderson. */
    RON::ip_conduit* conduit;
   
    /** WHat routing metric should we use for routing? RON_ROUTE_BW, RON_ROUTE_LAT etc. */
    int routingMetric;

    bool parseConfigFile (char* peerListFileName);

    
    void completedSetup (RONOCDStateInfo *vn);

    void handleSetupRequestBlock (RONSetupRequestBlock *setupBlk);
    void handleSetupReplyBlock (RONSetupReplyBlock *setupBlk);

    void handleProxyBlock ( RONProxyBlock *blk, int blkLen);

    
    void recoverFromCrash();

    void handleRONPacket(char *ip_hdr, int len);

    void sendSetupReply (RONOCDStateInfo *ocdState, char *optionBlocks, int optionLen);

    RONOCDStateInfo* lookupRemoteVN_remote (long flowId);
    RONOCDStateInfo* lookupRemoteVN_local (struct in_addr srcAddr);
    
    void addRemoteVN_local (RONOCDStateInfo *ronVN);
    void addRemoteVN_remote (RONOCDStateInfo *ronVN, long flowId);

    static void setRONBlock (RONBlock *ronBlk, uint8_t ronMsgType, uint16_t blkLen);


    friend class RONOCDStateInfo;
    friend class RON::forw;
    friend class RON::ip_conduit;

};


#endif
