/**
 *  \file SampleOCD.cpp
 *  \brief The functions associated with the SampleOCD are defined here.
 */
#ifndef _WIN32
#	include <sys/types.h>
#	include <sys/socket.h>
#else
#	include "netwrap.h"
#endif

#include <errno.h>

#include "OCD_include.h"

#include "SampleOCD.h"
#include "SampleOCDContext.h"
#include "SampleOCDStateInfo.h"
#include "SampleOCD_debug.h"

SampleOCD::SampleOCD(
        string type, string id, string desc, 
        string fileName, 
        string urlSuffixes, bool handleLegacyURLs
    ) :OCD(    type, id, desc, fileName, 
                urlSuffixes, handleLegacyURLs
            ) {

    // Constructs the SampleOCDSpecific code.
    // The bulk of SampleOCD specific processing is handled by the context
    // module.  
    this->context = new SampleOCDContext(this, fileName);
   
}

/**
  * Construct a new SampleOCD with fields taken from the
  * specified overlay network.
  */
SampleOCD::SampleOCD(OCD* o): OCD() {


    this->type = o->type;
    this->id = o->id;
    this->description = o->description;
    this->configFileName = o->configFileName;
    this->urlSuffixes = o->urlSuffixes;
    this->handleLegacyURLs = o->handleLegacyURLs;

    this->configPropertyMap = o->configPropertyMap;
    this->myNameList = o->myNameList;
}



/**
  * Initializes the SampleOCD 
  * @param proxy A pointer to the proxy of which this overlay is a part.
  * @param oci  The OCI associated with this OCD.
  * @param recoverFromCrash This parameter is set to true by the proxy if it is 
  * is recovering from a crash. The OCD is expected to reconstruct its
  * state from the journal.  This flag is usually simply passed on to
  * the OCD's JournalManager which is instantiated in the init() function .
  */
void SampleOCD::init(Proxy *prxy, OCI* oci, bool recoverFromCrash) {
 
    this->proxy = prxy;
    this->oci = oci;


    // Initializes the SampleOCD specific processing.
    // A socket is opened for the SampleOCD to listen on.
    this->context->init();
    
    // If you need to do crash recovery, you need to instantiate a JournalManager for the OCD.
    // Since this is an advanced topic and not necessary for basic operation of the OCD, we are not
    // describing it here.  Please refer the I3OverlayNetwork::init() for an example of how to use
    // the JournalManager.
}


/**
  * Gives a string representation (pretty printed) of the SampleOCD.
  * @param tmpStr a character buffer in which the string is stored.
  * Sufficient space should be allocated for tmpStr.
  * @return pointer to tmpStr
  */
char* SampleOCD::toString(char *tmpStr) {

    sprintf( tmpStr,
            "Sample Overlay Network\n"
            "\ttype\t\t:\t%s\n"
            "\tid\t\t:\t%s\n"
            "\tdescription\t:\t%s\n"
            "\tconfig file\t:\t%s\n",
            type.c_str(),
            id.c_str(),
            description.c_str(),
            configFileName.c_str()
           );
    return tmpStr;
}

/**
  * Setup up a tunnel over the SampleOCD to the specified name.
  * If a tunnel already exists, reuse it if it makes sense to do so in the specific overlay.
  * @param ociHandle The oci to be notified through a callback when the tunnel setup succeeds
  * or fails.
  * @parentLayer The layer which is setting up this new layer.
  * @return a handle to the tunnel.
  */
OCDHandle SampleOCD::setup (
                            const string endPointName, 
                            int endPointNameType,
                            char *optionBlocks, 
                            int optionLen, 
                            OCIHandle ociHandle,
                            OCI* oci) {

    if (endPointNameType == OCI_NAME) {
	    // The end point specified in the parameter endPointName is 
        // the name of an end point.  It should be resolved to
        // an OCD specific ID in an OCD specific manner.

        // Check if there already exists a tunnel to the end point
        SampleOCDStateInfo *ocdState = (SampleOCDStateInfo *) lookupEndPoint (endPointName);
        OCDHandle ocdHandle = NULL_HANDLE;
        
        if (ocdState == NULL) {
           // No tunnel exists,  create a new one.
 
            // First get a new unique handle.
	        ocdHandle = getNewOCDHandle();
                    
            // Now create the OCD State for this new tunnel
        	ocdState = new SampleOCDStateInfo (ocdHandle, oci, this, endPointName);
            //add to the list of ocd states present in this OCD.
            addOCDState (ocdState);
		
            SAMPLE_OCD_PRINT_DEBUG2 (
                    DEBUG_LEVEL_SUPER, 
                    "In SampleOCD setup, creating a new tunnel handle %d, endpointName  = %s\n", 
                    ocdHandle,
                    ocdState->endPointName
                );

            // Add the specified ociHandle to the list of OCIStates waiting for this OCD setup to complete
            ocdState->addParentHandle (ociHandle);

	        // Add the OCDState of the newly created tunnel to the handle-->state hash table
            putStateInfo (ocdHandle, ocdState);

            // add newly created state to hash table keyed by end point name.
            putEndPoint  (endPointName, ocdState);

	        // set the name of the end point of the tunnel.
            strcpy (ocdState->endPointName, endPointName.c_str());

       	    // NOTE: optionBlocks are currently not used.
            ocdState->setupOptionBlocks = optionBlocks;
            ocdState->setupOptionLen = optionLen;

  
	        // Do the OCD specific tunnel to the ocdState's remoteEndpointName
            context->setupTunnel (ocdState);
	
            // After the tunnel setup is complete, the ocdState->notifySetupSuccess()
            // function is called.  Please note that ocdState->notifySetupSuccess()
            // should not be called in this function immediately.  
            // It should be called only after this function returns and appropriately
            // sets the ocd handle.  

        } else {
    
            // Tunnel to end point already exists.  Reuse it if it makes sense to do
  	        // so in the current overlay.

            if (ocdState->status == OCD_STATE_STATUS_READY) {
                
                // Add the specified ociHandle to the list of OCIStates waiting for this OCD setup to complete
                ocdState->addParentHandle (ociHandle);


                // Do something to refresh the tunnel state.  In the case of SampleOCD, we 
                // don't need to do anything.  In the case of i3, we would have refreshed
                // the private trigger pair at this point.
                ocdState->refreshState();

            } else {
                
                // Tunnel existed, but no longer active
                SAMPLE_OCD_PRINT_DEBUG2 (
                    DEBUG_LEVEL_SUPER, 
                    "In SampleOCD setup, refreshing tunnel handle %d, endpointName  = %s\n", 
                    ocdHandle,
                    ocdState->endPointName
                );

                // Add the specified ociHandle to the list of OCIStates waiting for this OCD setup to complete
                ocdState->addParentHandle (ociHandle);

	            // Do the OCD specific tunnel to the ocdState's remoteEndpointName
                context->setupTunnel (ocdState);
	
                // After the tunnel setup is complete, the ocdState->notifySetupSuccess()
                // function is called.  Please note that ocdState->notifySetupSuccess()
                // should not be called in this function immediately.  
                // It should be called only after this function returns and appropriately
                // sets the ocd handle.  


            }

            ocdHandle = ocdState->myHandle;

        }
        
        // Return the handle of the newly created tunnel or the pre-existing tunnel.
        return ocdHandle;
    
    } else {
    
        // The endPointName specified in the setup call is already an overlay specific id.
        // There is no need for further resolution.
	    // Steps similar to if part.
        // So we are leaving it unimplemented here.

        return NULL_HANDLE;
    }
}
    
/**
  * This function is called when some other layer wants to send anything via SampleOCD.
  */
void SampleOCD::send ( OCDHandle handle, char *data, int dataLen) {

    SampleOCDStateInfo *ocdState = (SampleOCDStateInfo *) getStateInfo (handle);

    if (ocdState == NULL) {
        SAMPLE_OCD_PRINT_DEBUG1 (
                        DEBUG_LEVEL_MINIMAL,
                        "Trying to send through unknown OCD with handle %d\n",
                        handle
                    );
        return;
    } 
  
    SAMPLE_OCD_PRINT_DEBUG2 (DEBUG_LEVEL_VERY_VERBOSE,
                "Sending data packet of size %d to %s.\n",
                dataLen, ocdState->endPointName
            );

    // Send the dataLen bytes contained in dataLen through the specific OCD.
    context->sendDataPacket (ocdState, data, dataLen);
}


/**
  * Initializes the state necessary for inbound names associated with this 
  * SampleOCD.
  */
void SampleOCD::initMyNames () {

    // Iterate over the incoming names assocated with this OCD 
    list<MyName *>::iterator itr = myNameList.begin();
    while (itr != myNameList.end()) {

        MyName *currName = (*itr);

	    // The OCD specific properties associated with the name
   	    // can be retrieved by invoking getProperty on the MyName object.
        //char *sampleProp = currName->getProperty (SAMPLE_PROPERTY_NAME); 

	    // OCD specific inbound name setup steps.

    	itr++;
    }
}


// Register the OCD with the proxy when the code is dynamically loaded
// The following line sets the name of the OCD's main class
#define MY_OCD_CLASS_NAME SampleOCD
// The following line sets the type name of the OCD
#define MY_OCD_TYPE_STR "sampleOCD"
#include "OCDRegister.h"
